# ------------------------------------------------------------------------------
#  font.tcl
#  This file is part of Unifix BWidget Toolkit
# ------------------------------------------------------------------------------
#  Index of commands:
#     - SelectFont::create
#     - SelectFont::cget
#     - SelectFont::_draw
#     - SelectFont::_destroy
#     - SelectFont::_modstyle
#     - SelectFont::_update
#     - SelectFont::_getattributes
#     - SelectFont::_init
# ------------------------------------------------------------------------------

source init.tcl
source bwidgets/dialog.tcl
source bwidgets/labelframe.tcl
source bwidgets/combobox.tcl
source bwidgets/bitmap.tcl

namespace eval SelectFont {
    Dialog::use
    LabelFrame::use
    ScrolledWindow::use

    Widget::declare SelectFont {
        {-title      String     "Font selection" 0}
        {-parent     String     "" 0}
        {-background TkResource "" 0 frame}
        {-family     String     ""            0 label}
        {-size       String     ""            0 label}
        {-weight     String     ""            0 label}
        {-slant      String     ""            0 label}
        {-position   String     ""            0 label}
        {-command    String     ""            0}
        {-sampletext String     "Sample Text" 0}
        {-bg         Synonym    -background}
    }

    proc ::SelectFont { path args } { return [eval SelectFont::create $path $args] }
    proc use {} {}

    variable _families {arial times arialn lucidatypewriter symbol}
    variable _sizes    {12 14 16 18 20 22 24 26 28 30 32 34 36 40 46 52 60 70 80 100 120}
    variable _styles   {bold italic up down}
    variable _helpText
    set _helpText(bold) "Bold"
    set _helpText(italic) "Italic"
    set _helpText(up) "Up"
    set _helpText(down) "Down"

    variable _attributes
    variable _values

    variable _widget
}


# ------------------------------------------------------------------------------
#  Command SelectFont::create
# ------------------------------------------------------------------------------
proc SelectFont::create { path args } {
    variable _families
    variable _sizes
    variable _styles
    variable _helpText
    variable _attributes

    Widget::init SelectFont "$path#SelectFont" $args
    set bg [Widget::getoption "$path#SelectFont" -background]
  
    frame $path -relief flat -borderwidth 0 -background $bg
    bind $path <Destroy> "SelectFont::_destroy $path"
    set lbf [ComboBox::create $path.font \
		 -highlightthickness 0 -takefocus 0 -background $bg \
		 -values $_families \
		 -textvariable SelectFont::_attributes(family) \
		 -editable 0 \
		 -modifycmd "SelectFont::_update $path"]
    set lbs [ComboBox::create $path.size \
		 -highlightthickness 0 -takefocus 0 -background $bg \
		 -width    4 \
		 -values   $_sizes \
		 -textvariable SelectFont::_attributes(size) \
		 -editable 0 \
		 -modifycmd "SelectFont::_update $path"]
    pack $lbf -side left -anchor w
    pack $lbs -side left -anchor w -padx 4

    foreach st $_styles {
	Button $path.$st \
	    -highlightthickness 0 -takefocus 0 -padx 0 -pady 0 -bd 1 \
	    -background $bg -relief flat\
	    -image  [Bitmap::get ../gifs/$st.gif] \
	    -command "SelectFont::_modstyle $path $st" \
	    -helptext $_helpText($st)
	pack $path.$st -side left -anchor w
	bind $path.$st <Enter> {+
	    if {[string compare [%W cget -relief] "sunken"] == 0} {
		%W configure -bg grey
	    } else {
		%W configure -relief raised
	    }
	}
	bind $path.$st <Leave> {+
	    if {[string compare [%W cget -relief] "sunken"] == 0} {
		%W configure -bg LightGrey
	    } else {
		%W configure -relief flat
	    }
	}
    }
    set data(label) ""
    set data(lbf)   $lbf
    set data(lbs)   $lbs

    _getattributes $path

    rename $path ::$path:cmd
    proc ::$path { cmd args } "return \[eval SelectFont::\$cmd $path \$args\]"

    return $path
    
}

# ------------------------------------------------------------------------------
#  Command SelectFont::configure
# ------------------------------------------------------------------------------
proc SelectFont::configure { path args } {
    variable _styles

    set option [lindex $args 0]
    set value [lindex $args 1]

    set res [Widget::configure "$path#SelectFont" $args]

    switch -- $option {
	-family {set SelectFont::_attributes(family) $value}
	-size {set SelectFont::_attributes(size) $value}
	-weight {
	    set SelectFont::_attributes(weight) $value
	    if {[expr {[string compare $value "bold"] == 0}]} {
		$path.bold configure -relief sunken -bg LightGrey
		set SelectFont::_values(bold) 1
	    } else {
		$path.bold configure -relief flat -bg $Config::bgColor
		set SelectFont::_values(bold) 0
	    }
	}
	-slant {
	    set SelectFont::_attributes(slant) $value
	    if {[expr {[string compare $value "italic"] == 0}]} {
		$path.italic configure -relief sunken -bg LightGrey
		set SelectFont::_values(italic) 1
	    } else {
		$path.italic configure -relief flat -bg $Config::bgColor
		set SelectFont::_values(italic) 0
	    }
	}
	-position {
	    set SelectFont::_attributes(position) $value
	    if {[expr {[string compare $value "up"] == 0}]} {
		$path.up configure -relief sunken -bg LightGrey
		$path.down configure -relief flat -bg $Config::bgColor
		set SelectFont::_values(up) 1
		set SelectFont::_values(down) 0
	    } elseif {[expr {[string compare $value "down"] == 0}]} {
		$path.up configure -relief flat -bg $Config::bgColor
		$path.down configure -relief sunken -bg LightGrey
		set SelectFont::_values(up) 0
		set SelectFont::_values(down) 1
	    } else {
		$path.up configure -relief flat -bg $Config::bgColor
		$path.down configure -relief flat -bg $Config::bgColor
		set SelectFont::_values(up) 0
		set SelectFont::_values(down) 0
	    }
	}
	default {puts default}
    } 
   return $res

}

# ------------------------------------------------------------------------------
#  Command SelectFont::cget
# ------------------------------------------------------------------------------
proc SelectFont::cget { path option } {
    variable _attributes
    
    switch -- $option {
	-family {return [Widget::cget "$path#SelectFont" $option]}
	-size {return [Widget::cget "$path#SelectFont" $option]}
	-weight {return $_attributes(weight)}
	-slant {return $_attributes(slant)}
	-position {return $_attributes(position)}
	default {puts default}
    }
}

# ------------------------------------------------------------------------------
#  Command SelectFont::_draw
# ------------------------------------------------------------------------------
proc SelectFont::_draw { path } {
    variable _attributes

    $data(lbf) selection clear 0 end
    $data(lbf) selection set $_attributes(family)
    $data(lbf) activate $_attributes(family)
    $data(lbf) see $(_attributesfamily)
    $data(lbs) selection clear 0 end
    $data(lbs) selection set $_attributes(size)
    $data(lbs) activate $_attributes(size)
    $data(lbs) see $_attributes(size)
    _update $path

    if { [Dialog::draw $path] == 0 } {
        set result [Widget::getoption "$path#SelectFont" -family]
    } else {
        set result ""
    }
    unset data
    Widget::destroy "$path#SelectFont"

    return $result
}


# ------------------------------------------------------------------------------
#  Command SelectFont::_destroy
# ------------------------------------------------------------------------------
proc SelectFont::_destroy { path } {

    Widget::destroy "$path#SelectFont"
}


# ------------------------------------------------------------------------------
#  Command SelectFont::_modstyle
# ------------------------------------------------------------------------------
proc SelectFont::_modstyle { path style } {
    variable _values
    variable _attributes

    if { $_values($style) == 1 } {
	switch $style {
	    bold {set _attributes(weight) "normal"}
	    italic {set _attributes(slant) "roman"}
	    up {set _attributes(position) "normal"}
	    down {set _attributes(position) "normal"
}
	}
        $path.$style configure -relief flat -bg $Config::bgColor
        set _values($style) 0
    } else {
	switch $style {
	    bold {set _attributes(weight) "bold"}
	    italic {set _attributes(slant) "italic"}
	    up {
		$path.down configure -relief flat -bg $Config::bgColor
		set _values(down) 0
		set _attributes(position) "up"
	    }
	    down {
		$path.up configure -relief flat -bg $Config::bgColor
		set _values(up) 0
		set _attributes(position) "down"
	    }
	}
	$path.$style configure -relief sunken -bg LightGrey
	set _values($style) 1
    }

    uplevel \#0 [Widget::getoption "$path#SelectFont" -command]
}


# ------------------------------------------------------------------------------
#  Command SelectFont::_update
# ------------------------------------------------------------------------------
proc SelectFont::_update { path } {
    variable _families
    variable _sizes
    variable _attributes


    Widget::setoption "$path#SelectFont" -family $_attributes(family) 
    Widget::setoption "$path#SelectFont" -size $_attributes(size)  
    
    uplevel \#0 [Widget::getoption "$path#SelectFont" -command]
}


# ------------------------------------------------------------------------------
#  Command SelectFont::_getattributes
# ------------------------------------------------------------------------------
proc SelectFont::_getattributes { path } {
    variable _families
    variable _styles
    variable _sizes
    variable _attributes
    variable _values

    set _attributes(family)   [Widget::getoption "$path#SelectFont" -family]
    set _attributes(size)     [Widget::getoption "$path#SelectFont" -size]
    set _attributes(weight)   [Widget::getoption "$path#SelectFont" -weight]
    set _attributes(slant)    [Widget::getoption "$path#SelectFont" -slant]
    set _attributes(position) [Widget::getoption "$path#SelectFont" -position]

    set _values(bold)       [expr {[string compare $_attributes(weight) "bold"] == 0}]
    set _values(italic)     [expr {[string compare $_attributes(slant)  "italic"]  == 0}]
    set _values(up)         [expr {[string compare $_attributes(position)  "up"]  == 0}]
    set _values(down)       [expr {[string compare $_attributes(position)  "down"]  == 0}]

    foreach st $_styles {
	$path.$st configure -relief [expr {$_values($st) ? "sunken":"flat"}]
    }
}


