/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************/

#include "UCE_Interface.h"


void UCE::failToRead() {
	std::cerr << "UCE::UCE_Interface::failToRead - error reading from server" << std::endl;
  #ifdef WIN32
	std::cerr << WSAGetLastError() << std::endl;
  #endif
}


//
// class ReceiveBlock
//
char UCE::ReceiveBlock::recvbuf[1000];

UCE::ReceiveBlock::ReceiveBlock(int sock) : recvError(false), datalength(0) {
  // read message-header
  if (readn(sock, &recvbuf, 11) < 11) { 
    failToRead();
    recvError = true;
    return;
  }
  
  uint8*  p8  = (uint8*)recvbuf;
  uint16* p16 = (uint16*)(recvbuf + 5);
  uint32* p32 = (uint32*)(recvbuf + 1);

  sID  = p8[0];
  mID  = ntohl(p32[0]);
  code = ntohs(p16[0]);

  p32 = (uint32*)(recvbuf + 7);
  datalength = ntohl(p32[0]);
  
  // read message
  if (datalength > 0) {
    data = new char[datalength];
    if (readn(sock, data, datalength) < datalength) { 
      failToRead();
      recvError = true;
      return;
    }
  }
}




//
// class UCE::UCE_Interface
//
UCE::UCE_Interface* UCE::UCE_Interface::ucei = NULL;


UCE::UCE_Interface::~UCE_Interface() {
  unlinksocket(sock); 
  close(sock); 
}


void UCE::UceHandleIncomingData(ClientData clientData, int mask) {
  UCE_Interface::ucei->handleIncomingData(clientData, mask);
}


void UCE::UCE_Interface::handleIncomingData(ClientData clientData, int mask) {
  int sock = (int)clientData;

  #ifdef WIN32
    clearReadSocketMessageQueue(sock);
  #endif
  
  ReceiveBlock recv(sock);
  if (recv.isReceiveError()) {
    #ifdef WIN32
      restoreReadSocketMessageQueue(sock);
    #endif
    return;
  }

  if (recv.isError()) {
    if (recv.isFatal()) {
			std::cerr << "UCE::UCE_Interface::handleIncomingData - receiving fatal error: " << recv.getErrorMsg().c_str() << "\n  -> connection closed" << std::endl;
      close(sock);
      unlinksocket(sock);
    } else {
      std::cerr << "UCE::UCE_Interface::handleIncomingData - receiving error: " << recv.getErrorMsg().c_str() << std::endl;
    }
    #ifdef WIN32
      restoreReadSocketMessageQueue(sock);
    #endif
    return;
  }

  uint16 code = recv.getCode();
  std::cout << code << std::endl;
  
  switch (code) {
  case 007: { // synch info
    if (recv.isEmpty()) break;

    // zweiten header aus message auslesen (wegen UCE-Interface notwendig)
    char*   data = recv.getData();
    uint8*  p8   = (uint8*)data;
    uint32* p32  = (uint32*)(data + 1);

    uint8  sID = p8[0];
    uint32 mID = ntohl(p32[0]);
    code = p8[5];

    p32 = (uint32*)(data + 6);
    uint32 datalen = ntohl(p32[0]);
 
    // synchinfo fuer whiteboard und relevant?
    if ((sID != wbID) || (code != 3)) break;  
    
    // whiteboard - cmd auslesen
    uint8 wbcmd = p8[10];   
    switch (wbcmd) {
    case 2: { process_SET_ACTIVEPAGE(data + 11); break; }
    case 3: { process_SET_TELEPOINTER(data + 11); break; }
    }

    break;
  }
  case 104: { // connection close
    close(sock);
    unlinksocket(sock);
    std::cout << "UCE::UCE_Interface::handleIncomingData - connection closed";
    break;
  }
  case 248: { // status information whiteboard
    uint32* p32 = (uint32*)(recv.getData() + 1);
    uint32 num = p32[0];
    for (int i = 0; i < num; i++) {
      uint32 id = p32[i];
      // -> id speichern == diese Folien wurden bereits vom mlb hochgeladen
    }
    break;
  }
  }

  #ifdef WIN32
    restoreReadSocketMessageQueue(sock);
  #endif
}


void UCE::UCE_Interface::process_SET_ACTIVEPAGE(char* data) {
  uint32* p32 = (uint32*)data;
  uint32  id  = ntohl(p32[0]);
  mlb->setActivePage(id);
}


void UCE::UCE_Interface::process_SET_TELEPOINTER(char* data) {
  uint8*  p8  = (uint8*)data;
  uint32* p32 = (uint32*)data;

  uint32 x      = ntohl(p32[0]);
  uint32 y      = ntohl(p32[1]);
  uint32 width  = ntohl(p32[2]);
  uint32 height = ntohl(p32[3]);
  bool pressed  = (bool)p8[16];
  float xRel = (float)x / width;
  float yRel = (float)y / height;
  if (xRel < 0) xRel = 0; if (xRel > 1) xRel = 1;
  if (yRel < 0) yRel = 0; if (yRel > 1) yRel = 1;

  mlb->setTelePointer(xRel, yRel, pressed);
}


int UCE::UCE_Interface::send(char* data, uint32 size) {
  if (writen(sock, data, size) < size) {
    std::cerr << "UCE::UCE_Interface::send - error writing to server" << std::endl;
    #ifdef WIN32
      std::cerr << WSAGetLastError() << std::endl;
    #endif
    return -1;
  }
  return size;
}


int UCE::UCE_Interface::send(uint8 sID, uint8 cmd, char* data, uint32 size) {
  char* buf = new char[100+size];
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)(buf+1);

  static uint32 mIDctr = 0;
  p8[0]  = sID;
  p32[0] = htonl(++mIDctr);
  p8[5]  = cmd;

  p32 = (uint32*)(buf+6);
  p32[0] = htonl(size);
  memcpy(buf + 10, data, size);
  
  int i = send(buf, size + 10);
  delete[] buf;
  return i;
}


int UCE::UCE_Interface::cmd_open_connection(){
  // init socket
  #ifdef WIN32
    WSADATA WSAdata;
    if (WSAStartup(MAKEWORD(2, 0), &WSAdata)) {
      std::cerr << "UCE::UCE_Interface::UCE_Interface - Windows Sockets init failed" << std::endl;
      std::cerr << WSAGetLastError() << std::endl;
      return 1;
    }
  SetMessageQueue(64);
  #endif
  
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
    std::cerr << "UCE::UCE_Interface::UCE_Interface - error intializing socket" << std::endl;
    return 1;
  }

  // create address
  sockaddr_in a;
  memset((char*)&a, 0, sizeof(a));	
  a.sin_family = AF_INET;
  a.sin_port = htons(port);
  a.sin_addr.s_addr = inet_addr(address.c_str());

  if (a.sin_addr.s_addr == INADDR_NONE) {
    std::cerr << "UCE::UCE_Interface:UCE_Interface - given addresss is not a valid IP-Address" << std::endl;
    return 1;
  }

   // connect to server
  if (connect(sock, (sockaddr*)&a, sizeof(a)) < 0) {
    std::cerr << "UCE::UCE_Interface::UCE_Interface - could not connect to server" << std::endl;
    #ifdef WIN32
      std::cerr << WSAGetLastError() << std::endl;
    #endif
    return 1;
  }
  
  // sending UCE-Header
  char buf[100];
  uint8*  p8; 
  uint16* p16; 
  uint32* p32; 
  
  std::string prot = "UCEP";
  p16 = (uint16*)buf;
  p16[0] = htons(prot.length());
  memcpy(buf + 2, prot.c_str(), prot.length());
  
  p8 = (uint8*)(buf + prot.length() + 2);
  p8[0] = 2;
  p8[1] = 0;
  p8[2] = 0;
  
  p16 = (uint16*)(buf + prot.length() + 5);
  p16[0] = htons(username.length());
  memcpy(buf + prot.length() + 7, username.c_str(), username.length());
    
  send(buf, prot.length() + username.length() + 7);
  

  // receiving service list
  char recvbuf[1000]; 
  if (readn(sock, &recvbuf, 2) < 2) { failToRead(); return 1; }
  p16 = (uint16*)recvbuf;
  uint16 length = ntohs(p16[0]);
  if (readn(sock, &recvbuf, length) < length) { failToRead(); return 1; }
  if (std::string(recvbuf, length).compare("WLC") != 0) {
    std::cerr << "UCE::UCE_Interface::UCE_Interface - handshake failed" << std::endl;
    return 1;
  }
  
  bool wbservice = false;
  int8 num = -1;
  do {
    if (readn(sock, &recvbuf, 1) < 1) { failToRead(); return 1; }
    int8* p8 = (int8*)recvbuf;   
    num = p8[0];
    if (num >= 0) {
      if (readn(sock, &recvbuf, 2) < 2) { failToRead(); return 1; }
      p16 = (uint16*)recvbuf;
      uint16 length = ntohs(p16[0]);
      if (readn(sock, &recvbuf, length) < length) { failToRead(); return 1; }
      if (std::string(recvbuf, length).compare("Whiteboard") == 0) { wbservice = true; wbID = num; }
    }
  } while(num >= 0);

  if (!wbservice) {
    std::cerr << "UCE::UCE_Interface::UCE_Interface - Whiteboard service not provided by server" << std::endl;
    return 1;
  }

  // register 
  p16 = (uint16*)buf;
  p16[0] = htons(pass.length());
  memcpy(buf + 2, pass.c_str(), pass.length());
  send(wbID, 1, buf, pass.length() + 2);
  
  ReceiveBlock recv(sock); 
  if (recv.isReceiveError() || recv.isError()) {
    std::cerr << "UCE::UCE_Interface::UCE_Interface - could not register to Whiteboard" << std::endl;
    return 1;
  }
  
  // access general and whiteboard-service
  for (int i = 0; i < 2; i++) {
    p8 = (uint8*)buf;
    p8[0] = 3;
    send((i == 0) ? 0 : wbID, 5, buf, 1); 
    ReceiveBlock recv(sock); 
    if (recv.isReceiveError() || recv.isError() || (recv.getCode() != 108)) {
      std::cerr << "UCE::UCE_Interface::UCE_Interface - wrong access permission"  << std::endl;
      return 1;
    }  
    uint8* p8 = (uint8*)recv.getData();
    uint8 access = p8[0];
    if (access != 7) {
      std::cerr << "UCE::UCE_Interface::UCE_Interface - wrong access permission"  << std::endl;
      return 1;
    }
  }

  linksocket(sock, TCL_READABLE, UceHandleIncomingData);
  
  // get status information (whiteboard)
  p8 = (uint8*)buf;
  p8[0] = 2;
  send(wbID, 5, buf, 1); 
 
  // acceptSynchMsg
  p8[0]  = 2;
  p32 = (uint32*)(buf + 1);
  p32[0] = htonl(1);
  send(0, 4, buf, 5);

  // get synch -> no resynch
  p8[0] = 3;
  send(0, 6, buf, 1);
  
  p8[0] = 4;
  send(wbID, 5, buf, 1);

  return 0;
}
 


int UCE::UCE_Interface::cmd_load(uint32 id, std::string name, uint32 order, uint32 size, char* page) {
  // id bereits hochgeladen ?

  char*   buf = new char[100+size];
  uint8*  p8  = (uint8*)buf; 
  uint16* p16 = (uint16*)(buf + 5); 
  uint32* p32 = (uint32*)(buf + 1); 
  
  p8[0]  = 4;
  p32[0] = htonl(id);

  p16[0] = htons(name.length());
  memcpy(buf + 7, name.c_str(), name.length());
  
  p32 = (uint32*)(buf + name.length() + 7);
  p32[0] = htonl(order);
  p32[1] = htonl(size);

  memcpy(buf + name.length() + 15, page, size);
  int s = send(wbID, 3, buf, name.length() + 15 + size);
  delete[] buf;
  return s;
}


int UCE::UCE_Interface::cmd_delete(uint32 id) {
  char buf[100];
  uint8*  p8  = (uint8*)buf; 
  uint32* p32 = (uint32*)(buf + 1); 
  p8[0]  = 5;
  p32[0] = htonl(id);
  return send(wbID, 3, buf, 5);
}


int UCE::UCE_Interface::cmd_rename(uint32 id, std::string name) {
  char buf[100];
  uint8*  p8  = (uint8*)buf; 
  uint16* p16 = (uint16*)(buf + 5); 
  uint32* p32 = (uint32*)(buf + 1); 
  p8[0]  = 6;
  p32[0] = htonl(id);
  p16[0] = htons(name.length());
  memcpy(buf + 7, name.c_str(), name.length());
  return send(wbID, 3, buf, name.length() + 7);
}


int UCE::UCE_Interface::cmd_set_active(uint32 id) {
  char buf[100];
  uint8*  p8  = (uint8*)buf; 
  uint32* p32 = (uint32*)(buf + 1); 
  p8[0]  = 2;
  p32[0] = htonl(id);
  return send(wbID, 3, buf, 5);
}





