/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : chat_manager.cpp
 authors : Juergen Vogel
 date    : 
 desc    : 

***********************************************************************************************************/


#ifdef WIN32
#include <winsock.h>
#include <util/win32.h>
#else
#include <arpa/inet.h>
#endif

#include <math.h>

#include "SMP.h"
#include "general.h"

#include <rtpi/rtpiException.h>
 
#include "mlb.h"
#include "chat_manager.h"
#include "mlb_communication.h"
#include "mlb_participants.h"
#include "mlb_gui.h"

#define DEBUG
#define DEBUG_RTPI
//#define DEBUG_RTCPI


CHAT::Message::Message(string _nickname, string _message) : nickname(_nickname), message(_message) {
}


void CHAT_handleReceivedRtpi(SMP::Indication ind) {
  if (CHAT::ChatManager::cm) CHAT::ChatManager::cm->HandleReceivedRtpi(ind);
}


void CHAT_handleReceivedRtcpi(SMP::Indication ind) {
  if (CHAT::ChatManager::cm) CHAT::ChatManager::cm->HandleReceivedRtcpi(ind);
}


CHAT::ChatManager* CHAT::ChatManager::cm = NULL;
char* CHAT::ChatManager::sendBuffer = new char[10000];
const unsigned int CHAT::ChatManager::T = 500;

CHAT::ChatManager::ChatManager() : openConnection(false), init(true) {
  chatGui = new GUI::Chat();
}


CHAT::ChatManager::~ChatManager() {
  for (std::list<Message*>::iterator i = history.begin(); i != history.end(); ++i)
    delete *i;

  delete chatGui;
  CloseRTPI();
}


bool CHAT::ChatManager::OpenRTPI() {
  if (openConnection || !MLB::CommunicationManager::cm->openSMP || !MLB::CommunicationManager::cm->openIDServer) {
    GUI::MessageWindow::Error("ChatManager::OpenRTPI - RTP/I Connection cannot be set up\n");
    return false;
  }

  SMP::SMP* smp = MLB::CommunicationManager::cm->smp;

  // setup SMP
  if ((rtpiSocket = smp->Open(::CHAT_handleReceivedRtpi)) < 0) {
    GUI::MessageWindow::Error("ChatManager::Open - could not connect to smp\n");
    return false;
  }

  if ((rtcpiSocket = smp->Open(::CHAT_handleReceivedRtcpi)) < 0) {
    GUI::MessageWindow::Error("ChatManager::Open - could not connect to smp\n");
    return false;
  }

  // setup RTP/I
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::JoinGroup - joining %q - %u - %u\n", MLB::CommunicationManager::cm->ip.c_str(), 
		 MLB::CommunicationManager::cm->port+7, MLB::CommunicationManager::cm->ttl);
  #endif

  if (!smp->JoinGroup(rtpiSocket, MLB::CommunicationManager::cm->ip.c_str(), MLB::CommunicationManager::cm->port+7, 
		      MLB::CommunicationManager::cm->ttl, MLB::CommunicationManager::cm->bandwidth/20, 'A', false, false, false)) {
    GUI::MessageWindow::Error("ChatManager::Open - smp JoinGroup failed\n");
    smp->Close(rtpiSocket);
    smp->Close(rtcpiSocket);
    return false;
  }
  if (!smp->JoinGroup(rtcpiSocket, MLB::CommunicationManager::cm->ip.c_str(), MLB::CommunicationManager::cm->port+8, 
		      MLB::CommunicationManager::cm->ttl, MLB::CommunicationManager::cm->bandwidth/20, 'A', false, false, false)) {
    GUI::MessageWindow::Error("ChatManager::Open - smp JoinGroup failed\n");
    smp->Close(rtpiSocket);
    smp->Close(rtcpiSocket);
    return false;
  }

  rtpi = new RTPI::Rtpi();
  rtpi->join_group( MLB::CommunicationManager::cm->username, this, this, new RTPI::TransparentReliableTransport());

  // register subcomponent (there is only one!)
  rtpi->add_subcomponent(0, "participant list", true);

  openConnection = true;

  // Initialize -> light-weight late join
  SendStateQuery(3);
  MLB::TimerManager::tm->SetTimer(this, new MLB::TimerObject(2000, MLB::TimerObject::INIT));
  
  return true;
}


bool CHAT::ChatManager::CloseRTPI() {
  if (!openConnection) {
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::CloseRTPI - session already closed\n");
    #endif
    return false;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::LeaveGroup - leaving group\n");
  #endif
  
  rtpi->leave_group();

  MLB::CommunicationManager::cm->smp->LeaveGroup(rtpiSocket);
  MLB::CommunicationManager::cm->smp->LeaveGroup(rtcpiSocket);
  MLB::CommunicationManager::cm->smp->Close(rtpiSocket);
  MLB::CommunicationManager::cm->smp->Close(rtcpiSocket);

  delete rtpi;
  
  openConnection = false;
  init = true;

  for (std::list<Message*>::iterator i = history.begin(); i != history.end(); ++i)
    delete *i;
  history.clear();

  return true;
}


void CHAT::ChatManager::SendMessage(char* nickname, char* message) {
  if (!openConnection) return;
  
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::SendMessage - send %q %q\n", nickname, message);
  #endif

  RememberMessage(nickname, message);

  int bytes = FlushAddMessage(sendBuffer, nickname, message);
  SendEvent(sendBuffer, bytes);
}


void CHAT::ChatManager::RememberMessage(char* nickname, char* message) {
  if (history.size() == HISTORY_SIZE) {
    std::list<Message*>::iterator del = history.begin();
    delete *del;
    history.pop_front();
  }

  history.push_back(new Message(nickname, message));
}


void CHAT::ChatManager::DisplayHistory() {

  chatGui->ClearMessages();

  for (std::list<Message*>::iterator m = history.begin(); m != history.end(); ++m) {
    chatGui->ShowMessages("\"%q\" \"%q\"", (*m)->nickname.c_str(), (*m)->message.c_str());
  }
}


void CHAT::ChatManager::HandleReceivedRtpi(SMP::Indication ind) {
  if (!openConnection) return;

  switch(ind.ServiceType) {
  case SMP::Indication::DATAIND: {	
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got data %u\n", ind.buf->getSize());
    #endif

    RefCount<RTPI::RtpiAduVector> rtpiAdus;
    try {
      rtpiAdus = rtpi->create_rtpi_adus(ind.buf);
    } catch (RTPI::RtpiException e) {
      GUI::MessageWindow::Error("ChatManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    if (!rtpiAdus) 
      return; // adu not complete, received only fragments

    for (RTPI::RtpiAduVector::iterator i = rtpiAdus->begin(); i != rtpiAdus->end(); ++i) {
      if ((*i)->GetAduHeader()->GetPayloadType() != APPLICATION_TYPE) {
	GUI::MessageWindow::Error("ChatManager::HandleReceivedRtpi - wrong payload type %u\n", 
		       (*i)->GetAduHeader()->GetPayloadType());
	continue;
      }

      switch ((*i)->GetType()) {
      case RTPI::RtpiHeader::PKT_EVENT: {
	char* buf   = (*i)->GetPayload();
	uint8*  p8  = (uint8*)buf;
	uint16* p16 = (uint16*)buf;
	uint32* p32 = (uint32*)buf;

	if (((p8[0] & CHAT_MASK_V) >> 6) != VERSION) {
	  GUI::MessageWindow::Error("ChatManager::ReceiveEvent - wrong version of chat protocol\n");
	  continue;
	}

	int type = p8[0] & CHAT_MASK_TYPE;
	switch (type) {
	case ADD_MESSAGE: {
	  int nlength = ntohs(p16[2]);
	  int padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
	  char* nickname = new char[nlength + 1];
	  memcpy(nickname, buf + 8, nlength);
	  nickname[nlength] = '\0';

	  
	  int mlength = ntohs(p16[3]);
	  char* message = new char[mlength + 1];
	  memcpy(message, buf + 8 + nlength + padd, mlength);
	  message[mlength] = '\0';
	  
	  RememberMessage(nickname, message);

	  GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got nickname %q message %q\n", nickname, message);

	  chatGui->ShowMessages("\"%q\" \"%q\"", nickname, message);
	  delete[] nickname;
	  delete[] message;
	  break;
	}
	}
	break;
      }
      case RTPI::RtpiHeader::PKT_STATE: {
	char* buf   = (*i)->GetPayload();
	uint8*  p8  = (uint8*)buf;
	uint16* p16 = (uint16*)buf;
	uint32* p32 = (uint32*)buf;

	if (((p8[0] & CHAT_MASK_V) >> 6) != VERSION) {
	  GUI::MessageWindow::Error("ChatManager::ReceiveState - wrong version of chat protocol\n");
	  continue;
	}

	if (init) {
	  // we are late joiner -> init our history
	  int length = ntohs(p16[1]);
	  int nlength;
	  int mlength;
	  char* nickname;
	  char* message;
	  int offset = 4;
	  int padd;
	  for (int i = 1; i <= length; ++i) {
	    nlength = ntohs(p16[offset/2]);
	    mlength = ntohs(p16[offset/2 + 1]);
	    offset += 4;

	    nickname = new char[nlength + 1];
	    nickname[nlength] = '\0';
	    memcpy(nickname, buf + offset, nlength);
	    padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
	    offset += (nlength + padd);

	    message  = new char[mlength + 1];
	    message[mlength] = '\0';
	    memcpy(message, buf + offset, mlength);
	    padd = (mlength % 4 == 0) ? 0 : 4 - (mlength % 4);
	    offset += (mlength + padd);

	    RememberMessage(nickname, message);
	    delete[] nickname;
	    delete[] message;
	  }

	  init = false;
	  DisplayHistory();
	  
	  break;
	}

	if (MLB::TimerManager::tm->TimerRunning(this, MLB::TimerObject::STATE_REPLY)) {
	  // we have an own state reply timer -> delete 
	  MLB::TimerManager::tm->DeleteTimer(this, MLB::TimerObject::STATE_REPLY);
	} 

	// we are no late joiner -> do nothing
	break;
      }

      case RTPI::RtpiHeader::PKT_STATE_QUERY: {
	#ifdef DEBUG
	GUI::MessageWindow::Info("ChatManager::ReceiveStateQuery - got State Query\n");
	#endif

	if (init) continue; // we don't know yet if we are late joiner or the session is empty

	double x = drand48();
	int n    = MLB::ParticipantManager::pm->GetNumberOfParticipants();
	double N = (n == 1) ? 1 : n - 1;

	#ifdef DEBUG
	GUI::MessageWindow::Info("ChatManager::ReceiveStateQuery - x: %f, N: %f\n", x, N);
	#endif

	if ((n == 1) || (N == 1) || (x < (1.0 / N))) {
	  // send State immediately
	  int bytes = FlushState(sendBuffer);
	  SendState(((RTPI::RtpiStateQuery*)&(**i))->GetAduHeader()->GetPriority(), sendBuffer, bytes);

  	  #ifdef DEBUG
	  GUI::MessageWindow::Info("ChatManager::ReceiveStateQuery - answer immediately\n");
	  #endif
	} else {
	  // set state reply timer
	  int runningTime = (int)(T * (1 + log(x) / log(N)));
	  MLB::TimerManager::tm->SetTimer(this, new MLB::TimerObject(runningTime , MLB::TimerObject::STATE_REPLY, 
								     (*i)->GetSubcomponentID(), 
								     ((RTPI::RtpiStateQuery*)&(**i))->GetAduHeader()->GetPriority()));

  	  #ifdef DEBUG
	  GUI::MessageWindow::Info("ChatManager::ReceiveStateQuery - answer later, set timer with running time %u\n", runningTime);
	  #endif
	}
	break;
      }
      default:
	GUI::MessageWindow::Error("ChatManager::HandleReceivedRtpi - wrong adu type %u\n", 
		       (*i)->GetType());
      }
    }

    try {
      rtpi->handle_received_rtpi(rtpiAdus, ind.senderIPAddr);
    } catch (RTPI::RtpiException e) {
      GUI::MessageWindow::Error("ChatManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    return;
  }

  
  case SMP::Indication::UNRLDATAIND: {	
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got unreliable data %u\n", ind.buf->getSize());
    #endif
    return;
  }
  

  case SMP::Indication::LJNEEDED: 
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got LJNEEDED\n");
    #endif
    return;
														 

  case SMP::Indication::DADUIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got DADUIND\n");
    #endif
    return;

  case SMP::Indication::DADUFAIL:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got DADUFAIL\n");
    #endif
    return;

  case SMP::Indication::ERR:
    GUI::MessageWindow::Error("ChatManager::HandleReceivedRtpi - got ERR / probably lost connection to smp\n");
    //unlinksocket(smp_socket);
    return;


  case SMP::Indication::NOSERVICE:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got NOSERVICE\n");
    #endif
    return;

  case SMP::Indication::JOININD:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got JOININD\n");
    #endif
    return;

  case SMP::Indication::LEAVIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got LEAVIND\n");
    #endif
    return;

  case SMP::Indication::ACKIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got ACKIND\n");
    #endif
    return;

  case SMP::Indication::DADUCONF:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got DADUCONF\n");
    #endif
    return;

  case SMP::Indication::CLOSEIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got CLOSEIND\n");
    #endif
    return;

  case SMP::Indication::ALIVE:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtpi - got ALIVE\n");
    #endif
    return;

  default:
    return;
  }
}


void CHAT::ChatManager::HandleReceivedRtcpi(SMP::Indication ind) {
  if (!openConnection) return;

  switch(ind.ServiceType) {
  case SMP::Indication::DATAIND: {	
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got data %u\n", ind.buf->getSize());
    #endif
    return;
  }

  case SMP::Indication::UNRLDATAIND: {	
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got unreliable data %u\n", ind.buf->getSize());
    #endif
    try {
      rtpi->handle_received_rtcpi(ind.buf, ind.senderIPAddr);
    } catch (RTPI::RtpiException e) {
      GUI::MessageWindow::Error("ChatManager::HandleReceivedRtcpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }
    return;
  }

  case SMP::Indication::LJNEEDED: 
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got LJNEEDED\n");
    #endif
    return;

  case SMP::Indication::DADUIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got DADUIND\n");
    #endif
    return;

  case SMP::Indication::DADUFAIL:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got DADUFAIL\n");
    #endif
    return;

  case SMP::Indication::ERR:
    GUI::MessageWindow::Error("ChatManager::HandleReceivedRtcpi - got ERR / probably lost connection to smp\n");
    //unlinksocket(smp_socket);
    return;

  case SMP::Indication::NOSERVICE:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got NOSERVICE\n");
    #endif
    return;

  case SMP::Indication::JOININD:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got JOININD\n");
    #endif
    return;

  case SMP::Indication::LEAVIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got LEAVIND\n");
    #endif
    return;

  case SMP::Indication::ACKIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got ACKIND\n");
    #endif
    return;

  case SMP::Indication::DADUCONF:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got DADUCONF\n");
    #endif
    return;

  case SMP::Indication::CLOSEIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got CLOSEIND\n");
    #endif
    return;

  case SMP::Indication::ALIVE:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("ChatManager::HandleReceivedRtcpi - got ALIVE\n");
    #endif
    return;

  default:
    return;
  }
}


void CHAT::ChatManager::SendState(int priority, char* payload, int size) {
  if (!openConnection) return;

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::SendState - sending state\n");
  #endif

  RefCount<RTPI::RtpiAdu> state = new RTPI::RtpiState(SUB_ID, APPLICATION_TYPE, 0, rtpitime(), priority, payload, size);
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;
  adus->push_back(state);

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adus);
  } catch (RTPI::RtpiException e) {
    GUI::MessageWindow::Error("ChatManager::SendState - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }

  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    MLB::CommunicationManager::cm->smp->Send(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


void CHAT::ChatManager::SendEvent(char* payload, int size) {
  if (!openConnection) return;

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::SendEvent - sending event\n");
  #endif

  RefCount<RTPI::RtpiAdu> event = new RTPI::RtpiEvent(SUB_ID, APPLICATION_TYPE, 0, rtpitime(), payload, size);
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;
  adus->push_back(event);

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adus);
  } catch (RTPI::RtpiException e) {
    GUI::MessageWindow::Error("ChatManager::SendEvent - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }
 
  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    MLB::CommunicationManager::cm->smp->Send(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


void CHAT::ChatManager::SendStateQuery(int priority) {
  if (!openConnection) return;

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("ChatManager::SendStateQuery - sending state query\n");
  #endif

  RefCount<RTPI::RtpiAdu> query = new RTPI::RtpiStateQuery(SUB_ID, APPLICATION_TYPE, rtpitime(), priority);
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;
  adus->push_back(query);

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adus);
  } catch (RTPI::RtpiException e) {
    GUI::MessageWindow::Error("ChatManager::SendStateQuery - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }
 
  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    MLB::CommunicationManager::cm->smp->Send(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


int CHAT::ChatManager::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (ChatManager::VERSION << 6);
  p8[0] = p8[0] | ChatManager::ADD_MESSAGE;

  p16[1] = htons(history.size());

  int nlength;
  int mlength;

  int offset = 4;
  int padd;

  for (std::list<Message*>::iterator i = history.begin(); i != history.end(); ++i) {
    nlength = (*i)->nickname.length();
    mlength = (*i)->message.length();

    p16[offset/2] = htons(nlength);
    p16[offset/2 + 1] = htons(mlength);

    offset += 4;

    memcpy(buf + offset, (*i)->nickname.c_str(), nlength);
  
    padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
    if (padd != 0) { // padding needed
      for (int i = 0; i < padd; ++i)
	p8[offset + nlength + i] = 0;
    }
  
    offset += (nlength + padd);

    memcpy(buf + offset, (*i)->message.c_str(), mlength);
  
    padd = (mlength % 4 == 0) ? 0 : 4 - (mlength % 4);
    if (padd != 0) { // padding needed
      for (int i = 0; i < padd; ++i)
	p8[offset + mlength + i] = 0;

    }

    offset += (mlength + padd);
  }
  
  return offset;
}


int CHAT::ChatManager::FlushAddMessage(char* buf, char* nickname, char* message) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (ChatManager::VERSION << 6);
  p8[0] = p8[0] | ChatManager::ADD_MESSAGE;

  int nlength = strlen(nickname);
  int mlength = strlen(message);

  p16[2] = htons(nlength);
  p16[3] = htons(mlength);

  memcpy(buf + 8, nickname, nlength);
  
  int padd1 = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
  if (padd1 != 0) { // padding needed
    for (int i = 0; i < padd1; ++i)
      p8[8 + nlength + i] = 0;
  }
  
  memcpy(buf + 8 + nlength + padd1, message, mlength);
  
  int padd2 = (mlength % 4 == 0) ? 0 : 4 - (mlength % 4);
  if (padd2 != 0) { // padding needed
    for (int i = 0; i < padd2; ++i)
      p8[8 + nlength + padd1 + mlength + i] = 0;
  }
  
  return 8 + nlength + padd1 + mlength + padd2;
}


void CHAT::ChatManager::send_rtcpi(RefCount<Buffer> buf) {
  if (!openConnection) return;
  MLB::CommunicationManager::cm->smp->SendUnreliable(rtcpiSocket, buf->getCont(), buf->getSize());
}


void CHAT::ChatManager::subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason) {
}


void CHAT::ChatManager::subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason) {
}


void CHAT::ChatManager::subcomponent_removed(uint32 id) {
}


void CHAT::ChatManager::participant_added(uint32 pid) {
}


void CHAT::ChatManager::participant_removed(uint32 pid) {
}


void CHAT::ChatManager::participant_changed(uint32 pid) {
}


void CHAT::ChatManager::ExecuteTimer(MLB::TimerObject* timer) {
  switch (timer->type) {
  case MLB::TimerObject::INIT: {
    init = false;

    chatGui->EnableMessageEntry();

    #ifdef DEBUG
    GUI::MessageWindow::Info("ChatManager::ExecuteTimer - Init Timer ran out\n");
    #endif
    break;
  }
  case MLB::TimerObject::STATE_QUERY: {
    break;
  }
  case MLB::TimerObject::STATE_REPLY: {
    int bytes = FlushState(sendBuffer);
    SendState(timer->priority, sendBuffer, bytes);

    #ifdef DEBUG
    GUI::MessageWindow::Info("ChatManager::ExecuteTimer - State Reply Timer ran out, sending state\n");
    #endif
    break;
  }
  }

  delete timer;
}


