/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : ft_manager.h 
 authors : Juergen Vogel
 date    : 09.05.2001
 desc    : 

 **********************************************************************************************************/


#ifndef __FT_MANAGER_H__
#define __FT_MANAGER_H__

#define FT_MASK_V    0xC0
#define FT_MASK_TYPE 0x3F
#define FT_MASK_VIS  0x01

#include <rtpi/rtpi.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpihandler.h>
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpiInterfaces.h>
#include <rtpi/rtpiTransport.h>
#include <rtpi/rtpiSeqNumber.h>
#include <util/types.h>
#include <util/Buffer.h>

#include <mlb_communication.h>

#include <tcl.h>

#include <vector>
#include <list>
#include <string>

namespace FT {
  class Answer {
  public:
    Answer() : value(0), pid(0), createTime(0) {}
    Answer(uint32 _pid, int _value, uint32 _createTime) : pid(_pid), value(_value), createTime(_createTime) {}

    int    value;      // selected value
    uint32 pid;        // originating participant
    uint32 createTime; // time answer was created in secs

    int FlushCreate(char* buf);
    int FlushDelete(char* buf);
  };

  class Question {
  public:
    Question(uint32 _id) : id (_id) {}
    Question(uint32 _id, uint32 _ttl, std::string text, int nbr, std::string _expl);

    void Update(std::string text, int nbr, std::string _expl);

    bool visible;                           // visibility of question
    uint32 id;                              // subid of question
    uint32 ttl;                             // time to live (received answers)
    std::string text;                       // question text
    std::vector<std::string> explanations;  // meaning of values
    std::list<Answer> answers;              // received answers

    double CalculateResult();

    void DisplayUpdate();
    void DisplayAdd();
    void DisplayResult();

    int FlushState(char* buf);
    int ParseState(char* buf, uint32 ts);

    int FlushDelete(char* buf);
    int FlushChangeVisibility(char* buf);
  };

  class FtManager : public RTPI::RtcpiReportInterface, public RTPI::RtpiFeedbackInterface, public MLB::TimerFeedbackInterface {
  public:
    enum types {
      // general
      VERSION = 0,
      APPLICATION_TYPE = 5, // ft
      
      // operation Types
      CHANGE_VISIBILITY = 0,
      DELETE_QUESTION   = 1,
      CREATE_ANSWER     = 2,
      DELETE_ANSWER     = 3
    };
    
    FtManager();
    ~FtManager();
    
    void Init();

    bool OpenConnection() {return openConnection;}

    bool OpenRTPI();
    bool CloseRTPI();

    // handle incoming data
    void HandleReceivedRtpi(SMP::Indication ind);
    void HandleReceivedRtcpi(SMP::Indication ind);

    // functions called by gui
    void CreateQuestion(int argc, const char** argv);
    void EditQuestion(int argc, const char** argv);
    void DeleteQuestion(uint32 id);
    void ChangeVisibility(uint32 id, bool visible);
    void ChangeAnswer(uint32 id, int value);
    void RebuildScores();

    // outgoing data
    void SendState(uint32 id, int priority, char* payload, int size);
    void SendEvent(uint32 id, char* payload, int size);
    void SendStateQuery(uint32 id, int priority);

    // functions implemented due to  RtcpiReportInterface
    void    send_rtcpi(RefCount<Buffer>);
    u_int32 get_local_address(int i=0)   {return MLB::CommunicationManager::cm->get_local_address(i);}
    int     addr_is_local(u_int32 addr)  {return MLB::CommunicationManager::cm->addr_is_local(addr);}
    u_int32 get_session_bandwith()       {return MLB::CommunicationManager::cm->bandwidth/10;}
    u_int32 get_new_subid()              {return MLB::CommunicationManager::cm->get_new_subid();} 
    u_int32 get_new_pid()                {return MLB::CommunicationManager::cm->pid;}
    string  get_session_name(string ="") {return MLB::CommunicationManager::cm->name;}
    string  get_session_email()          {return MLB::CommunicationManager::cm->email;}
    string  get_session_phone()          {return MLB::CommunicationManager::cm->phone;}
    string  get_session_loc()            {return MLB::CommunicationManager::cm->location;}
    string  get_session_note()           {return ""; }
    string  get_session_tool()           {return "feedback"; }

    // functions implemented due to RtpiFeedbackInterface
    void subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason);
    void subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason);
    void subcomponent_removed(uint32 id);
    void participant_added(uint32 pid);
    void participant_removed(uint32 pid);
    void participant_changed(uint32 pid);

    static FtManager* ftm;

    static const unsigned int T;
    static char* sendBuffer;

  private:
    // helper functions
    void RegisterSubcomponent(uint32 subID);
    void UnregisterSubcomponent(uint32 subID);
    void ChangeSubcomponentActiveness(uint32 subID, bool active);

    // list of questions
    Question* GetQuestion(uint32 id);
    std::list<Question*> questions;
    
    // UDP 
    bool openConnection;

    // SMP 
    int rtpiSocket;
    int rtcpiSocket;

    // RTP/I 
    bool init; // while init, don't answer state queries
    RTPI::Rtpi* rtpi;
    
    // timer management
    void ExecuteTimer(MLB::TimerObject* obj);
  };
}


#endif
