/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************/

#include <errno.h>

#ifdef WIN32
#else
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

#include "general.h"
#include <tcl.h>
#include <iostream> // cout


#include "iohelpers.h"

int readn(int fd, void *vptr, int n) {

  int nleft;
  int nread;
  char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = recv(fd, ptr, nleft, 0)) < 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nread = 0;
      else if (WSAGetLastError() == WSAEINTR) nread = 0;
      else return -1;
#else
      if (errno == EINTR) {
	//cerr << "readn() : interrupted system" << endl;
	nread = 0;
      } else if (errno == EWOULDBLOCK) {
	nread = 0;
      }
      else
	return(-1);
#endif
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int recvn(int fd, void *vptr, int n, int flags) {

  int nleft;
  int nread;
  char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = recv(fd, ptr, nleft, flags)) < 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nread = 0;
      else if (WSAGetLastError() == WSAEINTR) nread = 0;
      else return -1;
#else
      if (errno == EINTR) {
	//cerr << "readn() : interrupted system call" << endl;
	nread = 0;
      } else if (errno == EWOULDBLOCK) {
	nread = 0;
      } else
	return(-1);
#endif
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int writen(int fd, const void *vptr, int n) {

  int nleft;
  int nwritten;
  const char *ptr;

  ptr = (char*)vptr;
  nleft = n;
  while(nleft > 0) {

    if ((nwritten = send(fd, ptr, nleft, 0)) <= 0) {
#ifdef WIN32
      if (WSAGetLastError() == WSAEWOULDBLOCK) nwritten = 0;
      else if (WSAGetLastError() == WSAEINTR)  nwritten = 0;
      else return -1;
#else
      if (errno == EINTR || errno == EWOULDBLOCK) {
	nwritten = 0;
      } else
	return -1;
#endif
    }
    //cout << "writen " << nwritten << endl;
    nleft -= nwritten;
    ptr += nwritten;
  }

  return(n);
}

int _readn(Tcl_Channel fd, void *vptr, int n) {

  int nleft;
  int nread;
  char *ptr;
  int tclerrno;

  ptr = (char*)vptr;
  nleft = n;
  while (nleft > 0) {
    if ((nread = Tcl_Read(fd, ptr, nleft)) < 0) {
      tclerrno = Tcl_GetErrno();

      if (tclerrno == EINTR) {
	//cerr << "readn() : interrupted system" << endl;
	nread = 0;
      } else if (tclerrno == EWOULDBLOCK) {
	nread = 0;
      }
      else
	return(-1);
    } else if (nread == 0)
      break;

    nleft -= nread;
    ptr += nread;
  }
  return (n - nleft);
}

int _writen(Tcl_Channel fd, const void *vptr, int n) {

  int nleft;
  int nwritten;
  char *ptr;
  int tclerrno;

  ptr = (char*)vptr;
  nleft = n;
  while(nleft > 0) {

    
    if ((nwritten = Tcl_Write(fd, ptr, nleft)) <= 0) {
      tclerrno = Tcl_GetErrno();
      if (tclerrno == EINTR || tclerrno == EWOULDBLOCK) {
	nwritten = 0;
      } else
	return -1;
    }
    //cout << "writen " << nwritten << endl;
    nleft -= nwritten;
    ptr += nwritten;
  }

  return(n);
}


int is_multicast(char* saddr) {
  #ifdef DEBUG1
  cout << "tools::is_multicast("<<saddr<<")" << endl;
  #endif

  unsigned int ipaddr = inet_addr(saddr);
  if ((ntohl(ipaddr) & 0xE0000000) == 0xE0000000) {
    #ifdef DEBUG1
    cout << "tools::is_multicast() " << saddr << " is a multicast address." << endl;
    #endif
    return 1;
  } else {
    #ifdef DEBUG1
    cout << "tools::is_multicast() " << saddr << " is NOT a multicast address." << endl;
    #endif
    return 0;
  };
}

#ifndef WIN32
int closesocket(int sock) {
	return ::close(sock);
}
#endif

