/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_adu.cpp
 authors : Juergen Vogel
 date    :
 desc    :

 *********************************************************************************************************/

#include <util/tools.h>

#include "mlb_adu.h"
#include "mlb.h"
#include "mlb_participants.h"
#include "mlb_communication.h"
#include "mlb_session.h"
#include "mlb_folder.h"
#include "mlb_page.h"
#include "mlb_objects.h"
#include "mlb_uce.h"
#include "mlb_undomanager.h"

//#define ENABLE_TW

//#define DEBUG
//#define DEBUG_ADU
//#define DEBUG_TW
//#define DEBUG_CONS
//#define DEBUG_ORDER
//#define DEBUG_RECORDING
//#define DEBUG_ADU_ORDERING

//
// class AduManager
//

MLB::AduManager* MLB::AduManager::am = NULL;


void MLB::AduManager::Reset() {
  TimerManager::tm->DeleteTimer(this, TimerObject::ADU_DUE);
  aq.clear();
}


void MLB::AduManager::HandleState(RefCount<RTPI::RtpiAdu> adu) {
  RTPI::RtpiState* state = (RTPI::RtpiState*)&(*adu);
  uint8*  p8   = (uint8*)(state->GetPayload());
  int  version = (p8[0] & MLB_MASK_V) >> 6;
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];

  if (version != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("AduManager::HandleState - wrong version of mlb protocol %u\n", version);
    return;
  }

  Session* session = NULL;
  bool shared = true;
  if (state->ObjectIDIsSet()) {
    GenericObject* obj = SessionManager::sm->GetObject(state->GetObjectID());
    if (obj == NULL) {
      // state for this object was faster than the state of its parent (i.e. page)
      // this can happen only in shared sessions (e.g., recorded sessions).
      session = SessionManager::sm->GetSharedSession();
    } else {
      session = SessionManager::sm->GetParentSession(SessionManager::sm->GetObject(state->GetObjectID()));
      shared = (session->GetSessionType() == Session::SHARED);
    }
  } else {
    session = SessionManager::sm->GetSharedSession();
  }
  Page* page = NULL;

  if (TimerManager::tm->TimerRunning(CommunicationManager::cm, TimerObject::STATE_REPLY, state->GetSubcomponentID()))
    TimerManager::tm->DeleteTimer(CommunicationManager::cm, TimerObject::STATE_REPLY, state->GetSubcomponentID());

  if (TimerManager::tm->TimerRunning(CommunicationManager::cm, TimerObject::STATE_QUERY, state->GetSubcomponentID()))
    TimerManager::tm->DeleteTimer(CommunicationManager::cm, TimerObject::STATE_QUERY, state->GetSubcomponentID());


  // register participant activity 
  ParticipantManager::pm->UpdateParticipantActiveness(adu->GetParticipantID());


  if (opType == CommunicationManager::ADU_STATE) {
    GenericObject* tmp = SessionManager::sm->GetObject(state->GetSubcomponentID());
    GenericObject* test;

    #ifdef DEBUG_ADU
    if (state->ObjectIDIsSet())
      SessionManager::sm->message->Info("AduManager::HandleState - handle state for obj %u / parent %u / ts %u\n",
					state->GetSubcomponentID(), state->GetObjectID(), state->GetAduHeader()->GetTimestamp());
    else
      SessionManager::sm->message->Info("AduManager::HandleState - handle state for obj %u / ts %u\n", 
					state->GetSubcomponentID(), state->GetAduHeader()->GetTimestamp());
    #endif

    #ifdef DEBUG_RECORDING
    if (state->ObjectIDIsSet())
      SessionManager::sm->message->Info("AduManager::HandleState - handle state for obj %u / parent %u / ts %u\n",
					state->GetSubcomponentID(), state->GetObjectID(), state->GetAduHeader()->GetTimestamp());
    else
      SessionManager::sm->message->Info("AduManager::HandleState - handle state for obj %u / ts %u\n", 
					state->GetSubcomponentID(), state->GetAduHeader()->GetTimestamp());
    #endif

    switch (objType) {
    //
    // Folder
    //
    case CommunicationManager::FOLDER: {
      if (tmp && (tmp->GetType() != GenericObject::FOLDER))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create folder, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      GenericContainer* parent;
      if (state->ObjectIDIsSet()) {
	parent = (GenericContainer*)session->GetObject(state->GetObjectID());
	if (!parent) {
	  SessionManager::sm->message->Error("AduManager::HandleState - cannot create folder, parent object with %u id does not exist\n",
					     state->GetObjectID());
	  return;
	}
      } else {
	parent = session;
      }

      Folder* obj = (Folder*)tmp;
      if (!obj) {
	obj = new Folder(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());

	test = parent->GetObjectWithOrder(obj->GetOrder());
	if (test) // somebody else was faster, sort of consistency problem
	  obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));

	parent->InsertContainer(obj);
	SessionManager::sm->DrawPageTree();
	  if (shared) CommunicationManager::cm->RegisterSubcomponent(obj->GetID(), CommunicationManager::FOLDER);
      } else {
	if (state->GetAduHeader()->GetPriority() == 3) {
	  // consistency control
	  SessionManager::sm->GetParentContainer(obj->GetID())->DeleteObject(obj->GetID());
	  obj->ParseState(state->GetPayload());

	  test = parent->GetObjectWithOrder(obj->GetOrder());
	  if (test) // somebody else was faster, sort of consistency problem
	    obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));

	  parent->InsertContainer(obj);
	  SessionManager::sm->ClearPageTree();
	  SessionManager::sm->DrawPageTree();
	    if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else if (state->GetAduHeader()->GetPriority() == 2) {
	  // late join -> reorder
	  SessionManager::sm->GetParentContainer(obj->GetID())->DeleteObject(obj->GetID());
	  test = parent->GetObjectWithOrder(obj->GetOrder());
	  if (test) // somebody else was faster, sort of consistency problem
	    obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));
	  
	  parent->InsertContainer(obj);
	  SessionManager::sm->ClearPageTree();
	  SessionManager::sm->DrawPageTree();
	} // else ignore
      }
        if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create folder with id %u and parent %u and order %u\n",
					obj->GetID(), parent->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Page
    //
    case CommunicationManager::PAGE: {
      if (tmp && (tmp->GetType() != GenericObject::PAGE))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create page, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      GenericContainer* parent;
      if (state->ObjectIDIsSet()) {
	parent = (GenericContainer*)session->GetObject(state->GetObjectID());
	if (!parent) {
	  SessionManager::sm->message->Error("AduManager::HandleState - cannot create page, parent object with %u id does not exist\n",
					     state->GetObjectID());
	  return;
	}
      } else {
	parent = session;
      }

      bool changed = false;
      bool active;
      Page* obj = (Page*)tmp;
      if (!obj) {
	obj = new Page(state->GetSubcomponentID(), Session::SHARED);
	obj->ParseState(state->GetPayload(), active);

	test = parent->GetObjectWithOrder(obj->GetOrder());
	if (test) {// somebody else was faster, sort of consistency problem
	  obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));
	  changed = true;
	}

	parent->InsertContainer(obj);
	if (shared) CommunicationManager::cm->RegisterSubcomponent(obj->GetID(), CommunicationManager::PAGE);
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  SessionManager::sm->GetParentContainer(obj->GetID())->DeleteObject(obj->GetID());
	  obj->ParseState(state->GetPayload(), active);

	  test = parent->GetObjectWithOrder(obj->GetOrder());
	  if (test) {// somebody else was faster, sort of consistency problem
	    obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));
	    changed = true;
	  }

	  parent->InsertContainer(obj);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());

	  if (obj == SessionManager::sm->GetActivePage()) {
	    // ReDraw
	    obj->Activate(true);
	  }

	  SessionManager::sm->ClearPageTree();
	  SessionManager::sm->DrawPageTree();


	} else if (state->GetAduHeader()->GetPriority() == 2) {
	  // late join -> reorder
	  SessionManager::sm->GetParentContainer(obj->GetID())->DeleteObject(obj->GetID());
	  obj->ParseState(state->GetPayload(), active);
	  
	  test = parent->GetObjectWithOrder(obj->GetOrder());
	  if (test) // somebody else was faster, sort of consistency problem
	    obj->SetOrder(parent->GetNextObjectOrder(test->GetID()));
	  
	  parent->InsertContainer(obj);

	  SessionManager::sm->ClearPageTree();
	  SessionManager::sm->DrawPageTree();
	  if (active) {
	    ActivatePage(obj);
	  }

	  return;
	} else {// else ignore
	  return;
	}
      }

      if (active && (obj != SessionManager::sm->GetActivePage())) {
	ActivatePage(obj);
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), active);
      if (shared) CommunicationManager::cm->rtpi->send_rtcpi_note();

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create page with id %u and parent %u and order %u, changed %u\n",
					obj->GetID(), parent->GetID(), obj->GetOrder(), changed);
      #endif
      return;
    }

    //
    // Group
    //
    case CommunicationManager::GROUP: {
      tmp = session->GetGroup(state->GetSubcomponentID());
      if (tmp && (tmp->GetType() != GenericObject::GROUP))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create group, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page) {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create group, parent object with %u id does not exist\n",
		       state->GetObjectID());
	return;
      }

      Group* obj = (Group*)tmp;
      if (!obj) { // first state
	obj = new Group(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::GROUP, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
	if (shared) CommunicationManager::cm->RegisterSubcomponent(obj->GetID(), CommunicationManager::GROUP);
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create group with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Rectangle
    //
    case CommunicationManager::RECTANGLE: {
      if (tmp && (tmp->GetType() != GenericObject::RECTANGLE))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create rectangle, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      Rectangle* obj = (Rectangle*)tmp;
      if (!obj) { // first state
	obj = new Rectangle(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::RECTANGLE, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) 
	  return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create rectangle with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Oval
    //
    case CommunicationManager::OVAL: {
      if (tmp && (tmp->GetType() != GenericObject::OVAL))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create oval, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      Oval* obj = (Oval*)tmp;
      if (!obj) { // first state
	obj = new Oval(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::OVAL, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create oval with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Line
    //
    case CommunicationManager::LINE: {
      if (tmp && (tmp->GetType() != GenericObject::LINE))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create line, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      Line* obj = (Line*)tmp;
      if (!obj) { // first state
	obj = new Line(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::LINE, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create line with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // PolyLine
    //
    case CommunicationManager::POLYLINE: {
      if (tmp && (tmp->GetType() != GenericObject::POLYLINE) && (tmp->GetType() != GenericObject::FREEHAND))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create polyline, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      PolyLine* obj = (PolyLine*)tmp;
      if (!obj) { // first state
	obj = new PolyLine(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::POLYLINE, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) 
	  return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create polyline with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Polygon
    //
    case CommunicationManager::POLYGON: {
      if (tmp && (tmp->GetType() != GenericObject::POLYGON))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create polygon, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      Polygon* obj = (Polygon*)tmp;
      if (!obj) { // first state
	obj = new Polygon(state->GetSubcomponentID());
	obj->ParseState(state->GetPayload());
	if (HandleNewDrawObject(page, obj, CommunicationManager::POLYGON, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create polygon with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // Text
    //
    case CommunicationManager::TEXT: {
      tmp = SessionManager::sm->GetGroup(state->GetSubcomponentID());
      if (tmp && (tmp->GetType() != GenericObject::GROUP))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create text, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      Group* obj = (Group*)tmp;
      if (!obj) { // first state of text
	obj = new Group(state->GetSubcomponentID());
	obj->ParseStateTG(state->GetPayload(), page);
	obj->UpdateDimension();
	if (HandleNewDrawObject(page, obj, CommunicationManager::TEXT, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
	obj->DrawTextCursor(page->GetCanvas());
      } else { // object already exists
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseStateTG(state->GetPayload(), page);
	  obj->UpdateDimension();
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  obj->DrawTextCursor(page->GetCanvas());
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create text with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }

    //
    // PhotoImage
    //
    case CommunicationManager::IMAGE: {
      if (tmp && (tmp->GetType() != GenericObject::IMAGE))   {
	SessionManager::sm->message->Error("AduManager::HandleState - cannot create image, object %u already exists with different type\n",
		       state->GetSubcomponentID());
	return;
      }

      page = (Page*)session->GetObject(state->GetObjectID());
      if (!page)
	page = CreateDefaultPage(state->GetObjectID());

      PhotoImage* obj = (PhotoImage*)tmp;
      if (!obj) { // first state

	try {
	  obj = new PhotoImage( state->GetSubcomponentID(), state->GetPayload(), page->GetCanvas() );
	} catch ( Magick::Exception& exc ) {
	  GUI::MessageWindow::Error( "AduManager::HandleState - cannot create image: %q", exc.what() );
	  return;
	}

	if (HandleNewDrawObject(page, obj, CommunicationManager::IMAGE, shared, state->GetAduHeader()->GetPriority(), state->GetParticipantID())) return;
      } else {
	// consistency control
	if (state->GetAduHeader()->GetPriority() == 3) {
	  page->DeleteObject(obj);
	  obj->ParseState(state->GetPayload(), page->GetCanvas());
	  page->AppendObject(obj);
	  ActivatePage(page, true);
	  if (shared) CommunicationManager::cm->ResetEventSequenceNumber(state->GetSubcomponentID(), state->GetParticipantID());
	} else {// else ignore
	  return;
	}
      }

      if (shared) CommunicationManager::cm->ChangeSubcomponentActiveness(obj->GetID(), true);

      #ifdef DEBUG_ORDER
      SessionManager::sm->message->Info("AduManager::HandleState - Create image with id %u and parent %u and order %u\n",
					obj->GetID(), page->GetID(), obj->GetOrder());
      #endif
      return;
    }
    }
    return;
  }

  // error
}


bool MLB::AduManager::HandleNewDrawObject(Page* page, DrawObject* obj, bool shared, int type, int priority, uint32 participantID) {
  GenericObject* test = page->GetObjectWithOrder(obj->GetOrder());
  if (test) // somebody else was faster, sort of consistency problem
    obj->SetOrder(page->GetNextObjectOrder(test->GetID()));
  
  bool own = (participantID == CommunicationManager::cm->GetParticipantID());
  if (own && UndoManager::um->GetAnnotation(obj->GetID())) obj->SetAnnotation(true);

  if ((page != SessionManager::sm->GetActivePage()) && (priority == 3)) 
    // activate page necessary for replay of recorded objects, since it is likely that we reveive
    // the objects before the page and need to create a default page; this page needs then to
    // be activated
    // do not activate page if lj object has arrived, since here there will eventually arrive
    // the state of the active page
    ActivatePage(page); 

  page->AppendObject(obj);

  if (shared) CommunicationManager::cm->RegisterSubcomponent(obj->GetID(), type);
  
  if (page != SessionManager::sm->GetActivePage()) return true; // obj is not visible -> finished
  
  if (obj->Annotation() && !SessionManager::sm->AnnotationMode()) return true; // obj is not visible

  if (obj->GetOrder() == page->GetEndObjectOrder()) {
    // to draw last obj is sufficient
    obj->Draw(page->GetCanvas()); 
  } else {
    obj->Draw(page->GetCanvas()); 
    page->ReDraw();
  }

  DisplayParticipantHint(page, participantID, obj);

  return false;
}


void MLB::AduManager::HandleEvent(RefCount<RTPI::RtpiAdu> adu) {
  RTPI::RtpiEvent* event = (RTPI::RtpiEvent*)&(*adu);
  char* buf    = event->GetPayload();
  uint8*  p8   = (uint8*)buf;
  uint16* p16  = (uint16*)buf;
  uint32* p32  = (uint32*)buf;
  int  version = (p8[0] & MLB_MASK_V) >> 6;
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];
  int  subtype = p8[2];

  if (version != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("AduManager::HandleEvent - wrong version of mlb protocol %u\n", version);
    return;
  }

  GenericObject* tmp = SessionManager::sm->GetObject(event->GetSubcomponentID());
  if (!tmp) {
    SessionManager::sm->message->Error("AduManager::HandleEvent - object type %u  with id %u does not exist\n", objType, event->GetSubcomponentID());

    return;
  }
  Session* session = SessionManager::sm->GetParentSession(tmp);
  if (!session) {
    SessionManager::sm->message->Error("AduManager::HandleEvent - parent session for object type %u  with id %u does not exist\n", 
				       objType, event->GetSubcomponentID());
    return;
  }
  bool shared  = (session->GetSessionType() == Session::SHARED);
  
  bool own     = (event->GetParticipantID() == CommunicationManager::cm->GetParticipantID()) ? true : false;


  #ifdef DEBUG_ADU_ORDERING
  static lastSN = 0;
  if (adu->GetPacketHeader()->GetSequenceNumber() < lastSN)
    SessionManager::sm->message->Error("AduManager::HandleEvent - event for object with id %u with ts %u and sn %u\n",
				       adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp(), adu->GetPacketHeader()->GetSequenceNumber());
  else
    SessionManager::sm->message->Info("AduManager::HandleEvent - event for object with id %u with ts %u and sn %u\n",
				      adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp(), adu->GetPacketHeader()->GetSequenceNumber());

  lastSN = adu->GetPacketHeader()->GetSequenceNumber();
  #endif


  // register participant activity 
  ParticipantManager::pm->UpdateParticipantActiveness(adu->GetParticipantID());


  switch (objType) {
  //
  // Folder
  //
  case CommunicationManager::FOLDER: {
    if (!tmp || (tmp->GetType() != GenericObject::FOLDER)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - folder with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Folder* obj = (Folder*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for folder %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::SET_ACTIVE: {
	if (!SessionManager::sm->SynchMode()) {
	  SessionManager::sm->SetGlobalFolder(obj->GetID());
	  if (shared) TimeWarp::tw->InsertAdu(obj->GetStateAdu(rtpitime(), 3));
	  return;
	}
	if (obj != SessionManager::sm->GetActiveFolder()) {
	  //if (SessionManager::sm->GetActivePage() && !(SessionManager::sm->IsParentOf(obj->GetNodePath(),
	  //								      SessionManager::sm->GetActivePage()->GetNodePath()))) {
	  if (SessionManager::sm->GetActivePage()) {
	    if (shared && SessionManager::sm->GetParentSession(SessionManager::sm->GetActivePage())->GetSessionType() == Session::SHARED)
	      CommunicationManager::cm->DeactivateAllSubcomponents();
	    SessionManager::sm->GetActivePage()->DeActivate();
	  }
	  //if (SessionManager::sm->GetActiveFolder()
	  //    && (SessionManager::sm->GetParentSession(SessionManager::sm->GetActiveFolder())->GetSessionType() == Session::SHARED))
	  //  CommunicationManager::cm->ChangeSubcomponentActiveness(SessionManager::sm->GetActiveFolder()->GetID(), false);

	  SessionManager::sm->RemoteSelectContainer(obj->GetNodePath());

	  #ifdef ENABLE_TW
	  if (SessionManager::sm->GetActivePage())
	    SessionManager::sm->GetActivePage()->StoreAllObjectsInTWQ(rtpitime());
	  #endif

	  session->Draw();
	  SessionManager::sm->DrawPageTree();
	  if (shared) TimeWarp::tw->InsertAdu(obj->GetStateAdu(rtpitime(), 3));
	  //}
	}
	return;
      }
      case CommunicationManager::CHG_NAME: {
	int nlength = p8[3];
	char* tmp = new char[nlength + 1];
	tmp[nlength] = '\0';
	memcpy(tmp, buf + 4, nlength);
	obj->SetName(tmp);
	SessionManager::sm->DrawPageTree();
	delete[] tmp;

	// UCE connection
	if (UceManager::connected) {
	  UCE::UCE_Interface::ucei->cmd_rename(obj->GetID(), obj->GetName());
	}

	return;
      }
      case CommunicationManager::CHG_PARENT: {
	obj->SetOrder(ntohl(p32[1]));
	SessionManager::sm->RemoteChangeHierachy(obj->GetID(), event->GetObjectID());

        #ifdef DEBUG_ORDER
	SessionManager::sm->message->Info("AduManager::HandleEvent -  folder %u with new parent %u and order %u\n",
					obj->GetID(), event->GetObjectID(), obj->GetOrder());
        #endif
	return;
      }
      }


    if (opType == CommunicationManager::ADU_DELETE) {
      //obj->ChangeSubcomponentsActiveness(false);
      SessionManager::sm->RemoteDeleteContainer(obj);

      // UCE connection
      if (UceManager::connected) {
	UCE::UCE_Interface::ucei->cmd_delete(obj->GetID());
      }
    }

    return;
  }

  //
  // Page
  //
  case CommunicationManager::PAGE: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::PAGE)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - page with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Page* obj = (Page*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for page %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::SET_ACTIVE: {
	if (!p8[3]) return; // page set inactive!
	if (!SessionManager::sm->SynchMode()) {
	  SessionManager::sm->SetGlobalPage(obj->GetID());
	  return;
	}
	if (obj != SessionManager::sm->GetActivePage()) {
	  #ifdef ENABLE_TW
	  if (shared) obj->StoreAllObjectsInTWQ(rtpitime());
	  #endif
  	  ActivatePage(obj);
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	obj->SetPageDimension(ntohs(p16[2]), ntohs(p16[3]));
	return;
      }
      case CommunicationManager::CHG_BACKGRCOLOR: {
	obj->SetBackgroundColor(ColoredObject::ParseColor((char*)(p8 + 4)));
	return;
      }
      case CommunicationManager::CHG_NAME: {
	int nlength = p8[3];
	char* tmp = new char[nlength + 1];
	tmp[nlength] = '\0';
	memcpy(tmp, buf + 4, nlength);
	obj->SetName(tmp);
	SessionManager::sm->DrawPageTree();
	delete[] tmp;
	return;
      }
      case CommunicationManager::CHG_PARENT: {
	if (SessionManager::sm->GetActivePage()) {
	  if (shared) CommunicationManager::cm->DeactivateAllSubcomponents();
	  //SessionManager::sm->GetActivePage()->ChangeSubcomponentsActiveness(false);
	  SessionManager::sm->GetActivePage()->DeActivate();
	}

	obj->SetOrder(ntohl(p32[1]));
	if (event->ObjectIDIsSet()) SessionManager::sm->RemoteChangeHierachy(obj->GetID(), event->GetObjectID());
	else SessionManager::sm->RemoteChangeHierachy(obj->GetID(), 1);

	SessionManager::sm->ClearPageTree();
	SessionManager::sm->DrawPageTree();

        #ifdef DEBUG_ORDER
	SessionManager::sm->message->Info("AduManager::HandleEvent -  page %u with new parent %u and order %u\n",
					obj->GetID(), event->GetObjectID(), obj->GetOrder());
        #endif
	return;
      }
      case CommunicationManager::CHG_ACTIVE : {
	DrawObject*  active = (DrawObject*)obj->GetObject(event->GetObjectID());
	if (!active) active = (DrawObject*)obj->GetGroup(event->GetObjectID());
	if (active) {
	  if (active->GetType() == GenericObject::GROUP) obj->SetActiveObject(((Group*)active)->GetFirst());
	  else obj->SetActiveObject(active);
	} else {
	  SessionManager::sm->message->Error("AduManager::HandleEvent - CHG_ACTIVE: object with %u id does not exist\n",
			 event->GetObjectID());
	}
	return;
      }
      }


    if (opType == CommunicationManager::ADU_DELETE) {
      if (shared) obj->ChangeSubcomponentsActiveness(false);
      SessionManager::sm->RemoteDeleteContainer(obj);
    }

    return;
  }

  //
  // Group 
  //
  /*case CommunicationManager::GROUP: {
    Group* obj = (Group*)session->GetGroup(event->GetSubcomponentID());
    if (!obj) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - group with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for group %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find page for group with id %u\n", obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE) {
      switch (subtype) {
      case CommunicationManager::MOVE: {
	obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	return;
      }
      case CommunicationManager::CHG_ACTIVE : {
	if (p8[3] == 1) obj->SetFirstActive();
	else if (p8[3] == 2) obj->SetLastActive();
	else if (p8[3] == 3) obj->SetLeftActive();
	else if (p8[3] == 4) obj->SetRightActive();
	return;
      }
      case CommunicationManager::RAISE : {
	return;
      }
      case CommunicationManager::LOWER : {
	return;
      }
      }
    }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) {
	page->UnSelectObjects();
	obj->DeleteSelection(page->GetCanvas());
      }
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }

    return;
  }
*/

  //
  // Rectangle
  //
  case CommunicationManager::RECTANGLE: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::RECTANGLE)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - rectangle with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Rectangle* obj = (Rectangle*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for rectangle %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for rectangle with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	int width  = ntohs(p16[4]);
	int height = ntohs(p16[5]);
	if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
	if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ModifyCoords(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[3]), width, height);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	  obj->SetWidth(width);
	  obj->SetHeight(height);
	}
	return;
      }
      case CommunicationManager::CHG_LINEWIDTH: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineWidth(page->GetCanvas(), p8[7]);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineWidth(p8[7]);
	}
	return;
      }
      case CommunicationManager::CHG_LINESTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineStyle(page->GetCanvas(), (char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineStyle((char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_FILLCOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeFillColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetFillColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if ((page == SessionManager::sm->GetActivePage()) && (obj->IsSelected())) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }

    return;
  }

  //
  // Oval
  //
  case CommunicationManager::OVAL: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::OVAL)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - oval with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Oval* obj = (Oval*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for oval %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for oval with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	int width  = ntohs(p16[4]);
	int height = ntohs(p16[5]);
	if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
	if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ModifyCoords(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[3]), width, height);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	  obj->SetWidth(width);
	  obj->SetHeight(height);
	}
	return;
      }
      case CommunicationManager::CHG_LINEWIDTH: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineWidth(page->GetCanvas(), p8[7]);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineWidth(p8[7]);
	}
	return;
      }
      case CommunicationManager::CHG_LINESTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineStyle(page->GetCanvas(), (char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineStyle((char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_FILLCOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeFillColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetFillColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }
    return;
  }

  //
  // Line
  //
  case CommunicationManager::LINE: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::LINE)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - line with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Line* obj = (Line*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for line %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for line with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT_1: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), 1, ntohs(p16[2]), ntohs(p16[3]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->MovePoint(page->GetCanvas(), 1, ntohs(p16[2]), ntohs(p16[3]), false);
	}
	return;
      }
      case CommunicationManager::MOVE_POINT_2: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), 2, ntohs(p16[2]), ntohs(p16[3]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->MovePoint(page->GetCanvas(), 2, ntohs(p16[2]), ntohs(p16[3]), false);
	}
	return;
      }
      case CommunicationManager::CHG_LINEWIDTH: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineWidth(page->GetCanvas(), p8[7]);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineWidth(p8[7]);
	}
	return;
      }
      case CommunicationManager::CHG_LINESTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineStyle(page->GetCanvas(), (char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineStyle((char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_FILLCOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeFillColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetFillColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_CAPSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeCapStyle(page->GetCanvas(), (char*)Line::ParseCapStyle((char*)(p8 + 3)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetCapStyle((char*)Line::ParseCapStyle((char*)(p8 + 3)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_JOINSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeJoinStyle(page->GetCanvas(), (char*)Line::ParseJoinStyle((char*)(p8 + 3)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetJoinStyle((char*)Line::ParseJoinStyle((char*)(p8 + 3)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_ARROWSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeArrowStyle(page->GetCanvas(), (char*)Line::ParseArrowStyle((char*)(p8 + 3)).c_str());
	  obj->ChangeArrowShape(page->GetCanvas(), (char*)Line::ParseArrowShape((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetArrowStyle((char*)Line::ParseArrowStyle((char*)(p8 + 3)).c_str());
	  obj->SetArrowShape((char*)Line::ParseArrowShape((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }

    return;
  }

  //
  // Polyline
  //
  case CommunicationManager::POLYLINE: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || ((tmp->GetType() != GenericObject::FREEHAND) && (tmp->GetType() != GenericObject::POLYLINE))) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - polyline with %u id does not exist (ts %u)\n",
		     event->GetSubcomponentID(), event->GetAduHeader()->GetTimestamp());
      return;
    }

    PolyLine* obj = (PolyLine*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for polyline %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for polyline with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]), false);
	}
	return;
      }
      case CommunicationManager::CHG_LINEWIDTH: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineWidth(page->GetCanvas(), p8[7]);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineWidth(p8[7]);
	}
	return;
      }
      case CommunicationManager::CHG_LINESTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineStyle(page->GetCanvas(), (char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineStyle((char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  obj->ReDraw(page->GetCanvas());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_CAPSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeCapStyle(page->GetCanvas(), (char*)Line::ParseCapStyle((char*)(p8 + 3)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetCapStyle((char*)Line::ParseCapStyle((char*)(p8 + 3)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_JOINSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeJoinStyle(page->GetCanvas(), (char*)Line::ParseJoinStyle((char*)(p8 + 3)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetJoinStyle((char*)Line::ParseJoinStyle((char*)(p8 + 3)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_ARROWSTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeArrowStyle(page->GetCanvas(), (char*)Line::ParseArrowStyle((char*)(p8 + 3)).c_str());
	  obj->ChangeArrowShape(page->GetCanvas(), (char*)Line::ParseArrowShape((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetArrowStyle((char*)Line::ParseArrowStyle((char*)(p8 + 3)).c_str());
	  obj->SetArrowShape((char*)Line::ParseArrowShape((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::ADD_POINT: {
	obj->AppendNewPoint(ntohs(p16[2]), ntohs(p16[3]));
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->DrawNewSegment(page->GetCanvas());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }

    return;
  }

  //
  // Polygon
  //
  case CommunicationManager::POLYGON: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::POLYGON)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - polygon with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    Polygon* obj = (Polygon*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for polygon %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for polygon with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]), false);
	}
	return;
      }
      case CommunicationManager::CHG_LINEWIDTH: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineWidth(page->GetCanvas(), p8[7]);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineWidth(p8[7]);
	}
	return;
      }
      case CommunicationManager::CHG_LINESTYLE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineStyle(page->GetCanvas(), (char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineStyle((char*)ColoredObject::ParseLineStyle((char*)(p8 + 7)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeLineColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  obj->ReDraw(page->GetCanvas());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetLineColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::CHG_FILLCOLOR: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ChangeFillColor(page->GetCanvas(), (char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetFillColor((char*)ColoredObject::ParseColor((char*)(p8 + 4)).c_str());
	}
	return;
      }
      case CommunicationManager::ADD_POINT: {
	obj->AppendNewPoint(ntohs(p16[2]), ntohs(p16[3]));
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->DrawNewSegment(page->GetCanvas());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::CLOSE_POLYGON: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ReDraw(page->GetCanvas());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
    }

    return;
  }

  //
  // Text
  //
  case CommunicationManager::TEXT: {
    Group* group = (Group*)SessionManager::sm->GetGroup(event->GetSubcomponentID());
    if (!group) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - group with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Info("AduManager::HandleEvent - received event %u for group %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page*  page  = (Page*)session->GetContainer(group->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find page for group with id %u\n",
		     group->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE) {
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {group->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) group->Hide(page->GetCanvas()); else group->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  group->Move(page->GetCanvas(), ntohs(p16[2]) - group->GetX(), ntohs(p16[3]) - group->GetY());
	  group->UpdateDimension();
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	} else {
	  group->Move(page->GetCanvas(), ntohs(p16[2]) - group->GetX(), ntohs(p16[3]) - group->GetY(), false);
	  group->UpdateDimension();
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	return;
      }
      case CommunicationManager::INSERT_FONT: {
	string fontFamily, fontSize, fontWeight, fontSlant, textPos;
	Color color;
	group->ParseInsertNewFontTG(buf, color, fontFamily, fontSize, fontWeight, fontSlant, textPos);
	group->RemoteInsertNewFont(page, fontFamily, fontSize, fontWeight, fontSlant, textPos, own);
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::CHG_FONT: {
	string fontFamily, fontSize, fontWeight, fontSlant, textPos;
	group->ParseChangeFontTG(buf, fontFamily, fontSize, fontWeight, fontSlant);
	group->RemoteChangeFont(page, fontFamily, fontSize, fontWeight, fontSlant, textPos);
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::CHG_FONTFAMILY: {
	return;
      }
      case CommunicationManager::CHG_FONTSIZE: {
	return;
      }
      case CommunicationManager::CHG_LINECOLOR: {
	group->RemoteChangeColor(page, ColoredObject::ParseColor(buf + 4));
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::INSERT_FCOLOR: {
	group->RemoteInsertNewColor(page, ColoredObject::ParseColor(buf + 4), own);
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::CHG_FONTWEIGHT: {
	return;
      }
      case CommunicationManager::CHG_FONTSLANT: {
	return;
      }
      case CommunicationManager::CHG_FONTPOS: {
	return;
      }
      case CommunicationManager::INSERT_CHARS: {
	if (p8[3]) {
	  group->RemoteNewLine(page, own);
	} else {
	  int length = ntohs(p16[3]);
	  char* tmp = new char[length + 1];
	  tmp[length] = '\0';
	  strncpy(tmp, buf + 8, length);
	  group->RemoteInsertNewChars(tmp);
	  delete[] tmp;
	}
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::DELETE_CHARS: {
	if (p8[3] == 1) group->RemoteBackspace(page, own);
	else return;
	group->UpdateDimension();
	if (page == SessionManager::sm->GetActivePage()) {
	  group->ReDraw(page->GetCanvas());
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::RAISE: {
	uint32 order = ntohl(p32[1]);
	page->DeleteObjectFromContainer(group);
	group->SetOrder(order);
	page->AppendObject(group);

	Text* txt;
	std::vector<uint32>* ids = group->GetObjectIDs();
	for (std::vector<uint32>::iterator i = ids->begin(); i != ids->end(); ++i) {
	  txt = (Text*)(page->GetObject(*i));
	  page->DeleteObjectFromContainer(txt);
	  txt->SetOrder(order);
	  page->AppendObject(txt);
	  order += MLB_OO_OFFSET;
	}
	delete ids;

	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	uint32 order = ntohl(p32[1]);
	page->DeleteObjectFromContainer(group);
	group->SetOrder(order);
	page->AppendObject(group);

	Text* txt;
	std::vector<uint32>* ids = group->GetObjectIDs();
	for (std::vector<uint32>::iterator i = ids->begin(); i != ids->end(); ++i) {
	  txt = (Text*)(page->GetObject(*i));
	  page->DeleteObjectFromContainer(txt);
	  txt->SetOrder(order);
	  page->AppendObject(txt);
	  order -= MLB_OO_OFFSET;
	}

	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      case CommunicationManager::CHG_ACTIVE : {
	if (p8[3] == 1) group->RemoteGoHome(own);
	else if (p8[3] == 2) group->RemoteGoEnd(own);
	else if (p8[3] == 3) group->RemoteGoLeft(own);
	else if (p8[3] == 4) group->RemoteGoRight(own);
	else if (p8[3] == 5) group->RemoteGoUp(own);
	else if (p8[3] == 6) group->RemoteGoDown(own);
	else if (p8[3] == 7) group->RemoteSetActive(ntohs(p16[2]), ntohs(p16[3]));
	if (page == SessionManager::sm->GetActivePage()) {
	  group->DrawTextCursor(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), group);
	}
	return;
      }
      }
    }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(group->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(group->GetID());
      if (page == SessionManager::sm->GetActivePage()) {
	if (group->IsSelected()) {
	  page->UnSelectObjects();
	  group->DeleteSelection(page->GetCanvas());
	}
	DisplayParticipantHint(page, event->GetParticipantID(), group);
	group->Hide(page->GetCanvas());
	group->DeleteTextCursor(page->GetCanvas());
      }
      std::vector<uint32>* groupObjects = group->GetObjectIDs();
      for (std::vector<uint32>::iterator j = groupObjects->begin(); j != groupObjects->end(); ++j) {
	page->GetAnimations().DeleteObject((DrawObject*)(page->GetObject(*j)));
	delete page->RemoteDeleteObject(*j);
      }
      group->DeleteAll(page->GetCanvas());
      page->RemoteDeleteObject(group);
      page->GetAnimations().DeleteObject(group);
      delete group;
    }

    return;
  }

  //
  // Image
  //
  case CommunicationManager::IMAGE: {
    tmp = session->GetObject(event->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::IMAGE)) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - image with %u id does not exist\n",
		     event->GetSubcomponentID());
      return;
    }

    PhotoImage* obj = (PhotoImage*)tmp;

    #ifdef DEBUG_ADU
    SessionManager::sm->message->Error("AduManager::HandleEvent - received event %u for object %u\n",
		   opType, event->GetSubcomponentID());
    #endif

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleEvent - did not find container for object with id %u\n",
		     obj->GetID());
      return;
    }

    if (shared) CommunicationManager::cm->UpdateEventSequenceNumber(event->GetSubcomponentID(), event->GetParticipantID(),
							event->GetPacketHeader()->GetSequenceNumber());

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::CHG_VISIBILITY: {
	bool vis = (p8[0] & MLB_MASK_VIS) >> 1;
	if (vis) {obj->SetVisible(); if (SessionManager::sm->GetActivePage() == page) page->ReDraw(); } 
	else {if (SessionManager::sm->GetActivePage() == page) obj->Hide(page->GetCanvas()); else obj->SetInvisible();}
	return;
      }
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	int width  = ntohs(p16[4]);
	int height = ntohs(p16[5]);
	if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
	if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ModifyCoords(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[3]), width, height);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	  obj->SetWidth(width);
	  obj->SetHeight(height);
	}
	return;
      }
      case CommunicationManager::RAISE: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      case CommunicationManager::LOWER: {
	page->DeleteObjectFromContainer(obj);
	obj->SetOrder(ntohl(p32[1]));
	page->AppendObject(obj);
	if (page == SessionManager::sm->GetActivePage()) {
	  page->ReDraw();
	  DisplayParticipantHint(page, event->GetParticipantID(), obj);
	}
	return;
      }
      }

    if (opType == CommunicationManager::ADU_DELETE) {
      if (!own) UndoManager::um->DeleteAdu(obj->GetID());
      if (shared) CommunicationManager::cm->UnregisterSubcomponent(obj->GetID());
      if (obj->IsSelected()) page->UnSelectObjects();
      if (page == SessionManager::sm->GetActivePage()) DisplayParticipantHint(page, event->GetParticipantID(), obj);
      page->RemoteDeleteObject(obj);
      page->GetAnimations().DeleteObject(obj);
      delete obj;
   }
    return;
  }


  default:
    SessionManager::sm->message->Error("AduManager::HandleEvent - wrong object type %u\n", objType);
  }
}


void MLB::AduManager::HandleCue(RefCount<RTPI::RtpiAdu> adu) {
  RTPI::RtpiCue* cue = (RTPI::RtpiCue*)&(*adu);
  uint8*  p8   = (uint8*)(cue->GetPayload());
  uint16* p16  = (uint16*)(cue->GetPayload());
  uint32* p32  = (uint32*)(cue->GetPayload());
  int  version = (p8[0] & MLB_MASK_V) >> 6;
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];
  int  subtype = p8[2];

  if (version != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("AduManager::HandleCue - wrong version of mlb protocol %u\n", version);
    return;
  }

  Session* session = SessionManager::sm->GetSharedSession();
  GenericObject* tmp;

  switch (objType) {
  //
  // Group
  //
  case CommunicationManager::GROUP: {
    Group* obj = (Group*)session->GetGroup(cue->GetSubcomponentID());
    if (!obj) {
      SessionManager::sm->message->Error("AduManager::HandleCue - group with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for group %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for group with id %u\n",
					 obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE) {
      switch (subtype) {
      case CommunicationManager::MOVE: {
	obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	return;
      }
      }
    }

    return;
  }


  //
  // Rectangle
  //
  case CommunicationManager::RECTANGLE: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::RECTANGLE)) {
      SessionManager::sm->message->Error("AduManager::HandleCue - rectangle with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for rectangle %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Rectangle* obj = (Rectangle*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for rectangle with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	int width  = ntohs(p16[4]);
	int height = ntohs(p16[5]);
	if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
	if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ModifyCoords(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[3]), width, height);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	  obj->SetWidth(width);
	  obj->SetHeight(height);
	}
	return;
      }
      }
    return;
  }

  //
  // Oval
  //
  case CommunicationManager::OVAL: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::OVAL)) {
      SessionManager::sm->message->Error("AduManager::HandleCue - oval with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for oval %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Oval* obj = (Oval*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for oval with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::CHG_SIZE: {
	int width  = ntohs(p16[4]);
	int height = ntohs(p16[5]);
	if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
	if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

	if (page == SessionManager::sm->GetActivePage()) {
	  obj->ModifyCoords(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[3]), width, height);
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	  obj->SetWidth(width);
	  obj->SetHeight(height);
	}
	return;
      }
      }
    return;
  }

  //
  // Line
  //
  case CommunicationManager::LINE: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::LINE)) {
      SessionManager::sm->message->Error("AduManager::HandleCue - line with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for line %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Line* obj = (Line*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for line with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT_1: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), 1, ntohs(p16[2]), ntohs(p16[3]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->MovePoint(page->GetCanvas(), 1, ntohs(p16[2]), ntohs(p16[3]), false);
	}
	return;
      }
      case CommunicationManager::MOVE_POINT_2: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), 2, ntohs(p16[2]), ntohs(p16[3]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->MovePoint(page->GetCanvas(), 2, ntohs(p16[2]), ntohs(p16[3]), false);
	}
	return;
      }
      }
    return;
  }

  //
  // Polyline
  //
  case CommunicationManager::POLYLINE: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || ((tmp->GetType() != GenericObject::FREEHAND) && (tmp->GetType() != GenericObject::POLYLINE))) {
      SessionManager::sm->message->Error("AduManager::HandleCue - polyline with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for polyline %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    PolyLine* obj = (PolyLine*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for object with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]), false);
	}
	return;
      }
      }
    return;
  }

  //
  // Polygon
  //
  case CommunicationManager::POLYGON: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::POLYGON)) {
      SessionManager::sm->message->Error("AduManager::HandleCue - polygon with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for polygon %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Polygon* obj = (Polygon*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for polygon with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->SetX(ntohs(p16[2]));
	  obj->SetY(ntohs(p16[3]));
	}
	return;
      }
      case CommunicationManager::MOVE_POINT: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]));
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->MovePoint(page->GetCanvas(), ntohs(p16[2]), ntohs(p16[4]), ntohs(p16[5]), false);
	}
	return;
      }
      }
    return;
  }

  //
  // Text
  //
  case CommunicationManager::TEXT: {
    Group* obj = (Group*)SessionManager::sm->GetGroup(cue->GetSubcomponentID());
    if (!obj) {
      SessionManager::sm->message->Error("AduManager::HandleCue - group with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for group %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    Page*  page  = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for group with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  obj->UpdateDimension();
	  obj->ReDraw(page->GetCanvas());
	  obj->DrawTextCursor(page->GetCanvas());
	} else {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY(), false);
	  obj->UpdateDimension();
	}
	return;
      }
      }
    return;
  }


  //
  // Image
  //
  case CommunicationManager::IMAGE: {
    tmp = session->GetObject(cue->GetSubcomponentID());
    if (!tmp || (tmp->GetType() != GenericObject::IMAGE)) {
      SessionManager::sm->message->Error("AduManager::HandleCue - image with %u id does not exist\n",
		     cue->GetSubcomponentID());
      return;
    } else {
      #ifdef DEBUG_ADU
      SessionManager::sm->message->Info("AduManager::HandleCue - received cue %u for image %u\n",
		     opType, cue->GetSubcomponentID());
      #endif
    }

    PhotoImage* obj = (PhotoImage*)tmp;

    Page* page = (Page*)session->GetContainer(obj->GetID());
    if (!page) {
      SessionManager::sm->message->Error("AduManager::HandleCue - did not find container for image with id %u\n",
		     obj->GetID());
      return;
    }

    if (opType == CommunicationManager::ADU_CHANGE)
      switch (subtype) {
      case CommunicationManager::MOVE: {
	if (page == SessionManager::sm->GetActivePage()) {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY());
	  if (obj->IsSelected()) obj->DrawSelection(page->GetCanvas());
	} else {
	  obj->Move(page->GetCanvas(), ntohs(p16[2]) - obj->GetX(), ntohs(p16[3]) - obj->GetY(), false);
	}
	return;
      }
      }
    return;
  }


  default:
    SessionManager::sm->message->Error("AduManager::HandleCue - wrong object type\n");
  }
}


MLB::Page* MLB::AduManager::CreateDefaultPage(uint32 id) {
  SessionManager::sm->message->Info("AduManager::CreateDefaultPage - for %u\n", id);
  Page* page = new Page(id, Session::SHARED);
  SessionManager::sm->GetSharedSession()->InsertContainerEnd(page);
  SessionManager::sm->ClearPageTree();
  SessionManager::sm->DrawPageTree();
  return page;
}


void MLB::AduManager::ActivatePage(Page* page, bool redraw) {
  if (!redraw && (SessionManager::sm->GetActivePage() == page)) return; // already active

  if (SessionManager::sm->GetActivePage()) {
    if (SessionManager::sm->GetParentSession(SessionManager::sm->GetActivePage())->GetSessionType() == Session::SHARED) {
      CommunicationManager::cm->DeactivateAllSubcomponents();
    }
    SessionManager::sm->GetActivePage()->DeActivate();
  }
  SessionManager::sm->SetGlobalPage(page->GetID());
  ((Folder*)SessionManager::sm->GetParentContainer(page->GetID()))->SetCurrentPage(page);
  SessionManager::sm->DrawPageTree();
  SessionManager::sm->RemoteSelectContainer(page->GetNodePath());
  SessionManager::sm->ClearPageTree();
  SessionManager::sm->DrawPageTree();
  
  // shared session ?
  if (CommunicationManager::cm->rtpi) CommunicationManager::cm->rtpi->send_rtcpi_note(); 

  // UCE connection
  if (UceManager::connected) {
    UCE::UCE_Interface::ucei->cmd_set_active(page->GetID());
  }
}


void MLB::AduManager::InsertAdu(RefCount<RTPI::RtpiAdu> adu) {
  if (!adu) return;
  _InsertAdu(adu);
  CheckForDueAdus();
  SetExecutionTimer();
} 


void MLB::AduManager::_InsertAdu(RefCount<RTPI::RtpiAdu> adu) {
  #ifdef DEBUG
  if (adu->GetType() == RTPI::RtpiHeader::PKT_EVENT) {
    static lastSN = 0;
    if (adu->GetPacketHeader()->GetSequenceNumber() < lastSN)
      SessionManager::sm->message->Error("AduManager::_InsertAdu - insert event for object with id %u with ts %u and sn %u\n",
					adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp(), adu->GetPacketHeader()->GetSequenceNumber());
    else
      SessionManager::sm->message->Info("AduManager::_InsertAdu - insert event for object with id %u with ts %u and sn %u\n",
					adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp(), adu->GetPacketHeader()->GetSequenceNumber());
      
    lastSN = adu->GetPacketHeader()->GetSequenceNumber();
  } else
    SessionManager::sm->message->Info("AduManager::_InsertAdu - insert adu for object with id %u with ts %u\n",
				      adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp());

  #endif

  if (aq.empty()) {
    aq.push_back(adu);
    return; 
  }

  // sort by timestamp and pid
  std::list<RefCount<RTPI::RtpiAdu> >::iterator pos = aq.begin();
  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() > (*pos)->GetAduHeader()->GetTimestamp()))
    ++pos;

  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() > (*pos)->GetParticipantID()))
    ++pos;

  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() == (*pos)->GetParticipantID()))
    ++pos;

  if (pos == aq.end())
    aq.push_back(adu);
  else
    aq.insert(pos, adu);

  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::_InsertAdu - queue size %u\n", aq.size());
  #endif
}


void MLB::AduManager::InsertAdus(RefCount<RTPI::RtpiAduVector> adus) {
  if (!adus || !adus->size()) return;

  _InsertAdus(adus);
  CheckForDueAdus();
  SetExecutionTimer();
}


void MLB::AduManager::_InsertAdus(RefCount<RTPI::RtpiAduVector> adus) {
  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::_InsertAdu - insert %u adus\n", adus->size());
  #endif

  if (aq.empty()) {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i)
      aq.push_back(*i);
    return;
  }

  // sort by timestamp and pid
  RefCount<RTPI::RtpiAdu> adu = adus->front();
  std::list<RefCount<RTPI::RtpiAdu> >::iterator pos = aq.begin();
  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() > (*pos)->GetAduHeader()->GetTimestamp()))
    ++pos;

  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() > (*pos)->GetParticipantID()))
    ++pos;

  while ((pos != aq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() == (*pos)->GetParticipantID()))
    ++pos;

  if (pos == aq.end()) {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i)
      aq.push_back(*i);
  } else {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i)
      pos = aq.insert(pos, *i);
  }

  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::_InsertAdu - queue size %u\n", aq.size());
  #endif
}


void MLB::AduManager::PrintAduQueue() {
  GUI::MessageWindow::Info("AduManager::PrintAduQueue  -  queue size %u\n", aq.size());
  for (std::list<RefCount<RTPI::RtpiAdu> >::iterator i = aq.begin(); i != aq.end(); ++i) {
    GUI::MessageWindow::Info("ADU (%u) - ts %u pid %u\n", (*i)->GetType(), (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetParticipantID());
  }
}

 
void MLB::AduManager::CheckForDueAdus() {
  uint32 currentTime;

  std::list<RefCount<RTPI::RtpiAdu> >::iterator i = aq.begin();
  while (i != aq.end()) {
    #ifdef WIN32
    currentTime = rtpitime() + 20;
    #else
    currentTime = rtpitime() + 6;
    #endif
    if ((*i)->GetAduHeader()->GetTimestamp() <= currentTime) {
      switch ((*i)->GetType()) {
      case RTPI::RtpiHeader::PKT_EVENT: {
	#ifdef DEBUG
	SessionManager::sm->message->Info("AduManager::CheckForDueAdus - Event for Object %u is due (ts - ct)\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp(), currentTime);
	#endif
	HandleEvent(*i);
	break;
      }
      case RTPI::RtpiHeader::PKT_CUE: {
	#ifdef DEBUG
	SessionManager::sm->message->Info("AduManager::CheckForDueAdus - Cue for Object %u is due (ts - ct)\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp(), currentTime);
	#endif
	HandleCue(*i);
	break;
      }
      case RTPI::RtpiHeader::PKT_STATE: {
	#ifdef DEBUG
	SessionManager::sm->message->Info("AduManager::CheckForDueAdus - State for Object %u is due (ts - ct)\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp(), currentTime);
	#endif
	HandleState(*i);
	break;
      }
      }

      RefCount<RTPI::RtpiAdu> adu = *i;
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_CUE) {
	// cues do not neet to be stored
	//SessionManager::sm->message->Info("AduManager::CheckForDueAdus - delete payload for cue from %u\n", (*i)->GetParticipantID());
	//delete (*i)->GetPayload();
      } else {
	TimeWarp::tw->InsertAdu(*i);
      }
      i = aq.erase(i);
    } else {
      break; // all following adus have a greater timestamp
    }
  }

  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::CheckForDueAdus - queue size %u\n", aq.size());
  #endif
}


RefCount<RTPI::RtpiAduList> MLB::AduManager::CheckForIntentionViolation(RefCount<RTPI::RtpiAdu> adu) {
  RefCount<RTPI::RtpiAduList> icAdus = new RTPI::RtpiAduList;

  for (RTPI::RtpiAduList::iterator i = aq.begin(); i != aq.end(); ++i) {
    if (CommunicationManager::cm->IntentionViolation(adu, *i))
      icAdus->push_back(*i);
  }

  return icAdus;
}


void MLB::AduManager::SetExecutionTimer() {
  if (aq.empty()) return;
  if (timerRunning) return;

  std::list<RefCount<RTPI::RtpiAdu> >::iterator first = aq.begin();

  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::SetExecutionTimer - runningTime %u, currentTime %u\n",
		 (*first)->GetAduHeader()->GetTimestamp() - rtpitime(), rtpitime());
  #endif

  timerRunning = true;
  TimerManager::tm->SetTimer(this, new TimerObject((*first)->GetAduHeader()->GetTimestamp() - rtpitime(), TimerObject::ADU_DUE));
}


void MLB::AduManager::ExecuteTimer(TimerObject* timer) {
  #ifdef DEBUG
  SessionManager::sm->message->Info("AduManager::ExecuteTimer - timer run out (runningTime %u, currentTime %u)\n",
		 timer->runningTime, rtpitime());
  #endif

  switch (timer->type) {
  case TimerObject::ADU_DUE: {
    timerRunning = false;
    delete timer;
    CheckForDueAdus();
    SetExecutionTimer();
    break;
  }
  }
}


void MLB::AduManager::DisplayParticipantHint(Page* parent, uint32 pid, DrawObject* obj) {
  if (!participantHints) return;
  if (pid == CommunicationManager::cm->pid) return;

  if (!obj) return;
  if ((obj->GetType() == GenericObject::FOLDER) || (obj->GetType() == GenericObject::PAGE)) return;

  if (!parent) return;

  int x = obj->GetX();
  int y = obj->GetY();

  std::string participant;
  Participant* p = ParticipantManager::pm->GetParticipant(pid);
  if (!p) return;

  if (!(p->name.empty())) participant = p->name;
  else participant = p->cname;

  if (Et_GlobalEvalF(Et_Interp, "Canvas::DisplayParticipantHint \"%q\" %u %u \"%q\"",
		     parent->GetCanvas()->GetWidget(),
		     int(x * SessionManager::sm->GetZoom()),
		     int(y * SessionManager::sm->GetZoom()),
		     participant.c_str()) != TCL_OK)
    GUI::MessageWindow::Error("AduManager::DisplayParticipantHint - tcl error \"%q\"\n", Et_Interp->result);
  if (Et_GlobalEvalF(Et_Interp,"unset pList") != TCL_OK)
    GUI::MessageWindow::Error("AduManager::DisplayParticipantHint - tcl error \"%q\"\n", Et_Interp->result);
}


//
// class TimeWarp
//

MLB::TimeWarp* MLB::TimeWarp::tw = NULL;


void MLB::TimeWarp::Reset() {
  TimerManager::tm->DeleteTimer(this, TimerObject::TW_CLEAN);
  //for (RTPI::RtpiAduList::iterator i = twq.begin(); i != twq.end(); ++i)
  //  delete (*i)->GetPayload();
  twq.clear();
}


void MLB::TimeWarp::InsertAdu(RefCount<RTPI::RtpiAdu> adu) {
  #ifndef ENABLE_TW
  return;
  #else

  if (!adu) return;

  #ifdef DEBUG_CONS
  if (adu->GetType() == RTPI::RtpiHeader::PKT_STATE)
    SessionManager::sm->message->Info("TimeWarp::InsertAdu - insert state for %u with ts %u\n",
		   adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp());
  #endif

  if (twq.empty()) {
    twq.push_back(adu);
    return;
  }

  // sort by timestamp and pid
  std::list<RefCount<RTPI::RtpiAdu> >::iterator pos = twq.begin();
  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() > (*pos)->GetAduHeader()->GetTimestamp()))
    ++pos;

  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() > (*pos)->GetParticipantID()))
    ++pos;

  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() == (*pos)->GetParticipantID()))
    ++pos;

  if (pos == twq.end())
    twq.push_back(adu);
  else
    twq.insert(pos, adu);

  //RemoveOldAdus();

  #ifdef DEBUG_TW
  int bytes = 0;
  for (std::list<RefCount<RTPI::RtpiAdu> >::iterator i = twq.begin(); i != twq.end(); ++i) bytes += (*i)->GetPayloadSize();
  SessionManager::sm->message->Info("TimeWarp::InsertAdu - %u ADUs in TimeWarp-Queue with %u total bytes\n", twq.size(), bytes);
  #endif

  #endif  // #ifndef ENABLE_TW
}


void MLB::TimeWarp::InsertAdus(RefCount<RTPI::RtpiAduVector> adus) {
  #ifndef ENABLE_TW
  return;
  #else

  if (!adus || !adus->size()) return;

  if (twq.empty()) {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
      #ifdef DEBUG_CONS
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE)
	SessionManager::sm->message->Info("TimeWarp::InsertAdu - insert state for %u with ts %u\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp());
      twq.push_back(*i);
      #endif
    }
    return;
  }

  // sort by timestamp and pid
  RefCount<RTPI::RtpiAdu> adu = adus->front();
  std::list<RefCount<RTPI::RtpiAdu> >::iterator pos = twq.begin();
  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() > (*pos)->GetAduHeader()->GetTimestamp()))
    ++pos;

  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() > (*pos)->GetParticipantID()))
    ++pos;

  while ((pos != twq.end()) && (adu->GetAduHeader()->GetTimestamp() == (*pos)->GetAduHeader()->GetTimestamp())
	 && (adu->GetParticipantID() == (*pos)->GetParticipantID()))
    ++pos;

  if (pos == twq.end()) {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE)
        #ifdef DEBUG_CONS
	SessionManager::sm->message->Info("TimeWarp::InsertAdu - insert state for %u with ts %u\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp());
        #endif
      twq.push_back(*i);
    }
  } else {
    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE)
        #ifdef DEBUG_CONS
	SessionManager::sm->message->Info("TimeWarp::InsertAdu - insert state for %u with ts\n",
		       (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp());
        #endif
      pos = twq.insert(pos, *i);
    }
  }

  //RemoveOldAdus();

  #ifdef DEBUG_TW
  int bytes = 0;
  for (std::list<RefCount<RTPI::RtpiAdu> >::iterator i = twq.begin(); i != twq.end(); ++i) bytes += (*i)->GetPayloadSize();
  SessionManager::sm->message->Info("TimeWarp::InsertAdu - %u ADUs in TimeWarp-Queue with %u total bytes\n", twq.size(), bytes);
  #endif

  #endif  // #ifndef ENABLE_TW
}


void MLB::TimeWarp::SetCleanTimer() {
  #ifndef ENABLE_TW
  return;
  #else

  #ifdef DEBUG_TW
  SessionManager::sm->message->Info("TimeWarp::SetCleanTimer - runningTime %u, currentTime %u\n",
		 10000, rtpitime());
  #endif

  TimerManager::tm->SetTimer(this, new TimerObject(10000, TimerObject::TW_CLEAN));

  #endif  // #ifndef ENABLE_TW
}


void MLB::TimeWarp::ExecuteTimer(TimerObject* timer) {
  #ifndef ENABLE_TW
  return;
  #else

  #ifdef DEBUG_TW
  SessionManager::sm->message->Info("TimeWarp::ExecuteTimer - timer run out (runningTime %u, currentTime %u)\n",
		 timer->runningTime, rtpitime());
  #endif

  switch (timer->type) {
  case TimerObject::TW_CLEAN: {
    delete timer;
    RemoveOldAdus();
    SetCleanTimer();
    break;
  }
  }

  #endif  // #ifndef ENABLE_TW
}


void MLB::TimeWarp::RemoveOldAdus() {
  #ifndef ENABLE_TW
  return;
  #else

  RefCount<RTPI::RtpiAdu> adu;
  RefCount<RTPI::RtpiAduList> newAdus = new RTPI::RtpiAduList;
  GenericObject* obj = NULL;
  uint32 currentTime = rtpitime();

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);

  for (std::list<RefCount<RTPI::RtpiAdu> >::iterator i = twq.begin(); i != twq.end(); ++i) {
    adu = *i;
    if ((adu->GetAduHeader()->GetTimestamp() + 3*storageTime) < currentTime) {
      #ifdef DEBUG_TW
      SessionManager::sm->message->Info("TimeWarp::RemoveOldAdus - finally delete adu for object %u created at time %u\n",
		     adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp());
      #endif

      #ifdef DEBUG_CONS
      if (adu->GetType() == RTPI::RtpiHeader::PKT_STATE)
	SessionManager::sm->message->Info("TimeWarp::RemoveOldAdus - delete state for %u with ts %u\n", adu->GetSubcomponentID(),
		       adu->GetAduHeader()->GetTimestamp());
      #endif

      i = twq.erase(i);
      //delete *i not needed -> refcount
    } else
    if (((adu->GetAduHeader()->GetTimestamp() + storageTime) < currentTime)
	&& !StateIsStored(adu->GetSubcomponentID(), currentTime-storageTime)) {
      if (adu->GetType() == RTPI::RtpiHeader::PKT_STATE) {
	//if (!StateIsStored(adu->GetSubcomponentID())) {
	// get new state and store in twq
	obj = SessionManager::sm->GetObject(adu->GetSubcomponentID());
	if (!obj) obj = SessionManager::sm->GetGroup(adu->GetSubcomponentID());
	if (obj) {
          #ifdef DEBUG_TW
	  SessionManager::sm->message->Info("TimeWarp::RemoveOldAdus - get new state for object %u created at time %u\n",
			 adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp());
          #endif
      	  newAdus->push_back(obj->GetStateAdu(rtpitime(), 3));
	} // else: obj has been deleted in the meantime
	//} // else: we have a newer state for this object left in the timewarp queue
      }
    }
  }

  #ifdef DEBUG_TW
  SessionManager::sm->message->Info("TimeWarp::RemoveOldAdus - insert %u new state adus\n", newAdus->size());
  #endif

  for (RTPI::RtpiAduList::iterator j = newAdus->begin(); j != newAdus->end(); ++j) {
    if (!(*j)) continue;
    InsertAdu(*j);
  }

  SessionManager::sm->cursor->Reset();

  #endif  // #ifndef ENABLE_TW
}


bool MLB::TimeWarp::StateIsStored(uint32 id, uint32 oldestTime) {
  for (std::list<RefCount<RTPI::RtpiAdu> >::reverse_iterator i = twq.rbegin(); i != twq.rend(); ++i) {
    if ((*i)->GetAduHeader()->GetTimestamp() < oldestTime) return false;
    if (((*i)->GetSubcomponentID() == id) && ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE)) return true;
  }
  return false;
}


bool MLB::TimeWarp::StateIsStored(uint32 id) {
  for (std::list<RefCount<RTPI::RtpiAdu> >::iterator i = twq.begin(); i != twq.end(); ++i) {
    if (((*i)->GetSubcomponentID() == id) && ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE))
      return true;
  }

  return false;
}


RefCount<RTPI::RtpiAduList> MLB::TimeWarp::CheckForIntentionViolation(RefCount<RTPI::RtpiAdu> adu, uint32 oldestTime) {
  RefCount<RTPI::RtpiAduList> icAdus = new RTPI::RtpiAduList;

  for (RTPI::RtpiAduList::reverse_iterator i = twq.rbegin(); i != twq.rend(); ++i) {
    if ((*i)->GetAduHeader()->GetTimestamp() < oldestTime) break;
    if (CommunicationManager::cm->IntentionViolation(adu, *i))
      icAdus->push_back(*i);
  }

  return icAdus;
}


RefCount<RTPI::RtpiAduList> MLB::TimeWarp::GetPossiblyConflictingAdus(uint32 ts) {
  RefCount<RTPI::RtpiAduList> cAdus = new RTPI::RtpiAduList;

  for (std::list<RefCount<RTPI::RtpiAdu> >::reverse_iterator i = twq.rbegin(); i != twq.rend(); ++i) {
    if (ts < (*i)->GetAduHeader()->GetTimestamp())
      cAdus->push_back(*i);
    else break; // finished;
  }

  return cAdus;
}


RefCount<RTPI::RtpiAduList> MLB::TimeWarp::GetRestoreConsistencyAduList(uint32 triggerID, std::list<uint32>* ids, uint32 ts) {
  if (ids->empty()) return NULL;

  RefCount<RTPI::RtpiAduList> l;
  std::list<uint32>::iterator i = ids->begin();

  RefCount<RTPI::RtpiAduList> rAdus = (*i == triggerID) ? GetRestoreConsistencyAduList(*i, ts, true) : GetRestoreConsistencyAduList(*i, ts);
  if (!rAdus) return NULL;
  ++i;

  while (i != ids->end()) {
    l = (*i == triggerID) ? GetRestoreConsistencyAduList(*i, ts, true) : GetRestoreConsistencyAduList(*i, ts);
    if (!l) return NULL;
    rAdus = MergeLists(rAdus, l);
    ++i;
  }

  #ifdef DEBUG_CONS
  {
  SessionManager::sm->message->Info("GetRestoreConsistencyList - complete restore list\n");
  for (RTPI::RtpiAduList::reverse_iterator i = rAdus->rbegin(); i != rAdus->rend(); ++i) {
    SessionManager::sm->message->Info("type %u ts%u pid%u sn%u\n",(*i)->GetType(),
		   (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetParticipantID(), (*i)->GetPacketHeader()->GetSequenceNumber());
  }
  }
  #endif

  return rAdus;
}


RefCount<RTPI::RtpiAduList> MLB::TimeWarp::GetRestoreConsistencyAduList(uint32 id, uint32 ts, bool trigger) {
  if (twq.empty()) return NULL;

  RefCount<RTPI::RtpiAduList> rAdus = new RTPI::RtpiAduList;

  #ifdef DEBUG_CONS
  SessionManager::sm->message->Info("TimeWarp::GetResoreConsistencyAduList - search for state for object %u earlier than %u\n",id,ts);
  for (std::list<RefCount<RTPI::RtpiAdu> >::reverse_iterator i = twq.rbegin(); i != twq.rend(); ++i) {
    if (((*i)->GetType() ==  RTPI::RtpiHeader::PKT_STATE) && ((*i)->GetSubcomponentID() == id))
      SessionManager::sm->message->Info("found state with ts %u\n",(*i)->GetAduHeader()->GetTimestamp());
  }
  #endif

  // 1. extract most recent state of object before late ADU
  std::list<RefCount<RTPI::RtpiAdu> >::iterator pos = twq.end();
  --pos;

  while (true) {
    if (trigger &&
	((*pos)->GetType() == RTPI::RtpiHeader::PKT_STATE) && ((*pos)->GetSubcomponentID() == id) && ((*pos)->GetAduHeader()->GetTimestamp() > ts)) {
      // state is invalid
      #ifdef DEBUG_CONS
      SessionManager::sm->message->Info("TimeWarp::GetRestoreConsistencyAduList - delete state for %u with ts %u\n",
		     id, (*pos)->GetAduHeader()->GetTimestamp());
      #endif

      pos = twq.erase(pos);
      --pos;
    }
    if (((*pos)->GetType() == RTPI::RtpiHeader::PKT_STATE) && ((*pos)->GetSubcomponentID() == id) && ((*pos)->GetAduHeader()->GetTimestamp() <= ts)) {
      // found most recent state before the late ADU
      rAdus->push_front(*pos);
      ((RTPI::RtpiState*)&(**pos))->GetAduHeader()->SetPriority(3);
      break;
    }

    if (pos == twq.begin()) break;
    --pos;
  }

  if (rAdus->empty()) {
    // there is no state BEFORE the latecoming adu -> get first state AFTER (obj was created afterwards)
    pos = twq.end(); --pos;
    while (true) {
      if ((*pos)->GetAduHeader()->GetTimestamp() < ts) break; // did not find any state
      if (((*pos)->GetType() == RTPI::RtpiHeader::PKT_STATE) && ((*pos)->GetSubcomponentID() == id)) {
	rAdus->push_front(*pos);
	((RTPI::RtpiState*)&(**pos))->GetAduHeader()->SetPriority(3);
	break;
      }
     if (pos == twq.begin()) break;
     --pos;
    }
  }

  if (rAdus->empty()) {
    SessionManager::sm->message->Error("GetRestoreConsistencyList - did not find state for object%u\n",id);
    return NULL; // error, did not find the necessary state to start with
  }

  // 2. extract following events
  ++pos;
  while (true) {
    if (pos == twq.end()) break;

    if (((*pos)->GetType() == RTPI::RtpiHeader::PKT_EVENT) && ((*pos)->GetSubcomponentID() == id)) {
      // found another event
      rAdus->push_front(*pos);
    }

    ++pos;
  }

  if (rAdus->size() == 0) return NULL; // error

  #ifdef DEBUG_CONS
  SessionManager::sm->message->Info("GetRestoreConsistencyList - list for object%u\n",id);
  for (RTPI::RtpiAduList::reverse_iterator i = rAdus->rbegin(); i != rAdus->rend(); ++i) {
    if ((*i)->GetType() == RTPI::RtpiHeader::PKT_EVENT)
      SessionManager::sm->message->Info("event ts %u pid %u sn %u\n",
		     (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetParticipantID(), (*i)->GetPacketHeader()->GetSequenceNumber());
    else
      if (((RTPI::RtpiStateHeader*)(*i)->GetAduHeader())->sns.size() != 0)
	SessionManager::sm->message->Info("state ts %u pid %u sn %u highest sn %u\n",
		       (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetParticipantID(), (*i)->GetPacketHeader()->GetSequenceNumber(),
		       ((RTPI::RtpiStateHeader*)(*i)->GetAduHeader())->sns[0]);
      else
	SessionManager::sm->message->Info("state ts %u pid %u sn %u\n",
		       (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetParticipantID(), (*i)->GetPacketHeader()->GetSequenceNumber());
  }
  #endif

  return rAdus;
}


RefCount<RTPI::RtpiAduList> MLB::TimeWarp::MergeLists(RefCount<RTPI::RtpiAduList> list1, RefCount<RTPI::RtpiAduList> list2) {
  RTPI::RtpiAduList::reverse_iterator i = list1->rbegin();
  RTPI::RtpiAduList::reverse_iterator j = list2->rbegin();
  RefCount<RTPI::RtpiAduList> ret = new RTPI::RtpiAduList;

  while (true) {
    if (i == list1->rend()) {
      for (; j != list2->rend(); ++j) ret->push_front(*j);
      break;
    }

    if (j == list2->rend()) {
      for (; i != list1->rend(); ++i) ret->push_front(*i);
      break;
    }

    if (((*i)->GetAduHeader()->GetTimestamp() < (*j)->GetAduHeader()->GetTimestamp())
	|| (((*i)->GetAduHeader()->GetTimestamp() == (*j)->GetAduHeader()->GetTimestamp()) && ((*i)->GetParticipantID() < (*j)->GetParticipantID()))) {
      ret->push_front(*i);
      ++i;
    } else {
      ret->push_front(*j);
      ++j;
    }
  }

  /*
  while (j != list2->rend()) {
    if (i == list1->rend()) {
      for (; j != list2->rend(); ++j) list1->push_back(*j);
      return list1;
    }

    while ((i != list1->rend()) && ((*j)->GetAduHeader()->GetTimestamp() > (*i)->GetAduHeader()->GetTimestamp()))
      ++i;

    while ((i != list1->rend()) && ((*j)->GetAduHeader()->GetTimestamp() == (*i)->GetAduHeader()->GetTimestamp())
	   && ((*j)->GetParticipantID() > (*i)->GetParticipantID()))
      ++i;

    while ((i != list1->rend()) && ((*j)->GetAduHeader()->GetTimestamp() == (*i)->GetAduHeader()->GetTimestamp())
	   && ((*j)->GetParticipantID() == (*i)->GetParticipantID()))
      ++i;

    if (i == list1->rend()) list1->push_back(*j);
    else list1->insert(i, *j);
    ++j;
  }

  return list1;
  */

  return ret;
}


bool MLB::AduManager::StateReceived(uint32 id) {
  for (RTPI::RtpiAduList::iterator i = aq.begin(); i != aq.end(); ++i) {
    if (((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE) && ((*i)->GetSubcomponentID() == id)) return true;
  }
  return false;
}
