/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_adu.h 
 authors : Juergen Vogel
 date    : 
 desc    : 

***********************************************************************************************************/


#ifndef __MLB_ADU_H__
#define __MLB_ADU_H__

#include <rtpi/rtpiPacket.h>
#include <util/types.h>
#include <util/Buffer.h>
#include <tcl.h>
#include <tk.h>
#include <time.h>
#include <stdlib.h>

#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <stdio.h>
#include <string>
#include <list>

#include "mlb_timermanager.h"
#include "mlb_page.h"


namespace MLB {
  class AduManager : public TimerFeedbackInterface {
  public:
    AduManager() : timerRunning(false), localLag(100), participantHints(false) {}
    ~AduManager() {}

    void Reset();

    void   SetLocalLag(uint32 lag) {localLag = lag;}
    uint32 GetLocalLag() {return localLag;}

    void SetParticipantHints(bool hint) {participantHints = hint;}
    bool GetParticipantHints() {return participantHints;}  

    // manager adu queue
    void InsertAdu(RefCount<RTPI::RtpiAdu> adu);
    void InsertAdus(RefCount<RTPI::RtpiAduVector> adus);

    // check adu for a possible intention violation
    RefCount<RTPI::RtpiAduList> CheckForIntentionViolation(RefCount<RTPI::RtpiAdu> adu);

    // handle both local and remote ADUs
    void HandleState(RefCount<RTPI::RtpiAdu> adu);
    void HandleEvent(RefCount<RTPI::RtpiAdu> adu);
    void HandleCue(RefCount<RTPI::RtpiAdu> adu);

    bool StateReceived(uint32 id); // state for obj with id in adu-queue?

    static AduManager* am;

  private:
    Page* CreateDefaultPage(uint32 id);
    void  ActivatePage(Page* page, bool redraw = false);

    bool HandleNewDrawObject(Page* page, DrawObject* obj, bool shared, int type, int priority, uint32 participantID);

    void _InsertAdu(RefCount<RTPI::RtpiAdu> adu);
    void _InsertAdus(RefCount<RTPI::RtpiAduVector> adus);
    void CheckForDueAdus();  // check for adus to be executed

    void SetExecutionTimer();
    void ExecuteTimer(TimerObject* timer);

    bool timerRunning;

    uint32 localLag;
 
    RTPI::RtpiAduList aq; // adu queue with both remote and local adus
    void PrintAduQueue();

    void DisplayParticipantHint(Page* parent, uint32 pid, DrawObject* obj);
    bool participantHints;
  };

    
  class TimeWarp : public TimerFeedbackInterface {
  public:
    TimeWarp() : storageTime(1800000) {SetCleanTimer();}
    ~TimeWarp() {}

    void Reset();

    void   SetStorageTime(uint32 time) {storageTime = time;}
    uint32 GetStorageTime() {return storageTime;}

    void   SetStateSaveTime(uint32 time) {stateSaveTime = time;}
    uint32 GetStateSaveTime() {return stateSaveTime;}

    void InsertAdu(RefCount<RTPI::RtpiAdu> adu);
    void InsertAdus(RefCount<RTPI::RtpiAduVector> adus);

    // check adu for a possible intention violation
    RefCount<RTPI::RtpiAduList> CheckForIntentionViolation(RefCount<RTPI::RtpiAdu> adu, uint32 ts);

    RefCount<RTPI::RtpiAduList> GetPossiblyConflictingAdus(uint32 ts);

    RefCount<RTPI::RtpiAduList> GetRestoreConsistencyAduList(uint32 triggerID, std::list<uint32>* ids, uint32 ts);

    static TimeWarp* tw;

  private:
    void SetCleanTimer();
    void ExecuteTimer(TimerObject* timer);
    void RemoveOldAdus();
    bool StateIsStored(uint32 id);
    bool StateIsStored(uint32 id, uint32 oldestTime);
    RefCount<RTPI::RtpiAduList> GetRestoreConsistencyAduList(uint32 id, uint32 ts, bool trigger = false);
    RefCount<RTPI::RtpiAduList> MergeLists(RefCount<RTPI::RtpiAduList> list1, RefCount<RTPI::RtpiAduList> list2);

    uint32 storageTime;            // length of timewarp queue in millisecs
    uint32 stateSaveTime;          // frequency of state saves in millisecs
    RTPI::RtpiAduList twq; // timewarp queue
  };
}

#endif

