/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_animations.cpp
 authors : Andrius Kurtinaitis
 date    : 07.08.2001
 desc    : Presentation animations implementation.

 *********************************************************************************************************/

#include "mlb_objects.h"
#include "mlb_page.h"
#include "mlb_adu.h"
#include "mlb_session.h"
#include "mlb_animations.h"

//-------------------------------------------------------------------------//

/**   @return the step with given name or EndStep().
  **/
MLB::Animations::step_iterator MLB::Animations::GetStep( const string& name )
{
  MLB::Animations::step_iterator stepIt=BeginStep();
  for(; stepIt!=EndStep() && stepIt->GetName()!=name; ++stepIt );
  return stepIt;
} // GetStep()

//-------------------------------------------------------------------------//

/**   Runs all animation steps.
  **/
void MLB::Animations::RunFirstStep( bool run )
{
  bool changed = false;

  // run
  step_iterator iStep=BeginStep();
  if( iStep==EndStep() ) return;
  if( iStep->IsRun() != run ) changed = iStep->Run(run);

  // redraw
  if( changed &&
      SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE &&
      SessionManager::sm->GetActivePage() == &_page
    ) _page.ReDraw();

} // RunFirstStep()

//-------------------------------------------------------------------------//

/**   Resets all animation steps so that all original objects are visible.
  **/
void MLB::Animations::Reset()
{
  bool changed = false;

  // run
  for( step_iterator i=BeginStep(); i!=EndStep(); ++i )
    if( i->Reset() ) changed = true;

  // redraw
  if( changed &&
      SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE &&
      SessionManager::sm->GetActivePage()                      == &_page
    ) _page.ReDraw();

} // Reset()

//-------------------------------------------------------------------------//

/**   Runs all animation steps.
  **/
void MLB::Animations::RunAllSteps( bool run )
{
  bool changed = false;

  // run
  for( step_iterator i=BeginStep(); i!=EndStep(); ++i )
    if( i->IsRun() != run ) { changed = i->Run(run) || changed; }

  // redraw
  if( changed &&
      SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE &&
      SessionManager::sm->GetActivePage() == &_page
    ) _page.ReDraw();

} // RunAllSteps()

//-------------------------------------------------------------------------//

/**   Runs the given step.
  **/
bool MLB::Animations::RunStep( const string& stepName, bool run )
{
  bool changed = false;

  // run the found step
  step_iterator i=GetStep( stepName );
  if( i!=EndStep() && i->IsRun() != run ) { changed = i->Run( run ) || changed; }

  // redraw
  if( changed &&
      SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE  &&
      SessionManager::sm->GetActivePage() == &_page
    ) _page.ReDraw();

  
  if (i !=EndStep()) return true;
  else return false;

} // RunStep()

//-------------------------------------------------------------------------//

void MLB::Animations::RemoveObject( const string& stepName, const string& objectName )
{
  step_iterator gr = GetStep( stepName );
  if( EndStep()==gr ) return;
  gr->RemoveObject( objectName );
} // RemoveObject()

//-------------------------------------------------------------------------//

/**   @return name of animation step for this object
  *           or empty string if no such object in animations
  **/
string MLB::Animations::GetStepNameForObject( const string& objName )
{
  // show objects
  step_iterator i;
  for( i=BeginStep(); i!=EndStep(); ++i )
    if( i->GetObject(objName, true) != i->EndObject(true) ) return i->GetName();

  // hide objects
  for( i=BeginStep(); i!=EndStep(); ++i )
    if( i->GetObject(objName, false) != i->EndObject(false) ) return i->GetName();

  return "";
} // GetStepNameForObject()

//-------------------------------------------------------------------------//

/**   @return space-delimited list of names of existing animation steps **/
string MLB::Animations::GetStepNames()
{
  string ret="";
  for( step_iterator i=BeginStep(); i!=EndStep(); ++i ) ret+=i->GetName()+" ";
  return ret;
} // GetStepNames()

//-------------------------------------------------------------------------//

/**   @return space-delimited list of names of already run animation steps **/
string MLB::Animations::GetRunStepNames()
{
  string ret="";
  for( step_iterator i=BeginStep(); i!=EndStep(); ++i )
    if( i->IsRun() ) ret+=i->GetName()+" ";
  return ret;
} // GetRunStepNames()

//-------------------------------------------------------------------------//

/**   @return   true if some step has objects in the given sequence **/
bool MLB::Animations::HasObjectsIn( const DrawObject::object_list& objs )
{
/*
  for( step_iterator i=BeginStep(); i!=EndStep(); ++i )
    if( std::find_first_of( i->BeginObject(), i->EndObject(), objs.begin(), objs.end() ) != i->EndObject() ) return true;
  return false;
  */
  return false;
} // HasObjectsIn()

//-------------------------------------------------------------------------//

/**   @return space-delimited list of object names of given animation step **/
string MLB::Animations::GetObjectNames( const string& stepName, bool show )
{
  // find the step we need
  step_iterator stepIt=GetStep(stepName);
  if( stepIt==EndStep() ) return "";
  return stepIt->GetObjectNames(show);
} // GetObjectNames()

//-------------------------------------------------------------------------//

string MLB::Animations::GetObjectTitle( const string& stepName, const string& objectName )
{
  // find the step we need
  step_iterator stepIt=GetStep(stepName);
  if( stepIt==EndStep() ) return "";
  return stepIt->GetObjectTitle(objectName);
} // GetObjectTitle()

//-------------------------------------------------------------------------//

/**   @return title for step with the given name  **/
string MLB::Animations::GetStepTitle( const string& stepName )
{
  // find the step we need
  step_iterator stepIt=GetStep(stepName);
  if( stepIt==EndStep() ) return "";
  else return stepIt->GetTitle();
} // GetStepTitle()

//-------------------------------------------------------------------------//

void MLB::Animations::SetStepTitle( const string& stepName, const string& title )
{
  // find the step we need
  step_iterator stepIt=GetStep( stepName );
  if( stepIt==EndStep() ) return;
  else stepIt->SetTitle( title );
} // SetStepTitle()

//-------------------------------------------------------------------------//

/** @return   true if step is displayed as open tree folder, false otherwise **/
bool MLB::Animations::IsStepOpen( const string& stepName )
{
  // find the step we need
  step_iterator stepIt=GetStep( stepName );
  if( stepIt==EndStep() ) return false;
  else return stepIt->IsOpen();
} // IsStepOpen()

//-------------------------------------------------------------------------//

/** @return   true if step is displayed as open tree folder, false otherwise **/
void MLB::Animations::SetStepOpen( const string& stepName, bool open )
{
  // find the step we need
  step_iterator stepIt=GetStep( stepName );
  if( stepIt==EndStep() ) return;
  else stepIt->SetOpen( open );
} // SetStepOpen()

//-------------------------------------------------------------------------//

void MLB::Animations::MoveStep( const string& stepName, int position )
{
  if( position < 0 ) return;

  // find the step we need
  step_iterator step=GetStep( stepName );
  if( step==EndStep() ) return;

  // find the place to insert
  step_iterator stepPos=BeginStep();
  for( int pos=0; stepPos!=EndStep() && pos<position; ++stepPos, ++pos );
  //if( stepPos==EndStep() ) return;

  // replace
  _steps.insert( stepPos, *step );
  _steps.erase ( step );

} // MoveStep()

//-------------------------------------------------------------------------//

/**   Adds all selected objects to the step with given name.
  *   @return space delimited names of the added objects.
  **/
string MLB::Animations::AddSelectedToAnimationStep( const string& stepName, bool show )
{
  // return val
  string ret="";

  // find the step we need
  step_iterator stepIt=GetStep( stepName );
  if( stepIt==EndStep() ) return ret;

  // add all selected objects
  DrawObject::object_list selectedObjects = _page.GetSelectedObjects();
  for( DrawObject::object_iterator objIter = selectedObjects.begin(); objIter!=selectedObjects.end(); ++objIter )
  {
    if( !((*objIter)->Annotation()) && !stepIt->HasObject(*objIter) )
    {
      stepIt->AddObject( *objIter, show );
      ret += stepIt->GetObjectName(*objIter);
      ret += " ";
    }
  }
  return ret;

} // AddSelectedToAnimationStep()

//-------------------------------------------------------------------------//

void MLB::Animations::SelectAnimationStepObjects( const string& stepName )
{
  _page.UnSelectObjects();

  // find the step we need
  step_iterator stepIt=GetStep( stepName );
  if( stepIt==EndStep() ) return;

  // select all show objects in this step
  DrawObject::object_iterator i;
  for( i=stepIt->BeginObject(true); i!=stepIt->EndObject(true); ++i )
  {
    DrawObject* obj = (*i)->Select();
    _page.SelectObject(obj);
  } // for

  // select all hide objects in this step
  for( i=stepIt->BeginObject(false); i!=stepIt->EndObject(false); ++i )
  {
    DrawObject* obj = (*i)->Select();
    _page.SelectObject(obj);
  } // for

  _page.DrawSelected();

} // SelectAnimationStepObjects()

//-------------------------------------------------------------------------//

void MLB::Animations::SelectAnimationObject( const string& tkRef )
{
  _page.UnSelectObjects();

  DrawObject* obj = _page.GetObject(tkRef);
  if( !obj ) return;

  // draw selection
  obj = obj->Select();
  if( obj ) {
    _page.SelectObject(obj);
    _page.DrawSelected();
  }

} // SelectAnimationObject()

//-------------------------------------------------------------------------//

void MLB::Animations::Save( FILE *_fp, int _level )
{
  std::string tab;
  for( int k=0; k<_level; ++k ) tab += "  ";

  // Animations
  if( BeginStep()!=EndStep() )
  {
    fprintf(_fp,"%s<ANIMATIONS>\n", (char*)tab.c_str() );
    for( Animations::step_iterator i=BeginStep(); i!=EndStep(); ++i ) i->Save( _fp, _level+1 );
    fprintf(_fp,"%s</ANIMATIONS>\n", (char*)tab.c_str() );
  } // if any animations

} // Save()

//-------------------------------------------------------------------------//

/**   Loads animations for a page.
  *   @param  _root sgml element ANIMATIONS
  *   @return       multimap: for each object name - list of animation steps where it belongs.
  **/
/*
StringStepMap MLB::Animations::Load( SGMLElement *_root )
{
  StringStepMap externalNaming;
  SGMLElement*  hilfSonSon;
  SGMLElement*  hilfSon;

  for( hilfSon = hilf->son; hilfSon != NULL; hilfSon = hilfSon->next )
  {
    char* tmp = hilfSon->GetAttribute( "name" );
    if( hilfSon->Name == "STEP" && tmp )
    {
      // name
      string stepName( tmp );
      AnimationStep* newStep = &(*(animations.AddNewStep()));

      // title
      tmp = hilfSon->GetAttribute( "title" );
      if( tmp ) newStep->SetTitle( tmp );

      // insert
      externalNaming.insert( StringStepMap::value_type( stepName, newStep ) );
    }
  } // steps
} // Load()
*/
//=========================================================================//



void MLB::Animations::DeleteObject(DrawObject* obj) { 
  // delete object from all steps
  for (step_iterator i = _steps.begin(); i != _steps.end(); ++i ) 
    i->DeleteObject(obj); 

  // update gui if active page
  if (SessionManager::sm->GetActivePage() && (_page.GetID() == SessionManager::sm->GetActivePage()->GetID()))
    Et_GlobalEvalF(Et_Interp,"Animations::update");
}


 
