/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_animationstep.cpp
 authors : Andrius Kurtinaitis
 date    : 17.09.2001
 desc    : Presentation animation step implementation.

**********************************************************************************************************/

#include "mlb_objects.h"
#include "mlb_page.h"
#include "mlb_adu.h"
#include "mlb_session.h"
#include "mlb_animationstep.h"

int MLB::AnimationStep::_count = 0;

//-------------------------------------------------------------------------//

/**   Runs the step the given direction.
  *   @return true if some objects were set visible and need redraw, false otherwise
  **/
bool MLB::AnimationStep::Run( bool direction )
{
  bool redraw = false;

  // shared
  if (SessionManager::sm->GetParentSession(&_page)->GetSessionType() == Session::SHARED )
  {
    RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

    // gather adus
    AppendVisibilityAdus( adus, true,  direction );
    AppendVisibilityAdus( adus, false, direction );

    // send adus
    AduManager::am->InsertAdus(adus);
    CommunicationManager::cm->SendAdusReliable(adus);
  }
  else // private or annotation
  {
    // show objects
    DrawObject::object_iterator i;
    for( i=BeginObject(true); i!=EndObject(true); ++i )
      redraw = SetLocalObjectVisible( *i, direction ) || redraw;

    // hide objects
    for( i=BeginObject(false); i!=EndObject(false); ++i )
      redraw = SetLocalObjectVisible( *i, !direction ) || redraw;

  } // else - private

  _run = direction;

  if (MLB::SessionManager::sm->recording && !MLB::SessionManager::sm->paused) {
    // write step in evq and obj file
    // GUI::MessageWindow::Debug("MLB::AnimationStep::Run\n");
    FILE *objFp, *evqFp;
    int time, pageNumber;
    Page* currentPage;
    Color background;
    Telepointer* currentPointer;

#ifdef WIN32
    time = MLB::SessionManager::sm->aofRecording->GetRecordTime();
#else
    time = MLB::SessionManager::sm->aofRecording->GetTime();
#endif
    currentPage = MLB::SessionManager::sm->GetActivePage();
    pageNumber = currentPage->GetPageNumber();
    objFp = MLB::SessionManager::sm->GetObjFp();
    evqFp = MLB::SessionManager::sm->GetEvqFp();
    background = currentPage->GetColor();
    fprintf(evqFp,"%d %d %d %d",time,pageNumber,CalculateColor(background),currentPage->RecordableObjects());
    currentPage->WriteObjectFile();

    currentPointer=currentPage->GetLocalTelepointer();
    if (currentPointer->IsActive()) {
      currentPointer->SetUnrecorded();
      currentPointer->SaveAsAOF(objFp);
      fprintf(evqFp," %d",currentPointer->GetRecordID());
      //GUI::MessageWindow::Debug("RecordID vom OnlinePointer %d\n",currentPointer->GetRecordID());
    }
    
    fprintf(evqFp,"\n");
    fflush(objFp);
    fflush(evqFp);
  }

  return redraw;
} // RunStep()

//-------------------------------------------------------------------------//

/**   Resets the step. It means, stops animation and shows the objects as they were.
  *   @return true if some objects were set visible and need redraw, false otherwise
  **/
bool MLB::AnimationStep::Reset()
{
  bool redraw = false;

  // shared
  if (SessionManager::sm->GetParentSession(&_page)->GetSessionType() == Session::SHARED )
  {
    RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

    // gather adus
    AppendVisibilityAdus( adus, true,  true   );
    AppendVisibilityAdus( adus, false, false  );

    // send adus
    AduManager::am->InsertAdus(adus);
    CommunicationManager::cm->SendAdusReliable(adus);
  }
  else // private or annotation
  {
    // show objects
    DrawObject::object_iterator i;
    for( i=BeginObject(true); i!=EndObject(true); ++i )
      redraw = SetLocalObjectVisible( *i, true ) || redraw;

    // hide objects
    for( i=BeginObject(false); i!=EndObject(false); ++i )
      redraw = SetLocalObjectVisible( *i, true ) || redraw;

  } // else - private

  _run = false;
  return redraw;

} // Reset()

//-------------------------------------------------------------------------//

void MLB::AnimationStep::AppendVisibilityAdus( RefCount<RTPI::RtpiAduVector>& adus, bool visible, bool direction )
{
  RefCount<Buffer> buf;
  for( DrawObject::object_iterator i=BeginObject(visible); i!=EndObject(visible); ++i )
  {
    if( (*i)->Annotation() ) continue;
    buf = new Buffer( (*i)->GetChangeVisibilitySize() );

    if( (*i)->IsVisible() != !(visible^direction) ) {
      if( (*i)->GetType() == GenericObject::GROUP ) ((Group*)*i)->FlushChangeVisibilityTG( buf->getCont(), !(visible^direction) );
      else (*i)->FlushChangeVisibility  ( buf->getCont(), !(visible^direction) );
      adus->push_back( CommunicationManager::cm->CreateEvent((*i)->GetID(), buf) );
    }
  } // for

} // AppendVisibilityAdus()

//-------------------------------------------------------------------------//

bool MLB::AnimationStep::SetLocalObjectVisible( DrawObject* obj, bool visible )
{
  if( obj->Annotation() ) return false;
  if( visible ) { if( !obj->IsVisible() ) { obj->SetVisible(); return true; } }
  else { obj->Hide( _page.GetCanvas() ); return false; }

} // SetLocalObjectVisible()

//-------------------------------------------------------------------------//

//  void MLB::AnimationStep::Update()
//  {
//    // show objects
//    int j;
//    for( j=0; j<_showObjects.size(); ++j )
//      for( DrawObject::object_iterator i=BeginObject(true); i!=EndObject(true); ++i )
//        if( (*i)->IsDeleted() ) { _showObjects.erase( i ); break; }

//    // hide objects
//    for( j=0; j<_hideObjects.size(); ++j )
//      for( DrawObject::object_iterator i=BeginObject(false); i!=EndObject(false); ++i )
//        if( (*i)->IsDeleted() ) { _hideObjects.erase( i ); break; }

//  } // Update()


void MLB::AnimationStep::DeleteObject(DrawObject* obj) {
  // show objects
  int j;
  for( j=0; j<_showObjects.size(); ++j )
    for( DrawObject::object_iterator i=BeginObject(true); i!=EndObject(true); ++i )
      if( *i == obj ) { _showObjects.erase( i ); break; }

  // hide objects
  for( j=0; j<_hideObjects.size(); ++j )
    for( DrawObject::object_iterator i=BeginObject(false); i!=EndObject(false); ++i )
      if( *i == obj ) { _hideObjects.erase( i ); break; }

}



//-------------------------------------------------------------------------//

MLB::DrawObject::object_iterator MLB::AnimationStep::GetObject( const string& name )
{
  DrawObject::object_iterator i = GetObject( name, true );
  if( EndObject(true)==i )    i = GetObject( name, false );
  return i;
} // GetObject()

//-------------------------------------------------------------------------//

MLB::DrawObject::object_iterator MLB::AnimationStep::GetObject( const string& name, bool show )
{
  DrawObject::object_iterator i = BeginObject(show);
  for(; i!=EndObject(show) && GetObjectName(*i)!=name; ++i );
  return i;
} // GetObject()

//-------------------------------------------------------------------------//

void MLB::AnimationStep::RemoveObject( const string& name )
{
  // show objects
  DrawObject::object_iterator i = GetObject( name, true );
  if( i != EndObject(true) ) { /*(*i)->SetAnimationStepName(""); */ _showObjects.erase( i ); }

  // hide objects
  i = GetObject( name, false );
  if( i != EndObject(false) ) { /*(*i)->SetAnimationStepName("");*/ _hideObjects.erase( i ); }

} // RemoveObject()

//-------------------------------------------------------------------------//

/**   @return space-delimited list of object names of this step **/
string MLB::AnimationStep::GetObjectNames( bool show )
{
  string ret="";
  for( DrawObject::object_iterator i=BeginObject(show); i!=EndObject(show); ++i )
    ret += GetObjectName(*i) + string(" ");
  return ret;
} // GetObjectNames()

//-------------------------------------------------------------------------//

/**   @return title for the object with the given name **/
string MLB::AnimationStep::GetObjectTitle( const string& objectName )
{
  // find the object
  DrawObject::object_iterator i=GetObject(objectName);
  if( i==EndObject(true) || i==EndObject(false) ) return string("");
  return GetObjectTitle(*i);
} // GetObjectTitles()

//-------------------------------------------------------------------------//

void MLB::AnimationStep::Save( FILE *_fp, int _level )
{

  std::string tab;
  for( int k=0; k<_level; ++k ) tab += "  ";

  // begin step
  fprintf( _fp, "%s<STEP name=\"%s\" title=\"%s\">\n", tab.c_str(), GetName().c_str(), GetTitle().c_str() );

  // show objects
  DrawObject::object_iterator i;
  for( i=BeginObject(true); i!=EndObject(true); ++i )
    fprintf( _fp, "%s  <OBJECT name=\"%s\" onAnimate=\"show\"></OBJECT>\n", tab.c_str(), (*i)->GetTkRef() );

  // hide objects
  for( i=BeginObject(false); i!=EndObject(false); ++i )
    fprintf( _fp, "%s  <OBJECT name=\"%s\" onAnimate=\"hide\"></OBJECT>\n", tab.c_str(), (*i)->GetTkRef() );

  // end step
  fprintf( _fp, "%s</STEP>\n", tab.c_str() );

} // Save()

//=========================================================================//
