/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_char.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 3.12.1999 last changes 14.03.2001
 desc    : Functions to draw and manipulate characters (created by TrueType-Fonts)

 *********************************************************************************************************/

#include "freetype.h"

#include "mlb.h"
#include "mlb_objects.h"
#include "mlb_page.h"
#include "mlb_session.h"
#include "mlb_charfactory.h"


uint32 MLB::Char::count = 1;

MLB::Char::Char(MLB::Font *_font,char *_color,char _ascii) : GenericObject(0) {
  char buf[100];
  sprintf(buf,"%u",count);

  tkref = "Char";
  tkref += buf;

  ascii  = _ascii;
  color  = _color;
  drawed = false;
  font   = NULL;

  Init(_font);
  id = count;
  ++count;
}


MLB::Char::~Char() { 
  DeInit(); 
}


void MLB::Char::Delete() { 
  DeInit(); 
}


void MLB::Char::DeInit() {
  CharFactory::DeleteCharBitmap(font, color, ascii);
  tkref = "";
  charTag = "";
} 


void MLB::Char::Init(Font* _font) {
  char buf[100];

  if (!_font) return;
  font = _font;


  bool ok = CharFactory::GetCharBitmap(font, color, ascii, &width, &ascent, &descent, &lbearing, &bbox, &charTag);
  if (!ok) return;

  width    = double2int(width    / SessionManager::sm->GetZoom());
  ascent   = double2int(ascent   / SessionManager::sm->GetZoom());
  descent  = double2int(descent  / SessionManager::sm->GetZoom()); 
  lbearing = double2int(lbearing / SessionManager::sm->GetZoom());  
  height = ascent + descent;
}


void MLB::Char::Move(GUI::Canvas *_canvas, int _diffX, int _diffY, bool redraw) {
  x += _diffX;
  y += _diffY;

  if (redraw) _canvas->MoveObject(tkref.c_str(),SessionManager::sm->GetZoom()*_diffX,SessionManager::sm->GetZoom()*_diffY);
}


void MLB::Char::Draw(GUI::Canvas *_canvas, int _x, int _y) {
  x = _x;
  y = _y;

  if(!GUI::Image::Exists(charTag.c_str())) {
    Init(font);
  }
  _canvas->DrawImage(tkref.c_str(),
		     SessionManager::sm->GetZoom() * (x-lbearing),
		     SessionManager::sm->GetZoom() * y,
		     charTag.c_str());
  drawed = true;
}


void MLB::Char::DrawTextCursor(GUI::Canvas *_canvas, uint32 _id) {
  _canvas->DrawTextCursor(_id,
			  SessionManager::sm->GetZoom() * (x+width),
			  SessionManager::sm->GetZoom() * y,
			  0,
			  SessionManager::sm->GetZoom() * (ascent+descent), 
			  color.c_str());
}


void MLB::Char::ReDraw(GUI::Canvas *_canvas, int _x, int _y) {
  Hide(_canvas);
  Draw(_canvas,_x,_y);
  drawed = true;
}


void MLB::Char::Hide(GUI::Canvas *_canvas) {
  if (!drawed) return;
  _canvas->DeleteObject(tkref.c_str());
  drawed = false;
}


bool MLB::Char::Overlap(GUI::Canvas *_canvas,int _x, int _y, bool *_smaller){
  if ((_x < x) || (_y < y) || (_x > x+width) || (_y > y+height)) return false;

  if (_x < (x+(width/2)))
    *_smaller = true;
  else
    *_smaller = false;

  return true;
}


bool MLB::Char::IsEditablePoint(int _x, int _y, bool *_smaller) {
  if ((_x < x) || (_y < y) || (_x > x+width) || (_y > y+height)) return false;

  if (_x < (x+(width/2)))
    *_smaller = true;
  else
    *_smaller = false;

  return true;
}


void MLB::Char::ChangeFont(GUI::Canvas *_canvas, MLB::Font *_font) {
  _canvas->DeleteImage(charTag.c_str());
  CharFactory::DeleteCharBitmap(font, color, ascii);
  Init(_font);
}


void MLB::Char::ChangeColor(GUI::Canvas *_canvas,char *_color) {
  _canvas->DeleteImage(charTag.c_str());
  CharFactory::DeleteCharBitmap(font, color, ascii);
  color = _color;
  Init( font );
}

