/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_communication.cpp
 authors : Juergen Vogel
 date    : 
 desc    : 

**********************************************************************************************************/


#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif

#ifdef WIN32
#include <winsock.h>
#else
#include <arpa/inet.h>
#include <sys/types.h>
#include <unistd.h>
#endif

#include <math.h>
#include <memory.h>
#include <stdlib.h>
#include <time.h>

#include <rtpi/rtpiException.h>
#ifndef IDSERVER
#include <util/random.h>
#endif

#include <mlb_communication.h>
#include <mlb_adu.h>
#include <mlb_participants.h>
#include <mlb_objects.h>
#include <mlb_page.h>
#include <mlb_session.h>
#include <mlb.h>

//#define DEBUG 
//#define DEBUG_RTPI
//#define DEBUG_RTCPI
//#define DEBUG_RECORDING
//#define ENABLE_TW
//#define DEBUG_CONS
//#define INFO_CONS
#define SIMPLE_LJ

//
// class CommunicationManager
//
void handleReceivedRtpi(SMP::Indication ind) {
  MLB::CommunicationManager::cm->HandleReceivedRtpi(ind);
}

void handleReceivedRtcpi(SMP::Indication ind) {
  MLB::CommunicationManager::cm->HandleReceivedRtcpi(ind);
}

const unsigned int MLB::CommunicationManager::T = 500;
double MLB::CommunicationManager::cueSendRate = 0.1;
MLB::CommunicationManager* MLB::CommunicationManager::cm = NULL;


MLB::CommunicationManager::CommunicationManager()
  : openConnection(false), ip(""), port(0), ttl(0), bandwidth(300000), 
    name(""), email(""), phone(""), location(""), openSMP(false), smp(NULL), 
    rtpi(NULL), checkIntentionViolation(true) {

  ParticipantManager::pm = new ParticipantManager();

#ifdef WIN32
  SMP::SMP::StartupSockets();
#endif

  // 1. get local addresses
  struct hostent* host;
  struct in_addr a;
  char buf[100];
  if (gethostname(buf, sizeof(buf)) != 0) {
    GUI::MessageWindow::Error("CommunicationManager::CommunicationManager - gethostname failed\n");
    #ifdef WIN32
    GUI::MessageWindow::Error("%q\n", WSAGetLastError());
    #endif
    exit(0);
  }
  localhost = buf;
  host = gethostbyname(buf);
  for (int i = 0; host->h_addr_list[i] != NULL; ++i) {
    memcpy(&(a.s_addr), host->h_addr_list[i], host->h_length);
    localAddresses.push_back(ntohl(a.s_addr));
  }

  // 2. get local user name
  char user[50];
  #ifdef WIN32
  unsigned long size = sizeof(user);
  GetUserName(user, &size);
  #else
  struct passwd* pwd = getpwuid(getuid());
  if (pwd == NULL) strcpy (user, "unknown");
  else strcpy (user, pwd->pw_name);
  #endif

  username = user;

  // 3. init id server
  #ifdef IDSERVER
  openIDServer = false;
  uniqueIdClient = NULL;
  localIDServer = false;
  #else
  openIDServer = true;
  #endif
}


MLB::CommunicationManager::~CommunicationManager() {
  CloseRTPI();
  #ifdef IDSERVER
  CloseIDServer();
  #endif
  CloseSMP();

  if (smp) delete smp;
  delete ParticipantManager::pm;

#ifdef WIN32
  WSACleanup();
#endif
}

void MLB::CommunicationManager::Shutdown() {
  CloseRTPI();
  #ifdef IDSERVER
  CloseIDServer();
  #endif
  CloseSMP();
}


void MLB::CommunicationManager::SetParticipantName(char* _name)  {
  if (_name) name = _name; 
  else name = "";

  if (!rtpi) return;

  RTPI::RtpiParticipant* local = rtpi->get_own_participant();
  ParticipantManager::pm->UpdateParticipant(local->get_pid(), local->get_cname(), local->get_name(), 
					    local->get_email(), local->get_phone(), local->get_loc(), local->get_tool());
}


void MLB::CommunicationManager::SetParticipantEmail(char* _email) {
  if (_email) email = _email; 
  else email = "";

  if (!rtpi) return;

  RTPI::RtpiParticipant* local = rtpi->get_own_participant();
  ParticipantManager::pm->UpdateParticipant(local->get_pid(), local->get_cname(), local->get_name(), 
					    local->get_email(), local->get_phone(), local->get_loc(), local->get_tool());
}


void MLB::CommunicationManager::SetParticipantPhone(char* _phone) {
  if (_phone) phone = _phone; 
  else phone = "";

  if (!rtpi) return;

  RTPI::RtpiParticipant* local = rtpi->get_own_participant();
  ParticipantManager::pm->UpdateParticipant(local->get_pid(), local->get_cname(), local->get_name(), 
					    local->get_email(), local->get_phone(), local->get_loc(), local->get_tool());
}


void MLB::CommunicationManager::SetParticipantLocation(char* _location) {
  if (_location) location = _location; 
  else location = "";

  if (!rtpi) return;

  RTPI::RtpiParticipant* local = rtpi->get_own_participant();
  ParticipantManager::pm->UpdateParticipant(local->get_pid(), local->get_cname(), local->get_name(), 
					    local->get_email(), local->get_phone(), local->get_loc(), local->get_tool());
}


bool MLB::CommunicationManager::OpenSMP() {
  if (openSMP) {
    GUI::MessageWindow::Error("CommunicationManager::OpenSMP - Connection to SMP-Daemon is already set up\n");
    return false;
  }

  smp = new SMP::SMP();

  if ((rtpiSocket = smp->Open(::handleReceivedRtpi)) < 0) {
    GUI::MessageWindow::Error("CommunicationManager::OpenSMP - could not connect to smp service\n");
    delete smp;
    return false;
  }

  if ((rtcpiSocket = smp->Open(::handleReceivedRtcpi)) < 0) {
    GUI::MessageWindow::Error("CommunicationManager::OpenSMP - could not connect to smp service\n");
    delete smp;
    return false;
  }

  openSMP = true;

  return true;
}


bool MLB::CommunicationManager::CloseSMP() {

  if (!openSMP || openConnection) {
    //GUI::MessageWindow::Error("CommunicationManager::CloseSMP - close not possible\n");
    return false;
  }

  smp->Close(rtpiSocket);
  smp->Close(rtcpiSocket);

  openSMP = false;

  return true;
}

#ifdef IDSERVER
bool MLB::CommunicationManager::OpenIDServer(std::string _unsIP, int _unsPort) {
  if (openIDServer) {
    GUI::MessageWindow::Error("CommunicationManager::OpenIDServer - Connection to ID-Server is already set up\n");
    return true;
  }

  unsIP = _unsIP;
  unsPort = _unsPort;

  if (StartIdClient(unsIP, unsPort)) {
    openIDServer = true;
    return true;
  }

  return false; // failed
}


bool MLB::CommunicationManager::OpenIDServer() {
  localIDServer = true;
  localID = 0;
  return true;
}


bool MLB::CommunicationManager::CloseIDServer() {

  if (localIDServer) return true;

  if (!openIDServer || openConnection) {
    //GUI::MessageWindow::Error("CommunicationManager::CloseIDServer - close not possible\n");
    return false;
  }
  
  delete uniqueIdClient;
  openIDServer = false;
  uniqueIdClient = NULL;

  return true;
}
#endif

bool MLB::CommunicationManager::OpenRTPI(std::string _ip, int _port, int _ttl, int _bandwidth) {
  if (openConnection || !openSMP || !openIDServer) {
    GUI::MessageWindow::Error("CommunicationManager::OpenRTPI - Cannot set up RTP/I Connection\n");
    return false;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::JoinGroup - joining %q - %u - %u\n", _ip.c_str(), _port, _ttl);
  #endif

  ip = _ip; port = _port; ttl = _ttl; bandwidth = _bandwidth; // validiation necessary!
  if (!smp->JoinGroup(rtpiSocket, ip.c_str(), port, ttl, bandwidth, 'A', false, false, false)) {
    GUI::MessageWindow::Error("CommunicationManager::Open - smp JoinGroup failed\n");
    smp->Close(rtpiSocket);
    smp->Close(rtcpiSocket);
    return false;
  }
  if (!smp->JoinGroup(rtcpiSocket, ip.c_str(), port+1, ttl, bandwidth/10, 'A', false, false, false)) {
    GUI::MessageWindow::Error("CommunicationManager::Open - smp JoinGroup failed\n");
    smp->Close(rtpiSocket);
    smp->Close(rtcpiSocket);
    return false;
  }

  rtpi = new RTPI::Rtpi();
  RTPI::RtpiParticipant* local = rtpi->join_group(username, this, this, new RTPI::TransparentReliableTransport());

  ParticipantManager::pm->AddParticipant(local->get_pid());
  ParticipantManager::pm->UpdateParticipant(local->get_pid(), local->get_cname(), local->get_name(), 
					    local->get_email(),local->get_phone(), local->get_loc(), local->get_tool());

  Et_GlobalEvalF(Et_Interp,"Participant::SetCName %q", local->get_cname().c_str());

  openConnection = true;

  return true;
}


bool MLB::CommunicationManager::CloseRTPI() {
  if (!openConnection) {
    //GUI::MessageWindow::Error("CommunicationManager::CloseRTPI - close not possible\n");
    return false;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::LeaveGroup - leaving group\n");
  #endif

  rtpi->leave_group();

  smp->LeaveGroup(rtpiSocket);
  smp->LeaveGroup(rtcpiSocket);

  AduManager::am->Reset();
  TimeWarp::tw->Reset();

  delete rtpi; rtpi = NULL;

  openConnection = false;

  ParticipantManager::pm->RemoveAllParticipants();

  return true;
}


RTPI::RtpiAdu* MLB::CommunicationManager::CreateState(uint32 subID, uint32 objectID, int type, int priority, RefCount<Buffer> buf) {
  // register after object has been created (ll)
//    if (!rtpi->subcomponent_exists(subID)) {
//      // new subcomponent -> register
//      RegisterSubcomponent(subID, type);
//    }
  
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateState - create state for subcomponent with id %u and parent %u\n", 
		 subID, objectID);
  #endif

  RTPI::RtpiState* state = new RTPI::RtpiState(subID, APPLICATION_TYPE, 0, rtpitime() + AduManager::am->GetLocalLag(), priority, buf);
  state->GetPacketHeader()->SetParticipantID(pid);

  if (objectID != 1) // do not set parent id for the highest hierachy level -> session object does not exist 
    state->SetObjectID(objectID);

  std::map<uint32, std::map<uint32, uint16> >::iterator sn = seqNumbers.find(subID);
  if (sn != seqNumbers.end()) {
    for (std::map<uint32, uint16>::iterator i = sn->second.begin(); i != sn->second.end(); ++i) 
      state->AddSequenceNumberToHeader(i->first, i->second);
  }

  return state;
} 


RTPI::RtpiAdu* MLB::CommunicationManager::CreateState(uint32 subID, uint32 objectID, int type, int priority, RefCount<Buffer> buf, uint32 timestamp) {
//    if (!rtpi->subcomponent_exists(subID)) {
//      // new subcomponent -> register
//      RegisterSubcomponent(subID, type); // 
//    }
  
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateState - create state for subcomponent with id %u and parent %u\n", 
		 subID, objectID);
  #endif

  RTPI::RtpiState* state = new RTPI::RtpiState(subID, APPLICATION_TYPE, 0, timestamp + AduManager::am->GetLocalLag(), priority, buf);
  state->GetPacketHeader()->SetParticipantID(pid);

  if (objectID != 1) // do not set parent id for the highest hierachy level -> session object does not exist 
    state->SetObjectID(objectID);

  std::map<uint32, std::map<uint32, uint16> >::iterator sn = seqNumbers.find(subID);
  if (sn != seqNumbers.end()) {
    for (std::map<uint32, uint16>::iterator i = sn->second.begin(); i != sn->second.end(); ++i) 
      state->AddSequenceNumberToHeader(i->first, i->second);
  }

  return state;
} 


RTPI::RtpiAdu* MLB::CommunicationManager::CreateEvent(uint32 subID, RefCount<Buffer> buf) {
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateEvent - create event for subcomponent with id %u\n", subID);
  #endif

  RTPI::RtpiEvent* event = new RTPI::RtpiEvent(subID, APPLICATION_TYPE, 0, rtpitime() + AduManager::am->GetLocalLag(), buf);
  event->GetPacketHeader()->SetParticipantID(pid);

  return event;
}


RTPI::RtpiAdu* MLB::CommunicationManager::CreateEvent(uint32 subID, uint32 objectID, RefCount<Buffer> buf) {
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateEvent - create event for subcomponent with id %u and parent %u\n", 
		 subID, objectID);
  #endif

  RTPI::RtpiEvent* event = new RTPI::RtpiEvent(subID, APPLICATION_TYPE, 0, rtpitime() + AduManager::am->GetLocalLag(), buf);
  event->GetPacketHeader()->SetParticipantID(pid);

  if (objectID != 1) // do not set parent id for the highest hierachy level -> session object does not exist //J
    event->SetObjectID(objectID);

  return event;
}


RTPI::RtpiAdu* MLB::CommunicationManager::CreateCue(uint32 subID, RefCount<Buffer> buf) {
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateCue - create cue for subcomponent with id %u\n", subID);
  #endif

  RTPI::RtpiCue* cue = new RTPI::RtpiCue(subID, APPLICATION_TYPE, 0, rtpitime() + AduManager::am->GetLocalLag(), buf);
  cue->GetPacketHeader()->SetParticipantID(pid);

  return cue;
}


RTPI::RtpiAdu* MLB::CommunicationManager::CreateStateQuery(uint32 subID, int priority) {
  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::CreateStateQuery - create state query for subcomponent with id %u\n", subID);
  #endif

  RTPI::RtpiStateQuery* query = new RTPI::RtpiStateQuery(subID, APPLICATION_TYPE, rtpitime(), priority);
  query->GetPacketHeader()->SetParticipantID(pid);

  return query;
}


void MLB::CommunicationManager::SendAduReliable(RefCount<RTPI::RtpiAdu> adu) {
  if (!adu) return;
  if (!openConnection) return;

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adu); 

    if (adu->GetType() == RTPI::RtpiHeader::PKT_EVENT) 
      UpdateEventSequenceNumber(adu->GetSubcomponentID(), adu->GetParticipantID(), adu->GetPacketHeader()->GetSequenceNumber());

  } catch (RtpiException e) {
    GUI::MessageWindow::Error("CommunicationManager::SendAduReliable - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }

  #ifdef DEBUG_RECORDING
  if (adu->GetType() == RTPI::RtpiHeader::PKT_STATE)
    if (((RTPI::RtpiStateBaseHeader*)adu->GetAduHeader())->ObjectIDIsSet())
      SessionManager::sm->message->Info("CommunicationManager::SendAduReliable - send state for obj %u / parent %u / ts %u\n",
					adu->GetSubcomponentID(), 
					((RTPI::RtpiStateBaseHeader*)adu->GetAduHeader())->GetObjectID(), 
					adu->GetAduHeader()->GetTimestamp());
    else
      SessionManager::sm->message->Info("CommunicationManager::SendAduReliable - send state for obj %u / ts %u\n", 
					adu->GetSubcomponentID(), adu->GetAduHeader()->GetTimestamp());
  #endif

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::SendAduReliable - send adu for subcomponent with id %u\n", 
		 adu->GetSubcomponentID());
  #endif

  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    smp->Send(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}

 
void MLB::CommunicationManager::SendAdusReliable(RefCount<RTPI::RtpiAduVector> adus) {
  if (!openConnection) return;
  if (!adus || !adus->size()) return;

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adus);  

    for (RTPI::RtpiAduVector::iterator i = adus->begin(); i != adus->end(); ++i) {
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_EVENT) 
	UpdateEventSequenceNumber((*i)->GetSubcomponentID(), (*i)->GetParticipantID(), (*i)->GetPacketHeader()->GetSequenceNumber());

      #ifdef DEBUG_RECORDING
      if ((*i)->GetType() == RTPI::RtpiHeader::PKT_STATE)
	if (((RTPI::RtpiStateBaseHeader*)(*i)->GetAduHeader())->ObjectIDIsSet())
	  SessionManager::sm->message->Info("CommunicationManager::SendAdusReliable - send state for obj %u / parent %u / ts %u\n",
					    (*i)->GetSubcomponentID(), 
					    ((RTPI::RtpiStateBaseHeader*)(*i)->GetAduHeader())->GetObjectID(), 
					    (*i)->GetAduHeader()->GetTimestamp());
	else
	  SessionManager::sm->message->Info("CommunicationManager::SendAdusReliable - send state for obj %u / ts %u\n", 
					    (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp());
      #endif
    }

  } catch (RtpiException e) {
    GUI::MessageWindow::Error("CommunicationManager::SendAdusReliable - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::SendAdusReliable - send %u adus\n",  bufVector->size());
  #endif

  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    smp->Send(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


void MLB::CommunicationManager::SendAduUnreliable(RefCount<RTPI::RtpiAdu> adu) {
  if (!openConnection) return;
  if (!adu) return;

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adu); 
  } catch (RtpiException e) {
    GUI::MessageWindow::Error("CommunicationManager::SendAduUnreliable - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::SendAduUnreliable - send adu for subcomponent with id %u\n", 
		 adu->GetSubcomponentID());
  #endif

  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    smp->SendUnreliable(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


void MLB::CommunicationManager::SendAdusUnreliable(RefCount<RTPI::RtpiAduVector> adus) {
  if (!openConnection) return;
  if (!adus || !adus->size()) return;

  RefCount<RTPI::BufferVector> bufVector = NULL;

  try {
    bufVector = rtpi->create_rtpi_packets(adus); 
  } catch (RtpiException e) {
    GUI::MessageWindow::Error("CommunicationManager::SendAdusUnreliable - RtpiException occured: %u / %q\n", 
		   e.errCode, e.errMsg.c_str());
    return;
  }

  #ifdef DEBUG_RTPI
  GUI::MessageWindow::Info("CommunicationManager::SendAdusUnreliable - send %u adus\n",  rtpiPackets->size());
  #endif

  for (RTPI::BufferVector::iterator i = bufVector->begin(); i != bufVector->end(); ++i)
    smp->SendUnreliable(rtpiSocket, (*i)->getCont(), (*i)->getSize());
}


void MLB::CommunicationManager::QueryState(uint32 id, int priority) {
  if (!openConnection) return;

  double x = drand48();
  int n    = ParticipantManager::pm->GetNumberOfParticipants();
  double N = (n == 1) ? 1 : n - 1;
  
  #ifdef DEBUG
  GUI::MessageWindow::Info("CommunicationManager::QueryState for object %u - x: %f, N: %f\n", id, x, N);
  #endif
 
  if ((n == 1) || (N == 1) || (x < (1.0 / N))) {
    // 1. send state query immediately
    SendStateQuery(id, priority);

    #ifdef DEBUG
    GUI::MessageWindow::Info("CommunicationManager::QueryState - send immediately\n");
    #endif
  } else {
    // 2. set state query timer

    int runningTime = (int)(T * (1 + log(x) / log(N)));
    TimerManager::tm->SetTimer(this, new TimerObject(runningTime, TimerObject::STATE_QUERY, id, priority));

    #ifdef DEBUG
    GUI::MessageWindow::Info("CommunicationManager::QueryState - send later, set timer with running time %u\n", runningTime);
    #endif
  }
}


void MLB::CommunicationManager::SendStateQuery(uint32 id, int priority) {
  if (!openConnection) return;

  #ifdef DEBUG
  GUI::MessageWindow::Info("CommunicationManager::SendStateQuery - send state query for %u with pri %u\n", id, priority);
  #endif

  RefCount<RTPI::RtpiAdu> adu = CreateStateQuery(id, priority);
  SendAduReliable(adu);
}


void MLB::CommunicationManager::HandleStateQuery(RTPI::RtpiStateQuery* query) {
  GenericObject* obj = SessionManager::sm->GetSharedSession()->GetObject(query->GetSubcomponentID());
  if (!obj) obj = SessionManager::sm->GetSharedSession()->GetGroup(query->GetSubcomponentID());
  if (!obj) {
    GUI::MessageWindow::Error("CommunicationManager::HandleStateQuery - did not find object with id %u\n", 
		   query->GetSubcomponentID());
    return; // do not insert in pending adu list
  }

  GenericContainer* parent = SessionManager::sm->GetParentContainer(obj->GetID());
  if (!parent) {
    GUI::MessageWindow::Error("CommunicationManager::HandleStateQuery - did not find parent for object with id %u\n", 
		   obj->GetID());
    return; // do not insert in pending adu list
  }

  int priority = query->GetAduHeader()->GetPriority();

  double x = drand48();
  int n    = ParticipantManager::pm->GetNumberOfParticipants();
  double N = (n == 1) ? 1 : n - 1;
  
  #ifdef DEBUG
  GUI::MessageWindow::Info("CommunicationManager::HandleStateQuery - x: %f, N: %f\n", x, N);
  #endif
 
  if ((n == 1) || (N == 1) || (x < (1.0 / N))) {
    // 1. send State immediately
    AnswerStateQuery(obj->GetID(), priority);

    #ifdef DEBUG
    GUI::MessageWindow::Info("CommunicationManager::HandleStateQuery - answer immediately\n");
    #endif
  } else {
    // 2. set state reply timer

    int runningTime = (int)(T * (1 + log(x) / log(N)));
    TimerManager::tm->SetTimer(this, new TimerObject(runningTime , TimerObject::STATE_REPLY, obj->GetID(), priority));

    #ifdef DEBUG
    GUI::MessageWindow::Info("CommunicationManager::HandleStateQuery - answer later, set timer with running time %u\n", runningTime);
    #endif
  }

  return;
}


void MLB::CommunicationManager::AnswerStateQuery(uint32 id, int priority) {
  GenericObject* obj = SessionManager::sm->GetSharedSession()->GetObject(id);
  if (!obj) obj = SessionManager::sm->GetSharedSession()->GetGroup(id);
  if (!obj) {
    // could have been deleted in the meantime
    GUI::MessageWindow::Error("CommunicationManager::AnswerStateQuery - did not find object with id %u\n", id);
    return;
  }
 
  if ((priority == 0) || (priority == 1)) priority = 0;
  //else priority = 3;

  #ifdef DEBUG
  GUI::MessageWindow::Info("CommunicationManager::AnswerStateQuery - send state for %u with pri %u\n", obj->GetID(), priority);
  #endif

  #ifdef DEBUG_RECORDING
  GUI::MessageWindow::Info("CommunicationManager::AnswerStateQuery - send state for %u with pri %u\n", obj->GetID(), priority);
  #endif

  RefCount<RTPI::RtpiAdu> adu;
  uint32 timestamp = rtpitime();

  if (obj->GetType() == GenericObject::FOLDER) {
    adu = obj->GetStateAdu(rtpitime(), priority);
    TimeWarp::tw->InsertAdu(adu);
    SendAduReliable(adu);

    if (priority == 3) 
      ((Folder*)obj)->SendAllObjects(timestamp, priority);
  } else if (obj->GetType() == GenericObject::PAGE) {
    adu = obj->GetStateAdu(rtpitime(), priority);
    TimeWarp::tw->InsertAdu(adu);
    SendAduReliable(adu);

    //if (priority == 3) 
    ((Page*)obj)->SendAllObjects(timestamp, priority);
  } else {
    adu = obj->GetStateAdu(rtpitime(), priority);
    TimeWarp::tw->InsertAdu(adu);
    SendAduReliable(adu);
  }
}


void MLB::CommunicationManager::HandleReceivedRtpi(SMP::Indication ind) {
  if (!openConnection) return;

  #ifdef INFO_CONS
  static uint32 opCount               = 0;
  static uint32 intentionCount        = 0;
  static uint32 lateOpCount           = 0;
  static uint32 lateOpTWQEmptyCount   = 0;
  static uint32 lateOpNoConflictCount = 0;
  static uint32 lateOpOverwriteCount  = 0;
  static uint32 timewarpCount         = 0;
  static uint32 timewarpTime          = 0;
  static uint32 totalTimewarpTime     = 0;
  static uint32 stateRequestCount     = 0;  
  #endif

  switch(ind.ServiceType) {
  case SMP::Indication::DATAIND: {	
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got data %u\n", ind.buf->getSize());
    #endif

    RefCount<RTPI::RtpiAduVector> rtpiAdus;
    try {
      rtpiAdus = rtpi->create_rtpi_adus(ind.buf);
    } catch (RtpiException e) {
      GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    if (!rtpiAdus) 
      return; // adu not complete, received only fragments

    uint32 currentTime = rtpitime();

    for (RTPI::RtpiAduVector::iterator i = rtpiAdus->begin(); i != rtpiAdus->end(); ++i) {
      if ((*i)->GetAduHeader()->GetPayloadType() != APPLICATION_TYPE) {
	GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - wrong payload type %u\n", 
		       (*i)->GetAduHeader()->GetPayloadType());
	return;
      }

      //ParticipantManager::pm->AddDelay((*i)->GetParticipantID(), (*i)->GetAduHeader()->GetTimestamp());

      #ifdef INFO_CONS
      ++opCount;
      #endif

      #ifdef DEBUG_RECORDING
      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received ADU(%u) for %u with ts %u at time %u\n", 
			       (*i)->GetType(), (*i)->GetSubcomponentID(), (*i)->GetAduHeader()->GetTimestamp(), currentTime);
      #endif

      #ifdef DEBUG_CONS
      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - receive ADU from %u with ts %u at time %u\n", 
			       (*i)->GetParticipantID(), (*i)->GetAduHeader()->GetTimestamp(), currentTime);
      #endif

      switch ((*i)->GetType()) {
      //
      // Event
      //
      case RTPI::RtpiHeader::PKT_EVENT: {
	#ifdef DEBUG_RTPI
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received Event\n");
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() >= currentTime) {
	  AduManager::am->InsertAdu(*i);

	  if (checkIntentionViolation) {
	    RefCount<RTPI::RtpiAduList> icAdus1 = NULL;
	    icAdus1 = AduManager::am->CheckForIntentionViolation(*i);

	    RefCount<RTPI::RtpiAduList> icAdus2 = NULL;
	    icAdus2 = TimeWarp::tw->CheckForIntentionViolation(*i, (*i)->GetAduHeader()->GetTimestamp() - AduManager::am->GetLocalLag());

	    if ((!icAdus1->empty()) && (!icAdus2->empty())) {
              #ifdef DEBUG_CONS
	      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for object %u caused intention violation\n",
			   (*i)->GetSubcomponentID());
              #endif

	      #ifdef INFO_CONS
	      ++intentionCount;
	      #endif

	      DisplayIntentionConflicts((*i)->GetSubcomponentID(), icAdus1, icAdus2);
	    } 
	  }
	  break;
	}
	  
	#ifndef ENABLE_TW
	AduManager::am->InsertAdu(*i);
	break;
	#endif

	#ifdef INFO_CONS
	++lateOpCount;
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() < currentTime - TimeWarp::tw->GetStorageTime()) {
	  // inconsistency cannot be repaired -> outside of twq, state request is needed !
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for object %u was too late to be repaired\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++stateRequestCount;
	  #endif

	  RefCount<RtpiAdu> query = CreateStateQuery((*i)->GetSubcomponentID(), 3);
	  SendAduReliable(query);
	  break;
	}

	// possibly an inconsistency occured
	RefCount<RTPI::RtpiAduList> pcAdus = TimeWarp::tw->GetPossiblyConflictingAdus((*i)->GetAduHeader()->GetTimestamp());
	if (pcAdus->empty()) {
	  // did not find any adus in the conflicting period of time
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for object %u was late, but no possibly conflicting adus\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++lateOpTWQEmptyCount;
	  #endif

	  AduManager::am->InsertAdu(*i);
	  break;
	}

	RefCount<RTPI::RtpiAduList> cAdus = GetConflictingAdus(*i, pcAdus);
	if (cAdus->empty()) {
	  // did not find any conflicting adus
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for object %u was late, but no conflicting adus\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++lateOpNoConflictCount;
	  #endif

	  AduManager::am->InsertAdu(*i);
	  break;
	}
	
	if (OverwritingAdu(*i, cAdus)) {
	  // no timewarp necessary, ignore late adu
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for object %u was late but can be ignored\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++lateOpOverwriteCount;
	  #endif

	  TimeWarp::tw->InsertAdu(*i);
	  break;
	}

	// execute timewarp
        #ifdef DEBUG_CONS
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - event for %u with ts %u\n"
		       "was late and TimeWarp is necessary",(*i)->GetAduHeader()->GetTimestamp(), (*i)->GetSubcomponentID());
        #endif

	#ifdef INFO_CONS
	++timewarpCount;
	timewarpTime = rtpitime();
	#endif

	SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);

	TimeWarp::tw->InsertAdu(*i);

	int objType = ParseObjectType((*i)->GetPayload());
	if ((objType == FOLDER) || (objType == PAGE)) {
	  // 1. get IDs of concerned objects
	  Folder* parent;
	  std::list<uint32>* concernedObjects;
	  if (ParseSubType((*i)->GetPayload()) == CHG_PARENT) {
	    parent = (Folder*)SessionManager::sm->GetSharedSession()->GetObject(((RTPI::RtpiEvent*)&(**i))->GetObjectID());
	    if (!parent) {
	      SessionManager::sm->cursor->Reset();
	      break;
	    }
	    concernedObjects = parent->GetObjectIDList();
	    concernedObjects->push_back((*i)->GetSubcomponentID());
	  } else {
	    parent = (Folder*)SessionManager::sm->GetSharedSession()->GetContainer((*i)->GetSubcomponentID());
	    if (!parent) {
	      SessionManager::sm->cursor->Reset();
	      break;
	    }
	    concernedObjects = parent->GetObjectIDList();
	  }

	  // 2. filter last state and following events from twq for all concerned objects
	  RefCount<RTPI::RtpiAduList> restoreAdus = TimeWarp::tw->GetRestoreConsistencyAduList((*i)->GetSubcomponentID(), 
											       concernedObjects, 
											       (*i)->GetAduHeader()->GetTimestamp());
	  if (!restoreAdus || restoreAdus->empty()) {
	    // state request needed !
	    #ifdef DEBUG_CONS
	    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - did not find all necessary adus for timewarp (1)\n");
	    #endif
	    SessionManager::sm->cursor->Reset();
	    break;
	  }
	  
	  // 3. execute ADUs
	  for (RTPI::RtpiAduList::reverse_iterator a = restoreAdus->rbegin(); a != restoreAdus->rend(); ++a) {
	    switch ((*a)->GetType()) {
	    case RTPI::RtpiHeader::PKT_STATE:
	      AduManager::am->HandleState(*a);
	      break;
	    case RTPI::RtpiHeader::PKT_EVENT:
	      AduManager::am->HandleEvent(*a);
	      break;
	    }
	  }

	  //parent->StoreAllObjectsInTWQ(rtpitime());
	  delete concernedObjects;
	} else {
	  // 1. get IDs of concerned objects
	  Page* parent = (Page*)(SessionManager::sm->GetSharedSession()->GetContainer((*i)->GetSubcomponentID()));
	  if (!parent) {
	    SessionManager::sm->cursor->Reset();
	    break;
	  }
	  std::list<uint32>* concernedObjects = parent->GetObjectIDList();

	  // 2. filter last state and following events from twq for all concerned objects
	  RefCount<RTPI::RtpiAduList> restoreAdus = TimeWarp::tw->GetRestoreConsistencyAduList((*i)->GetSubcomponentID(), 
											       concernedObjects, 
											       (*i)->GetAduHeader()->GetTimestamp());
	  if (!restoreAdus || restoreAdus->empty()) {
	    // state request needed !
	    #ifdef DEBUG_CONS
	    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - did not find all necessary adus for timewarp (2)\n");
	    #endif
	    SessionManager::sm->cursor->Reset();
	    break;
	  }

	  // 3. Reset affected parent page
	  parent->Clear();

	  // 4. execute ADUs
	  for (RTPI::RtpiAduList::reverse_iterator a = restoreAdus->rbegin(); a != restoreAdus->rend(); ++a) {
	    switch ((*a)->GetType()) {
	    case RTPI::RtpiHeader::PKT_STATE:
	      AduManager::am->HandleState(*a);
	      break;
	    case RTPI::RtpiHeader::PKT_EVENT:
	      AduManager::am->HandleEvent(*a);
	      break;
	    }
	  }

	  //parent->StoreAllObjectsInTWQ(rtpitime());
	  delete concernedObjects;
	}
	SessionManager::sm->cursor->Reset();

	#ifdef INFO_CONS
	timewarpTime = rtpitime() - timewarpTime;
	totalTimewarpTime += timewarpTime;
	#endif

	break;
      }


      //
      // Cue
      //
      case RTPI::RtpiHeader::PKT_CUE: {
	#ifdef DEBUG_RTPI
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received Cue (reliable)\n");
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() >= currentTime) {
	  AduManager::am->InsertAdu(*i);

	  if (checkIntentionViolation) {
	    RefCount<RTPI::RtpiAduList> icAdus1 = NULL;
	    icAdus1 = AduManager::am->CheckForIntentionViolation(*i);

	    RefCount<RTPI::RtpiAduList> icAdus2 = NULL;
	    icAdus2 = TimeWarp::tw->CheckForIntentionViolation(*i, (*i)->GetAduHeader()->GetTimestamp() - AduManager::am->GetLocalLag());

	    if ((!icAdus1->empty()) && (!icAdus2->empty())) {
              #ifdef DEBUG_CONS
	      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - cue for object %u caused intention violation\n",
			   (*i)->GetSubcomponentID());
              #endif

	      #ifdef INFO_CONS
	      ++intentionCount;
	      #endif

	      DisplayIntentionConflicts((*i)->GetSubcomponentID(), icAdus1, icAdus2);
	    } 
	  }
	  break;
	} 

	#ifdef INFO_CONS
	++lateOpCount;
	++lateOpOverwriteCount;
	#endif

	// cue is too late but can be ignored due to it's volatile character
	break;
      }


      //
      // State
      //
      case RTPI::RtpiHeader::PKT_STATE: {
	#ifdef DEBUG_RTPI
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received State\n");
	#endif

	#ifdef SIMPLE_LJ
	if (TimerManager::tm->TimerRunning(this, TimerObject::STATE_QUERY, (*i)->GetSubcomponentID()))
	  TimerManager::tm->DeleteTimer(this, TimerObject::STATE_QUERY, (*i)->GetSubcomponentID());
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() >= currentTime) {
	  AduManager::am->InsertAdu(*i);

	  if (checkIntentionViolation) {
	    RefCount<RTPI::RtpiAduList> icAdus1 = NULL;
	    icAdus1 = AduManager::am->CheckForIntentionViolation(*i);

	    RefCount<RTPI::RtpiAduList> icAdus2 = NULL;
	    icAdus2 = TimeWarp::tw->CheckForIntentionViolation(*i, (*i)->GetAduHeader()->GetTimestamp() - AduManager::am->GetLocalLag());

	    if ((!icAdus1->empty()) && (!icAdus2->empty())) {
              #ifdef DEBUG_CONS
	      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u caused intention violation\n",
			   (*i)->GetSubcomponentID());
              #endif

	      #ifdef INFO_CONS
	      ++intentionCount;
	      #endif

	      DisplayIntentionConflicts((*i)->GetSubcomponentID(), icAdus1, icAdus2);
	    } 
	  }
	  break;
	} 

	#ifndef ENABLE_TW
	AduManager::am->InsertAdu(*i);
	break;
	#endif

	#ifdef INFO_CONS
	++lateOpCount;
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() < currentTime - TimeWarp::tw->GetStorageTime()) {
	  // inconsistency cannot be repaired -> outside of twq, state request needed
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u was too late to be repaired\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++stateRequestCount;
	  #endif

	  RefCount<RtpiAdu> query = CreateStateQuery((*i)->GetSubcomponentID(), 3);
	  SendAduReliable(query);
	  break;
	}

	// possibly an inconsistency occured
	RefCount<RTPI::RtpiAduList> pcAdus = TimeWarp::tw->GetPossiblyConflictingAdus((*i)->GetAduHeader()->GetTimestamp());
	if (pcAdus->empty()) {
	  // did not find any adus in the conflicting period of time
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u was late, but no possibly conflicting adus\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++lateOpTWQEmptyCount;
	  #endif

	  AduManager::am->InsertAdu(*i);
	  break;
	}

	RefCount<RTPI::RtpiAduList> cAdus = GetConflictingAdus(*i, pcAdus);
	if (cAdus->empty()) {
	  // did not find any conflicting adus
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u was late, but no conflicting adus\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  ++lateOpNoConflictCount;
	  #endif

	  AduManager::am->InsertAdu(*i);
	  break;
	}
	
	if (OverwritingAdu(*i, cAdus)) {
	  // no timewarp necessary, ignore late adu
          #ifdef DEBUG_CONS
	  GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u was late but can be ignored\n",
			 (*i)->GetSubcomponentID());
          #endif

	  #ifdef INFO_CONS
	  lateOpOverwriteCount;
	  #endif

	  TimeWarp::tw->InsertAdu(*i);
	  break;
	}

	// execute timewarp
        #ifdef DEBUG_CONS
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for %u with ts %u was late\n"
		       " and TimeWarp is necessary", (*i)->GetAduHeader()->GetTimestamp(), (*i)->GetSubcomponentID());
        #endif

        #ifdef INFO_CONS
	++timewarpCount;
	timewarpTime = rtpitime();
        #endif

	SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);

	TimeWarp::tw->InsertAdu(*i);

	int objType = ParseObjectType((*i)->GetPayload());
	if ((objType == FOLDER) || (objType == PAGE)) {
	  // 1. get IDs of concerned objects
	  Folder* parent = (Folder*)SessionManager::sm->GetSharedSession()->GetObject(((RTPI::RtpiState*)&(**i))->GetObjectID());
	  if (!parent) {
	    SessionManager::sm->cursor->Reset();
	    break;
	  }
	  std::list<uint32>* concernedObjects = parent->GetObjectIDList();
	  concernedObjects->push_back((*i)->GetSubcomponentID());;

	  // 2. filter last state and following events from twq for all concerned objects
	  RefCount<RTPI::RtpiAduList> restoreAdus = TimeWarp::tw->GetRestoreConsistencyAduList((*i)->GetSubcomponentID(), 
											       concernedObjects, 
											       (*i)->GetAduHeader()->GetTimestamp());
	  if (!restoreAdus || restoreAdus->empty()) {
	    #ifdef DEBUG_CONS
	    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - did not find all necessary adus for timewarp (3)\n");
	    #endif
	    SessionManager::sm->cursor->Reset();
	    break;
	  }
	  
	  // 3. execute ADUs
	  for (RTPI::RtpiAduList::reverse_iterator a = restoreAdus->rbegin(); a != restoreAdus->rend(); ++a) {
	    switch ((*a)->GetType()) {
	    case RTPI::RtpiHeader::PKT_STATE:
	      AduManager::am->HandleState(*a);
	      break;
	    case RTPI::RtpiHeader::PKT_EVENT:
	      AduManager::am->HandleEvent(*a);
	      break;
	    }
	  }

	  //parent->StoreAllObjectsInTWQ(rtpitime());
	  delete concernedObjects;
	} else {
	  // 1. get IDs of concerned objects
	  Page* parent = (Page*)SessionManager::sm->GetSharedSession()->GetObject(((RTPI::RtpiState*)&(**i))->GetObjectID());
	  if (!parent) {
	    SessionManager::sm->cursor->Reset();
	    break;
	  }
	  std::list<uint32>* concernedObjects = parent->GetObjectIDList();
	  concernedObjects->push_back((*i)->GetSubcomponentID());;

	  // 2. filter last state and following events from twq for all concerned objects
	  RefCount<RTPI::RtpiAduList> restoreAdus = TimeWarp::tw->GetRestoreConsistencyAduList((*i)->GetSubcomponentID(), 
											       concernedObjects, 
											       (*i)->GetAduHeader()->GetTimestamp());
	  if (!restoreAdus || restoreAdus->empty()) {
	    #ifdef DEBUG_CONS
	    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - did not find all necessary adus for timewarp (4)\n");
	    #endif
	    SessionManager::sm->cursor->Reset();
	    break;
	  }

	  // 3. Reset affected parent page
	  parent->Clear();

	  // 4. execute ADUs
	  for (RTPI::RtpiAduList::reverse_iterator a = restoreAdus->rbegin(); a != restoreAdus->rend(); ++a) {
	    switch ((*a)->GetType()) {
	    case RTPI::RtpiHeader::PKT_STATE:
	      AduManager::am->HandleState(*a);
	      break;
	    case RTPI::RtpiHeader::PKT_EVENT:
	      AduManager::am->HandleEvent(*a);
	      break;
	    }
	  }

	  //parent->StoreAllObjectsInTWQ(rtpitime());
	  delete concernedObjects;
	}
	SessionManager::sm->cursor->Reset();

        #ifdef INFO_CONS
	timewarpTime = rtpitime() - timewarpTime;
	totalTimewarpTime += timewarpTime;
	#endif

	break;
      }
      //
      // State Query
      //
      case RTPI::RtpiHeader::PKT_STATE_QUERY: {
	#ifdef DEBUG_RTPI
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received State Query\n");
	#endif
	HandleStateQuery((RTPI::RtpiStateQuery*)&(**i));
	break;
      }
      default:
	GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - wrong adu type %u\n", 
		       (*i)->GetType());
      }
    }

    #ifdef INFO_CONS
    GUI::MessageWindow::InfoBlue("received ops %u, intention conflicts %u, late ops %u, state requests %u, twq empty %u, no conflict %u, overwite %u, timewarps %u - average timewarp time %u\n",
		   opCount, intentionCount, lateOpCount, stateRequestCount, lateOpTWQEmptyCount, lateOpNoConflictCount, 
		   lateOpOverwriteCount, timewarpCount, (uint32)((double)totalTimewarpTime/timewarpCount));
    #endif

    try {
      uint8* p8; 
      int opType;
      for (RTPI::RtpiAduVector::iterator i = rtpiAdus->begin(); i != rtpiAdus->end(); ++i) {
	if ((*i)->GetType() == RTPI::RtpiHeader::PKT_EVENT) {
	  p8 = (uint8*)((RTPI::RtpiEvent*)&(**i))->GetPayload();
	  opType = (p8[0] & MLB_MASK_OPT) >> 2;
	  if (opType == ADU_DELETE) continue; // ignore delete events; they would trigger a late join
	}	
	rtpi->handle_received_rtpi(*i, ind.senderIPAddr);
      }
    } catch (RtpiException e) {
      GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    return;
  }

  
  case SMP::Indication::UNRLDATAIND: {	
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got unreliable data %u\n", ind.buf->getSize());
    #endif

    RefCount<RTPI::RtpiAduVector> rtpiAdus;
    try {
      rtpiAdus = rtpi->create_rtpi_adus(ind.buf);
    } catch (RtpiException e) {
      GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    if (!rtpiAdus) 
      return; // adu not complete, received only fragments

    try {
      rtpi->handle_received_rtpi(rtpiAdus, ind.senderIPAddr);
    } catch (RtpiException e) {
      GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }

    for (RTPI::RtpiAduVector::iterator i = rtpiAdus->begin(); i != rtpiAdus->end(); ++i) {
      if ((*i)->GetAduHeader()->GetPayloadType() != APPLICATION_TYPE) {
	GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - wrong payload type %u\n", 
		       (*i)->GetAduHeader()->GetPayloadType());
	return;
      }

      //ParticipantManager::pm->AddDelay((*i)->GetParticipantID(), (*i)->GetAduHeader()->GetTimestamp());

      uint32 currentTime = rtpitime();

      #ifdef INFO_CONS
      ++opCount;
      #endif

      switch ((*i)->GetType()) {
      case RTPI::RtpiHeader::PKT_CUE: {
	#ifdef DEBUG_RTPI
	GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - received Cue\n");
	#endif

	if ((*i)->GetAduHeader()->GetTimestamp() >= currentTime) {
	  AduManager::am->InsertAdu(*i);

	  if (checkIntentionViolation) {
	    RefCount<RTPI::RtpiAduList> icAdus1 = NULL;
	    icAdus1 = AduManager::am->CheckForIntentionViolation(*i);

	    RefCount<RTPI::RtpiAduList> icAdus2 = NULL;
	    icAdus2 = TimeWarp::tw->CheckForIntentionViolation(*i, (*i)->GetAduHeader()->GetTimestamp() - AduManager::am->GetLocalLag());

	    if ((!icAdus1->empty()) && (!icAdus2->empty())) {
              #ifdef DEBUG_CONS
	      GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - state for object %u caused intention violation\n",
			   (*i)->GetSubcomponentID());
              #endif

	      #ifdef INFO_CONS
	      ++intentionCount;
	      #endif

	      DisplayIntentionConflicts((*i)->GetSubcomponentID(), icAdus1, icAdus2);
	    } 
	  }
	  break;
	} 

	#ifdef INFO_CONS
	++lateOpCount;
	++lateOpOverwriteCount;
	#endif

	// cue is too late but can be ignored due to it's volatile character
	break;
      }
      default:
	GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - wrong adu type %u\n", 
		       (*i)->GetType());
      }
    }

    #ifdef INFO_CONS
    GUI::MessageWindow::InfoBlue("received ops %u, intention conflicts %u, late ops %u, state requests %u, twq empty %u, no conflict %u, overwite %u, timewarps %u - average timewarp time %u\n",
		   opCount, intentionCount, lateOpCount, stateRequestCount, lateOpTWQEmptyCount, lateOpNoConflictCount, 
		   lateOpOverwriteCount, timewarpCount, (uint32)((double)totalTimewarpTime/timewarpCount));
    #endif

    return;
  }
  

  case SMP::Indication::LJNEEDED: 
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got LJNEEDED\n");
    #endif
    return;
														 

  case SMP::Indication::DADUIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got DADUIND\n");
    #endif
    return;

  case SMP::Indication::DADUFAIL:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got DADUFAIL\n");
    #endif
    return;

  case SMP::Indication::ERR:
    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtpi - got ERR / probably lost connection to smp\n");
    //unlinksocket(smp_socket);
    return;


  case SMP::Indication::NOSERVICE:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got NOSERVICE\n");
    #endif
    return;

  case SMP::Indication::JOININD:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got JOININD\n");
    #endif
    return;

  case SMP::Indication::LEAVIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got LEAVIND\n");
    #endif
    return;

  case SMP::Indication::ACKIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got ACKIND\n");
    #endif
    return;

  case SMP::Indication::DADUCONF:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got DADUCONF\n");
    #endif
    return;

  case SMP::Indication::CLOSEIND:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got CLOSEIND\n");
    #endif
    return;

  case SMP::Indication::ALIVE:
    #ifdef DEBUG_RTPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtpi - got ALIVE\n");
    #endif
    return;

  default:
    return;
  }
}


void MLB::CommunicationManager::HandleReceivedRtcpi(SMP::Indication ind) {
  if (!openConnection) return;

  switch(ind.ServiceType) {
  case SMP::Indication::DATAIND: {	
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got data %u\n", ind.buf->getSize());
    #endif
    return;
  }

  case SMP::Indication::UNRLDATAIND: {	
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got unreliable data %u\n", ind.buf->getSize());
    #endif
    try {
      rtpi->handle_received_rtcpi(ind.buf, ind.senderIPAddr);
    } catch (RtpiException e) {
      GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtcpi - RtpiException occured: %u / %q\n", 
		     e.errCode, e.errMsg.c_str());
      return;
    }
    return;
  }

  case SMP::Indication::LJNEEDED: 
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got LJNEEDED\n");
    #endif
    return;

  case SMP::Indication::DADUIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got DADUIND\n");
    #endif
    return;

  case SMP::Indication::DADUFAIL:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got DADUFAIL\n");
    #endif
    return;

  case SMP::Indication::ERR:
    GUI::MessageWindow::Error("CommunicationManager::HandleReceivedRtcpi - got ERR / probably lost connection to smp\n");
    //unlinksocket(smp_socket);
    return;

  case SMP::Indication::NOSERVICE:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got NOSERVICE\n");
    #endif
    return;

  case SMP::Indication::JOININD:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got JOININD\n");
    #endif
    return;

  case SMP::Indication::LEAVIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got LEAVIND\n");
    #endif
    return;

  case SMP::Indication::ACKIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got ACKIND\n");
    #endif
    return;

  case SMP::Indication::DADUCONF:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got DADUCONF\n");
    #endif
    return;

  case SMP::Indication::CLOSEIND:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got CLOSEIND\n");
    #endif
    return;

  case SMP::Indication::ALIVE:
    #ifdef DEBUG_RTCPI
    GUI::MessageWindow::Info("CommunicationManager::HandleReceivedRtcpi - got ALIVE\n");
    #endif
    return;

  default:
    return;
  }
}


void MLB::CommunicationManager::send_rtcpi(RefCount<Buffer> buf) {
  if (!openConnection) return;
  smp->SendUnreliable(rtcpiSocket, buf->getCont(), buf->getSize());
}


u_int32 MLB::CommunicationManager::get_local_address(int i) {
  if (i + 1 > localAddresses.size()) return 0;
  return localAddresses[i];
}


int MLB::CommunicationManager::addr_is_local(u_int32 addr) {
  for (std::vector<uint32>::iterator i = localAddresses.begin(); i != localAddresses.end(); ++i)
    if (*i == addr) return 1;
  return 0;
}


u_int32 MLB::CommunicationManager::get_session_bandwith() {
  return bandwidth;
}


u_int32 MLB::CommunicationManager::get_new_subid() {
  return GetObjectID();
}


u_int32 MLB::CommunicationManager::get_new_pid() {
  #ifdef IDSERVER
  std::vector<uint32>* ids = uniqueIdClient->GetIds("participant", 1);

  if ((ids == NULL) || ids->empty()) {
    //Et_GlobalEvalF(Et_Interp,"MessageDlg .errorIdClient -icon error -type ok -title \"Error\""
    //		   "-message \"Unique Name Server not found. Please start UNS on machine %q port %u\"\n", unsIP, unsPort);
    //Et_GlobalEvalF(Et_Interp,"ExitMLB");
    exit(-1);
  }

  pid = ids->front();
  delete ids;
  return pid;
  #else
  pid = random32();
  return pid;
  #endif
}


void MLB::CommunicationManager::subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason) {
  #ifdef SIMPLE_LJ
  if (subcomp->second.is_active()) {
    GenericObject* obj = SessionManager::sm->GetSharedSession()->GetObject(subcomp->second.get_subid());
    if (obj) return; // we have object, nothing to do
    
    if (!AduManager::am->StateReceived(subcomp->second.get_subid())) 
      // obj is active and not available -> request state
      QueryState(subcomp->second.get_subid(), 2); // priority 2 for late join
  }
  #endif
}


void MLB::CommunicationManager::subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason) {
  #ifdef SIMPLE_LJ
  if (subcomp->second.is_active()) {
    GenericObject* obj = SessionManager::sm->GetSharedSession()->GetObject(subcomp->second.get_subid());
    if (obj) return; // we have object, nothing to do
    
    if (!AduManager::am->StateReceived(subcomp->second.get_subid()))
      // obj is active and not available -> request state
      QueryState(subcomp->second.get_subid(), 2); // priority 2 for late join
  }
  #endif
}


void MLB::CommunicationManager::subcomponent_removed(uint32 id) {
}


void MLB::CommunicationManager::participant_added(uint32 pid) {
  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::participant_added - new participant with pid %u\n", pid);
  #endif

  if (!rtpi->participant_exists(pid)) return; // not a valid participant

  ParticipantManager::pm->AddParticipant(pid);
}


void MLB::CommunicationManager::participant_removed(uint32 pid) {
  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::participant_removed - remove participant with pid %u\n", pid);
  #endif

  if (!rtpi->participant_exists(pid)) return; // not a valid participant

  ParticipantManager::pm->RemoveParticipant(pid);
}


void MLB::CommunicationManager::participant_changed(uint32 pid) {
  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::participant_changed - change participant with pid %u\n", pid);
  #endif

  RTPI::RtpiParticipant* participant = rtpi->get_participant(pid);
  if (!participant) return; // not a valid participant

  ParticipantManager::pm->UpdateParticipant(pid, participant->get_cname(), participant->get_name(), participant->get_email(),
					    participant->get_phone(), participant->get_loc(), participant->get_tool());
}


void MLB::CommunicationManager::RegisterSubcomponent(uint32 subID, int type) {
  if (!openConnection) return;

  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::RegisterSubcomponent - register sub with id %u\n", subID);
  #endif

  char tmp[3]; sprintf(tmp,"%u",type); std::string name  = tmp;
  rtpi->add_subcomponent(subID, name, true);
}


void MLB::CommunicationManager::UnregisterSubcomponent(uint32 subID) {
  if (!openConnection) return;

  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::UnregisterSubcomponent - unregister sub with id %u\n", subID);
  #endif

  rtpi->delete_subcomponent(subID);
}


void MLB::CommunicationManager::ChangeSubcomponentActiveness(uint32 subID, bool active) {
  if (!openConnection) return;

  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::ChangeSubcomponentActiveness - set sub with id %u active = %u\n", subID, active);
  #endif

  if (active) rtpi->activate_subcomponent(subID);
  else rtpi->deactivate_subcomponent(subID);
}


void MLB::CommunicationManager::DeactivateAllSubcomponents() {
  if (!openConnection) return;

  #ifdef DEBUG_RTCPI
  GUI::MessageWindow::Info("CommunicationManager::DeactivateAllSubcomponents\n");
  #endif

  rtpi->deactivate_all_subcomponents();
}


uint32 MLB::CommunicationManager::GetObjectID() {
  #ifdef IDSERVER
  if (localIDServer)
    return ++localID;

  if (!openIDServer) {
    GUI::MessageWindow::Error("CommunicationManager::GetObjectID - Connection to ID-Server is not available\n");
    return 0;
  }

  static uint32 id;
  static std::vector<uint32>* ids;
  static bool first = true;

  if (first) {
    ids = uniqueIdClient->GetIds("object", 100);

    if ((ids == NULL) || ids->empty()) {
      GUI::MessageWindow::Error("CommunicationManager::GetObjectID - Connection to ID-Server is not available\n");
      return 0;
    }

    first = false;
  }

  if (ids->empty()) {
    delete ids;
    ids = uniqueIdClient->GetIds("object", 100);

    if ((ids == NULL) || ids->empty()) {
      GUI::MessageWindow::Error("CommunicationManager::GetObjectID - Connection to ID-Server is not available\n");
      return 0;
    }
  }

  id = ids->back();
  ids->pop_back();

  if (id == 0) { // reserved for special use: documents id
    id = ids->back();
    ids->pop_back();
  }

  if (id == 1) { // reserved for special use: shared document id
    id = ids->back();
    ids->pop_back();
  }

  return id;
  #else
  return random32();
  #endif
}


#ifdef IDSERVER
bool MLB::CommunicationManager::StartIdClient(std::string ip, int port) {
  if (uniqueIdClient != 0) return true;
  unsIP = ip; unsPort = port;
    
  try {
    uniqueIdClient = new RTPI::UniqueIdClient(ip,port);
  } catch (RTPI::RtpiException e) {
    return false;
  }

  return uniqueIdClient->Alive();
}
#endif


void MLB::CommunicationManager::ExecuteTimer(TimerObject* timer) {
  switch (timer->type) {
  case TimerObject::STATE_REPLY: {
    AnswerStateQuery(timer->id, timer->priority);
    break;
  }
  case TimerObject::STATE_QUERY: {
    SendStateQuery(timer->id, timer->priority);
    break;
  }
  }

  delete timer;
}


bool MLB::CommunicationManager::StateIncludesEvent(RTPI::RtpiState* state, uint32 pid, uint16 snr) {
  std::vector<uint16>::iterator j = state->GetAduHeader()->sns.begin();
  for (std::vector<uint32>::iterator i = state->GetAduHeader()->pids.begin(); i != state->GetAduHeader()->pids.end(); ++i) {
    if ((*i == pid) && (*j == snr)) return true;
    ++j;
  }
  
  return false;
}


bool MLB::CommunicationManager::UpdateEventSequenceNumber(uint32 objID, uint32 pid, uint16 sn) {
  std::map<uint32, uint16> objSeqNrs = seqNumbers[objID]; // implicit insert if not exists
  objSeqNrs[pid] = sn;
  return true;

  /*
  std::map<uint32, uint16>::iterator pos = objSeqNumbers.find(pid);
  if (pos == seqNumbers.end()) {
    // first event
    objSeqNumbers[pid] = sn;
    return true;
  }
  
  if (pos->second == (sn + 1)) {
    ++(pos->second);
    return true;
  }
  
  // found gap
  return false;
  */
}


void MLB::CommunicationManager::ResetEventSequenceNumber(uint32 objID, uint32 pid) {
  std::map<uint32, std::map<uint32, uint16> >::iterator pos = seqNumbers.find(objID);
  if (pos == seqNumbers.end()) return;
  pos->second.erase(pid);
}



RefCount<RTPI::RtpiAduList> MLB::CommunicationManager::GetConflictingAdus(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAduList> pcAdus) {
  RefCount<RTPI::RtpiAduList> cAdus = new RTPI::RtpiAduList;

  for (RTPI::RtpiAduList::iterator i = pcAdus->begin(); i != pcAdus->end(); ++i) {
    if (Conflict(lateAdu, *i))
      cAdus->push_back(*i);
  }
  
  return cAdus;
}


bool MLB::CommunicationManager::Conflict(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAdu> adu) {
  if (lateAdu->GetParticipantID() == adu->GetParticipantID()) return false;

  int opT   = ParseOperationType(adu->GetPayload());
  int lopT  = ParseOperationType(lateAdu->GetPayload());

  int objT  = ParseObjectType(adu->GetPayload());
  int lobjT = ParseObjectType(lateAdu->GetPayload());

  int subT  = ParseSubType(adu->GetPayload());
  int lsubT = ParseSubType(lateAdu->GetPayload());
  
  uint32 pID;
  GenericContainer* p;
  if (opT == ADU_STATE) {
    pID = ((RTPI::RtpiStateHeader*)adu->GetAduHeader())->GetObjectID();
  } else {
    p = SessionManager::sm->GetParentContainer(adu->GetSubcomponentID());
    if (!p) return false;
    pID = p->GetID();
  }
  uint32 lpID;
  if (lopT = ADU_STATE) {
    lpID = ((RTPI::RtpiStateHeader*)lateAdu->GetAduHeader())->GetObjectID();
  } else {
    p = SessionManager::sm->GetParentContainer(lateAdu->GetSubcomponentID());
    if (!p) return false;
    lpID = p->GetID();
  }

  if (lateAdu->GetSubcomponentID() == adu->GetSubcomponentID()) {
    if (opT  == ADU_STATE) return true;    
    if (lopT == ADU_STATE) return true;
    if ((lopT == ADU_CHANGE) && (opT == ADU_CHANGE)) {
      if (subT == lsubT) return true;
      if (((objT == POLYGON) || (objT == POLYLINE)) 
	  && (((subT == MOVE_POINT) || (subT == ADD_POINT)) && ((lsubT == MOVE_POINT) || (lsubT == ADD_POINT)))) 
	return true;
      if ((objT == TEXT) 
	  && (((subT == CHG_FONT) || (subT == CHG_LINECOLOR) || (subT == INSERT_CHARS) || (subT == DELETE_CHARS) || (subT == CHG_ACTIVE)) 
	      && ((lsubT == CHG_FONT) || (lsubT == CHG_LINECOLOR) || (lsubT == INSERT_CHARS) || (lsubT == DELETE_CHARS) || (lsubT == CHG_ACTIVE)))) 
	return true;
    }
    if (opT == ADU_DELETE) return true;
    return false;
  }

  if (pID == lpID) {
    if ((lopT == ADU_STATE) && (opT == ADU_STATE)) return true;

    if ((lopT == ADU_STATE) && (opT == ADU_CHANGE) && (subT == RAISE)) return true;
    if ((opT == ADU_STATE) && (lopT == ADU_CHANGE) && (lsubT == LOWER)) return true;

    if ((lopT == ADU_STATE) && (opT == ADU_CHANGE) && (subT == CHG_PARENT) && (pID == lpID)) return true;
    if ((opT == ADU_STATE) && (lopT == ADU_CHANGE) && (lsubT == CHG_PARENT) && (pID == lpID)) return true;

    return false;
  }

  if ((opT == ADU_CHANGE) && (subT == CHG_PARENT) && (lopT == ADU_CHANGE) && (lsubT == CHG_PARENT) && (pID == lpID)) return true;

  if ((opT == ADU_CHANGE) && (subT == SET_ACTIVE) && (lopT == ADU_CHANGE) && (lsubT == SET_ACTIVE)) return true;

  return false;
}


bool MLB::CommunicationManager::OverwritingAdu(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAduList> cAdus) {
  for (RTPI::RtpiAduList::iterator i = cAdus->begin(); i != cAdus->end(); ++i) {
    if (Overwriting(lateAdu, *i)) return true;
  }

  // did not find an overwriting adu
  return false;
}


bool MLB::CommunicationManager::Overwriting(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAdu> adu) {
  int opT   = ParseOperationType(adu->GetPayload());
  int lopT  = ParseOperationType(lateAdu->GetPayload());

  int objT  = ParseObjectType(adu->GetPayload());
  int lobjT = ParseObjectType(lateAdu->GetPayload());

  int subT  = ParseSubType(adu->GetPayload());
  int lsubT = ParseSubType(lateAdu->GetPayload());
  
  if (lateAdu->GetSubcomponentID() == adu->GetSubcomponentID()) {
    if ((opT == ADU_STATE) && (lopT == ADU_CHANGE)) return true; 
    //if ((adu->GetType() == RTPI::RtpiHeader::PKT_STATE) && (lopT == ADU_CHANGE) 
    //&& StateIncludesEvent((RtpiState*)&(*adu), lateAdu->GetParticipantID(), lateAdu->GetPacketHeader()->GetSequenceNumber())) return true;

    if ((lopT == ADU_CHANGE) && (opT == ADU_CHANGE) && (subT == lsubT)) {
      if (((objT == POLYGON) || (objT == POLYLINE)) && ((subT == ADD_POINT) || (lsubT == ADD_POINT))) return false;
      if ((objT == TEXT) 
	  && (((subT == CHG_FONT) || (subT == CHG_LINECOLOR) || (subT == INSERT_CHARS) || (subT == DELETE_CHARS) || (subT == CHG_ACTIVE)) 
	      && ((lsubT == CHG_FONT) || (lsubT == CHG_LINECOLOR) || (lsubT == INSERT_CHARS) || (lsubT == DELETE_CHARS) || (lsubT == CHG_ACTIVE)))) 
	return false;
      return true;
    }

    if (opT == ADU_DELETE) return true;
  }

  if ((opT == ADU_CHANGE) && (subT == SET_ACTIVE) && (lopT == ADU_CHANGE) && (lsubT == SET_ACTIVE)) return true;

  return false;
}


bool MLB::CommunicationManager::IntentionViolation(RefCount<RTPI::RtpiAdu> adu1, RefCount<RTPI::RtpiAdu> adu2) {
  if (adu1->GetParticipantID() == adu2->GetParticipantID()) return false;

  int opT1  = ParseOperationType(adu1->GetPayload());
  int opT2  = ParseOperationType(adu2->GetPayload());

  int objT1 = ParseObjectType(adu1->GetPayload());
  int objT2 = ParseObjectType(adu2->GetPayload());

  int subT1 = ParseSubType(adu1->GetPayload());
  int subT2 = ParseSubType(adu2->GetPayload());

  GenericContainer* p;

  uint32 pID1;
  if (opT1 == ADU_STATE) {
    pID1 = ((RTPI::RtpiStateHeader*)adu1->GetAduHeader())->GetObjectID();
  } else {
    p = SessionManager::sm->GetParentContainer(adu1->GetSubcomponentID());
    if (!p) return false;
    pID1 = p->GetID();
  }
  uint32 pID2;
  if (opT2 == ADU_STATE) {
    pID2 = ((RTPI::RtpiStateHeader*)adu2->GetAduHeader())->GetObjectID();
  } else {
    p = SessionManager::sm->GetParentContainer(adu2->GetSubcomponentID());
    if (!p) return false;
    pID2 = p->GetID();
  }
 
  if (adu1->GetSubcomponentID() == adu2->GetSubcomponentID()) {
    if ((opT1 == ADU_DELETE) || (opT2 == ADU_DELETE)) return true;
    if ((opT1 == ADU_CHANGE) && (opT2 == ADU_CHANGE)) {
      if (subT1 == subT2) return true;
      if ((objT1 == TEXT) && ((subT1 == INSERT_CHARS) || (subT1 == DELETE_CHARS) || (subT1 == CHG_ACTIVE)) 
	                  && ((subT2 == INSERT_CHARS) || (subT2 == DELETE_CHARS) || (subT2 == CHG_ACTIVE)))
	return true;
    }
    return false;
  }
  
  if (pID1 == pID2) {
    if ((opT1 == ADU_STATE) && (opT2 == ADU_STATE)) return true;
    if ((opT1 == ADU_STATE) && (opT2 == ADU_CHANGE) && (subT2 == RAISE)) return true;
    if ((opT2 == ADU_STATE) && (opT1 == ADU_CHANGE) && (subT1 == RAISE)) return true;
    // not if subT == LOWER: only raise can conflict since new objects are inserted at the top
  }

  if ((opT1 == ADU_CHANGE) && (opT2 == ADU_CHANGE) && (subT1 == CHG_PARENT) && (subT2 == CHG_PARENT) && (pID1 == pID2)) return true;

  if ((opT1 == ADU_CHANGE) && (opT2 == ADU_CHANGE) && (subT1 == SET_ACTIVE) && (subT2 == SET_ACTIVE)) return true;
  
  return false;
}


void MLB::CommunicationManager::DisplayIntentionConflicts(uint32 objID, RefCount<RTPI::RtpiAduList> icAdus1, RefCount<RTPI::RtpiAduList> icAdus2) {
  // 1. get id's of conflicting participants
  std::list<uint32> conflictParticipants;
  RTPI::RtpiAduList::iterator ic;
  for (ic = icAdus1->begin(); ic != icAdus1->end(); ++ic) conflictParticipants.push_back((*ic)->GetParticipantID());
  for (ic = icAdus2->begin(); ic != icAdus2->end(); ++ic) conflictParticipants.push_back((*ic)->GetParticipantID());
  conflictParticipants.sort();
  conflictParticipants.unique(); // each participant only once

  // 2. build name list
  Participant* p;
  for (std::list<uint32>::iterator pids = conflictParticipants.begin(); pids != conflictParticipants.end(); ++pids) {
    p = ParticipantManager::pm->GetParticipant(*pids);
    if (!(p->name.empty())) {
      if (Et_GlobalEvalF(Et_Interp, "lappend pList %q", p->name.c_str()) != TCL_OK)
	GUI::MessageWindow::Error("CommunicationManager::DisplayIntentionConflicts - tcl error \"%q\"\n", Et_Interp->result);
    } else { // cname always there
      if (Et_GlobalEvalF(Et_Interp, "lappend pList %q", p->cname.c_str()) != TCL_OK)
	GUI::MessageWindow::Error("CommunicationManager::DisplayIntentionConflicts - tcl error \"%q\"\n", Et_Interp->result);
    }
  }

  // 3. get concerend object and hint display coordinates
  GenericObject* obj = SessionManager::sm->GetSharedSession()->GetObject(objID);
  if (!obj) obj = SessionManager::sm->GetSharedSession()->GetGroup(objID);
  if (!obj) return;
  int x = 0, y = 0; 
  if ((obj->GetType() != GenericObject::FOLDER) && (obj->GetType() != GenericObject::PAGE)) {
    x = ((DrawObject*)obj)->GetX();
    y = ((DrawObject*)obj)->GetY();
  }

  // 4. get parent page
  Page* page = (Page*)SessionManager::sm->GetParentContainer(objID);
  if (!page) return;

  // 4. display conflict
  if (Et_GlobalEvalF(Et_Interp, "Canvas::DisplayConflictHint \"%q\" %u %u $pList", 
		     page->GetCanvas()->GetWidget(), 
		     int(x * SessionManager::sm->GetZoom()), 
		     int(y * SessionManager::sm->GetZoom())) != TCL_OK)
    GUI::MessageWindow::Error("CommunicationManager::DisplayIntentionConflicts - tcl error \"%q\"\n", Et_Interp->result);
  if (Et_GlobalEvalF(Et_Interp,"unset pList") != TCL_OK) 
    GUI::MessageWindow::Error("CommunicationManager::DisplayIntentionConflicts - tcl error \"%q\"\n", Et_Interp->result);
}

