/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_communication.h
 authors : Juergen Vogel
 date    :
 desc    :

**********************************************************************************************************/


#ifndef __MLB_COMMUNICATION_H__
#define __MLB_COMMUNICATION_H__

#define MLB_MASK_V    0xC0
#define MLB_MASK_R    0x20
#define MLB_MASK_OPT  0x1C
#define MLB_MASK_VIS  0x02
#define MLB_MASK_T    0x01
#define MLB_MASK_CST  0xC0
#define MLB_MASK_JST  0x30
#define MLB_MASK_AST  0x0C
#define MLB_MASK_SW   0x02
#define MLB_MASK_SH   0x01
#define MLB_MASK_S    0x01
#define MLB_MASK_FWG  0xC0
#define MLB_MASK_FSL  0x30
#define MLB_MASK_FPS  0x0C
#define MLB_MASK_A    0x01


#define xIDSERVER

#include <rtpi/rtpi.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpihandler.h>
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpiInterfaces.h>
#include <rtpi/rtpiTransport.h>
#include <rtpi/rtpiSeqNumber.h>
#include <util/types.h>
#include <util/Buffer.h>

#ifdef IDSERVER
#include <rtpi/rtpiUniqueId.h>
#endif

#include <tcl.h>
#include <tk.h>
#include <time.h>
#include <stdlib.h>

#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <stdio.h>

#include <vector>
#include <map>
#include <string>

#include "SMP.h"
#include "general.h"

#include "mlb_timermanager.h"
#include "mlb_generic.h"


namespace HRT {
  class HrtManager;
}

namespace CHAT {
  class ChatManager;
}

namespace FT {
  class FtManager;
}

namespace ALT {
  class AltManager;
}

namespace VT {
  class VtManager;
}

namespace MLB {
  //  using namespace RTPI;
#ifdef WIN32
  using std::vector<uint32>;
#endif

  class PendingAdu {
  public:
    PendingAdu(RefCount<RTPI::RtpiAdu>_adu) : count(0), adu(_adu) {}
    uint16 count;
    RefCount<RTPI::RtpiAdu> adu;
  };


  class CommunicationManager : public RTPI::RtcpiReportInterface, public RTPI::RtpiFeedbackInterface, public TimerFeedbackInterface {
  public:
    enum types {
      // general
      VERSION = 0,
      APPLICATION_TYPE = 1, // mlb

      // operation Types
      ADU_STATE  = 0,
      ADU_CHANGE = 1,
      ADU_DELETE = 2,

      // ADU types
      SESSION     = 0,
      FOLDER      = 1,
      PAGE        = 2,
      RECTANGLE   = 4,
      OVAL        = 5,
      LINE        = 6,
      POLYLINE    = 7,
      POLYGON     = 8,
      TEXT        = 9,
      IMAGE       = 10,
      GROUP       = 11,

      // subtypes
      CHG_VISIBILITY  = 0,
      MOVE            = 1,
      CHG_SIZE        = 2,
      CHG_LINEWIDTH   = 3,
      CHG_LINESTYLE   = 4,
      CHG_LINECOLOR   = 5,
      CHG_FILLCOLOR   = 6,
      MOVE_POINT_1    = 7,
      MOVE_POINT_2    = 8,
      CHG_CAPSTYLE    = 9,
      CHG_JOINSTYLE   = 10,
      CHG_ARROWSTYLE  = 11,
      ADD_POINT       = 12,
      CHG_TYPE        = 13,
      MOVE_POINT      = 14,
      INSERT_FONT     = 31,
      INSERT_FCOLOR   = 32,
      CHG_FONT        = 30,
      CHG_FONTFAMILY  = 15,
      CHG_FONTSIZE    = 16,
      CHG_FONTWEIGHT  = 17,
      CHG_FONTSLANT   = 18,
      CHG_FONTPOS     = 19,
      INSERT_CHARS    = 20,
      DELETE_CHARS    = 21,
      CHG_BACKGRCOLOR = 22,
      CHG_NAME        = 23,
      SET_ACTIVE      = 24,
      RAISE           = 25,
      LOWER           = 26,
      CHG_PARENT      = 27,
      CHG_ACTIVE      = 28,
      CLOSE_POLYGON   = 29,

      // image subtypes
      GIF             = 0,
      PNG             = 1
    };

    CommunicationManager();
    ~CommunicationManager();

    void Shutdown();

    bool GetCheckIntentionViolation() {return checkIntentionViolation;}
    void SetCheckIntentionViolation(bool check) {checkIntentionViolation = check;}

    void SetParticipantName(char* _name);
    void SetParticipantEmail(char* _email);
    void SetParticipantPhone(char* _phone);
    void SetParticipantLocation(char* _location);

    std::string GetSessionIP() {return ip;}

    bool OpenSMP();
    bool CloseSMP();

    bool OpenRTPI(std::string _ip, int _port, int _ttl, int _bandwidth);
    bool CloseRTPI();

    bool OpenConnection() {return openConnection;}

    // outgoing data
    RTPI::RtpiAdu* CreateState(uint32 subID, uint32 objectID, int type, int priority, RefCount<Buffer> buf);
    RTPI::RtpiAdu* CreateState(uint32 subID, uint32 objectID, int type, int priority, RefCount<Buffer> buf, uint32 timestamp);
    RTPI::RtpiAdu* CreateEvent(uint32 subID, RefCount<Buffer> buf);
    RTPI::RtpiAdu* CreateEvent(uint32 subID, uint32 objectID, RefCount<Buffer> buf);
    RTPI::RtpiAdu* CreateCue(uint32 subID, RefCount<Buffer> buf);
    RTPI::RtpiAdu* CreateStateQuery(uint32 subID, int priority);

    void SendAduReliable(RefCount<RTPI::RtpiAdu> adu);
    void SendAdusReliable(RefCount<RTPI::RtpiAduVector> adus);
    void SendAduUnreliable(RefCount<RTPI::RtpiAdu> adu);
    void SendAdusUnreliable(RefCount<RTPI::RtpiAduVector> adus);

    // handle incoming data
    void HandleReceivedRtpi(SMP::Indication ind);
    void HandleReceivedRtcpi(SMP::Indication ind);

    void QueryState(uint32 subID, int priority);
    void SendStateQuery(uint32 subID, int priority);
    void HandleStateQuery(RTPI::RtpiStateQuery* query);
    void AnswerStateQuery(uint32 subID, int priority);

    // functions implemented due to  RtcpiReportInterface
    void    send_rtcpi(RefCount<Buffer>);
    u_int32 get_local_address(int i=0);
    int     addr_is_local(u_int32);
    u_int32 get_session_bandwith();
    u_int32 get_new_pid();
    u_int32 get_new_subid();
    std::string  get_session_name(std::string ="") { return name;}
    std::string  get_session_email()          { return email;}
    std::string  get_session_phone()          { return phone;}
    std::string  get_session_loc()            { return location; }
    std::string  get_session_note()           { return ""; }
    std::string  get_session_tool()           { return "mlb 0.9.1"; }

    // functions implemented due to RtpiFeedbackInterface
    void subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason);
    void subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason);
    void subcomponent_removed(uint32 id);
    void participant_added(uint32 pid);
    void participant_removed(uint32 pid);
    void participant_changed(uint32 pid);

    // helper functions
    void RegisterSubcomponent(uint32 subID, int type);
    void UnregisterSubcomponent(uint32 subID);
    void ChangeSubcomponentActiveness(uint32 subID, bool active);
    void DeactivateAllSubcomponents();

    // id functions
    #ifdef IDSERVER
    bool OpenIDServer(std::string _unsIP, int _unsPort);
    bool OpenIDServer();
    bool CloseIDServer();
    bool StartIdClient(std::string ip, int port);
    #endif
    uint32 GetObjectID();
    uint32 GetParticipantID() {return pid;}

    // parse payload of ADUS
    int ParseOperationType(char* buf) {return (((uint8*)buf)[0] & MLB_MASK_OPT) >> 2;}
    int ParseSubType(char* buf)       {return ((uint8*)buf)[2];}
    int ParseObjectType(char* buf)    {return ((uint8*)buf)[1];}

    // manage sequence numbers included in sent states
    bool UpdateEventSequenceNumber(uint32 objID, uint32 pid, uint16 snr);  // return false if consistency problem
    void ResetEventSequenceNumber(uint32 objID, uint32 pid);

    RTPI::Rtpi* GetRtpi() {return rtpi;}

    // static members
    static void SetCueRate(double cueRate) {cueSendRate = cueRate;}

    static const unsigned int T;
    static double cueSendRate;

    static CommunicationManager* cm;

  private:
    friend class HRT::HrtManager;
    friend class CHAT::ChatManager;
    friend class FT::FtManager;
    friend class ALT::AltManager;
    friend class VT::VtManager;
    friend class TelepointerManager;
    friend class AduManager;
    friend class TimeWarp;

    // UDP
    std::string localhost;
    std::vector<uint32> localAddresses;
    bool openConnection;
    std::string ip;
    int port;
    int ttl;
    int bandwidth;

    // SMP
    bool openSMP;
    SMP::SMP* smp;
    int rtpiSocket;
    int rtcpiSocket;

    // Unique Number Server
    bool openIDServer;
    #ifdef IDSERVER
    std::string unsIP;
    int    unsPort;
    RTPI::UniqueIdClient* uniqueIdClient;

    bool   localIDServer;
    uint32 localID;
    #endif

    // RTP/I
    uint32 pid;
    std::string username;
    std::string name;
    std::string email;
    std::string phone;
    std::string location;

    RTPI::Rtpi* rtpi;

    // manage sequence numbers included in sent states
    bool StateIncludesEvent(RTPI::RtpiState* state, uint32 pid, uint16 snr);
    std::map<uint32, std::map<uint32, uint16> > seqNumbers;           // <objID, PID, SNr>

    // consistency control
    RefCount<RTPI::RtpiAduList> GetConflictingAdus(RefCount<RTPI::RtpiAdu>, RefCount<RTPI::RtpiAduList>);
    bool Conflict(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAdu> adu);
    bool OverwritingAdu(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAduList> cAdus);
    bool Overwriting(RefCount<RTPI::RtpiAdu> lateAdu, RefCount<RTPI::RtpiAdu> adu);
    bool IntentionViolation(RefCount<RTPI::RtpiAdu> adu1, RefCount<RTPI::RtpiAdu> adu2);
    void DisplayIntentionConflicts(uint32 objID, RefCount<RTPI::RtpiAduList> icAdus1, RefCount<RTPI::RtpiAduList> icAdus2);
    bool checkIntentionViolation;

    // function implemented due to TimerFeedbackInterface
    void ExecuteTimer(TimerObject* timer);
  };
}


#endif
