/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_font.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 19.06.2000
 desc    : functions to create TrueType-fonts

**********************************************************************************************************/

#include "mlb_font.h"
#include "mlb_gui.h"

#define MaxGlyphs  65535

char *MLB::FontList::Fonts[] = {"arial.tt2", "arialbd.tt2", "ariali.tt2", "arialbi.tt2",
				"times.ttf", "timesbd.ttf", "timesi.ttf", "timesbi.ttf",
				"arialn.ttf", "arialnb.ttf", "arialni.ttf", "arialnbi.ttf",
				"lucidatypewriterregular.ttf", "lucidatypewriterbold.ttf","lucidatypewriteroblique.ttf","lucidatypewriterboldoblique.ttf",
				"symbol.ttf", "symbol.ttf", "symbol.ttf", "symbol.ttf",};

typedef struct _SegmentInfo {
  double
    x1,
    y1,
    x2,
    y2;
} SegmentInfo;


MLB::Font::Font(char *_filename,TT_Face _face, TT_Face_Properties _face_properties, unsigned int _size, double _zoom) {
  unsigned short
    x_resolution,
    y_resolution,
    encoding,
    platform;
  register int
    i;
  TT_Error
    error;

  filename = _filename;

  face = _face;
  face_properties = _face_properties;
  size = _size;
  zoom = _zoom;

  x_resolution = 64 * zoom;
  y_resolution = 64 * zoom;

  if (error=TT_New_Instance(face,&instance)) {
    GUI::MessageWindow::Error("Font::Font - cant create instance\n");
    return;
  }

  error |= TT_Set_Instance_Resolutions(instance,x_resolution,y_resolution);
  error |= TT_Set_Instance_CharSize(instance,(int) (64.0*size));
  if (error) {
    GUI::MessageWindow::Error("Font::Font - Cannot initialize TTF instance\n");
    return;
  }


  TT_Get_Instance_Metrics(instance,&instance_metrics);

  for (code=0; (int) code < (int)face_properties.num_CharMaps; code++) {
    TT_Get_CharMap_ID(face,code,&platform,&encoding);
    if (((platform == 3) && (encoding == 1)) ||
        ((platform == 0) && (encoding == 0))) {
      TT_Get_CharMap(face,code,&char_map);
      break;
    }
  }
  number_glyphs = 0;
  character_map = True;
  if (code == face_properties.num_CharMaps) {
    number_glyphs = face_properties.num_Glyphs;
    character_map = False;
  }
  glyphs = new TT_Glyph[MaxGlyphs]; 
  if (glyphs == (TT_Glyph *) NULL)
    GUI::MessageWindow::Error("Font::Font - Memory allocation failed\n");
  for (i=0; i < MaxGlyphs; i++)
    glyphs[i].z=(TT_Glyph *) NULL;

  ascent =  (instance_metrics.y_ppem*face_properties.horizontal->Ascender)/
    face_properties.header->Units_Per_EM+1;
  descent = (-instance_metrics.y_ppem*face_properties.horizontal->Descender)/
    face_properties.header->Units_Per_EM+1;
  /***/
}

int MLB::Font::Ascent()
{
  return ascent;
}

int MLB::Font::Descent()
{
  return descent;
}

char *MLB::Font::GetFilename()
{
  return filename;
}

unsigned char *MLB::Font::RenderGlyph(TT_Glyph glyph,int x_offset,int y_offset, int width, int height, TT_Glyph_Metrics *glyph_metrics)
{ 
  int
    y;

  register int
    i,
    x;

  register unsigned char
    *p,
    *q;

  unsigned char 
    *bitmap;

  TT_Raster_Map 
    character;
  SegmentInfo
    bounds;
  TT_F26Dot6  xmin, ymin;


  /*
    Render Glyph.
  */
  character.rows   = instance_metrics.y_ppem*(face_properties.horizontal->Ascender-
					      face_properties.horizontal->Descender) / 
                                              face_properties.header->Units_Per_EM;
  character.width  = (instance_metrics.x_ppem+32+3) & -4;
  character.flow   = TT_Flow_Down;
  character.cols   = character.width;
  character.size   = character.rows*character.width;
  character.bitmap = new char[character.size];

  if (!character.bitmap)
    GUI::MessageWindow::Error("Font::RenderGlyph - Memory allocation failed\n");

  q = (unsigned char *)character.bitmap;
  for (i=0; i < character.size; i++)
    *q++ = 0;

  bitmap = new unsigned char[width*height];
  p = bitmap;
  for (i=0; i < width*height; i++)
    *p++ = 0;

  xmin = glyph_metrics->bbox.xMin & -64;
  ymin = glyph_metrics->bbox.yMin & -64;
  TT_Get_Glyph_Pixmap(glyph,&character,-xmin,-ymin);
  /*
    Composite character on canvas.
  */
  x_offset += (glyph_metrics->bbox.xMin & -64)/64;
  y_offset = (-(glyph_metrics->bbox.yMin & -64)/64)-y_offset;
  bounds.x1 = x_offset < 0 ? -x_offset : 0;
  bounds.y1 = y_offset < 0 ? -y_offset : 0;
   bounds.x2 = width-x_offset;
  if (bounds.x2 > character.cols)
    bounds.x2 = character.cols;
  bounds.y2 = height-y_offset;
  if (bounds.y2 > character.rows)
    bounds.y2 = character.rows;
  if (bounds.x1 >= bounds.x2)
    return NULL;


  for (y=(int)bounds.y1; y < (int)bounds.y2; y++)
  {
    p = ((unsigned char *)character.bitmap)+y*character.cols+(int) bounds.x1;
    q =  bitmap+(y+y_offset)*width+(int) bounds.x1+x_offset;
    for (x=(int) bounds.x1; x < bounds.x2; x++) {
      *q = *p;
      q++; p++;
    }
  }

  delete[] character.bitmap;
  return bitmap;
}

TT_F26Dot6  xmin, ymin, xmax, ymax;


static void RenderBitmapGlyph(TT_Raster_Map *canvas,TT_Raster_Map *character,
  TT_Glyph glyph,int x_offset,int y_offset,TT_Glyph_Metrics *glyph_metrics)
{
  int
    y;

  register int
    i,
    x;

  register unsigned char
    *p,
    *q;

  SegmentInfo
    bounds;
  TT_F26Dot6  xmin, ymin;


  /*
    Render Glyph.
  */
  q = (unsigned char *)character->bitmap;
  for (i=0; i < character->size; i++)
    *q++ = 0;

  xmin = glyph_metrics->bbox.xMin & -64;
  ymin = glyph_metrics->bbox.yMin & -64;
  TT_Get_Glyph_Pixmap(glyph,character,-xmin,-ymin);
  /*
    Composite character on canvas.
  */
  x_offset += (glyph_metrics->bbox.xMin & -64)/64;
  y_offset = (-(glyph_metrics->bbox.yMin & -64)/64)-y_offset;
  bounds.x1 = x_offset < 0 ? -x_offset : 0;
  bounds.y1 = y_offset < 0 ? -y_offset : 0;
   bounds.x2 = canvas->cols-x_offset;
  if (bounds.x2 > character->cols)
    bounds.x2 = character->cols;
  bounds.y2 = canvas->rows-y_offset;
  if (bounds.y2 > character->rows)
    bounds.y2 = character->rows;
  if (bounds.x1 >= bounds.x2)
    return;
  for (y=(int)bounds.y1; y < (int)bounds.y2; y++)
  {
    p = ((unsigned char *)character->bitmap)+y*character->cols+(int) bounds.x1;
    q = ((unsigned char *)canvas->bitmap)+(y+y_offset)*canvas->cols+
      (int) bounds.x1+x_offset;
    for (x=(int) bounds.x1; x < bounds.x2; x++)
      *q++|=(*p++);
  }
}


std::string MLB::Font::GetCharBitmap(char ascii, int *width, int *ascent, int *descent, int *lbearing, TT_BBox *bbox)
{
  TT_Error
    error;
  TT_Raster_Map 
    canvas,
    character;
  TT_Glyph_Metrics
    glyph_metrics;
  std::string
    bitmapData;
  unsigned short
    unicode;
  register int
    i, j, x;
  int
    y;
  char
    buf[20];
  unsigned char
    value,
    *p, bitChar;
  int 
    border;

  unicode = (unsigned char)ascii;
  
  if (glyphs[unicode].z == (TT_Glyph *) NULL) {
    if (character_map) {
      code = TT_Char_Index(char_map,unicode);
    }
    else {
      code = ((int) unicode-' '+3) < 0 ? 0 : ((int) unicode-' '+3);
      if ((int) code >= number_glyphs)
	code = 0;
    }
    error = TT_New_Glyph(face,&glyphs[unicode]);
    error|= TT_Load_Glyph(instance,glyphs[unicode],code,
			  TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH);
    if (error)
      GUI::MessageWindow::Error("Font::GetCharBitmap - Cannot initialize TTF glyph");
  }

  border = 20;
  *lbearing = border/2;
  canvas.width = border;
  if (glyphs[unicode].z != (TT_Glyph *) NULL) {
    TT_Get_Glyph_Metrics(glyphs[unicode],&glyph_metrics);
    canvas.width += (glyph_metrics.bbox.xMax-glyph_metrics.bbox.xMin)/64;
    //    TT_Get_PS_Name(face,code,&PSName);
    //    printf ("%s\n",PSName);
  }
  canvas.rows = instance_metrics.y_ppem*(face_properties.horizontal->Ascender-
					 face_properties.horizontal->Descender) / 
                                         face_properties.header->Units_Per_EM;   

  canvas.flow = TT_Flow_Down;
  canvas.cols = canvas.width;
  canvas.size = canvas.rows*canvas.width;
  canvas.bitmap = new char[canvas.size];
  if (!canvas.bitmap)
    GUI::MessageWindow::Error("Font::GetCharBitmap - Memory allocation failed\n");

  p = (unsigned char *) canvas.bitmap;
  for (i=0; i < canvas.size; i++)
    *p++ = 0;
  character.rows   = canvas.rows;
  character.width  = (instance_metrics.x_ppem+32+3) & -4;
  character.flow   = TT_Flow_Down;
  character.cols   = character.width;
  character.size   = character.rows*character.width;
  character.bitmap = new char[character.size];

  if (!character.bitmap)
    GUI::MessageWindow::Error("Font::GetCharBitmap - Memory allocation failed\n");

  x = *lbearing;
  y = (-instance_metrics.y_ppem*face_properties.horizontal->Descender)/
    face_properties.header->Units_Per_EM+1;

  if (glyphs[unicode].z != (TT_Glyph *) NULL) {
    TT_Get_Glyph_Metrics(glyphs[unicode],&glyph_metrics);
    RenderBitmapGlyph(&canvas,&character,glyphs[unicode],x,y,&glyph_metrics);
  }

  bbox->xMin = glyph_metrics.bbox.xMin/64;
  bbox->yMin = glyph_metrics.bbox.yMin/64;
  bbox->xMax = glyph_metrics.bbox.xMax/64;
  bbox->yMax = glyph_metrics.bbox.yMax/64;

  *width = glyph_metrics.advance/64;
  *ascent =  (instance_metrics.y_ppem*face_properties.horizontal->Ascender)/
    face_properties.header->Units_Per_EM+1;
  *descent = (-instance_metrics.y_ppem*face_properties.horizontal->Descender)/
    face_properties.header->Units_Per_EM+1;
  /*
    Render label with a TrueType font.
  */

  bitmapData = "#define char_width ";
  sprintf(buf,"%d\n",canvas.width);
  bitmapData += buf;
  bitmapData += "#define char_height ";
  sprintf(buf,"%d\n",canvas.rows);
  bitmapData += buf;
  bitmapData += "static unsigned char char_bits[] = {\n  ";
  p = (unsigned char *)canvas.bitmap;
  i = 0;
  j = 0;
  bitChar = 0;
  for (y=0; y < (int) canvas.rows; y++) {
    for (x=0; x < (int) canvas.width; x++) {
      bitChar>>=1;
      value = (*p) > 1 ? 1 : 0;
      if (value) 
 	bitChar |= 0x80;
      i++;
      if (i == 8) {
	sprintf(buf,"0x%2.2x, ",bitChar);
	bitmapData += buf ;
	i = 0;
	bitChar = 0;
	j++;
	if (j%12 == 0) bitmapData +="\n  ";
      }
      p++;
    }
    if (i!=0) {
      bitChar>>=(8-i);
      sprintf(buf,"0x%2.2x, ",bitChar);
      bitmapData += buf;
      i = 0;
      bitChar = 0;
      j++;
      if (j%12 == 0) bitmapData +="\n  ";
    }

  }
  bitmapData.erase(bitmapData.size()-2,2);
  bitmapData += "};";

  /*
    Free TrueType resources.
  */
  delete[] canvas.bitmap;
  delete[] character.bitmap;

  return bitmapData;
}


void *MLB::Font::GetCharPixmap(char ascii, int *cols, int *rows, int *width, int *ascent, int *descent, int *lbearing, TT_BBox *bbox)
{
  TT_Glyph_Metrics
    glyph_metrics;
  TT_Error
    error;
  unsigned short
    unicode;
  register int
    x;
  int
    y;
  unsigned char
    *bitmap;
  int 
    border; 


  unicode = 0;
  
  unicode = (unsigned char)ascii;
  
  if (glyphs[unicode].z == (TT_Glyph *) NULL) {
    if (character_map) {
      code = TT_Char_Index(char_map,unicode);
    }
    else {
      code = ((int) unicode-' '+3) < 0 ? 0 : ((int) unicode-' '+3);
      if ((int) code >= number_glyphs)
	code = 0;
    }
    error = TT_New_Glyph(face,&glyphs[unicode]);
    error|= TT_Load_Glyph(instance,glyphs[unicode],code,
			  TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH);
    if (error)
      GUI::MessageWindow::Error("Font::GetCharPixmap - Cannot initialize TTF glyph");
  }   

  border = 20;
  *lbearing = border/2;  
  *cols = border;
  if (glyphs[unicode].z != (TT_Glyph *) NULL) {
    TT_Get_Glyph_Metrics(glyphs[unicode],&glyph_metrics);
    *cols += (glyph_metrics.bbox.xMax-glyph_metrics.bbox.xMin)/64;
  }
  *rows = instance_metrics.y_ppem*(face_properties.horizontal->Ascender-
				    face_properties.horizontal->Descender) /
                                    face_properties.header->Units_Per_EM;


  x = *lbearing;
  y = (-instance_metrics.y_ppem*face_properties.horizontal->Descender)/
    face_properties.header->Units_Per_EM+1;

  if (glyphs[unicode].z != (TT_Glyph *) NULL) {
    TT_Get_Glyph_Metrics(glyphs[unicode],&glyph_metrics);
    bitmap = RenderGlyph(glyphs[unicode],x,y,*cols,*rows,&glyph_metrics);
  }

  bbox->xMin = glyph_metrics.bbox.xMin/64;
  bbox->yMin = glyph_metrics.bbox.yMin/64;
  bbox->xMax = glyph_metrics.bbox.xMax/64;
  bbox->yMax = glyph_metrics.bbox.yMax/64;

  *width =  glyph_metrics.advance/64;
  *ascent =  (instance_metrics.y_ppem*face_properties.horizontal->Ascender)/
    face_properties.header->Units_Per_EM+1;
  *descent = (-instance_metrics.y_ppem*face_properties.horizontal->Descender)/
    face_properties.header->Units_Per_EM+1;
  /*
    Render label with a TrueType font.
  */
  /*
    Free TrueType resources.
  */

  return bitmap;
}

MLB::Font::~Font()
{
  register int
    i;

  TT_Done_Instance(instance);

  for (i=0; i < MaxGlyphs; i++)
    TT_Done_Glyph(glyphs[i]); 

  delete[] glyphs;
}

int MLB::Font::GetSize()
{
  return size;
}

int MLB::Font::GetHeight()
{
  int
    height;

  height = ((instance_metrics.y_ppem*
	     (face_properties.horizontal->Ascender-face_properties.horizontal->Descender+face_properties.horizontal->Line_Gap))/
	    face_properties.header->Units_Per_EM+1);

  return height;
}

MLB::FontFamily::FontFamily(TT_Engine engine, char *fontPath, char *_family, char *_weight, char *_slant)
{
  char 
    *filename;
  int
    i;

  error = 0;
  
  name = _family;
  weight = _weight;
  slant = _slant;

  i = 0;
  if (name != "arial") {
    i++;
    if (name != "times") {
      i++;
      if (name != "arialn") {
	i++;
	if (name != "lucidatypewriter") {
	  i++;
	  if (name != "symbol") {
	    i++;
	  }
	}
      }
    }
  }

  i = i*4;
  if (weight == "bold") i = i+1;
  if (slant == "italic") i = i+2;
  
  fontFilename = FontList::Fonts[i];
#ifndef WIN32
  filename = new char[strlen(MLB::FontList::Fonts[i])+strlen(fontPath)+2];
  sprintf(filename,"%s/%s",fontPath,MLB::FontList::Fonts[i]);
#else
  filename = new char[strlen(MLB::FontList::Fonts[i])+7];
  sprintf(filename,"fonts\\%s",MLB::FontList::Fonts[i]);
#endif

  if (error = TT_Open_Face(engine,filename,&face)) {   
    GUI::MessageWindow::Error("FontFamily::FontFamily - cant open face\n");
    free(filename);
    return;
  }

  TT_Get_Face_Properties(face,&face_properties);
  TT_Load_PS_Names( face, &ttpost );
  delete[] filename;
}

MLB::FontFamily::~FontFamily()
{
  std::list<MLB::Font *>::iterator 
    temp;

  name.erase();
  weight.erase();
  slant.erase();

  temp = fonts.begin();
  while (temp != fonts.end()) {
    delete((*temp));
    temp++;
  }
  fonts.clear();

  if (!error)
    TT_Close_Face(face);
}

MLB::Font* MLB::FontFamily::FindFontInFamily(char *_family, char *_size, char *_weight, char *_slant, double zoom) {
  if (error) return NULL;

  if ((name != _family) || (weight != _weight) || (slant != _slant)) return NULL;

  int size = atoi(_size);
  for (std::list<Font*>::iterator i = fonts.begin(); i != fonts.end(); ++i)
    if ((size == (*i)->GetSize()) && (zoom == (*i)->GetZoom())) return *i;

  return AddFont(size, zoom);
}

MLB::Font* MLB::FontFamily::AddFont(int size, double zoom) {
  if (error) return NULL;

  Font* font = new Font(fontFilename, face, face_properties, size, zoom);
  fonts.push_back(font);
  return font;
}

MLB::FontList::FontList(char *path) {
  TT_Error
    error;
#ifndef WIN32
  char
    *zw;
#endif
  unsigned char  
    virtual_palette[5] = { 0, 1, 2, 3, 4 };

  fontPath = NULL;
  if (error=TT_Init_FreeType(&engine)) { 
    GUI::MessageWindow::Error("FontFamily::FontList - Cannot initialize TTF engine\n");
    return;
  }
  if (error=TT_Init_Post_Extension(engine)) { 
    GUI::MessageWindow::Error("FontFamily::FontList - Cannot initialize TTF-PostScript\n");
    return;
  }

  TT_Set_Raster_Gray_Palette( engine, virtual_palette );

#ifndef WIN32

  zw = getenv("TTF_PATH");
  if (zw) fontPath = strdup(zw);
  else    fontPath = NULL;
  if (!fontPath || (Tcl_Access(fontPath,R_OK) < 0)) {
    if (fontPath) delete[] fontPath;
    fontPath = new char[strlen(path) + 7];
    sprintf(fontPath,"%s/fonts",path);
    GUI::MessageWindow::Info("No TTF_PATH (TrueTypeFont-Path) specified, set path to %s\n",fontPath);
  }
#endif

}


MLB::FontList::~FontList() {
  TT_Done_FreeType(engine);
  if (fontPath) delete[] fontPath;
}


MLB::Font* MLB::FontList::GetFont(char* _family, char* _size, char* _weight, char* _slant, double zoom) {
  Font* font = NULL;
  for (std::list<FontFamily*>::iterator i = families.begin(); i != families.end(); ++i) 
    if (font = (*i)->FindFontInFamily(_family,_size,_weight,_slant, zoom)) break;

  if (!font) {
    FontFamily* family = new MLB::FontFamily(engine,fontPath,_family,_weight,_slant);
    font = family->AddFont(atoi(_size), zoom);
    families.push_back(family);
  }

  return font;
}
