/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_generic.cpp
 authors : Juergen Vogel, Gabriela Maass
 date    : 08.03.2000
 desc    :

 **********************************************************************************************************/

#include <iostream>
#include <iterator>

#include "mlb.h"
#include "mlb_generic.h"
#include "mlb_session.h"

uint32 MLB::StringToUint32(char* str) {
  char* e = &str[strlen(str)];
  return strtoul(str, &e, 10);
}

/**************************************
 *
 **************************************/

MLB::GenericObject::GenericObject(uint32 _id, string _tkref)
  : id(_id), tkref (_tkref) {
}

MLB::GenericObject::GenericObject(uint32 _id)
  : id(_id) {
}

MLB::GenericObject::~GenericObject() {
}


void MLB::GenericObject:: ChangeFlushedObjectOrder(char* buf, uint32 order) { //J
  uint32* p32 = (uint32*)buf;
  p32[1] = htonl(order);
}


/**************************************
 *
 **************************************/
MLB::GenericContainer::GenericContainer(char *_name, uint32 _id, string tkref)
  : GenericObject(_id, tkref), current(0), lastObjectOrder(0) {
  containerName = _name;
}


MLB::GenericContainer::GenericContainer(char *_name, uint32 _id)
  : GenericObject(_id) , current(0), lastObjectOrder(0) {
  containerName = _name;
}


MLB::GenericContainer::GenericContainer(uint32 _id, string tkref)
  : GenericObject(_id, tkref), current(0), lastObjectOrder(0) {
}


MLB::GenericContainer::GenericContainer(uint32 _id)
  : GenericObject(_id), current(0), lastObjectOrder(0) {
}


MLB::GenericContainer::~GenericContainer() {

  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    delete *i;
  subContainer.clear();

  containerName.erase();
  nodeName.erase();
}


void MLB::GenericContainer::PrintActiveSubcomponents() {
  RTPI::RtpiSubComponent* sub;
  
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    sub = CommunicationManager::cm->GetRtpi()->get_subcomponent((*i)->GetID());
    if (sub->is_active()) GUI::MessageWindow::Debug(" O%u", (*i)->GetID());
    (*i)->PrintActiveSubcomponents();
  }
}


std::list<MLB::GenericContainer*>::iterator MLB::GenericContainer::GetCurrentContainerIterator() {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    if ((*i)->GetID() == current) return i;
  return subContainer.end();
}


MLB::GenericContainer* MLB::GenericContainer::GetCurrentContainer() {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    if ((*i)->GetID() == current) return *i;
  return NULL;
}


void MLB::GenericContainer::Init() {
  if (subContainer.empty()) {
    current = 0;
    return;
  }

  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    (*i)->Init();

  current = (subContainer.front())->GetID();
}


void MLB::GenericContainer::InsertContainer(GenericContainer* container) { 

  current = container->GetID();

  if (subContainer.empty()) {
    subContainer.push_back(container);
    return;
  }

  std::list<GenericContainer*>::iterator pos = subContainer.begin();
  while (pos != subContainer.end() && (container->GetOrder() >= (*pos)->GetOrder())) ++pos;

  subContainer.insert(pos, container);
}


void MLB::GenericContainer::InsertContainerCurrent(MLB::GenericContainer *container) {

  if (subContainer.empty()) {
    subContainer.push_back(container);
    container->SetOrder(MLB_OO_MIDDLE);
    current = container->GetID();
    return;
  }

  std::list<GenericContainer*>::iterator pos = GetCurrentContainerIterator();
  if (pos == subContainer.end()) --pos; // error, but push_back container to prevent worse

  if (*pos != subContainer.back()) {
    uint32 order = (*pos)->GetOrder() / 2;
    ++pos;
    order += ((*pos)->GetOrder() / 2);
    container->SetOrder(order);
    subContainer.insert(pos, container);
  } else {
    if (subContainer.back()->GetOrder() == MLB_OO_END) {
      if (subContainer.size() == 1) {
        container->SetOrder(MLB_OO_MIDDLE);
        subContainer.insert(--subContainer.end(), container);
      } else {
        container->SetOrder((*(--(--subContainer.end())))->GetOrder() + MLB_OO_OFFSET);
        subContainer.insert(--(--subContainer.end()), container);
      }
    } else {
    container->SetOrder((subContainer.back())->GetOrder() + MLB_OO_OFFSET);
    subContainer.push_back(container);
  }
  }
  current = container->GetID();
}


void MLB::GenericContainer::InsertContainerEnd(GenericContainer* container) {
  if (subContainer.empty()) {
    container->SetOrder(MLB_OO_MIDDLE);
  subContainer.push_back(container);
  } else {
    if (subContainer.back()->GetOrder() == MLB_OO_END) {
      if (subContainer.size() == 1) {
        container->SetOrder(MLB_OO_MIDDLE);
        subContainer.insert(--subContainer.end(), container);
      } else {
        container->SetOrder((*(--(--subContainer.end())))->GetOrder() + MLB_OO_OFFSET);
        subContainer.insert(--(--subContainer.end()), container);
      }
    } else {
      container->SetOrder((subContainer.back())->GetOrder() + MLB_OO_OFFSET);
      subContainer.push_back(container);
    }
  }
  current = container->GetID();
}


void MLB::GenericContainer::SetPageDimension(int _width, int _height) {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    (*i)->SetPageDimension(_width, _height);
}


MLB::GenericObject* MLB::GenericContainer::GetObject(uint32 id) {
  if (id == this->GetID()) return this;

  GenericObject* obj = NULL;
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) return *i;
    obj = (*i)->GetObject(id);
    if (obj) return obj;
  }

  return NULL;
}


MLB::GenericObject* MLB::GenericContainer::GetGroup(uint32 id) {
   GenericObject* obj = NULL;
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    obj = (*i)->GetGroup(id);
    if (obj) return obj;
  }

  return NULL;
}


uint32 MLB::GenericContainer::GetObjectOrder(uint32 id) {
  for (std::list<MLB::GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) return (*i)->GetOrder();
  }

  return 0; // not found
}


uint32 MLB::GenericContainer::GetPushBackObjectOrder() {
  if (subContainer.empty()) {
    if (lastObjectOrder == 0) lastObjectOrder =  MLB_OO_MIDDLE;
    else lastObjectOrder += MLB_OO_OFFSET;
  } else {
    uint32 order = subContainer.back()->GetOrder();
    if (order <= lastObjectOrder) lastObjectOrder += MLB_OO_OFFSET;
    else lastObjectOrder = order + MLB_OO_OFFSET;
  }
  return lastObjectOrder;
}


uint32 MLB::GenericContainer::GetInsertObjectOrder() {
  if (subContainer.empty()) {
    if (lastObjectOrder == 0) lastObjectOrder =  MLB_OO_MIDDLE;
    else lastObjectOrder += MLB_OO_OFFSET;
  } else {
    std::list<MLB::GenericContainer*>::iterator pos = GetCurrentContainerIterator();
    if ((pos == subContainer.end()) || (*pos == subContainer.back())) {
      uint32 order = subContainer.back()->GetOrder();
      if (order <= lastObjectOrder) lastObjectOrder +=  MLB_OO_OFFSET;
      else lastObjectOrder = order + MLB_OO_OFFSET;
    } else {
      lastObjectOrder = ((*pos)->GetOrder() / 2);
      ++pos;
      lastObjectOrder += ((*pos)->GetOrder() / 2);
    }
  }

  return lastObjectOrder;
}


uint32  MLB::GenericContainer::GetEndObjectOrder() {
  if (subContainer.empty()) return 0;
  else return (subContainer.back())->GetOrder();
}


uint32  MLB::GenericContainer::GetFirstObjectOrder() {
  if (subContainer.empty()) return 0;
  else return (subContainer.front())->GetOrder();
}


uint32 MLB::GenericContainer::GetNewObjectOrder(int pos) {
  if (subContainer.empty()) {
    return MLB_OO_MIDDLE;
  } else if (pos == 0) {
    return (subContainer.front())->GetOrder() - MLB_OO_OFFSET;
  } else if (pos >= GetSize()) {
    return (subContainer.back())->GetOrder() + MLB_OO_OFFSET;
  } else {
    std::list<MLB::GenericContainer*>::iterator iter = subContainer.begin();
    for (int i = 0; i < pos - 1; ++i) ++iter;
    uint32 order = (*iter)->GetOrder() / 2;
    ++iter;
    order += (*iter)->GetOrder() / 2;
    return order;
  }
}


uint32 MLB::GenericContainer::GetNextObjectOrder(uint32 _id) {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == _id) {
      if (*i == subContainer.back()) return (*i)->GetOrder() + MLB_OO_OFFSET;
      uint32 order = (*i)->GetOrder() / 2;
      ++i;
      return order + (*i)->GetOrder() / 2;
    }
  }
  return 0;
}


MLB::GenericObject* MLB::GenericContainer::GetObjectWithOrder(uint32 order) {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetOrder() == order) return *i;
    if ((*i)->GetOrder() > order) return NULL; // not found
  }
  return NULL; // not found
}


// Find direct parent container of Object obj
MLB::GenericContainer* MLB::GenericContainer::GetContainer(uint32 id) {
  GenericContainer* container = NULL;
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) return this;
    container = (*i)->GetContainer(id);
    if (container) return container;
  }

  return NULL;
}


MLB::GenericContainer* MLB::GenericContainer::GetFirstPage() {
  GenericContainer* ret = NULL;

  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetType() == GenericObject::PAGE) return *i;
    ret = (*i)->GetFirstPage();
    if (ret) return ret;
  }

  return NULL; // no page in here
}


MLB::GenericContainer* MLB::GenericContainer::GetPrevious(uint32 id) {
  GenericContainer* ret = NULL;

  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) {
      // found reference container
      if (i == subContainer.begin()) return NULL; // there is no previous container
      --i;
      return *i;
    }

    ret = (*i)->GetPrevious(id);
    if (ret) return ret; // else not found
  }

  return NULL; // not found
}


MLB::GenericContainer* MLB::GenericContainer::GetNext(uint32 id) {
  GenericContainer* ret = NULL;

  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) {
      // found reference container
      ++i;
      if (i == subContainer.end()) return NULL; // there is no next container
      return *i;
    }

    ret = (*i)->GetNext(id);
    if (ret) return ret; // else not found
  }

  return NULL; // not found
}


void MLB::GenericContainer::DeleteTelepointerOfParticipant(uint32 participantID) {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    (*i)->DeleteTelepointerOfParticipant(participantID);
  }
}


MLB::GenericObject* MLB::GenericContainer::DeleteObject(uint32 id) {
  GenericObject* obj = NULL;
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
    if ((*i)->GetID() == id) {
      obj = *i;
      subContainer.erase(i);
      if (id == current) current = 0;
      return obj;
    }
    obj = (*i)->DeleteObject(id);
    if (obj) return obj;
  }

  return NULL;
}


void MLB::GenericContainer::ClearNodeNames() {
  for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i)
    (*i)->ClearNodeNames();

  nodeName = "";
}

//-------------------------------------------------------------------------//

/**   If shows or hides all the animated objects on container's pages.
  **/
void MLB::GenericContainer::SetAnimationEditMode( bool mode )
{
  for( std::list<GenericContainer*>::iterator i=subContainer.begin(); i!=subContainer.end(); i++ )
    (*i)->SetAnimationEditMode(mode);
} // SetAnimationObjectsVisible()

//=========================================================================//
