/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_generic.h
 authors : Juergen Vogel, Gabriela Maass
 date    : 08.03.2000
 desc    :

 *********************************************************************************************************/

#ifndef __MLB_GENERIC__
#define __MLB_GENERIC__

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif

#include <stdio.h>
#include <string>
#include <list>
#include <vector>

#include <tcl.h>

#include <util/types.h>
#include <rtpi/rtpiPacket.h>

#ifdef WIN32
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include <mlb_gui.h>


#define MLB_OO_MIDDLE 2147483648u
#define MLB_OO_END    4294967295u
#define MLB_OO_OFFSET 8192

namespace MLB {
  //using namespace std;

  class DrawObject;
  class Page;

  class GenericObject {
  protected:
    uint32 id;
    uint32 order;
    std::string tkref;

  public:
    enum type {
      SESSION,
      FOLDER,
      PAGE,
      RECTANGLE,
      OVAL,
      LINE,
      IMAGE,
      ASCIITEXT,
      CHAR,
      FREEHAND,
      POLYLINE,
      POLYGON,
      GROUP,
      TELEPOINTER,
      TELEBOX
    };

    GenericObject() : id (0) {}
    GenericObject(char* name, uint32 id, std::string tkref);
    GenericObject(char* name, uint32 id);
    GenericObject(uint32 id, std::string tkref);
    GenericObject(uint32 id);

    virtual ~GenericObject();

    virtual int   GetType() = 0;

    virtual void  SetTkRef(char*_tkref) {tkref.erase(); tkref  = _tkref;}
    virtual char* GetTkRef() {return (char *)tkref.c_str();}

    uint32 GetID() {return id;}
    void   SetID(uint32 _id) {id = _id;}

    uint32 GetOrder() {return order;}
    virtual void SetOrder(uint32 _order) {order = _order;}

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority) {return NULL;}

    virtual int FlushState(char* buf) {return -1;}
    virtual int GetStateSize() {return -1;}
    virtual int ParseState(char* buf) {return -1;}

    static void ChangeFlushedObjectOrder(char* buf, uint32 order);

    virtual int FlushDelete(char* buf) {return -1;}
    static int GetDeleteSize() {return 4;}
  };

  class GenericContainer : public GenericObject {
  protected:
    std::list<GenericContainer*>::iterator GetCurrentContainerIterator();

    std::list<GenericContainer*> subContainer;
    uint32 current;
    std::string nodeName;
    std::string nodePath;
    std::string containerName;

    uint32 lastObjectOrder;

  public:
    GenericContainer();
    GenericContainer(char* name, uint32 id, std::string tkref);
    GenericContainer(char* name, uint32 id);
    GenericContainer(uint32 id, std::string tkref);
    GenericContainer(uint32 id);
    virtual ~GenericContainer();

    virtual void Init();
    virtual bool Empty() {return subContainer.empty();}
    virtual int  GetSize() {return subContainer.size();}

    virtual void PrintActiveSubcomponents();

    /*** General Functions ***/
    virtual void SetName(char* _name)=0;
    virtual char *GetName() {return (char *)containerName.c_str();}

    virtual GenericContainer* GetCurrentContainer();
    virtual GenericContainer* SetCurrentContainer(std::vector<uint32>* idVector, int pos) = 0;

    virtual void SetPageDimension(int _width, int _height);

    /*** draw the document structure as tree ***/
    virtual void DrawPageTree(std::string _nodeNumber, std::string _nodeData, std::string parentNode, int pos, bool *opne) = 0;
    virtual void ClearPageTree() {}
    virtual void ClearNodeNames();

    std::string GetNodeName() {return nodeName;}
    void ClearNodeName() {nodeName = "";}
    std::string GetNodePath() {return nodePath;}

    /*** management of container and object hierachy ***/
    virtual uint32 GetObjectOrder(uint32 id);
    virtual uint32 GetPushBackObjectOrder();
    virtual uint32 GetInsertObjectOrder();
    virtual uint32 GetEndObjectOrder();
    virtual uint32 GetFirstObjectOrder();
    virtual uint32 GetNewObjectOrder(int pos);
    virtual uint32 GetNextObjectOrder(uint32 id);
    virtual GenericObject* GetObjectWithOrder(uint32 order);

    virtual void InsertContainer(GenericContainer* container);
    virtual void InsertContainerCurrent(GenericContainer* container);
    virtual void InsertContainerEnd(GenericContainer* container);

    virtual GenericObject* GetObject(uint32 id);
    virtual GenericObject* GetGroup(uint32 id);
    virtual GenericContainer* GetContainer(uint32 id);
    virtual int GetNbrOfPages() {return 0;}

    virtual GenericContainer* GetFirstPage();
    virtual GenericContainer* GetPrevious(uint32 id);
    virtual GenericContainer* GetNext(uint32 id);

    virtual GenericObject* DeleteObject(uint32 id);
    virtual void DeleteAll(GUI::Canvas* _canvas) {}
    virtual void DeleteTelepointerOfParticipant(uint32 participantID);
    virtual void DeleteObjectsFromRTPI() = 0;

    virtual void ReplaceIDs() = 0;

    /*** save functions ***/
    virtual void SaveAsPs(FILE *fp) = 0;
    virtual void SaveAsHtml(char* directory, char* docname, FILE* indexframe, int& nbr, int last) = 0;
    virtual void SaveAsImage(char* name, char* type, int& nbr) = 0;
    virtual void SaveContainer(FILE *fp, int level) = 0;

    /*** ADU functions ***/
    virtual int FlushChangeName(char* buf, std::string name) {return -1;}
    virtual int GetChangeNameSize(std::string name) {return -1;}

    /*** AOF Recording ***/
    virtual void ResetRecording() = 0;

    /*** Animation ***/
    virtual void SetAnimationEditMode( bool mode );
  };
  
  uint32 StringToUint32(char* str);
}

#endif
