/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_group.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 11.03.2001
 desc    : Object to group objects

**********************************************************************************************************/

#include "mlb.h"

#include "mlb_objects.h"
#include "mlb_session.h"
#include "mlb_adu.h"


MLB::Group::Group(uint32 _id) :  DrawObject(0,0,0,0,_id) {
  char buf[100];
  sprintf(buf,"%u",_id);
  tkref = "Group";
  tkref += buf;

  textGroup = false;
  activeItem = 0;
  saved = false;

  recorded=false;
}


MLB::DrawObject* MLB::Group::GetActiveItem() {
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    if ((*i)->GetID() == activeItem) return *i;
  return NULL;
}


std::list<MLB::DrawObject*>::iterator MLB::Group::GetActiveItemIterator() {
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    if ((*i)->GetID() == activeItem) return i;
  return groupObjects.end();
}


bool MLB::Group::HasObject(uint32 id) {
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    if ((*i)->GetID() == id) return true;
  return false;
}


MLB::DrawObject* MLB::Group::GetObject(int idx) {
  if ((idx < 0) || (idx >= groupObjects.size())) return NULL;

  std::list<DrawObject*>::iterator obj = groupObjects.begin();
  for (int i = 0; i < idx; ++i) ++obj;

  return *obj;
}


void MLB::Group::SetVisible() {
  visible = true;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    (*i)->SetVisible();
}


void MLB::Group::SetInvisible() {
  DrawObject::SetInvisible();

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    (*i)->SetInvisible();
}


std::vector<uint32>* MLB::Group::GetObjectIDs() {
  std::vector<uint32>* ids = new std::vector<uint32>;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    ids->push_back((*i)->GetID());
  return ids;
}


bool MLB::Group::AllEmpty()
{
  if (!textGroup) return false;

  if (groupObjects.empty()) return true;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    if (!((Text*)*i)->Empty()) return false; // found one being not empty

  return true;
}


void MLB::Group::SetOrder(uint32 _order) {
  order = _order;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    (*i)->SetOrder(_order);
    _order++;
  }
}



MLB::DrawObject* MLB::Group::Copy() {
  DrawObject* obj;
  Group* group = new Group(CommunicationManager::cm->GetObjectID());

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    obj = (*i)->Copy();
    if (obj) group->AddObject(obj);
  }

  return group;
}


MLB::DrawObject* MLB::Group::Copy(Page* page) {
  DrawObject* obj;
  Group* group = new Group(CommunicationManager::cm->GetObjectID());
  group->SetAnnotation(SessionManager::sm->AnnotationMode());

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::GROUP) {
      obj = ((Group*)*i)->Copy(page);
    }
    else {
      obj = (*i)->Copy();
    }
    if (obj) {
      obj->SetAnnotation(SessionManager::sm->AnnotationMode());
      obj->SetOrder(page->GetPushBackObjectOrder());
      page->AppendObject(obj);
      group->AddObject(obj);
    }
  }

  return group;
}


void MLB::Group::SendCopy(Page* page) {
  if (groupObjects.empty()) return;

  uint32 objID;
  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    switch ((*i)->GetType()) {
    case MLB::GenericObject::RECTANGLE: {
      buf   = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::RECTANGLE, 3, buf));
      break;
    }
    case MLB::GenericObject::OVAL: {
      buf   = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::OVAL, 3, buf));
      break;
    }
    case MLB::GenericObject::LINE: {
      buf   = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::LINE, 3, buf));
     break;
    }
    case MLB::GenericObject::FREEHAND:
    case MLB::GenericObject::POLYLINE: {
      buf   = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 3, buf));
      break;
    }
    case MLB::GenericObject::POLYGON: {
      buf   = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYGON, 3, buf));
      break;
    }
    case MLB::GenericObject::IMAGE: {
      buf = new Buffer((*i)->GetStateSize());
      ((PhotoImage*)(*i))->FlushState(buf->getCont());
      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), page->GetPushBackObjectOrder());
      objID = CommunicationManager::cm->GetObjectID();
      adus->push_back(CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::IMAGE, 3, buf));
      break;
    }
    }
  }

  AduManager::am->InsertAdus(adus);
  CommunicationManager::cm->SendAdusReliable(adus);
}


int MLB::Group::GetObjectOrder(DrawObject* obj) {
  int nbr = 0;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if (*i == obj) return nbr;
    ++nbr;
  }

  return -1; // not found
}


void MLB::Group::SetObjectOrder(DrawObject* obj, int newOrder) {
  if (newOrder > groupObjects.size()) return;

  // remove object from old pos
  groupObjects.remove(obj);

  // search new pos
  std::list<DrawObject*>::iterator pos = groupObjects.begin();
  for (int index = 0; index <= newOrder; ++index) ++pos;

  // insert object
  groupObjects.insert(pos, obj);
}


void MLB::Group::Draw(GUI::Canvas *_canvas) {
  if (!visible) return;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) 
    (*i)->Draw(_canvas);

  if (selected) DrawSelection(_canvas);
}


void MLB::Group::ReDraw(GUI::Canvas *_canvas) {
  if (!visible) return;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) 
    (*i)->ReDraw(_canvas);

  if (selected) DrawSelection(_canvas);
  if (textGroup) DrawTextCursor(_canvas);
}


void MLB::Group::Move(GUI::Canvas *_canvas, int diffX, int diffY, bool redraw) {
  if (!visible) return;

  recorded = false;

  x += diffX;
  y += diffY;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) 
    (*i)->Move(_canvas,diffX,diffY,redraw);

  if (redraw) {
    DeleteTextCursor(_canvas);
    if (selected) DrawSelection(_canvas);
  }
}


void MLB::Group::SendMove(bool cue, int x, int y) {
  if (textGroup) return;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  for (std::list<MLB::DrawObject *>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    buf   = new Buffer((*i)->GetMoveSize());
    (*i)->FlushMove(buf->getCont(), x, y);
    if (cue) adus->push_back(CommunicationManager::cm->CreateCue((*i)->GetID(), buf));
    else adus->push_back(CommunicationManager::cm->CreateEvent((*i)->GetID(), buf));
  }

  AduManager::am->InsertAdus(adus);
  if (cue) CommunicationManager::cm->SendAdusUnreliable(adus);
  else CommunicationManager::cm->SendAdusReliable(adus);
}


void MLB::Group::SendDelete() {
  if (textGroup) return;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    buf   = new Buffer((*i)->GetDeleteSize());
    (*i)->FlushDelete(buf->getCont());
    adus->push_back(CommunicationManager::cm->CreateEvent((*i)->GetID(), buf));
  }

  AduManager::am->InsertAdus(adus);
  CommunicationManager::cm->SendAdusReliable(adus);
}


void MLB::Group::SendAllObjects(uint32 timestamp, int priority) {
  /*
  if (textGroup) return;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.begin(); ++i) {
    switch ((*i)->GetType()) {
    case GenericObject::ASCIITEXT: {
      bytes = (*i)->GetStateSize();
      buf   = new Buffer()char[bytes];
      (*i)->FlushState(buf);
      adus->push_back(CommunicationManager::cm->CreateState((*i)->GetID(), GetID(), CommunicationManager::TEXT, priority, buf, bytes, timestamp));
      break;
    }
    }
  }
  TimeWarp::tw->InsertAdus(adus);
  CommunicationManager::cm->SendAdusReliable(adus);
  */
}


void MLB::Group::ChangeSubcomponentsActiveness(bool active) {
  CommunicationManager::cm->ChangeSubcomponentActiveness(GetID(), active);
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    CommunicationManager::cm->ChangeSubcomponentActiveness((*i)->GetID(), active);
}


void MLB::Group::ReplaceIDs() {
  uint32 newID;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    newID = CommunicationManager::cm->GetObjectID();
    if ((*i)->GetID() == activeItem) activeItem = newID;
    (*i)->SetID(newID);
  }
}


void MLB::Group::DeleteObjectsFromRTPI() {
  CommunicationManager::cm->UnregisterSubcomponent(GetID());

  if (textGroup) return; // done

  for (std::list<MLB::DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    CommunicationManager::cm->UnregisterSubcomponent((*i)->GetID());
}


void MLB::Group::SaveObject(FILE *fp, int level) {
  if (groupObjects.empty()) return;

  char* tab = new char[level*2+1];
  *tab = '\0';
  for (int j = 0; j < level; ++j) strcat(tab, "  ");

  fprintf(fp,"%s<GROUP%s>\n", tab, GetAnimationStepSGMLText().c_str() );

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    (*i)->SaveObject(fp,(level+1));
  }

  fprintf(fp,"%s</GROUP>\n", tab);
  delete tab;
}


void MLB::Group::SetUnrecorded()
{
  recorded = false;
  for (std::list<MLB::DrawObject *>::iterator iter = groupObjects.begin(); iter != groupObjects.end(); ++iter) {
      (*iter)->SetUnrecorded();
  }
}

int MLB::Group::RecordableObjects()
{
  if (groupObjects.empty() || !textGroup) visible = false;

  if (!visible) return 0;

  return 1;
}

void MLB::Group::SaveAsAOF(FILE *_fp)
{
  std::list<MLB::DrawObject *>::iterator
    iter;
  int
    lastX, lastY,
    lastWidth,
    lastHeight,
    actX, actY,
    offX, offY,
    line;

  if (!visible || groupObjects.empty() || !textGroup || recorded) return;

  recorded = true;

  recID = MLB::SessionManager::sm->recCount;
  MLB::SessionManager::sm->recCount++;

  iter = groupObjects.begin();
  lastX = (*iter)->GetX();
  lastY = (*iter)->GetY();

  fprintf(_fp,"<TEXT x=%d y=%d width=%d height=%d>\n",lastX,lastY,width,height);

  lastX = (*iter)->GetX();
  lastY = (*iter)->GetY();
  lastWidth = 0;
  lastHeight = 0;
  line = 0;
  offX = offY = 0;
  for (iter = groupObjects.begin(); iter != groupObjects.end(); ++iter) {
    actX = (*iter)->GetX();
    actY = (*iter)->GetY();
    if (actY > lastY) {
      line++;
      offX = 0;
      offY += (actY-lastY)-lastHeight;
    }
    else {
      offX = actX-(lastX+lastWidth);
    }
    ((Text *)(*iter))->SaveAsAOF(_fp,line,offX,offY);
    lastX = actX;
    lastY = actY;
    lastWidth = (*iter)->GetWidth();
    lastHeight = (*iter)->GetHeight();
  }

  fprintf(_fp,"</TEXT>\n");
}

void MLB::Group::SaveAsPs(FILE *fp, int pageHeight, int pageWidth, bool rotate) {
}


void MLB::Group::AddObject(DrawObject *obj) {
  int x1, y1, x2, y2;

  if (obj->GetType() == GenericObject::ASCIITEXT) {
    if (groupObjects.empty() || textGroup) textGroup = true;
    else textGroup = false;
  }

  obj->GetBBox(&x1,&y1,&x2,&y2);
  if (groupObjects.empty()) {
    x = x1;
    y = y1;
    width = (x2-x);
    height = (y2-y);
  } else {
    if (x1 < x) x = x1;
    if (y1 < y) y = y1;
    if ((x2-x) > width) width = (x2-x);
    if ((y2-y) > height) height = (y2-y);
  }

  obj->SetGroup(this);

  if (!activeItem) groupObjects.push_front(obj); 
  else {
  std::list<DrawObject*>::iterator pos = GetActiveItemIterator();
  if (pos == groupObjects.end()) groupObjects.push_back(obj);
  else {
    ++pos;
    if (pos == groupObjects.end()) groupObjects.push_back(obj);
    else groupObjects.insert(pos, obj);
  }
  }
  activeItem = obj->GetID();
}


MLB::GenericObject* MLB::Group::DeleteObject(GenericObject* obj) { 
  if (groupObjects.empty()) return NULL;

  if (textGroup) {
    std::list<DrawObject*>::iterator toDel = groupObjects.begin();
    while ((toDel != groupObjects.end()) && (*toDel != obj)) {
      ++toDel;
    }

    if (toDel == groupObjects.end()) return NULL; // not found

    ((DrawObject*)obj)->SetGroup(NULL);

    int offX = (*toDel)->GetWidth();
    Text* txt;
    for (std::list<DrawObject*>::iterator i = toDel; i != groupObjects.end(); ++i) {
      txt = (Text*)*i;
      if (txt->GetY() != (*toDel)->GetY()) break;
      txt->SetX(txt->GetX() - offX);
    }

    if (activeItem == (*toDel)->GetID()) {
      std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
      if (groupObjects.size() == 1) {
	activeItem = 0;
      } else if (actIter != groupObjects.begin()) {
	--actIter; 
	activeItem = (*actIter)->GetID();
      } else {
	++actIter;
	activeItem = (*actIter)->GetID();
      }
    }

    groupObjects.erase(toDel);
    UpdateDimension();

    return obj;

  } else {

    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i) == obj) {
	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}
	groupObjects.erase(i);
	UpdateDimension();
	return obj;
      }
    }

    return NULL;
  }
}


MLB::DrawObject* MLB::Group::DeleteObject(uint32 id) { //J
  if (groupObjects.empty()) return NULL;

  if (textGroup) {
    DrawObject* obj = NULL;

    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetID() == id) {
	obj = *i;

	if (obj->GetType() == GenericObject::ASCIITEXT) {
	  int offX = obj->GetWidth();
	  Text* txt;

	  std::list<MLB::DrawObject*>::iterator j = i;
	  for (++j; j != groupObjects.end(); ++j) {
	    txt = (Text*)*j;
	    if (txt->GetY() != obj->GetY()) break;
	    txt->SetX(txt->GetX() - offX);
	  }
	}

	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}
	
	obj->SetGroup(NULL);
	groupObjects.erase(i);
	UpdateDimension();
	break;
      }
    }

    return obj;

  } else {

    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetID() == id) {
	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}
	(*i)->SetGroup(NULL);
	groupObjects.erase(i);
	UpdateDimension();
	return *i;
      }
    }

    return NULL;
  }
}


void MLB::Group::RemoteDeleteObject(MLB::DrawObject* obj) { //J
  if (groupObjects.empty()) return;

  if (textGroup) {
    std::list<MLB::DrawObject *>::iterator toDel = groupObjects.begin();
    while ((toDel != groupObjects.end()) && (*toDel != obj)) {
      ++toDel;
    }
    if (toDel == groupObjects.end()) return; // not found

    int offX = (*toDel)->GetWidth();
    Text* txt;

    for (std::list<DrawObject*>::iterator i = toDel; i != groupObjects.end(); ++i) {
      txt = (Text*)*i;
      txt->SetX(txt->GetX() - offX);
    }

    if (activeItem == (*toDel)->GetID()) {
      std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
      if (groupObjects.size() == 1) {
	activeItem = 0;
      } else if (actIter != groupObjects.begin()) {
	--actIter; 
	activeItem = (*actIter)->GetID();
      } else {
	++actIter;
	activeItem = (*actIter)->GetID();
      }
    }

    (*toDel)->SetGroup(NULL);
    groupObjects.erase(toDel);
    UpdateDimension();

  } else {

    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i) == obj) {
	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}
	(*i)->SetGroup(NULL);
	groupObjects.erase(i);
	UpdateDimension();
	return;
      }
    }
  }
}


MLB::DrawObject* MLB::Group::RemoteDeleteObject(uint32 id) { //J
  if (groupObjects.empty()) return NULL;

  if (textGroup) {

    DrawObject* obj = NULL;

    for (std::list<MLB::DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetID() == id) {
	obj = *i;
	if (obj->GetType() == GenericObject::ASCIITEXT) {
	  int offX = obj->GetWidth();
	  Text* txt;

	  std::list<MLB::DrawObject*>::iterator j = i;
	  for (++j; j != groupObjects.end(); ++j) {
	    txt = (Text*)*j;
	    txt->SetX(txt->GetX() - offX);
	  }
	}

	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}

	(*i)->SetGroup(NULL);
	groupObjects.erase(i);
	break;
      }
    }

    UpdateDimension();
    return obj;

  } else {

    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetID() == id) {
	if (activeItem == (*i)->GetID()) {
	  std::list<DrawObject*>::iterator actIter = GetActiveItemIterator();
	  if (groupObjects.size() == 1) {
	    activeItem = 0;
	  } else if (actIter != groupObjects.begin()) {
	    --actIter; 
	    activeItem = (*actIter)->GetID();
	  } else {
	    ++actIter;
	    activeItem = (*actIter)->GetID();
	  }
	}
	(*i)->SetGroup(NULL);
	groupObjects.erase(i);
	UpdateDimension();
	return *i;
      }
    }
    return NULL;
  }
}


void MLB::Group::DeleteAll(GUI::Canvas *_canvas) {
//    DeleteTextCursor(_canvas);

//    if (groupObjects.empty()) return;

//    for (std::list<MLB::DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
//      /* Objekte einer Gruppe sind nicht , sondern nur auf der Seite */
//      (*i)->Hide(_canvas);
//    }
  groupObjects.clear();
  activeItem = 0;
}


void MLB::Group::DrawSelection(GUI::Canvas *_canvas) {
  if (!visible) return;

  if (selected) DeleteSelection(_canvas);

  selected = true;

  selectPoints[0] = new Point(x,y);
  selectPoints[1] = new Point(x+width,y);
  selectPoints[2] = new Point(x+width,y+height);
  selectPoints[3] = new Point(x,y+height);
  for (int i = 0; i < 4; i++) _canvas->DrawSelectionRectangle(tkref.c_str(),
							      SessionManager::sm->GetZoom() * selectPoints[i]->x,
							      SessionManager::sm->GetZoom() * selectPoints[i]->y);
}


void MLB::Group::DeleteSelection(GUI::Canvas *_canvas) {
  if (!visible || !selected) return;
  
  selected = false;

  _canvas->DeleteSelection(tkref.c_str());
}


void MLB::Group::Hide(GUI::Canvas *_canvas) {
  if (!visible) return;

  if (textGroup) DeleteTextCursor(_canvas);
  if (selected) DeleteSelection(_canvas);

  visible = false;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i)
    (*i)->Hide(_canvas);
}


bool MLB::Group::IsEditablePoint(int _x, int _y) {
  if (!visible) return false;

  for (int i = 0; i < 4; ++i) {
    if ((_x > selectPoints[i]->x-TOLERANZ) && (_x < selectPoints[i]->x+TOLERANZ)) {
      if ((_y > selectPoints[i]->y-TOLERANZ) && (_y < selectPoints[i]->y+TOLERANZ)) {
	selectCount = i;
	return true;
      }
    }
  }
  return false;
}


MLB::DrawObject* MLB::Group::PointIsInText(int _x, int _y) {
  if (!visible || groupObjects.empty()) return NULL;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->PointIsInText(_x,_y)) {
      activeItem = (*i)->GetID();
      std::list<DrawObject*>::iterator k;
      for (k = groupObjects.begin(); k != i; ++k) ((Text*)(*k))->GoEnd();
      for (++k; k != groupObjects.end(); ++k)     ((Text*)(*k))->GoHome();
      return (*i);
    }
  }

  return NULL;
}


bool MLB::Group::Overlap(GUI::Canvas *_canvas,int _x, int _y) {
  if (!visible) return false;

  int x1, y1, x2, y2;
  GetBBox(&x1,&y1,&x2,&y2);
  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return false;

  bool overlap = false;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->Overlap(_canvas,_x,_y)) {
      overlap = true;
      activeItem = (*i)->GetID();
      break;
    }
  }

  return overlap;
}

void MLB::Group::Update() {
  if (!visible || !textGroup || groupObjects.empty()) return;

  std::list<DrawObject*>::iterator i = groupObjects.begin();
  int newX = (*i)->GetX();
  int newY = (*i)->GetY();
  int offX = (*i)->GetWidth();
  int offY = (*i)->GetHeight();
  int oldY = (*i)->GetY();
  ++i;

  while (i != groupObjects.end()) {
    if ((*i)->GetY() != oldY) {
      offX = 0;
      newY += offY;
    }
    oldY = (*i)->GetY();
    (*i)->SetX(newX+offX);
    (*i)->SetY(newY);
    offX += (*i)->GetWidth();
    offY = (*i)->GetHeight();
    ++i;
  }

  UpdateDimension();
}

void MLB::Group::UpdateDimension() {
  if (!visible || groupObjects.empty()) return;

  int x1, y1, x2, y2;
  (*groupObjects.begin())->GetBBox(&x1,&y1,&x2,&y2);

  x = x1;
  y = y1;
  width = x2-x1;
  height = y2-y1;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    (*i)->GetBBox(&x1,&y1,&x2,&y2);
    if (x1 < x) x = x1;
    if (y1 < y) y = y1;
    if ((x2-x) > width) width = x2 - x;
    if ((y2-y) > height) height = y2 - y;
  }
}


void MLB::Group::SetActiveItem(DrawObject* obj) {
  if (!visible || groupObjects.empty()) return;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i) == obj) {
      activeItem = obj->GetID();
      break;
    }
  }
}


void MLB::Group::DrawTextCursor(GUI::Canvas *_canvas) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  DeleteTextCursor(_canvas);
  Text* text = (Text*)GetActiveItem();
  if (text) text->DrawTextCursor(_canvas);
}


void MLB::Group::DeleteTextCursor(GUI::Canvas *_canvas) {
  if (!visible || !textGroup || groupObjects.empty()) return;
  _canvas->DeleteTextCursor(GetID());
}


bool MLB::Group::IsTextGroup() {
  return textGroup;
}


void MLB::Group::InsertNewFont(Page* parentPage, string family, string size, string weight, string slant, string position) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf   = new Buffer(GetChangeFontSizeTG(family));
    FlushInsertNewFontTG(buf->getCont(), actText->GetColor(), family, size, weight, slant, position);
    RefCount<RTPI::RtpiAdu> adu  = CommunicationManager::cm->CreateEvent(GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

  } else { // PRIVATE Session or Annotation Mode
    
    Text* text;
    if ((text = actText->Split(parentPage->GetCanvas())) != NULL) {
      parentPage->AppendObject(text);
      AddObject(text);
      activeItem = actText->GetID();
    }

    int newX = actText->GetX() + actText->GetWidth();
    int newY = actText->GetY();
    text = new Text(newX,newY,0,0,actText->GetColor(),NULL,family,size,weight,slant,position, CommunicationManager::cm->GetObjectID());
    text->SetAnnotation(Annotation()); // same mode as group

    parentPage->AppendObject(text);
    parentPage->SetActiveObject(text);
    AddObject(text);
    ReDraw(parentPage->GetCanvas());
    DrawTextCursor(parentPage->GetCanvas());
  }
}


void MLB::Group::RemoteInsertNewFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos, bool own) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  Text* text;
  if ((text=actText->Split(page->GetCanvas())) != NULL) {
    page->AppendObject(text);
    AddObject(text);
    activeItem = actText->GetID();
  }

  int newX = actText->GetX() + actText->GetWidth();
  int newY = actText->GetY();
  text = new Text(newX,newY,0,0,actText->GetColor(),NULL, fontfamily, fontsize, fontweight, fontslant, textpos, CommunicationManager::cm->GetObjectID());

  page->AppendObject(text);
  if (own) {
    if (Et_GlobalEvalF(Et_Interp, "Attributes::UpdateFont %q %q %q %q %q", fontfamily.c_str(), fontsize.c_str(), fontweight.c_str(), fontslant.c_str(), textpos.c_str()) != TCL_OK) {
      GUI::MessageWindow::Error("TclError in AduManager::HandleEvent: %q\n",Et_Interp->result);
    }
    page->SetActiveObject(text);
  }
  AddObject(text);
}


void MLB::Group::ChangeFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf = new Buffer(GetChangeFontSizeTG(fontfamily));
    FlushChangeFontTG(buf->getCont(), ((Text*)*groupObjects.begin())->GetColor(), fontfamily, fontsize, fontweight, fontslant);
    RefCount<RTPI::RtpiAdu> adu  = CommunicationManager::cm->CreateEvent(GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);
  }
  else { // PRIVATE Session or Annotation Mode
    recorded = false;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::ASCIITEXT)
	((Text*)(*i))->ChangeFont(page->GetCanvas(),fontfamily,fontsize,fontweight,fontslant);
    }
    Update();
    ReDraw(page->GetCanvas());
  }
}


void MLB::Group::RemoteChangeFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT)
      ((Text*)(*i))->ChangeFont(page->GetCanvas(),fontfamily,fontsize,fontweight,fontslant);
  }
  Update();
}


void MLB::Group::InsertNewColor(Page* parentPage, Color color) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf  = new Buffer(GetChangeFontColorSizeTG());
    FlushInsertNewFontColorTG(buf->getCont(), color);
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);
    return;
  } else { // PRIVATE Session or Annotation Mode

    Text* text;
    if ((text = actText->Split(parentPage->GetCanvas())) != NULL) {
      parentPage->AppendObject(text);
      AddObject(text);
      activeItem = actText->GetID();
    }

    int newX = actText->GetX() + actText->GetWidth();
    int newY = actText->GetY();
    text = new Text(newX,newY,0,0,color,NULL,actText->GetFamily(),actText->GetSize(),actText->GetWeight(),
		    actText->GetSlant(),actText->GetPosition(), CommunicationManager::cm->GetObjectID());
    text->SetAnnotation(Annotation());

    parentPage->AppendObject(text);
    parentPage->SetActiveObject(text);
    AddObject(text);
    ReDraw(parentPage->GetCanvas());
    DrawTextCursor(parentPage->GetCanvas());
  }
}


void MLB::Group::RemoteInsertNewColor(Page* page, Color color, bool own) {
  if (!visible || groupObjects.empty() || !textGroup) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;
  Text* text;

  if ((text = actText->Split(page->GetCanvas())) != NULL) {
    page->AppendObject(text);
    AddObject(text);
    activeItem = actText->GetID();
  }

  int newX = actText->GetX() + actText->GetWidth();
  int newY = actText->GetY();

  text = new Text(newX,newY,0,0,color,NULL,actText->GetFamily(),actText->GetSize(),actText->GetWeight(),
		  actText->GetSlant(),actText->GetPosition(), CommunicationManager::cm->GetObjectID());

  page->AppendObject(text);
  if (own) page->SetActiveObject(text);
  AddObject(text);
}


void MLB::Group::ChangeLineColor(GUI::Canvas *_canvas, char *_color) {
  if (!visible || groupObjects.empty()) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;

    if (textGroup) {
      buf = new Buffer(GetChangeFontColorSizeTG());
      FlushChangeFontColorTG(buf->getCont(), _color);
      adu = CommunicationManager::cm->CreateEvent(GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    } else {
      for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
	if (!(*i)->Annotation()) {
	  if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT)) {
	    buf   = new Buffer(((ColoredObject*)(*i))->GetChangeLinecolorSize());
	    ((ColoredObject*)(*i))->FlushChangeLinecolor(buf->getCont(), _color);
	    adu  = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	  }
	} else {//Annotation
	  if ((*i)->GetType() != GenericObject::IMAGE)
	    ((ColoredObject *)(*i))->ChangeLineColor(_canvas,_color);
	}
      }
    }
  }
  else { // PRIVATE Session
    recorded = false;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetType() != GenericObject::IMAGE) {
	((ColoredObject *)(*i))->ChangeLineColor(_canvas,_color);
	if ((*i)->GetType() == GenericObject::ASCIITEXT) (*i)->ReDraw(_canvas);
      }
    }
  }
}


void MLB::Group::RemoteChangeColor(Page* page, Color color) {
  if (!visible || groupObjects.empty()) return;
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) {
      ((Text*)(*i))->ChangeLineColor(page->GetCanvas(),(char *)color.c_str());
      (*i)->ReDraw(page->GetCanvas());
    }
  }
  Update();
}


void MLB::Group::ChangeFillColor(GUI::Canvas *_canvas, char *_color) {
  if (!visible || groupObjects.empty() || textGroup) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if (!(*i)->Annotation()) {
	if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT)) {
	  buf   = new Buffer(((ColoredObject*)(*i))->GetChangeFillcolorSize());
	  ((ColoredObject*)(*i))->FlushChangeFillcolor(buf->getCont(), _color);
	  adu  = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      } else {//Annotation
	if ((*i)->GetType() != GenericObject::IMAGE)
	  ((ColoredObject *)(*i))->ChangeFillColor(_canvas,_color);
      }
    }
  }
  else { // PRIVATE Session or Annotation Mode
    recorded = false;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if ((*i)->GetType() != GenericObject::IMAGE && ((*i)->GetType() != GenericObject::ASCIITEXT))
	((ColoredObject *)(*i))->ChangeFillColor(_canvas,_color);
    }
  }
}


void MLB::Group::ChangeLineStyle(GUI::Canvas *_canvas, char *_lineStyle) {
  if (!visible || groupObjects.empty() || textGroup) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if (!(*i)->Annotation()) {
	if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT)) {
	  buf   = new Buffer(((ColoredObject*)(*i))->GetChangeLinestyleSize());
	  ((ColoredObject*)(*i))->FlushChangeLinestyle(buf->getCont(), _lineStyle);
	  adu  = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      } else {//Annotation
	if ((*i)->GetType() != GenericObject::IMAGE && ((*i)->GetType() != GenericObject::ASCIITEXT))
	  ((ColoredObject *)(*i))->ChangeLineStyle(_canvas, _lineStyle);
      }
    }
  }
  else { // PRIVATE Session or Annotation Mode

    recorded = false;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT))
	((ColoredObject *)(*i))->ChangeLineStyle(_canvas,_lineStyle);
    }
  }
}


void MLB::Group::ChangeLineWidth(GUI::Canvas *_canvas, int _lineWidth) {
  if (!visible || groupObjects.empty() || textGroup) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if (!(*i)->Annotation()) {
	if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT)) {
	  buf   = new Buffer(((ColoredObject*)(*i))->GetChangeLinewidthSize());
	  ((ColoredObject*)(*i))->FlushChangeLinewidth(buf->getCont(), _lineWidth);
	  adu  = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      } else {//Annotation
	if ((*i)->GetType() != GenericObject::IMAGE && ((*i)->GetType() != GenericObject::ASCIITEXT))
	  ((ColoredObject *)(*i))->ChangeLineWidth(_canvas, _lineWidth);
      }
    }
  }
  else { // PRIVATE Session or Annotation Mode

    recorded = false;
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      if (((*i)->GetType() != GenericObject::IMAGE) && ((*i)->GetType() != GenericObject::ASCIITEXT))
	((ColoredObject *)(*i))->ChangeLineWidth(_canvas,_lineWidth);
    }
  }
}


void MLB::Group::Change(Page* page, int argc, char *argv[], DrawObject* active, int _type) {
  if (!visible || groupObjects.empty() || !textGroup) return;
  if (!HasObject(active->GetID())) return;

  std::list<DrawObject*>::iterator i;
  Text* actText;
  Text* text;
  int offX, offY;
  bool deleted;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;

  SetUnrecorded();

  if (activeItem != active->GetID()) {
    activeItem = active->GetID();
  }

  if (textGroup) {
    actText = (Text*)active;

    switch (_type) {
    case MLB::Text::INSERT: {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
	char* newChar = argv[3];
	buf   = new Buffer(GetInsertCharactersSizeTG(newChar));
	FlushInsertCharactersTG(buf->getCont(), newChar);
	adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else { // PRIVATE Session or Annotation Mode
	offX = actText->GetWidth();
	actText->Change(page,argc,argv,_type);
	offX = actText->GetWidth() - offX;
	i = GetActiveItemIterator();
	++i;
	while (i != groupObjects.end()) {
	  text = (Text*)(*i);
	  if (text->GetY() != actText->GetY()) break; // next text line reached; no more adjustments necessary
	  text->SetX(text->GetX()+offX);
	  ++i;
	}
      }
      break;
    }
    case MLB::Text::BACKSPACE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
	buf   = new Buffer(GetDeleteCharactersSizeTG());
	FlushDeleteActiveTG(buf->getCont());
	adu = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else { // PRIVATE Session or Annotation Mode
	if (!actText->firstPos)
	  deleted = actText->DeleteChar(page->GetCanvas());
	else
	  deleted = false;

	while (!deleted) {
	  i = GetActiveItemIterator();
	  if (i == groupObjects.begin()) break;
	  --i; activeItem = (*i)->GetID();
	  text = (Text*)(*i);
	  text->GoEnd();
	  if (actText->GetY() == text->GetY())
	    deleted = text->DeleteChar(page->GetCanvas());
	  else
	    deleted = true;
	}

	text = (Text*)GetActiveItem();
	offY = actText->GetY() - text->GetY();
	offX = text->GetX() + text->GetWidth();

	i = GetActiveItemIterator();
	++i;
	while (i != groupObjects.end()) {
	  text = (Text*)(*i);
	  if (text->GetY() != actText->GetY()) break;
	  text->SetX(offX);
	  offX += text->GetWidth();
	  ++i;
	}

	if (offY > 0) {
	  i = GetActiveItemIterator();
	  ++i;
	  while (i != groupObjects.end()) {
	    text = (Text*)(*i);
	    text->SetY(text->GetY()-offY);
	    ++i;
	  }
	}
	page->SetActiveObject(GetActiveItem());
      }
      break;
    }
    case MLB::Text::LEFT :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	buf   = new Buffer(GetChangeActiveSizeTG());
	FlushSetLeftActiveTG(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else {
	  i = GetActiveItemIterator();
	      while (!actText->GoLeft()) {
	  if (i == groupObjects.begin()) {
	    ((Text*)(*i))->GoHome();
	    break;
          } else {
	  --i; activeItem = (*i)->GetID();
            if ((*i)->GetY() != actText->GetY()) {
              actText = (Text*)(*i);
              actText->GoEnd();
              break;
            } else {
	  actText = (Text*)(*i);
            }
          }
	}
	page->SetActiveObject(actText);
	break;
      }
    case MLB::Text::RIGHT :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	buf   = new Buffer(GetChangeActiveSizeTG());
	FlushSetRightActiveTG(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else {
	  i = GetActiveItemIterator();
	      while (!actText->GoRight()) {
	  if (*i == groupObjects.back()) {
	    ((Text*)(*i))->GoEnd();
	    break;
          } else {
	  ++i; activeItem = (*i)->GetID();
	          if ((*i)->GetY() != actText->GetY()) {
	  actText = (Text*)(*i);
	  actText->GoHome();
              break;
            } else {
              actText = (Text*)(*i);
            }            
          }
	}
	page->SetActiveObject(actText);
	break;
      }
    case MLB::Text::UP :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	      buf   = new Buffer(GetChangeActiveSizeTG());
	      FlushSetUpActiveTG(buf->getCont());
	      adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
        return;
      } else {
	      int x = actText->GetX() + actText->GetCurrentX();  
        i  = GetActiveItemIterator();

        int x1, y1, x2, y2;
        (*i)->GetBBox(&x1,&y1,&x2,&y2);
        if (y1 == GetY()) return;
        
        int offX = (*i)->GetX();
        while ((offX > x) || ((*i)->GetY() == actText->GetY())) {
          ((Text*)(*i))->GoHome();
          --i;
          offX = (*i)->GetX();
        }
        
        activeItem = (*i)->GetID();
	      actText    = (Text*)(*i);
	      page->SetActiveObject(actText);
      
        actText->GoEnd();
        while (offX + actText->GetCurrentX() > x) actText->GoLeft();
      break;
      }
    case MLB::Text::DOWN :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	      buf   = new Buffer(GetChangeActiveSizeTG());
	      FlushSetDownActiveTG(buf->getCont());
	      adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
      	return;
      } else {
	      int x = actText->GetX() + actText->GetCurrentX();  
        i  = GetActiveItemIterator();
	        
        int x1, y1, x2, y2;
        (*i)->GetBBox(&x1,&y1,&x2,&y2);
        if (y2 == (GetY() + GetHeight())) return;

        while ((*i)->GetY() == actText->GetY()) {
          ((Text*)(*i))->GoEnd();
          ++i;
        }
         
        int y    = (*i)->GetY(); 
        int offX = (*i)->GetX();
        
        while ((offX + (*i)->GetWidth() < x) && ((*i)->GetY() == y)) {
          if (*i == groupObjects.back()) break;
          ((Text*)(*i))->GoEnd();
          ++i;
          
          offX = (*i)->GetX();
        }
          
        if ((*i)->GetY() != y) {
          ((Text*)(*i))->GoHome();
          --i;
        }
        
        activeItem = (*i)->GetID();
	      actText    = (Text*)(*i);
	      page->SetActiveObject(actText);
        
        actText->GoEnd();
        while (offX + actText->GetCurrentX() > x) actText->GoLeft();
      break;
      }
    case MLB::Text::END :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	buf   = new Buffer(GetChangeActiveSizeTG());
	FlushSetLastActiveTG(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else {
        for (i = GetActiveItemIterator(); i != groupObjects.end(); ++i) 
          ((Text*)(*i))->GoEnd();
	      --i; 
        activeItem = (*i)->GetID();
	actText = (Text*)(*i);
	page->SetActiveObject(actText);
	break;
      }
    case MLB::Text::HOME :
      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !Annotation()) {
	buf   = new Buffer(GetChangeActiveSizeTG());
	FlushSetFirstActiveTG(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	return;
      }
      else {
        for (i = GetActiveItemIterator(); i != groupObjects.begin(); --i) 
          ((Text*)(*i))->GoHome();
	activeItem = (*i)->GetID();
	actText = (Text*)(*i);
	actText->GoHome();
	page->SetActiveObject(actText);
	break;
      }
    }

    UpdateDimension();
    ReDraw(page->GetCanvas());
    DrawTextCursor(page->GetCanvas());
  }
  else {
    active->Change(page,argc,argv,_type);
  }
}


void MLB::Group::RemoteInsertNewChars(char* newChars) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  int offX = actText->GetWidth();
  actText->RemoteInsertNewChars(newChars);
  offX = actText->GetWidth() - offX;

  Text* text;
  std::list<DrawObject*>::iterator i = GetActiveItemIterator();
  ++i;
  while (i != groupObjects.end()) {
    text = (Text*)(*i);
    if (text->GetY() != actText->GetY()) break; // next text line reached; no more adjustments necessary
    text->SetX(text->GetX()+offX);
    ++i;
  }
}


void MLB::Group::RemoteGoLeft(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  std::list<DrawObject*>::iterator i = GetActiveItemIterator();

  while (!actText->GoLeft()) {
    if (i == groupObjects.begin()) {
      ((Text*)(*i))->GoHome();
      break;
    } else {
    --i; activeItem = (*i)->GetID();
      if ((*i)->GetY() != actText->GetY()) {
        actText = (Text*)(*i);
        actText->GoEnd();
        break;
      } else {
    actText = (Text*)(*i);
  }
    }
  }
  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteGoRight(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  std::list<DrawObject*>::iterator i = GetActiveItemIterator();

  while (!actText->GoRight()) {
	  if (*i == groupObjects.back()) {
	    ((Text*)(*i))->GoEnd();
	    break;
    } else {
    ++i; activeItem = (*i)->GetID();
	    if ((*i)->GetY() != actText->GetY()) {
        actText = (Text*)(*i);
        actText->GoHome();
        break;
      } else {
    actText = (Text*)(*i);
  }
    }
  }
  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteGoHome(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  std::list<DrawObject*>::iterator i = GetActiveItemIterator(); 
  for (; i != groupObjects.begin(); --i) {
    ((Text*)(*i))->GoHome();
  }
  activeItem = (*i)->GetID();
	Text* actText = (Text*)(*i);
  actText->GoHome();

  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteGoEnd(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  std::list<DrawObject*>::iterator i = GetActiveItemIterator();
  for (; i != groupObjects.end(); ++i) {
    ((Text*)(*i))->GoEnd();
  }
  --i;
  activeItem = (*i)->GetID();
	Text* actText = (Text*)(*i);

  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteGoUp(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;
  int x = actText->GetX() + actText->GetCurrentX();  
  
  std::list<DrawObject*>::iterator i  = GetActiveItemIterator();

  int x1, y1, x2, y2;
  (*i)->GetBBox(&x1,&y1,&x2,&y2);
  if (y1 == GetY()) return;
  
  int offX = (*i)->GetX();
  while ((offX > x) || ((*i)->GetY() == actText->GetY())) {
    ((Text*)(*i))->GoHome();
    --i;
    offX = (*i)->GetX();
  }
  
  activeItem = (*i)->GetID();
	actText    = (Text*)(*i);

  actText->GoEnd();
  while (offX + actText->GetCurrentX() > x) actText->GoLeft();

  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteGoDown(bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;
	int x = actText->GetX() + actText->GetCurrentX();  

  std::list<DrawObject*>::iterator i  = GetActiveItemIterator();
	  
  int x1, y1, x2, y2;
  (*i)->GetBBox(&x1,&y1,&x2,&y2);
  if (y2 == (GetY() + GetHeight())) return;

  while ((*i)->GetY() == actText->GetY()) {
    ((Text*)(*i))->GoEnd();
    ++i;
  }
   
  int y    = (*i)->GetY(); 
  int offX = (*i)->GetX();
  
  while ((offX + (*i)->GetWidth() < x) && ((*i)->GetY() == y)) {
    if (*i == groupObjects.back()) break;
    ((Text*)(*i))->GoEnd();
    ++i;
    
    offX = (*i)->GetX();
  }
    
  if ((*i)->GetY() != y) {
    ((Text*)(*i))->GoHome();
    --i;
  }
  
  activeItem = (*i)->GetID();
	actText    = (Text*)(*i);
	
  actText->GoEnd();
  while (offX + actText->GetCurrentX() > x) actText->GoLeft();

  if (own) SessionManager::sm->GetActivePage()->SetActiveObject(actText);
}


void MLB::Group::RemoteSetActive(int x, int y) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  //((Text*)active)->DeleteTextCursor(canvas);
  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    if ((*i)->PointIsInText(x, y)) {
      activeItem = (*i)->GetID();
      std::list<DrawObject*>::iterator k;
      for (k = groupObjects.begin(); k != i; ++k) ((Text*)(*k))->GoEnd();
      for (++k; k != groupObjects.end(); ++k)     ((Text*)(*k))->GoHome();
    }
  } 
}


void MLB::Group::RemoteBackspace(Page* page, bool own) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* text;
  std::list<DrawObject*>::iterator i;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  bool deleted = false;
  if (!actText->firstPos) deleted = actText->DeleteChar(page->GetCanvas());

  while (!deleted) {
    i = GetActiveItemIterator();
    if (i == groupObjects.begin()) break;
    --i; activeItem = (*i)->GetID();
	  text = (Text*)(*i);
	  text->GoEnd();
	  if (actText->GetY() == text->GetY())
	    deleted = text->DeleteChar(page->GetCanvas());
	  else
	    deleted = true;
  }

	text = (Text*)GetActiveItem();
	int offY = actText->GetY() - text->GetY();
	int offX = text->GetX() + text->GetWidth();

  i = GetActiveItemIterator();
  ++i;
  while (i != groupObjects.end()) {
    text = (Text*)(*i);
    if (text->GetY() != actText->GetY()) break;
	  text->SetX(offX);
	  offX += text->GetWidth();
    ++i;
  }

	if (offY > 0) {
  i = GetActiveItemIterator();
	  ++i;
  while (i != groupObjects.end()) {
    text = (Text*)(*i);
	    text->SetY(text->GetY()-offY);
    ++i;
  }
	}
	if (own) page->SetActiveObject(GetActiveItem());
}


void MLB::Group::InsertNewLine(Page* page) {
  if (!visible || !textGroup || groupObjects.empty()) return;

  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !Annotation()) {
    RefCount<Buffer> buf = new Buffer(GetInsertNewlineSizeTG());
    FlushInsertNewlineTG(buf->getCont());
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);
    return;

  } else { // PRIVATE Session or Annotation Mode
    Text* text;

    bool firstPos = false;
    int newX = GetX();
    int newY = actText->GetY();
    
    if ((text = actText->Split(page->GetCanvas())) == NULL) {
      firstPos = actText->firstPos;
      if (firstPos) {
        std::list<DrawObject*>::iterator i = GetActiveItemIterator();
        if (i != groupObjects.begin()) {
          --i; 
          activeItem = (*i)->GetID(); 
        } else {
          activeItem = 0;
        }
    }

    text = new Text(newX,newY,0,0,actText->GetColor(),NULL,actText->GetFamily(),
		    actText->GetSize(),actText->GetWeight(),actText->GetSlant(),actText->GetPosition(),
		    CommunicationManager::cm->GetObjectID());
    }

    text->SetAnnotation(Annotation());
    page->AppendObject(text);
    AddObject(text);

    if (firstPos) activeItem = actText->GetID();
    else page->SetActiveObject(text);

    int offY = text->GetHeight();
    int offX = 0;
    int actY = actText->GetY();
    
    std::list<DrawObject*>::iterator i = GetActiveItemIterator();
    while (i != groupObjects.end()) {
      if ((*i)->GetY() == actY) {
      	(*i)->SetX(newX+offX);
	      offX += (*i)->GetWidth();
      }
      (*i)->SetY((*i)->GetY() + offY);
      ++i;
    }
    
    UpdateDimension();
    ReDraw(page->GetCanvas());
    DrawTextCursor(page->GetCanvas());
  }
}


//MB
void MLB::Group::RemoteNewLine(Page* page, bool own) {
  Text* actText = (Text*)GetActiveItem();
  if (!actText) return;

  Text* text;

  bool firstPos = false;
  int newX = GetX();
  int newY = actText->GetY();
  
  if ((text = actText->Split(page->GetCanvas())) == NULL) {
    firstPos = actText->firstPos;
    if (firstPos) {
      std::list<DrawObject*>::iterator i = GetActiveItemIterator();
      if (i != groupObjects.begin()) {
        --i; 
        activeItem = (*i)->GetID(); 
      } else {
        activeItem = 0;
      }
  }

  text = new Text(newX,newY,0,0,actText->GetColor(),NULL,actText->GetFamily(),
		  actText->GetSize(),actText->GetWeight(),actText->GetSlant(),actText->GetPosition(),
		  CommunicationManager::cm->GetObjectID());
  }

  page->AppendObject(text);
  AddObject(text);

  if (firstPos) {
    activeItem = actText->GetID();
  } else {
    if (own) page->SetActiveObject(text);
  }
  
  int offY = text->GetHeight();
  int offX = 0;
  int actY = actText->GetY();
  
  std::list<DrawObject*>::iterator i = GetActiveItemIterator();
  while (i != groupObjects.end()) {
    if ((*i)->GetY() == actY) {
      (*i)->SetX(newX+offX);
	    offX += (*i)->GetWidth();
    }
    (*i)->SetY((*i)->GetY() + offY);
    ++i;
  }
}


void MLB::Group::LoadObjects(GUI::Canvas *_canvas, MLB::GenericContainer *page, SGMLElement *_root) {
  SGMLElement
    *hilf;
  DrawObject
    *obj;
  char
    *tmp;
  int
    length;

  for (hilf = _root->son; hilf != NULL; hilf = hilf->next) {
    obj = NULL;
    if (hilf->Name == "LINE")           obj = Line::Load(hilf);
    else if (hilf->Name == "RECTANGLE") obj = Rectangle::Load(hilf);
    else if (hilf->Name == "OVAL")      obj = Oval::Load(hilf);
    else if (hilf->Name == "POLYLINE")  obj = PolyLine::Load(hilf);
    else if (hilf->Name == "POLYGON")   obj = Polygon::Load(hilf);
    else if (hilf->Name == "IMAGE")     obj = PhotoImage::Load(hilf);
    else if (hilf->Name == "TEXT")      obj = Text::Load(hilf);
    else if (hilf->Name == "GROUP") {
      obj = Group::Load(hilf);
      ((Group *)obj)->LoadObjects(_canvas,page,hilf);
    }
    if (obj) {
      obj->SetOrder(((Page*)page)->GetPushBackObjectOrder()); //J
      ((MLB::Page *)page)->AppendObject(obj);
      AddObject(obj);
    }
  }
}


void MLB::Group::AdjustGui() {
  if (textGroup && !groupObjects.empty()) {
    Text* txt = (Text*)(*groupObjects.begin());
    txt->AdjustGui();
  }
}


RefCount<RTPI::RtpiAdu> MLB::Group::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;

  if (textGroup) {
    RefCount<Buffer> buf = new Buffer(GetStateSizeTG());
    FlushStateTG(buf->getCont());
    return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::TEXT, priority, buf, timestamp);
  } else {
    RefCount<Buffer> buf = new Buffer(GetStateSize());
    FlushState(buf->getCont());
    return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::GROUP, priority, buf, timestamp);
  }
}


int MLB::Group::FlushState(char* buf, uint32 order, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);

  p8[1] = CommunicationManager::GROUP;

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  return 12;
}


int MLB::Group::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);

  p8[1] = CommunicationManager::GROUP;

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  return 12;
}


int MLB::Group::ParseState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("Group::ParseState - wrong version of mlb protocol");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    SessionManager::sm->message->Error("Group::ParseState - wrong operation type");
    return -1;
  }

  if (p8[1] != CommunicationManager::GROUP) {
    SessionManager::sm->message->Error("Group::ParseState - wrong object type %d", (int)p8[1]);
    return -1;
  }

  order = ntohl(p32[1]);

  x = ntohs(p16[4]);
  y = ntohs(p16[5]);

  return 12;
}


int MLB::Group::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::GROUP;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}


int MLB::Group::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::GROUP;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Group::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::GROUP;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Group::FlushDelete(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::GROUP;

  return 4;
}



int MLB::Group::FlushStateTG(char* buf, uint32 order, bool visible, int x, int y, int width, int height, Color linecolor,
			     string fontfamily, string fontsize, string fontweight, string fontslant, string textpos) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = 0;
  p8[3] = 1;

  p32[1] = htonl(order);

  p16[4] = htons(width);
  p16[5] = htons(height);

  p8[12]  = 0;
  p8[13]  = 1;
  p16[7] = htons(0);

  p16[8] = htons(x);
  p16[9] = htons(y);

  ColoredObject::FlushColor(buf + 20, linecolor);
  p8[23] = atoi(fontsize.c_str());

  p8[24] = 0;
  Text::FlushFontWeight(buf + 24, fontweight);
  Text::FlushFontSlant(buf + 24, fontslant);
  Text::FlushFontPosition(buf + 24, textpos);

  int fflength = fontfamily.length();
  p8[25] = fflength;
  memcpy(buf + 26, fontfamily.c_str(), fflength);

  int padd = ((fflength + 2) % 4 == 0) ? 0 : 4 - ((fflength  + 2) % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[26 + fflength + i] = 0;
  }

  p32[(26 + fflength + padd)/4] = 0; // no text (tlength = 0 etc.)

  return 30 + fflength + padd;
}


int MLB::Group::GetStateSizeTG(string fontFamily) {
  int fflength = fontFamily.length();
  return 30 + fflength + (((fflength + 2) % 4 == 0) ? 0 : 4 - ((fflength  + 2) % 4));
}


int MLB::Group::FlushStateTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::TEXT;

  p16[1] = 0;
  p8[3]  = groupObjects.size();

  p32[1] = htonl(order);

  p16[4] = htons(width);
  p16[5] = htons(height);

  Text* active = (Text*)GetActiveItem();
  p8[12]  = (active) ? GetObjectOrder(active) : 0;
  p8[13]  = (active) ? ((active->firstPos) ? 1 : 0) : 0;
  p16[7]  = (active) ? htons(active->GetCurrentIndex()) : 0;

  int fflength;
  int tlength;
  int padd1;
  int padd2;
  int offset = 16;
  char* tmp;

  Text* text;
  for (std::list<DrawObject*>::iterator t = groupObjects.begin(); t != groupObjects.end(); ++t) {
    text = (Text*)*t;

    p16[offset/2]     = htons(text->x);
    p16[offset/2 + 1] = htons(text->y);

    ColoredObject::FlushColor(buf + offset + 4, text->color);

    p8[offset + 7] = atoi(text->fontSize.c_str());

    p8[offset + 8] = 0;
    Text::FlushFontWeight(buf + offset + 8, text->fontWeight);
    Text::FlushFontSlant(buf + offset + 8, text->fontSlant);
    Text::FlushFontPosition(buf + offset + 8, text->textPosition);

    fflength = text->fontFamily.length();
    p8[offset + 9] = fflength;
    memcpy(buf + offset + 10, text->fontFamily.c_str(), fflength);

    padd1 = ((fflength + 2) % 4 == 0) ? 0 : 4 - ((fflength  + 2) % 4);
    if (padd1 != 0) { // padding needed
      for (int i = 0; i < padd1; ++i)
	p8[offset + 10 + fflength + i] = 0;
    }

    offset += 10 + fflength + padd1;

    tmp = text->GetString();
    tlength = strlen(tmp);
    p16[offset/2] = htons(tlength);

    memcpy(buf + offset + 2, tmp, tlength);

    delete[] tmp;

    padd2 = ((tlength + 2) % 4 == 0) ? 0 : 4 - ((tlength  + 2) % 4);
    if (padd2 != 0) { // padding needed
      for (int i = 0; i < padd2; ++i)
	p8[offset + 2 + tlength + i] = 0;
    }

    offset += 2 + tlength + padd2;
  }

  return offset;
}


int MLB::Group::GetStateSizeTG() {
  Text* text;
  int flength;
  int tlength;
  int bytes = 16;

  for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
    text = (Text*)*i;
    bytes += 8;
    flength = text->fontFamily.length();
    tlength = text->GetStringSize();
    bytes += (2 + flength + (((flength + 2) % 4 == 0) ? 0 : 4 - ((flength  + 2) % 4)));
    bytes += (2 + tlength + (((tlength + 2) % 4 == 0) ? 0 : 4 - ((tlength  + 2) % 4)));
  }

  return bytes;
}


int MLB::Group::ParseStateTG(char* buf, Page* page) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("Text::ParseState - wrong version of mlb protocol");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    SessionManager::sm->message->Error("Text::ParseState - wrong operation type");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;

  if (p8[1] != CommunicationManager::TEXT) {
    SessionManager::sm->message->Error("Text::ParseState - wrong object type %d", (int)p8[1]);
    return -1;
  }

  if (!groupObjects.empty()) {
    for (std::list<DrawObject*>::iterator i = groupObjects.begin(); i != groupObjects.end(); ++i) {
      (*i)->SetGroup(NULL); // prevent recursive delete call 
      page->DeleteObject(*i);
      delete *i;
    }
    groupObjects.clear();
  }

  order = ntohl(p32[1]);

  x = ntohs(p16[8]); //pos of first text element
  y = ntohs(p16[9]);

  int nbr = p8[3];

  Text* text;
  char* txt;
  int _x, _y;
  string fontFamily;
  string fontSize;
  string fontWeight;
  string fontSlant;
  string textPosition;
  Color color;
  char tmp[4]; // 0..255
  char* tmp2;
  int fflength;
  int tlength;
  int offset, padd1, padd2;
  uint32 _order = order += MLB_OO_OFFSET;

  offset = 16;

  for (int i = 1; i <= nbr; ++i) {
    _x = ntohs(p16[offset/2]);
    _y = ntohs(p16[offset/2 + 1]);

    color = ColoredObject::ParseColor(buf + offset + 4);

    sprintf(tmp,"%u", p8[offset + 7]);
    fontSize = string(tmp);

    fontWeight   = Text::ParseFontWeight(buf + offset + 8);
    fontSlant    = Text::ParseFontSlant(buf + offset + 8);
    textPosition = Text::ParseFontPosition(buf + offset + 8);

    fflength = p8[offset + 9];
    tmp2 = new char[fflength + 1];
    memcpy(tmp2, buf + offset + 10, fflength);
    tmp2[fflength] = '\0';
    fontFamily = string(tmp2);
    delete[] tmp2;

    padd1 = ((fflength + 2) % 4 == 0) ? 0 : 4 - ((fflength  + 2) % 4);

    offset += 10 + fflength + padd1;

    tlength = ntohs(p16[offset/2]);
    txt = new char[tlength + 1];
    txt[tlength] = '\0';
    strncpy(txt, buf + offset + 2, tlength);

    text = new Text(_x, _y, 0, 0, color, txt, fontFamily, fontSize, fontWeight, fontSlant, textPosition, CommunicationManager::cm->GetObjectID());
    delete txt;

    text->SetOrder(_order); _order += MLB_OO_OFFSET;
    page->AppendObject(text);
    AddObject(text);

    padd2 = ((tlength + 2) % 4 == 0) ? 0 : 4 - ((tlength  + 2) % 4);
    offset += 2 + tlength + padd2;
  }

  // set active item and cursor position
  int actIdx = p8[12];
  if (actIdx >= groupObjects.size()) {
    activeItem = 0;
  } else {
    std::list<DrawObject*>::iterator a = groupObjects.begin();
    for (int j = 0; j < actIdx; ++j) ++a;
    activeItem = (*a)->GetID();

    if (p8[13]) ((Text*)*a)->GoHome();
    else ((Text*)*a)->SetActive(ntohs(p16[7]));
  }

  return offset;
}



int MLB::Group::FlushChangeVisibilityTG(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}


int MLB::Group::FlushMoveTG(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::MOVE;
  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}


int MLB::Group::FlushChangeSizeTG(char* buf, int x, int y, int width, int height) {
  // not supported yet
  return -1;
}


int MLB::Group::FlushInsertNewFontTG(char* buf, Color linecolor, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::TEXT;

  p16[1] = 0;
  p8[2] = CommunicationManager::INSERT_FONT;

  Text::FlushFontWeight(buf + 3, fontweight);
  Text::FlushFontSlant(buf + 3, fontslant);
  Text::FlushFontPosition(buf + 3, textpos);

  ColoredObject::FlushColor(buf + 4, linecolor);

  p8[7] = atoi(fontsize.c_str());

  int fflength = fontfamily.length();
  p8[8] = fflength;
  memcpy(buf + 9, fontfamily.c_str(), fflength);

  int padd = ((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[9 + fflength + i] = 0;
  }

  return 9 + fflength + padd;
}


int MLB::Group::ParseInsertNewFontTG(char* buf, Color& linecolor, string& fontFamily, string& fontSize, string& fontWeight,
				     string& fontSlant, string& textpos) {
  uint8*  p8  = (uint8*)buf;

  // font size
  char tmp[4]; // 0..255
  sprintf(tmp,"%u", p8[7]);
  fontSize = string(tmp);

  // font family
  int fflength = p8[8];
  char* tmp2 = new char[fflength + 1];
  memcpy(tmp2, buf + 9, fflength);
  tmp2[fflength] = '\0';
  fontFamily = string(tmp2);
  delete[] tmp2;

  fontWeight = Text::ParseFontWeight(buf + 3);
  fontSlant  = Text::ParseFontSlant(buf + 3);
  textpos    = Text::ParseFontPosition(buf + 3);

  linecolor  = ColoredObject::ParseColor(buf + 4);

  int padd = ((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4);
  return 9 + fflength + padd;
}


int MLB::Group::FlushChangeFontTG(char* buf, Color linecolor, string fontfamily, string fontsize, string fontweight, string fontslant) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::TEXT;

  p16[1] = 0;
  p8[2] = CommunicationManager::CHG_FONT;

  Text::FlushFontWeight(buf + 3, fontweight);
  Text::FlushFontSlant(buf + 3, fontslant);

  ColoredObject::FlushColor(buf + 4, linecolor);

  p8[7] = atoi(fontsize.c_str());

  int fflength = fontfamily.length();
  p8[8] = fflength;
  memcpy(buf + 9, fontfamily.c_str(), fflength);

  int padd = ((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[9 + fflength + i] = 0;
  }

  return 9 + fflength + padd;
}


int MLB::Group::ParseChangeFontTG(char* buf, string& fontFamily, string& fontSize, string& fontWeight, string& fontSlant) {
  uint8*  p8  = (uint8*)buf;

  // font size
  char tmp[4]; // 0..255
  sprintf(tmp,"%u", p8[7]);
  fontSize = string(tmp);

  // font family
  int fflength = p8[8];
  char* tmp2 = new char[fflength + 1];
  memcpy(tmp2, buf + 9, fflength);
  tmp2[fflength] = '\0';
  fontFamily = string(tmp2);
  delete[] tmp2;

  fontWeight = Text::ParseFontWeight(buf + 3);
  fontSlant  = Text::ParseFontSlant(buf + 3);

  int padd = ((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4);
  return 9 + fflength + padd;
}


int MLB::Group::GetChangeFontSizeTG(string fontFamily) {
  int fflength = fontFamily.length();
  return 9 + fflength + (((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4));
}


int MLB::Group::FlushChangeFontFamilyTG(char* buf, string fontfamily) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_FONTFAMILY;
  p8[3] = 0;

  int fflength = fontfamily.length();
  p8[4] = fflength;
  memcpy(buf + 5, fontfamily.c_str(), fflength);

  int padd = ((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[5 + fflength + i] = 0;
  }

  return 5 + fflength + padd;
}


int MLB::Group::GetChangeFontFamilySizeTG(string fontFamily) {
  int fflength = fontFamily.length();
  return 5 + fflength + (((fflength + 1) % 4 == 0) ? 0 : 4 - ((fflength  + 1) % 4));
}


int MLB::Group::FlushChangeFontSizeTG(char* buf, string fontsize) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_FONTSIZE;
  p8[3] = 0;

  p8[15] = atoi(fontsize.c_str());

  return 8;
}


int MLB::Group::FlushChangeFontColorTG(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_LINECOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Group::FlushInsertNewFontColorTG(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::INSERT_FCOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Group::FlushChangeFontWeightTG(char* buf, string fontweight) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_FONTWEIGHT;

  p8[3] = 0;
  Text::FlushFontWeight(buf + 3, fontweight);

  return 4;
}


int MLB::Group::FlushChangeFontSlantTG(char* buf, string fontslant) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_FONTSLANT;

  p8[3] = 0;
  Text::FlushFontSlant(buf + 3, fontslant);

  return 4;
}


int MLB::Group::FlushChangeFontPositionTG(char* buf, string fontposition) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_FONTPOS;

  p8[3] = 0;
  Text::FlushFontPosition(buf + 3, fontposition);

  return 4;
}


int MLB::Group::FlushInsertCharactersTG(char* buf, char* newChars) {
  #ifdef DEBUG
  SessionManager::sm->message->Info("Text::FlushInsertCharacters - insert at position %u text %q ", idx, newChars);
  #endif

  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::INSERT_CHARS;
  p8[3] = 0;

  int length = strlen(newChars);

  p16[2] = 0;
  p16[3] = htons(length);

  strncpy(buf + 8, newChars, length);

  int padd = (length % 4 == 0) ? 0 : 4 - (length % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[8 + length + i] = 0;
  }

  return 8 + length + padd;
}


int MLB::Group::GetInsertCharactersSizeTG(char* newChars) {
  int length = strlen(newChars);
  return 8 + length + ((length % 4 == 0) ? 0 : 4 - (length % 4));
}


int MLB::Group::FlushInsertNewlineTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::INSERT_CHARS;
  p8[3] = 1;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}

int MLB::Group::FlushDeleteCharactersTG(char* buf, int nbr) {
  #ifdef DEBUG
  SessionManager::sm->message->Info("Text::FlushDeleteCharacters - delete from %u to %u", startidx, stopidx);
  #endif

  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::DELETE_CHARS;
  p8[3] = 0;

  p16[2] = 0;
  p16[3] = htons(nbr);

  return 8;
}


int MLB::Group::FlushDeleteActiveTG(char* buf) {
  #ifdef DEBUG
  SessionManager::sm->message->Info("Text::FlushDeleteCharacters - delete active");
  #endif

  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::DELETE_CHARS;
  p8[3] = 1;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushChangeParentObjectTG(char* buf, int nbr) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_PARENT;
  p8[3] = 0;

  p16[2] = htons(nbr);
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetFirstActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 1;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetLastActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 2;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetLeftActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 3;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetRightActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 4;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetUpActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 5;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetDownActiveTG(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 6;

  p16[2] = 0;
  p16[3] = 0;

  return 8;
}


int MLB::Group::FlushSetActiveTG(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 7;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

    
int MLB::Group::FlushRaiseTG(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Group::FlushLowerTG(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::TEXT;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Group::FlushDeleteTGS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::TEXT;

  return 4;
}


MLB::DrawObject *MLB::Group::Load(SGMLElement *hilf)
{
  Group
    *obj;

  obj = new MLB::Group(CommunicationManager::cm->GetObjectID());

  LoadAnimationStep( obj, hilf );

  return obj;
}

