/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_gui.h
 authors : Gabriela Maass, Juergen Vogel
 date    : 22.11.2000 
 desc    : Functions to change tcl values and widgets

***********************************************************************************************************/

#ifndef __MLB_GUI__
#define __MLB_GUI__


#ifdef WIN32
#  pragma warning (disable: 4786)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <list>
#include <vector>

#ifndef WIN32
#  include <unistd.h>
#else
#  include <windows.h>
#endif

#ifdef _ImageMagick
  namespace MagickLib {
    #undef PACKAGE
    #undef VERSION
    #undef class
  }
  #include <Magick++.h>
#endif

#include "tk.h"
#include "tcl.h"

#include <util/types.h>
#include "mlb.h"
#include "mlb_types.h"

namespace GUI {


  class Attributes
    {
    public:
      Attributes();
      ~Attributes();

      static void SetFont(char *family, char *size, char *weight, char *slant, char *position);
      static void SetOutlineColor(char *color);
      static void SetFillColor(char *color);
      static void SetLineWidth(char *lineWidth);
      static void SetLineStyle(char *lineStyle);
      static void SetCapStyle(char *capStyle);
      static void SetJoinStyle(char *joinStyle);
      static void SetArrowStyle(char *arrowStyle);
      static void SetArrowShape(char *arrowShape);
      static void SetTool(char *tool);
      static void SetPageBackground(const char *color);
      static void SetPageDimension(const int width, const int height);
      static void ResetAction();
    };

  class Cursor
    {
    protected:
      Tk_Window
	tkWin;
      static Tk_Cursor
	watch,
	point,
	pointer_ro,
	pointer_ru,
	pointer_lo,
	pointer_lu;
      int
	xHot,
	yHot;

    public:
      enum type {
	WATCH
      };
      Cursor(char *widget=NULL);
      ~Cursor();

      void Set(int which);
      void Set(const char* which, const char* color);
      void Reset();
      int GetXHot();
      int GetYHot();
    };

  class Progress
    {
    public:
      Progress();
      ~Progress();

      void ShowProgress(char *title, int max);
      void IncrementProgress();
      void HideProgress();

    };

  class MessageWindow
    {
    public:
      MessageWindow();
      ~MessageWindow();


      static void Error(char *message, ...);
      static void Info(char *message, ...);
      static void InfoBlue(char *format, ...);
      static void Debug(char *message, ...);
    };

  class Record
    {
    protected:
      int
	time;

    public:
      Record() {}
      ~Record() {}

      static void Stop();
    };

  class Toplevel
    {
      Toplevel() {}
      ~Toplevel() {}

    public:
      static void SetMenuState(char *menustate);
      static void SetStatus(char *status);
    };

  class Canvas
    {
    protected:
     Tk_Window
	tkWin;
      char
	*widget;
      int
	width,
	height;


    public:
      Cursor
	*cursor;

      Canvas();
      Canvas(uint32 id, int width, int height, const char *color);
      ~Canvas();

      int  Create(uint32 id, int _width, int _height, const char *color, bool thumbnailPage = false);
      int  Create();
      void Delete();
      char *GetWidget();
      void SetDimension(int _width, int _height);
      void Hide();
      void Show();
      
      bool GetAsImage(Magick::Image& image);
      void SaveAsImage(char* name, char* type);
      void Clear(int width, int height);
      void SetBackground(const char *color);

      void BeginTicks(int x, int y);
      void MoveTicks(int x, int y);
      void EndTicks(int x, int y);

      void BeginSelectRegion(int x, int y);
      void ModifySelectRegion(int firstX, int firstY,int x, int y);
      void DeleteSelectRegion();

      void BeginHelpLine(int x, int y);
      void ModifyHelpLine(int firstX, int firstY,int x, int y);
      void DeleteHelpLine();

      void SetCursor(const char* name, const char* color);
      void ResetCursor();

      void MoveObject(const char *ref, double diffX, double diffY);
      bool Overlap(const char *ref, int x, int y);
      /*** Draw graphic objects ***/

      void DrawLine(const char *ref, int x, int y, int width, int height, const char *color, int lineWidth,
		    const char *dash, const char *capStyle, const char *joinStyle, const char *arrowStyle, const char *arrowShape);
      void DrawRectangle(const char *ref, int x, int y, int width, int height, const char *color, const char *fillColor, int lineWidth, const char *lineStyle);
      void DrawRectangle(const char *ref, int x, int y, int width, int height, const char *color, int lineWidth, const char *lineStyle);
      void DrawOval(const char *ref, int x, int y, int width, int height, const char *color, const char *fillColor, int lineWidth, const char *lineStyle);
      void DrawOval(const char *ref, int x, int y, int width, int height, const char *color, int lineWidth, const char *lineStyle);
      void DrawImage(const char *ref, int x, int y, const char *imgTag);

      void DrawPolyLine(const char *ref, std::list<MLB::Point *> points, const char *color, int lineWidth, const char *lineStyle,
			const char *capStyle, const char *joinStyle, const char *arrowStyle, const char *arrowShape);

      void DrawPolygon(const char *ref, std::list<MLB::Point*> points, const char *color, int lineWidth,const char *lineStyle);
      void DrawPolygon(const char *ref, std::list<MLB::Point*> points, const char *color, const char *fillColor, int lineWidth,const char *lineStyle);
      void DrawLineSegment(const char *ref, int x, int y, int width, int height, const char *color, int lineWidth, const char *lineStyle);
      void DeleteLineSegments(const char *ref);
      void DrawSelectionRectangle(const char *ref, int x, int y);
      void DeleteSelection(const char *ref);
      void DeleteObject(const char *ref);
      void DeleteImage(const char *tag);
      void UnsetObject(const char *ref);
      void DrawTextCursor(uint32 id, int x, int y, int ascent, int descent, const char *color);
      void DeleteTextCursor(uint32 id);

      void DrawTelepointer(const char *ref, const char *name, uint32 id, int x, int y, const char *color);

      void ChangeFillColor(const char *ref, const char *color);
      void ChangeLineColor(const char *ref, const char *color);
      void ChangeLineWidth(const char *ref, int lineWidth);
      void ChangeLineStyle(const char *ref, const char *lineStyle);
      void ChangeCapStyle(const char *ref, const char *capStyle);
      void ChangeJoinStyle(const char *ref, const char *joinStyle);
      void ChangeArrowStyle(const char *ref, const char *arrowStyle);
      void ChangeArrowShape(const char *ref, const char *arrowShape);
      void ModifyCoords(const char *ref, int x, int y, int width, int height);
      void ModifyPosition(const char *ref, int x, int y);
      void Scale(const char *ref, float scaleFactor);
    };

  class PageView
    {
    public:

      PageView();
      ~PageView();

      static void Init(const char *nodeName);
      static void Delete(const char *nodeName);
      static void InsertPageNode(char *nodeNumber, char *nodeName,char *containerName, char *parentNode, char *nodeData, int pos);
      static void DrawPageNode(char *nodeNumber, char *nodeName,char *containerName, char *nodeData);

      static void InsertSession(const char *nodeName,const char *containerName, const char *parentNode, const char *nodePath, int pos);
      static void InsertFolderNode(const char *nodeName,const char *containerName, const char *parentNode, const char *nodeData, int pos, bool thumbnailFolder = false);
      static void DrawFolderNode(const char *nodeName,const char *containerName, const char *nodeData);
      static void OpenFolderNode(const char *nodeName);

      static void SelectNode(const char *nodeName);
      static void SelectFolderNode(const char *nodeName);

      static void Redraw();
      static void ActivateNewPageButton(bool on);
      static void ActivateAnnotationButton(bool on);
      static void ActivateSynchButton(bool on);

      /*

	ClearPage

       */
    };


  class Chat
    {
    public:
      Chat();
      ~Chat();

      void ClearMessages();
      void ShowMessages(char *format, ...);
      void EnableMessageEntry();
    };

  class Image
    {
    public:
      Image() {};
      ~Image() {};

      static void Create(const char *tag);
      static void Destroy(const char *tag);
      static bool Exists    ( const char *tag );
      static void DestroyImages();
      static void DestroyCharImages();
    };
}

#endif
