/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_image.cpp
 authors : Gabriela Maass, Juergen Vogel, Andrius Kurtinaitis
 date    : 04.04.2000
 desc    : Functions to read, draw and manipulate images

***********************************************************************************************************/

//#include <Magick++.h>

#include "mlb.h"
#include "mlb_session.h"
#include "mlb_objects.h"
#include "mlb_adu.h"
#include "mlb_misc.h"
#include "mlb_tkphoto.h"

static int zaehl = 0;

/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

MLB::PhotoImage::PhotoImage( uint32 id, char* payload, GUI::Canvas* canvas )

  : DrawObject(id),
    filename(""),
    defaultImg(false),
    lastZoom(0.0)
{
  InitTkTags( id );
  ParseState( payload, canvas );

} // PhotoImage()

//-------------------------------------------------------------------------//

/**   Copy constructor
  **/
MLB::PhotoImage::PhotoImage( const MLB::PhotoImage& _image )

  : DrawObject( _image.x, _image.y, _image.width, _image.height, CommunicationManager::cm->GetObjectID() ),
    filename  ( _image.filename ),
    blob      ( _image.blob     ),
    tkBlock   ( false           ),
    defaultImg( _image.defaultImg ),
    lastZoom(0.0)
{
  InitTkTags  ( id );
  //PutTkBlock();

} // PhotoImage()

//-------------------------------------------------------------------------//

MLB::PhotoImage::PhotoImage( uint32 _id,    char*   _filename,
                             int    _x,     int     _y,
                             int    _width, int     _height,
                             char*  _blob,  unsigned int _blength  )

  : DrawObject( _x, _y, _width, _height, _id ),
    filename  ( _filename ),
    tkBlock   ( false ),
    defaultImg( false ),
    lastZoom(0.0)
{
  InitTkTags  ( id );
  InitFileName();
  blob.update( _blob, _blength );

} // PhotoImage()

//-------------------------------------------------------------------------//

MLB::PhotoImage::PhotoImage( uint32 _id,    char*   _filename,
                             int    _x,     int     _y,
                             Magick::Image& _image            )

  : DrawObject( _x, _y, _image.columns(), _image.rows(), _id ),
    filename  ( _filename ),
    tkBlock   ( false ),
    defaultImg( false ),
    lastZoom(0.0)
{
  InitTkTags  ( id );
  InitFileName();
  layer = 1;
  // Blob
  _image.quality(100);
  _image.write( &blob, "PNG" );

} // PhotoImage()

//-------------------------------------------------------------------------//

MLB::PhotoImage::~PhotoImage()
{ DeInit(); } // ~PhotoImage()

//-------------------------------------------------------------------------//

/**   Usually called from destructor and form Delete(). **/
void MLB::PhotoImage::DeInit()
{
  if( tkBlock ) RemoveTkBlock();
  imgTag.erase();
  filename.erase();
  blob.updateNoCopy( NULL, 0 );

} // DeInit()

//-------------------------------------------------------------------------//
/**   Usually called when a exception is triggered **/
void MLB::PhotoImage::LoadDefaultImage() 
{
  filename = "../gifs/default_img.gif";
  
  try { 
    // read image
    std::vector<Magick::Image> imageList;
    Magick::readImages(&imageList, filename); 
    Magick::Image image = *imageList.begin();

    x     += (width  - image.columns()) / 2;
    y     += (height - image.rows()   ) / 2;
    width  = image.columns();
    height = image.rows();

    InitFileName();

    // Blob
    image.quality(100);
    image.write(&blob, "PNG");

    defaultImg = true;
  } catch (Magick::Exception& exc) {
    GUI::MessageWindow::Error("Cannot read images: %q", exc.what());
  }
} // LoadDefaultImage()

//-------------------------------------------------------------------------//

/**   Init Tk tag strings used to identify Tk objects.
  **/
void MLB::PhotoImage::InitTkTags( int id )
{
  char buf[20];

  sprintf( buf, "%u", id );
  tkref = "Image";
  tkref += buf;

  imgTag = "imgTag";
  imgTag += buf;

} // InitTkTags()

//-------------------------------------------------------------------------//

/**   Init filename.
  **/
void MLB::PhotoImage::InitFileName()
{
  /* cut path, change suffix */
  string::size_type pos;
//  #ifdef WIN32
//    pos = filename.find_last_of("\\");
//  #else
//    pos = filename.find_last_of("/");
//  #endif
  pos = filename.find_last_of("/");
  if (pos != string::npos) {
    pos++;
    filename.replace(0,pos,"");
  }

  pos = filename.find_last_of(".");
  if (pos != string::npos) {
    pos++;
    filename.replace(pos,string::npos,"png");
  }

} // InitFileName()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::PutTkBlock()
{
  try {
  // image
  Magick::Image image( blob );
  image.quality(100);

  // resize
  if( image.columns() != width  ||
      image.rows()    != height ||
      SessionManager::sm->GetZoom() != 1.0)
  {
    Magick::Geometry newSize( SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height );
    newSize.aspect( true );
    image.scale( newSize );
  }

  // put
  TkImageBlock  block( image );
  block.putBlock( imgTag );
  tkBlock = true;
  } catch (Magick::Exception& exc) {
    GUI::MessageWindow::Error("Cannot put Tk photo block: %q\n", exc.what());
    LoadDefaultImage();
    if (defaultImg) PutTkBlock();
  }
} // PutTkBlock()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::RemoveTkBlock()
{
  if( !imgTag.empty() ) { GUI::Image::Destroy( imgTag.c_str() ); }
  tkBlock = false;

} // RemoveTkBlock()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::Delete()
{
  DeInit();
} // Delete()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::Update(GUI::Canvas *_canvas)
{
  if (!visible) return;
  ModifyCoords(_canvas,x,y,width,height);
}

//-------------------------------------------------------------------------//

MLB::DrawObject *MLB::PhotoImage::Copy()
{
  PhotoImage* ret = NULL;
  try { ret = new PhotoImage( *this ); }
  catch ( Magick::Exception& exc )
  { GUI::MessageWindow::Error( "PhotoImage::Copy - cannot create image: %q", exc.what() ); }
  catch ( string& str )
  { GUI::MessageWindow::Error( "PhotoImage::Copy - cannot create image: %q", str.c_str() ); }
  catch ( ... )
  { GUI::MessageWindow::Error( "PhotoImage::Copy - cannot create image." ); }

  return ret;
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::Draw(GUI::Canvas *_canvas)
{
  if (!visible) return;

  if (lastZoom != SessionManager::sm->GetZoom()) {
    lastZoom = SessionManager::sm->GetZoom();
    if ( tkBlock ) RemoveTkBlock();
  }

  if( !GUI::Image::Exists( imgTag.c_str() ) ) PutTkBlock();
  _canvas->DrawImage(tkref.c_str(),
		     SessionManager::sm->GetZoom()*x,
		     SessionManager::sm->GetZoom()*y,
		     imgTag.c_str());

  if (selected) DrawSelection(_canvas);
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::ReDraw(GUI::Canvas *_canvas)
{
  if (!visible) return;

  _canvas->DeleteObject(tkref.c_str());
  Draw(_canvas);
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::Move(GUI::Canvas *_canvas, int _diffX, int _diffY, bool redraw)
{
  if (!visible) return;

  recorded = false;

  x += _diffX;
  y += _diffY;
  
  if (redraw) {
    _canvas->MoveObject(tkref.c_str(),SessionManager::sm->GetZoom()*_diffX,SessionManager::sm->GetZoom()*_diffY);
    if (selected) DrawSelection(_canvas);
  }
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::ModifyCoords(GUI::Canvas *_canvas,int _x, int _y , int _width, int _height)
{
  if (!visible) return;

  x = _x;
  y = _y;

  try   { Resize(_width,_height); }
  catch ( Magick::Exception& exc )
  {
    GUI::MessageWindow::Error( "PhotoImage::ModifyCoords - cannot resize image: %q", exc.what() );
  }

  _canvas->ModifyPosition(tkref.c_str(),SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);

} // ModifyCoords()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::ModifyCoords(int _x, int _y , int _width, int _height)
{
  if (!visible) return;

  x = _x;
  y = _y;

  try   { Resize(_width,_height); }
  catch ( Magick::Exception& exc )
  {
    GUI::MessageWindow::Error( "PhotoImage::ModifyCoords - cannot resize image: %q", exc.what() );
  }
} // ModifyCoords()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::Resize(int _width, int _height)
{
  if (!visible) return;

  // set new size
  width  = _width;
  height = _height;

  // update Tk block
  RemoveTkBlock();
  PutTkBlock();

} // Resize()

//-------------------------------------------------------------------------//

void MLB::PhotoImage::DrawSelection(GUI::Canvas *_canvas)
{
  int i;

  if (!visible) return;
  if (selected) DeleteSelection(_canvas);

  selected = true;

  selectPoints[0] = new Point(x,y);
  selectPoints[1] = new Point(x+(width/2),y);
  selectPoints[2] = new Point(x+width,y);
  selectPoints[3] = new Point(x+width,y+(height/2));
  selectPoints[4] = new Point(x+width,y+height);
  selectPoints[5] = new Point(x+(width/2),y+height);
  selectPoints[6] = new Point(x,y+height);
  selectPoints[7] = new Point(x,y+(height/2));

  for (i=0; i < 8; i++)
    _canvas->DrawSelectionRectangle(tkref.c_str(),
				    SessionManager::sm->GetZoom() * selectPoints[i]->x,
				    SessionManager::sm->GetZoom() * selectPoints[i]->y);
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::DeleteSelection(GUI::Canvas *_canvas)
{
  if (!visible || !selected) return;

  selected = false;

  _canvas->DeleteSelection(tkref.c_str());
}

//-------------------------------------------------------------------------//

bool MLB::PhotoImage::IsEditablePoint(int _x, int _y) {
  if (!visible) return false;

  for (int i = 0; i < 8; ++i) {
    if ((_x > selectPoints[i]->x-TOLERANZ) && (_x < selectPoints[i]->x+TOLERANZ)) {
      if ((_y > selectPoints[i]->y-TOLERANZ) && (_y < selectPoints[i]->y+TOLERANZ)) {
	selectCount = i;
	return true;
      }
    }
  }
  return false;
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::SaveObject(FILE *fp, int level)
{
  std::string
    tab;
  int
    i;

  if (!visible) return;

  for (i=0; i<level; i++)
    tab += "  ";

//if (blob) {
    fprintf(fp,"%s<IMAGE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" filename=\"%s\" length=\"%d\"%s>\n",tab.c_str(),x,y,width,height,(char *)filename.c_str(), blob.length(), GetAnimationStepSGMLText().c_str() );
    fprintf(fp,"\"");
    fwrite( blob.data(), 1, blob.length(), fp);
    fprintf(fp,"\"");
/*
  }
  else {
    fprintf(fp,"%s<IMAGE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" filename=\"%s\" length=0>\n",tab.c_str(),x,y,width,height,(char *)filename.c_str(),blobLength);
  }
*/
  fprintf(fp,"\n%s</IMAGE>\n",tab.c_str());

  tab.erase();
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::RecordableObjects()
{
  if (!visible) return 0;

  return 1;
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::SaveAsAOF(FILE *fp)
{
  if (recorded) return;

  recorded = true;

  recID = MLB::SessionManager::sm->recCount;
  MLB::SessionManager::sm->recCount++;
  fprintf(fp,"<IMAGE x=%d y=%d width=%d height=%d fname=%s layer=%d></IMAGE>\n",x,y,width,height,filename.c_str(),layer);
}

//-------------------------------------------------------------------------//

void MLB::PhotoImage::SaveAsPs(FILE *fp, int pageWidth, int pageHeight, bool rotate)
{
  try {

    if( !GUI::Image::Exists( imgTag.c_str() ) ) PutTkBlock();
    TkImageBlock    block( imgTag );
    unsigned char*  p;
    int             i, j, k;

  fprintf(fp,"/origstate save def\n");
  fprintf(fp,"20 dict begin\n");
  fprintf(fp,"/pix %d string def\n",3*width);
  fprintf(fp,"/grays %d string def\n",width);
  fprintf(fp,"/npixls 0 def\n");
  fprintf(fp,"/rgbindx 0 def\n");
  if (rotate) {
    fprintf(fp,"%d %d translate\n",y+height,x);
    fprintf(fp,"%d %d scale\n",height,width);
  }
  else {
    fprintf(fp,"%d %d translate\n",x,pageHeight-(y+height));
    fprintf(fp,"%d %d scale\n",width,height);
  }
  fprintf(fp,"/colorimage where\n");
  fprintf(fp,"  { pop }\n");
  fprintf(fp,"  {\n");
  fprintf(fp,"    /colortogray {\n");
  fprintf(fp,"      /rgbdata exch store\n");
  fprintf(fp,"      rgbdata length 3 idiv\n");
  fprintf(fp,"      /npixls exch store\n");
  fprintf(fp,"      /rgbindx 0 store\n");
  fprintf(fp,"      0 1 npixls 1 sub {\n");
  fprintf(fp,"        grays exch\n");
  fprintf(fp,"        rgbdata rgbindx       get 20 mul\n");
  fprintf(fp,"        rgbdata rgbindx 1 add get 32 mul\n");
  fprintf(fp,"        rgbdata rgbindx 2 add get 12 mul\n");
  fprintf(fp,"        add add 64 idiv\n");
  fprintf(fp,"        put\n");
  fprintf(fp,"        /rgbindx rgbindx 3 add store\n");
  fprintf(fp,"      } for\n");
  fprintf(fp,"      grays 0 npixls getinterval\n");
  fprintf(fp,"    } bind def\n");
  fprintf(fp,"    /mergeprocs { \n");
  fprintf(fp,"      dup length\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      length\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      5 1 roll\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      add\n");
  fprintf(fp,"      array cvx\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      3 -1 roll\n");
  fprintf(fp,"      0 exch\n");
  fprintf(fp,"      putinterval\n");
  fprintf(fp,"      dup\n");
  fprintf(fp,"      4 2 roll\n");
  fprintf(fp,"      putinterval\n");
  fprintf(fp,"    } bind def\n");
  fprintf(fp,"/colorimage {\n");
  fprintf(fp,"      pop pop\n");
  fprintf(fp,"      {colortogray} mergeprocs\n");
  fprintf(fp,"      image\n");
  fprintf(fp,"    } bind def\n");
  fprintf(fp,"  } ifelse\n");
  fprintf(fp,"%d %d 8\n",width,height);
  fprintf(fp,"[%d 0 0 %d 0 %d]\n",width,-height,height);
  if (rotate)
    fprintf(fp,"90 rotate\n");
  fprintf(fp,"{currentfile pix readhexstring pop}\n");
  fprintf(fp,"false 3 colorimage\n");

    p = block.pixelPtr;
    for (i=0; i<block.height; i++) {
      for (j=0; j<block.width; j++) {
        for (k=0; k<3; k++) {
          fprintf(fp,"%.2x",*p);
          p++;
        }
        p++;
      }
      fprintf(fp,"\n");
    }
    fprintf(fp,"end origstate restore 1 1 scale\n");
  }
  catch ( Magick::Exception& exc )
  {
    GUI::MessageWindow::Error( "PhotoImage::SaveAsPs - cannot save image. %q", exc.what() );
}

} // SaveAsPs()

//-------------------------------------------------------------------------//

RefCount<RTPI::RtpiAdu> MLB::PhotoImage::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;

  RefCount<Buffer> buf  = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::IMAGE, priority, buf, timestamp);
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::GetStateSize()
{
  int bLength = blob.length();
  return 20 + bLength + ((bLength % 4 == 0) ? 0 : 4 - (bLength % 4));

} // GetStateSize()

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::PNG;
  p8[3] = 0;

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  p16[6] = htons(width);
  p16[7] = htons(height);

  p32[4] = htonl(blob.length());

  memcpy(buf + 20, blob.data(), blob.length() );

  int padd = (blob.length() % 4 == 0) ? 0 : 4 - (blob.length() % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[20 + blob.length() + i] = 0;
  }

  return 20 + blob.length() + padd;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::ParseState( char* buf, GUI::Canvas* _canvas )
{
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("PhotoImage::ParseState - wrong version of mlb protocol");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    SessionManager::sm->message->Error("PhotoImage::ParseState - wrong operation type");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;

  if (p8[1] != CommunicationManager::IMAGE) {
    SessionManager::sm->message->Error("PhotoImage::ParseState - wrong object type %u", (int)p8[1]);
    return -1;
  }

  order = ntohl(p32[1]);

  // set type
  x = ntohs(p16[4]);
  y = ntohs(p16[5]);

  width  = ntohs(p16[6]);
  height = ntohs(p16[7]);

  int length = ntohl(p32[4]);

  try { blob.update( buf+20, length ); }
  catch ( Magick::Exception& exc )
  {
    GUI::MessageWindow::Error( "PhotoImage::ParseState - cannot build image: %q", exc.what() );
    return -1;
  }

  int padd = (length % 4 == 0) ? 0 : 4 - (length % 4);
  return 20 + length + padd;

} // ParseState()

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushChangeVisibility(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushChangeSize(char* buf, int x, int y, int width, int height) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::CHG_SIZE;

  p8[3] = 0;
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p16[2] = htons(x);
  p16[3] = htons(y);

  p16[4] = htons(abs(width));
  p16[5] = htons(abs(height));

  return 12;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::IMAGE;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}

//-------------------------------------------------------------------------//

int MLB::PhotoImage::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::IMAGE;

  return 4;
}

/**** Static Functions *****************************************************/

MLB::DrawObject *MLB::PhotoImage::Load(SGMLElement *hilf)
{
  PhotoImage* obj;
  char*       tmp = NULL;
  int         x, y,
              width, height, length;
  char*       filename;

  tmp = hilf->GetAttribute("x");
  if (tmp) x = atoi(tmp);
  else     x = 0;
  tmp = hilf->GetAttribute("y");
  if (tmp) y = atoi(tmp);
  else     y = 0;
  tmp = hilf->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = 0;
  tmp = hilf->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = 0;
  tmp = hilf->GetAttribute("filename");
  if (tmp) filename = strdup(tmp);
  else     filename = strdup("");

  if( !hilf->Parameter.empty() ) {
    // the image is included in the mlb document read blob
    tmp = hilf->GetAttribute("length");
    if (tmp) length = atoi(tmp);
    else     length = 0;

    try {
      obj = new MLB::PhotoImage( CommunicationManager::cm->GetObjectID(),
				 filename, x, y, width, height,
				 (char*) hilf->Parameter.c_str(), length );
      obj->SetLayer(0);
    }
    catch ( Magick::Exception& exc )
      {
	GUI::MessageWindow::Error( "Cannot build image. %q", exc.what() );
        delete[] filename;
        return NULL;
      }

  } else {
    // the image is an external file

    try {
      std::string name = SessionManager::sm->GetActiveSession()->GetDocumentPath();
      name += filename;
      Magick::Image image (name);
      obj = new MLB::PhotoImage( CommunicationManager::cm->GetObjectID(),
				 filename, x, y, image );
      obj->SetLayer(0);
    }
    catch ( Magick::Exception& exc )
      {
	GUI::MessageWindow::Error( "Cannot build image. %q", exc.what() );
        delete[] filename;
        return NULL;
      }
  }

  LoadAnimationStep( obj, hilf );

  delete[] filename;
  return obj;

} // Load()

//=========================================================================//


char* MLB::PhotoImage::GetARGB(int& size) {
  Magick::Image image( blob );
  Magick::Geometry g = image.size();
  size = 8 + 4 * g.width() * g.height();
  char* buf = new char[size];
  uint32* p = (uint32*)buf;
  p[0] = htonl(g.width());
  p[1] = htonl(g.height());
  image.write(0, 0, g.width(), g.height(), "ARGB", Magick::CharPixel, buf + 8);
  return buf;
}


void MLB::PhotoImage::WriteImage() {
  if (!visible) return;

  std::string file = SessionManager::sm->GetDataPath() + filename;
  FILE* fp = fopen(file.c_str(),"wb");
  if (!fp) {
    GUI::MessageWindow::Error("PhotoImage::WriteImage - Can't open %s\n", file.c_str());
  }
  fwrite(blob.data(), 1, blob.length(), fp);
  fclose(fp);
}
