/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_line.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 3.12.1999
 desc    : Functions to draw and manipulate objects on a tcl-canvas

***********************************************************************************************************/


#include "mlb.h"

#include "mlb_page.h"
#include "mlb_session.h"
#include "mlb_objects.h"
#include "mlb_communication.h"
#include "mlb_adu.h"
#include "mlb_gui.h"


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

MLB::Line::Line(uint32 id) : ColoredObject(id), type(0), capStyle(""), joinStyle(""), arrowStyle(""), arrowShape("") {
  char buf[100];
  sprintf(buf,"%u",id);
  tkref = "Line";
  tkref += buf;
  layer = 1;
  fillColor = "none";
  recorded = false;
}

MLB::Line::Line(int _x, int _y, int _width, int _height, char *_color,
		int _lineWidth, char *_lineStyle, char *_capStyle, char *_joinStyle,
		char *_arrowStyle, char *_arrowShape, uint32 _id)
  :ColoredObject(_x,_y,_width,_height,_color,"none",_lineWidth,_lineStyle,_id)
{
  char buf[100];
  sprintf(buf,"%u",_id);
  tkref = "Line";
  tkref += buf;
  type        = 0;

  capStyle    = _capStyle;
  joinStyle   = _joinStyle;
  arrowStyle  = _arrowStyle;
  arrowShape  = _arrowShape;
  layer = 1;

  if ((width==0) && (height==0)) width = 1;
}

MLB::Line::~Line() {
}

MLB::DrawObject *MLB::Line::Copy()
{
  MLB::Line
    *ret;

  ret = new Line(x,y,width,height,(char *)color.c_str(),lineWidth,(char *)lineStyle.c_str(),(char *)capStyle.c_str(),
		 (char *)joinStyle.c_str(),(char *)arrowStyle.c_str(),(char *)arrowShape.c_str(),CommunicationManager::cm->GetObjectID());

  return (DrawObject *)ret;
}

void MLB::Line::Draw(GUI::Canvas *_canvas)
{
  if (!visible) return;

  // calculate arrowShape
  char as[12];
  unsigned a1, a2, a3;
  sscanf(arrowShape.c_str(), "%u %u %u", &a1, &a2, &a3);
  a1 *= SessionManager::sm->GetZoom();
  a2 *= SessionManager::sm->GetZoom();
  a3 *= SessionManager::sm->GetZoom();
  sprintf(as, "%u %u %u", a1, a2, a3);

  _canvas->DrawLine(tkref.c_str(),
		    SessionManager::sm->GetZoom()*x,
		    SessionManager::sm->GetZoom()*y,
		    SessionManager::sm->GetZoom()*width,
		    SessionManager::sm->GetZoom()*height,
		    color.c_str(),
		    SessionManager::sm->GetZoom()*lineWidth,
		    lineStyle.c_str(),
		    capStyle.c_str(),
		    joinStyle.c_str(),
		    arrowStyle.c_str(),
		    as);

  if (selected) DrawSelection(_canvas);

}

void MLB::Line::ReDraw(GUI::Canvas *canvas)
{
  if (!visible) return;

  canvas->DeleteObject(tkref.c_str());
  Draw(canvas);
}

void MLB::Line::DrawSelection(GUI::Canvas *canvas) {
  if (!visible) return;

  if (selected) DeleteSelection(canvas);

  selectPoints[0] = new Point(x,y);
  selectPoints[1] = new Point(x+width,y+height);
  for (int i=0; i<2; ++i)
    canvas->DrawSelectionRectangle(tkref.c_str(),
				   SessionManager::sm->GetZoom()*selectPoints[i]->x,
				   SessionManager::sm->GetZoom()*selectPoints[i]->y);
  selected = true;
}

void MLB::Line::DeleteSelection(GUI::Canvas *canvas){
  if (!visible || !selected) return;

  selected = false;

  canvas->DeleteSelection(tkref.c_str());
}


bool MLB::Line::IsEditablePoint(int _x, int _y) {
  if (!visible) return false;

  for (int i = 0; i < 2; ++i) {
    if ((_x > selectPoints[i]->x-TOLERANZ) && (_x < selectPoints[i]->x+TOLERANZ)) {
      if ((_y > selectPoints[i]->y-TOLERANZ) && (_y < selectPoints[i]->y+TOLERANZ)) {
	selectCount = i;
	return true;
      }
    }
  }

  return false;
}


void MLB::Line::MovePoint(GUI::Canvas *_canvas, int point, int _x, int _y, bool redraw) {
  recorded = false;

  if (point == 1) {
    width  = x + width - _x;
    height = y + height - _y;

    x = _x;
    y = _y;
  }
  else {
    width  = _x - x;
    height = _y - y;
  }

  if (redraw) ReDraw(_canvas);
}


void MLB::Line::SaveObject(FILE *fp,int level) {
  int
    i;

  if (!visible) return;

  for (i=0; i<level; i++) fprintf(fp,"  ");

  if (arrowStyle != "none") {
    fprintf(fp,"<LINE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\" arrowStyle=\"%s\" arrowShape=\"%s\"%s></LINE>\n",
	    x,y,width,height,color.c_str(),lineWidth,lineStyle.c_str(),arrowStyle.c_str(),arrowShape.c_str(), GetAnimationStepSGMLText().c_str());
  }
  else {
    fprintf(fp,"<LINE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\"%s></LINE>\n",
	    x,y,width,height,color.c_str(),lineWidth,lineStyle.c_str(), GetAnimationStepSGMLText().c_str() );
  }
}
 #include <math.h>


void MLB::Line::ConfigureArrows()
{

    double dx, dy, length, sinTheta, cosTheta, temp;
    double fracHeight;                  /* Line width as fraction of
                                         * arrowhead width. */
    double backup;                      /* Distance to backup end points
                                         * so the line ends in the middle
                                         * of the arrowhead. */
    double vertX, vertY;                /* Position of arrowhead vertex. */
    double shapeA, shapeB, shapeC;      /* Adjusted coordinates (see
                                         * explanation below). */
    double lwidth;
    int i;
  size_t
    pos1,
    pos2;
  std::string
    zahl;


    pos1 = 0;
    pos2 = arrowShape.find(" ",pos1,1);
    if (pos2 > arrowShape.size()) {
      pos2 = arrowShape.size();
    }
    zahl = arrowShape.substr(pos1,(pos2-pos1));
    shapeA = atoi(zahl.c_str());

    pos1 = pos2;
    pos1++;
    pos2 = arrowShape.find(" ",pos1,1);
    if (pos2 > arrowShape.size()) {
      pos2 = arrowShape.size();
    }
    zahl = arrowShape.substr(pos1,(pos2-pos1));
    shapeB = atoi(zahl.c_str());

    pos1 = pos2;
    pos1++;
    pos2 = arrowShape.find(" ",pos1,1);
    if (pos2 > arrowShape.size()) {
      pos2 = arrowShape.size();
    }
    zahl = arrowShape.substr(pos1,(pos2-pos1));
    shapeC = atoi(zahl.c_str());

    lwidth = lineWidth;

    shapeA += 0.001;
    shapeB += 0.001;
    shapeC += lwidth/2.0 + 0.001;

    /*
     * If there's an arrowhead on the first point of the line, compute
     * its polygon and adjust the first point of the line so that the
     * line doesn't stick out past the leading edge of the arrowhead.
     */

    fracHeight = (lwidth/2.0)/shapeC;
    backup     = fracHeight*shapeB + shapeA*(1.0 - fracHeight)/2.0;

    if (arrowStyle != "last") {
      firstArrow[0] = firstArrow[10] = x;
      firstArrow[1] = firstArrow[11] = y;

      dx = -width;
      dy = -height;
      length = hypot(dx, dy);
      if (length == 0) {
	sinTheta = cosTheta = 0.0;
      } else {
	sinTheta = dy/length;
	cosTheta = dx/length;
      }
      vertX = firstArrow[0] - shapeA*cosTheta;
      vertY = firstArrow[1] - shapeA*sinTheta;
      temp = shapeC*sinTheta;
      firstArrow[2] = firstArrow[0] - shapeB*cosTheta + temp;
      firstArrow[8] = firstArrow[2] - 2*temp;
      temp = shapeC*cosTheta;
      firstArrow[3] = firstArrow[1] - shapeB*sinTheta - temp;
      firstArrow[9] = firstArrow[3] + 2*temp;
      firstArrow[4] = firstArrow[2]*fracHeight + vertX*(1.0-fracHeight);
      firstArrow[5] = firstArrow[3]*fracHeight + vertY*(1.0-fracHeight);
      firstArrow[6] = firstArrow[8]*fracHeight + vertX*(1.0-fracHeight);
      firstArrow[7] = firstArrow[9]*fracHeight + vertY*(1.0-fracHeight);

      for (i=0; i<12; i+=2) {
	firstArrow[i] += backup*cosTheta;
	firstArrow[i+1] += backup*sinTheta;

      }
      /*
       * Polygon done.  Now move the first point towards the second so
       * that the corners at the end of the line are inside the
       * arrowhead.
       */
    }
    /*
     * Similar arrowhead calculation for the last point of the line.
     */

    if (arrowStyle != "first") {
      lastArrow[0] = lastArrow[10] = x+width;
      lastArrow[1] = lastArrow[11] = y+height;
      dx = width;
      dy = height;
      length = hypot(dx, dy);
      if (length == 0) {
	sinTheta = cosTheta = 0.0;
      } else {
	sinTheta = dy/length;
	cosTheta = dx/length;
      }
      vertX = lastArrow[0] - shapeA*cosTheta;
      vertY = lastArrow[1] - shapeA*sinTheta;
      temp = shapeC*sinTheta;
      lastArrow[2] = lastArrow[0] - shapeB*cosTheta + temp;
      lastArrow[8] = lastArrow[2] - 2*temp;
      temp = shapeC*cosTheta;
      lastArrow[3] = lastArrow[1] - shapeB*sinTheta - temp;
      lastArrow[9] = lastArrow[3] + 2*temp;
      lastArrow[4] = lastArrow[2]*fracHeight + vertX*(1.0-fracHeight);
      lastArrow[5] = lastArrow[3]*fracHeight + vertY*(1.0-fracHeight);
      lastArrow[6] = lastArrow[8]*fracHeight + vertX*(1.0-fracHeight);
      lastArrow[7] = lastArrow[9]*fracHeight + vertY*(1.0-fracHeight);
      for (i=0; i<12; i+=2) {
	lastArrow[i] += backup*cosTheta;
	lastArrow[i+1] += backup*sinTheta;

      }
    }
}

int MLB::Line::RecordableObjects()
{
  if (width==0 && height==0) visible = false;

  if (!visible) return 0;

  return 1;
}

void MLB::Line::SaveAsPs(FILE *fp, int pageWidth, int pageHeight, bool rotate)
{
  XColor
    *xColor;
  float
    red, green, blue;
  int
    psX1, psX2, psY1, psY2;


  if (!visible) return;
  if ((width==0) && (height==0)) return;

  xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)color.c_str());
  red   = (float)xColor->red/65535;
  green = (float)xColor->green/65535;
  blue  = (float)xColor->blue/65535;

  if (rotate) {
    psX1 = y;
    psX2 = y+height;
    psY1 = x;
    psY2 = x+width;
  }
  else {
    psX1 = x;
    psX2 = x+width;
    psY1 = pageHeight-y;
    psY2 = pageHeight-(y+height);
  }

  fprintf(fp,"\n");
  fprintf(fp,"gsave\n");
  fprintf(fp,"%d setlinewidth 0 setlinejoin 2 setlinecap\n",lineWidth);
  fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
  fprintf(fp,"newpath\n");
  fprintf(fp,"%d %d moveto\n",psX1,psY1);
  fprintf(fp,"%d %d lineto\n",psX2,psY2);
  fprintf(fp,"stroke\n");
  fprintf(fp,"grestore\n");
  fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);

  if (arrowStyle != "none") {
    ConfigureArrows();
    fprintf(fp,"1 setlinewidth 0 setlinejoin 2 setlinecap\n");
    if (arrowStyle!="last") {
      fprintf(fp,"newpath\n");
	if (rotate) {
	  psX1 = (int)firstArrow[1];
	  psY1 = (int)firstArrow[0];
	}
	else {
	  psX1 = (int)firstArrow[0];
	  psY1 = pageHeight-(int)firstArrow[1];
	}
      fprintf(fp,"%d %d moveto\n",psX1,psY1);
      for (int i=2; i<12; i+=2) {
	if (rotate) {
	  psX1 = (int)firstArrow[i+1];
	  psY1 = (int)firstArrow[i];
	}
	else {
	  psX1 = (int)firstArrow[i];
	  psY1 = pageHeight-(int)firstArrow[i+1];
	}
	fprintf(fp,"%d %d lineto\n",psX1,psY1);
      }
      fprintf(fp,"closepath fill\n");
    }
    if (arrowStyle!="first") {
      fprintf(fp,"newpath\n");
	if (rotate) {
	  psX1 = (int)lastArrow[1];
	  psY1 = (int)lastArrow[0];
	}
	else {
	  psX1 = (int)lastArrow[0];
	  psY1 = pageHeight-(int)lastArrow[1];
	}
      fprintf(fp,"%d %d moveto\n",psX1,psY1);
      for (int i=2; i<12; i+=2) {
	if (rotate) {
	  psX1 = (int)lastArrow[i+1];
	  psY1 = (int)lastArrow[i];
	}
	else {
	  psX1 = (int)lastArrow[i];
	  psY1 = pageHeight-(int)lastArrow[i+1];
	}
	fprintf(fp,"%d %d lineto\n",psX1,psY1);
      }
      fprintf(fp,"closepath fill\n");
    }
  }

  fprintf(fp,"stroke\n");
  fprintf(fp,"grestore\n");
}

void MLB::Line::SaveAsAOF(FILE *fp)
{
  int
    zx, zy,
    zw, zh,
    dir, ori;

   GUI::MessageWindow::Debug("Line::SaveAsAOF\n");

  if (!visible || recorded) return;

  recorded = true;

  zx = x;
  zy = y;
  zw = width;
  zh = height;

  dir    = 0;
  ori = (width*height>0);
  if (width < 0) {
    zx = x + width;
    zw = -width;
    dir = 1;
  }
  if (width == 0 && height > 0) dir=1;
  if (height < 0) {
    zy = y+height;
    zh = -height;
  }

  recID = MLB::SessionManager::sm->recCount;
  MLB::SessionManager::sm->recCount++;

  if (arrowStyle == "none") {
    fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=0 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(),layer);
  }
  else if (arrowStyle == "first") {
    if (dir) {
      fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=1 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(), layer);
    }
    else  {
      fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=2 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(), layer);
    }
  }
  else if (arrowStyle == "last") {
    if (dir) {
      fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=2 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(), layer);
    }
    else {
      fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=1 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(), layer);
    }
  }
  else if (arrowStyle == "both") {
    fprintf(fp,"<LINE x=%d y=%d width=%d height=%d ori=%d dir=%d fcolor=%d linewidth=%d linestyle=%d arrowstyle=3 layer=%d></LINE>\n",zx,zy,zw,zh,ori,dir,CalculateColor(color),lineWidth,CalculateLineStyle(), layer);
  }
}

void MLB::Line::ChangeLineColor(GUI::Canvas *canvas,char *_color)
{
  if (!visible) return;

  recorded = false;
  color    = _color;

  canvas->ChangeFillColor(tkref.c_str(),color.c_str());
}

void MLB::Line::ChangeCapStyle(GUI::Canvas *canvas, char *_style)
{
  if (!visible) return;

  recorded = false;
  capStyle = _style;

  canvas->ChangeCapStyle(tkref.c_str(),capStyle.c_str());
}

void MLB::Line::ChangeJoinStyle(GUI::Canvas *canvas, char *_style)
{
  if (!visible) return;

  recorded  = false;
  joinStyle = _style;

  canvas->ChangeJoinStyle(tkref.c_str(),joinStyle.c_str());
}

void MLB::Line::ChangeArrowStyle(GUI::Canvas *canvas, char *_style)
{
  if (!visible) return;

  recorded   = false;
  arrowStyle = _style;

  canvas->ChangeArrowStyle(tkref.c_str(),arrowStyle.c_str());
}

void MLB::Line::ChangeArrowShape(GUI::Canvas *canvas, char *_shape)
{
  if (!visible) return;

  recorded   = false;
  arrowShape = _shape;

  // calculate arrowShape
  char as[12];
  unsigned a1, a2, a3;
  sscanf(arrowShape.c_str(), "%u %u %u", &a1, &a2, &a3);
  a1 *= SessionManager::sm->GetZoom();
  a2 *= SessionManager::sm->GetZoom();
  a3 *= SessionManager::sm->GetZoom();
  sprintf(as, "%u %u %u", a1, a2, a3);
  canvas->ChangeArrowShape(tkref.c_str(),as);
}

bool MLB::Line::Overlap(GUI::Canvas *canvas, int _x, int _y)
{
  int
    x1, y1,
    x2, y2;

  if (!visible) return false;

  GetBBox(&x1,&y1,&x2,&y2);
  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return false;

  return canvas->Overlap(tkref.c_str(),SessionManager::sm->GetZoom()*_x,SessionManager::sm->GetZoom()*_y);
}

void MLB::Line::AdjustGui()
{

  char buf[20];

  GUI::Attributes::SetOutlineColor((char *)color.c_str());
  sprintf(buf,"%d",lineWidth);
  GUI::Attributes::SetLineWidth(buf);
  GUI::Attributes::SetLineStyle((char *)lineStyle.c_str());
  GUI::Attributes::SetCapStyle((char *)capStyle.c_str());
  GUI::Attributes::SetJoinStyle((char *)joinStyle.c_str());
  GUI::Attributes::SetArrowStyle((char *)arrowStyle.c_str());
  GUI::Attributes::SetArrowShape((char *)arrowShape.c_str());

  GUI::Attributes::SetTool("Line");
}


void MLB::Line::FlushCapStyle(char* buf, Style& capstyle) {
  uint8*  p8  = (uint8*)buf;
  if (capstyle.compare("butt") == 0)
    p8[0] = p8[0] | (0 << 6);
  else if (capstyle.compare("round") == 0)
    p8[0] = p8[0] | (1 << 6);
  else if (capstyle.compare("projecting") == 0)
    p8[0] = p8[0] | (2 << 6);
}

MLB::Style MLB::Line::ParseCapStyle(char* buf) {
  uint8*  p8  = (uint8*)buf;
  switch ((p8[0] & MLB_MASK_CST) >> 6) {
  case 0: return "butt";
  case 1: return "round";
  case 2: return "projecting";
  default: return "";
  }
}

void MLB::Line::FlushJoinStyle(char* buf, Style& joinstyle) {
  uint8*  p8  = (uint8*)buf;
  if (joinstyle.compare("bevel") == 0)
    p8[0] = p8[0] | (0 << 4);
  else if (joinstyle.compare("miter") == 0)
    p8[0] = p8[0] | (1 << 4);
  else if (joinstyle.compare("round") == 0)
    p8[0] = p8[0] | (2 << 4);
	   }

MLB::Style MLB::Line::ParseJoinStyle(char* buf) {
  uint8*  p8  = (uint8*)buf;
  switch ((p8[0] & MLB_MASK_JST) >> 4) {
  case 0: return "bevel";
  case 1: return "miter";
  case 2: return "round";
  default: return "";
  }
}

void MLB::Line::FlushArrowStyle(char* buf, Style& arrowstyle) {
  uint8*  p8  = (uint8*)buf;
  if (arrowstyle.compare("none") == 0)
      p8[0] = p8[0] | (0 << 2);
  else if (arrowstyle.compare("first") == 0)
    p8[0] = p8[0] | (1 << 2);
  else if (arrowstyle.compare("last") == 0)
    p8[0] = p8[0] | (2 << 2);
  else if (arrowstyle.compare("both") == 0)
    p8[0] = p8[0] | (3 << 2);
}

MLB::Style MLB::Line::ParseArrowStyle(char* buf) {
  uint8*  p8  = (uint8*)buf;
  switch ((p8[0] & MLB_MASK_AST) >> 2) {
  case 0: return "none";
  case 1: return "first";
  case 2: return "last";
  case 3: return "both";
  default: return "";
  }
}

void MLB::Line::FlushArrowShape(char* buf, Style& arrowshape) {
  uint8*  p8  = (uint8*)buf;
  unsigned a1, a2, a3;
  sscanf(arrowshape.c_str(), "%u %u %u", &a1, &a2, &a3);
  p8[0] = a1;
  p8[1] = a2;
  p8[2] = a3;
}

MLB::Style MLB::Line::ParseArrowShape(char* buf) {
  uint8*  p8  = (uint8*)buf;
  char tmp[12];
  sprintf(tmp, "%u %u %u", p8[0], p8[1], p8[2]);
  return string(tmp);
}


RefCount<RTPI::RtpiAdu> MLB::Line::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;
  RefCount<Buffer> buf = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::LINE, priority, buf, timestamp);
}


int MLB::Line::FlushState(char* buf, uint32 order, bool visible, int x, int y, int width, int height,
			   Color linecolor, int linewidth, Style linestyle,
			   Style capstyle, Style joinstyle, Style arrowstyle, Style arrowshape) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::LINE;

  Line::FlushCapStyle(buf + 3, capstyle);
  Line::FlushJoinStyle(buf + 3, joinstyle);
  Line::FlushArrowStyle(buf + 3, arrowstyle);

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  p16[6] = htons(x+width);
  p16[7] = htons(y+height);

  ColoredObject::FlushColor(buf + 16, linecolor);
  ColoredObject::FlushLineStyle(buf + 19, linestyle);

  Line::FlushArrowShape(buf + 20, arrowshape);
  p8[23] = linewidth;

  return 24;
}


int MLB::Line::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);

  p8[1] = CommunicationManager::LINE;

  Line::FlushCapStyle(buf + 3, capStyle);
  Line::FlushJoinStyle(buf + 3, joinStyle);
  Line::FlushArrowStyle(buf + 3, arrowStyle);

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  p16[6] = htons(x+width);
  p16[7] = htons(y+height);

  ColoredObject::FlushColor(buf + 16, color);
  ColoredObject::FlushLineStyle(buf + 19, lineStyle);

  Line::FlushArrowShape(buf + 20, arrowShape);
  p8[23] = lineWidth;

  return 24;
}


int MLB::Line::ParseState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("Line::ParseCreate - wrong version of mlb protocol");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    SessionManager::sm->message->Error("Line::ParseCreate - wrong operation type");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;

  if (p8[1] != CommunicationManager::LINE) {
    SessionManager::sm->message->Error("Line::ParseCreate - wrong object type %u", (int)p8[1]);
    return -1;
  }

  capStyle = Line::ParseCapStyle(buf + 3);
  joinStyle = Line::ParseJoinStyle(buf + 3);
  arrowStyle = Line::ParseArrowStyle(buf + 3);

  order = ntohl(p32[1]);

  x = ntohs(p16[4]);
  y = ntohs(p16[5]);

  width  = htons(p16[6]) - x;
  height = htons(p16[7]) - y;

  color     = ColoredObject::ParseColor(buf + 16);
  lineStyle = ColoredObject::ParseLineStyle(buf + 19);

  arrowShape = Line::ParseArrowShape(buf + 20);
  lineWidth = p8[23];

  return 24;
}


int MLB::Line::FlushChangeVisibility(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}


int MLB::Line::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

int MLB::Line::FlushMovePoint(char* buf, int point, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[3] = 0;

  if (point == 1) {
    p8[2] = CommunicationManager::MOVE_POINT_1;
  } else {
    p8[2] = CommunicationManager::MOVE_POINT_2;
  }

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}


int MLB::Line::FlushChangeLinewidth(char* buf, int linewidth) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_LINEWIDTH;

  p8[7] = linewidth;

  return 8;
}


int MLB::Line::FlushChangeLinestyle(char* buf, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_LINESTYLE;

  ColoredObject::FlushLineStyle(buf + 7, linestyle);

  return 8;
}


int MLB::Line::FlushChangeLinecolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_LINECOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Line::FlushChangeCapstyle(char* buf, Style capstyle) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_CAPSTYLE;

  p8[3] = 0;
  Line::FlushCapStyle(buf + 3, capstyle);

  return 4;
}


int MLB::Line::FlushChangeJoinstyle(char* buf, Style joinstyle ) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_JOINSTYLE;

  p8[3] = 0;
  Line::FlushJoinStyle(buf + 3, joinstyle);

  return 4;
}


int MLB::Line::FlushChangeArrowstyle(char* buf, Style arrowstyle, Style arrowshape) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::CHG_ARROWSTYLE;

  p8[3] = 0;
  Line::FlushArrowStyle(buf + 3, arrowstyle);

  p32[1] = 0;
  Line::FlushArrowShape(buf + 4, arrowshape);

  return 8;
}


int MLB::Line::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Line::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::LINE;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Line::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::LINE;

  return 4;
}


MLB::DrawObject *MLB::Line::Load(SGMLElement *hilf)
{
  Line
    *obj;
  char
    *tmp;
  int
    x, y,
    width, height,
    lineWidth;
  char
    *color,
    *lineStyle,
    *capStyle,
    *joinStyle,
    *arrowStyle,
    *arrowShape;

  tmp = hilf->GetAttribute("x");
  if (tmp) x = atoi(tmp);
  else     x = 0;
  tmp = hilf->GetAttribute("y");
  if (tmp) y = atoi(tmp);
  else     y = 0;
  tmp = hilf->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = 0;
  tmp = hilf->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = 0;
  tmp = hilf->GetAttribute("color");
  if (tmp) color = strdup(tmp);
  else     color = strdup("#ff0000");
  tmp = hilf->GetAttribute("lineWidth");
  if (tmp) lineWidth = atoi(tmp);
  else     lineWidth = 2;
  tmp = hilf->GetAttribute("lineStyle");
  if (tmp) lineStyle = strdup(tmp);
  else     lineStyle = strdup(" ");
  tmp = hilf->GetAttribute("capStyle");
  if (tmp) capStyle = strdup(tmp);
  else     capStyle = strdup("butt");
  tmp = hilf->GetAttribute("joinStyle");
  if (tmp) joinStyle = strdup(tmp);
  else     joinStyle = strdup("bevel");
  tmp = hilf->GetAttribute("arrowStyle");
  if (tmp) arrowStyle = strdup(tmp);
  else     arrowStyle = strdup("none");
  tmp = hilf->GetAttribute("arrowShape");
  if (tmp) arrowShape = strdup(tmp);
  else     arrowShape = strdup("8 10 4");
  obj = new Line(x,y,width,height,color,lineWidth,lineStyle,
		 capStyle,joinStyle,arrowStyle,arrowShape,
		 CommunicationManager::cm->GetObjectID());
  obj->SetLayer(0);

  LoadAnimationStep( obj, hilf );

  delete[] color;
  delete[] lineStyle;
  delete[] capStyle;
  delete[] joinStyle;
  delete[] arrowStyle;
  delete[] arrowShape;

  return obj;
}
