/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_objects.cpp
 authors : Juergen Vogel, Gabriela Maass
 date    : 3.12.1999
 desc    : Functions to draw and manipulate objects on a tcl-canvas

 **********************************************************************************************************/


#include "mlb.h"
#include "mlb_objects.h"
#include "mlb_page.h"
#include "mlb_session.h"


MLB::DrawObject::DrawObject(uint32 id) 
  : GenericObject(id), x(0), y(0), width(0), height(0), visible(false), selected(false), annotation(false), group(NULL) {}

/**********************************************************************
 Name:      MLB::DrawObject::DrawObject(int _x, int _y, int _width, int _height, int _id)
 Param:     _x, _y:          Coordinates (in Pixel)
            _width, _height: Dimensions (in Pixel)
	    _id:         Identification-number of the draw-object
 Desc:      initializes variables
 Return:  
**********************************************************************/
MLB::DrawObject::DrawObject(int _x, int _y, int _width, int _height, uint32 _id) 
  : GenericObject(_id), annotation(false) {
  x        = _x;
  y        = _y;
  width    = _width;
  height   = _height;
  visible  = true;
  selected = false;  
  recorded = false;

  group = NULL;
}


MLB::DrawObject::~DrawObject() {
}


void MLB::DrawObject::DrawSelection(GUI::Canvas *_canvas) {
  if (!visible) return;
  
  if (selected) DeleteSelection(_canvas);

  selectPoints[0] = new Point(x,y);
  selectPoints[1] = new Point(x+(width/2),y);
  selectPoints[2] = new Point(x+width,y);
  selectPoints[3] = new Point(x+width,y+(height/2));
  selectPoints[4] = new Point(x+width,y+height);
  selectPoints[5] = new Point(x+(width/2),y+height);
  selectPoints[6] = new Point(x,y+height);
  selectPoints[7] = new Point(x,y+(height/2));

  for (int i= 0; i < 8; ++i) _canvas->DrawSelectionRectangle(tkref.c_str(),
							     SessionManager::sm->GetZoom()*selectPoints[i]->x,
							     SessionManager::sm->GetZoom()*selectPoints[i]->y);
  selected = true;
}


void MLB::DrawObject::DeleteSelection(GUI::Canvas *_canvas) {
  if (!visible || !selected) return;
  
  selected = false;

  _canvas->DeleteSelection(tkref.c_str());
}


bool MLB::DrawObject::IsEditablePoint(int _x, int _y) {
  if (!visible) return false;
  
  for (int i = 0; i < 8; ++i) {
    if ((_x > selectPoints[i]->x-TOLERANZ) && (_x < selectPoints[i]->x+TOLERANZ)) {
      if ((_y > selectPoints[i]->y-TOLERANZ) && (_y < selectPoints[i]->y+TOLERANZ)) {
	selectCount = i;
	return true;
      }
    }
  }

  return false;
}


void MLB::DrawObject::ComputeCoords(int _x, int _y, int& _newX, int& _newY, int& _newWidth, int& _newHeight) {
  if (!visible) return;

  int x1 = x;
  int x2 = x+width;
  int y1 = y;
  int y2 = y+height;

  switch (selectCount) {
  case 0:
  case 7:
  case 6:
    if (_x > x2) {
      x1 = x2;
      x2 = _x;
      selectCount = (10-selectCount)%8;
    }
    else 
      x1 = _x;
    break;
  case 2:
  case 3:
  case 4:
    if (_x < x1) {
      x2 = x1;
      x1 = _x;
      selectCount = (10-selectCount)%8;
    }
    else 
      x2 = _x;
    break;
  }

  switch (selectCount) {
  case 0:
  case 1:
  case 2:
    if (_y > y2) {
      y1 = y2;
      y2 = _y;
      selectCount = 6-selectCount;
    }
    else
      y1 = _y;
    break;
  case 6:
  case 5:
  case 4:
    if (_y < y1) {
      y2 = y1;
      y1 = _y;
      selectCount = 6-selectCount;
    }
    else
      y2 = _y;
    break;
  }

  _newX = x = x1;
  _newY = y = y1;
  _newWidth  = width  = x2-x1;
  _newHeight = height = y2-y1;
}


MLB::DrawObject* MLB::DrawObject::Select(bool select) {
  if (group) {
    selected = false;
    return group->Select(select);
  }
  else {
    selected = select;
    return this;
  }
}


int MLB::DrawObject::FlushMoveS(char* buf, int type, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
 
  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = type;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}


/**********************************************************************
 Name:      MLB::DrawObject::Hide(GUI::Canvas *canvas)
 Param:     widget:          name of the tcl-canvas
 Desc:      delete the object from the tcl-canvas; automatically redrawn
 Return:  
**********************************************************************/
void MLB::DrawObject::Hide(GUI::Canvas *_canvas) {
  if (!visible) return;
  
  if (group) {
    group->Hide(_canvas);
  }
  else {
    if (selected) DeleteSelection(_canvas);
    _canvas->DeleteObject(tkref.c_str());
    visible = false;
  }
}

/**********************************************************************
 Name:      MLB::DrawObject::ModifyCoords(GUI::Canvas *canvas, int _x, int _y, int _width, int _height)
 Param:     canvas:          GUI-Canvas
	    _x, _y:          new coordinates
	    _width, _height: new dimensions
 Desc:      change the coordinates of the object; automatically redrawn
 Return:  
**********************************************************************/
void MLB::DrawObject::ModifyCoords(GUI::Canvas *_canvas, int _x, int _y, int _width, int _height)
{
  if (!visible) return;
  
  x      = _x;
  y      = _y;
  width  = _width;
  height = _height;

  _canvas->ModifyCoords(tkref.c_str(),
			SessionManager::sm->GetZoom()*x,
			SessionManager::sm->GetZoom()*y,
			SessionManager::sm->GetZoom()*width,
			SessionManager::sm->GetZoom()*height);

  if (selected) DrawSelection(_canvas);
}


/**********************************************************************
 Name:      MLB::DrawObject::ModifyCoords(int _x, int _y, int _width, int _height)
 Param:   
	    _x, _y:          new coordinates
	    _width, _height: new dimensions
 Desc:      change the coordinates of the object; automatically redrawn
 Return:  
**********************************************************************/
void MLB::DrawObject::ModifyCoords(int _x, int _y, int _width, int _height)
{
  if (!visible) return;
  
  x      = _x;
  y      = _y;
  width  = _width;
  height = _height;
}

/**********************************************************************
 Name:      MLB::DrawObject::Resize(GUI::Canvas *canvas, int _width, int _height)
 Param:     canvas:          GUI-Canvas
	    _width, _height: new dimensions
 Desc:      resize the object; automatically redrawn
 Return:  
**********************************************************************/
void MLB::DrawObject::Resize(GUI::Canvas *_canvas, int _width, int _height)
{
  if (!visible) return;
  
  width  = _width;
  height = _height;

  _canvas->ModifyCoords(tkref.c_str(),
			SessionManager::sm->GetZoom()*x,
			SessionManager::sm->GetZoom()*y,
			SessionManager::sm->GetZoom()*width,
			SessionManager::sm->GetZoom()*height);
}

/**********************************************************************
 Name:      MLB::DrawObject::Move(GUI::Canvas *canvas, int diffX, int diffY)
 Param:     canvas:          GUI-Canvas
	    diffX, diffY:    the x and y difference
 Desc:      move the object; automatically redrawn
 Return:  
**********************************************************************/
void MLB::DrawObject::Move(GUI::Canvas *_canvas, int _diffX, int _diffY, bool redraw)
{
  if (!visible) return;
  
  recorded = false;

  x += _diffX;
  y += _diffY;

  if (redraw) {
    _canvas->MoveObject(tkref.c_str(),SessionManager::sm->GetZoom()*_diffX,SessionManager::sm->GetZoom()*_diffY);
    if (selected) DrawSelection(_canvas);
  }
}


/**********************************************************************
 Name:      MLB::DrawObject::SelectObject(int _x, int _y)
 Param:     length: length of selectList
            selectList: list of selected objects
 Desc:      scale the object from 0,0 with the factor "scaleFactor"; automatically redrawn
 Return:    -1 if tkRef of DrawObject is not in selectList, else 1
**********************************************************************/
int MLB::DrawObject::SelectObject(int _x, int _y)
{
  int 
    x1, y1,
    x2, y2;

  if (!visible) return 0;
  GetBBox(&x1,&y1,&x2,&y2);
  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return 0;
  
  return 1;
}
#include "mlb_session.h"

void MLB::DrawObject::GetBBox(int *minX, int *minY, int *maxX, int *maxY)
{  
  if ((x+width) < x) {
    *minX = x+width-3;
    *maxX = x+3;
  }
  else {
    *minX = x-3;
    *maxX = x+width+3;
  }
  if ((y+height) < y) {
    *minY = y+height-3;
    *maxY = y+3;
  }
  else {
    *minY = y-3;
    *maxY = y+height+3;
  }
}

bool MLB::DrawObject::IsInRegion(int _x1, int _y1, int _x2, int _y2)
{
  int 
    x1, y1,
    x2, y2;

  if (!visible) return false;

  GetBBox(&x1,&y1,&x2,&y2);
  if ((x1 < _x1) || (y1 < _y1) || (x2 > (_x2)) || (y2 > (_y2))) return false;

  return true;
}

bool MLB::DrawObject::Overlap(GUI::Canvas *_canvas,int _x, int _y) 
{
  int 
    x1, y1,
    x2, y2;

  if (!visible) return false;

  GetBBox(&x1,&y1,&x2,&y2);
  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return false;

  return _canvas->Overlap(tkref.c_str(),
			  SessionManager::sm->GetZoom()*_x,
			  SessionManager::sm->GetZoom()*_y);
}


bool MLB::DrawObject::Find(std::list<DrawObject*>& objlist, DrawObject* obj) {
  for (std::list<DrawObject*>::iterator i = objlist.begin(); i != objlist.end(); ++i) {
    if (*i == obj) return true;
  } 

  return false;
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

MLB::ColoredObject::ColoredObject(uint32 id)
  : DrawObject(id), color(""), fillColor(""), lineWidth(0), lineStyle("") {}


MLB::ColoredObject::ColoredObject(int _x, int _y, int _width, int _height, char *_color, 
				char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id)
  : DrawObject(_x,_y,_width,_height,_id) 
{
  color     = _color;
  fillColor = _fillColor;
  lineWidth = _lineWidth;
  lineStyle = _lineStyle;
}

MLB::ColoredObject::~ColoredObject()
{
  if (!color.empty())     color.erase();
  if (!fillColor.empty()) fillColor.erase();
  if (!lineStyle.empty()) lineStyle.erase();
}

bool MLB::ColoredObject::IsFilled() 
{
  if (fillColor == "none") 
    return false;
  else 
    return true;
}

void MLB::ColoredObject::ChangeFillColor(GUI::Canvas *_canvas,char *_fillColor)
{
  if (!visible || (fillColor == "none")) return;

  recorded = false;

  fillColor = _fillColor;
  _canvas->ChangeFillColor(tkref.c_str(),fillColor.c_str());
}

void MLB::ColoredObject::ChangeLineColor(GUI::Canvas *_canvas,char *_color)
{
  if (!visible) return;
  
  recorded = false;

  color = _color;
  _canvas->ChangeLineColor(tkref.c_str(),color.c_str());
}

void MLB::ColoredObject::ChangeLineWidth(GUI::Canvas *_canvas,int _lineWidth)
{
  if (!visible) return;
  
  recorded = false;

  lineWidth = _lineWidth;
  _canvas->ChangeLineWidth(tkref.c_str(),SessionManager::sm->GetZoom()*lineWidth);
}

void MLB::ColoredObject::ChangeLineStyle(GUI::Canvas *_canvas,char *_lineStyle)
{
  if (!visible) return;
  
  recorded = false;

  lineStyle = _lineStyle;

  if (lineStyle == "none") lineStyle = "";
  _canvas->ChangeLineStyle(tkref.c_str(),lineStyle.c_str());
}


int MLB::ColoredObject::FlushColor(char* buf, MLB::Color color) {
  uint8* p8 = (uint8*)buf;
  string tmp;
  int i;
  // write r
  tmp.assign(color, 1, 2);
  sscanf(tmp.c_str(), "%x", &i);
  p8[0] = (uint8)i;
  // write g
  tmp.assign(color, 3, 2);
  sscanf(tmp.c_str(), "%x", &i);
  p8[1] = (uint8)i;
  // write b
  tmp.assign(color, 5, 2);
  sscanf(tmp.c_str(), "%x", &i);
  p8[2] = (uint8)i;
  
  return 3; 
}


MLB::Color MLB::ColoredObject::ParseColor(char* buf) {
  uint8* p8 = (uint8*)buf;
  char tmp[6];
  string color = "#";
  sprintf(tmp, "%02x%02x%02x", p8[0], p8[1], p8[2]);
  color.append(tmp);
  return color;
}


int MLB::ColoredObject::FlushLineStyle(char* buf, MLB::Style linestyle) {
  uint8* p8 = (uint8*)buf; 
  
  if ((linestyle.compare("") == 0) || (linestyle.compare("none") == 0))
    p8[0] = 0;
  else if (linestyle.compare(".") == 0)
    p8[0] = 1;
  else if (linestyle.compare(",") == 0)
    p8[0] = 2;
  else if (linestyle.compare("-") == 0)
    p8[0] = 3;
  else if (linestyle.compare("_") == 0)
    p8[0] = 4;
  else if (linestyle.compare(".-") == 0)
    p8[0] = 5;
  else if (linestyle.compare("..-") == 0)
    p8[0] = 6;
  else return -1;

  return 1;
}


MLB::Style MLB::ColoredObject::ParseLineStyle(char* buf) {
  uint8* p8 = (uint8*)buf; 
  Style linestyle;

  switch (p8[0]) {
  case 0:
    linestyle = "";
    break;
  case 1:
    linestyle = ".";
    break;
  case 2:
    linestyle = ",";
    break;
  case 3:
    linestyle = "-";
    break;
  case 4:
    linestyle = "_";
    break;
  case 5:
    linestyle = ".-";
    break;
  case 6:
    linestyle = "..-";
    break;
  default:
    return "";
  }

  return linestyle;
}


int MLB::ColoredObject::CalculateLineStyle() {
  if (lineStyle == "none") {
    return 0;
  }
  else if (lineStyle == ".") {
    return 1;
  }
  else if (lineStyle == ",") {
    return 2;
  }
  else if (lineStyle == "-") {
    return 3;
  }
  else if (lineStyle == "_") {
    return 5;
  }
  else if (lineStyle == ".-") {
    return 4;
  }
  else if (lineStyle == "..-") {
    return 4;
  }
}

