/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_objects.h
 authors : Juergen Vogel, Gabriela Maass
 date    : 3.12.1999
 desc    :

 **********************************************************************************************************/

#ifndef __MLB_OBJECT__
#define __MLB_OBJECT__

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif

#include <stdlib.h>
#include <stdio.h>
#include <list>
#include <vector>

#ifndef WIN32
#  include <unistd.h>
#else
#include <windows.h>
#endif

#include "tk.h"
#include "tcl.h"

#ifdef _ImageMagick
  namespace MagickLib {
    #undef PACKAGE
    #undef VERSION
    #undef class
  }
  #include <Magick++.h>
#endif


#include "mlb_communication.h"
#include "mlb_types.h"
#include "mlb_generic.h"
#include "mlb_font.h"
#include "mlb_char.h"
#include "mlb_gui.h"
#include "sgml.h"


namespace MLB {

  using namespace RTPI;

  class Page;
  class Group;

  class DrawObject : public GenericObject {
  public:
    typedef std::list<DrawObject*>  object_list;
    typedef std::list<Group*>       group_list;
    typedef object_list::iterator   object_iterator;

  protected:
    RefCount<Point> selectPoints[8];
    int selectCount;
    int x, y;              /* x- and y-Coordinate (Pixel) */
    int width, height;     /* width and height (Pixel) */
    int recID;
    int layer; /* AOF-Recording Layer=0 orginal Dokument, Layer=1 lokale Annotation, Layer=2 private Annotation */
    //string      animationGroupName;

    bool visible;            /* true=visible, false=invisible */
    bool selected;           /* true=selected, false=not selected */
    bool recorded;           /* true=recorded, false=not recorded */
    bool annotation;

    Group* group;

  public:
    enum type {
      TOLERANZ = 3
    };


    DrawObject(uint32 id);
    DrawObject(int _x, int _y, int _width, int _height, uint32 id);
    ~DrawObject();

    virtual DrawObject* Copy() {return NULL;}    /* returns a copy of the object */

    virtual int  GetX() {return x;}
    virtual void SetX(int _x) {x = _x;}
    virtual int  GetY() {return y;}
    virtual void SetY(int _y) {y = _y;}
    virtual int  GetWidth() {return width;}
    virtual void SetWidth(int _width) {width = _width;}
    virtual int  GetHeight() {return height;}
    virtual void SetHeight(int _height) {height = _height;}

    virtual void GetBBox(int *minX, int *minY, int *maxX, int *maxY);

    void SetAnnotation(bool ann) {annotation = ann;}
    bool Annotation() {return annotation;}

    virtual void SetVisible() {visible = true;}
    virtual void SetInvisible() {visible = false;}
    bool IsVisible() {return visible;}

    virtual void Draw(GUI::Canvas *canvas) {};     /* Draw object on the canvas */
    virtual void ReDraw(GUI::Canvas *canvas) {}    /* Delete and Draw */
    virtual void Hide(GUI::Canvas *canvas);        /* hidden the object (delete from canvas) */

    virtual int  SelectObject(int _x, int _y);    /* If objects tkref is in list, select object return 1 */
    virtual void DrawSelection(GUI::Canvas *canvas);      /* Draw markers to show that the object was selected */
    virtual void DeleteSelection(GUI::Canvas *canvas);    /* Delete markers */
    virtual int  GetSelectCount() {return selectCount;}
    virtual int  GetSelectedPoint() {return selectCount;}
    virtual bool IsSelected() {return selected;}
    virtual MLB::DrawObject* Select(bool select = true);

    virtual void SetGroup(Group* _group) {group = _group;}
    virtual Group* GetGroup() {return group;}

    virtual bool IsEditablePoint(int _x, int _y);
    virtual DrawObject *PointIsInText(int _x, int _y) {return NULL;};
    virtual bool IsInRegion(int _x1, int _y1, int _x2, int _y2);
    virtual bool Overlap(GUI::Canvas *canvas,int _x, int _y);

    virtual void ComputeCoords(int _x, int _y, int& _newX, int& _newY, int& _newWidth, int& _newHeight);
    virtual void ModifyCoords(GUI::Canvas *canvas,int _x, int _y, int _width, int _height);    /* Change the coordinates of the object */
    virtual void ModifyCoords(int _x, int _y, int _width, int _height);    /* Change the coordinates of the object */
    virtual void Resize(GUI::Canvas *canvas, int _width, int _height);    /* Change the size of the object */
    virtual void Move(GUI::Canvas *canvas, int diffX, int diffY, bool redraw = true);    /* Move the object */
    virtual void Change(Page* page,int argc, char *argv[], int ttype) {};

    virtual void SaveObject(FILE *fp, int level=0) {}    /* write object to the Document */
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false) {}      /* save object as postscript */
    virtual void SaveAsAOF(FILE *fp) = 0;    /* save object to an aof-document */
    virtual int  RecordableObjects() = 0;
    virtual int  GetRecordID() {return recID;}
    virtual void SetRecordID (int ID) {recID=ID;}
    virtual int  GetLayer() {return layer;}  /* get AOF-layer of an object */
    virtual void SetLayer (int i) {layer=i;} /* set AOF-layer to an object */
    virtual void SetUnrecorded() {recorded = false;}
    virtual void SetRecorded() {recorded = true;}

    virtual int FlushChangeVisibility(char* buf, bool vis) {return 0;}
    virtual int GetChangeVisibilitySize() {return 4;}
    virtual int FlushMove(char* buf, int x, int y) {return 0;}
    virtual int GetMoveSize() {return 8;}
    static  int FlushMoveS(char* buf, int type, int x, int y);
    static  int GetMoveSizeS() {return 8;}
    virtual int FlushChangeSize(char* buf, int x, int y, int width, int height) {return 0;}
    virtual int GetChangeSizeSize() {return 12;}
    virtual int FlushRaise(char* buf, uint32 order) {return 0;}
    virtual int GetRaiseSize() {return 8;}
    virtual int FlushLower(char* buf, uint32 order) {return 0;}
    virtual int GetLowerSize() {return 8;}

    virtual int FlushDelete(char* buf) {return 0;}
    
    virtual void AdjustGui() {};

    //inline void   SetAnimationStepName( const string& name ) { animationGroupName = name; }
    //inline string GetAnimationStepName() { return animationGroupName; }
    inline string GetAnimationStepSGMLText() { return " name=\""+string(GetTkRef())+"\""; }// { return ( animationGroupName=="") ? "" : " animation=\""+animationGroupName+"\""+" onAnimate=\""+animationGroupName+"\""; }
    inline static void LoadAnimationStep( DrawObject* obj, SGMLElement* hilf ) { /* char* tmp = hilf->GetAttribute( "animation" ); if( tmp ) obj->animationGroupName = string(tmp);*/ }

    static bool Find(std::list<DrawObject*>& objlist, DrawObject* obj);
  };

/***************************************************************************************************/

  class ColoredObject : public DrawObject {
  protected:
    Color color;
    Color fillColor;
    Style lineStyle;
    int   lineWidth;

  public:
    ColoredObject(uint32 id);
    ColoredObject(int _x, int _y, int _width, int _height, char *color,
		  char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id);
    ~ColoredObject();

    Color GetLineColor() {return color;}
    void  SetLineColor(Color _color) {color = _color;}
    Color GetFillColor() {return fillColor;}
    void  SetFillColor(Color _color) {fillColor = _color;}
    Style GetLineStyle() {return lineStyle;}
    void  SetLineStyle(Style _style) {lineStyle = _style;}
    int   GetLineWidth() {return lineWidth;}
    void  SetLineWidth(int _width) {lineWidth = _width;}

    bool IsFilled();

    virtual void ChangeLineColor(GUI::Canvas *canvas,char *_color);
    virtual void ChangeFillColor(GUI::Canvas *canvas,char *_fillColor);
    virtual void ChangeLineWidth(GUI::Canvas *canvas,int  _lineWidth);
    virtual void ChangeLineStyle(GUI::Canvas *canvas,char *_lineStyle);

    static int   FlushColor(char* buf, Color color);
    static Color ParseColor(char* buf);

    static int   FlushLineStyle(char* buf, Style linestyle);
    static Style ParseLineStyle(char* buf);

    virtual int FlushChangeLinewidth(char* buf, int linewidth) {return 0;}
    virtual int GetChangeLinewidthSize() {return 8;}
    virtual int FlushChangeLinestyle(char* buf, Style linestyle) {return 0;}
    virtual int GetChangeLinestyleSize() {return 8;}
    virtual int FlushChangeLinecolor(char* buf, Color color) {return 0;}
    virtual int GetChangeLinecolorSize() {return 8;}
    virtual int FlushChangeFillcolor(char* buf, Color color) {return 0;}
    virtual int GetChangeFillcolorSize() {return 8;}

    int CalculateLineStyle();
  };

/***************************************************************************************************/

  class Rectangle : public ColoredObject {
  private:
    int thumbnailID;

  public:
    Rectangle(uint32 id);
    Rectangle(int _x, int _y, int _width, int _height, char *_color,
	      char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id);
    ~Rectangle();

    virtual int GetType() {return GenericObject::RECTANGLE;};

    void SetThumbnailID(int id) { thumbnailID = id; }
    int  GetThumbnailID() { return thumbnailID; }
    bool IsThumbnail() { return (thumbnailID != -1); }

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);
    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();
    virtual bool Overlap(GUI::Canvas *canvas,int _x, int _y);

    virtual void AdjustGui();

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    static int FlushState(char* buf, uint32 order, bool visible, int x, int y, int width, int height,
			   Color linecolor, Color fillcolor, int linewidth, Style linestyle);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 24;}
    int GetStateSize() {return 24;}
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    int FlushChangeSize(char* buf, int x, int y, int width, int height);
    static int SFlushChangeSize(char* buf, int x, int y, int width, int height);
    static int SGetChangeSizeSize() {return 12;}
    int FlushChangeLinewidth(char* buf, int linewidth);
    int FlushChangeLinestyle(char* buf, Style linestyle);
    int FlushChangeLinecolor(char* buf, Color color);
    int FlushChangeFillcolor(char* buf, Color color);
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    static DrawObject *Load(SGMLElement *hilf);
  };

/***************************************************************************************************/

  class Oval : public ColoredObject {
  public:
    Oval(uint32 id);
    Oval(int _x, int _y, int _width, int _height, char *_color,
	 char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id);

    ~Oval();

    virtual int GetType() {return GenericObject::OVAL;}

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);
    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();
    virtual bool Overlap(GUI::Canvas *canvas,int _x, int _y);

    virtual void AdjustGui();

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    static int FlushState(char* buf, uint32 order, bool visible, int x, int y, int width, int height,
			   Color linecolor, Color fillcolor, int linewidth, Style linestyle);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 24;}
    int GetStateSize() {return 24;}
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    int FlushChangeSize(char* buf, int x, int y, int width, int height);
    static int SFlushChangeSize(char* buf, int x, int y, int width, int height);
    static int SGetChangeSizeSize() {return 12;}
    int FlushChangeLinewidth(char* buf, int linewidth);
    int FlushChangeLinestyle(char* buf, Style linestyle);
    int FlushChangeLinecolor(char* buf, Color color);
    int FlushChangeFillcolor(char* buf, Color color);
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    static DrawObject *Load(SGMLElement *hilf);
  };

/***************************************************************************************************/

  class Line : public ColoredObject {
  protected:
    int type;              /* normal line or freehand */
    Style capStyle;
    Style joinStyle;
    Style arrowStyle;
    Style arrowShape;
    double firstArrow[12];
    double lastArrow[12];

  public:

    Line(uint32 id);
    Line(int _x, int _y, int _width, int _height, char *_color,
	 int _lineWidth, char *_lineStyle, char *_capStyle, char *_joinStyle,
	 char *_arrowStyle, char *_arrowShape, uint32 _id);
    ~Line();


    virtual int GetType() {return GenericObject::LINE;};
    
    Style GetCapStyle() {return capStyle;}
    void  SetCapStyle(Style style) {capStyle = style;}
    Style GetJoinStyle() {return joinStyle;}
    void  SetJoinStyle(Style style) {joinStyle = style;}
    Style GetArrowStyle() {return arrowStyle;}
    void  SetArrowStyle(Style style) {arrowStyle = style;}
    Style GetArrowShape() {return arrowShape;}
    void  SetArrowShape(Style style) {arrowShape = style;}

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);

    virtual void DrawSelection(GUI::Canvas *canvas);
    virtual void DeleteSelection(GUI::Canvas *canvas);

    virtual bool IsEditablePoint(int _x, int _y);
    virtual void MovePoint(GUI::Canvas *canvas, int point, int _x, int _y, bool redraw = true);

    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();

    virtual void ChangeLineColor(GUI::Canvas *canvas,char *_color);
    virtual void ChangeCapStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeJoinStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeArrowStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeArrowShape(GUI::Canvas *canvas, char *_shape);
    virtual bool Overlap(GUI::Canvas *canvas, int _x, int _y);

    virtual void AdjustGui();

    void ConfigureArrows();

    static void FlushCapStyle(char* buf, Style& capstyle);
    static Style ParseCapStyle(char* buf);
    static void FlushJoinStyle(char* buf, Style& joinstyle);
    static Style ParseJoinStyle(char* buf);
    static void FlushArrowStyle(char* buf, Style& arrowstyle);
    static Style ParseArrowStyle(char* buf);
    static void FlushArrowShape(char* buf, Style& arrowshape);
    static Style ParseArrowShape(char* buf);

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    static int FlushState(char* buf, uint32 id, bool visible, int x, int y, int width, int height,
			   Color linecolor, int linewidth, Style linestyle,
			   Style capstyle, Style joinstyle, Style arrowstyle, Style arrowshape);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 24;}
    int GetStateSize() {return 24;}
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    static int FlushMovePoint(char* buf, int point, int x, int y);
    static int GetMovePointSize() {return 8;}
    int FlushChangeLinewidth(char* buf, int linewidth);
    int FlushChangeLinestyle(char* buf, Style linestyle);
    int FlushChangeLinecolor(char* buf, Color color);
    int FlushChangeCapstyle(char* buf, Style capstyle);
    int GetChangeCapstyleSize() {return 4;}
    int FlushChangeJoinstyle(char* buf, Style joinstyle);
    int GetChangeJoinstyleSize() {return 4;}
    int FlushChangeArrowstyle(char* buf, Style arrowstyle, Style arrowshape);
    int GetChangeArrowstyleSize() {return 8;}
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    static DrawObject *Load(SGMLElement *hilf);
  };

  /***************************************************************************************************/

  class PolyObject : public ColoredObject {
  protected:
    std::list<Point*>::iterator selectedPoint;
    std::list<Point*> _points;
    double lastZoom;

  public:
    std::list<Point*>  points;

    PolyObject(uint32 id) : ColoredObject(id) {}
    PolyObject(int _x, int _y, int _width, int _height, char *_color,
	       char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id);

    virtual ~PolyObject();

    virtual void DrawSelection(GUI::Canvas *canvas);
    virtual void DeleteSelection(GUI::Canvas *canvas);
    virtual int  GetSelectedPoint();

    virtual void Move(GUI::Canvas *canvas, int diffX, int diffY, bool redraw = true);
    virtual void MovePoint(GUI::Canvas *canvas,  int point,int x, int y, bool redraw = true);

    virtual bool IsEditablePoint(int _x, int _y);

    void AppendNewPoint(int _x, int _y);
    void DrawNewSegment(GUI::Canvas *canvas);
    void DeleteSegments(GUI::Canvas *canvas);
    virtual bool Overlap(GUI::Canvas *canvas, int _x, int _y);

    virtual int FlushMovePoint(char* buf, int point, int x, int y) {return 0;}
    virtual int GetMovePointSize() {return 12;}
  };

/***************************************************************************************************/

  class PolyLine : public PolyObject {
  protected:
    bool smooth;
    bool isFreehand;
    Style capStyle;
    Style joinStyle;
    Style arrowStyle;
    Style arrowShape;

  public:
    PolyLine(uint32 id, bool _isFreehand=false);
    PolyLine(int _x, int _y, int _width, int _height, char *_color,
	     int _lineWidth, char *_lineStyle, char *_capStyle, char *_joinStyle,
	     char *_arrowStyle, char *_arrowShape, bool _smooth, uint32 _id, bool _isFreehand=false);
    ~PolyLine();

    virtual int GetType();

    Style GetCapStyle() {return capStyle;}
    void  SetCapStyle(Style style) {capStyle = style;}
    Style GetJoinStyle() {return joinStyle;}
    void  SetJoinStyle(Style style) {joinStyle = style;}
    Style GetArrowStyle() {return arrowStyle;}
    void  SetArrowStyle(Style style) {arrowStyle = style;}
    Style GetArrowShape() {return arrowShape;}
    void  SetArrowShape(Style style) {arrowShape = style;}

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);
    virtual void Hide(GUI::Canvas *canvas);

    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();
    virtual void ChangeLineColor(GUI::Canvas *canvas,char *_color);
    virtual void ChangeLineStyle(GUI::Canvas *canvas,char *_lineStyle);
    virtual void ChangeCapStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeJoinStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeArrowStyle(GUI::Canvas *canvas, char *_style);
    virtual void ChangeArrowShape(GUI::Canvas *canvas, char *_shape);

    virtual void AdjustGui();

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    static int FlushState(char* buf, uint32 order, bool visible, int x, int y, Color linecolor, int linewidth, Style linestyle,
			   bool smooth, Style capstyle, Style joinstyle, Style arrowstyle, Style arrowshape);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 28;}
    int GetStateSize() {return 24 + points.size()*4;}
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    int FlushMovePoint(char* buf, int point, int x, int y);
    int FlushChangeLinewidth(char* buf, int linewidth);
    int FlushChangeLinestyle(char* buf, Style linestyle);
    int FlushChangeLinecolor(char* buf, Color color);
    int FlushChangeCapstyle(char* buf, Style capstyle);
    int GetChangeCapstyleSize() {return 4;}
    int FlushChangeJoinstyle(char* buf, Style joinstyle);
    int GetChangeJoinstyleSize() {return 4;}
    int FlushChangeArrowstyle(char* buf, Style arrowstyle, Style arrowshape);
    int GetChangeArrowstyleSize() {return 4;}
    static int FlushAddPoint(char* buf, int x, int y);
    static int GetAddPointSize() {return 8;}
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    static DrawObject *Load(SGMLElement *hilf);
  };

/***************************************************************************************************/

  class Polygon : public PolyObject {
  public:
    Polygon(uint32 id);
    Polygon(int _x, int _y, int _width, int _height, char *_color,
	    char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id);
    ~Polygon();

    virtual int GetType() {return GenericObject::POLYGON;};

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);

    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();

    virtual void ChangeLineColor(GUI::Canvas *canvas,char *_color);
    virtual void AdjustGui();

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    static int FlushState(char* buf, uint32 order, bool visible, int x, int y, Color linecolor, Color fillcolor, int linewidth, Style linestyle);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 28;}
    int GetStateSize() {return 24 + points.size()*4;}
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    int FlushMovePoint(char* buf, int point, int x, int y);
    int FlushChangeLinewidth(char* buf, int linewidth);
    int FlushChangeLinestyle(char* buf, Style linestyle);
    int FlushChangeLinecolor(char* buf, Color color);
    int FlushChangeFillcolor(char* buf, Color color);
    static int FlushAddPoint(char* buf, int x, int y);
    static int GetAddPointSize() {return 8;}
    static int FlushClosePolygon(char* buf);
    static int GetClosePolygonSize() {return 4;}
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    static DrawObject *Load(SGMLElement *hilf);
  };

/***************************************************************************************************/


  class Text : public ColoredObject {
  private:
    friend class Group;

    Char* GetActiveChar();
    std::list<Char*>::iterator GetActiveCharIterator();

    std::list<Char*> text;
    uint32 activeChar;

    std::vector<int> lineHeight;

    Font* font;

    std::string fontFamily;
    std::string fontSize;
    std::string fontWeight;
    std::string fontSlant;
    std::string textPosition;

    int ascent, descent;
    int offY;

    double lastZoom;

    void ZoomFont(GUI::Canvas *canvas);

  public:
    enum type {
      INSERT,
      BACKSPACE,
      RETURN,
      RIGHT,
      LEFT,
      UP,
      DOWN,
      HOME,
      END,
      RESOLUTION = 4
    };

    bool firstPos;

    Text(uint32 id);
    Text(int _x, int _y, int _width, int _height, char *_color,
	 char *_textString, char *_family, char *_size, char *_weight, char *_slant, char *_position, uint32 _id);
    Text(int _x, int _y, int _width, int _height, Color _color,
	 char* _textString, string _family, string _size, string _weight, string _slant, string _position, uint32 _id);
    ~Text();


    std::string GetFontFamily() {return fontFamily;}
    void SetFontFamily(std::string family) {fontFamily = family;}
    std::string GetFontSize() {return fontSize;}
    void SetFontSize(std::string size) {fontSize = size;}
    std::string GetFontWeight() {return fontWeight;}
    void SetFontWeight(std::string weight) {fontWeight = weight;}
    std::string GetFontSlant() {return fontSlant;}
    void SetFontSlant(std::string slant) {fontSlant = slant;}
    std::string GetTextPosition() {return textPosition;}
    void SetTextPosition(std::string pos) {textPosition = pos;}

    void InitFont(double scale);

    ///virtual void SetVisible();
    //virtual void SetInvisible();

    virtual bool IsVisible();
    virtual void Update();

    virtual int GetType() {return GenericObject::ASCIITEXT;};
    virtual int GetHeight();
    virtual int GetWidth() { return width; }

    virtual DrawObject *Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);

    virtual void DrawSelection(GUI::Canvas *canvas);
    virtual void DeleteSelection(GUI::Canvas *canvas);

    virtual bool IsEditablePoint(int _x, int _y);
    virtual DrawObject* PointIsInText(int _x, int _y);

    virtual void Hide(GUI::Canvas *canvas);
    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *_fp);
    virtual void SaveAsAOF(FILE *_fp, int _line, int _offX, int _offY);
    virtual int RecordableObjects();
    char *CalculatePSFontname();
    char *CalculateFontname(int &size);
    virtual int  SelectObject(int _x, int _y);
    //virtual int  ObjectIsInList(int length,char *selectList);
    virtual void Move(GUI::Canvas *canvas, int diffX, int diffY, bool redraw = true);
    virtual void ChangeLineColor(GUI::Canvas *canvas, char *_color);
    virtual void Change(Page* page,int argc, char *argv[], int ttype);
    virtual bool Overlap(GUI::Canvas *canvas,int _x, int _y);

    void AppendNewChars(GUI::Canvas *canvas, char* _string);
    void InsertNewChars(GUI::Canvas *canvas, char* _string, char *family, char *size, char *weight, char *slant, char *position);
    bool DeleteChar(GUI::Canvas *canvas);

    void RemoteInsertNewChars(char* newChars);
    //void RemoteDeleteChars(GUI::Canvas *_canvas, int nbr);
    //void RemoteDeleteAllChars(GUI::Canvas *_canvas);
    //void RemoteDeleteActive(GUI::Canvas *_canvas);

    
    std::list<Char*> GetText() { return text; }
    bool Empty() {return text.empty();}

    char *GetTextRef();
    bool GoRight();
    bool GoLeft();
    bool GoEnd();
    bool GoHome();
    int  GetPos();
    void Update(GUI::Canvas *canvas);
    void GetBBox(int *minX, int *minY, int *maxX, int *maxY);
    void UpdateItemList();
    void UpdateLineHeights(GUI::Canvas *canvas);
    void ChangeFont(GUI::Canvas *canvas, int argc, char *argv[]);
    void ChangeFont(GUI::Canvas *canvas, string family, string size, string weight, string slant);
    void InsertNewTextColor(GUI::Canvas *canvas,char *_color);
    void NewLine(GUI::Canvas *canvas);
    void SplitCurrentItem(GUI::Canvas *canvas,int pos);
    void DrawTextCursor(GUI::Canvas *canvas);
    void DeleteTextCursor(GUI::Canvas *_canvas);
    void ChangeString(GUI::Canvas *canvas, char *_textString);
    int  EndPos();
    char *GetString();
    char *GetColor();
    char *GetFamily();
    char *GetSize();
    char *GetWeight();
    char *GetSlant();
    char *GetPosition();
    Font* GetFont() {return font;}

    void AdjustGui();
    int  GetStringSize() {return text.size();}
    int  GetCurrentX();
    int  GetCurrentIndex();
    int  GetEndIndex();
    char GetChar(int idx);

    void SetActive(int idx);

    Text* Split(GUI::Canvas *canvas);
    void  Split(char** first, char** last);

    int GetLength();

    static void FlushFontWeight(char* buf, std::string& fontWeight);
    static std::string ParseFontWeight(char* buf);

    static void FlushFontSlant(char* buf, std::string& fontSlant);
    static std::string ParseFontSlant(char* buf);

    static void FlushFontPosition(char* buf, std::string& fontPosition);
    static std::string ParseFontPosition(char* buf);

    static DrawObject *Load(SGMLElement *hilf);
  };

/***************************************************************************************************/

  class Group : public DrawObject {
  private:
    friend class SessionManager;
    friend class Text;

    std::list<DrawObject*>::iterator GetActiveItemIterator();

    std::list<DrawObject*> groupObjects;
    uint32 activeItem;
    bool textGroup;

  public:
    Group(uint32 _id);

    virtual int GetType() {return GenericObject::GROUP;};

    virtual void SetVisible();
    virtual void SetInvisible();

    DrawObject* Copy();
    DrawObject* Copy(Page* page);
    void SendCopy(Page* page);

    void SetOrder(uint32 order);
    int  GetObjectOrder(DrawObject* object);
    void SetObjectOrder(DrawObject* object, int newOrder);

    bool HasObject(uint32 id);
    DrawObject* GetObject(int idx);
    DrawObject* GetFirst() {if (groupObjects.empty()) return NULL; return groupObjects.front();}
    DrawObject* GetActiveItem();
    std::vector<uint32>* GetObjectIDs();
    std::list<DrawObject*> GetGroupObjects() { return groupObjects; }

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);
    virtual void DrawSelection(GUI::Canvas *canvas);
    virtual void DeleteSelection(GUI::Canvas *canvas); 
    virtual void Move(GUI::Canvas *canvas, int diffX, int diffY, bool redraw = true);

    virtual bool IsEditablePoint(int _x, int _y);
    virtual DrawObject* PointIsInText(int _x, int _y);

    virtual void SendMove(bool cue, int x, int y);
    virtual void SendDelete();
    virtual void SendAllObjects(uint32 timestamp, int priority = 3);

    void ChangeSubcomponentsActiveness(bool active);
    void ReplaceIDs();

    virtual void DeleteObjectsFromRTPI();

    virtual void SaveObject(FILE *fp, int level=0);
    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();
    virtual void SetUnrecorded();

    virtual bool Overlap(GUI::Canvas *canvas,int _x, int _y);
    virtual void Hide(GUI::Canvas *canvas);
    virtual void AdjustGui();

    void AddObject(DrawObject *obj);
    GenericObject* DeleteObject(GenericObject* obj);
    DrawObject* DeleteObject(uint32 id);
    void DeleteAll(GUI::Canvas *_canvas);

    void RemoteDeleteObject(DrawObject* obj);
    DrawObject* RemoteDeleteObject(uint32 id);

    bool AllEmpty();
    bool Empty() {return groupObjects.empty();}
    int GetSize() {return groupObjects.size();}

    bool IsTextGroup();
    void LoadObjects(GUI::Canvas *canvas, GenericContainer *page, SGMLElement *root);
    void UpdateDimension();
    void Update();

    void SetActiveItem(DrawObject* obj);

    /* Functions for Text-Object in Group */
    void InsertNewFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos);
    void ChangeFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant);
    void InsertNewColor(Page* page, Color color);
    void InsertNewLine(Page* page);
    void Change(Page* page, int argc, char *argv[], DrawObject *active, int _type);

    virtual void ChangeLineColor(GUI::Canvas *_canvas,char *_color);
    virtual void ChangeFillColor(GUI::Canvas *_canvas,char *_color);
    virtual void ChangeLineStyle(GUI::Canvas *_canvas, char *_lineStyle);
    virtual void ChangeLineWidth(GUI::Canvas *_canvas, int _lineWidth);

    void RemoteInsertNewChars(char* newChars);
    void RemoteGoLeft(bool own = false);
    void RemoteGoRight(bool own = false);
    void RemoteGoHome(bool own = false);
    void RemoteGoEnd(bool own = false);
    void RemoteGoUp(bool own = false);
    void RemoteGoDown(bool own = false);
    void RemoteSetActive(int x, int y);
    void RemoteBackspace(Page* page, bool own = false);
    void RemoteNewLine(Page* page, bool own = false);
    void RemoteInsertNewFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos, bool own = false);
    void RemoteChangeFont(Page* page, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos);
    void RemoteInsertNewColor(Page* page, Color color, bool own = false);
    void RemoteChangeColor(Page* page, Color color);

    void DrawTextCursor(GUI::Canvas *canvas);
    void DeleteTextCursor(GUI::Canvas *canvas);

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    // functions for regular Group
    static int FlushState(char* buf, uint32 order, int x, int y);
    int FlushState(char* buf);
    static int GetStateSizeS() {return 12;}
    int GetStateSize() {return 12;}
    int ParseState(char* buf);

    int FlushMove(char* buf, int x, int y);

    static int GetChangeActiveSize() {return 4;}
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    int FlushDelete(char* buf);

    // functions for Textgroup
    static int FlushStateTG(char* buf, uint32 order, bool visible, int x, int y, int width, int height, Color linecolor, string fontfamily, string fontsize,
			   string fontweight, string fontslant, string textpos);
    int FlushStateTG(char* buf);
    static int GetStateSizeTG(string fontFamily);
    int GetStateSizeTG();
    int ParseStateTG(char* buf, Page* page);

    int FlushChangeVisibilityTG(char* buf, bool vis);
    int FlushMoveTG(char* buf, int x, int y);
    int FlushChangeSizeTG(char* buf, int x, int y, int width, int height);
    int FlushChangeFontTG(char* buf, Color linecolor, string fontfamily, string fontsize, string fontweight, string fontslant);
    int ParseChangeFontTG(char* buf, string& fontFamily, string& fontSize, string& fontWeight, string& fontSlant);
    int FlushInsertNewFontTG(char* buf, Color linecolor, string fontfamily, string fontsize, string fontweight, string fontslant, string textpos);
    int ParseInsertNewFontTG(char* buf, Color& linecolor, string& fontFamily, string& fontSize, string& fontWeight, string& fontSlant, string& textpos);
    int GetChangeFontSizeTG(string fontfamily);
    int FlushChangeFontFamilyTG(char* buf, string fontfamily);
    int GetChangeFontFamilySizeTG(string fontfamily);
    int FlushChangeFontSizeTG(char* buf, string fontsize);
    int GetChangeFontSizeSizeTG() {return 8;}
    int FlushChangeFontColorTG(char* buf, Color color);
    int FlushInsertNewFontColorTG(char* buf, Color color);
    int GetChangeFontColorSizeTG() {return 8;}
    int FlushChangeFontWeightTG(char* buf, string fontweight);
    int GetChangeFontWeightSizeTG() {return 4;}
    int FlushChangeFontSlantTG(char* buf, string fontslant);
    int GetChangeFontSlantSizeTG() {return 4;}
    int FlushChangeFontPositionTG(char* buf, string fontposition);
    int GetChangeFontPositionSizeTG() {return 4;}
    int FlushInsertCharactersTG(char* buf, char* newChars);
    int GetInsertCharactersSizeTG(char* newChars);
    int FlushInsertNewlineTG(char* buf);
    int GetInsertNewlineSizeTG() {return 8;}
    int FlushDeleteCharactersTG(char* buf, int nbr);
    int FlushDeleteActiveTG(char* buf);
    int GetDeleteCharactersSizeTG() {return 8;}
    int FlushChangeParentObjectTG(char* buf, int nbr);
    int GetChangeParentObjectSizeTG() {return 8;}
    int FlushSetFirstActiveTG(char* buf);
    int FlushSetLastActiveTG(char* buf);
    int FlushSetLeftActiveTG(char* buf);
    int FlushSetRightActiveTG(char* buf);
    int FlushSetUpActiveTG(char* buf);
    int FlushSetDownActiveTG(char* buf);
    int FlushSetActiveTG(char* buf, int x, int y);
    int GetChangeActiveSizeTG() {return 8;}
    int FlushRaiseTG(char* buf, uint32 order);
    int FlushLowerTG(char* buf, uint32 order);

    static int FlushDeleteTGS(char* buf);
    int FlushDeleteTG(char* buf) { return FlushDeleteTGS(buf); }

    static DrawObject *Load(SGMLElement *hilf);

    bool saved;
  };


/***************************************************************************************************/

  class PhotoImage : public DrawObject {
  private:
    std::string
      filename,
      imgTag;
    Magick::Blob  blob;
    /** False if Tk photo block is not on the Tk side, true if we are not sure. **/
    bool          tkBlock;
    double        lastZoom;
    bool          defaultImg;

    void DeInit       ();
    void LoadDefaultImage ();
    void InitFileName ();
    void InitTkTags   ( int _id                 );
    void Resize       ( int _width, int _height );

  public:
    PhotoImage( const PhotoImage& image );
    PhotoImage( uint32 _id, char* _payload,  GUI::Canvas* _canvas );
    PhotoImage( uint32 _id, char* _filename, int _x, int _y, Magick::Image& _image );
    PhotoImage( uint32 _id, char* _filename, int _x, int _y, int _width, int _height, char*  _blob, unsigned int _blength );
    ~PhotoImage();

    char* GetARGB(int& size);

    virtual void  Delete        ();
    void          PutTkBlock    ();
    void          RemoveTkBlock ();

    uint32 GetBlobLength() {return blob.length();};
    char*  GetBlobData() {return (char*)blob.data();}

    void ModifyCoords(GUI::Canvas *canvas, int _x, int _y, int _width, int _height);
    void ModifyCoords(int _x, int _y, int _width, int _height);
    void Update(GUI::Canvas *canvas);

    virtual int GetType() {return GenericObject::IMAGE;}

    virtual DrawObject* Copy();

    virtual void Draw(GUI::Canvas *canvas);
    virtual void ReDraw(GUI::Canvas *canvas);

    virtual void DrawSelection(GUI::Canvas *canvas);
    virtual void DeleteSelection(GUI::Canvas *canvas);

    virtual bool IsEditablePoint(int _x, int _y);

    virtual void Move(GUI::Canvas *canvas, int diffX, int diffY, bool redraw = true);

    virtual void SaveObject(FILE *fp, int level=0);

    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false);
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects();

    virtual void AdjustGui() {};

    virtual RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

    //static int FlushState(char* buf, uint32 order, bool visible, int x, int y, int width, int height, char* blob, uint32 blobLength);
    int FlushState(char* buf);
    //static int GetStateSize(uint32 blobLength);
    int GetStateSize();
    int ParseState(char* buf, GUI::Canvas *canvas);

    int FlushChangeVisibility(char* buf, bool vis);
    int FlushMove(char* buf, int x, int y);
    int FlushChangeSize(char* buf, int x, int y, int width, int height);
    int FlushRaise(char* buf, uint32 order);
    int FlushLower(char* buf, uint32 order);

    static int FlushDeleteS(char* buf);
    int FlushDelete(char* buf) { return FlushDeleteS(buf); }

    /* Function for AOF-Recording */
    void WriteImage();

    static DrawObject *Load(SGMLElement *hilf);
  };

}

#endif
