/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_page.cpp
 authors : Juergen Vogel, Gabriela Maass
 date    : 06.03.2001
 desc    :

***********************************************************************************************************/

#ifdef WIN32
#pragma warning(disable: 4786)
#endif

#include <set>
#include <map>
#include <tcl.h>
#include <math.h>
#include <util/types.h>

#include "freetype.h"
#include "pointer.h"

#include "mlb.h"
#include "mlb_page.h"
#include "mlb_session.h"
#include "mlb_adu.h"
#include "mlb_gui.h"
#include "mlb_participants.h"
#include "mlb_undomanager.h"


//#define _DEBUG1
//#define DEBUG_OBJ

#define MAX(x, y) (((x) > (y)) ? (x) : (y))
#define MIN(x, y) (((x) < (y)) ? (x) : (y))


MLB::Page::Page(uint32 _id)
  : MLB::GenericContainer(_id), onlinePointer(NULL), onlineBox(NULL), animations(*this) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Page %u\n",_id);
  #endif

  SetName("Page");

  active     = NULL;
  width      = 0;
  height     = 0;
  background = "#FFFFFF";

  canvas        = NULL;
  onlinePointer = NULL;
  onlineBox     = NULL;

  SessionManager::sm->pageCount++;
  pageNumber = SessionManager::sm->pageCount;

  firstTime  = -1;
}


MLB::Page::Page(uint32 _id, int _sessionType, char *_name, int _width, int _height, string _color)
  : MLB::GenericContainer(_id), animations(*this) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Page %u %s %d %d %s\n",_id,_name,_width,_height,_color.c_str());
  #endif

  SetName(_name);

  active     = NULL;
  width      = _width;
  height     = _height;
  background = _color;

  Init();

  SessionManager::sm->pageCount++;
  pageNumber = SessionManager::sm->pageCount;

  firstTime  = -1;
}


MLB::Page::Page(uint32 _id, int _sessionType, SGMLElement *_root)
  : MLB::GenericContainer(_id), animations(*this) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Page %u %d\n",_id,_sessionType);
  #endif

  active = NULL;
  int thumbnailID;

  char* tmp = _root->GetAttribute("name");
  SetName(tmp);

  tmp = _root->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = A4_WIDTH;

  tmp = _root->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = A4_HEIGHT;

  tmp = _root->GetAttribute("color");
  if (tmp) background = tmp;
  else     background = "#ffffff";

  tmp = _root->GetAttribute("thumbnailID");
  if (tmp) thumbnailID = atoi(tmp);
  else     thumbnailID = -1;

  if (thumbnailID != -1) {
    SessionManager::sm->GetActiveSession()->thumbnailIDs[thumbnailID] = _id;
  }
  
  Init();

  LoadObjects(_root);
  //SetAnimationEditMode( SessionManager::sm->GetAnimationEditMode() );

  SessionManager::sm->pageCount++;
  pageNumber = SessionManager::sm->pageCount;

  firstTime = -1;
}


MLB::Page::~Page() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::~Page\n");
  #endif

  DeleteAll(canvas);

  containerName.erase();
  nodeName.erase();

  delete canvas;
}


void MLB::Page::DeleteAll(GUI::Canvas* _canvas) {
  UnSelectObjects();

  bool hide = false;
  if (SessionManager::sm->GetActivePage() == this) hide = true;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if (hide) (*i)->Hide(canvas);
    delete *i;
  }
  UndoManager::um->DeleteAdus(GetID());
  objects.clear();
  selectedObjects.clear();
  active = NULL;

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    delete *j;
  }
  groups.clear();

  if (onlinePointer) {
    if (hide) onlinePointer->Hide(this);
    delete onlinePointer;
    onlinePointer = NULL;
  }

  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k) {
    if (hide) (*k)->Hide(this);
    delete *k;
  }
  pointers.clear();

  if (onlineBox) {
    if (hide) onlineBox->Hide(this);
    delete onlineBox;
    onlineBox = NULL;
  }

  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l) {
    if (hide) (*l)->Hide(this);
    delete *l;
  }
  boxes.clear();
}


void MLB::Page::SetActiveObject(DrawObject* obj) {
  active = obj;
}


void MLB::Page::PrintAllObjects() {
  GUI::MessageWindow::Debug("Page::PrintALlObjects - P%u ", GetID());
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) GUI::MessageWindow::Debug("O%u ", (*i)->GetID());
  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) GUI::MessageWindow::Debug("G%u ", (*j)->GetID());
  if (onlinePointer) GUI::MessageWindow::Debug("TP%u ", onlinePointer->GetID());
  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k) GUI::MessageWindow::Debug("TP%u ", (*k)->GetID());
  if (onlineBox) GUI::MessageWindow::Debug("TB%u ", onlineBox->GetID());
  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l) GUI::MessageWindow::Debug("TB%u ", (*l)->GetID());
  GUI::MessageWindow::Debug("\n");
}


void MLB::Page::PrintActiveSubcomponents() {
  RTPI::RtpiSubComponent* sub; 
  
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    sub = CommunicationManager::cm->GetRtpi()->get_subcomponent((*i)->GetID());
    if (sub && sub->is_active()) GUI::MessageWindow::Debug(" O%u", (*i)->GetID());
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    sub = CommunicationManager::cm->GetRtpi()->get_subcomponent((*j)->GetID());
    if (sub && sub->is_active()) GUI::MessageWindow::Debug(" G%u", (*j)->GetID());
  }
}


void MLB::Page::Init() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Init\n");
  #endif

  InitTelepointer();
  
  canvas = new GUI::Canvas(GetID(),SessionManager::sm->GetZoom()*width,SessionManager::sm->GetZoom()*height,background.c_str());
}


void MLB::Page::InitTelepointer() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::InitTelepointer\n");
  #endif

  onlinePointer = new Telepointer(0,0,"none",MLB::Telepointer::INACTIVE,
				  CommunicationManager::cm->GetObjectID(),
				  CommunicationManager::cm->GetParticipantID());
  InitTelebox();
}


void MLB::Page::InitTelebox() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::InitTelebox\n");
  #endif

  onlineBox = new Telebox(0,0,0,0,"none",0,"", MLB::Telepointer::INACTIVE,
				  CommunicationManager::cm->GetObjectID(),
				  CommunicationManager::cm->GetParticipantID());
}


void MLB::Page::LoadObjects(SGMLElement *_root) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::LoadObjects\n");
  #endif

  SGMLElement
    *hilf,
    *hilfSon,
    *hilfSonSon;
  DrawObject
    *obj;
  char
    *tmp;
  int
    length;

  // animations
  Animations::StringStepMap showObjectStepsMap;
  Animations::StringStepMap hideObjectStepsMap;

  for (hilf = _root->son; hilf != NULL; hilf = hilf->next) {

    if( hilf->Name=="ANIMATIONS" )
    {
      for( hilfSon = hilf->son; hilfSon != NULL; hilfSon = hilfSon->next )
      {
        char* tmp = hilfSon->GetAttribute( "name" );
        if( hilfSon->Name == "STEP" && tmp )
        {
          // name
          string stepName( tmp );
          AnimationStep* newStep = &(*(animations.AddNewStep()));

          // title
          tmp = hilfSon->GetAttribute( "title" );
          if( tmp ) newStep->SetTitle( tmp );

          // animation objects
          for( hilfSonSon = hilfSon->son; hilfSonSon != NULL; hilfSonSon = hilfSonSon->next) {
            if( hilfSonSon->Name=="OBJECT" )
            {
              // read name
              tmp = hilfSonSon->GetAttribute( "name" );
              if( !tmp ) continue;
              string objName(tmp);

              // read command
              tmp = hilfSonSon->GetAttribute( "onAnimate" );
              if( !tmp ) continue;
              string command(tmp);

              // add to one of the maps
              if     ( "show" == command ) showObjectStepsMap.insert( Animations::StringStepMap::value_type( objName, newStep ) );
              else if( "hide" == command ) hideObjectStepsMap.insert( Animations::StringStepMap::value_type( objName, newStep ) );

            } // if object
          } // for objects

          // insert
          //externalNaming.insert( StringStepMap::value_type( stepName, newStep ) );
        }
      } // steps
    }
    else if( hilf->Name=="LAYER" ) {

      for (hilfSon = hilf->son; hilfSon != NULL; hilfSon = hilfSon->next) {
        obj = NULL;
        if (hilfSon->Name == "LINE")           obj = Line::Load(hilfSon);
        else if (hilfSon->Name == "RECTANGLE") obj = Rectangle::Load(hilfSon);
        else if (hilfSon->Name == "OVAL")      obj = Oval::Load(hilfSon);
        else if (hilfSon->Name == "POLYLINE")  obj = PolyLine::Load(hilfSon);
        else if (hilfSon->Name == "POLYGON")   obj = Polygon::Load(hilfSon);
        else if (hilfSon->Name == "IMAGE"     ) obj = PhotoImage::Load(hilfSon);
        else if (hilfSon->Name == "TEXT") { // Text only appear in Group
          Group* group = new Group(CommunicationManager::cm->GetObjectID());
          obj = Text::Load(hilfSon);
          obj->SetOrder(GetPushBackObjectOrder()); 
          AppendObject(obj);
          group->AddObject(obj);
          obj = group;
        }
        else if (hilfSon->Name == "GROUP") {
          obj = Group::Load(hilfSon);
          ((Group *)obj)->LoadObjects(GetCanvas(),this,hilfSon);
        }

        if (obj) {
          obj->SetOrder(GetPushBackObjectOrder()); 
          AppendObject(obj);

          // read object name
          tmp = hilfSon->GetAttribute( "name" );
          if( !tmp ) continue;
          string objName(tmp);

          // ANIMATIONS
          /////////////
          Animations::StringStepMap::iterator iMap;

          // add object to steps to show
          for( iMap=showObjectStepsMap.lower_bound(objName); iMap!=showObjectStepsMap.upper_bound(objName); iMap++ )
            iMap->second->AddObject( obj, true );

          // add object to steps to hide
          for( iMap=hideObjectStepsMap.lower_bound(objName); iMap!=hideObjectStepsMap.upper_bound(objName); iMap++ )
            iMap->second->AddObject( obj, false );

/*
          // ANIMATIONS
          const string aName = obj->GetAnimationStepName();
          if( obj && ""!=aName )
          {
            // find needed animation step
            StringStepMap::iterator ag = externalNaming.find( aName );
            //if( ag==externalNaming.end() ) {
            //  AnimationStep* newStep = &(*(animations.AddNewStep()));
            //  ag = externalNaming.insert( StringStepMap::value_type( aName, newStep ) ).first;
            //}
            if( ag!=externalNaming.end() ) (*ag).second->AddObject( obj, true );
*/
        } // if obj not null
      } // for objects
    } // if layer
  } // for page sons
} // LoadObjects()


void MLB::Page::SetName(char *_name) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SetName\n");
  #endif

  if (_name == NULL) containerName = "Page";
  else containerName = _name;
}


MLB::GenericContainer* MLB::Page::SetCurrentContainer(std::vector<uint32>* _idVector, int _pos) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SetCurrentContainer\n");
  #endif

  Activate();

  return this;
}


void MLB::Page::Activate(bool _rebuild) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Activate %d\n",_rebuild);
  #endif

  Page* oldPage = SessionManager::sm->GetActivePage();
  if (oldPage) {
    if (!_rebuild && (oldPage == this)) return;
    oldPage->DeActivate();
  }

  active = NULL;
  UnSelectObjects();

  SessionManager::sm->SetActivePage(this);
  //canvas->Show();
  //canvas->Delete();
  //canvas->Create(id, SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height, background.c_str());
  //canvas->Show();

  Session* session = SessionManager::sm->GetParentSession(this);
  bool thumbnailPage = session->thumbnailFolder && SessionManager::sm->IsParentOf(session->thumbnailFolder->GetNodePath(), GetNodePath());
  
  if (thumbnailPage) {
    const char** argv = new const char*[2];
    argv[0] = "";
    argv[1] = "1";
    session->SelectThumbnail(1, argv);
    delete[] argv;
  }

  
  ReDraw();

  GUI::Attributes::SetPageBackground(background.c_str());
  GUI::Attributes::SetPageDimension(width,height);
  SaveAsAOF();

  #ifdef DEBUG_OBJ
  PrintAllObjects();
  SessionManager::sm->PrintActiveSubcomponents();
  #endif

  return;
}

void MLB::Page::DeActivate() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DeActivate\n");
  #endif

  onlinePointer->SetInactive(this);
  onlineBox->SetInactive(this);
  if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
    int size = onlinePointer->FlushChangeVisibility(TelepointerManager::sendBuffer);
    TelepointerManager::tm->SendEvent(onlinePointer->GetID(), TelepointerManager::sendBuffer, size);
    size = onlineBox->FlushChangeVisibility(TelepointerManager::sendBuffer);
    TelepointerManager::tm->SendEvent(onlineBox->GetID(), TelepointerManager::sendBuffer, size);
  }

  if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
  }

  canvas->Clear(SessionManager::sm->GetZoom()*width,SessionManager::sm->GetZoom()*height);
  //canvas->Hide();
  canvas->Delete();
  GUI::Image::DestroyImages();
  SessionManager::sm->SetActivePage(NULL);
}


void MLB::Page::ShowTicks(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ShowTicks\n");
  #endif

  int status = atoi(argv[1]);
  int x      = atoi(argv[2]);
  int y      = atoi(argv[3]);

  switch (status) {
  case MLB::Action::BEGIN:
    canvas->BeginTicks(x,y);
    break;
  case MLB::Action::CONTINUE:
    canvas->MoveTicks(x,y);
    break;
  case MLB::Action::END:
    canvas->EndTicks(x,y);
    break;
  }
}


void MLB::Page::CopyObjects(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::CopyObjects\n");
  #endif

  static int lastX, lastY;
  static std::list<DrawObject*> copyObjects;

  if (objects.empty()) return;

  std::list<DrawObject*>::iterator i;
  DrawObject* obj;
  uint32 objID;
  bool selected;

  int status = atoi(argv[1]);
  int x      = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y      = atoi(argv[3]) / SessionManager::sm->GetZoom();

  if (status == Action::BEGIN) {
    lastX = x;
    lastY = y;
    selected = false;

    copyObjects.clear();

    for (i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
      if ((*i)->Overlap(canvas, x, y)) {
	selected = true;
	break;
      }
    }

    if (selected) {
      for (i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
	if (!DrawObject::Find(copyObjects, *i)) copyObjects.push_back((*i));
      }
    } else UnSelectObjects();

    if (copyObjects.empty()) {
      i = objects.end();
      do {
	--i;
	if ((*i)->Overlap(canvas,x,y)) {
	  if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not edit images
	  obj = (*i);
	  while (obj->GetGroup())  obj = obj->GetGroup();
	  if (!DrawObject::Find(copyObjects, obj)) copyObjects.push_back(obj);
	  break;
	}
      } while (i != objects.begin());
    }
  }

  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) {
    static std::list<uint32> copyBuffer;
    static std::list<int> objTypesBuffer;
    static int count = 0;
    static double test;
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;
    switch (status) {
    case Action::BEGIN: {
      copyBuffer.clear();
      objTypesBuffer.clear();
      for (std::list<DrawObject*>::iterator i = copyObjects.begin(); i != copyObjects.end(); ++i) {
	if (!(*i)->Annotation()) {
	  switch ((*i)->GetType()) {
	  case MLB::GenericObject::RECTANGLE: {
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::RECTANGLE, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::RECTANGLE);
	    break;
	  }
	  case MLB::GenericObject::OVAL: {
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::OVAL, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::OVAL);
	    break;
	  }
	  case MLB::GenericObject::LINE: {
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::LINE, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::LINE);
	    break;
	  }
	  case MLB::GenericObject::FREEHAND:
	  case MLB::GenericObject::POLYLINE: {
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::POLYLINE);
	    break;
	  }
	  case MLB::GenericObject::POLYGON: {
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYGON, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::POLYGON);
	    break;
	  }
	  case MLB::GenericObject::IMAGE: {
	    if (SessionManager::sm->FreezeImages()) continue; // do not edit images 
	    buf   = new Buffer((*i)->GetStateSize());
	    (*i)->FlushState(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::IMAGE, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	    objTypesBuffer.push_back(CommunicationManager::IMAGE);
	    break;
	  }
	  case MLB::GenericObject::GROUP: {
	    if (((Group*)*i)->IsTextGroup()) {
	      buf   = new Buffer(((Group*)*i)->GetStateSizeTG());
	      ((Group*)*i)->FlushStateTG(buf->getCont());
	      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	      objID = CommunicationManager::cm->GetObjectID();
	      adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::TEXT, 3, buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	      objTypesBuffer.push_back(CommunicationManager::TEXT);
	    } else {
	      /*
	      buf   = new Buffer((*i)->GetStateSize());
	      (*i)->FlushState(buf->getCont());
	      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	      GenericObject::FlushCangedObjectOrder(buf, GetPushBackObjectOrder());
	      objID = CommunicationManager::cm->GetObjectID();
	      adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::GROUP, 3, buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	      */
	      ((Group*)*i)->SendCopy(this);
	      objTypesBuffer.push_back(CommunicationManager::GROUP);
	    }
	    break;
	  }
	  }
	      
          aduList->push_back(UndoManager::um->CreateStateInverse(adu));

	} else { // Annotation
	  if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not edit images 
	  if ((*i)->GetType() == GenericObject::GROUP) {
	    obj = ((Group*)*i)->Copy(this);
	    objID = obj->GetID();
	    obj->SetAnnotation(SessionManager::sm->AnnotationMode());
	    obj->SetOrder(GetPushBackObjectOrder());
	    AppendObject(obj);
	    obj->Draw(canvas);
	    if (((Group*)*i)->IsTextGroup()) ((Group*)*i)->DrawTextCursor(GetCanvas());
	  } else {
	    obj = (*i)->Copy();
	    if( !obj ) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
	    objID = obj->GetID();
	    obj->SetAnnotation((*i)->Annotation());
	    obj->SetOrder(GetPushBackObjectOrder());
	    AppendObject(obj);
	    obj->Draw(canvas);
	  }
          
          RefCount<Buffer> undoBuf = new Buffer(GenericObject::GetDeleteSize());
          obj->FlushDelete(undoBuf->getCont());
          aduList->push_back(CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf));
	}

	copyBuffer.push_back(objID);
      }
      if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

      break;
    }
    case MLB::Action::CONTINUE : {
      std::list<int>::iterator j = objTypesBuffer.begin();
      for (std::list<uint32>::iterator i = copyBuffer.begin(); i != copyBuffer.end(); ++i) {
	if (!SessionManager::sm->AnnotationMode()) {
	  buf   = new Buffer(DrawObject::GetMoveSizeS());
	  DrawObject::FlushMoveS(buf->getCont(), *j, x, y);
	  adu   = CommunicationManager::cm->CreateCue(*i, buf);
	  AduManager::am->InsertAdu(adu);
	  test  = CommunicationManager::cueSendRate * count;
	  if (floor(test) == ceil(test)) {
	    CommunicationManager::cm->SendAduUnreliable(adu);
	  }
	  if (count == 100) count = 1;
	  else ++count;
	  ++j;
	}
	else { // Annotation
	  obj = (DrawObject*)GetObject(*i);
	  if (!obj) obj = (DrawObject*)GetGroup(*i);

	  obj->DeleteSelection(canvas);
	  obj->Move(canvas, x-lastX, y-lastY);
	  obj->DrawSelection(canvas);
	}
      }
      lastX = x;
      lastY = y;
      break;
    }
    case MLB::Action::END : {
      std::list<int>::iterator j = objTypesBuffer.begin();
      for (std::list<uint32>::iterator i = copyBuffer.begin(); i != copyBuffer.end(); ++i) {
	if (!SessionManager::sm->AnnotationMode()) {
	  buf   = new Buffer(DrawObject::GetMoveSizeS());
	  DrawObject::FlushMoveS(buf->getCont(), *j, x, y);
	  adu   = CommunicationManager::cm->CreateEvent(*i, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  ++j;
	} else { // Annotation
	  //obj->DeleteSelection(canvas);
	}
      }
      objTypesBuffer.clear();
      copyBuffer.clear();
      Draw();
      break;
    }
    }
  } else {  // Private
    switch (status) {
    case Action::BEGIN : {
      for (std::list<DrawObject*>::iterator i = copyObjects.begin(); i != copyObjects.end(); ++i) {
	if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not edit images
	if ((*i)->GetType() == MLB::GenericObject::GROUP) {
	  obj = ((Group*)*i)->Copy(this);
	  if (!obj) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
	} else {
	  obj = (*i)->Copy();
	  if (!obj) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
	}
	obj->SetOrder(GetPushBackObjectOrder());
	obj->SetX(obj->GetX() + 10); obj->SetY(obj->GetY() + 10);
	AppendObject(obj);
	obj->Select();
	obj->Draw(canvas);
	*i = obj;

        RefCount<Buffer> undoBuf = new Buffer(GenericObject::GetDeleteSize());
        obj->FlushDelete(undoBuf->getCont());
        aduList->push_back(CommunicationManager::cm->CreateEvent(obj->GetID(), GetID(), undoBuf));
      }
      if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
      break;
    }
    case Action::CONTINUE : {
      for (std::list<DrawObject*>::iterator i = copyObjects.begin(); i != copyObjects.end(); ++i) {
	(*i)->Move(canvas, x-lastX, y-lastY);
      }
      lastX = x;
      lastY = y;
      break;
    }
    case Action::END : {
      SaveAsAOF();
      for (std::list<DrawObject*>::iterator i = copyObjects.begin(); i != copyObjects.end(); ++i)
	(*i)->Select(false);
      copyObjects.clear();
      Draw();
      break;
    }
    }
  }
}


void MLB::Page::CutObjects() {
  if (objects.empty()) return;

#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::CutObjects\n");
#endif

  DrawObject* obj;

  if (!SessionManager::sm->pasteBuffer.empty()) {
    for (std::list<DrawObject*>::iterator i = SessionManager::sm->pasteBuffer.begin(); i != SessionManager::sm->pasteBuffer.end(); ++i) {
      delete (*i);
    }
    SessionManager::sm->pasteBuffer.clear();
  }

  if (selectedObjects.empty()) return;

  if (active && (active->GetType()== GenericObject::ASCIITEXT)) ((Text*)active)->DeleteTextCursor(canvas);
  if (active && active->IsSelected()) active->DeleteSelection(canvas);
  active = NULL;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (std::list<DrawObject*>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
    if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not edit images
      
    if ((*i)->GetType() == GenericObject::GROUP) {
      if (((Group*)*i)->IsTextGroup()) {
	buf = new Buffer((*i)->GetDeleteSize());
	((Group*)*i)->FlushDeleteTG(buf->getCont());
      } else {
	buf = new Buffer((*i)->GetDeleteSize());
	(*i)->FlushDelete(buf->getCont());
      }
    } else {
      buf = new Buffer((*i)->GetDeleteSize());
      (*i)->FlushDelete(buf->getCont());
    }
	  
    adu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), buf);
    obj = (*i)->Copy();
    if (!obj) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
 
    SessionManager::sm->pasteBuffer.push_back(obj);
    aduList->push_back(UndoManager::um->CreateEventInverse(adu));

    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) {
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    } else { // PRIVATE or Annotation
      (*i)->DeleteSelection(canvas);
      DeleteObject(*i);
      animations.DeleteObject(*i);
      delete *i;
    }
  }
  
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
    
  SaveAsAOF();

  selectedObjects.clear();
}


void MLB::Page::CopyObjects() {
  // there are no special actions for shared documents since existing objects are not manipulated yet

#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::CopyObjects\n");
#endif

  if (objects.empty()) return;

  std::list<DrawObject*>::iterator i;
  DrawObject* obj;

  if (!SessionManager::sm->pasteBuffer.empty()) {
    for (i = SessionManager::sm->pasteBuffer.begin(); i != SessionManager::sm->pasteBuffer.end(); ++i) {
      delete (*i);
    }
    SessionManager::sm->pasteBuffer.clear();
  }

  if (!selectedObjects.empty()) {
    for (i= selectedObjects.begin(); i != selectedObjects.end(); ++i) {
      if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not edit images

      GUI::MessageWindow::Debug("Page::CopyObjects %u %u\n", (*i)->GetID(), (*i)->GetOrder());

      obj = (*i)->Copy();
      if (!obj) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
      SessionManager::sm->pasteBuffer.push_back(obj);
    }
    SaveAsAOF();
    UnSelectObjects();
  }

  if (active && active->IsSelected()) active->DeleteSelection(canvas);
  active = NULL;
}

void MLB::Page::PasteObjects() {
  if (SessionManager::sm->pasteBuffer.empty()) return;

#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::PasteObjects\n");
#endif

  DrawObject* obj;

  UnSelectObjects();
  active = NULL;

  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) {
    uint32 objID;
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;

    for (std::list<DrawObject*>::iterator i= SessionManager::sm->pasteBuffer.begin(); i != SessionManager::sm->pasteBuffer.end(); ++i) {
      if (!(*i)->Annotation()) {
	switch ((*i)->GetType()) {
	case MLB::GenericObject::RECTANGLE: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::RECTANGLE, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::OVAL: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::OVAL, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::LINE: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::LINE, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::FREEHAND:
	case MLB::GenericObject::POLYLINE: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::POLYGON: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYGON, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::IMAGE: {
	  buf   = new Buffer((*i)->GetStateSize());
	  (*i)->FlushState(buf->getCont());
	  GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	  objID = CommunicationManager::cm->GetObjectID();
	  adu = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::IMAGE, 3, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  break;
	}
	case MLB::GenericObject::GROUP: {
	  if (((Group*)*i)->IsTextGroup()) {
	    buf   = new Buffer(((Group*)*i)->GetStateSizeTG());
	    ((Group*)*i)->FlushStateTG(buf->getCont());
	    GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	    objID = CommunicationManager::cm->GetObjectID();
	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::TEXT, 3, buf);
	    AduManager::am->InsertAdu(adu);
	    CommunicationManager::cm->SendAduReliable(adu);
	  } else {
	    /*
	      buf   = new Buffer((*i)->GetStateSize());
	      (*i)->FlushState(buf->getCont());
	      GenericObject::ChangeFlushedObjectOrder(buf->getCont(), GetPushBackObjectOrder());
	      GenericObject::FlushCangedObjectOrder(buf, GetPushBackObjectOrder());
	      objID = CommunicationManager::cm->GetObjectID();
	      adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::GROUP, 3, buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    */
	    ((Group*)*i)->SendCopy(this);
	  }
	  break;
	}
	}
        aduList->push_back(UndoManager::um->CreateStateInverse(adu));

      } else { // Annotation
	if ((*i)->GetType() == MLB::GenericObject::GROUP)
	  obj = ((Group*)*i)->Copy(this);
	else
	  obj = (*i)->Copy();
          if( !obj ) { GUI::MessageWindow::Debug("MLB::Page:: ... cannot copy an object \n"); return; }
	AppendObject(obj);
	obj->Draw(canvas);
      
        RefCount<Buffer> undoBuf = new Buffer(GenericObject::GetDeleteSize());
        obj->FlushDelete(undoBuf->getCont());
        aduList->push_back(CommunicationManager::cm->CreateEvent(obj->GetID(), GetID(), undoBuf));
      }
      if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
    }

  } // PRIVATE
  else {

    for (std::list<DrawObject*>::iterator i= SessionManager::sm->pasteBuffer.begin(); i != SessionManager::sm->pasteBuffer.end(); ++i) {
      if ((*i)->GetType() == MLB::GenericObject::GROUP)
	obj = ((Group*)*i)->Copy(this);
      else
	obj = (*i)->Copy();
      AppendObject(obj);
      obj->Draw(canvas);

      RefCount<Buffer> undoBuf = new Buffer(GenericObject::GetDeleteSize());
      obj->FlushDelete(undoBuf->getCont());
      aduList->push_back(CommunicationManager::cm->CreateEvent(obj->GetID(), GetID(), undoBuf));
    }
    if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
  }

  SaveAsAOF();
}


void MLB::Page::Draw(){
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::Draw %u\n", id);
  #endif

  if (objects.empty()) return;

  MLB::SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);

  canvas->SetDimension(SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height);
  canvas->Clear(SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height);

  for (std::list<MLB::DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->Annotation()) {
      if (SessionManager::sm->AnnotationMode())
	(*i)->Draw(canvas);
    }
    else {
      (*i)->Draw(canvas);
    }
  }

  if (onlinePointer) onlinePointer->Draw(this);
  for (std::list<Telepointer*>::iterator j = pointers.begin(); j != pointers.end(); ++j)
    (*j)->Draw(this);

  if (onlineBox) onlineBox->Draw(this);
  for (std::list<Telebox*>::iterator k = boxes.begin(); k != boxes.end(); ++k)
    (*k)->Draw(this);

  MLB::SessionManager::sm->cursor->Reset();
}


void MLB::Page::ReDraw() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ReDraw %u\n", id);
  #endif

  MLB::SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);

  //canvas->SetDimension(SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height);
  //canvas->Clear(SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height);
  //canvas->SetBackground(background.c_str());

  active = NULL;

  Session* session = SessionManager::sm->GetParentSession(this);
 bool thumbnailPage = session->thumbnailFolder && SessionManager::sm->IsParentOf(session->thumbnailFolder->GetNodePath(), GetNodePath());
  
  canvas->Delete();
  canvas->Create(id, SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height, background.c_str(), thumbnailPage);
  canvas->Clear(SessionManager::sm->GetZoom()*width, SessionManager::sm->GetZoom()*height);
  canvas->Show();

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->Annotation()) {
      if (SessionManager::sm->AnnotationMode())	(*i)->Draw(canvas);
    } else {
      (*i)->Draw(canvas);
    }
  }

  if (onlinePointer) onlinePointer->Draw(this);
  for (std::list<Telepointer*>::iterator j = pointers.begin(); j != pointers.end(); ++j)
    (*j)->Draw(this);

  if (onlineBox) onlineBox->Draw(this);
  for (std::list<Telebox*>::iterator k = boxes.begin(); k != boxes.end(); ++k)
    (*k)->Draw(this);

  MLB::SessionManager::sm->cursor->Reset();
}


void MLB::Page::DrawSelected() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DrawSelected\n");
  #endif

  if (selectedObjects.empty()) return;

  for (std::list<MLB::DrawObject*>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i)
    (*i)->DrawSelection(canvas);
}


void MLB::Page::DrawPageTree(string _nodeNumber, string _nodePath, string _parentNode, int _pos, bool *_open) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DrawPageTree %s %s %s %d\n",_nodeNumber.c_str(),_nodePath.c_str(),_parentNode.c_str(),_pos);
  #endif

  char buf[20];

  sprintf(buf,".%u",id);
  nodePath = _nodePath + buf;

  if (nodeName.empty()) {
    sprintf(buf,"n:%u",id);
    nodeName = buf;
    GUI::PageView::InsertPageNode((char *)_nodeNumber.c_str(),(char *)nodeName.c_str(),
				  (char *)containerName.c_str(),(char *)_parentNode.c_str(),
				  (char *)nodePath.c_str(),_pos);
  } else {
    GUI::PageView::DrawPageNode((char *)_nodeNumber.c_str(),(char *)nodeName.c_str(),
				(char *)containerName.c_str(),(char *)nodePath.c_str());
  }

  if (this == SessionManager::sm->GetActivePage())
    GUI::PageView::SelectNode(nodeName.c_str());
}


void MLB::Page::ClearPageTree() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ClearPageTree\n");
  #endif

  GUI::PageView::Delete(GetNodeName().c_str());
  ClearNodeName();
}


void MLB::Page::DeleteObject(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DeleteObject\n");
  #endif

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  if (objects.empty()) return;
  
  if (active && (active->GetType() == GenericObject::GROUP)) ((Group*)active)->DeleteTextCursor(canvas);
  if (active && active->IsSelected()) active->DeleteSelection(canvas);
  active = NULL;

  if (!selectedObjects.empty()) {
    for (std::list<DrawObject*>::iterator i= selectedObjects.begin(); i != selectedObjects.end(); ++i) {
      (*i)->DeleteSelection(canvas);
      if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
      if ((*i)->GetType() == GenericObject::GROUP) {
	Group* group = (Group*)*i;
	buf   = new Buffer(group->GetDeleteSize());
	group->FlushDeleteTG(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(group->GetID(), GetID(), buf);
      } else {
	buf   = new Buffer((*i)->GetDeleteSize());
	(*i)->FlushDelete(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), buf);
      }
	
      aduList->push_back(UndoManager::um->CreateInverse(adu));
	
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*i)->Annotation()) {
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
      }	else { // PRIVATE session or Annotation Mode
	if  ((*i)->GetType() == GenericObject::GROUP) {
	  (*i)->Hide(GetCanvas());
	  ((Group*)(*i))->DeleteTextCursor(GetCanvas());
	  
	  std::vector<uint32>* groupObjects = ((Group*)(*i))->GetObjectIDs();
	  for (std::vector<uint32>::iterator j = groupObjects->begin(); j != groupObjects->end(); ++j) {
	    GetAnimations().DeleteObject((DrawObject*)(GetObject(*j)));
	    delete DeleteObject(*j);
	  }
	  ((Group*)(*i))->DeleteAll(GetCanvas());
	  DeleteObject(*i);
	  GetAnimations().DeleteObject(*i);
	  delete *i;
	} else {
	  if ((*i)->GetGroup()) (*i)->GetGroup()->DeleteObject(*i);
	  DeleteObject(*i);
	  animations.DeleteObject(*i);
	  delete *i;
	}
      }
    }
    
    if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
    
    SaveAsAOF();
  }

  selectedObjects.clear();
}


MLB::GenericObject* MLB::Page::DeleteObject(GenericObject* obj) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DeleteObject\n");
  #endif

  if (active == obj) active = NULL;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if (*i == obj) {
      if ((*i)->GetGroup()) (*i)->GetGroup()->DeleteObject(*i);
      (*i)->Hide(canvas);
      objects.erase(i);
      return obj;
    }
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if (*j == obj) {
      (*j)->Hide(canvas);
      groups.erase(j);
      return obj;
    }
  }

  return NULL;
}


MLB::GenericObject* MLB::Page::RemoteDeleteObject(DrawObject* obj) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::RemoteDeleteObject\n");
  #endif

  if (active == obj) active = NULL;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if (obj == *i) {
      if (obj->GetGroup()) obj->GetGroup()->DeleteObject(obj);
      if (SessionManager::sm->GetActivePage() == this) obj->Hide(canvas); 
      objects.erase(i);
      return obj;
    }
  }
  
  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if (*j == obj) {
      if (SessionManager::sm->GetActivePage() == this) (*j)->Hide(canvas);
      groups.erase(j);
      return obj;
    }
  }

  return NULL;
}


MLB::GenericObject* MLB::Page::DeleteObject(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DeleteObject %u\n",_id);
  #endif

  GenericObject* obj;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetID() == _id) {
      if ((*i)->GetGroup()) (*i)->GetGroup()->DeleteObject(*i);
      (*i)->Hide(canvas);
      obj = *i;
      if (active == obj) active = NULL;
      objects.erase(i);
      return obj;
    }
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->GetID() == _id) {
      (*j)->Hide(canvas);
      obj = *j;
      if (active == obj) active = NULL;
      groups.erase(j);
      return obj;
    }
  }

  return NULL;
}


MLB::GenericObject* MLB::Page::RemoteDeleteObject(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::DeleteObject %u\n",_id);
  #endif

  GenericObject* obj;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetID() == _id) {
      if ((*i)->GetGroup()) (*i)->GetGroup()->DeleteObject(*i);
      if (SessionManager::sm->GetActivePage() == this) (*i)->Hide(canvas);
      obj = *i;
      if (active == obj) active = NULL;
      objects.erase(i);
      return obj;
    }
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->GetID() == _id) {
      if (SessionManager::sm->GetActivePage() == this) (*j)->Hide(canvas);
      obj = *j;
      if (active == obj) active = NULL;
      groups.erase(j);
      return obj;
    }
  }

  return NULL;
}


void MLB::Page::Clear() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page:Clear\n");
  #endif

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) 
    (*j)->Hide(canvas);

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) 
    (*i)->Hide(canvas);
}


MLB::GenericObject* MLB::Page::DeleteObjectFromContainer(GenericObject* obj) {
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i) == obj) {
      objects.erase(i);
      return obj;
    }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) 
    if ((*j) == obj) {
      groups.erase(j);
      return obj;
    }
  }

  return NULL;
}


void MLB::Page::SaveContainer(FILE *_fp, int _level) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SaveContainer %d\n",_level);
  #endif

  std::string tab;
  for (int k = 0; k <_level; ++k) tab += "  ";

  fprintf(_fp,"%s<PAGE name=\"%s\" width=\"%d\" height=\"%d\" color=\"%s\" thumbnailID=\"%d\">\n",
	  (char*)tab.c_str(), (char*)containerName.c_str(), width, height, (char*)background.c_str(),pageNumber);

  GetAnimations().Save( _fp, _level+1 );

  fprintf(_fp,"  %s<LAYER name=\"\">\n", (char*)tab.c_str());

  // objects
  Group* group;
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) {
      group = (Group*)((*i)->GetGroup());
      if (!group->saved) {
	group->SaveObject(_fp,(_level+2));
	group->saved = true;
      }
    } else {
      (*i)->SaveObject(_fp,(_level+2));
    }
  }

  fprintf(_fp,"  %s</LAYER>\n", (char*)tab.c_str());
  fprintf(_fp,"%s</PAGE>\n", (char*)tab.c_str());

  for (std::list<DrawObject*>::iterator j = objects.begin(); j != objects.end(); ++j) {
    if ((*j)->GetType() == GenericObject::ASCIITEXT) {
      ((Group*)((*j)->GetGroup()))->saved = false;
    }
  }
}


void MLB::Page::SaveAsPs(FILE* _fp) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SaveAsPs\n");
  #endif

  fprintf(_fp,"%%%%Page: %d %d\n",pageNumber,pageNumber);
  fprintf(_fp,"save\n");
  fprintf(_fp,"%.2f %.2f scale\n",1.0,1.0);
  fprintf(_fp,"%d %d translate\n",0,0);

  bool rotate = (width > height) ? true : false;
  for (std::list<MLB::DrawObject *>::iterator iter = objects.begin(); iter != objects.end(); ++iter) {
    (*iter)->SaveAsPs(_fp,width,height,rotate);
  }

  fprintf(_fp,"showpage\n");
}


void MLB::Page::SaveAsImage(char* name, char* type, int& nbr) {
  char tmp[100];
  sprintf(tmp, "%s_%u.%s", name, nbr, type);
  Activate();
  MLB::SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  canvas->SaveAsImage(tmp, type);
  SessionManager::sm->progress->IncrementProgress();
  ++nbr;
}


MLB::GenericObject* MLB::Page::GetObject(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::GetObject %u\n",_id);
  #endif

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetID() == _id) {
      return (*i);
    }
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->GetID() == _id) {
      return (*j);
    }
  }

  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k) {
    if ((*k)->GetID() == _id) {
      return (*k);
    }
  }

  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l) {
    if ((*l)->GetID() == _id) {
      return (*l);
    }
  }

  return NULL;
}


MLB::DrawObject* MLB::Page::GetObject(const string& tkRef) {

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetTkRef() == tkRef) {
      return (*i);
    }
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->GetTkRef() == tkRef) {
      return (*j);
    }
  }

  return NULL;
}


MLB::GenericObject* MLB::Page::GetGroup(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::GetGroup %u\n",_id);
  #endif

  GenericObject* obj;

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->GetID() == _id) {
      return (*j);
    }
  }

  return NULL;
}


uint32 MLB::Page::GetObjectOrder(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::GetObjectOrder %u\n",_id);
  #endif

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetID() == _id) return (*i)->GetOrder();
  }

  return 0; // not found
}


uint32 MLB::Page::GetPushBackObjectOrder() {
  if (objects.empty()) {
    if (lastObjectOrder == 0) lastObjectOrder =  MLB_OO_MIDDLE;
    else lastObjectOrder += MLB_OO_OFFSET;
  } else {
    uint32 order = objects.back()->GetOrder();
    if (order <= lastObjectOrder) lastObjectOrder += MLB_OO_OFFSET;
    else lastObjectOrder = order + MLB_OO_OFFSET;
  }
  return lastObjectOrder;
}


uint32 MLB::Page::GetEndObjectOrder() {
  if (objects.empty()) return 0;
  return (objects.back())->GetOrder();
}


uint32 MLB::Page::GetFirstObjectOrder() {
  if (objects.empty()) return 0;
  return (objects.front())->GetOrder();
}


uint32 MLB::Page::GetNextObjectOrder(uint32 _id) {
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetID() == _id) {
      if (*i == objects.back()) return (*i)->GetOrder() + MLB_OO_OFFSET;
      uint32 order = (*i)->GetOrder() / 2;
      ++i;
      return order + (*i)->GetOrder() / 2;
    }
  }
  return 0;
}


MLB::GenericObject* MLB::Page::GetObjectWithOrder(uint32 order) {
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetOrder() == order) return *i;
    if ((*i)->GetOrder() >  order) return NULL; // not found
  }
  return NULL; // not found
}


MLB::Telepointer* MLB::Page::GetTelepointer(uint32 id) {
  for (std::list<Telepointer*>::iterator i = pointers.begin(); i != pointers.end(); ++i) {
    if ((*i)->GetID() == id) return (Telepointer*)(*i);
  }
  return NULL;
}


MLB::Telepointer* MLB::Page::DeleteTelepointer(uint32 id) {
  for (std::list<Telepointer*>::iterator i = pointers.begin(); i != pointers.end(); ++i) {
    if ((*i)->GetID() == id) {
      Telepointer* tp = (Telepointer*)(*i);
      pointers.erase(i);
      return tp;
    }
  }
  return NULL;
}


void MLB::Page::DeleteTelepointerOfParticipant(uint32 pid) {
  #ifdef _DEBUG1
  cout << "MLB::Page::DeleteTelepointer" << _participantID << endl;
  #endif

  for (std::list<Telepointer*>::iterator i = pointers.begin(); i != pointers.end(); ++i) {
    if ((*i)->GetParticipantID() == pid) {
      (*i)->SetInactive(this);
      delete *i;
      pointers.erase(i);
      return;
    }
  }
}


void MLB::Page::AddTelepointer(MLB::Telepointer* tp) {
  if (GetTelepointer(tp->GetID())) return; // already here
  pointers.push_back(tp);
}


MLB::Telebox* MLB::Page::GetTelebox(uint32 id) {
  for (std::list<Telebox*>::iterator i = boxes.begin(); i != boxes.end(); ++i) {
    if ((*i)->GetID() == id) return (Telebox*)(*i);
  }
  return NULL;
}


MLB::Telebox* MLB::Page::DeleteTelebox(uint32 id) {
  for (std::list<Telebox*>::iterator i = boxes.begin(); i != boxes.end(); ++i) {
    if ((*i)->GetID() == id) {
      Telebox* b = (Telebox*)(*i);
      boxes.erase(i);
      return b;
    }
  }
  return NULL;
}


void MLB::Page::DeleteTeleboxOfParticipant(uint32 pid) {
  #ifdef _DEBUG1
  cout << "MLB::Page::DeleteTelebox" << _participantID << endl;
  #endif

  for (std::list<Telebox*>::iterator i = boxes.begin(); i != boxes.end(); ++i) {
    if ((*i)->GetParticipantID() == pid) {
      (*i)->SetInactive(this);
      delete *i;
      boxes.erase(i);
      return;
    }
  }
}


void MLB::Page::AddTelebox(MLB::Telebox* b) {
  if (GetTelebox(b->GetID())) return; // already here
  boxes.push_back(b);
}


void MLB::Page::SetPageDimension(int _width, int _height) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SetDimension %u %u\n",_width,_height);
  #endif

  width  = _width;
  height = _height;
  canvas->SetDimension(SessionManager::sm->GetZoom()*width,SessionManager::sm->GetZoom()*height);

  if (SessionManager::sm->GetActivePage() == this) Draw();
}


void MLB::Page::SetBackgroundColor(Color _color) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SetBackgroundColor %s\n",_color);
  #endif

  background = _color;

  canvas->SetBackground(background.c_str());
}


void MLB::Page::SetWindowPointer(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SetWindowPointer\n");
  #endif

  const char* name  = argv[1];
  const char* color = argv[2];

  canvas->SetCursor(name,color);
  if (strcmp(name,"point") != 0)
    onlinePointer->Set(name,canvas->cursor->GetXHot(),canvas->cursor->GetYHot());
}


void MLB::Page::UnSetWindowPointer(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::UnSetWindowPointer\n");
  #endif

  onlinePointer->SetInactive(this);
  canvas->ResetCursor();
  if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
    int bytes = onlinePointer->FlushChangeVisibility(TelepointerManager::sendBuffer);
    TelepointerManager::tm->SendEvent(onlinePointer->GetID(), TelepointerManager::sendBuffer, bytes);
  }
}


MLB::DrawObject* MLB::Page::GetActiveObject() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::GetActiveObject\n");
  #endif

  if (objects.empty()) return NULL;

  return active;
}


MLB::PhotoImage* MLB::Page::GetFirstImage() {
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i)
    if ((*i)->GetType() == GenericObject::IMAGE) return ((PhotoImage*)*i);
  return NULL;
}


MLB::GenericContainer* MLB::Page::GetContainer(uint32 _id) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::GetContainer %u\n",_id);
  #endif

  // check object list
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i)
    if ((*i)->GetID() == _id) return this;

  // check group list
  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j)
    if ((*j)->GetID() == _id) return this;

  // check telepointer list
  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k)
    if ((*k)->GetID() == _id) return this;

  // check telebox list
  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l)
    if ((*l)->GetID() == _id) return this;

  return NULL;
}


void MLB::Page::AppendObject(DrawObject *obj) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::AppendObject\n");
  #endif

  if (GetObject(obj->GetID()) || GetGroup(obj->GetID())) return; // already here

  if (obj->GetType() == GenericObject::GROUP) {
    groups.push_back((Group*)obj);
    return;
  }

  if (objects.empty()) {
    objects.push_back(obj);
    return;
  }

  std::list<DrawObject*>::iterator pos = objects.begin();
  while (pos != objects.end() && (obj->GetOrder() >= (*pos)->GetOrder())) ++pos;
  objects.insert(pos, obj);
}


void MLB::Page::AppendNewObject(int argc, char* argv[], int _type) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::AppendNewObject\n");
  #endif

  static int firstX;
  static int firstY;
  static int count = 1;
  static double test;
  static bool annotation = false;
  static DrawObject* obj = NULL;
  static uint32 objID = 0; // 0 is reserved as id for the session; real objects start with id 2

  char
    *color,
    *fill,
    *lineStyle,
    *capStyle,
    *joinStyle,
    *arrowStyle,
    *arrowShape,
    *family,
    *size,
    *weight,
    *slant,
    *position;

  bool smooth;
  int linewidth;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;

  UnSelectObjects();

  int status = atoi(argv[1]);
  int x      = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y      = atoi(argv[3]) / SessionManager::sm->GetZoom();


  lineStyle = NULL;
  switch (status) {
  case MLB::Action::BEGIN : {
    DeleteEmptyTextObjects();
    annotation = SessionManager::sm->AnnotationMode();
    color = argv[4];
    switch (_type) {
    case MLB::GenericObject::RECTANGLE : {
      fill      = strdup(argv[5]);
      linewidth = atoi(argv[6]);
      lineStyle = argv[7];

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(Rectangle::GetStateSizeS());
	Rectangle::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, 0, 0, color, fill, linewidth, lineStyle);
	objID = CommunicationManager::cm->GetObjectID();
	adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::RECTANGLE, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new Rectangle(x, y , 0, 0, color, fill, linewidth, lineStyle, objID = CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	AppendObject(obj);
	obj->Draw(canvas);
	active = obj;
      }

      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      Rectangle::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      delete[] fill;
      break;
    }
    case MLB::GenericObject::OVAL : {
      fill      = strdup(argv[5]);
      linewidth = atoi(argv[6]);
      lineStyle = argv[7];

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(Oval::GetStateSizeS());
	Oval::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, 0, 0, color, fill, linewidth, lineStyle);
	objID = CommunicationManager::cm->GetObjectID();
	adu = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::RECTANGLE, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new Oval(x, y, 0, 0, color, fill, linewidth, lineStyle, objID = CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	AppendObject(obj);
	obj->Draw(canvas);
	active = obj;
      }

      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      Oval::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      delete[] fill;
      break;
    }
    case MLB::GenericObject::FREEHAND : {
      linewidth  = atoi(argv[5]);
      lineStyle  = argv[6];
      capStyle   = argv[7];
      joinStyle  = argv[8];
      arrowStyle = argv[9];
      arrowShape = argv[10];
      if (strncmp(argv[11],"smooth",6)==0)
	smooth = true;
      else
	smooth = false;

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(PolyLine::GetStateSizeS());
	PolyLine::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, color, linewidth, lineStyle, false,
			     capStyle, joinStyle, arrowStyle, arrowShape);
	objID = CommunicationManager::cm->GetObjectID();
	adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new PolyLine(x, y ,0 , 0, color, linewidth, lineStyle, capStyle, joinStyle, arrowStyle, arrowShape, smooth,
            			         objID = CommunicationManager::cm->GetObjectID(),true);
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	((PolyLine*)obj)->AppendNewPoint(x,y);
	SavePointAsAOF((PolyObject *)obj,x,y,true);
	AppendObject(obj);
	obj->Draw(canvas);
	active = obj;
      }
      
      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      PolyLine::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      break;
    }
    case MLB::GenericObject::POLYLINE : {
      linewidth  = atoi(argv[5]);
      lineStyle  = argv[6];
      capStyle   = argv[7];
      joinStyle  = argv[8];
      arrowStyle = argv[9];
      arrowShape = argv[10];
      if (strncmp(argv[11],"smooth",6)==0)
	smooth = true;
      else
	smooth = false;

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(PolyLine::GetStateSizeS());
	PolyLine::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, color, linewidth, lineStyle, false,
			     capStyle, joinStyle, arrowStyle, arrowShape);
	objID = CommunicationManager::cm->GetObjectID();
	adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new PolyLine(x, y ,0 , 0, color, linewidth, lineStyle, capStyle, joinStyle, arrowStyle, arrowShape, smooth,
			         objID = CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	((PolyLine*)obj)->AppendNewPoint(x,y);
	SavePointAsAOF((PolyObject *)obj,x,y,true);
	AppendObject(obj);
	obj->Draw(canvas);
	canvas->BeginHelpLine(SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);
	active = obj;
      }

      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      PolyLine::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      break;
    }
    case MLB::GenericObject::POLYGON : {
      fill      = strdup(argv[5]);
      linewidth = atoi(argv[6]);
      lineStyle = argv[7];

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(Polygon::GetStateSizeS());
	Polygon::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, color, fill, linewidth, lineStyle);
	objID = CommunicationManager::cm->GetObjectID();
	adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYGON, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new Polygon(x, y, 0, 0, color, fill, linewidth, lineStyle, objID = CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	((Polygon*)obj)->AppendNewPoint(x,y);
	SavePointAsAOF((PolyObject *)obj,x,y,true);
	AppendObject(obj);
	obj->Draw(canvas);
	canvas->BeginHelpLine(SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);
	active = obj;
      }
      
      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      Polygon::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      delete[] fill;
      break;
    }
    case MLB::GenericObject::LINE : {
      linewidth  = atoi(argv[5]);
      lineStyle  = argv[6];
      capStyle   = argv[7];
      joinStyle  = argv[8];
      arrowStyle = argv[9];
      arrowShape = argv[10];

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer(Line::GetStateSizeS());
	Line::FlushState(buf->getCont(), GetPushBackObjectOrder(), true, x, y, 0, 0, color, linewidth, lineStyle,
			 capStyle, joinStyle, arrowStyle, arrowShape);
	objID = CommunicationManager::cm->GetObjectID();
	adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::LINE, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
	AduManager::am->InsertAdu(adu);
      }
      else { // PRIVATE session or Annotation Mode
	obj = new Line(x, y, 0, 0, color, linewidth, lineStyle, capStyle, joinStyle, arrowStyle, arrowShape,
		                   objID = CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	AppendObject(obj);
	obj->Draw(canvas);
	active = obj;
      }

      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      Line::FlushDeleteS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      break;
    }
    case MLB::GenericObject::TELEPOINTER: 
    case MLB::GenericObject::TELEBOX: {
      return;
    }
    case MLB::GenericObject::ASCIITEXT : {
      family   = argv[5];
      size     = argv[6];
      weight   = argv[7];
      slant    = argv[8];
      position = argv[9];

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	buf   = new Buffer( Group::GetStateSizeTG(family));
	Group::FlushStateTG(buf->getCont(), GetPushBackObjectOrder(), true, x, y, 0, 0, color, family, size, weight, slant, position);
	objID = CommunicationManager::cm->GetObjectID();
	adu = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::TEXT, 3, buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	buf   = new Buffer(GetChangeActiveItemSize());
	FlushChangeActiveItem(buf->getCont());
	AduManager::am->InsertAdu(CommunicationManager::cm->CreateEvent(GetID(), objID, buf)); // only local change
      }
      else { // PRIVATE session or Annotation Mode
	Group* group = new Group(objID = CommunicationManager::cm->GetObjectID());
	group->SetAnnotation(annotation);
	group->SetOrder(GetPushBackObjectOrder());
	AppendObject(group);

	obj = new Text(x, y, 0, 0, color, NULL, family, size, weight, slant, position, CommunicationManager::cm->GetObjectID());
	obj->SetAnnotation(annotation);
        if (annotation) obj->SetLayer(2);
	obj->SetOrder(GetPushBackObjectOrder());
	AppendObject(obj);
	active = obj;
	group->AddObject(obj);

	group->Draw(canvas);
	group->DrawTextCursor(GetCanvas());
      }

      undoBuf = new Buffer(GenericObject::GetDeleteSize());  
      Group::FlushDeleteTGS(undoBuf->getCont());
      undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu); 
      
      firstX = x;
      firstY = y;

      return;
    }
    }

    firstX = x;
    firstY = y;

    return;
  }
  case MLB::Action::CONTINUE : {
    switch (_type) {
    case MLB::GenericObject::RECTANGLE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(Rectangle::SGetChangeSizeSize());
	  Rectangle::SFlushChangeSize(buf->getCont(), firstX, firstY, x-firstX, y-firstY);
	  adu   = CommunicationManager::cm->CreateCue(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  test  = CommunicationManager::cueSendRate * count;
	  if (floor(test) == ceil(test)) {
	    CommunicationManager::cm->SendAduUnreliable(adu);
	  }
	  if (count == 100) count = 1;
	  else ++count;
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj) obj->ModifyCoords(canvas,firstX,firstY,x-firstX,y-firstY);
      }
      return;
    }
    case MLB::GenericObject::OVAL : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(Oval::SGetChangeSizeSize());
	  Oval::SFlushChangeSize(buf->getCont(), firstX, firstY, x-firstX, y-firstY);
	  adu   = CommunicationManager::cm->CreateCue(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  test  = CommunicationManager::cueSendRate * count;
	  if (floor(test) == ceil(test)) {
	    CommunicationManager::cm->SendAduUnreliable(adu);
	  }
	  if (count == 100) count = 1;
	  else ++count;
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj)
	  obj->ModifyCoords(canvas,firstX,firstY,x-firstX,y-firstY);
      }
      return;
    }
    case MLB::GenericObject::LINE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(Line::GetMovePointSize());
	  Line::FlushMovePoint(buf->getCont(), 2, x, y);
	  adu   = CommunicationManager::cm->CreateCue(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  test  = CommunicationManager::cueSendRate * count;
	  if (floor(test) == ceil(test)) {
	    CommunicationManager::cm->SendAduUnreliable(adu);
	  }
	  if (count == 100) count = 1;
	  else ++count;
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj)
	  obj->ModifyCoords(canvas,firstX,firstY,x-firstX,y-firstY);
      }
      return;
    }
    case MLB::GenericObject::FREEHAND : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(PolyLine::GetAddPointSize());
	  PolyLine::FlushAddPoint(buf->getCont(), x, y);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  if (obj) // Multi-layer recording
	    SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj) {
	  ((PolyLine*)obj)->AppendNewPoint(x,y);
	  ((PolyLine*)obj)->DrawNewSegment(canvas);
	  SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }

      firstX = x;
      firstY = y;

      return;
    }
    case MLB::GenericObject::POLYLINE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(PolyLine::GetAddPointSize());
	  PolyLine::FlushAddPoint(buf->getCont(), x, y);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  if (obj) // Multi-layer recording
	    SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj) {
	  ((PolyLine*)obj)->AppendNewPoint(x,y);
	  ((PolyLine*)obj)->DrawNewSegment(canvas);
	  SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }

      firstX = x;
      firstY = y;

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE || annotation) { 
	canvas->ModifyHelpLine(SessionManager::sm->GetZoom()*firstX,SessionManager::sm->GetZoom()*firstY,
			       SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);
      }
      return;
    }
    case MLB::GenericObject::POLYGON : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(Polygon::GetAddPointSize());
	  Polygon::FlushAddPoint(buf->getCont(), x, y);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	  if (obj) // Multi-layer recording
	    SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }
      else { // PRIVATE session or Annotation Mode
	if (obj) {
	  ((Polygon*)obj)->AppendNewPoint(x,y);
	  ((Polygon*)obj)->DrawNewSegment(canvas);
	  SavePointAsAOF((PolyObject *)obj,x,y);
	}
      }

      firstX = x;
      firstY = y;

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE || annotation) { 
	canvas->ModifyHelpLine(SessionManager::sm->GetZoom()*firstX,SessionManager::sm->GetZoom()*firstY,
			       SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);
      }
      return;
    }
    }
    return;
  }
  case MLB::Action::SHOWLINE : {
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE || annotation) { 
      canvas->ModifyHelpLine(SessionManager::sm->GetZoom()*firstX,SessionManager::sm->GetZoom()*firstY,
			     SessionManager::sm->GetZoom()*x,SessionManager::sm->GetZoom()*y);
    }
    return;
  }
  case MLB::Action::END : {
    switch (_type) {
    case MLB::GenericObject::RECTANGLE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
      if (objID) {
	  buf   = new Buffer(Rectangle::SGetChangeSizeSize());
	  Rectangle::SFlushChangeSize(buf->getCont(), firstX, firstY, x-firstX, y-firstY);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      }
      break;
    }
    case MLB::GenericObject::OVAL : {
	    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
      if (objID) {
	  buf   = new Buffer(Oval::SGetChangeSizeSize());
	  Oval::SFlushChangeSize(buf->getCont(), firstX, firstY, x-firstX, y-firstY);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      }
      break;
    }
    case MLB::GenericObject::LINE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
      if (objID) {
	  buf   = new Buffer(Line::GetMovePointSize());
	  Line::FlushMovePoint(buf->getCont(), 2, x, y);
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      }
      break;
    }
    case MLB::GenericObject::FREEHAND : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	// additional state at the end of a drawing operation can be used by the recorder as starting 
	// point in the case of random access -> more efficient than replay of event sequence
//  	if (objID) {
//  	  GenericObject* tmp = GetObject(objID);
//  	  if (tmp) { 
//  	    buf   = new Buffer(tmp->GetStateSize());
//  	    tmp->FlushState(buf->getCont());
//  	    adu   = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::POLYLINE, 0, buf, rtpitime());
//  	    CommunicationManager::cm->SendAduReliable(adu);
//  	  }
//  	}
      }
      else { // PRIVATE session or Annotation Mode
        if (obj) ((PolyObject *)obj)->ReDraw(canvas);
      }
      break;
    }
    case MLB::GenericObject::POLYLINE : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() != Session::SHARED || annotation) {
        if (obj) ((PolyObject *)obj)->ReDraw(canvas);
	  canvas->DeleteHelpLine();
	}
      break;
    }
    case MLB::GenericObject::POLYGON : {
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
	if (objID) {
	  buf   = new Buffer(Polygon::GetClosePolygonSize());
	  Polygon::FlushClosePolygon(buf->getCont());
	  adu   = CommunicationManager::cm->CreateEvent(objID, buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
      } else {
	if (obj) ((PolyObject *)obj)->ReDraw(canvas);
	  canvas->DeleteHelpLine();
	}
      break;
    }
    case MLB::GenericObject::TELEBOX:
    case MLB::GenericObject::TELEPOINTER: {
      break;
    }
    }
    obj = NULL;

    SaveAsAOF();

    objID = 0;
    break;
  }
  }
}


void MLB::Page::SelectObject(DrawObject *obj) {
  if (obj) {
    selectedObjects.push_back(obj);
    obj->AdjustGui();
  }
} 


void MLB::Page::SelectObjects(int _x, int _y, int _status) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SelectObject %d %d %d\n",_x,_y,_status);
  #endif

  if (objects.empty()) return;

  static bool edit = false;
  static int firstX, firstY;
  static int count = 1;
  static double test;

  DrawObject* obj;
  std::list<DrawObject*>::iterator i;

  int newX, newY;
  int newWidth, newHeight;
  int x1, y1, x2, y2;

  if (_status == Action::BEGIN) {
    edit = false;
    if (active && active->GetType() == GenericObject::ASCIITEXT)
      ((Text*)active)->DeleteTextCursor(canvas);

    if (!selectedObjects.empty()) {
      for (i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
	if (active = (*i)->PointIsInText(_x,_y)) {
          if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !active->Annotation()) {
              RefCount<Buffer> buf        = new Buffer(((Group*)active)->GetChangeActiveSizeTG());
	            ((Group*)active)->FlushSetActiveTG(buf->getCont(), _x, _y);
	            RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	            AduManager::am->InsertAdu(adu);
	            CommunicationManager::cm->SendAduReliable(adu);
          }  
	  edit = true;
	  active->AdjustGui();
	  break;
	}
	if ((*i)->IsEditablePoint(_x,_y)) {
	  active = (*i);
	  edit = true;
	  break;
	}
      }
    }

    firstX = _x;
    firstY = _y;
    if (!edit) {
      UnSelectObjects();
      canvas->BeginSelectRegion(SessionManager::sm->GetZoom()*_x,SessionManager::sm->GetZoom()*_y);
    }
  }
  else {
    RefCount<Buffer> undoBuf;
    RefCount<RTPI::RtpiAdu> undoAdu;
    int undoX, undoY, undoWidth, undoHeight;
    
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) {
      RefCount<Buffer> buf;
      RefCount<RTPI::RtpiAdu> adu;
      switch (_status) {
      case MLB::Action::CONTINUE : {
	// change size of selected object
	if (edit && active) {
	  switch (active->GetType()) {
	  case MLB::GenericObject::RECTANGLE:
	  case MLB::GenericObject::OVAL: {
	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    if (!active->Annotation()) {
	      buf   = new Buffer(active->GetChangeSizeSize());
	      active->FlushChangeSize(buf->getCont(), newX, newY, newWidth, newHeight);
	      adu   = CommunicationManager::cm->CreateCue(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      test  = CommunicationManager::cueSendRate * count;
	      if (floor(test) == ceil(test)) {
		CommunicationManager::cm->SendAduUnreliable(adu);
	      }
	      if (count == 100) count = 1;
	      else ++count;
	    }
	    else { // Annotation
	      active->ModifyCoords(canvas, newX, newY, newWidth, newHeight);
	    }
	    break;
	  }
	  case MLB::GenericObject::LINE: {
	    if (!active->Annotation()) {
	      buf   = new Buffer(((Line*)active)->GetMovePointSize());
	      ((Line*)active)->FlushMovePoint(buf->getCont(), active->GetSelectedPoint() + 1, _x, _y);
	      adu   = CommunicationManager::cm->CreateCue(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      test = CommunicationManager::cueSendRate * count;
	      if (floor(test) == ceil(test)) {
		CommunicationManager::cm->SendAduUnreliable(adu);
	      }
	      if (count == 100) count = 1;
	      else ++count;
	    }
	    else { // Annotation
	      ((Line*)active)->MovePoint(canvas, active->GetSelectedPoint() + 1, _x, _y);
	      //active->ModifyCoords(canvas, newX, newY, newWidth, newHeight);
	    }
	    break;
	  }
	  case MLB::GenericObject::FREEHAND:
	  case MLB::GenericObject::POLYLINE:
	  case MLB::GenericObject::POLYGON: {
	    if (!active->Annotation()) {
	      buf   = new Buffer(((PolyObject*)active)->GetMovePointSize());
	      ((PolyObject*)active)->FlushMovePoint(buf->getCont(), active->GetSelectedPoint(), _x, _y);
	      adu   = CommunicationManager::cm->CreateCue(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      test  = CommunicationManager::cueSendRate * count;
	      if (floor(test) == ceil(test)) {
		CommunicationManager::cm->SendAduUnreliable(adu);
	      }
	      if (count == 100) count = 1;
	      else ++count;
	    }
	    else { // Annotation
	      ((PolyObject*)active)->MovePoint(canvas, active->GetSelectedPoint(), _x, _y);
	    }
	    break;
	  }
	  case MLB::GenericObject::IMAGE: {
	    if (SessionManager::sm->FreezeImages()) return; // do not resize images 

	    // repaint only slelection rectangle; rendering is too costly and done at the end of the operation
	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    active->DrawSelection(canvas);
	    break;
	  }
//MB
// Telepointer kann nicht selektiert werden
/*	  case MLB::GenericObject::TELEPOINTER: {
	    ((Telepointer*)active)->Update(_x, _y);
	    active->ReDraw(canvas);
	    if (!active->Annotation()) {
	      test = CommunicationManager::cueSendRate * count;
	      if (floor(test) == ceil(test)) {
		int bytes = ((Telepointer*)active)->FlushMove(TelepointerManager::sendBuffer);
		TelepointerManager::tm->SendCue(active->GetID(), TelepointerManager::sendBuffer, bytes);
	      }
	      if (count == 100) count = 1;
	      else            ++count;
	    }
	    break;
	  } */
	  case MLB::GenericObject::ASCIITEXT: {
	    // nothing to be done
	    break;
	  }
	  }
	}
	else {
	  canvas->ModifySelectRegion(SessionManager::sm->GetZoom()*firstX,SessionManager::sm->GetZoom()*firstY,
				     SessionManager::sm->GetZoom()*_x,SessionManager::sm->GetZoom()*_y);
	}
	break;
      }
      case MLB::Action::END : {
	if (edit && active) {
	  switch (active->GetType()) {
	  case MLB::GenericObject::RECTANGLE:
	  case MLB::GenericObject::OVAL: {
            undoBuf = new Buffer(active->GetChangeSizeSize());
	    active->ComputeCoords(firstX, firstY, undoX, undoY, undoWidth, undoHeight);
	    active->FlushChangeSize(undoBuf->getCont(), undoX, undoY, undoWidth, undoHeight);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);
            
	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    if (!active->Annotation()) {
	      buf   = new Buffer(active->GetChangeSizeSize());
	      active->FlushChangeSize(buf->getCont(), newX, newY, newWidth, newHeight);
	      adu   = CommunicationManager::cm->CreateEvent(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // Annotation
	      active->ModifyCoords(canvas, newX, newY, newWidth, newHeight);
	      active->DrawSelection(canvas);
	    }
	    break;
	  }
	  case MLB::GenericObject::LINE: {
            undoBuf = new Buffer(((Line*)active)->GetMovePointSize());
	    ((Line*)active)->FlushMovePoint(undoBuf->getCont(), active->GetSelectedPoint() + 1, firstX, firstY);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    if (!active->Annotation()) {
	      buf   = new Buffer(((Line*)active)->GetMovePointSize());
	      ((Line*)active)->FlushMovePoint(buf->getCont(), active->GetSelectedPoint() + 1, _x, _y);
	      adu   = CommunicationManager::cm->CreateEvent(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // Annotation
	      ((Line*)active)->MovePoint(canvas,  active->GetSelectedPoint() + 1, _x, _y);
	      active->DrawSelection(canvas);
	    }
	    break;
	  }
	  case MLB::GenericObject::FREEHAND:
	  case MLB::GenericObject::POLYLINE:
	  case MLB::GenericObject::POLYGON: {
            undoBuf = new Buffer(((PolyObject*)active)->GetMovePointSize());
	    ((PolyObject*)active)->FlushMovePoint(undoBuf->getCont(), active->GetSelectedPoint(), firstX, firstY);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    if (!active->Annotation()) {
	      buf   = new Buffer(((PolyObject*)active)->GetMovePointSize());
	      ((PolyObject*)active)->FlushMovePoint(buf->getCont(), active->GetSelectedPoint(), _x, _y);
	      adu   = CommunicationManager::cm->CreateEvent(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { //Annotation
	      ((PolyObject*)active)->MovePoint(canvas, active->GetSelectedPoint(), _x, _y);
	      active->DrawSelection(canvas);
	    }
	    break;
	  }
	  case MLB::GenericObject::IMAGE: {
	          if (SessionManager::sm->FreezeImages()) return; // do not resize images 

            undoBuf = new Buffer(active->GetChangeSizeSize());
	    active->ComputeCoords(firstX, firstY, undoX, undoY, undoWidth, undoHeight);
	    active->FlushChangeSize(undoBuf->getCont(), undoX, undoY, undoWidth, undoHeight);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    if (!active->Annotation()) {
	      buf   = new Buffer(active->GetChangeSizeSize());
	      active->FlushChangeSize(buf->getCont(), newX, newY, newWidth, newHeight);
	      adu   = CommunicationManager::cm->CreateEvent(active->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // Annotation
	      ((PhotoImage*)active)->Update(canvas);
	      active->DrawSelection(canvas);
	    }
	    break;
	  }
//MB
// Telepointer kann nicht selektiert werden
/*    case MLB::GenericObject::TELEPOINTER: {
	    ((Telepointer*)active)->Update(_x, _y);
	    active->ReDraw(canvas);
	    if (!active->Annotation()) {
	      int bytes = ((Telepointer*)active)->FlushMove(TelepointerManager::sendBuffer);
	      TelepointerManager::tm->SendEvent(active->GetID(), TelepointerManager::sendBuffer, bytes);
	    }
	    break;
	  }*/
	  case MLB::GenericObject::ASCIITEXT: {
	    ((Text *)active)->DrawTextCursor(canvas);
	    break;
	  }
	  }
	  if (active->GetType() != MLB::GenericObject::ASCIITEXT) {
	    active = NULL;
	  }
	}
	else {
	  x1 = MIN(_x,firstX);
	  x2 = MAX(_x,firstX);
	  y1 = MIN(_y,firstY);
	  y2 = MAX(_y,firstY);
	  canvas->DeleteSelectRegion();
	  for (i = objects.begin(); i != objects.end(); ++i) {
	    if ((*i)->IsInRegion(x1,y1,x2,y2)) {
	      if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
	      obj = (*i)->Select();
	      if (obj) 
		if (!DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
	    }
	  }
	  if (selectedObjects.empty()) {
	    i = objects.end();
	    do {
	      --i;
	      if ((*i)->Overlap(canvas,_x,_y)) {
		if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
		obj = (*i)->Select();
		if (obj) {
		  if (!DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
		  obj->AdjustGui();
		}
		break;
	      }
	    } while (i != objects.begin());
	  }

	  DrawSelected();

	}
	break;
      }
      }
    }
    else { // PRIVATE
      switch (_status) {
      case MLB::Action::CONTINUE : {
	if (edit && active) {
	  switch (active->GetType()) {
	  case MLB::GenericObject::RECTANGLE:
	  case MLB::GenericObject::OVAL: {
	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    active->ModifyCoords(canvas, newX, newY, newWidth, newHeight);
	    break;
 	  }
	  case MLB::GenericObject::LINE: {
	    ((Line*)active)->MovePoint(canvas, active->GetSelectedPoint() + 1, _x, _y);
	    break;
	  }
	  case MLB::GenericObject::FREEHAND:
	  case MLB::GenericObject::POLYLINE:
	  case MLB::GenericObject::POLYGON: {
	    ((PolyObject*)active)->MovePoint(canvas, active->GetSelectedPoint(), _x, _y);
	    break;
	  }
	  case MLB::GenericObject::IMAGE: {
	    if (SessionManager::sm->FreezeImages()) return; // do not resize images 

	    // repaint only slelection rectangle; rendering is too costly and done at the end of the operation
	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    active->DrawSelection(canvas);
	    break;
	  }
	  case MLB::GenericObject::ASCIITEXT: {
	    // nothing to be done
	    break;
	  }
	}
	}
	else {
	  canvas->ModifySelectRegion(SessionManager::sm->GetZoom()*firstX,SessionManager::sm->GetZoom()*firstY,
				     SessionManager::sm->GetZoom()*_x,SessionManager::sm->GetZoom()*_y);
	}
	break;
      }
      case MLB::Action::END : {
	if (edit && active) {
	  switch (active->GetType()) {
	  case MLB::GenericObject::RECTANGLE:
	  case MLB::GenericObject::OVAL: {
            undoBuf = new Buffer(active->GetChangeSizeSize());
	    active->ComputeCoords(firstX, firstY, undoX, undoY, undoWidth, undoHeight);
	    active->FlushChangeSize(undoBuf->getCont(), undoX, undoY, undoWidth, undoHeight);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    active->ModifyCoords(canvas, newX, newY, newWidth, newHeight);
	    active->DrawSelection(canvas);
	    break;
	  }
	  case MLB::GenericObject::LINE: {
            undoBuf = new Buffer(((Line*)active)->GetMovePointSize());
	    ((Line*)active)->FlushMovePoint(undoBuf->getCont(), active->GetSelectedPoint() + 1, firstX, firstY);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    ((Line*)active)->MovePoint(canvas,  active->GetSelectedPoint() + 1, _x, _y);
	    active->DrawSelection(canvas);
	    break;
	  }
	  case MLB::GenericObject::FREEHAND:
	  case MLB::GenericObject::POLYLINE:
	  case MLB::GenericObject::POLYGON: {
            undoBuf = new Buffer(((PolyObject*)active)->GetMovePointSize());
	    ((PolyObject*)active)->FlushMovePoint(undoBuf->getCont(), active->GetSelectedPoint(), firstX, firstY);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    ((PolyObject*)active)->MovePoint(canvas,  active->GetSelectedPoint(), _x, _y);
	    active->DrawSelection(canvas);
	    break;
	  }
	  case MLB::GenericObject::IMAGE: {
	    if (SessionManager::sm->FreezeImages()) return; // do not resize images 

            undoBuf = new Buffer(active->GetChangeSizeSize());
	    active->ComputeCoords(firstX, firstY, undoX, undoY, undoWidth, undoHeight);
	    active->FlushChangeSize(undoBuf->getCont(), undoX, undoY, undoWidth, undoHeight);
	    undoAdu = CommunicationManager::cm->CreateEvent(active->GetID(), GetID(), undoBuf);
            UndoManager::um->InsertAdu(undoAdu);

	    active->ComputeCoords(_x, _y, newX, newY, newWidth, newHeight);
	    ((PhotoImage*)active)->Update(canvas);
	    active->DrawSelection(canvas);
	    break;
	  }
	  case MLB::GenericObject::ASCIITEXT: {
	    ((Text *)active)->DrawTextCursor(canvas);
	    break;
	  }
	  }
	  if (active->GetType() != MLB::GenericObject::ASCIITEXT) {
	    active = NULL;
	  }
	}
	else {
	  x1 = MIN(_x,firstX);
	  x2 = MAX(_x,firstX);
	  y1 = MIN(_y,firstY);
	  y2 = MAX(_y,firstY);
	  canvas->DeleteSelectRegion();
	  for (i = objects.begin(); i != objects.end(); ++i) {
	    if ((*i)->IsInRegion(x1,y1,x2,y2)) {
	      if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
	      obj = (*i)->Select();
	      if (obj)
		if (!DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
	    }
	  }
	  if (selectedObjects.empty()) {
	    i = objects.end();
	    do {
	      --i;
	      if ((*i)->Overlap(canvas,_x,_y)) {
		if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
		obj = (*i)->Select();
		if (obj) {
		  if (!DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
		  obj->AdjustGui();
		}
		break;
	      }
	    } while (i != objects.begin());
	  }

	  DrawSelected();
	}
	break;
      }
      }
    }
  }
}


void MLB::Page::SelectObjects(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SelectObject\n");
  #endif

  int status = atoi(argv[1]);
  int x      = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y      = atoi(argv[3]) / SessionManager::sm->GetZoom();

  SelectObjects(x,y,status);
}


void MLB::Page::UnSelectObjects() {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::UnSelectObject\n");
  #endif

  if (active && (active->GetType() == GenericObject::ASCIITEXT)) ((Text*)active)->DeleteTextCursor(canvas);
  active = NULL;

  for (std::list<DrawObject*>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i)
    (*i)->DeleteSelection(canvas);
  selectedObjects.clear();
}


void MLB::Page::ChangeObject(int argc, const char* argv[], int _type) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeObject\n");
  #endif

  if (active) {
    Group* group = (Group*)active->GetGroup();
    if (group) {
      group->Change(this, argc, (char**)argv, active, _type);
      //active->ReDraw(canvas);
    }
    else {
      active->Change(this, argc, (char**)argv, _type);
      active->ReDraw(canvas);
    }

    SaveAsAOF();
  }
}


void MLB::Page::ChangeFont(int argc, char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeFont\n");
  #endif

  Group
    *group;
  char
    *family,
    *size,
    *weight,
    *slant,
    *position;

  if (objects.empty()) return;

  family   = argv[1];
  size     = argv[2];
  weight   = argv[3];
  slant    = argv[4];
  position = argv[5];

  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  if (active && (active->GetType() == GenericObject::ASCIITEXT)) { // 1. insert new text element into active text object
    group = (Group *)active->GetGroup();
    if (!group) return; // error

    Text* txt = (Text*)group->GetActiveItem();
    if (!txt) return;
    undoBuf   = new Buffer(group->GetChangeFontSizeTG(txt->GetFontFamily()));
    group->FlushInsertNewFontTG(undoBuf->getCont(), 
				txt->GetColor(), 
				txt->GetFontFamily(), 
				txt->GetFontSize(), 
				txt->GetFontWeight(), 
				txt->GetFontSlant(), 
				txt->GetTextPosition());
    undoAdu   = CommunicationManager::cm->CreateEvent(group->GetID(), GetID(), undoBuf);
    UndoManager::um->InsertAdu(undoAdu);
    
    group->InsertNewFont(this,family,size,weight,slant,position);
    SaveAsAOF();
  }
  else { // 2. change font of whole selected textgroup
    if (!selectedObjects.empty()) {
      for (std::list<DrawObject *>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
	if ((*i)->GetType() == GenericObject::GROUP) {

          Text* txt = (Text*)((Group*)(*i))->GetFirst();
          if (!txt) return;
          undoBuf   = new Buffer(((Group*)(*i))->GetChangeFontSizeTG(txt->GetFontFamily()));
          ((Group*)(*i))->FlushChangeFontTG(undoBuf->getCont(), 
					    txt->GetColor(), 
					    txt->GetFontFamily(), 
					    txt->GetFontSize(), 
					    txt->GetFontWeight(), 
					    txt->GetFontSlant());
          undoAdu   = CommunicationManager::cm->CreateEvent(((Group*)(*i))->GetID(), GetID(), undoBuf);
          aduList->push_back(undoAdu);

	  ((Group *)(*i))->ChangeFont(this,family,size,weight,slant);
	}
      }
      if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

      SaveAsAOF();
    }
  }
}


void MLB::Page::ChangeLineColor(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeLineColor\n");
  #endif

  if (objects.empty()) return;

  char* color = (char*)argv[1];

  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;

  if (active && (active->GetType() == GenericObject::ASCIITEXT)) {
    if (active->GetGroup()) {
      Text* txt = (Text*)((Group*)active->GetGroup())->GetActiveItem();
      if (!txt) return;
      undoBuf   = new Buffer(((Group*)active->GetGroup())->GetChangeFontColorSizeTG());
      ((Group*)active->GetGroup())->FlushInsertNewFontColorTG(undoBuf->getCont(), txt->GetColor());
      undoAdu   = CommunicationManager::cm->CreateEvent(((Group*)active->GetGroup())->GetID(), GetID(), undoBuf);
      UndoManager::um->InsertAdu(undoAdu);

      ((Group*)active->GetGroup())->InsertNewColor(this,color);
    }
  }
  else if (!selectedObjects.empty()) {
    int bytes;
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;
    RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();
    for (std::list<DrawObject*>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
      if ((*i)->GetType() != GenericObject::IMAGE) {
        if ((*i)->GetType() == GenericObject::GROUP) { 
          Text* txt = (Text*)((Group*)*i)->GetFirst();
          if (!txt) return;
          undoBuf   = new Buffer(((Group*)*i)->GetChangeFontColorSizeTG());
          ((Group*)*i)->FlushChangeFontColorTG(undoBuf->getCont(), txt->GetColor());
          undoAdu   = CommunicationManager::cm->CreateEvent(((Group*)*i)->GetID(), GetID(), undoBuf);
          aduList->push_back(undoAdu);

          ((Group*)*i)->ChangeLineColor(canvas,color);
        } else {
          if ((*i)->GetType() != GenericObject::TELEPOINTER) {
            undoBuf = new Buffer(((ColoredObject*)*i)->GetChangeLinecolorSize());
            ((ColoredObject*)*i)->FlushChangeLinecolor(undoBuf->getCont(), ((ColoredObject*)*i)->GetLineColor());
	    undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);
          }
          
          if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !(*i)->Annotation()) {
	    if ((*i)->GetType() == GenericObject::TELEPOINTER) {
	      ((ColoredObject*)*i)->ChangeLineColor(canvas,color);
              bytes = ((Telepointer*)*i)->FlushChangeLineColor(TelepointerManager::sendBuffer);
              TelepointerManager::tm->SendEvent((*i)->GetID(), TelepointerManager::sendBuffer, bytes);
	      
            } else {
	      buf   = new Buffer(((ColoredObject*)*i)->GetChangeLinecolorSize());
	      ((ColoredObject*)*i)->FlushChangeLinecolor(buf->getCont(), color);
	      adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	  }
	  else { // PRIVATE or Annotation
	    ((ColoredObject*)*i)->ChangeLineColor(canvas,color);
	  }
        }
      }
    }
    if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

    SaveAsAOF();
  }
}


void MLB::Page::ChangeFillColor(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeColor\n");
  #endif

  if (objects.empty() || selectedObjects.empty()) return;

  std::list<MLB::DrawObject*>::iterator iter;
  char* fillColor = (char*)argv[1];

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if ((*iter)->GetType() != GenericObject::IMAGE && (*iter)->GetType() != GenericObject::ASCIITEXT &&
        (*iter)->GetType() != GenericObject::LINE) {
      if ((*iter)->GetType() == GenericObject::GROUP) {
	      //TG
        //((Group*)*iter)->ChangeFillColor(canvas, fillColor);
      } else {
        if (((ColoredObject*)*iter)->IsFilled()) {
          undoBuf = new Buffer(((ColoredObject*)*iter)->GetChangeFillcolorSize());
          ((ColoredObject*)*iter)->FlushChangeFillcolor(undoBuf->getCont(), ((ColoredObject*)*iter)->GetFillColor());
	        undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	        aduList->push_back(undoAdu);

          if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !(*iter)->Annotation()) {
            buf = new Buffer(((ColoredObject*)*iter)->GetChangeFillcolorSize());
	          ((ColoredObject*)*iter)->FlushChangeFillcolor(buf->getCont(), fillColor);
	          adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	          AduManager::am->InsertAdu(adu);
	          CommunicationManager::cm->SendAduReliable(adu);
	        } else { // Annotation
             ((ColoredObject*)*iter)->ChangeFillColor(canvas, fillColor);
          }
        }
      }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

  SaveAsAOF();
}


void MLB::Page::ChangeLineWidth(int argc, const char* argv[]) {

  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeLineWidth\n");
  #endif

  std::list<MLB::DrawObject*>::iterator iter;
  int lineWidth;

  if (objects.empty() || selectedObjects.empty()) return;

  lineWidth = atoi(argv[1]);

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if (((*iter)->GetType() != GenericObject::ASCIITEXT) && 
        ((*iter)->GetType() != GenericObject::GROUP) && 
        ((*iter)->GetType() != GenericObject::IMAGE)) {
	    
      undoBuf = new Buffer(((ColoredObject*)*iter)->GetChangeLinewidthSize());
	    ((ColoredObject*)*iter)->FlushChangeLinewidth(undoBuf->getCont(), ((ColoredObject*)*iter)->GetLineWidth());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) &&
          !(*iter)->Annotation()) {
	      
          buf = new Buffer(((ColoredObject*)*iter)->GetChangeLinewidthSize());
	        ((ColoredObject*)*iter)->FlushChangeLinewidth(buf->getCont(), lineWidth);
	        adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
          AduManager::am->InsertAdu(adu);
	        CommunicationManager::cm->SendAduReliable(adu);
	    
      } else { // PRIVATE or Annotation
	      ((ColoredObject*)*iter)->ChangeLineWidth(canvas, lineWidth);
 	    }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
  
  SaveAsAOF();
}


void MLB::Page::ChangeLineStyle(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeLineStyle\n");
  #endif

  std::list<MLB::DrawObject*>::iterator iter;
  char *lineStyle;

  if (objects.empty() || selectedObjects.empty()) return;

  lineStyle = (char*)argv[1];

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if (((*iter)->GetType() != GenericObject::IMAGE) && 
        ((*iter)->GetType() != GenericObject::GROUP) && 
        ((*iter)->GetType() != GenericObject::ASCIITEXT)) {

      undoBuf = new Buffer(((ColoredObject*)*iter)->GetChangeLinestyleSize());
      ((ColoredObject*)*iter)->FlushChangeLinestyle(undoBuf->getCont(), ((ColoredObject*)*iter)->GetLineStyle());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) &&
          !(*iter)->Annotation()) {

	        buf = new Buffer(((ColoredObject*)*iter)->GetChangeLinestyleSize());
	        ((ColoredObject*)*iter)->FlushChangeLinestyle(buf->getCont(), lineStyle);
	        adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	        AduManager::am->InsertAdu(adu);
	        CommunicationManager::cm->SendAduReliable(adu);

      } else { // PRIVATE or Annotation
	      if ((*iter)->GetType() == GenericObject::GROUP)
	        ((Group*)*iter)->ChangeLineStyle(canvas, lineStyle);
	      else
	        ((ColoredObject*)*iter)->ChangeLineStyle(canvas, lineStyle);
	    }
    }
  }
  if (!aduList->empty())  UndoManager::um->InsertAduList(aduList);

  SaveAsAOF();
}


void MLB::Page::ChangeCapStyle(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeCapStyle\n");
  #endif

  std::list<MLB::DrawObject*>::iterator
    iter;
  char
    *capStyle;

  if (objects.empty() || selectedObjects.empty()) return;

  capStyle = (char*)argv[1];

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if ((*iter)->GetType() == GenericObject::LINE) {

      undoBuf = new Buffer(((Line*)*iter)->GetChangeCapstyleSize());
      ((Line*)*iter)->FlushChangeCapstyle(undoBuf->getCont(), ((Line*)*iter)->GetCapStyle());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
 	      buf = new Buffer(((Line*)*iter)->GetChangeCapstyleSize());
	      ((Line*)*iter)->FlushChangeCapstyle(buf->getCont(), capStyle);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    } else { // PRIVATE or Annotation
	      ((Line*)*iter)->ChangeCapStyle(canvas, capStyle);
	    }
    }
    else if ((*iter)->GetType() == GenericObject::POLYLINE || (*iter)->GetType() == GenericObject::FREEHAND) {

      undoBuf = new Buffer(((PolyLine*)*iter)->GetChangeCapstyleSize());
      ((PolyLine*)*iter)->FlushChangeCapstyle(undoBuf->getCont(), ((PolyLine*)*iter)->GetCapStyle());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((PolyLine*)*iter)->GetChangeCapstyleSize());
	      ((PolyLine*)*iter)->FlushChangeCapstyle(buf->getCont(), capStyle);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    } else { // PRIVATE or Annotation
	      ((PolyLine*)*iter)->ChangeCapStyle(canvas, capStyle);
	    }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

  SaveAsAOF();
}


void MLB::Page::ChangeJoinStyle(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeJoinStyle\n");
  #endif

  std::list<MLB::DrawObject*>::iterator
    iter;
  char
    *joinStyle;

  if (objects.empty() || selectedObjects.empty()) return;

  joinStyle = (char*)argv[1];
  
  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();
  
  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if ((*iter)->GetType() == GenericObject::LINE) {
	  
      undoBuf = new Buffer(((Line*)*iter)->GetChangeJoinstyleSize());
      ((Line*)*iter)->FlushChangeJoinstyle(undoBuf->getCont(), ((Line*)*iter)->GetJoinStyle());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((Line*)*iter)->GetChangeJoinstyleSize());
	      ((Line*)*iter)->FlushChangeJoinstyle(buf->getCont(), joinStyle);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // PRIVATE or Annotation
	      ((Line*)*iter)->ChangeJoinStyle(canvas, joinStyle);
	    }
    }
    else if ((*iter)->GetType() == GenericObject::POLYLINE || (*iter)->GetType() == GenericObject::FREEHAND) {
	    
      undoBuf = new Buffer(((PolyLine*)*iter)->GetChangeJoinstyleSize());
      ((PolyLine*)*iter)->FlushChangeJoinstyle(undoBuf->getCont(), ((PolyLine*)*iter)->GetJoinStyle());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((PolyLine*)*iter)->GetChangeJoinstyleSize());
	      ((PolyLine*)*iter)->FlushChangeJoinstyle(buf->getCont(), joinStyle);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    } else { // PRIVATE or Annotation
	      ((PolyLine*)*iter)->ChangeJoinStyle(canvas, joinStyle);
	    }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

  SaveAsAOF();
}


void MLB::Page::ChangeArrowStyle(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeArrowStyle\n");
  #endif

  std::list<MLB::DrawObject*>::iterator
    iter;
  char
    *arrowStyle;

  if (objects.empty() || selectedObjects.empty()) return;

  arrowStyle = (char*)argv[1];

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();
  
  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if ((*iter)->GetType() == GenericObject::LINE) {
      
      undoBuf = new Buffer(((Line*)*iter)->GetChangeArrowstyleSize());
      ((Line*)*iter)->FlushChangeArrowstyle(undoBuf->getCont(), ((Line*)*iter)->GetArrowStyle(), ((Line*)*iter)->GetArrowShape());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((Line*)*iter)->GetChangeArrowstyleSize());
	      ((Line*)*iter)->FlushChangeArrowstyle(buf->getCont(), arrowStyle, ((Line*)*iter)->GetArrowShape());
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // PRIVATE or Annotation 
	      ((Line*)*iter)->ChangeArrowStyle(canvas, arrowStyle);
	    }
    } else if ((*iter)->GetType() == GenericObject::POLYLINE || (*iter)->GetType() == GenericObject::FREEHAND) {

      undoBuf = new Buffer(((PolyLine*)*iter)->GetChangeArrowstyleSize());
      ((PolyLine*)*iter)->FlushChangeArrowstyle(undoBuf->getCont(), ((PolyLine*)*iter)->GetArrowStyle(), ((PolyLine*)*iter)->GetArrowShape());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((PolyLine*)*iter)->GetChangeArrowstyleSize());
	      ((PolyLine*)*iter)->FlushChangeArrowstyle(buf->getCont(), arrowStyle, ((PolyLine*)*iter)->GetArrowShape());
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // PRIVATE or Annotation
	      ((PolyLine*)*iter)->ChangeArrowStyle(canvas, arrowStyle);
	    }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

  SaveAsAOF();
}


void MLB::Page::ChangeArrowShape(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ChangeArrowShape\n");
  #endif

  std::list<MLB::DrawObject*>::iterator
    iter;
  char
    *arrowShape;

  if (objects.empty() || selectedObjects.empty()) return;

  arrowShape = (char*)argv[1];

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  for (iter = selectedObjects.begin(); iter != selectedObjects.end(); ++iter) {
    if ((*iter)->GetType() == GenericObject::LINE) {
	    
      undoBuf = new Buffer(((Line*)*iter)->GetChangeArrowstyleSize());
      ((Line*)*iter)->FlushChangeArrowstyle(undoBuf->getCont(), ((Line*)*iter)->GetArrowStyle(), ((Line*)*iter)->GetArrowShape());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((Line*)*iter)->GetChangeArrowstyleSize());
	      ((Line*)*iter)->FlushChangeArrowstyle(buf->getCont(), ((Line*)*iter)->GetArrowStyle(), arrowShape);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // PRIVATE or Annotation 
	      ((Line*)*iter)->ChangeArrowShape(canvas, arrowShape);
	    }
    }
    else if ((*iter)->GetType() == GenericObject::POLYLINE || (*iter)->GetType() == GenericObject::FREEHAND) {

      undoBuf = new Buffer(((PolyLine*)*iter)->GetChangeArrowstyleSize());
      ((PolyLine*)*iter)->FlushChangeArrowstyle(undoBuf->getCont(), ((PolyLine*)*iter)->GetArrowStyle(), ((PolyLine*)*iter)->GetArrowShape());
	    undoAdu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), GetID(), undoBuf);
	    aduList->push_back(undoAdu);

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !(*iter)->Annotation()) {
	      buf = new Buffer(((PolyLine*)*iter)->GetChangeArrowstyleSize());
	      ((PolyLine*)*iter)->FlushChangeArrowstyle(buf->getCont(), ((PolyLine*)*iter)->GetArrowStyle(), arrowShape);
	      adu = CommunicationManager::cm->CreateEvent((*iter)->GetID(), buf);
	      AduManager::am->InsertAdu(adu);
	      CommunicationManager::cm->SendAduReliable(adu);
	    }
	    else { // PRIVATE or Annotation
	      ((PolyLine*)*iter)->ChangeArrowShape(canvas, arrowShape);
      }
    }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);  

  SaveAsAOF();
}


void MLB::Page::MoveObjects(int argc, const char* argv[]) {
  // in shared mode moving groups with several objects is not possible since coordinates
  // are absolute and not relative

  #ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::MoveObjects\n");
  #endif

  if (objects.empty()) return;

  static bool move;
  static int lastX, lastY;
  static int firstX, firstY;
  static std::list<DrawObject*> moveObjects;
  static std::list<Delta> deltas;

  static double test;
  static int count = 0;
  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  std::list<DrawObject*>::iterator i;
  DrawObject* obj;
  bool selected;

  int status = atoi(argv[1]);
  int x      = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y      = atoi(argv[3]) / SessionManager::sm->GetZoom();
  
  int _x, _y;
  
  //
  // BEGIN
  //
  if (status == Action::BEGIN) {
    if (active && active->GetType() == GenericObject::ASCIITEXT) ((Text *)active)->DeleteTextCursor(canvas);
    active = NULL;

    lastX = atoi(argv[4]) / SessionManager::sm->GetZoom();
    lastY = atoi(argv[5]) / SessionManager::sm->GetZoom();
    
    selected = false;
    moveObjects.clear();

    for (i = selectedObjects.begin(); i != selectedObjects.end() && !selected; ++i) {
      if ((*i)->Overlap(canvas,x,y)) selected = true;
    }

    if (selected) {
      for (i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
	if (!DrawObject::Find(moveObjects, *i)) moveObjects.push_back((*i));
      }
    } else {
      UnSelectObjects();
    }

    if (moveObjects.empty()) {
      i = objects.end();
      do {
	--i;
	if ((*i)->Overlap(canvas,x,y)) {
	  if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
	  obj = (*i);
	  while (obj->GetGroup()) obj = obj->GetGroup();
	  if (!DrawObject::Find(moveObjects, obj)) moveObjects.push_back(obj);
	  break;
	}
      } while (i != objects.begin());
    }

    firstX = x;
    firstY = y;
    if (moveObjects.empty()) {
      move = false;
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE) {
	canvas->BeginSelectRegion(SessionManager::sm->GetZoom() * x, SessionManager::sm->GetZoom() * y);
      }
    } else {
      move = true;
      deltas.clear();
      if (!moveObjects.empty()) {
	for (i = moveObjects.begin(); i != moveObjects.end(); ++i) {
	  deltas.push_back(Delta((*i)->GetX() - x, (*i)->GetY() - y));
	}
      }
    }

    //
    // CONTINUE
    //
  } else if (status == Action::CONTINUE) {
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) {
      if (moveObjects.empty()) return;

      std::list<Delta>::iterator j = deltas.begin();
      for (i = moveObjects.begin(); i != moveObjects.end(); ++i) {
	if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) {
	  // do not move
	  lastX = x;
	  lastY = y;
	  continue;
	}

	if (!(*i)->Annotation()) {
	  _x = x + j->dx; if (_x < 0) _x = 0;
	  _y = y + j->dy; if (_y < 0) _y = 0;
	  buf   = new Buffer((*i)->GetMoveSize());
	  if (((*i)->GetType() == GenericObject::GROUP) && ((Group*)*i)->IsTextGroup()) ((Group*)*i)->FlushMoveTG(buf->getCont(), _x, _y);
	  else (*i)->FlushMove(buf->getCont(), _x, _y);
	  adu   = CommunicationManager::cm->CreateCue((*i)->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  test  = CommunicationManager::cueSendRate * count;
	  if (floor(test) == ceil(test)) {
	    CommunicationManager::cm->SendAduUnreliable(adu);
	  }
	  if (count == 100) count = 1;
	  else ++count;
	} else { // Annotation
	  (*i)->Move(canvas, x-lastX, y-lastY);
	}
	++j;
      }
    } else { // PRIVATE
      if (move) {

	for (i = moveObjects.begin(); i != moveObjects.end(); ++i) {
	  if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue; // do not move images
	  (*i)->Move(canvas, x-lastX, y-lastY);
	}

      } else {
	canvas->ModifySelectRegion(SessionManager::sm->GetZoom() * firstX, SessionManager::sm->GetZoom() * firstY,
				   SessionManager::sm->GetZoom() * x, SessionManager::sm->GetZoom() * y);
      }
    }

    lastX = x;
    lastY = y;

    //
    // END
    //
  } else if (status == Action::END) {
    
    if (move) {
      if (moveObjects.empty()) return;
      SaveAsAOF(); // Multi-layer recording

      std::list<Delta>::iterator j = deltas.begin();
      for (i = moveObjects.begin(); i != moveObjects.end(); ++i) {
	if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) {
	  // do not move images
	  lastX = x;
	  lastY = y;
	  continue;
	}

	if ((SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED) && !(*i)->Annotation()) {
	  _x = x + j->dx; if (_x < 0) _x = 0;
	  _y = y + j->dy; if (_y < 0) _y = 0;
	  buf   = new Buffer((*i)->GetMoveSize());
	  if (((*i)->GetType() == GenericObject::GROUP) && ((Group*)*i)->IsTextGroup()) ((Group*)*i)->FlushMoveTG(buf->getCont(), _x, _y);
	  else (*i)->FlushMove(buf->getCont(), _x, _y);
	  adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
        
        } else { // PRIVATE or Annotation
	  (*i)->Move(canvas, x-lastX, y-lastY);
	}
        
        _x = firstX + j->dx; if (_x < 0) _x = 0;
	_y = firstY + j->dy; if (_y < 0) _y = 0;
	undoBuf = new Buffer((*i)->GetMoveSize());  
        if (((*i)->GetType() == GenericObject::GROUP) && ((Group*)*i)->IsTextGroup()) ((Group*)*i)->FlushMoveTG(undoBuf->getCont(), _x, _y);
	else (*i)->FlushMove(undoBuf->getCont(), _x, _y);
        undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
        aduList->push_back(undoAdu); 

	++j;
      }
      if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);

      SaveAsAOF();

    } else { // not move

      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::PRIVATE) {
	int x1 = MIN(x,firstX);
	int x2 = MAX(x,firstX);
	int y1 = MIN(y,firstY);
	int y2 = MAX(y,firstY);
	canvas->DeleteSelectRegion();
	for (i = objects.begin(); i != objects.end(); ++i) {
	  if ((*i)->IsInRegion(x1,y1,x2,y2)) {
	    if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
	    obj = (*i)->Select();
	    if (obj && !DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
	  }
	}
	if (selectedObjects.empty()) {
	  i = objects.end();
	  do {
	    --i;
	    if ((*i)->Overlap(canvas, x, y)) {
	      if (((*i)->GetType() == GenericObject::IMAGE) && SessionManager::sm->FreezeImages()) continue;
	      obj = (*i)->Select();
	      if (obj) {
		if (!DrawObject::Find(selectedObjects, obj)) selectedObjects.push_back(obj);
		obj->AdjustGui();
	      }
	      break;
	    }
	  } while (i != objects.begin());
	}

	DrawSelected();
      }
    }
  }
}


void MLB::Page::RaiseObjects() {
  if (objects.empty()) return;
  if (selectedObjects.empty()) return;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  uint32 last = GetEndObjectOrder();
  last += MLB_OO_OFFSET;

  for (std::list<DrawObject*>::iterator i = selectedObjects.begin(); i != selectedObjects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::GROUP) {
      undoBuf = new Buffer((*i)->GetRaiseSize());  
      ((Group*)*i)->FlushRaiseTG(undoBuf->getCont(), (*i)->GetOrder());
      undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
      aduList->push_back(undoAdu); 
      
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && (!(*i)->Annotation())) {
	buf = new Buffer((*i)->GetRaiseSize());
	((Group*)*i)->FlushRaiseTG(buf->getCont(), last);
	adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);

	last += MLB_OO_OFFSET;

      } else { // PRIVATE or Annotation 
      DeleteObjectFromContainer(*i);
      (*i)->SetOrder(last);
      AppendObject(*i);
      last += MLB_OO_OFFSET;

	Text* txt;
	std::vector<uint32>* ids = ((Group*)*i)->GetObjectIDs();
	for (std::vector<uint32>::iterator j = ids->begin(); j != ids->end(); ++j) {
	  txt = (Text*)(GetObject(*j));
	  DeleteObjectFromContainer(txt);
	  txt->SetOrder(last);
	  AppendObject(txt);
	  ++last;
	}
	delete ids;
      }
    } else { // not GROUP
      undoBuf = new Buffer((*i)->GetRaiseSize());  
      (*i)->FlushRaise(undoBuf->getCont(), (*i)->GetOrder());
      undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
      aduList->push_back(undoAdu); 
      
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && (!(*i)->Annotation())) {
        buf = new Buffer((*i)->GetRaiseSize());
	(*i)->FlushRaise(buf->getCont(), last);
	adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	last += MLB_OO_OFFSET;

      } else { // PRIVATE or Annotation
        DeleteObjectFromContainer(*i);
        (*i)->SetOrder(last);
        AppendObject(*i);
        last += MLB_OO_OFFSET;
    }
  }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
  ReDraw();
}


void MLB::Page::LowerObjects() {
  if (objects.empty()) return;
  if (selectedObjects.empty()) return;

  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<Buffer> undoBuf;
  RefCount<RTPI::RtpiAdu> undoAdu;
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList();

  uint32 first = GetFirstObjectOrder();
  first -= MLB_OO_OFFSET;

  for (std::list<DrawObject*>::reverse_iterator i = selectedObjects.rbegin(); i != selectedObjects.rend(); ++i) {
    if ((*i)->GetType() == GenericObject::GROUP) {
      undoBuf = new Buffer((*i)->GetLowerSize());  
      ((Group*)*i)->FlushLowerTG(undoBuf->getCont(), (*i)->GetOrder());
      undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
      aduList->push_back(undoAdu); 

    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && (!(*i)->Annotation())) {
	buf = new Buffer((*i)->GetRaiseSize());
	((Group*)*i)->FlushLowerTG(buf->getCont(), first);
	adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
	first -= MLB_OO_OFFSET;

      } else { // PRIVATE or Annotation
      DeleteObjectFromContainer(*i);
      (*i)->SetOrder(first);
      AppendObject(*i);
      first -= MLB_OO_OFFSET;

	Text* txt;
	std::vector<uint32>* ids = ((Group*)*i)->GetObjectIDs();
	for (std::vector<uint32>::iterator j = ids->begin(); j != ids->end(); ++j) {
	  txt = (Text*)(GetObject(*j));
	  DeleteObjectFromContainer(txt);
	  txt->SetOrder(first);
	  AppendObject(txt);
	  ++first;
	}
	delete ids;
      }
    } else { // not GROUP
      undoBuf = new Buffer((*i)->GetLowerSize());  
      (*i)->FlushLower(undoBuf->getCont(), (*i)->GetOrder());
      undoAdu = CommunicationManager::cm->CreateEvent((*i)->GetID(), GetID(), undoBuf);
      aduList->push_back(undoAdu); 
  
      if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && (!(*i)->Annotation())) {
        buf = new Buffer((*i)->GetLowerSize());
	(*i)->FlushLower(buf->getCont(), first);
	adu   = CommunicationManager::cm->CreateEvent((*i)->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
        first -= MLB_OO_OFFSET;
      
      } else { // PRIVATE or Annotation
        DeleteObjectFromContainer(*i);
        (*i)->SetOrder(first);
        AppendObject(*i);
        first -= MLB_OO_OFFSET;
    }
  }
  }
  if (!aduList->empty()) UndoManager::um->InsertAduList(aduList);
  ReDraw();
}


void MLB::Page::PlaceImage(int argc, const char* argv[]) {
  static MLB::PhotoImage* img=NULL;
  static int lastX, lastY;

  if (!active) return;

  int status = atoi(argv[1]);
  int x = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y = atoi(argv[3]) / SessionManager::sm->GetZoom();

  switch (status) {
  case MLB::Action::BEGIN:
    img = (PhotoImage *)active;
    if (!img) return;
    img->SetX(x);
    img->SetY(y);
    img->Draw(canvas);
    lastX = x;
    lastY = y;
    break;
  case MLB::Action::CONTINUE:
    if (!img) return;
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !img->Annotation()) {
      RefCount<Buffer> buf = new Buffer(img->GetMoveSize());
      img->FlushMove(buf->getCont(), x-lastX, y-lastY);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateCue(img->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      // send only end position in order to minimize computational effort
      //CommunicationManager::cm->SendAduUnreliable(adu);
    }
    else { // PRIVATE
      img->Move(canvas,x-lastX,y-lastY);
    }
    lastX = x;
    lastY = y;
    break;
  case MLB::Action::END:
    if (!img) return;
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !img->Annotation()) {
      RefCount<Buffer> buf = new Buffer(img->GetMoveSize());
      img->FlushMove(buf->getCont(), x-lastX, y-lastY);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(img->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }
    else { // PRIVATE
      img->Move(canvas,x-lastX,y-lastY);
    }
    active = NULL;
    img    = NULL;
    break;
  }
}


void MLB::Page::InsertTextLine(int argc, const char* argv[])
{
#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::InsertTextLine\n");
#endif

  if (objects.empty()) return;

  if (active && (active->GetType() == MLB::GenericObject::ASCIITEXT)) {
    ((Group*)active->GetGroup())->InsertNewLine(this);
  }
}


void MLB::Page::EndText(int argc, const char* argv[])
{
//  #ifdef _DEBUG1
//    GUI::MessageWindow::Debug("MLB::Page::EndText\n");
//  #endif

//    active = NULL;
//    DeleteTextCursor();
}

void MLB::Page::InsertImage( Magick::Image& _image, char *_filename, int _x, int _y, bool _single) {
  bool annotation = SessionManager::sm->AnnotationMode();

  UnSelectObjects();

  uint32 objID;
  
  if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
    uint32 blength = 0;
    string type;

    try {
      PhotoImage image( CommunicationManager::cm->GetObjectID(), _filename, 10, 10, _image );

      RefCount<Buffer> buf = new Buffer( image.GetStateSize() );
      //image.SetVisible();
      image.SetOrder  ( GetPushBackObjectOrder() );
      image.FlushState( buf->getCont()           );

      objID = CommunicationManager::cm->GetObjectID();
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(objID, GetID(), CommunicationManager::IMAGE, 3, buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);

      buf   = new Buffer(GetChangeActiveItemSize());
      FlushChangeActiveItem(buf->getCont());
      adu   = CommunicationManager::cm->CreateEvent(GetID(), objID, buf);
      AduManager::am->InsertAdu(adu);
    }
    catch ( Magick::Exception& exc )
    {
      GUI::MessageWindow::Error( "Page::InsertImage - Cannot build blob. %q", exc.what() );
      return;
    }
  }
  else { // PRIVATE session or Annotation Mode

    PhotoImage* img;
    objID = CommunicationManager::cm->GetObjectID();

    try { img = new PhotoImage(objID, _filename, _x, _y, _image );  }
    catch ( Magick::Exception& exc ) {
      GUI::MessageWindow::Error( "Page::InsertImage - cannot create image: %q", exc.what() );
      return;
    }

    img->SetAnnotation(annotation); 
    img->SetOrder(GetPushBackObjectOrder());
    AppendObject(img);
    if (_single) {
      active = img;
    }
    else {
      SetPageDimension(img->GetWidth(),img->GetHeight());
      active = NULL;
    }
  }
  
  RefCount<Buffer> undoBuf        = new Buffer(GenericObject::GetDeleteSize());  
  PhotoImage::FlushDeleteS(undoBuf->getCont());
  RefCount<RTPI::RtpiAdu> undoAdu = CommunicationManager::cm->CreateEvent(objID, GetID(), undoBuf);
  UndoManager::um->InsertAdu(undoAdu); 
}


void MLB::Page::ActivateOnlinePointer(int argc, const char* argv[]) {
  int status = atoi(argv[1]);
  int x = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y = atoi(argv[3]) / SessionManager::sm->GetZoom();
  char* color = (char*)argv[4];

  static bool removed = false;

  switch (status) {
  case MLB::Action::BEGIN : {
    color = (char*)argv[4];
    onlinePointer->Hide(this);
    onlinePointer->Update(color);
    onlinePointer->Update(x,y);
    onlinePointer->SetActive(this);

    if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
      int bytes = onlinePointer->FlushState(TelepointerManager::sendBuffer);
      TelepointerManager::tm->SendState(onlinePointer->GetID(), GetID(), 3, TelepointerManager::sendBuffer, bytes);
    }

    SaveAsAOF();
    break;
  }
  case MLB::Action::CONTINUE : {
    onlinePointer->Update(x,y);
    onlinePointer->ReDraw(this);

    if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
      // register participant activity 
      ParticipantManager::pm->UpdateParticipantActiveness(CommunicationManager::cm->GetParticipantID());

      static int count = 0;
      static double test;
      test = CommunicationManager::cueSendRate * count;
      if (floor(test) == ceil(test)) {
	int bytes = onlinePointer->FlushMove(TelepointerManager::sendBuffer);
	TelepointerManager::tm->SendCue(onlinePointer->GetID(), TelepointerManager::sendBuffer, bytes);
      }
      if (count == 100) count = 1;
      else ++count;
    }

    SaveAsAOF();
    break;
  }
  case MLB::Action::END : {
    onlinePointer->ReDraw(this);

    if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
      // register participant activity 
      ParticipantManager::pm->UpdateParticipantActiveness(CommunicationManager::cm->GetParticipantID());

      int bytes = onlinePointer->FlushState(TelepointerManager::sendBuffer);
      TelepointerManager::tm->SendState(onlinePointer->GetID(), GetID(), 3, TelepointerManager::sendBuffer, bytes);

      //int bytes = onlinePointer->FlushChangeVisibility(CommunicationManager::sendBuffer);
      //TelepointerManager::tm->SendEvent(onlinePointer->GetID(), CommunicationManager::sendBuffer, bytes);
      //bytes = onlinePointer->FlushMove(CommunicationManager::sendBuffer);
      //TelepointerManager::tm->SendEvent(onlinePointer->GetID(), CommunicationManager::sendBuffer, bytes);
    }
    SaveAsAOF();
    break;
  }
  }
}


void MLB::Page::ActivateOnlineBox(int argc, const char* argv[]) {
  static int firstX, firstY;
  
  int status      = atoi(argv[1]);
  int x           = atoi(argv[2]) / SessionManager::sm->GetZoom();
  int y           = atoi(argv[3]) / SessionManager::sm->GetZoom();
  char* color     = (char*)argv[4];
  int   lineWidth = atoi(argv[5]);
  char* lineStyle = (char*)argv[6];

  switch (status) {
  case MLB::Action::BEGIN : {
    color = (char*)argv[4];
    onlineBox->Hide(this);
    onlineBox->Update(color, lineWidth, lineStyle);
    onlineBox->Update(x,y,0,0);
    onlineBox->SetActive(this);

    if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
      int bytes = onlineBox->FlushState(TelepointerManager::sendBuffer);
      TelepointerManager::tm->SendState(onlineBox->GetID(), GetID(), 3, TelepointerManager::sendBuffer, bytes);
    }

    firstX = x;
    firstY = y;

    //?? anpassen ??
    //SaveAsAOF();
    break;
  }
  case MLB::Action::CONTINUE : {
    onlineBox->Update(firstX,firstY,x-firstX,y-firstY);
    onlineBox->ReDraw(this);

    if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
      // register participant activity 
      ParticipantManager::pm->UpdateParticipantActiveness(CommunicationManager::cm->GetParticipantID());

      static int count = 0;
      static double test;
      test = CommunicationManager::cueSendRate * count;
      if (floor(test) == ceil(test)) {
	      int bytes = onlineBox->FlushChangeSize(TelepointerManager::sendBuffer);
	      TelepointerManager::tm->SendCue(onlineBox->GetID(), TelepointerManager::sendBuffer, bytes);
      }
      if (count == 100) count = 1;
      else ++count;
    }
 
    //?? anpassen ??
    //SaveAsAOF();
    break;
  }
  case MLB::Action::END : {
    // box-pointer moved ?
    if ((x == firstX) && (y == firstY)) {
      onlineBox->SetInactive(this);

      if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
        int bytes = onlineBox->FlushChangeVisibility(TelepointerManager::sendBuffer);
        TelepointerManager::tm->SendEvent(onlineBox->GetID(), TelepointerManager::sendBuffer, bytes);
      }
    
    } else {
      onlineBox->ReDraw(this);

      if (SessionManager::sm->GetParentSession(this)->GetSessionType() == Session::SHARED) {
        // register participant activity 
        ParticipantManager::pm->UpdateParticipantActiveness(CommunicationManager::cm->GetParticipantID());

        int bytes = onlineBox->FlushState(TelepointerManager::sendBuffer);
        TelepointerManager::tm->SendState(onlineBox->GetID(), GetID(), 3, TelepointerManager::sendBuffer, bytes);
      }
    }
    //?? anpassen ??
    //SaveAsAOF();
    break;
  }
  }
}


void MLB::Page::SaveAsAOF(int _time)
{
#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SaveAsAOF\n");
#endif

  FILE
    *objFp,
    *evqFp;
  int
    time;

  if (MLB::SessionManager::sm->recording && !MLB::SessionManager::sm->paused) {
#ifdef WIN32
    time = MLB::SessionManager::sm->aofRecording->GetRecordTime();
#else
    time = MLB::SessionManager::sm->aofRecording->GetTime();
#endif

 //     GUI::MessageWindow::Debug("Page::SaveAsAOF\n");

    objFp = MLB::SessionManager::sm->GetObjFp();
    evqFp = MLB::SessionManager::sm->GetEvqFp();
    if (firstTime < 0) firstTime = time;
    MLB::SessionManager::sm->AddPageAsRecorded(this,time);
    fprintf(evqFp,"%d %d %d %d",time,pageNumber,CalculateColor(background),RecordableObjects());
    WriteObjectFile();

    if (onlinePointer->IsActive()) {
      onlinePointer->SetUnrecorded();
      onlinePointer->SaveAsAOF(objFp);
      fprintf(evqFp," %d",onlinePointer->GetRecordID());
      GUI::MessageWindow::Debug("RecordID vom OnlinePointer %d\n",onlinePointer->GetRecordID());
    }

    //Telebox ???

    fprintf(evqFp,"\n");
    fflush(objFp);
    fflush(evqFp);
  }
}


int MLB::Page::RecordableObjects()
{
  int
    ret;

  ret = 0;

//  GUI::MessageWindow::Debug("AnnotationMode: %d *****************\n",SessionManager::sm->AnnotationMode());
  for (std::list<MLB::DrawObject *>::iterator i = objects.begin(); i != objects.end(); ++i) {
	if (((*i)->GetLayer()==2 && SessionManager::sm->AnnotationMode()==1) || (SessionManager::sm->AnnotationMode()==0 && ((*i)->GetLayer()==0 || (*i)->GetLayer()==1))) 
    ret += (*i)->RecordableObjects();
//	  GUI::MessageWindow::Debug("Anzahl Objecte =%d",(*i)->RecordableObjects());
 //     GUI::MessageWindow::Debug("ObjectCounter RecordID=%d\n",(*i)->GetRecordID());
  }

  if (onlinePointer->IsActive()) ret++;
  
  // Telebox ??

  return ret;
}


void MLB::Page::ResetRecording()
{
#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::ClearRecordedObjects\n");
#endif

  std::list<MLB::DrawObject *>::iterator 
    iter;

  iter = objects.begin();
  while (iter != objects.end()) {
    (*iter)->SetUnrecorded();
    if ((*iter)->GetType() == GenericObject::IMAGE) {
      ((PhotoImage *)(*iter))->WriteImage();
    }
    iter++;
  }
  firstTime = -1;
}


void MLB::Page::WriteObjectFile()
{
  FILE
    *objFp,
    *evqFp;

  objFp = MLB::SessionManager::sm->GetObjFp();
  evqFp = MLB::SessionManager::sm->GetEvqFp();

  for (std::list<MLB::DrawObject *>::iterator iter = objects.begin(); iter != objects.end(); ++iter) {
    if ((*iter)->IsVisible()) {
			  if (((*iter)->GetLayer()==2 && SessionManager::sm->AnnotationMode()==1) || (SessionManager::sm->AnnotationMode()==0 && ((*iter)->GetLayer()==0 || (*iter)->GetLayer()==1)))  {
       
      (*iter)->SaveAsAOF(objFp);
      GUI::MessageWindow::Debug("Layer: %d\n",(*iter)->GetLayer());
	//  GUI::MessageWindow::Debug("RecordID Abfrage 1: %d\n",(*iter)->GetRecordID());
//	  if (((*iter)->GetLayer()==2 && SessionManager::sm->AnnotationMode()==1) || (SessionManager::sm->AnnotationMode()==0 && ((*iter)->GetLayer()==0 || (*iter)->GetLayer()==1)))  {
      fprintf(evqFp," %d",(*iter)->GetRecordID());
        GUI::MessageWindow::Debug("RecordID Abfrage 1 %d\n",(*iter)->GetRecordID());
//	  }
//	  GUI::MessageWindow::Debug("RecordID Abfrage 2: %d+1\n",(*iter)->GetRecordID());
      if (((*iter)->GetType() != IMAGE) && ((*iter)->GetType() != GROUP) && ((ColoredObject *)(*iter))->IsFilled()) {
	fprintf(evqFp," %d",(*iter)->GetRecordID()+1);
        GUI::MessageWindow::Debug("+1\n");
	}}
  }
}
}


void MLB::Page::SavePointAsAOF(MLB::PolyObject *obj, int _x, int _y,int first)
{
#ifdef _DEBUG1
  GUI::MessageWindow::Debug("MLB::Page::SavePointAsAOF\n");
#endif
  static int
    lastTime = 0;
  std::list<MLB::DrawObject *>::iterator
    iter;
  std::list<Point *>::iterator
    ptTemp;
  FILE
    *objFp,
    *evqFp;
  int
    zX, zY,
    time;

  if (MLB::SessionManager::sm->recording) {
#ifdef WIN32
    time = MLB::SessionManager::sm->aofRecording->GetRecordTime();
#else
    time = MLB::SessionManager::sm->aofRecording->GetTime();
#endif
    if (time < lastTime) lastTime = 0;
    if (time-lastTime < 40) return;
    lastTime = time;
    objFp = MLB::SessionManager::sm->GetObjFp();
    evqFp = MLB::SessionManager::sm->GetEvqFp();
    if (!objects.empty()) {
      obj->SetInvisible();
      fprintf(evqFp,"%d %d %d %d",time,pageNumber,CalculateColor(background),RecordableObjects() + obj->points.size());
      WriteObjectFile();
      obj->SetVisible();
      ptTemp=obj->points.begin();
      zX = (*ptTemp)->x;
      zY = (*ptTemp)->y;
      for (ptTemp=obj->points.begin(); ptTemp != obj->points.end(); ++ptTemp) {
	if ((*ptTemp)->nr < 0) {
	  (*ptTemp)->nr = SessionManager::sm->recCount;
	  fprintf(objFp,"<LINE x=%d y=%d width=%d height=%d fcolor=%d linewidth=%d ori=1 direction=0 layer=%d></LINE>\n",zX,zY,(*ptTemp)->x-zX,(*ptTemp)->y-zY,CalculateColor(obj->GetLineColor()),obj->GetLineWidth(),obj->GetLayer());
	  SessionManager::sm->recCount++;
	}
	zX = (*ptTemp)->x;
	zY = (*ptTemp)->y;
	if ((obj->GetLayer()==2 && SessionManager::sm->AnnotationMode()==1) || (SessionManager::sm->AnnotationMode()==0 && (obj->GetLayer()==0 || obj->GetLayer()==1)))  {
	fprintf(evqFp," %d",(*ptTemp)->nr);
        GUI::MessageWindow::Debug("RecordID Abfrage 3 %d\n",(*ptTemp)->nr);
	}
      }
      fprintf(evqFp,"\n");
    }
  }
  else
    lastTime = 0;
}


void MLB::Page::WriteHtml(Record::AOF *_aofRecording, Page *first, Page *prev, Page *next, Page *last)
{
  FILE
    *slideFp,
    *apFp;
  std::string
    apName,
    slideName,
    gifName;
  char
    suffix[20];

  sprintf(suffix,"_%d.tmpl",pageNumber);

  apName = _aofRecording->GetApPath();
  apName += _aofRecording->GetPrefix();
  apName += suffix;

  apFp = fopen((char *)apName.c_str(),"w");
  fprintf(apFp,"AOFPH/%s %d\n",_aofRecording->GetAofName(),firstTime);
  fclose (apFp);

  sprintf(suffix,"_%d.ap",pageNumber);
  apName = _aofRecording->GetApPath();
  apName += _aofRecording->GetPrefix();
  apName += suffix;

  apFp = fopen((char *)apName.c_str(),"w");
  fprintf(apFp,"%s%s %d\n",_aofRecording->GetDataPath(),_aofRecording->GetAofName(),firstTime);
  fclose (apFp);

  sprintf(suffix,"_%d.png",pageNumber);
  gifName = _aofRecording->GetMiscPath();
  gifName += _aofRecording->GetPrefix();
  gifName += suffix;

  sprintf(suffix,"_%d.htm",pageNumber);
  slideName  = _aofRecording->GetMiscPath();
  slideName += "slidefs";
  slideName += suffix;

  if (!(slideFp = fopen((char *)slideName.c_str(),"w"))) return;

  fprintf(slideFp,"<html>\n");
  fprintf(slideFp,"  <head>\n");
  fprintf(slideFp,"    <title>%s: %s (%d:%d)</title>\n","Unknown","Unknown",0,0);
  fprintf(slideFp,"  </head>\n");
  fprintf(slideFp,"\n");
  fprintf(slideFp,"  <frameset cols=\"%d,*\" frameborder=\"1\" framespacing=\"1\" border=\"1\">\n",width/4+40);
  fprintf(slideFp,"    <frame name=\"indexframe\" src=\"indexframe.htm#%d\" scrolling=\"auto\" border=0 NORESIZE>\n",pageNumber);
  fprintf(slideFp,"    <frame name=\"viewframe\" src=\"slide_%d.htm\" border=0>\n",pageNumber);
  fprintf(slideFp,"\n");
  fprintf(slideFp,"    <noframes>\n");
  fprintf(slideFp,"      <body>\n");
  fprintf(slideFp,"        <H2>You should enable frames in your browser. </H2>\n");
  fprintf(slideFp,"      </body>\n");
  fprintf(slideFp,"    </noframes>\n");
  fprintf(slideFp,"\n");
  fprintf(slideFp,"   </frameset>\n");
  fprintf(slideFp,"</html>\n");
  fclose(slideFp);

  if (!prev) {
    slideName = _aofRecording->GetMiscPath();
    slideName += "index.htm";
    if (!(slideFp = fopen((char *)slideName.c_str(),"w"))) return;
    fprintf(slideFp,"<html>\n");
    fprintf(slideFp,"  <head>\n");
    fprintf(slideFp,"    <title>%s: %s (%d:%d)</title>\n","Unknown","Unknown",0,0);
    fprintf(slideFp,"  </head>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  <frameset cols=\"%d,*\" frameborder=\"1\" framespacing=\"1\" border=\"1\">\n",width/4+40);
    fprintf(slideFp,"    <frame name=\"indexframe\" src=\"indexframe.htm#%d\" scrolling=\"auto\" border=0 NORESIZE>\n",pageNumber);
    fprintf(slideFp,"    <frame name=\"viewframe\" src=\"slide_%d.htm\" border=0>\n",pageNumber);
    fprintf(slideFp,"\n");
    fprintf(slideFp,"    <noframes>\n");
    fprintf(slideFp,"      <body>\n");
    fprintf(slideFp,"        <H2>You should enable frames in your browser. </H2>\n");
    fprintf(slideFp,"      </body>\n");
    fprintf(slideFp,"    </noframes>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"   </frameset>\n");
    fprintf(slideFp,"</html>\n");
    fclose(slideFp);
  }


  slideName = _aofRecording->GetMiscPath();
  slideName += "slide";
  slideName += suffix;
  if (slideFp=fopen((char *)slideName.c_str(),"w")) {
    canvas->SaveAsImage((char *)gifName.c_str(), "png");
    fprintf(slideFp,"<html>\n");
    fprintf(slideFp,"  <head></head>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  <body bgcolor=\"#c0c0c0\">\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"    <table cellpadding=1 cellspacing=1 border=0>\n");
    fprintf(slideFp,"      <tr valign=top>\n");
    if (prev) {
      if (first) {
	fprintf(slideFp,"        <td><a href=\"slidefs_%d.htm\" target=_top><img border=0 src=\"leftend.gif\" alt=\"Go to first slide: %s\"></a></td>\n",first->GetPageNumber(),first->GetName());
      }
      else {
	fprintf(slideFp,"        <td><img border=0 src=\"leftend_dis.gif\"></td>\n");
      }
      fprintf(slideFp,"        <td><a href=\"slidefs_%d.htm\" target=_top><img border=0 src=\"left.gif\" alt=\"Go to previous slide: %s\"></a></td>\n",prev->GetPageNumber(),prev->GetName());
      }
      else {
        fprintf(slideFp,"        <td><img border=0 src=\"leftend_dis.gif\"></td>\n");
        fprintf(slideFp,"        <td><img border=0 src=\"left_dis.gif\"></td>\n");
      }
      if (next) {
        fprintf(slideFp,"        <td><a href=\"slidefs_%d.htm\" target=_top><img border=0 src=\"right.gif\" alt=\"Go to next slide: %s\"></a></td>\n",next->GetPageNumber(),next->GetName());
        if (last) {
          fprintf(slideFp,"        <td><a href=\"slidefs_%d.htm\" target=_top><img border=0 src=\"rightend.gif\" alt=\"Go to last slide: %s\"></a></td>\n",last->GetPageNumber(),last->GetName());
	}
        else {
          fprintf(slideFp,"        <td><img border=0 src=\"rightend_dis.gif\"></td>\n");
        }
      }
      else {
        fprintf(slideFp,"        <td><img border=0 src=\"right_dis.gif\"></td>\n");
        fprintf(slideFp,"        <td><img border=0 src=\"rightend_dis.gif\"></td>\n");
      }        fprintf(slideFp,"        <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n");
      fprintf(slideFp,"        <td valign=center><font size=5 face=\"Helvetica,Arial\">%s</font></td>\n",GetName());
      fprintf(slideFp,"      </tr>\n");
      fprintf(slideFp,"    </table>\n");
      fprintf(slideFp,"\n");
      fprintf(slideFp,"  <hr>\n");
      fprintf(slideFp,"  <img border=1 src=\"%s_%d.png\">\n",_aofRecording->GetPrefix(),pageNumber);
      fprintf(slideFp,"\n");
      fprintf(slideFp,"  </body>\n");
      fprintf(slideFp,"</html>\n");

      fclose(slideFp);
  }
}


void MLB::Page::SaveAsHtml(char* directory, char* docname, FILE* indexframe, int& nbr, int last) {
  FILE* slideFp;
  std::string slideName;
  std::string pngName;
  char suffix[20];


#ifdef WIN32
  sprintf(suffix,"\\i%d.png", nbr);
#else
  sprintf(suffix,"/i%d.png", nbr);
#endif
  pngName = directory;
  pngName += suffix;

  Activate();

  //
  // 1. write frame file
  // 
#ifdef WIN32
  sprintf(suffix,"\\f%d.htm", nbr);
#else
  sprintf(suffix,"/f%d.htm", nbr);
#endif
  slideName  = directory;
  slideName += suffix;

  if (!(slideFp = fopen((char *)slideName.c_str(),"w"))) {
    GUI::MessageWindow::Error("Page::SaveAsHtml - Can't create file %s\n", slideName.c_str());
    return;
  }

  fprintf(slideFp,"<html>\n");
  fprintf(slideFp,"  <head>\n");
  fprintf(slideFp,"    <title>%s</title>\n", docname);
  fprintf(slideFp,"  </head>\n");
  fprintf(slideFp,"\n");
  fprintf(slideFp,"  <frameset cols=\"%d,*\" frameborder=\"1\" framespacing=\"1\" border=\"1\">\n",width/4+40);
  fprintf(slideFp,"    <frame name=\"indexframe\" src=\"indexframe.htm#%d\" scrolling=\"auto\" border=0 NORESIZE>\n", nbr);
  fprintf(slideFp,"    <frame name=\"viewframe\" src=\"s%d.htm\" border=0>\n", nbr);
  fprintf(slideFp,"\n");
  fprintf(slideFp,"    <noframes>\n");
  fprintf(slideFp,"      <body>\n");
  fprintf(slideFp,"        <H2>You should enable frames in your browser. </H2>\n");
  fprintf(slideFp,"      </body>\n");
  fprintf(slideFp,"    </noframes>\n");
  fprintf(slideFp,"\n");
  fprintf(slideFp,"   </frameset>\n");
  fprintf(slideFp,"</html>\n");
  fclose(slideFp);

  if (nbr == 0) {
    slideName = directory;
#ifdef WIN32
    slideName += "\\index.htm";
#else
    slideName += "/index.htm";
#endif
    if (!(slideFp = fopen((char *)slideName.c_str(),"w"))) {
      GUI::MessageWindow::Error("Page::SaveAsHtml - Can't create file %s\n", slideName.c_str());
      return;
    }
    fprintf(slideFp,"<html>\n");
    fprintf(slideFp,"  <head>\n");
    fprintf(slideFp,"    <title>%s</title>\n", docname);
    fprintf(slideFp,"  </head>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  <frameset cols=\"%d,*\" frameborder=\"1\" framespacing=\"1\" border=\"1\">\n", width/4+40);
    fprintf(slideFp,"    <frame name=\"indexframe\" src=\"indexframe.htm#%d\" scrolling=\"auto\" border=0 NORESIZE>\n", nbr);
    fprintf(slideFp,"    <frame name=\"viewframe\" src=\"s%d.htm\" border=0>\n", nbr);
    fprintf(slideFp,"\n");
    fprintf(slideFp,"    <noframes>\n");
    fprintf(slideFp,"      <body>\n");
    fprintf(slideFp,"        <H2>You should enable frames in your browser. </H2>\n");
    fprintf(slideFp,"      </body>\n");
    fprintf(slideFp,"    </noframes>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"   </frameset>\n");
    fprintf(slideFp,"</html>\n");
    fclose(slideFp);
  }

  //
  // 2. write slide file
  //
#ifdef WIN32
  sprintf(suffix,"\\s%d.htm", nbr);
#else
  sprintf(suffix,"/s%d.htm", nbr);
#endif
  slideName  = directory;
  slideName += suffix;
  if (slideFp=fopen((char *)slideName.c_str(),"w")) {
    canvas->SaveAsImage((char *)pngName.c_str(), "png");
    fprintf(slideFp,"<html>\n");
    fprintf(slideFp,"  <head></head>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  <body bgcolor=\"#CBD9ED\">\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"    <table cellpadding=1 cellspacing=1 border=0>\n");
    fprintf(slideFp,"      <tr valign=top>\n");

    fprintf(slideFp,"        <td>&nbsp;&nbsp;&nbsp;&nbsp;</td>\n");
    fprintf(slideFp,"        <td valign=center><font size=5 face=\"Helvetica,Arial\">%s</font></td>\n", GetName());
    fprintf(slideFp,"      </tr>\n");
    fprintf(slideFp,"    </table>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  <hr>\n");
    fprintf(slideFp,"  <img border=1 src=\"i%d.png\">\n", nbr);
    fprintf(slideFp,"  <hr>\n");
    fprintf(slideFp,"    <table cellpadding=1 cellspacing=1 border=0>\n");
    fprintf(slideFp,"      <tr valign=top>\n");
    if (nbr != 0) {
      fprintf(slideFp,"        <td><font size=5 face=\"Helvetica,Arial\"><a href=\"f0.htm\" target=_top>&lt;&lt;</font></a></td>\n");
      fprintf(slideFp,"        <td><font size=5 face=\"Helvetica,Arial\"><a href=\"f%d.htm\" target=_top>&lt;</font></a></td>\n", nbr - 1);
    }
    if (nbr != last) {
      fprintf(slideFp,"        <td><font size=5 face=\"Helvetica,Arial\"><a href=\"f%d.htm\" target=_top>&gt;</font></a></td>\n", nbr + 1);
      fprintf(slideFp,"        <td><font size=5 face=\"Helvetica,Arial\"><a href=\"f%d.htm\" target=_top>&gt;&gt;</font></a></td>\n", last);
    }
    fprintf(slideFp,"      </tr>\n");
    fprintf(slideFp,"\n");
    fprintf(slideFp,"  </body>\n");
    fprintf(slideFp,"</html>\n");
    
    fclose(slideFp);
  }

  // 
  // 3. write indexframe
  //
  fprintf(indexframe,"\n");
  fprintf(indexframe,"      <tr><td><hr></td></tr>\n");
  fprintf(indexframe,"\n");
  fprintf(indexframe,"      <tr>\n");
  fprintf(indexframe,"        <td align=center><strong><font face=\"Helvetica,Arial\"><a name=\"%d\">%s</font></strong><p>\n", nbr, GetName());
  fprintf(indexframe,"        <a href=\"s%d.htm\">\n", nbr);
  fprintf(indexframe,"        <img width=%d height=%d border=1 src=\"i%d.png\" alt=\"Display slide: i%d.png\"></a><p>\n", 
	  GetWidth()/4, GetHeight()/4, nbr, nbr);
  fprintf(indexframe,"\n");
  fprintf(indexframe,"        </td>\n");
  fprintf(indexframe,"      </tr>\n");

  //
  // 4. prepare for next slide
  //
  SessionManager::sm->progress->IncrementProgress();
  ++nbr;
}


std::list<uint32>*  MLB::Page::GetObjectIDList() {
  std::list<uint32>* objList = new std::list<uint32>;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i)
    if (((*i)->GetType() != GenericObject::ASCIITEXT)) objList->push_back((*i)->GetID());

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) 
    objList->push_back((*j)->GetID());

  return objList;
}


RefCount<RTPI::RtpiAdu> MLB::Page::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;
  RefCount<Buffer> buf = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::PAGE, priority, buf, timestamp);
}


void MLB::Page::StoreAllObjectsInTWQ(uint32 timestamp) {
  //SessionManager::sm->message->Info("Page::StoreAllObjectsInTWQ - store now\n");

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  adus->push_back(GetStateAdu(timestamp, 3));

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;
    if ((*i)->Annotation()) continue;

    adu = (*i)->GetStateAdu(timestamp, 3);
    if (!adu) continue;

    adus->push_back(adu);
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->Annotation()) continue;

    adu = (*j)->GetStateAdu(timestamp, 3);
    if (!adu) continue;

    adus->push_back(adu);
  }

  TimeWarp::tw->InsertAdus(adus);

  SessionManager::sm->cursor->Reset();
}


void MLB::Page::SendAllObjects(uint32 timestamp, int priority) {
  RefCount<RTPI::RtpiAdu> adu;
  RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;
    if ((*i)->Annotation()) continue;

    adu = (*i)->GetStateAdu(timestamp, priority);
    if (!adu) continue;

    adus->push_back(adu);
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->Annotation()) continue;

    adu = (*j)->GetStateAdu(timestamp, 3);
    if (!adu) continue;

    adus->push_back(adu);
  }

  TimeWarp::tw->InsertAdus(adus);
  CommunicationManager::cm->SendAdusReliable(adus);
}


void MLB::Page::ChangeSubcomponentsActiveness(bool active) {
  CommunicationManager::cm->ChangeSubcomponentActiveness(GetID(), active);

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;
    if ((*i)->Annotation()) continue;

    CommunicationManager::cm->ChangeSubcomponentActiveness((*i)->GetID(), active);
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    if ((*j)->Annotation()) continue;

    CommunicationManager::cm->ChangeSubcomponentActiveness((*j)->GetID(), active);
  }

  if (onlinePointer && onlinePointer->IsVisible()) TelepointerManager::tm->ChangeSubcomponentActiveness(onlinePointer->GetID(), active);
  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k) 
    TelepointerManager::tm->ChangeSubcomponentActiveness((*k)->GetID(), active);

  if (onlineBox && onlineBox->IsVisible()) TelepointerManager::tm->ChangeSubcomponentActiveness(onlineBox->GetID(), active);
  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l) 
    TelepointerManager::tm->ChangeSubcomponentActiveness((*l)->GetID(), active);

}


int MLB::Page::FlushState(char* buf, int width, int height, Color background, std::string name, uint32 order, bool active) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  if (active)
    p8[0] = p8[0] | 1; // i am the active page

  p8[1] = CommunicationManager::PAGE;

  p32[1] = htonl(order);

  p16[4] = htons(width);
  p16[5] = htons(height);

  ColoredObject::FlushColor(buf + 12, background);

  int nlength = name.length();
  p8[15] = nlength;
  memcpy(buf + 16, name.c_str(), nlength);

  int padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[16 + nlength + i] = 0;
  }

  return 16 + nlength + padd;
}


int MLB::Page::GetStateSize(std::string name) {
  int nlength = name.length();
  return 16 + nlength + ((nlength % 4 == 0) ? 0 : 4 - (nlength % 4));
}


int MLB::Page::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  if (SessionManager::sm->GetActivePage() && (SessionManager::sm->GetActivePage() == this))
    p8[0] = p8[0] | 1; // i am the active page

  p8[1] = CommunicationManager::PAGE;

  p32[1] = htonl(order);

  p16[4] = htons(width);
  p16[5] = htons(height);

  ColoredObject::FlushColor(buf + 12, background);

  int nlength = containerName.length();
  p8[15] = nlength;
  memcpy(buf + 16, containerName.c_str(), nlength);

  int padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[16 + nlength + i] = 0;
  }

  return 16 + nlength + padd;
}


int MLB::Page::GetStateSize() {
  int nlength = containerName.length();
  return 16 + nlength + ((nlength % 4 == 0) ? 0 : 4 - (nlength % 4));
}


int MLB::Page::ParseState(char* buf, bool& active) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    SessionManager::sm->message->Error("Page::ParseCreate - wrong version of mlb protocol");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    SessionManager::sm->message->Error("Page::ParseCreate - wrong operation type");
    return -1;
  }

  if (p8[1] != CommunicationManager::PAGE) {
    SessionManager::sm->message->Error("Page::ParseCreate - wrong object type %d", (int)p8[1]);
    return -1;
  }

  active = (p8[0] & MLB_MASK_A);

  order = ntohl(p32[1]);

  width  = ntohs(p16[4]);
  height = ntohs(p16[5]);

  background = ColoredObject::ParseColor(buf + 12);

  int nlength = p8[15];
  char* tmp2 = new char[nlength + 1];
  memcpy(tmp2, buf + 16, nlength);
  tmp2[nlength] = '\0';
  containerName = string(tmp2);
  delete[] tmp2;

  int padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);

  return 16 + nlength + padd;
}


int MLB::Page::FlushSetActive(char* buf, bool active) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::SET_ACTIVE;

  if (active) p8[3] = 1;
  else p8[3] = 0;

  return 4;
}


int MLB::Page::FlushChangeSize(char* buf, int width, int height) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::CHG_SIZE;
  p8[3] = 0;

  p16[2] = htons(width);
  p16[3] = htons(height);

  return 8;
}


int MLB::Page::FlushChangeColor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::CHG_BACKGRCOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Page::FlushChangeName(char* buf, string name) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::CHG_NAME;

  int nlength = name.length();
  p8[3] = nlength;

  memcpy(buf + 4, name.c_str(), nlength);

  int padd = (nlength % 4 == 0) ? 0 : 4 - (nlength % 4);
  if (padd != 0) { // padding needed
    for (int i = 0; i < padd; ++i)
      p8[4 + nlength + i] = 0;
  }

  return 4 + nlength + padd;
}


int MLB::Page::GetChangeNameSize(string name) {
  int nlength = name.length();
  return 4 + nlength + ((nlength % 4 == 0) ? 0 : 4 - (nlength % 4));
}


int MLB::Page::FlushChangeParentObject(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::CHG_PARENT;

  p32[1] = htonl(order);

  return 12;
}


int MLB::Page::FlushChangeActiveItem(char* buf) {
  uint8*  p8  = (uint8*)buf;
  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::PAGE;

  p8[2] = CommunicationManager::CHG_ACTIVE;
  p8[3] = 0;

  return 4;
}


int MLB::Page::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::PAGE;

  return 4;
}


void MLB::Page::DeleteObjectsFromRTPI() {
  CommunicationManager::cm->UnregisterSubcomponent(GetID());
  
  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;
    CommunicationManager::cm->UnregisterSubcomponent((*i)->GetID());
  }

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    CommunicationManager::cm->UnregisterSubcomponent((*j)->GetID());
  }

  for (std::list<Telepointer*>::iterator k = pointers.begin(); k != pointers.end(); ++k)
    TelepointerManager::tm->UnregisterSubcomponent((*k)->GetID());

  for (std::list<Telebox*>::iterator l = boxes.begin(); l != boxes.end(); ++l)
    TelepointerManager::tm->UnregisterSubcomponent((*l)->GetID());

  if (onlinePointer && onlinePointer->IsVisible()) TelepointerManager::tm->UnregisterSubcomponent(onlinePointer->GetID());
  if (onlineBox && onlineBox->IsVisible()) TelepointerManager::tm->UnregisterSubcomponent(onlineBox->GetID());
}


void MLB::Page::ReplaceIDs() {
  uint32 newID;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    newID = CommunicationManager::cm->GetObjectID();
    (*i)->SetID(newID);
  }  

  for (std::list<Group*>::iterator j = groups.begin(); j != groups.end(); ++j) {
    newID = CommunicationManager::cm->GetObjectID();
    (*j)->SetID(newID);
  }  
}


void MLB::Page::DeleteEmptyTextObjects() {
  RefCount<Buffer> buf;
  RefCount<RTPI::RtpiAdu> adu;
  Text* txt;
  Group* group;

  for (std::list<DrawObject*>::iterator i = objects.begin(); i != objects.end(); ++i) {
    if ((*i)->GetType() == GenericObject::ASCIITEXT) {
      txt   = (Text*)*i;
      group = (Group*)txt->GetGroup();

      if (group->AllEmpty()) {
	if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !group->Annotation()) {
	  buf   = new Buffer(group->GetDeleteSize());
	  group->FlushDeleteTG(buf->getCont());
	  adu   = CommunicationManager::cm->CreateEvent(group->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}
	else { // PRIVATE session or Annotation Mode
	  group->Hide(GetCanvas());
  	  group->DeleteSelection(GetCanvas());
  	  group->DeleteTextCursor(GetCanvas());
	  std::vector<uint32>* groupObjects = group->GetObjectIDs();
	  for (std::vector<uint32>::iterator j = groupObjects->begin(); j != groupObjects->end(); ++j) 
	    delete DeleteObject(*j);
	  group->DeleteAll(GetCanvas());
	  DeleteObject(group);
	  delete group;
  	  i = objects.begin(); // reposition iterator because object list has been changed
	}
      }
    }
  }
}

//-------------------------------------------------------------------------//

void MLB::Page::SetAnimationEditMode( bool mode )
{
  if( mode ) animations.Reset();
  else       animations.RunFirstStep( true );

} // SetAnimationEditMode()

//=========================================================================//
