/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_page.h
 authors : Juergen Vogel, Gabriela Maass
 date    : 18.02.2000
 desc    :

 **********************************************************************************************************/

#ifndef __MLB_PAGE__
#define __MLB_PAGE__

#ifdef WIN32
#  pragma warning (disable: 4786)
#endif

#include <list>
#include <vector>

#include "sgml.h"
#include "mlb_generic.h"
#include "mlb_objects.h"
#include "mlb_telepointer.h"
#include "mlb_gui.h"
#include "mlb_record.h"
#include "mlb_animations.h"

namespace MLB {

  class Page : public GenericContainer  {
     protected:
      friend class SessionManager;
      friend class Session;

      GUI::Canvas* canvas;                      // gui representation of page
      GUI::PageView* pageView;                  // gui representation of document tree

      std::list<DrawObject*> objects;           // all objects ever seen on this page
      std::list<Group*>      groups;            // all object groups that are on this page
      std::list<DrawObject*> selectedObjects;   // currently selected objects
      DrawObject* active;                       // currently edited object

      Telepointer* onlinePointer;               // my own pointer;
      std::list<Telepointer*> pointers;         // pointers of remote participants

      Telebox* onlineBox;                       // my own box
    	std::list<Telebox*> boxes;                // boxes of remote participants

      int pageNumber;
      int firstTime;
      int width;
      int height;
      Color background;   
      int sessionType;

      Animations animations;

    public:

      enum type {
	A4_WIDTH  = 594,
	A4_HEIGHT = 842
      };

      Page(uint32 _id);
      Page(uint32 _id, int _sessionType, char *_title="Page", int _width=594, int _height=842, string color = "#ffffff");
      Page(uint32 _id, int _sessionType, SGMLElement *root);
      ~Page();

      void PrintAllObjects();
      void PrintActiveSubcomponents();
      
      /*** General Functions ***/
      virtual int GetType() {return MLB::GenericObject::PAGE;}

      /*** General Functions ***/
      virtual void SetName(char *_name);
      virtual GenericContainer* SetCurrentContainer(std::vector<uint32>* idVector, int pos);

      /*** Draw ***/
      virtual void DrawPageTree(std::string _nodeNumber, std::string _nodeData, std::string parentNode,int pos,bool *open);
      virtual void ClearPageTree();

      /*** Save Functions ***/
      virtual void SaveContainer(FILE *fp, int level);
      virtual void SaveAsPs(FILE *fp);
      virtual void SaveAsHtml(char* directory, char* docname, FILE* indexframe, int& nbr, int last);
      virtual void SaveAsImage(char* name, char* type, int& nbr);

      /*** Find Functions ***/
      virtual GenericObject* GetObject(uint32 id);
      DrawObject* GetObject(const string& tkRef);
      virtual GenericObject* GetGroup(uint32 id);
      virtual GenericContainer* GetContainer(uint32 id);

      virtual uint32 GetObjectOrder(uint32 id);
      virtual uint32 GetPushBackObjectOrder();
      virtual uint32 GetEndObjectOrder();
      virtual uint32 GetFirstObjectOrder();
      virtual uint32 GetNextObjectOrder(uint32 id);
      virtual GenericObject* GetObjectWithOrder(uint32 order);

      virtual GenericContainer* GetFirstPage() {return NULL;}

      DrawObject* GetActiveObject();
      void SetActiveObject(DrawObject* obj);
 
      PhotoImage* GetFirstImage();

      /*** Animation functions ***/
      Animations& GetAnimations() { return animations; };
      virtual void SetAnimationEditMode( bool mode );

      /*** Telepointer Functions ***/
      Telepointer* GetTelepointer(uint32 id);
      Telepointer* DeleteTelepointer(uint32 id);
      virtual void DeleteTelepointerOfParticipant(uint32 participantID);
      void AddTelepointer(Telepointer* tp);

      Telebox* GetTelebox(uint32 id);
      Telebox* DeleteTelebox(uint32 id);
      virtual void DeleteTeleboxOfParticipant(uint32 participantID);
      void AddTelebox(Telebox* tp);

      /*** AOF-Recording Functions ***/
      virtual void ResetRecording();
      void WriteObjectFile();

      /*** General Functions ***/
      void Init();
      void InitTelepointer();
      void InitTelebox();
      void LoadObjects(SGMLElement *root);
      void CopyObjects(int argc, const char* argv[]);
      void ShowTicks(int argc, const char* argv[]);

      /*** Attributes ***/
      void SetPageDimension(int _width, int _height);
      void SetBackgroundColor(Color color);
      void SetWindowPointer(int argc, const char* argv[]);
      void UnSetWindowPointer(int argc, const char* argv[]);

      /*** Return private variables ***/
      GUI::Canvas* GetCanvas() {return canvas;}
      int  GetWidth() {return width;}
      int  GetHeight() {return height;}
      Color GetColor() {return background;}

      /*** Functions concerned objects ***/
      void CutObjects();
      void CopyObjects();
      void PasteObjects();

      /*** Draw ***/
      void Draw();
      void ReDraw();
      void Activate(bool rebuild=false);
      void DeActivate();
      void DrawSelected();

      /*** Insert ***/
      void AppendObject(DrawObject *obj);
      //     void AppendGroup(Group *grp);
      void AppendNewObject(int argc, char* argv[], int _type);

      /*** Select ***/
      inline DrawObject::object_list GetObjects        () { return objects;                   }
      inline DrawObject::group_list  GetGroups         () { return groups;                    }
      inline DrawObject::object_list GetSelectedObjects() { return selectedObjects;           }
      inline bool                    HasSelectedObjects() { return !selectedObjects.empty();  }

      void SelectObject (DrawObject *obj);
      void SelectObjects(int argc, const char* argv[]);
      void SelectObjects(int x, int y, int status);
      void UnSelectObjects();

      /*** Change ***/
      void ChangeObject(int argc, const char* argv[], int _type);
      void ChangeFont(int argc, char* argv[]);
      void ChangeLineColor(int argc, const char* argv[]);
      void ChangeFillColor(int argc, const char* argv[]);
      void ChangeLineWidth(int argc, const char* argv[]);
      void ChangeLineStyle(int argc, const char* argv[]);
      void ChangeCapStyle(int argc, const char* argv[]);
      void ChangeJoinStyle(int argc, const char* argv[]);
      void ChangeArrowStyle(int argc, const char* argv[]);
      void ChangeArrowShape(int argc, const char* argv[]);
      void MoveObjects(int argc, const char* argv[]);
      void InsertTextLine(int argc, const char* argv[]);
      void EndText(int argc, const char* argv[]);
      void DeleteEmptyTextObjects();
      void InsertImage(Magick::Image& image, char *filename, int x, int y, bool single=true);
      void PlaceImage(int argc, const char* argv[]);
      void ActivateOnlinePointer(int argc, const char* argv[]);
      void MLB::Page::ActivateOnlineBox(int argc, const char* argv[]);
      void RaiseObjects();
      void LowerObjects();

      std::list<uint32>* GetObjectIDList();

      /*** Delete ***/
      void DeleteObject(int argc, const char* argv[]);
      GenericObject* DeleteObject(uint32 id);
      GenericObject* DeleteObject(GenericObject *obj);
      GenericObject* RemoteDeleteObject(uint32 id);
      GenericObject* RemoteDeleteObject(DrawObject* obj);
      GenericObject* DeleteObjectFromContainer(GenericObject* obj);
      virtual void DeleteAll(GUI::Canvas* _canvas);
      virtual void DeleteObjectsFromRTPI();
      void Clear();

      void SendAllObjects(uint32 timestamp, int priority = 3);
      void StoreAllObjectsInTWQ(uint32 timestamp);
      RefCount<RTPI::RtpiAdu> GetStateAdu(uint32 timestamp, int priority);

      void ChangeSubcomponentsActiveness(bool active);
      void ReplaceIDs();

      static int FlushState(char* buf, int width, int height, Color background, std::string name, uint32 order, bool active);
      int FlushState(char* buf);
      static int GetStateSize(std::string name);
      int GetStateSize();
      int ParseState(char* buf, bool& active);

      static int FlushSetActive(char* buf, bool active = true);
      static int GetSetActiveSize() {return 4;}
      static int FlushChangeParentObject(char* buf, uint32 order);
      static int GetChangeParentObjectSize() {return 12;}

      int FlushChangeSize(char* buf, int width, int height);
      int GetChangeSizeSize() {return 8;}
      int FlushChangeColor(char* buf, Color color);
      int GetChangeColorSize() {return 8;}
      int FlushChangeName(char* buf, string name);
      int GetChangeNameSize(std::string name);
      int FlushChangeActiveItem(char* buf);
      int GetChangeActiveItemSize() {return 4;}

      static int FlushDeleteS(char* buf);
      int FlushDelete(char* buf) { return FlushDeleteS(buf); }
      
      /*** AOF-Recording Functions ***/
      void SaveAsAOF(int _time=-1);
      void SavePointAsAOF(PolyObject *obj, int _x, int _y, int first=false);
      Telepointer* GetLocalTelepointer() {return onlinePointer;}
      int RecordableObjects();
      int GetPageNumber() {return pageNumber;}
      void WriteHtml(Record::AOF *_aofRecording, Page *first, Page *prev, Page *next, Page *last);
    };
}

#endif
