/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_participants.cpp
 authors : Juergen Vogel
 date    : 
 desc    : 

 **********************************************************************************************************/


#ifdef WIN32
#  pragma warning (disable: 4786 4181)
#endif

#include <util/tools.h>

#include <mlb.h>
#include <mlb_gui.h>
#include <mlb_participants.h>
#include <mlb_session.h>

#define DEBUG

const uint32 MLB::Participant::defaultDelay = 100;
const double MLB::Participant::alpha = 0.7;

MLB::ParticipantManager* MLB::ParticipantManager::pm = NULL;


MLB::ParticipantManager::~ParticipantManager() {
 for (int i = 0; i < participants.size(); ++i)
    delete participants[i];
}

void MLB::ParticipantManager::AddParticipant(uint32 pid) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) {
      // participant already exists
      if ((*i)->status == Participant::EXITED) {
	UpdateParticipantActiveness(pid);
	ParticipantExitIndicatorOff(pid);
      }
      return; 
    }

  #ifdef DEBUG
  GUI::MessageWindow::Info("ParticipantManager::AddParticipant - add %u\n", pid);
  #endif

  participants.push_back(new Participant(pid));
}


void MLB::ParticipantManager::RemoveParticipant(uint32 pid) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) {
      #ifdef DEBUG
      GUI::MessageWindow::Info("ParticipantManager::RemoveParticipant - remove %u\n", pid);
      #endif

      (*i)->status = Participant::EXITED;
      ParticipantExitIndicatorOn(pid);

      return;
    }

  // error, participant not found
}


void MLB::ParticipantManager::RemoveExitedParticipant(uint32 pid) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) {
      #ifdef DEBUG
      GUI::MessageWindow::Info("ParticipantManager::RemoveExitedParticipant - remove %u\n", pid);
      #endif

      delete *i;
      participants.erase(i);
      DisplayParticipantList();

      return;
    }

  // error, participant not found
}


void MLB::ParticipantManager::RemoveAllParticipants() {
  #ifdef DEBUG
  GUI::MessageWindow::Info("ParticipantManager::RemoveAllParticipants\n");
  #endif

  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i) 
    delete *i;

  participants.clear();

  DisplayParticipantList();
}


void MLB::ParticipantManager::RemoveAllExitedParticipants() {
  #ifdef DEBUG
  GUI::MessageWindow::Info("ParticipantManager::RemoveAllExitedParticipants\n");
  #endif

  std::vector<Participant*>::iterator i = participants.begin();
  while (i != participants.end()) {
    if ((*i)->status == Participant::EXITED) {
      delete *i;
      i = participants.erase(i);
    } else {
      ++i;
    }
  }

  DisplayParticipantList();
}


void MLB::ParticipantManager::UpdateParticipant(uint32 pid, std::string cname, std::string name, std::string email, 
						std::string phone, std::string location, std::string tool) {

  // check if there is someone with the same cname 
  for (std::vector<Participant*>::iterator j = participants.begin(); j != participants.end(); ++j) {
    if (((*j)->cname == cname) && ((*j)->pid != pid) && ((*j)->status == Participant::EXITED)) {
      // we already know this guy - he reconnected after a crash and has a new pid now
      RemoveExitedParticipant((*j)->pid);
      DisplayParticipantList();
      break;
    }
  }

  // update the participants info
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i) {
    if ((*i)->pid == pid) {
      std::string oldName  = (*i)->name;
      std::string oldCName = (*i)->cname;
      (*i)->cname = cname;
      (*i)->name  = name;
      (*i)->email = email;
      (*i)->phone = phone;
      (*i)->location = location;
      (*i)->tool  = tool;

      if ((oldName != name) || (oldCName != cname)) 
	DisplayParticipantList();

      return;
    }
  }

  // error, participant not found
}


MLB::Participant* MLB::ParticipantManager::GetParticipant(uint32 pid) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) 
      return *i;

  return NULL; // not found
}


int MLB::ParticipantManager::GetIndex(uint32 pid) {
  int index = 0;
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i, ++index) 
    if ((*i)->pid == pid) return index;
  return participants.size();
}


void MLB::ParticipantManager::DisplayParticipant(uint32 pid) {
  Participant* p = GetParticipant(pid);
  if (!p) return;

  if (Et_GlobalEvalF(Et_Interp,"Participant::DisplayParticipant \"%q\" \"%q\" \"%q\" \"%q\" \"%q\" \"%q\"", 
		     p->cname.c_str(), p->name.c_str(), p->email.c_str(), p->phone.c_str(), p->location.c_str(), p->tool.c_str()) != TCL_OK)
    GUI::MessageWindow::Error("Participant::DisplayParticipant - tcl error \"%q\"\n", Et_Interp->result);
}


void  MLB::ParticipantManager::UpdateParticipantActiveness(uint32 pid) {
  Participant* p = GetParticipant(pid);
  if (!p) return;
  
  if (TimerManager::tm->TimerRunning(this, TimerObject::HALF_ACTIVENESS, pid)) {
    TimerManager::tm->DeleteTimer(this, TimerObject::HALF_ACTIVENESS, pid);
  }
  if (TimerManager::tm->TimerRunning(this, TimerObject::FULL_ACTIVENESS, pid)) {
    TimerManager::tm->DeleteTimer(this, TimerObject::FULL_ACTIVENESS, pid);
  }

  int runningTime = 2000;
  p->status = Participant::FULL_ACTIVE;
  DisplayActiveness(pid);
  TimerManager::tm->SetTimer(this, new TimerObject(runningTime, TimerObject::FULL_ACTIVENESS, pid, 0));
}


void MLB::ParticipantManager::DisplayActiveness(uint32 pid) {
  Participant* p = GetParticipant(pid);
  if (!p) return;

  switch (p->status) {
  case Participant::FULL_ACTIVE: {
    if (Et_GlobalEvalF(Et_Interp, "Participant::SetColor %u #000 #000", p->pid) != TCL_OK) 
      GUI::MessageWindow::Error("ParticipantManager::DisplayActiveness - tcl error \"%q\"\n", Et_Interp->result);
    break;
  }
  case Participant::HALF_ACTIVE: {
    if (Et_GlobalEvalF(Et_Interp, "Participant::SetColor %u #3C3C3C #3C3C3C",  p->pid) != TCL_OK) 
      GUI::MessageWindow::Error("ParticipantManager::DisplayActiveness - tcl error \"%q\"\n", Et_Interp->result);
    break;
  }
  case Participant::INACTIVE: {
    if (Et_GlobalEvalF(Et_Interp, "Participant::SetColor %u #787878 #787878",  p->pid) != TCL_OK) 
      GUI::MessageWindow::Error("ParticipantManager::DisplayActiveness - tcl error \"%q\"\n", Et_Interp->result);
    break;
  }
  }
}


void MLB::ParticipantManager::DisplayParticipantList() {
  if (participants.size() == 0) {
    if (Et_GlobalEvalF(Et_Interp,"Participant::ClearParticipantList") != TCL_OK) // clear list
      GUI::MessageWindow::Error("ParticipantManager::ClearParticipantList - tcl error \"%q\"\n", Et_Interp->result);
    return;
  }
  
  for (std::vector<Participant*>::iterator p = participants.begin(); p != participants.end(); ++p) {
    Color tpColor = "none";
    if (SessionManager::sm->GetActivePage()) {
      Telepointer* tp = SessionManager::sm->GetActivePage()->GetTelepointer((*p)->pid);
      if (tp) tpColor = tp->GetLineColor();
    }

    std::string exited = ((*p)->status == Participant::EXITED) ? "1" : "0";

    if (!((*p)->name.empty())) {
      if (Et_GlobalEvalF(Et_Interp, "lappend pList %u \"%q\" %s %s", (*p)->pid, (*p)->name.c_str(), tpColor.c_str(), exited.c_str()) != TCL_OK)
	GUI::MessageWindow::Error("ParticipantManager::DisplayParticipantList - tcl error \"%q\"\n", Et_Interp->result);
    } else { // cname always there
      if (Et_GlobalEvalF(Et_Interp, "lappend pList %u \"%q\" %s %s", (*p)->pid, (*p)->cname.c_str(), tpColor.c_str(), exited.c_str()) != TCL_OK)
	GUI::MessageWindow::Error("ParticipantManager::DisplayParticipantList - tcl error \"%q\"\n", Et_Interp->result);
    }
  }

  if (Et_GlobalEvalF(Et_Interp,"Participant::DisplayParticipantList $pList") != TCL_OK) 
    GUI::MessageWindow::Error("ParticipantManager::DisplayParticipantList - tcl error \"%q\"\n", Et_Interp->result);

  if (Et_GlobalEvalF(Et_Interp,"unset pList") != TCL_OK) 
    GUI::MessageWindow::Error("ParticipantManager::DisplayParticipantList - tcl error \"%q\"\n", Et_Interp->result);

  //for (int index = 0; index < participants.size(); ++index) 
  //if (participants[index]->status != Participant::INACTIVE) DisplayActiveness(index);
}


uint32 MLB::ParticipantManager::GetDelay(uint32 pid) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) { 
      if ((*i)->delay == 0) return Participant::defaultDelay;
      else return (*i)->delay;
    }

  return Participant::defaultDelay; // not found
}


void MLB::ParticipantManager::AddDelay(uint32 pid, uint32 timestamp) {
  uint32 delay = rtpitime() - timestamp;

  if (delay < 0) delay *= -1; //This prevents that participants with unsynchronized clocks destroy the time calculation
  
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->pid == pid) {
      if ((*i)->delay == 0) {
	// first delay
	(*i)->delay = delay;
      } else {
	// exponentielle Glttung
	(*i)->delay = Participant::alpha * delay + (1 - Participant::alpha) * (*i)->delay;
      }
      return;
    }
}  

uint32 MLB::ParticipantManager::GetMaxDelay() {
  uint32 max = Participant::defaultDelay;

  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    if ((*i)->delay > max) 
      max = (*i)->delay;

  return max;
}


uint32 MLB::ParticipantManager::GetAverageDelay() {
  if (participants.empty()) return Participant::defaultDelay;

  uint32 sum = 0;

  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i)
    sum += (*i)->delay;

  return (double)sum / participants.size();
  

}


void MLB::ParticipantManager::ExecuteTimer(TimerObject* timer) {
  for (std::vector<Participant*>::iterator i = participants.begin(); i != participants.end(); ++i) {
    if ((*i)->pid == timer->id) {
      switch (timer->type) {
      case TimerObject::FULL_ACTIVENESS: {
        (*i)->status = Participant::HALF_ACTIVE;
        int runningTime = 2000;
        TimerManager::tm->SetTimer(this, new TimerObject(runningTime, TimerObject::HALF_ACTIVENESS, (*i)->pid, 0));
        break;
      }
      case TimerObject::HALF_ACTIVENESS: {
        (*i)->status = Participant::INACTIVE;
        break;
      }
      }
      DisplayActiveness((*i)->pid);
      break;
    }
  } 
  delete timer;
}


void MLB::ParticipantManager::ParticipantExitIndicatorOn(uint32 pid) {
  if (Et_GlobalEvalF(Et_Interp,"Participant::ExitIndicatorOn %u", pid) != TCL_OK) 
    GUI::MessageWindow::Error("ParticipantManager::ExitIndicatorOn - tcl error %q\n", Et_Interp->result);    
}


void MLB::ParticipantManager::ParticipantExitIndicatorOff(uint32 pid) {
  if (Et_GlobalEvalF(Et_Interp,"Participant::ExitIndicatorOff %u", pid) != TCL_OK) 
    GUI::MessageWindow::Error("ParticipantManager::ExitIndicatorOff - tcl error %q\n", Et_Interp->result);    
}
