/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_polyobjects.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 3.12.1999
 desc    : Functions to draw and manipulate objects on a tcl-canvas

 **********************************************************************************************************/


#include "mlb.h"

#include "mlb_page.h"
#include "mlb_session.h"
#include "mlb_objects.h"
#include "mlb_adu.h"
#include "mlb_gui.h"


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

MLB::PolyObject::PolyObject(int _x, int _y, int _width, int _height, char *_color,
			    char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id)
  : ColoredObject(_x,_y,_width,_height,_color,_fillColor,_lineWidth,_lineStyle,_id) {
  selectedPoint = points.end();
  lastZoom = 0.0;
  layer = 1;
  recorded = false;
}


MLB::PolyObject::~PolyObject() {
  for (std::list<Point *>::iterator i = points.begin(); i != points.end(); ++i) delete *i;
  for (std::list<Point *>::iterator j = _points.begin(); j != _points.end(); ++j) delete *j;
}


void MLB::PolyObject::Move(GUI::Canvas *_canvas, int diffX, int diffY, bool redraw) {
  if (!visible) return;

  recorded = false;

  x += diffX;
  y += diffY;

  for (std::list<Point *>::iterator i = points.begin(); i != points.end(); ++i)  {
    (*i)->x += diffX;
    (*i)->y += diffY;
  }

  if (redraw) {
    _canvas->MoveObject(tkref.c_str(),SessionManager::sm->GetZoom()*diffX,SessionManager::sm->GetZoom()*diffY);
    if (selected) DrawSelection(_canvas);
  }
}


void MLB::PolyObject::DrawSelection(GUI::Canvas *_canvas) {
  if (!visible) return;

  if (selected) DeleteSelection(_canvas);

  selected = true;

  for (std::list<Point *>::iterator i = points.begin(); i != points.end(); ++i)  {
    _canvas->DrawSelectionRectangle(tkref.c_str(),
				    SessionManager::sm->GetZoom()*(*i)->x,
				    SessionManager::sm->GetZoom()*(*i)->y);
  }
}


void MLB::PolyObject::DeleteSelection(GUI::Canvas *_canvas) {
  if (!visible || !selected) return;

  selected = false;
  _canvas->DeleteSelection(tkref.c_str());
}


int MLB::PolyObject::GetSelectedPoint() {
  if (!visible) return -1;
  if (selectedPoint == points.end()) return -1;

  int point = 0;
  for (std::list<Point*>::iterator i = points.begin(); i != selectedPoint; ++i) ++point;
  return point;
}


bool MLB::PolyObject::IsEditablePoint(int _x, int _y) {
  if (!visible) return false;

  selectedPoint = points.end();

  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
    if ((_x > (*i)->x-TOLERANZ) && (_x < (*i)->x+TOLERANZ)) {
      if ((_y > (*i)->y-TOLERANZ) && (_y < (*i)->y+TOLERANZ)) {
	selectedPoint = i;
	return true;
      }
    }
  }

  return false;
}


void MLB::PolyObject::MovePoint(GUI::Canvas *_canvas, int point, int _x, int _y, bool redraw) {
  if (!visible) return;
  if (point >= points.size()) return;

  recorded = false;

  selectedPoint = points.begin();
  for (int j = 0; j < point; ++j) ++selectedPoint;

  (*selectedPoint)->x = _x;
  (*selectedPoint)->y = _y;

  // recalculate coords and width/height
  int x1 = _x;
  int x2 = _x;
  int y1 = _y;
  int y2 = _y;

  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
    if ((*i)->x < x1) x1 = (*i)->x;
    if ((*i)->y < y1) y1 = (*i)->y;
    if ((*i)->x > x2) x2 = (*i)->x;
    if ((*i)->y > y2) y2 = (*i)->y;
  }

  x = x1;
  y = y1;
  width  = x2-x1;
  height = y2-y1;

  if (redraw) ReDraw(_canvas);
}


void MLB::PolyObject::AppendNewPoint(int _x, int _y) {
  if (!visible) return;

  // recalculate coords and widht / height
  int x2 = x + width;
  int y2 = y + height;

  if (_x < x) x = _x;
  if (_y < y) y = _y;
  if (_x > x2) x2 = _x;
  if (_y > y2) y2 = _y;

  width  = x2 - x;
  height = y2 - y;

  points.push_back(new Point(_x, _y));
}


void MLB::PolyObject::DrawNewSegment(GUI::Canvas *_canvas) {
  if (!visible) return;

  if (points.size() > 1) {
    std::list<Point*>::iterator i1 = points.end(); --i1;
    std::list<Point*>::iterator i2 = i1; --i2;

    _canvas->DrawLineSegment(tkref.c_str(),
			     SessionManager::sm->GetZoom() * (*i2)->x,
			     SessionManager::sm->GetZoom() * (*i2)->y,
			     SessionManager::sm->GetZoom() * (*i1)->x,
			     SessionManager::sm->GetZoom() * (*i1)->y,
			     color.c_str(),
			     SessionManager::sm->GetZoom() * lineWidth,
			     lineStyle.c_str());
  }
}


void MLB::PolyObject::DeleteSegments(GUI::Canvas *_canvas) {
  if (!visible) return;
  _canvas->DeleteLineSegments(tkref.c_str());
}


bool MLB::PolyObject::Overlap(GUI::Canvas *_canvas, int _x, int _y) {
  if (!visible) return false;

  int x1, y1, x2, y2;
  GetBBox(&x1,&y1,&x2,&y2);

  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return false;
  return _canvas->Overlap(tkref.c_str(),SessionManager::sm->GetZoom()*_x,SessionManager::sm->GetZoom()*_y);
}


/*********************************************************************/
/*********************************************************************/
/*********************************************************************/

MLB::PolyLine::PolyLine(uint32 id, bool _isFreehand)
  : PolyObject(id), smooth(false), capStyle(""), joinStyle(""), arrowStyle(""), arrowShape("")
{
  char
    buf[10];

  sprintf(buf,"%u",id);
  tkref = "PolyLine";
  tkref += buf;

  isFreehand = _isFreehand;
}

MLB::PolyLine::PolyLine(int _x, int _y, int _width, int _height, char *_color,
			int _lineWidth, char *_lineStyle, char *_capStyle, char *_joinStyle,
			char *_arrowStyle, char *_arrowShape, bool _smooth, uint32 _id, bool _isFreehand)
  : PolyObject(_x,_y,_width,_height,_color,"none",_lineWidth,_lineStyle,_id)
{
  char
    buf[10];

  sprintf(buf,"%u",_id);
  tkref = "PolyLine";
  tkref += buf;
  smooth = _smooth;

  capStyle    = _capStyle;
  joinStyle   = _joinStyle;
  arrowStyle  = _arrowStyle;
  arrowShape  = _arrowShape;

  isFreehand = _isFreehand;

}


MLB::PolyLine::~PolyLine() {
}


int MLB::PolyLine::GetType() {
  if (isFreehand)
    return GenericObject::FREEHAND;
  else
    return GenericObject::POLYLINE;

}


MLB::DrawObject* MLB::PolyLine::Copy() {
  PolyLine* obj = new PolyLine(x,y,width,height,(char *)color.c_str(),lineWidth,(char *)lineStyle.c_str(),
		     (char *)capStyle.c_str(),(char *)joinStyle.c_str(),(char *)arrowStyle.c_str(),
		     (char *)arrowShape.c_str(),smooth,CommunicationManager::cm->GetObjectID());

  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
    obj->AppendNewPoint((*i)->x,(*i)->y);
  }

  return obj;
}


void MLB::PolyLine::Draw(GUI::Canvas *_canvas) {
  if (!visible) return;

  // calculate arrowShape
  static char as[12];
  unsigned a1, a2, a3;
  sscanf(arrowShape.c_str(), "%u %u %u", &a1, &a2, &a3);
  a1 *= SessionManager::sm->GetZoom();
  a2 *= SessionManager::sm->GetZoom();
  a3 *= SessionManager::sm->GetZoom();
  sprintf(as, "%u %u %u", a1, a2, a3);

  if ((lastZoom != SessionManager::sm->GetZoom()) || (points.size() != _points.size())) {
    lastZoom = SessionManager::sm->GetZoom();
  
  // calculate points
  for (std::list<Point*>::iterator j = _points.begin(); j != _points.end(); ++j) delete *j;
  _points.clear();
  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i)
    _points.push_back(new Point(SessionManager::sm->GetZoom() * (*i)->x, SessionManager::sm->GetZoom() * (*i)->y));
  }

  _canvas->DrawPolyLine(tkref.c_str(),
			_points,
			color.c_str(),
			SessionManager::sm->GetZoom()*lineWidth,
			lineStyle.c_str(),
			capStyle.c_str(),
			joinStyle.c_str(),
			arrowStyle.c_str(),
			as);

  if (selected) DrawSelection(_canvas);
}


void MLB::PolyLine::ReDraw(GUI::Canvas *_canvas) {
  if (!visible) return;

  _canvas->DeleteObject(tkref.c_str());
  _canvas->DeleteLineSegments(tkref.c_str());

  Draw(_canvas);
}


void MLB::PolyLine::Hide(GUI::Canvas *canvas) {
  if (!visible) return;

  if (selected) DeleteSelection(canvas);
  canvas->DeleteObject(tkref.c_str());
  canvas->DeleteLineSegments(tkref.c_str());
  visible = false;

}


void MLB::PolyLine::SaveObject(FILE *fp, int level) {
  std::list<Point *>::iterator
    temp;
  std::string
    ptBuf,
    tab;
  int
    i;
  char
    buf[20];

  if (!visible) return;

  for (i=0; i<level; i++)
    tab += "  ";

  temp = points.begin();
  while (temp != points.end()) {
    sprintf(buf,"{%d %d} ",(*temp)->x,(*temp)->y);
    ptBuf += buf;
    temp++;
  }
  fprintf(fp,"%s<POLYLINE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\" length=\"%d\"%s>\n",
	  tab.c_str(),x,y,width,height,color.c_str(),lineWidth,lineStyle.c_str(),ptBuf.size(), GetAnimationStepSGMLText().c_str() );


  fprintf(fp,"%s  \"",tab.c_str());
  fwrite(ptBuf.c_str(),1,ptBuf.size(),fp);
  fprintf(fp,"\"\n%s</POLYLINE>\n",tab.c_str());

  ptBuf.erase();
  tab.erase();
}

void MLB::PolyLine::SaveAsPs(FILE *fp, int pageWidth, int pageHeight, bool rotate)
{
  std::list<Point *>::iterator
    temp;
  XColor
    *xColor;
  float
    red, green, blue;

  if (!visible) return;

  if (points.size() <= 1) return;

  xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)color.c_str());
  red   = (float)xColor->red/65535;
  green = (float)xColor->green/65535;
  blue  = (float)xColor->blue/65535;

  fprintf(fp,"\n");
  fprintf(fp,"gsave\n");
  fprintf(fp,"%d setlinewidth 0 setlinejoin 2 setlinecap\n",lineWidth);
  fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
  fprintf(fp,"newpath\n");
  temp = points.begin();
  if (rotate)
    fprintf(fp,"%d %d moveto\n",(*temp)->y,(*temp)->x);
  else
    fprintf(fp,"%d %d moveto\n",(*temp)->x,pageHeight-(*temp)->y);
  temp++;
  while (temp != points.end()) {
    if (rotate)
      fprintf(fp,"%d %d lineto\n",(*temp)->y,(*temp)->x);
    else
      fprintf(fp,"%d %d lineto\n",(*temp)->x,pageHeight-(*temp)->y);
    temp++;
  }
  fprintf(fp,"stroke\n");
  fprintf(fp,"grestore\n");

}

int MLB::PolyLine::RecordableObjects()
{
  if (!visible) return 0;

  return 1;
}

void MLB::PolyLine::SaveAsAOF(FILE *fp)
{
#ifdef _DEBUG1
  cout << "MLB::PolyLine::SaveAsAOF" << endl;
#endif

  std::list<Point *>::iterator
    temp;

  if (!visible || recorded) return;

  recorded = true;

  recID = MLB::SessionManager::sm->recCount;
  MLB::SessionManager::sm->recCount++;
  fprintf(fp,"<FREEHAND x=%d y=%d width=%d height=%d fcolor=%d count=%d linewidth=%d linestyle=%d layer=%d>\n",
	  x,y,width,height,CalculateColor(color),points.size(),lineWidth,CalculateLineStyle(),layer);
  temp = points.begin();
  while (temp != points.end()) {
    fprintf(fp,"<pt x=%d y=%d></pt>\n",(*temp)->x,(*temp)->y);
    temp++;
  }
  fprintf(fp,"</FREEHAND>\n");
}

void MLB::PolyLine::ChangeLineColor(GUI::Canvas *_canvas,char *_color)
{
  if (!visible) return;
  recorded = false;

  color = _color;
  _canvas->ChangeFillColor(tkref.c_str(),color.c_str());
}

void MLB::PolyLine::ChangeLineStyle(GUI::Canvas *canvas,char *_lineStyle)
{
  if (!visible || isFreehand) return;

  lineStyle = _lineStyle;
  if (lineStyle == "none") lineStyle = "";
  canvas->ChangeLineStyle(tkref.c_str(),lineStyle.c_str());
}



void MLB::PolyLine::ChangeCapStyle(GUI::Canvas *_canvas, char *_style)
{
  if (!visible) return;
  recorded = false;

  capStyle = _style;
  _canvas->ChangeCapStyle(tkref.c_str(),capStyle.c_str());
}

void MLB::PolyLine::ChangeJoinStyle(GUI::Canvas *_canvas, char *_style)
{
  if (!visible) return;
  recorded = false;

  joinStyle = _style;
  _canvas->ChangeJoinStyle(tkref.c_str(),joinStyle.c_str());
}

void MLB::PolyLine::ChangeArrowStyle(GUI::Canvas *_canvas, char *_style)
{
  if (!visible) return;
  recorded = false;

  arrowStyle = _style;
  _canvas->ChangeArrowStyle(tkref.c_str(),arrowStyle.c_str());
}

void MLB::PolyLine::ChangeArrowShape(GUI::Canvas *_canvas, char *_shape)
{
  if (!visible) return;
  recorded = false;

  arrowShape = _shape;
  _canvas->ChangeArrowShape(tkref.c_str(),arrowShape.c_str());
}


void MLB::PolyLine::AdjustGui()
{

  GUI::Attributes::SetOutlineColor((char *)color.c_str());
  if (isFreehand)
    GUI::Attributes::SetTool("Freehand");
  else
    GUI::Attributes::SetTool("Polyline");
}


RefCount<RTPI::RtpiAdu> MLB::PolyLine::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;
  RefCount<Buffer> buf = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::POLYLINE, priority, buf, timestamp);
}

int MLB::PolyLine::FlushState(char* buf, uint32 order, bool visible, int x, int y, Color linecolor, int linewidth, Style linestyle,
			      bool smooth, Style capstyle, Style  joinstyle, Style arrowstyle, Style arrowshape) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | (int)smooth;

  p8[1] = CommunicationManager::POLYLINE;

  Line::FlushCapStyle(buf + 3, capstyle);
  Line::FlushJoinStyle(buf + 3, joinstyle);
  Line::FlushArrowStyle(buf + 3, arrowstyle);

  p32[1] = htonl(order);
  p32[2] = 0;

  ColoredObject::FlushColor(buf + 12, linecolor);
  ColoredObject::FlushLineStyle(buf + 15, linestyle);

  Line::FlushArrowShape(buf + 16, arrowshape);
  p8[19] = linewidth;

  p16[10] = htons(1);
  p16[11] = 0;

  p16[12] = htons(x);
  p16[13] = htons(y);

  return 28;
}


int MLB::PolyLine::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | (int)smooth;

  p8[1] = CommunicationManager::POLYLINE;

  Line::FlushCapStyle(buf + 3, capStyle);
  Line::FlushJoinStyle(buf + 3, joinStyle);
  Line::FlushArrowStyle(buf + 3, arrowStyle);
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p32[1] = htonl(order);

  p16[4] = htons(abs(width));
  p16[5] = htons(abs(height));

  ColoredObject::FlushColor(buf + 12, color);
  ColoredObject::FlushLineStyle(buf + 15, lineStyle);

  Line::FlushArrowShape(buf + 16, arrowShape);
  p8[19] = lineWidth;

  if (points.size() == 0) {
    // new polyline
    p16[10] = htons(1);
    p16[11] = 0;

    p16[12] = htons(x);
    p16[13] = htons(y);

    return 28;
  } else {
    p16[10] = htons(points.size());
    p16[11] = 0;

    int j = 0;
    for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
      p16[12 + j*2] = htons((*i)->x);
      p16[13 + j*2] = htons((*i)->y);
      ++j;
    }

    return 24 + points.size()*4;
  }
}


int MLB::PolyLine::ParseState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    GUI::MessageWindow::Error("PolyLine::ParseState - wrong version of mlb protocol\n");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    GUI::MessageWindow::Error("PolyLine::ParseState - wrong operation type\n");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;
  smooth  = p8[0] & MLB_MASK_S;

  if (p8[1] != CommunicationManager::POLYLINE) {
    GUI::MessageWindow::Error("PolyLine::ParseState - wrong object type %u\n", (int)p8[1]);
    return -1;
  }

  // reset
  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
    delete *i;
  }
  points.clear();

  capStyle = Line::ParseCapStyle(buf + 3);
  joinStyle = Line::ParseJoinStyle(buf + 3);
  arrowStyle = Line::ParseArrowStyle(buf + 3);

  order = ntohl(p32[1]);

  width  = ntohs(p16[4]);
  height = ntohs(p16[5]);
  if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
  if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

  color     = ColoredObject::ParseColor(buf + 12);
  lineStyle = ColoredObject::ParseLineStyle(buf + 15);

  arrowShape = Line::ParseArrowShape(buf + 16);
  lineWidth = p8[19];

  x = ntohs(p16[12]);
  y = ntohs(p16[13]);

  int count = ntohs(p16[10]);
  for (int j = 0; j < count; ++j) {
    points.push_back(new Point(ntohs(p16[12 + j*2]), ntohs(p16[13 + j*2])));
  }

  return 24 + count*4;
}


int MLB::PolyLine::FlushChangeVisibility(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}

int MLB::PolyLine::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

int MLB::PolyLine::FlushMovePoint(char* buf, int point, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::MOVE_POINT;

  p8[3] = 0;

  p16[2] = htons(point);
  p16[3] = 0;

  p16[4] = htons(x);
  p16[5] = htons(y);

  return 12;
}


int MLB::PolyLine::FlushChangeLinewidth(char* buf, int linewidth) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_LINEWIDTH;

  p8[7] = linewidth;

  return 8;
}


int MLB::PolyLine::FlushChangeLinestyle(char* buf, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_LINESTYLE;

  ColoredObject::FlushLineStyle(buf + 7, linestyle);

  return 8;
}


int MLB::PolyLine::FlushChangeLinecolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_LINECOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::PolyLine::FlushChangeCapstyle(char* buf, Style capstyle) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_CAPSTYLE;

  p8[3] = 0;
  Line::FlushCapStyle(buf + 3, capstyle);

  return 4;
}


int MLB::PolyLine::FlushChangeJoinstyle(char* buf, Style joinstyle) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_JOINSTYLE;

  p8[3] = 0;
  Line::FlushJoinStyle(buf + 3, joinstyle);

  return 4;
}


int MLB::PolyLine::FlushChangeArrowstyle(char* buf, Style arrowstyle, Style arrowshape) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::CHG_ARROWSTYLE;

  p8[3] = 0;
  Line::FlushArrowStyle(buf + 3, arrowstyle);

  p32[1] = 0;
  Line::FlushArrowShape(buf + 4, arrowshape);

  return 8;
}


int MLB::PolyLine::FlushAddPoint(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::ADD_POINT;
  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}


int MLB::PolyLine::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::PolyLine::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::PolyLine::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::POLYLINE;

  return 4;
}


MLB::DrawObject *MLB::PolyLine::Load(SGMLElement *hilf)
{
  PolyLine
    *obj;
  char
    *tmp,
    *p;
  int
    x, y,
    width, height,
    lineWidth;
  char
    *color,
    *lineStyle,
    *capStyle,
    *joinStyle,
    *arrowStyle,
    *arrowShape;

  tmp = hilf->GetAttribute("x");
  if (tmp) x = atoi(tmp);
  else     x = 0;
  tmp = hilf->GetAttribute("y");
  if (tmp) y = atoi(tmp);
  else     y = 0;
  tmp = hilf->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = 0;
  tmp = hilf->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = 0;
  tmp = hilf->GetAttribute("color");
  if (tmp) color = strdup(tmp);
  else     color = strdup("#ff0000");
  tmp = hilf->GetAttribute("lineWidth");
  if (tmp) lineWidth = atoi(tmp);
  else     lineWidth = 2;
  tmp = hilf->GetAttribute("lineStyle");
  if (tmp) lineStyle = strdup(tmp);
  else     lineStyle = strdup(" ");
  tmp = hilf->GetAttribute("capStyle");
  if (tmp) capStyle = strdup(tmp);
  else     capStyle = strdup("butt");
  tmp = hilf->GetAttribute("joinStyle");
  if (tmp) joinStyle = strdup(tmp);
  else     joinStyle = strdup("bevel");
  tmp = hilf->GetAttribute("arrowStyle");
  if (tmp) arrowStyle = strdup(tmp);
  else     arrowStyle = strdup("none");
  tmp = hilf->GetAttribute("arrowShape");
  if (tmp) arrowShape = strdup(tmp);
  else     arrowShape = strdup("8 10 4");

  obj = new MLB::PolyLine(x,y,width,height,color,lineWidth,lineStyle,
			  capStyle,joinStyle,arrowStyle,arrowShape,
			  false,CommunicationManager::cm->GetObjectID());
  obj->SetLayer(0);


  p = (char *)hilf->Parameter.c_str();
  while (*p != '\0') {
    while ((*p != '{') && (*p!='\0')) p++;
    if (*p=='\0') break;
    p++;
    tmp = p;
    while (*p != ' ') p++;
    *p = '\0';
    x = atoi(tmp);
    p++;
    tmp = p;
    while (*p != '}') p++;
    *p = '\0';
    y = atoi(tmp);
    obj->AppendNewPoint(x,y);
    p++;
  }

  LoadAnimationStep( obj, hilf );

  delete[] color;
  delete[] lineStyle;
  delete[] capStyle;
  delete[] joinStyle;
  delete[] arrowStyle;
  delete[] arrowShape;

  return obj;
}



/*********************************************************************/
/*********************************************************************/
/*********************************************************************/


MLB::Polygon::Polygon(uint32 id) : PolyObject(id)
{
  char
    buf[10];

  sprintf(buf,"%u",id);
  tkref = "Poly";
  tkref += buf;
}

MLB::Polygon::Polygon(int _x, int _y, int _width, int _height, char *_color,
				  char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id)
  : PolyObject(_x,_y,_width,_height,_color,_fillColor,_lineWidth,_lineStyle,_id)
{
  char
    buf[10];

  sprintf(buf,"%u",_id);
  tkref = "Poly";
  tkref += buf;
}

MLB::Polygon::~Polygon()
{
}

MLB::DrawObject *MLB::Polygon::Copy() {
  MLB::Polygon* ret = new Polygon(x,y,width,height,(char *)color.c_str(),(char *)fillColor.c_str(),
				  lineWidth,(char *)lineStyle.c_str(),CommunicationManager::cm->GetObjectID());

  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i)
    ret->AppendNewPoint((*i)->x,(*i)->y);

  return (DrawObject *)ret;
}


void MLB::Polygon::Draw(GUI::Canvas *_canvas) {
  if (!visible) return;

  if ((lastZoom != SessionManager::sm->GetZoom()) || (points.size() != _points.size())) {
    lastZoom = SessionManager::sm->GetZoom();
  // calculate points
  for (std::list<Point*>::iterator j = _points.begin(); j != _points.end(); ++j) delete *j;
  _points.clear();
  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i)
    _points.push_back(new Point(SessionManager::sm->GetZoom() * (*i)->x, SessionManager::sm->GetZoom() * (*i)->y));
  }

  if (fillColor == "none")
    _canvas->DrawPolygon(tkref.c_str(),
			 _points,
			 color.c_str(),
			 SessionManager::sm->GetZoom()*lineWidth,
			 lineStyle.c_str());
  else
    _canvas->DrawPolygon(tkref.c_str(),
			 _points,
			 color.c_str(),
			 fillColor.c_str(),
			 SessionManager::sm->GetZoom()*lineWidth,
			 lineStyle.c_str());

  if (selected) DrawSelection(_canvas);
}


void MLB::Polygon::ReDraw(GUI::Canvas *_canvas) {
  if (!visible) return;

  _canvas->DeleteObject(tkref.c_str());
  _canvas->DeleteLineSegments(tkref.c_str());

  Draw(_canvas);
}


void MLB::Polygon::SaveObject(FILE *fp, int level)
{
  std::list<Point *>::iterator
    temp;
  std::string
    ptBuf,
    tab;
  int
    j;
  char
    buf[20];

  if (!visible) return;

  for (j=0; j<level; j++)
    tab += "  ";

  for (std::list<Point *>::iterator i = points.begin(); i != points.end(); ++i) {
    sprintf(buf,"{%d %d} ",(*i)->x,(*i)->y);
    ptBuf += buf;
  }

  fprintf(fp,"%s<POLYGON x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" fillColor=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\" length=\"%d\"%s>\n",tab.c_str(),x,y,width,height,color.c_str(),fillColor.c_str(),lineWidth,lineStyle.c_str(),ptBuf.size(), GetAnimationStepSGMLText().c_str() );
  fprintf(fp,"%s  \"",tab.c_str());
  fwrite((char *)ptBuf.c_str(),1,ptBuf.size(),fp);
  fprintf(fp,"\"\n%s</POLYGON>\n",tab.c_str());

  ptBuf.erase();
  tab.erase();
}

void MLB::Polygon::SaveAsPs(FILE *fp, int pageWidth, int pageHeight, bool rotate)
{
  std::list<Point *>::iterator
    temp;
  XColor
    *xColor;
  float
    red, green, blue;

  if (!visible) return;

  if (points.size() <= 1) return;

  if (fillColor != "none") {
    xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)fillColor.c_str());
    red   = (float)xColor->red/65535;
    green = (float)xColor->green/65535;
    blue  = (float)xColor->blue/65535;

    fprintf(fp,"\n");
    fprintf(fp,"gsave\n");
    fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
    fprintf(fp,"newpath\n");
    temp = points.begin();
    if (rotate)
      fprintf(fp,"%d %d moveto\n",(*temp)->y,(*temp)->x);
    else
      fprintf(fp,"%d %d moveto\n",(*temp)->x,pageHeight-(*temp)->y);
    temp++;
    while (temp != points.end()) {
      if (rotate)
	fprintf(fp,"%d %d lineto\n",(*temp)->y,(*temp)->x);
      else
	fprintf(fp,"%d %d lineto\n",(*temp)->x,pageHeight-(*temp)->y);
      temp++;
    }
    fprintf(fp,"closepath fill\n");
    fprintf(fp,"stroke\n");
    fprintf(fp,"grestore\n");
  }

  xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)color.c_str());
  red   = (float)xColor->red/65535;
  green = (float)xColor->green/65535;
  blue  = (float)xColor->blue/65535;

  fprintf(fp,"\n");
  fprintf(fp,"gsave\n");
  fprintf(fp,"%d setlinewidth 0 setlinejoin 2 setlinecap\n",lineWidth);
  fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
  fprintf(fp,"newpath\n");
  temp = points.begin();
  if (rotate)
    fprintf(fp,"%d %d moveto\n",(*temp)->y,(*temp)->x);
  else
    fprintf(fp,"%d %d moveto\n",(*temp)->x,pageHeight-(*temp)->y);
  temp++;
  while (temp != points.end()) {
    if (rotate)
      fprintf(fp,"%d %d lineto\n",(*temp)->y,(*temp)->x);
    else
      fprintf(fp,"%d %d lineto\n",(*temp)->x,pageHeight-(*temp)->y);
    temp++;
  }
  fprintf(fp,"closepath\n");
  fprintf(fp,"stroke\n");
  fprintf(fp,"grestore\n");

}

int MLB::Polygon::RecordableObjects()
{
  if (!visible) return 0;

  if (fillColor == "none") /* Outline */
    return 1;
  else
    return 2;

}

void MLB::Polygon::SaveAsAOF(FILE *fp)
{
  std::list<Point *>::iterator
    i;

  if (recorded) return;
  recorded = true;

  recID = MLB::SessionManager::sm->recCount;

  if (fillColor != "none") {
    fprintf(fp,"<FILLEDPOLY x=%d y=%d width=%d height=%d fcolor=%d count=%d linewidth=%d linestyle=%d layer=%d>\n",
	    x,y,width,height,CalculateColor(fillColor),points.size(),lineWidth,CalculateLineStyle(),layer);
    for (i = points.begin(); i != points.end(); ++i) fprintf(fp,"<pt x=%d y=%d></pt>\n",(*i)->x,(*i)->y);
    fprintf(fp,"</FILLEDPOLY>\n");
    MLB::SessionManager::sm->recCount++;
  }

  fprintf(fp,"<POLYLINES x=%d y=%d width=%d height=%d fcolor=%d count=%d linewidth=%d linestyle=%d layer=%d>\n",
	  x,y,width,height,CalculateColor(color),points.size()+1,lineWidth,CalculateLineStyle(),layer);
  for (i = points.begin(); i != points.end(); ++i) fprintf(fp,"<pt x=%d y=%d></pt>\n",(*i)->x,(*i)->y);
  i = points.begin();
  fprintf(fp,"<pt x=%d y=%d></pt>\n",(*i)->x,(*i)->y);
  fprintf(fp,"</POLYLINES>\n");
  MLB::SessionManager::sm->recCount++;

}

void MLB::Polygon::ChangeLineColor(GUI::Canvas *_canvas,char *_color)
{
  if (!visible) return;

  recorded = false;
  color    = _color;

  if (fillColor == "none")
    _canvas->ChangeFillColor(tkref.c_str(),color.c_str());
  else
    _canvas->ChangeLineColor(tkref.c_str(),color.c_str());
}

void MLB::Polygon::AdjustGui()
{
  GUI::Attributes::SetOutlineColor((char *)color.c_str());

  if (fillColor != "none") {
    GUI::Attributes::SetFillColor((char *)fillColor.c_str());
    GUI::Attributes::SetTool("FilledPolygon");
  }
  else {
    GUI::Attributes::SetTool("OutlinePolygon");
  }
}

RefCount<RTPI::RtpiAdu> MLB::Polygon::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;
  RefCount<Buffer> buf = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::POLYGON, priority, buf, timestamp);
}

int MLB::Polygon::FlushState(char* buf, uint32 order, bool visible, int x, int y, Color linecolor, Color fillcolor, int linewidth, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | ((fillcolor.compare("none") == 0) ? 1 : 0); // transparent

  p8[1] = CommunicationManager::POLYGON;

  p32[1] = htonl(order);

  p32[2] = 0;

  ColoredObject::FlushColor(buf + 12, linecolor);
  ColoredObject::FlushLineStyle(buf + 15, linestyle);

  if (fillcolor.compare("none") == 0) p32[4] = 0;
  else ColoredObject::FlushColor(buf + 16, fillcolor);
  p8[19] = linewidth;

  p16[10] = htons(1);
  p16[11] = 0;

  p16[12] = htons(x);
  p16[13] = htons(y);

  return 28;
}


int MLB::Polygon::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | ((fillColor.compare("none") == 0) ? 1 : 0); // transparent

  p8[1] = CommunicationManager::POLYGON;

  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p32[1] = htonl(order);

  p16[4] = htons(abs(width));
  p16[5] = htons(abs(height));

  ColoredObject::FlushColor(buf + 12, color);
  ColoredObject::FlushLineStyle(buf + 15, lineStyle);

  if (fillColor.compare("none") == 0) p32[4] = 0;
  else ColoredObject::FlushColor(buf + 16, fillColor);
  p8[19] = lineWidth;

  if (points.size() == 0) {
    // new polyline
    p16[10] = htons(1);
    p16[11] = 0;

    p16[12] = htons(x);
    p16[13] = htons(y);

    return 28;
  } else {
    p16[10] = htons(points.size());
    p16[11] = 0;

    int j = 0;
    for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
      p16[12 + j*2] = htons((*i)->x);
      p16[13 + j*2] = htons((*i)->y);
      ++j;
    }

    return 24 + points.size()*4;
  }
}


int MLB::Polygon::ParseState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    GUI::MessageWindow::Error("Polygon::ParseState - wrong version of mlb protocol\n");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    GUI::MessageWindow::Error("Polygon::ParseState - wrong operation type\n");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;

  if (p8[1] != CommunicationManager::POLYGON) {
    GUI::MessageWindow::Error("Polygon::ParseState - wrong object type %u\n", (int)p8[1]);
    return -1;
  }

  // reset
  for (std::list<Point*>::iterator i = points.begin(); i != points.end(); ++i) {
    delete *i;
  }
  points.clear();

  order = ntohl(p32[1]);

  width  = ntohs(p16[4]);
  height = ntohs(p16[5]);

  if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
  if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

  color     = ColoredObject::ParseColor(buf + 12);
  lineStyle = ColoredObject::ParseLineStyle(buf + 15);

  if (p8[0] & MLB_MASK_T) // transparent -> no fill color
    fillColor = "none";
  else
    fillColor = ColoredObject::ParseColor(buf + 16);

  lineWidth = p8[19];

  x = ntohs(p16[12]);
  y = ntohs(p16[13]);

  int count = ntohs(p16[10]);
  for (int j = 0; j < count; ++j) {
    points.push_back(new Point(ntohs(p16[12 + j*2]),ntohs(p16[13 + j*2])));
  }

  return 24 + points.size()*4;
}


int MLB::Polygon::FlushChangeVisibility(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}

int MLB::Polygon::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

int MLB::Polygon::FlushMovePoint(char* buf, int point, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::MOVE_POINT;

  p8[3] = 0;

  p16[2] = htons(point);
  p16[3] = 0;

  p16[4] = htons(x);
  p16[5] = htons(y);

  return 12;
}

int MLB::Polygon::FlushChangeLinewidth(char* buf, int linewidth) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CHG_LINEWIDTH;

  p8[7] = linewidth;

  return 8;
}

int MLB::Polygon::FlushChangeLinestyle(char* buf, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CHG_LINESTYLE;

  ColoredObject::FlushLineStyle(buf + 7, linestyle);

  return 8;
}

int MLB::Polygon::FlushChangeLinecolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CHG_LINECOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}

int MLB::Polygon::FlushChangeFillcolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CHG_FILLCOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}

int MLB::Polygon::FlushAddPoint(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::ADD_POINT;
  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

int MLB::Polygon::FlushClosePolygon(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::CLOSE_POLYGON;
  p8[3] = 0;

  return 4;
}


int MLB::Polygon::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Polygon::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::POLYGON;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Polygon::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::POLYGON;

  return 4;
}


MLB::DrawObject *MLB::Polygon::Load(SGMLElement *hilf)
{
  Polygon
    *obj;
  char
    *tmp,
    *p;
  int
    x, y,
    width, height,
    lineWidth;
  char
    *color,
    *fillColor,
    *lineStyle;

  tmp = hilf->GetAttribute("x");
  if (tmp) x = atoi(tmp);
  else     x = 0;
  tmp = hilf->GetAttribute("y");
  if (tmp) y = atoi(tmp);
  else     y = 0;
  tmp = hilf->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = 0;
  tmp = hilf->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = 0;
  tmp = hilf->GetAttribute("color");
  if (tmp) color = strdup(tmp);
  else     color = strdup("#ff0000");
  tmp = hilf->GetAttribute("fillColor");
  if (tmp) fillColor = strdup(tmp);
  else     fillColor = strdup("none");
  tmp = hilf->GetAttribute("lineWidth");
  if (tmp) lineWidth = atoi(tmp);
  else     lineWidth = 2;
  tmp = hilf->GetAttribute("lineStyle");
  if (tmp) lineStyle = strdup(tmp);
  else     lineStyle = strdup(" ");

  obj = new MLB::Polygon(x,y,width,height,color,fillColor,lineWidth,lineStyle,
			 CommunicationManager::cm->GetObjectID());
  obj->SetLayer(0);

  p = (char *)hilf->Parameter.c_str();
  while (*p != '\0') {
    while ((*p != '{') && (*p!='\0')) p++;
    if (*p=='\0') break;
    p++;
    tmp = p;
    while (*p != ' ') p++;
    *p = '\0';
    x = atoi(tmp);
    p++;
    tmp = p;
    while (*p != '}') p++;
    *p = '\0';
    y = atoi(tmp);
    obj->AppendNewPoint(x,y);
    p++;
  }

  LoadAnimationStep( obj, hilf );

  delete[] color;
  delete[] fillColor;
  delete[] lineStyle;

  return obj;
}

