/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef WIN32
  #include <direct.h>
  #include <time.h>
  #include <sys/timeb.h>
#else
  #include <unistd.h>
  #include <sys/wait.h>
#endif

#include "AudioPort.h"
#include "AudioFile.h"

#include "mlb_session.h"

#include "mlb_record.h"

#define PAUSE 0
#define CONTINUE 1
#define ENDE 2

// #define BUFSIZE 512

static int 
  status;
static bool
  ready;
static int 
  totalFrames;
static time_t
  recorddate;


int Record::AOF::recCount = 0;

Record::AOF::AOF(char *_name, int _frameRate)
{
  std::string
    directory,
    zw;
  char
    count[10];

#ifdef WIN32
  recording = false;
#else
  recordingPID = 0;
#endif

  if (_name != NULL)
    prefix = _name;
  else 
    prefix = "unknown";

  sprintf(count,"_%d",recCount);
  directory = prefix + count;

  frameRate = _frameRate;

#ifdef WIN32
  while (_mkdir ((char *)directory.c_str()) == -1) {
    if (errno == EEXIST) {
      recCount++;
      sprintf(count,"_%d",recCount);
      directory = prefix + count;
    }
    else {
      GUI::MessageWindow::Error("ERROR: can't create directory %s\n",directory.c_str());
      return;
    }
  }
#else
  while (mkdir ((char *)directory.c_str(),S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
    if (errno == EEXIST) {
      recCount++;
      sprintf(count,"_%d",recCount);
      directory = prefix + count;
    }
    else {
      GUI::MessageWindow::Error("ERROR: can't create directory %s\n",directory.c_str());
      return;
    }
  }
#endif

  directory += "/";

  dataPath = directory + "data/";
  miscPath = directory + "misc/";
  apPath   = directory + "ap/";
  appsPath = directory + "apps/";
  
#ifdef WIN32
  if (_mkdir ((char *)dataPath.c_str()) == -1) {
    if (errno != EEXIST) {
      printf("ERROR: can't create directory data\n");
      return;
    }
  }
#else
  if (mkdir ((char *)dataPath.c_str(),S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
    if (errno != EEXIST) {
      printf("ERROR: can't create directory data\n");
      return;
    }
  }
#endif
  
  aofName   = prefix+".aof";
  audioName = prefix+".aif";
  evqName   = prefix+".evq";
  objName   = prefix+".obj";

  zw = dataPath+evqName;
  evqFp = fopen((char *)zw.c_str(),"wb");
  zw = dataPath+objName;
  objFp = fopen((char *)zw.c_str(),"wb");

  fprintf(objFp,"<WB orient=bottomleft>\n");
}

Record::AOF::~AOF()
{
  prefix.erase();

  dataPath.erase();
  miscPath.erase();
  apPath.erase();
  appsPath.erase();

  aofName.erase();
  audioName.erase();
  objName.erase();
  evqName.erase();
}

void Record::AOF::WriteAofFile(int _time)
{
  FILE  
    *aofFp;
  std::string
    zw;
  char
    buf[100];
  int
    min,
    sec,
    time,
    width, height;

  struct tm *detailrecorddate;
  
  if (MLB::SessionManager::sm && MLB::SessionManager::sm->GetActivePage()) {
    width = MLB::SessionManager::sm->GetActivePage()->GetWidth();
    height =  MLB::SessionManager::sm->GetActivePage()->GetHeight();
  }
  else {
    width = 594;
    height = 842;
  }

  zw = dataPath + aofName;
  aofFp = fopen((char *)zw.c_str(),"wb");

  if (aofFp == NULL) {
    sprintf(buf,"ERROR: can't open aoffile ");
    return;
  }

  time =_time/1000;
  min  = time/60;
  sec  = time%60;
  detailrecorddate = localtime(&recorddate);


  fprintf(aofFp,"<DOCUMENT>\n");
  fprintf(aofFp,"  <INFO>\n");
  fprintf(aofFp,"    <VERSION>1.1</VERSION>\n");
  fprintf(aofFp,"    <TITLE>%s</TITLE>\n",prefix.c_str());
  fprintf(aofFp,"    <AUTHOR></AUTHOR>\n");
  fprintf(aofFp,"    <COMMENTS>document description generated by MLB</COMMENTS>\n");
  fprintf(aofFp,"    <LENGTH>%d:%d</LENGTH>\n",min,sec);
  fprintf(aofFp,"    <RECORDDATE>%d.%d.%d %d:%d:%d</RECORDDATE>\n",detailrecorddate->tm_mday,detailrecorddate->tm_mon+1,detailrecorddate->tm_year+1900,detailrecorddate->tm_hour,detailrecorddate->tm_min,detailrecorddate->tm_sec);
  fprintf(aofFp,"  </INFO>\n");
  fprintf(aofFp,"  <FILES>\n");
  fprintf(aofFp,"    <AUDIO>%s</AUDIO>\n",audioName.c_str());
  fprintf(aofFp,"    <HELPERS>\n");
  fprintf(aofFp,"      <H>\n");
  fprintf(aofFp,"        <OPT>wbPlay</OPT>\n");
  fprintf(aofFp,"        <OPT>-G</OPT>\n");
  fprintf(aofFp,"        <OPT>%dx%d</OPT>\n",width,height);
  fprintf(aofFp,"        <OPT>%s</OPT>\n",objName.c_str());
  fprintf(aofFp,"        <OPT>%s</OPT>\n",evqName.c_str());
  fprintf(aofFp,"      </H>\n");
  if (MLB::SessionManager::sm->GetUseJedas()) {
    fprintf(aofFp,"      <H>\n");
    fprintf(aofFp,"        <OPT>java</OPT>\n");
    fprintf(aofFp,"        <OPT>jedas.io.AOFHelper</OPT>\n");
    fprintf(aofFp,"        <OPT>%s_jedas.ini</OPT>\n",prefix.c_str());
    fprintf(aofFp,"      </H>\n");
  }
  if (MLB::SessionManager::sm->GetUseAppPlay()) {
    fprintf(aofFp,"      <H>\n");
    fprintf(aofFp,"        <OPT>appPlay</OPT>\n");
    fprintf(aofFp,"        <OPT>%s_mgr.appevq</OPT>\n",prefix.c_str());
    fprintf(aofFp,"        <OPT>%s_mgr.appobj</OPT>\n",prefix.c_str());
    fprintf(aofFp,"      </H>\n");
  }
  fprintf(aofFp,"    </HELPERS>\n");
  fprintf(aofFp,"    <SLIDES>\n");
  fprintf(aofFp,"    </SLIDES>\n");
  fprintf(aofFp,"    <APPS>\n");
  fprintf(aofFp,"    </APPS>\n");
  fprintf(aofFp,"  </FILES>\n");
  fprintf(aofFp,"  <CONFIG>\n");
  fprintf(aofFp,"    <SLIDES>\n");
  fprintf(aofFp,"    </SLIDES>\n");
  fprintf(aofFp,"    <APPS>\n");
  fprintf(aofFp,"    </APPS>\n");
  fprintf(aofFp,"  </CONFIG>\n");
  fprintf(aofFp,"</DOCUMENT>\n");

  fclose (aofFp);
}

void Record::AOF::WritepaaFile(int _time)
{
  FILE  
    *paaFp;
  std::string
    zw;
  char
    buf[100];
  int
    min,
    sec,
    time,
    width, height;

  struct tm *detailrecorddate;
  
  if (MLB::SessionManager::sm && MLB::SessionManager::sm->GetActivePage()) {
    width = MLB::SessionManager::sm->GetActivePage()->GetWidth();
    height =  MLB::SessionManager::sm->GetActivePage()->GetHeight();
  }
  else {
    width = 594;
    height = 842;
  }

  zw = dataPath + prefix + ".paa";
  paaFp = fopen((char *)zw.c_str(),"wb");

  if (paaFp == NULL) {
    sprintf(buf,"ERROR: can't open paafile ");
    return;
  }

  time =_time/1000;
  min  = time/60;
  sec  = time%60;
  detailrecorddate = localtime(&recorddate);


  fprintf(paaFp,"<DOCUMENT>\n");
  fprintf(paaFp,"  <INFO>\n");
  fprintf(paaFp,"    <VERSION>1.0</VERSION>\n");
  fprintf(paaFp,"    <TITLE>%s</TITLE>\n",prefix.c_str());
  fprintf(paaFp,"    <AUTHOR></AUTHOR>\n");
  fprintf(paaFp,"    <COMMENTS>document description generated by MLB</COMMENTS>\n");
  fprintf(paaFp,"    <LENGTH>%d:%d</LENGTH>\n",min,sec);
  fprintf(paaFp,"    <RECORDDATE>%d.%d.%d %d:%d:%d</RECORDDATE>\n",detailrecorddate->tm_mday,detailrecorddate->tm_mon+1,detailrecorddate->tm_year+1900,detailrecorddate->tm_hour,detailrecorddate->tm_min,detailrecorddate->tm_sec);
  fprintf(paaFp,"  </INFO>\n");
  fprintf(paaFp,"  <FILES>\n");
  fprintf(paaFp,"    <HELPERS>\n");
  fprintf(paaFp,"      <H>\n");
  fprintf(paaFp,"        <OPT>wbPlay</OPT>\n");
  fprintf(paaFp,"        <OPT>-G</OPT>\n");
  fprintf(paaFp,"        <OPT>%dx%d</OPT>\n",width,height);
  fprintf(paaFp,"        <OPT>%s</OPT>\n",objName.c_str());
  fprintf(paaFp,"        <OPT>%s</OPT>\n",evqName.c_str());
  fprintf(paaFp,"      </H>\n");
  fprintf(paaFp,"    </HELPERS>\n");
  fprintf(paaFp,"</DOCUMENT>\n");

  fclose (paaFp);
}

#ifdef WIN32

unsigned int Record::AOF::GetRecordTime() 
{
// GUI::MessageWindow::Debug("totalFrames %d frameRate %d\n",totalFrames,frameRate);	
  return (int)((float)totalFrames/(((float)frameRate/1000)));
}

void Record::AOF::CreateProcess()
{
  CreateReadThread();
  CreateRecordThread();
  
  return;  
}

int Record::AOF::Start()
{
  if (!recording) {
    totalFrames = 0;
    ready = false;
	recorddate = time( (time_t *) NULL);
    CreatePipe(&readPipe,&writePipe,NULL,0);
    CreateProcess();
    recording = true;
    status = CONTINUE;
    while (!ready) Sleep(10);
    if (status == ENDE) {
      GUI::MessageWindow::Error("Cant't record audio\n");
      GUI::Record::Stop();  
    }
    ResetTime();
  }
  return 1;
}

int Record::AOF::Pause()
{
  if (recording) {
    WriteToPipe("PAUSE ");
    pauseTime = GetTime();
  } 
  
  return 1;
}

int Record::AOF::Continue()
{ 
  if (recording) {
    WriteToPipe("CONTINUE ");
    startTime += (GetTime() - pauseTime);
    pauseTime = 0;
  }

  return 1;
}

int Record::AOF::Stop()
{
  if (recording) {
    WriteToPipe("ENDE ");
    KillProcess();
    recording = false;
    WriteAofFile(GetTime());
    fprintf(objFp,"</WB>\n");
    fclose(objFp);
    objFp = 0;
    fclose(evqFp);
    evqFp = 0;
	if (MLB::SessionManager::sm->AnnotationMode())
		WritepaaFile(GetTime());
  }

  return 1;
}

void Record::AOF::WriteToPipe(char *text)
{
  DWORD
    written;

  WriteFile(writePipe,text,sizeof(text),&written,NULL);
}

LONG RecordThread ( LPVOID lpThreadData )
{
  char
    *buf;
  Audio::Port 
    *port;
  Audio::File
    *file; 
  int 
    i, nRead, nWrite, size, frames;
  HANDLE
    writePipe;
  DWORD
    written;
  char  
    *audioName;
  unsigned int 
    frequency;
  unsigned short 
    channel,
    encoding;

  audioName =	((Record::recordData *)lpThreadData)->audioName;
  writePipe = ((Record::recordData *)lpThreadData)->pipe;
  frequency = ((Record::recordData *)lpThreadData)->frequency;
  channel   = ((Record::recordData *)lpThreadData)->channel;
  encoding  = ((Record::recordData *)lpThreadData)->encoding;

  file = new Audio::File(frequency, encoding, channel);
  file->Open(audioName,"wb");
  file->PutAiffHeader(0);

  port = new Audio::Port();

#ifdef WIN32 
 
  if (port->Open(RECORD,frequency, channel, encoding) < 0) {
    GUI::MessageWindow::Error("Can not open audioport\n");
    port->Close();
    file->Close();
    delete port; 
    delete file;  
    status = ENDE;
    WriteFile(writePipe,"READY",5,&written,NULL);
    return -1;
  }
  WriteFile(writePipe,"READY",5,&written,NULL);
 
#else

  if (port->Open(RECORD,frequency,channel,encoding) < 0)
    GUI::MessageWindow::Error("Can not open port\n");

#endif

  size = port->GetBufSize();  
  buf = (char *)malloc(size);
  frames = 0;
  totalFrames = 0;
  i = 0;   
  GUI::MessageWindow::Info("Audio mute:%d\n",MLB::SessionManager::sm->GetAudioMute());
  while (status != ENDE) { 
	  if (status == CONTINUE) {
      nRead = port->Read((char *)buf, 256);
      if (nRead > 0){
		if (MLB::SessionManager::sm->GetAudioMute()==0)
          nWrite = file->Write(buf,nRead);
        if (nWrite != nRead && MLB::SessionManager::sm->GetAudioMute()==0)
          GUI::MessageWindow::Info("gelesen %d, geschrieben %d\n",nRead,nWrite); 
        frames += nRead; 
        totalFrames = frames;  
      }
    }   
  }  

  port->Stop(); 

  nRead = port->Read((char *)buf, 256);
  while (nRead >= 0) { 
    if (nRead > 0) {
      if (MLB::SessionManager::sm->GetAudioMute()==0)
		  nWrite = file->Write(buf,nRead);
      if (nWrite != nRead && MLB::SessionManager::sm->GetAudioMute()==0)
        GUI::MessageWindow::Info("gelesen %d, geschrieben %d\n",nRead,nWrite);
      frames += nRead; 
      totalFrames = frames;
    }
    nRead = port->Read((char *)buf, 256);
  }  

   port->Close(); 
 
  GUI::MessageWindow::Info("totalFrames: %d\n",totalFrames);
  file->SetLength(frames); 
  file->PutAiffHeader(frames);
  file->Close();
 
  delete port; 
  delete file; 
 
  GUI::MessageWindow::Info("Ennd Thread\n");
  return( 0L );      
}


BOOL Record::AOF::CreateRecordThread()
{
  HANDLE 
    hThread;            // thread that reads from pipe
  DWORD 
    dwThreadID;         // id of pipe thread
  static std::string 
    zw;
  recordData 
    *data;
  
  data = new recordData;
  zw = dataPath + audioName;
  data->audioName = strdup(zw.c_str());
  data->pipe = writePipe;
  data->frequency = frameRate;
  data->channel = Audio::MONO;
  data->encoding = Audio::LIN8OFFSET;
  hThread = CreateThread( NULL,           // security attributes
			  0,                          // initial stack size
			  (LPTHREAD_START_ROUTINE)RecordThread,
			  data,                       // argument
			  CREATE_SUSPENDED,           // creation flag
			  &dwThreadID);              // new thread's id
  
  if(!hThread) {
    GUI::MessageWindow::Error("Error creating record thread\n");
    return( FALSE );
  } 
  // set the thread's priority and let it run
  SetThreadPriority( hThread, THREAD_PRIORITY_BELOW_NORMAL );
  ResumeThread( hThread );
   
  return (TRUE);
}

void Record::AOF::KillProcess()
{
  if (!recording) return;

  //TerminateProcess((HANDLE) hProcess, 0);

  return;
}

LONG PipeThread ( LPVOID lpThreadData )
{ 
  HANDLE
    readPipe;
  DWORD	
    dwRead;
  BOOL	
    bTest;
  char 
    buf[20];


  readPipe = (HANDLE)lpThreadData;
  while( status != ENDE ) {
    bTest = ReadFile(readPipe, buf, sizeof(buf),	&dwRead, NULL );
    if( bTest ) {
      if (strncmp(buf,"ENDE",4) == 0) status = ENDE;
      if (strncmp(buf,"PAUSE",5) == 0) status = PAUSE;
      if (strncmp(buf,"CONTINUE",8) == 0) status = CONTINUE;
      if (strncmp(buf,"READY",5) == 0) ready = true;
    }
    else {
      GUI::MessageWindow::Error("Error in pipe thread\n");
      status = ENDE;	    	  // let the child end  
      return (0L);
    }
  }
  
  return( 0L );
}

BOOL Record::AOF::CreateReadThread()
{
  DWORD 
    dwThreadID;
  HANDLE   
    hThread; 
  
  hThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)PipeThread, readPipe, CREATE_SUSPENDED, &dwThreadID );

  if( ! hThread ) {
    GUI::MessageWindow::Error("Error creating read thread\n");
    return (FALSE);
  }

  // set the thread's priority and let it run
  SetThreadPriority( hThread, THREAD_PRIORITY_NORMAL );
  ResumeThread( hThread );
  
  return (TRUE);
}

#else /*** UNIX ***/

int Record::AOF::Start()
{
  if (recordingPID == 0) {
    CreatePipe();
    CreateProcess();
  }
  recorddate = time( (time_t *) NULL);
  ResetTime();
}

int Record::AOF::Pause()
{
  kill(recordingPID, SIGUSR1);   

  pauseTime = GetTime();

}

int Record::AOF::Continue()
{
  kill(recordingPID, SIGUSR1);    
  
  startTime += (GetTime() - pauseTime);
  pauseTime = 0;

}

int Record::AOF::Stop()
{
  if (recordingPID != 0) {
    WriteToPipe("ENDE ");
    KillProcess();
    recordingPID = 0;
    WriteAofFile(GetTime());
    fprintf(objFp,"</WB>\n");
    fclose(objFp);
    objFp = 0;
    fclose(evqFp);
    evqFp = 0;
	if (MLB::SessionManager::sm->AnnotationMode())
		WritepaaFile(GetTime());
  }
}

void Record::AOF::WriteToPipe(char *text)
{
  write(writePipe,text,strlen(text));
}

void Record::AOF::CreateProcess()
{
  if (recordingPID == 0)
    recordingPID = CreateRecordThread(readPipe);
}

void Record::AOF::KillProcess()
{
  int
    status;

  if (recordingPID == 0) return;

  kill(recordingPID, SIGINT);    
  waitpid(recordingPID,&status,0);

  return;
}

void sigint(int signal)
{
  status = ENDE;
}

void sigusr1(int signal)
{
  if (status == CONTINUE)
    status = PAUSE;
  else  
    status = CONTINUE;
}

void Record::AOF::RecordThread (int readPipe)
{
  short
    *buf;
  std::string
    zw;
  Audio::Port
    *port;
  Audio::File
    *file;
  int 
    i, nRead, size, frames;

  signal(SIGINT, sigint);
  signal(SIGUSR1, sigusr1);

  port = new Audio::Port();
  if (port->Open(RECORD,frameRate,Audio::MONO,Audio::LIN8) < 0) {
    printf("Can not open port\n");
    port->Close();
    delete port;
    return;
  }
  file = new Audio::File(frameRate, Audio::LIN8, 1);
  zw = dataPath + audioName;
  file->Open((char *)zw.c_str(),"w");
  file->PutAiffHeader(0);
  size = BUFSIZE;
  buf = (short *)malloc(size*sizeof(short));
  frames = 0;
  status = CONTINUE;
  while (status != ENDE) {
    if (status == CONTINUE) {
      nRead = port->Read((char *)buf, BUFSIZE);
      if ((nRead > 0) && (MLB::SessionManager::sm->GetAudioMute()==0)) {
	file->Write(buf,nRead);
	frames += nRead; 
      } 
    }
  }

  file->SetLength(frames);
  file->PutAiffHeader(frames);

  port->Close();
  file->Close();
  delete port;
  delete file; 

  return;
}

void Record::AOF::CreatePipe()
{
  int 
    pipeIds[2];

  if (pipe(pipeIds) != 0) {
    return;
  }
  
  fcntl(pipeIds[0], F_SETFD, FD_CLOEXEC);
  fcntl(pipeIds[1], F_SETFD, FD_CLOEXEC);
  fcntl(pipeIds[0], F_SETFD, O_NONBLOCK);
  fcntl(pipeIds[1], F_SETFD, O_NONBLOCK);
  
  readPipe  = pipeIds[0];
  writePipe = pipeIds[1];

  return;
}


int Record::AOF::CreateRecordThread(int readPipe)
{
  int 
    pid;
  char
    buf1[10],
    buf2[10];

  pid = fork();
  if (pid == -1) {
    perror("Cannot fork");
    abort();
  } 
  else if (pid == 0) {
    RecordThread(readPipe);
    exit(0);
  }

  return pid;
}

#endif
    
void Record::AOF::ResetTime()
{
  
  pauseTime = 0;
#ifdef WIN
  struct timeb t;
  
  ftime(&t);

  startTime = t.time*1000 + t.millitm;

#else

  struct timeval tv;
  struct timezone tz;
  
  (void) gettimeofday(&tv, &tz);

  startTime = tv.tv_sec*1000 + tv.tv_usec * 0.001;

#endif

}
   
unsigned int Record::AOF::GetTime()
{
  unsigned int
    time;

#ifdef WIN
  struct timeb t;
  
  ftime(&t);

  time = t.time*1000 + t.millitm;

#else  

  struct timeval tv;
  struct timezone tz;
  
  (void) gettimeofday(&tv, &tz);

  time = tv.tv_sec*1000 + tv.tv_usec * 0.001;

#endif

  return (time-startTime);
}
