/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_rectangles.cpp
 authors : Gabriela Maass, Juergen Vogel
 date    : 3.12.1999
 desc    : Functions to draw and manipulate objects on a tcl-canvas

 **********************************************************************************************************/


#include "mlb.h"

#include "mlb_objects.h"
#include "mlb_session.h"
#include "mlb_page.h"
#include "mlb_adu.h"
#include "mlb_gui.h"

MLB::Rectangle::Rectangle(uint32 id) : ColoredObject(id), thumbnailID(-1) {
  char buf[100];
  sprintf(buf,"%u",id);
  tkref = "Rect";
  tkref += buf;
  layer = 1;
  recorded = false;
}

MLB::Rectangle::Rectangle(int _x, int _y, int _width, int _height, char *_color,
				      char *_fillColor, int _lineWidth, char *_lineStyle, uint32 _id)
  : ColoredObject(_x,_y,_width,_height,_color,_fillColor,_lineWidth,_lineStyle,_id), thumbnailID(-1) {
  char buf[100];
  sprintf(buf,"%u",id);
  tkref = "Rect";
  tkref += buf;
  layer = 1;
}

MLB::Rectangle::~Rectangle() {
}

void MLB::Rectangle::Draw(GUI::Canvas *_canvas) {
  if (!visible) return;

  if (fillColor == "none")
    /* Outline Rectangle */
    _canvas->DrawRectangle(tkref.c_str(),
			   SessionManager::sm->GetZoom()*x,
			   SessionManager::sm->GetZoom()*y,
			   SessionManager::sm->GetZoom()*width,
			   SessionManager::sm->GetZoom()*height,
			   color.c_str(),
			   SessionManager::sm->GetZoom()*lineWidth,
			   lineStyle.c_str());
  else
    /* Filled Rectangle */
    _canvas->DrawRectangle(tkref.c_str(),
			   SessionManager::sm->GetZoom()*x,
			   SessionManager::sm->GetZoom()*y,
			   SessionManager::sm->GetZoom()*width,
			   SessionManager::sm->GetZoom()*height,
			   color.c_str(),
			   fillColor.c_str(),
			   SessionManager::sm->GetZoom()*lineWidth,
			   lineStyle.c_str());

  if (selected) DrawSelection(_canvas);
}


void MLB::Rectangle::ReDraw(GUI::Canvas *canvas) {
  if (!visible) return;

  canvas->DeleteObject(tkref.c_str());
  Draw(canvas);
}


MLB::DrawObject *MLB::Rectangle::Copy() {
  return new Rectangle(x, y, width, height, (char*)color.c_str(), (char*)fillColor.c_str(),
		      lineWidth, (char*)lineStyle.c_str(), CommunicationManager::cm->GetObjectID());
}


void MLB::Rectangle::SaveObject(FILE *fp, int level) {
  if (!visible) return;

  for (int i = 0; i < level; ++i) fprintf(fp,"  ");

  if (IsThumbnail()) {
    fprintf(fp,"<RECTANGLE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" fillColor=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\" thumbnailID=\"%d\"%s></RECTANGLE>\n",
	          x, y, width, height, color.c_str(), fillColor.c_str(), lineWidth, lineStyle.c_str(), thumbnailID, GetAnimationStepSGMLText().c_str());
  } else {
  fprintf(fp,"<RECTANGLE x=\"%d\" y=\"%d\" width=\"%d\" height=\"%d\" color=\"%s\" fillColor=\"%s\" lineWidth=\"%d\" lineStyle=\"%s\"%s></RECTANGLE>\n",
	  x, y, width, height, color.c_str(), fillColor.c_str(), lineWidth, lineStyle.c_str(), GetAnimationStepSGMLText().c_str());
}
}

void MLB::Rectangle::SaveAsPs(FILE *fp, int pageWidth, int pageHeight, bool rotate)
{
  XColor
    *xColor;
  float
    red, green, blue;
  int
    psX1, psX2,
    psY1, psY2;

  if (!visible) return;
  if ((width==0) || (height==0)) return;

  if (rotate) {
    psX1 = y;
    psX2 = y+height;
    psY1 = x;
    psY2 = x+width;
  }
  else {
    psX1 = x;
    psX2 = x+width;
    psY1 = pageHeight-y;
    psY2 = pageHeight-(y+height);
  }

  if (fillColor != "none") {  /* FilledRectangle */
    xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)fillColor.c_str());
    red   = (float)xColor->red/65535;
    green = (float)xColor->green/65535;
    blue  = (float)xColor->blue/65535;

    fprintf(fp,"\n");
    fprintf(fp,"gsave\n");
    fprintf(fp,"%d setlinewidth 0 setlinejoin 2 setlinecap\n",lineWidth);
    fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
    fprintf(fp,"newpath\n");
    fprintf(fp,"%d %d moveto\n",psX1,psY1);
    fprintf(fp,"%d %d lineto\n",psX2,psY1);
    fprintf(fp,"%d %d lineto\n",psX2,psY2);
    fprintf(fp,"%d %d lineto\n",psX1,psY2);
    fprintf(fp,"closepath fill\n");
    fprintf(fp,"stroke\n");
    fprintf(fp,"grestore\n");
  }

  xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)color.c_str());
  red   = (float)xColor->red/65535;
  green = (float)xColor->green/65535;
  blue  = (float)xColor->blue/65535;

  fprintf(fp,"gsave\n");
  fprintf(fp,"%d setlinewidth 0 setlinejoin 2 setlinecap\n",lineWidth);
  fprintf(fp,"%2.3f %2.3f %2.3f setrgbcolor\n",red,green,blue);
  fprintf(fp,"newpath\n");
  fprintf(fp,"%d %d moveto\n",psX1,psY1);
  fprintf(fp,"%d %d lineto\n",psX2,psY1);
  fprintf(fp,"%d %d lineto\n",psX2,psY2);
  fprintf(fp,"%d %d lineto\n",psX1,psY2);
  fprintf(fp,"closepath\n");
  fprintf(fp,"stroke\n");
  fprintf(fp,"grestore\n");
}


void MLB::Rectangle::SaveAsAOF(FILE *fp) {
  int
    zX, zY, 
    zWidth, zHeight;

  if (!visible || recorded) return;

  recorded = true;

  if (width < 0) {
    zX = x + width;
    zWidth = width*(-1);
  } 
  else {
    zX = x;
    zWidth = width;
  }

  if (height < 0) {
    zY = y + height;
    zHeight = height*(-1);
  } 
  else {
    zY = y;
    zHeight = height;
  }

  recID = MLB::SessionManager::sm->recCount;
  if (fillColor != "none") { /* Filled Rectangle */
    fprintf(fp,"<FILLEDRECT x=%d y=%d width=%d height=%d fcolor=%d layer=%d></FILLEDRECT>\n",zX,zY,zWidth,zHeight,CalculateColor(fillColor),layer);
    MLB::SessionManager::sm->recCount++;
  }

  fprintf(fp,"<OUTLINERECT x=%d y=%d width=%d height=%d linewidth=%d linestyle=%d fcolor=%d layer=%d></OUTLINERECT>\n",zX,zY,zWidth,zHeight,lineWidth,CalculateLineStyle(),CalculateColor(color),layer);
  MLB::SessionManager::sm->recCount++;
}


int MLB::Rectangle::RecordableObjects()
{
  if (width==0 && height==0) visible = false;
  if (!visible) return 0;

  if (fillColor == "none") /* Outline Rectangle */
    return 1;
  else
    return 2;
}


bool MLB::Rectangle::Overlap(GUI::Canvas *canvas, int _x, int _y)
{
  int
    x1, y1,
    x2, y2;

  if (!visible) return false;

  GetBBox(&x1,&y1,&x2,&y2);
  if ((_x < x1) || (_y < y1) || (_x > x2) || (_y > y2)) return false;

  return canvas->Overlap(tkref.c_str(), SessionManager::sm->GetZoom()*_x, SessionManager::sm->GetZoom()*_y);
}


void MLB::Rectangle::AdjustGui() {

  GUI::Attributes::SetOutlineColor((char *)color.c_str());
  if (fillColor != "none") {
    GUI::Attributes::SetFillColor((char *)fillColor.c_str());
    GUI::Attributes::SetTool("FillRect");
  }
  else {
    GUI::Attributes::SetTool("OutRect");
  }
}


RefCount<RTPI::RtpiAdu> MLB::Rectangle::GetStateAdu(uint32 timestamp, int priority) {
  GenericContainer* parent = SessionManager::sm->GetParentContainer(GetID());
  if (!parent) return NULL;
  RefCount<Buffer> buf = new Buffer(GetStateSize());
  FlushState(buf->getCont());
  return CommunicationManager::cm->CreateState(GetID(), parent->GetID(), CommunicationManager::RECTANGLE, priority, buf, timestamp);
}


int MLB::Rectangle::FlushState(char* buf, uint32 order, bool visible, int x, int y, int width, int height,
				Color linecolor, Color fillcolor, int linewidth, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | ((fillcolor.compare("none") == 0) ? 1 : 0); // transparent

  p8[1] = CommunicationManager::RECTANGLE;

  p16[1] = 0;
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  p16[6] = htons(abs(width));
  p16[7] = htons(abs(height));

  ColoredObject::FlushColor(buf + 16, linecolor);
  ColoredObject::FlushLineStyle(buf + 19, linestyle);

  if (fillcolor.compare("none") == 0)
    p32[5] = 0;
  else
    ColoredObject::FlushColor(buf + 20, fillcolor);
  p8[23] = linewidth;

  return 24;
}


int MLB::Rectangle::FlushState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_STATE << 2);
  p8[0] = p8[0] | (visible << 1);
  p8[0] = p8[0] | ((fillColor.compare("none") == 0) ? 1 : 0); // transparent

  p8[1] = CommunicationManager::RECTANGLE;

  p16[1] = 0;
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p32[1] = htonl(order);

  p16[4] = htons(x);
  p16[5] = htons(y);

  p16[6] = htons(abs(width));
  p16[7] = htons(abs(height));

  ColoredObject::FlushColor(buf + 16, color);
  ColoredObject::FlushLineStyle(buf + 19, lineStyle);

  if (fillColor.compare("none") == 0)
    p32[5] = 0;
  else
    ColoredObject::FlushColor(buf + 20, fillColor);
  p8[23] = lineWidth;

  return 24;
}


int MLB::Rectangle::ParseState(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  if (((p8[0] & MLB_MASK_V) >> 6) != CommunicationManager::VERSION) {
    GUI::MessageWindow::Error("Rectangle::ParseState - wrong version of mlb protocol\n");
    return -1;
  }

  // referenced
  if (((p8[0] & MLB_MASK_OPT) >> 2) != CommunicationManager::ADU_STATE) {
    GUI::MessageWindow::Error("Rectangle::ParseState - wrong operation type\n");
    return -1;
  }

  visible = (p8[0] & MLB_MASK_VIS) >> 1;

  if (p8[1] != CommunicationManager::RECTANGLE) {
    GUI::MessageWindow::Error("Rectangle::ParseState - wrong object type %u\n", (int)p8[1]);
    return -1;
  }

  order = ntohl(p32[1]);

  x = ntohs(p16[4]);
  y = ntohs(p16[5]);

  width  = htons(p16[6]);
  height = htons(p16[7]);

  if (((p8[3] & MLB_MASK_SW) >> 1) == 1) width = -width;
  if ((p8[3] & MLB_MASK_SH) == 1) height = -height;

  color     = ColoredObject::ParseColor(buf + 16);
  lineStyle = ColoredObject::ParseLineStyle(buf + 19);

  if (p8[0] & MLB_MASK_T) // transparent -> no fill color
    fillColor = "none";
  else
    fillColor = ColoredObject::ParseColor(buf + 20);

  lineWidth = p8[23];

  return 24;
}


int MLB::Rectangle::FlushChangeVisibility(char* buf, bool vis) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);
  p8[0] = p8[0] | (vis << 1);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_VISIBILITY;
  p8[3] = 0;

  return 4;
}


int MLB::Rectangle::FlushMove(char* buf, int x, int y) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::MOVE;

  p8[3] = 0;

  p16[2] = htons(x);
  p16[3] = htons(y);

  return 8;
}

int MLB::Rectangle::FlushChangeSize(char* buf, int x, int y, int width, int height) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_SIZE;

  p8[3] = 0;
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p16[2] = htons(x);
  p16[3] = htons(y);

  p16[4] = htons(abs(width));
  p16[5] = htons(abs(height));

  return 12;
}


int MLB::Rectangle::SFlushChangeSize(char* buf, int x, int y, int width, int height) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_SIZE;

  p8[3] = 0;
  if (width  < 0) p8[3] = p8[3] | (1 << 1);
  if (height < 0) p8[3] = p8[3] | 1;

  p16[2] = htons(x);
  p16[3] = htons(y);

  p16[4] = htons(abs(width));
  p16[5] = htons(abs(height));

  return 12;
}


int MLB::Rectangle::FlushChangeLinewidth(char* buf, int linewidth) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_LINEWIDTH;

  p8[7] = linewidth;

  return 8;
}


int MLB::Rectangle::FlushChangeLinestyle(char* buf, Style linestyle) {
  uint8*  p8  = (uint8*)buf;
  uint16* p16 = (uint16*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p32[1] = 0;

  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_LINESTYLE;

  ColoredObject::FlushLineStyle(buf + 7, linestyle);

  return 8;
}


int MLB::Rectangle::FlushChangeLinecolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_LINECOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Rectangle::FlushChangeFillcolor(char* buf, Color color) {
  uint8*  p8  = (uint8*)buf;

  p8[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::CHG_FILLCOLOR;
  p8[3] = 0;

  ColoredObject::FlushColor(buf + 4, color);
  p8[7] = 0;

  return 8;
}


int MLB::Rectangle::FlushRaise(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::RAISE;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Rectangle::FlushLower(char* buf, uint32 order) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_CHANGE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  p8[2] = CommunicationManager::LOWER;

  p32[1] = htonl(order);

  return 8;
}


int MLB::Rectangle::FlushDeleteS(char* buf) {
  uint8*  p8  = (uint8*)buf;
  uint32* p32 = (uint32*)buf;

  p32[0] = 0;
  p8[0] = p8[0] | (CommunicationManager::VERSION << 6);
  // referenced is 0
  p8[0] = p8[0] | (CommunicationManager::ADU_DELETE << 2);

  p8[1] = CommunicationManager::RECTANGLE;

  return 4;
}


MLB::DrawObject *MLB::Rectangle::Load(SGMLElement *hilf)
{
  Rectangle
    *obj;
  char
    *tmp;
  int
    x, y,
    width, height,
    lineWidth,
    thumbnailID;
  char
    *color,
    *fillColor,
    *lineStyle;

  tmp = hilf->GetAttribute("x");
  if (tmp) x = atoi(tmp);
  else     x = 0;
  tmp = hilf->GetAttribute("y");
  if (tmp) y = atoi(tmp);
  else     y = 0;
  tmp = hilf->GetAttribute("width");
  if (tmp) width = atoi(tmp);
  else     width = 0;
  tmp = hilf->GetAttribute("height");
  if (tmp) height = atoi(tmp);
  else     height = 0;
  tmp = hilf->GetAttribute("color");
  if (tmp) color = strdup(tmp);
  else     color = strdup("#ff0000");
  tmp = hilf->GetAttribute("fillColor");
  if (tmp) fillColor = strdup(tmp);
  else     fillColor = strdup("none");
  tmp = hilf->GetAttribute("lineWidth");
  if (tmp) lineWidth = atoi(tmp);
  else     lineWidth = 2;
  tmp = hilf->GetAttribute("lineStyle");
  if (tmp) lineStyle = strdup(tmp);
  else     lineStyle = strdup(" ");
  tmp = hilf->GetAttribute("thumbnailID");
  if (tmp) thumbnailID = atoi(tmp);
  else     thumbnailID = -1;

  obj = new Rectangle(x,y,width,height,color,fillColor,lineWidth,lineStyle,
		      CommunicationManager::cm->GetObjectID());
  obj->SetLayer(0);

  if (thumbnailID != -1) {
    obj->SetThumbnailID(thumbnailID);
    std::map<uint32,uint32>::iterator i = SessionManager::sm->GetActiveSession()->thumbnailIDs.find(thumbnailID);
    if (i != SessionManager::sm->GetActiveSession()->thumbnailIDs.end()) {
      SessionManager::sm->GetActiveSession()->thumbnailRefs[obj->GetID()] = i->second;
    }
  }

  LoadAnimationStep( obj, hilf );

  delete[] color;
  delete[] fillColor;
  delete[] lineStyle;

  return obj;
}






