/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_session.cpp
 authors : Juergen Vogel, Gabriela Maass, 
 date    : 18.02.2000
 desc    :

 ********************************** **********************************************************************/

#ifndef WIN32
#  include <sys/errno.h>
#  include <sys/wait.h>
#  include <unistd.h>
#endif


#include <vector>
#include <string>

#include <math.h>
#include <time.h>

#include "tcl.h"
#include "tk.h"

#ifdef WIN32
#  include <direct.h>
#include <util/refcount.cpp>
#else
#include <util/refcount.cc>
#endif

#include "mlb.h"

#include "mlb_types.h"
#include "mlb_session.h"
#include "mlb_folder.h"
#include "mlb_page.h"
#include "mlb_adu.h"
#include "mlb_gui.h"
#include "mlb_misc.h"
#include "mlb_uce.h"


//#define DEBUG

#define PENCOLOR 63
int MLB::CalculateColor(Color _color) {
  static int PenColor[PENCOLOR][3] = {
    {  0,  0,  0},{255,  0,  0},{  0,255,  0}, /* standard */
    {255,255,  0},{  0,  0,255},{255,  0,255},
    {  0,255,255},{255,255,255},
    {213,213,213},{190,190,190},{124,124,124}, /*grey */
    {103,103,103},{ 85, 85, 85},{ 51, 51, 51},
    { 28, 28, 28},
    {255,204,204},{255,154,154},{255,103,103}, /*red*/
    {255, 51, 51},{204,  0,  0}, {154,  0,  0},
    {103,  0,  0},{ 51,  0,  0},
    {204,255,204},{154,255,154},{103,255,103}, /*green*/
    { 51,255, 51},{  0,204,  0},{  0,154,  0},
    {  0,103,  0},{  0, 51,  0},
    {255,255,204},{255,255,154},{255,255,103}, /*yellow*/
    {255,255, 51},{204,204,  0},{154,154,  0},
    {103,103,  0},{ 51, 51,  0},
    {204,204,255},{154,154,255},{103,103,255}, /*blue*/
    { 51, 51,255},{  0,  0,204},{  0,  0,154},
    {  0,  0,103},{  0,  0, 51},
    {255,204,255},{255,154,255},{255,103,255}, /*pink*/
    {255, 51,255},{204,  0,204},{154,  0,154},
    {103,  0,103},{ 51,  0, 51},
    {204,255,255},{154,255,255},{103,255,255}, /*turquoise*/
    { 51,255,255},{  0,204,204},{  0,154,154},
    {  0,103,103},{  0, 51, 51}};
  XColor
    *xColor;
  int
    ret, i,
    min_dist,
    zw_dist,
    r, g, b;


   xColor = Tk_GetColor(Et_Interp,Tk_MainWindow(Et_Interp),(char *)_color.c_str());
   r = (xColor->red+1)/256;
   g = (xColor->green+1)/256;
   b = (xColor->blue+1)/256;

   ret = 0;
   min_dist = sqrt((PenColor[0][0]-r)*(PenColor[0][0]-r)+(PenColor[0][1]-g)*(PenColor[0][1]-g)+(PenColor[0][2]-b)*(PenColor[0][2]-b));
   for (i=1; i<PENCOLOR; i++) {
     zw_dist = sqrt((PenColor[i][0]-r)*(PenColor[i][0]-r)+(PenColor[i][1]-g)*(PenColor[i][1]-g)+(PenColor[i][2]-b)*(PenColor[i][2]-b));
     if (zw_dist < min_dist) {
       min_dist = zw_dist;
       ret = i;
     }
   }

   return ret;
}


MLB::Session::Session(uint32 _id, int _type, int _status)
  : MLB::Folder(_id,"Shared"), sessionType(_type), sessionStatus(_status), documentName("noname.mlb"), documentPath(""), thumbnailFolder(NULL) /*, animationEditMode(MLB::SessionManager::sm->GetAnimationEditMode())*/
{

  SessionManager::sm->SetActiveFolder(this);

  if (sessionType == PRIVATE) {
    SetName("Private");
  }

  GUI::PageView::ActivateNewPageButton(true);
  Init();
}


MLB::Session::~Session() {
  if (!subContainer.empty()) {
    for (std::list<MLB::GenericContainer *>::iterator temp = subContainer.begin(); temp != subContainer.end(); ++temp)
      delete((*temp));
    subContainer.clear();
  }
  documentName.erase();
  documentPath.erase();
  thumbnailRefs.clear();
  thumbnailIDs.clear();
}


void MLB::Session::SetName(char *_name) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SetName" << endl;
  #endif

  containerName.erase();

  if (!_name)
    containerName = "Session";
  else
    containerName = _name;
}


void MLB::Session::DrawPageTree(std::string _nodeNumber, std::string _nodePath, std::string parentNode, int pos, bool *_open) {
  #ifdef _DEBUG1
  cout << "MLB::Session::DrawPageTree" << endl;
  #endif

  std::string nodeNumber;
  char buf[200];
  int count;

  sprintf(buf,".%u",id);
  nodePath = _nodePath + buf;

  if (nodeName.empty()) {
    sprintf(buf, "%u", id);
    nodeName = std::string("n:") + buf;
    GUI::PageView::InsertSession(nodeName.c_str(),containerName.c_str(),parentNode.c_str(),nodePath.c_str(),pos);
  }

  if (this == SessionManager::sm->GetActiveFolder())
    GUI::PageView::SelectFolderNode(nodeName.c_str());

  if (!subContainer.empty()) {
    count = 1;
    for (std::list<GenericContainer*>::iterator i = subContainer.begin(); i != subContainer.end(); ++i) {
      sprintf (buf,"%u",count);
      nodeNumber = _nodeNumber + buf;
      (*i)->DrawPageTree(nodeNumber,nodePath,nodeName,count-1,_open);
      ++count;
    }
  }
  if (*_open) {
    OpenFolder();
  }

  nodeNumber.erase();

  GUI::PageView::Redraw();
}


void  MLB::Session::SaveContainer(FILE *fp, int level) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SaveContainer" << endl;
  #endif

  std::list<GenericContainer *>::iterator
    temp;

  MLB::SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  MLB::SessionManager::sm->progress->ShowProgress("Save Session",0);

  fprintf(fp,"<MLB name=\"%s\">\n",containerName.c_str());

  if (!subContainer.empty()) {
    MLB::SessionManager::sm->progress->IncrementProgress();
    temp = subContainer.begin();
    while (temp!=subContainer.end()) {
      (*temp)->SaveContainer(fp,(level+1));
      ++temp;
    }
  }

  fprintf(fp,"</MLB>\n");

  MLB::SessionManager::sm->cursor->Reset();
  MLB::SessionManager::sm->progress->HideProgress();
}


void MLB::Session::LoadSession(char *filename) {
  #ifdef _DEBUG1
  cout << "MLB::Session::LoadSession" << endl;
  #endif

  MLB::Folder
    *folder;
  MLB::Page
    *page;
  SGMLTree
    *tree;
  SGMLElement
    *root,
    *hilf;

  tree = new SGMLTree(filename);
  containerName.erase();

  root = tree->root;
  if (!root || (root->Name != "MLB")) {
    SetName("UNKNOWN");
    printf ("No current Session %s\n",filename);
    page = new Page(CommunicationManager::cm->GetObjectID(), Session::PRIVATE);
    InsertContainerCurrent(page);
    if (!currentPage) currentPage = page;
    //page->Activate(true);
    SessionManager::sm->DrawPageTree();
    return;
  }

  MLB::SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  MLB::SessionManager::sm->progress->ShowProgress("Load Document",0);

  SetName("Private");

  if (MLB::SessionManager::sm->GetActivePage()) MLB::SessionManager::sm->GetActivePage()->DeActivate();

  hilf = root->son;
  while (hilf) {
    MLB::SessionManager::sm->progress->IncrementProgress();
    if (hilf->Name == "FOLDER") {
      folder = new MLB::Folder(CommunicationManager::cm->GetObjectID(), hilf);
      InsertContainerCurrent(folder);
    } else if (hilf->Name == "THUMBNAILS") {
      thumbnailFolder = folder = new MLB::Folder(CommunicationManager::cm->GetObjectID(), hilf);
      thumbnailFolder->SetOrder(MLB_OO_END);
      InsertContainer(thumbnailFolder);
    } else if (hilf->Name == "PAGE") {
      page = new MLB::Page(CommunicationManager::cm->GetObjectID(), Session::PRIVATE, hilf);
      if (!MLB::SessionManager::sm->GetActivePage()) MLB::SessionManager::sm->SetActivePage(page);
      InsertContainerCurrent(page);
      if (!currentPage) currentPage = page;
      //page->Activate(true);
      SessionManager::sm->DrawPageTree();
    }
    else printf ("unknown container type %s\n",hilf->Name.c_str());

    hilf = hilf->next;
  }

  delete tree;

  MLB::SessionManager::sm->cursor->Reset();
  MLB::SessionManager::sm->progress->HideProgress();

  MLB::SessionManager::sm->GetActivePage()->Activate(true);
}


/* Functions to write PostScript */

void MLB::Session::InitPs(FILE *fp) {
  #ifdef _DEBUG1
  cout << "MLB::Session::InitPs" << endl;
  #endif

  fprintf(fp,"%%!PS-Adobe-2.0\n");
  fprintf(fp,"%%%%Creator: MLB\n");
  fprintf(fp,"%%%%Title: test.eps\n");
  fprintf(fp,"%%%%Pages: %d\n",GetNbrOfPages());
  fprintf(fp,"%%%%EndComments\n");
  fprintf(fp,"%%%%BeginProlog\n");
  fprintf(fp,"systemdict /ISOLatin1Encoding known not {\n");
  fprintf(fp,"  /ISOLatin1Encoding [\n");
  fprintf(fp,"    /space /breve /caron /dotaccent /dotlessi /fi /fl /fraction /hungarumlaut /Lslash /lslash\n");
  fprintf(fp,"    /minus /ogonek /ring /Zcaron /zcaron /space /space /space /space /space\n");
  fprintf(fp,"    /space /space /space /space /space /space /space /space /space\n");
  fprintf(fp,"    /space /space /space /exclam /quotedbl /numbersign /dollar /percent /ampersand\n");
  fprintf(fp,"    /quotesingle /parenleft /parenright /asterisk /plus /comma /hyphen /period /slash\n");
  fprintf(fp,"    /zero /one /two /three /four /five /six /seven /eight /nine /colon /semicolon /less\n");
  fprintf(fp,"    /equal /greater /question /at /A /B /C /D /E /F /G /H /I /J /K /L /M /N /O /P /Q /R\n");
  fprintf(fp,"    /S /T /U /V /W /X /Y /Z /bracketleft /backslash /bracketright /asciicircum /underscore\n");
  fprintf(fp,"    /grave /a /b /c /d /e /f /g /h /i /j /k /l /m /n /o /p /q /r /s /t /u /v /w /x /y /z\n");
  fprintf(fp,"    /braceleft /bar /braceright /asciitilde /space /Euro /space /quotesinglbase /florin\n");
  fprintf(fp,"    /quotedblbase /ellipsis /dagger /daggerdbl /circumflex /perthousand /Scaron /guilsinglleft\n");
  fprintf(fp,"    /OE /space /space /space /space /quoteleft /quoteright /quotedblleft /quotedblright\n");
  fprintf(fp,"    /bullet /endash /emdash /tilde /trademark /scaron /guilsinglright /oe /space /space /Ydieresis\n");
  fprintf(fp,"    /space /exclamdown /cent /sterling /currency /yen /brokenbar /section /dieresis /copyright\n");
  fprintf(fp,"    /ordfeminine /guillemotleft /logicalnot /space /registered /macron /degree /plusminus\n");
  fprintf(fp,"    /twosuperior /threesuperior /acute /mu /paragraph /periodcentered /cedilla /onesuperior\n");
  fprintf(fp,"    /ordmasculine /guillemotright /onequarter /onehalf /threequarters /questiondown /Agrave\n");
  fprintf(fp,"    /Aacute /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla /Egrave /Eacute /Ecircumflex\n");
  fprintf(fp,"    /Edieresis /Igrave /Iacute /Icircumflex /Idieresis /Eth /Ntilde /Ograve /Oacute /Ocircumflex\n");
  fprintf(fp,"    /Otilde /Odieresis /multiply /Oslash /Ugrave /Uacute /Ucircumflex /Udieresis /Yacute /Thorn\n");
  fprintf(fp,"    /germandbls /agrave /aacute /acircumflex /atilde /adieresis /aring /ae /ccedilla /egrave\n");
  fprintf(fp,"    /eacute /ecircumflex /edieresis /igrave /iacute /icircumflex /idieresis /eth /ntilde /ograve\n");
  fprintf(fp,"    /oacute /ocircumflex /otilde /odieresis /divide /oslash /ugrave /uacute /ucircumflex /udieresis\n");
  fprintf(fp,"    /yacute /thorn /ydieresis\n");
  fprintf(fp,"] def\n");
  fprintf(fp,"} if\n");
  fprintf(fp,"/ISOEncode {\n");
  fprintf(fp,"  dup length dict begin\n");
  fprintf(fp,"    {1 index /FID ne {def} {pop pop} ifelse} forall\n");
  fprintf(fp,"    /Encoding ISOLatin1Encoding def\n");
  fprintf(fp,"    currentdict\n");
  fprintf(fp,"  end\n");
  fprintf(fp,"  /Temporary exch definefont\n");
  fprintf(fp,"} bind def\n");
  fprintf(fp,"%%%%EndProlog\n");
  fprintf(fp,"\n");
}


void MLB::Session::EndPs(FILE *fp) {
  #ifdef _DEBUG1
  cout << "MLB::Session::EndPs" << endl;
  #endif

  fprintf(fp,"%%%%Trailer\n");
  fprintf(fp,"end\n");
  fprintf(fp,"%%%%EOF\n");
}


void MLB::Session::DrawPageTree() {
  #ifdef _DEBUG1
  cout << "MLB::Session::DrawPageTree" <<  endl;
  #endif

  std::string nodeNumber = "";
  std::string nodeName   = "";
  std::string nodePath   = "";

  bool open = false;

  this->DrawPageTree(nodeNumber, nodePath, nodeName, 0, &open);
}


void MLB::Session::SaveDocumentAsPS(const char *psName) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SaveDocumentAsPs" <<  endl;
  #endif

  FILE* f = fopen(psName,"wb");

  if (f) {
    InitPs(f);
    SaveAsPs(f);
    EndPs(f);
    fclose(f);
  }
}


void MLB::Session::SaveDocumentAsHtml(const char *directory) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SaveDocumentAsHtml" <<  endl;
  #endif

  if (subContainer.empty()) return; 

  FILE* indexFp;
  std::string index;
  std::string indexFrame;

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  SessionManager::sm->progress->ShowProgress("Export Document as HTML", GetNbrOfPages());

  char* docname = GetPraefix(GetFilename((char*)documentName.c_str()));

  Page* activePage = SessionManager::sm->GetActivePage();

  //
  // 1. create directory
  //
#ifdef WIN32
  if (_mkdir(directory) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Session::SaveDocumentAsHtml - Can't create directory %s\n", directory);
      return;
    }
  }
#else
  if (mkdir(directory, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Session::SaveDocumentAsHtml - Can't create directory %s\n", directory);
      return;
    }
  }
#endif

  //
  // 2. create index frame
  //
  indexFrame = directory;
#ifdef WIN32
  indexFrame += "\\indexframe.htm";
#else
  indexFrame += "/indexframe.htm";
#endif

  indexFp = fopen((char *)indexFrame.c_str(),"wb");
  if (!indexFp) {
    GUI::MessageWindow::Error("Session::SaveDocumentAsHtml - Can't open %s\n",indexFrame.c_str());
    return;
  }

  fprintf(indexFp,"<html>\n");
  fprintf(indexFp,"  <head>\n");
  fprintf(indexFp,"    <title>Presentation: Overview</title>\n");
  fprintf(indexFp,"    <base target=\"viewframe\">\n");
  fprintf(indexFp,"  </head>\n");
  fprintf(indexFp,"\n");
  fprintf(indexFp,"  <body bgcolor=\"#CBD9ED\">\n");
  fprintf(indexFp,"    <table cellpadding=2 cellspacing=1 border=0>\n");
  fprintf(indexFp,"\n");
  fprintf(indexFp,"      <tr><td>\n");
  fprintf(indexFp,"\n");
   fprintf(indexFp,"        <table cellpadding=1 cellspacing=2 border=0>\n");
  fprintf(indexFp,"          <tr>\n");
  fprintf(indexFp,"            <td valign=top><font face=\"Helvetica,Arial\"><strong>Title:</strong></td>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\">%s</td>\n", docname);
  fprintf(indexFp,"          </tr>\n");
  fprintf(indexFp,"          <tr>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\"><strong>Author:</strong></td>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\">%s</td>\n", CommunicationManager::cm->get_session_name().c_str());
  fprintf(indexFp,"          </tr>\n");
  fprintf(indexFp,"        </table>\n\n");
  fprintf(indexFp,"      </td></tr>\n\n");
  fprintf(indexFp,"\n");

  //
  // 3. save all pages as html
  //
  int nbr = 0;
  SaveAsHtml((char*)directory, docname, indexFp, nbr, GetNbrOfPages() - 1);


  //
  // 4. close index frame
  //
  fprintf(indexFp,"    </table>\n");
  fprintf(indexFp,"  </body>\n");
  fprintf(indexFp,"</html>\n");

  fclose (indexFp);
  if (activePage) activePage->Activate();
  DrawPageTree();
  delete docname;


  SessionManager::sm->progress->HideProgress();
  SessionManager::sm->cursor->Reset();
}


void MLB::Session::SaveDocumentAsImage(const char* name, const char* type) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SaveDocumentAsImage" <<  endl;
  #endif

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  SessionManager::sm->progress->ShowProgress("Export Document as Image", GetNbrOfPages());

  Page* active = SessionManager::sm->GetActivePage();

  int nbr = 0;
  SaveAsImage((char*)name, (char*)type, nbr);

  if (active) active->Activate();

  SessionManager::sm->progress->HideProgress();
  SessionManager::sm->cursor->Reset();
}


int MLB::Session::Save(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::Save" <<  endl;
  #endif

  char* filename = strdup(argv[1]);
  FILE* datei = fopen(filename,"wb");
  if (datei) {
    SaveContainer(datei);
    fclose(datei);
  }
  else {
    delete[] filename;
    return -1;
  }

  delete[] filename;
  return 1;
}


int MLB::Session::Open(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::Open" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
    if (SessionManager::sm->GetParentSession(SessionManager::sm->GetActivePage())->GetSessionType() == Session::SHARED) {
      CommunicationManager::cm->DeactivateAllSubcomponents();
    }
    SessionManager::sm->GetActivePage()->DeActivate();
  }

  SessionManager::sm->SetActiveFolder(NULL);

  documentName = argv[1];
  documentPath = MLB::GetPath((char*)argv[1]);
  SessionManager::sm->SetActiveFolder(this);
  LoadSession((char *)documentName.c_str());

  return 1;
}


int MLB::Session::Open(char *_filename) {
  #ifdef _DEBUG1
  cout << "MLB::Session::Open" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
    if (SessionManager::sm->GetParentSession(SessionManager::sm->GetActivePage())->GetSessionType() == Session::SHARED) {
      CommunicationManager::cm->DeactivateAllSubcomponents();
    }
    SessionManager::sm->GetActivePage()->DeActivate();
  }

  SessionManager::sm->SetActiveFolder(NULL);

  documentName = _filename;
  documentPath = MLB::GetPath(_filename);
  SessionManager::sm->SetActiveFolder(this);
  LoadSession((char *)documentName.c_str());

  return 1;
}


void MLB::Session::ChangeContainerName(std::string nodePath, char* name) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeContainerName" <<  endl;
  #endif

  std::vector<uint32>* idVector = SessionManager::sm->CreateIdVector(nodePath);
  if (!idVector) return; // error
  
  SessionManager::sm->SelectContainer(nodePath);
  GenericContainer* container = (GenericContainer*)GetObject(idVector->back());
  if (container) {
    if (sessionType == Session::SHARED) {
      RefCount<Buffer> buf = new Buffer(container->GetChangeNameSize(name));
      container->FlushChangeName(buf->getCont(), name);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(container->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    } else {
      container->SetName(name);
    }
  }

  delete idVector;
}


void MLB::Session::ChangeLineColor(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeColor" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeLineColor(argc,argv);
}


void MLB::Session::ChangeFillColor(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeColor" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeFillColor(argc, argv);
}


void MLB::Session::ChangeLineWidth(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeLineStyle" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeLineWidth(argc,argv);
}


void MLB::Session::ChangeLineStyle(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeLineStyle" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeLineStyle(argc,argv);
}


void MLB::Session::ChangeCapStyle(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeCapStyle" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeCapStyle(argc,argv);
}


void MLB::Session::ChangeJoinStyle(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeJoinStyle" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeJoinStyle(argc,argv);
}


void MLB::Session::ChangeArrowStyle(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeArrowStyle" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeArrowStyle(argc,argv);
}


void MLB::Session::ChangeArrowShape(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeArrowShape" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeArrowShape(argc,argv);
}


void MLB::Session::ChangeFont(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeFont" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeFont(argc,(char**)argv);
}


void MLB::Session::ChangeObject(int argc, const char* argv[], int _type) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ChangeObject" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ChangeObject(argc,argv,_type);
}


void MLB::Session::NewTextLine(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::NewTextLine" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->InsertTextLine(argc,argv);
}


void MLB::Session::EndText(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::EndText" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->EndText(argc,argv);
}


bool MLB::Session::LoadImage(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::LoadImage" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return false; // no new objects if not in synch mode

  if (thumbnailFolder && SessionManager::sm->GetActivePage())
    if (SessionManager::sm->IsParentOf(thumbnailFolder->GetNodePath(), SessionManager::sm->GetActivePage()->GetNodePath()))
      return false; // not possible

		typedef std::vector<Magick::Image> ImageList;

  ImageList   imageList;
  MLB::Page*  page;
  char
    *zw,
    *filename,
    *imageName;
  int
    x, y,
    i;

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  SessionManager::sm->progress->ShowProgress("Load Image",0);

  filename = (char*)argv[1];
  SessionManager::sm->progress->IncrementProgress();

  try { Magick::readImages( &imageList, filename ); }
  catch ( Magick::Exception& exc )
  {
    GUI::MessageWindow::Error( "Cannot read images: %q", exc.what() );
    SessionManager::sm->progress->HideProgress();
    SessionManager::sm->cursor->Reset();
    return false;
  }

  SessionManager::sm->progress->IncrementProgress();

  if( 1 <= imageList.size() ) {
    if( 1 == imageList.size() ) {
      // only one image
      imageName = filename;
      x = atoi(argv[2]) / SessionManager::sm->GetZoom();
      y = atoi(argv[3]) / SessionManager::sm->GetZoom();
      if (!SessionManager::sm->GetActivePage()) return false;
      SessionManager::sm->GetActivePage()->InsertImage(*imageList.begin(), imageName, x, y, true );
    }
    else {
      // more than one image
      // insert in new private document !
      Session* session = new Session(CommunicationManager::cm->GetObjectID(), Session::PRIVATE, Session::INACTIVE);
      SessionManager::sm->AddSession(session);
      SessionManager::sm->SetActiveSession(session->GetID());
      GUI::Attributes::ResetAction();

      if (SessionManager::sm->GetActivePage()) SessionManager::sm->GetActivePage()->DeActivate();

      i = 0;
      for (ImageList::iterator it = imageList.begin(); it !=imageList.end(); ++it) {
	SessionManager::sm->progress->IncrementProgress();
	page = new Page(CommunicationManager::cm->GetObjectID(), Session::PRIVATE);
        if (!SessionManager::sm->GetActivePage()) SessionManager::sm->SetActivePage(page);
	session->InsertContainerCurrent(page);
	zw =  GetPraefix(filename);
	imageName = new char[strlen(zw)+10];
	sprintf(imageName,"%s%d.png",zw,i++);
	delete[] zw;
	page->InsertImage( *it, imageName, 0, 0, false );
	delete[] imageName;
	//page->Activate(true);
	SessionManager::sm->DrawPageTree();
      }
      SessionManager::sm->GetActivePage()->Activate(true);
    }
  }

  SessionManager::sm->progress->HideProgress();
  SessionManager::sm->cursor->Reset();

  return true;
} // LoadImage()


void MLB::Session::PlaceImage(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::PlaceImage" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->PlaceImage(argc,argv);
}


void MLB::Session::AppendNewObject(int argc, const char* argv[], int _type) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return; // no new objects if not in synch mode

  #ifdef _DEBUG1
  cout << "MLB::Session::AppendNewObject" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->AppendNewObject(argc,(char**)argv,_type);
}


void MLB::Session::MoveObjects(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::MoveObject" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->MoveObjects(argc,argv);
}


void MLB::Session::CopyObjects(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::CopyObject" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) &&
      !SessionManager::sm->SynchMode() &&
      !SessionManager::sm->AnnotationMode())
    return;

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->CopyObjects(argc,argv);
}


void MLB::Session::RaiseObjects() {
  #ifdef _DEBUG1
  cout << "MLB::Session::RaiseObject" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) &&
      !SessionManager::sm->SynchMode() &&
      !SessionManager::sm->AnnotationMode())
    return;

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->RaiseObjects();
}


void MLB::Session::LowerObjects() {
  #ifdef _DEBUG1
  cout << "MLB::Session::LowerObject" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) &&
      !SessionManager::sm->SynchMode() &&
      !SessionManager::sm->AnnotationMode())
    return;

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->LowerObjects();
}


void MLB::Session::DeleteObject(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::DeleteObject" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->DeleteObject(argc,argv);
}


void MLB::Session::SetPageDimension(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::SetPageDimension" << endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
    int width  = atoi(argv[1]);
    int height = atoi(argv[2]);

    if (sessionType == Session::SHARED) {
      Page* page = SessionManager::sm->GetActivePage();
      RefCount<Buffer> buf = new Buffer(page->GetChangeSizeSize());
      page->FlushChangeSize(buf->getCont(), width, height);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(page->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    } else {
      SessionManager::sm->GetActivePage()->SetPageDimension(width,height);
    }
  }
}


void MLB::Session::SetPageColor(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::SetPageDimension" << endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
    if (sessionType == Session::SHARED) {
      Page* page = SessionManager::sm->GetActivePage();
      RefCount<Buffer> buf = new Buffer(page->GetChangeColorSize());
      page->FlushChangeColor(buf->getCont(), argv[1]);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(page->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    } else {
      SessionManager::sm->GetActivePage()->SetBackgroundColor(argv[1]);
    }
  }
}


void MLB::Session::ActivateOnlinePointer(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ActivateOnlinePointer" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ActivateOnlinePointer(argc,argv);
}


void MLB::Session::ActivateOnlineBox(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::ActivateOnlineBox" << endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ActivateOnlineBox(argc,argv);
}


void MLB::Session::SetWindowPointer(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::SetWindowPointer" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode())
    return;

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->SetWindowPointer(argc,argv);
}


void MLB::Session::UnSetWindowPointer(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::UnSetWindowPointer" << endl;
  #endif

  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode())
    return;

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->UnSetWindowPointer(argc,argv);
}


void MLB::Session::SelectObjects(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;

  #ifdef _DEBUG1
  cout << "MLB::Session::SelectObjects" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
   SessionManager::sm->GetActivePage()->SelectObjects(argc,argv);
  }
}


void MLB::Session::UnSelectObjects(int argc, const char* argv[]) {
  if ((GetSessionType() == Session::SHARED) && !SessionManager::sm->SynchMode() && !SessionManager::sm->AnnotationMode())
    return;
 
  #ifdef _DEBUG1
  cout << "MLB::Session::UnSelectObjects" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage()) {
    SessionManager::sm->GetActivePage()->UnSelectObjects();
  }
}


void MLB::Session::Draw() {
  #ifdef _DEBUG1
  cout << "MLB::Session::Draw" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->Draw();
}


void MLB::Session::DrawSelected() {
  #ifdef _DEBUG1
  cout << "MLB::Session::DrawSelected" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->DrawSelected();
}


void MLB::Session::ShowTicks(int argc, const char* argv[]) {
  #ifdef _DEBUG1
  cout << "MLB::Session::ShowTicks" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->ShowTicks(argc,argv);
}


void MLB::Session::CutObjects() {
  #ifdef _DEBUG1
  cout << "MLB::Session::CutObjects" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->CutObjects();
}


void MLB::Session::CopyObjects() {
  #ifdef _DEBUG1
  cout << "MLB::Session::CopyObjects" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->CopyObjects();
}


void MLB::Session::PasteObjects() {
  #ifdef _DEBUG1
  cout << "MLB::Session::PasteObjects" <<  endl;
  #endif

  if (SessionManager::sm->GetActivePage())
    SessionManager::sm->GetActivePage()->PasteObjects();
}


void MLB::Session::CreateThumbnails(int argc, const char* argv[]) {
  if (GetSessionType() == Session::SHARED) return;
    
  bool flat = (atoi(argv[1]) == 0);

  Page* active = SessionManager::sm->GetActivePage();
  if (active) active->DeActivate();

  SessionManager::sm->cursor->Set(GUI::Cursor::WATCH);
  SessionManager::sm->progress->ShowProgress("Create Thumbnails",0);
  SessionManager::sm->progress->IncrementProgress();

  if (!thumbnailFolder) {
    thumbnailFolder = new Folder(CommunicationManager::cm->GetObjectID(), "Thumbnails");
    thumbnailFolder->SetOrder(MLB_OO_END);
    InsertContainer(thumbnailFolder);
  } else {
    thumbnailFolder->DeleteAll(NULL);
    SessionManager::sm->ClearPageTree();
  }

  thumbnailRefs.clear();
  thumbnailIDs.clear();

  SessionManager::sm->DrawPageTree();

  int count = 0, x = 0, y = 0;
  ((Folder*)this)->CreateThumbnails(thumbnailFolder, count, x, y, flat);

  if (thumbnailFolder->GetFirstPage()) {
    SessionManager::sm->DrawPageTree();
    SessionManager::sm->SelectContainer(((Page*)thumbnailFolder->GetFirstPage())->GetNodePath());
  }

  SessionManager::sm->progress->HideProgress();
  SessionManager::sm->cursor->Reset();
  SessionManager::sm->DrawPageTree();
}


void MLB::Session::SelectThumbnail(int argc, const char* argv[]) {
  if (GetSessionType() == Session::SHARED) return;
 
  static int x = 1;
  static int y = 1;  
  
  int status = atoi(argv[1]);
  
  Page* active = SessionManager::sm->GetActivePage();
  bool thumbnailPage = thumbnailFolder && active && SessionManager::sm->IsParentOf(thumbnailFolder->GetNodePath(), active->GetNodePath());
  if (!thumbnailPage) return;
  
  switch (status) {
  case 1: { // select first thumbnail
    x = y = 1;
    if (!active->objects.empty()) {
    DrawObject* marker = active->objects.front();
    marker->Move(active->canvas, thumbnailFolder->thumbnailFirstX-marker->GetX(), thumbnailFolder->thumbnailFirstY-marker->GetY());
    }
    // mark shared pages
    // marker is the first object
    for (std::list<DrawObject*>::iterator i = ++active->objects.begin(); i != active->objects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::RECTANGLE) {
        // thumbnails are the first pages: all thumbnails found?
        if (((Rectangle*)*i)->GetThumbnailID() == -1) break;
    
        std::map<uint32,uint32>::iterator j = thumbnailRefs.find((*i)->GetID());
        if (j != thumbnailRefs.end()) {
          Page* page = (Page*)SessionManager::sm->GetObject(j->second);
          if (page) {
              if (SessionManager::sm->GetParentSession(page)->GetSessionType() == Session::SHARED) {
              ((ColoredObject*)*i)->SetFillColor("#C0C0C0");
              } else {
              ((ColoredObject*)*i)->SetFillColor("none");
            }
          }
        }
      }
    }
    break;
  }
  case 2: { // move marker (pressed key)
    int diffX  = atoi(argv[2]);
    int diffY  = atoi(argv[3]);
    
    int nbrImg = 0;
    // marker is the first object
    for (std::list<DrawObject*>::iterator i = ++active->objects.begin(); i != active->objects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::RECTANGLE) {
        if (((Rectangle*)*i)->GetThumbnailID() == -1) break;
        ++nbrImg;
      }
    }
    
    if ((x+diffX >= 1) && (nbrImg >= (y-1)*thumbnailFolder->thumbnailNbrX+x+diffX) && (x+diffX <= thumbnailFolder->thumbnailNbrX)) x += diffX;
    if ((y+diffY >= 1) && (nbrImg >= (y+diffY-1)*thumbnailFolder->thumbnailNbrX+x)) y += diffY;
    int _x = thumbnailFolder->thumbnailFirstX + (x-1)*thumbnailFolder->thumbnailIncrX;
    int _y = thumbnailFolder->thumbnailFirstY + (y-1)*thumbnailFolder->thumbnailIncrY;
    if (!active->objects.empty()) {
    DrawObject* marker = active->objects.front();
    marker->Move(active->canvas, _x-marker->GetX(), _y-marker->GetY());
    }
    break;
  }
  case 3:   // select thumbnail (x,y)
  case 4:   // dobble click 
  case 5: { // return key
    int _x, _y, x1, y1, x2, y2;
    for (std::list<DrawObject*>::iterator i = ++active->objects.begin(); i != active->objects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::RECTANGLE) {
        // all thumbnails found?
        if (((Rectangle*)*i)->GetThumbnailID() == -1) break;
  
        (*i)->GetBBox(&x1,&y1,&x2,&y2);
        switch (status) {
        case 3: { // select thumbnail (x,y)
          _x = atoi(argv[2]) / SessionManager::sm->GetZoom();
          _y = atoi(argv[3]) / SessionManager::sm->GetZoom();
          if ((_x >= x1) && (_x <= x2) && (_y >= y1) && (_y <= y2)) {
            x = (int)((double)(x1-thumbnailFolder->thumbnailFirstX+2) / thumbnailFolder->thumbnailIncrX) + 1;
            y = (int)((double)(y1-thumbnailFolder->thumbnailFirstY+2) / thumbnailFolder->thumbnailIncrY) + 1;
            _x = thumbnailFolder->thumbnailFirstX + (x-1)*thumbnailFolder->thumbnailIncrX;
            _y = thumbnailFolder->thumbnailFirstY + (y-1)*thumbnailFolder->thumbnailIncrY;
            if (!active->objects.empty()) {
            DrawObject* marker = active->objects.front();
            marker->Move(active->canvas, _x-marker->GetX(), _y-marker->GetY());
          }
          }
          break;
        }
        case 4: { // dobble click 
          _x = atoi(argv[2]) / SessionManager::sm->GetZoom();
          _y = atoi(argv[3]) / SessionManager::sm->GetZoom();
          if ((_x >= x1) && (_x <= x2) && (_y >= y1) && (_y <= y2)) {
            std::map<uint32,uint32>::iterator j = thumbnailRefs.find((*i)->GetID());
            if (j != thumbnailRefs.end()) {
              Page* page = (Page*)SessionManager::sm->GetObject(j->second);
              if (page) SessionManager::sm->SelectContainer(page->GetNodePath());
            }
          }      
          break;
        }
        case 5: { // return key
          DrawObject* marker = active->objects.front();
          _x = thumbnailFolder->thumbnailFirstX + (x-1)*thumbnailFolder->thumbnailIncrX + marker->GetWidth() / 2;
          _y = thumbnailFolder->thumbnailFirstY + (y-1)*thumbnailFolder->thumbnailIncrY + marker->GetWidth() / 2;
          if ((_x >= x1) && (_x <= x2) && (_y >= y1) && (_y <= y2)) {
            std::map<uint32,uint32>::iterator j = thumbnailRefs.find((*i)->GetID());
            if (j != thumbnailRefs.end()) {
                Page* page = (Page*)SessionManager::sm->GetObject(j->second);
              if (page) SessionManager::sm->SelectContainer(page->GetNodePath());
            }
          }
          break;
        }
        }
      }
    }
    break;
  }
  }
}


//-------------------------------------------------------------------------//

/**   Changes animation edit/presentation mode. If needed shows or hides
  *   all the objects on session's pages.
  *   @param mode   edit mode: true if edit, false if presentation
  **/
/*
void MLB::Session::SetAnimationEditMode( bool mode )
{
  // nothing to do
  if( animationEditMode == mode ) return;

  animationEditMode = mode;
  SetAnimationObjectsVisible( mode );

} // SetAnimationEditMode()
*/
//-------------------------------------------------------------------------//

//
// class SessionManager
//

MLB::SessionManager* MLB::SessionManager::sm = NULL;

MLB::SessionManager::SessionManager(const char* path)
: activeFolder(NULL), activePage(NULL), globalPage(0), globalFolder(0), nodeName(""),
  nodePath(""), annotationMode(false), synchMode(true), logPageChanges(false), pageCount(0), animationEditMode(true) {

  activeSession = sessions.end();
  programPath   = path;
  fontList      = new MLB::FontList((char *)programPath.c_str());
  zoom          = 1.0;
  recording     = false;
  paused        = false;
  cursor        = new GUI::Cursor();
  progress      = new GUI::Progress();
  aofRecording  = NULL;
  frameRate     = 16000;
  useJedas	= false;
  useAppPlay   	= false;
  mute          = 0;
}


MLB::SessionManager::~SessionManager() {
  DeleteAllSessions();

  delete fontList;
  delete cursor;
  delete progress;

  programPath.erase();
  recordedPages.clear();
}


void MLB::SessionManager::PrintActiveSubcomponents() {
  GUI::MessageWindow::Debug("SessionManager::PrintActiveSubcomponents : ");
  if (GetSharedSession()) GetSharedSession()->PrintActiveSubcomponents();
  GUI::MessageWindow::Debug("\n");
}


bool MLB::SessionManager::ImportDocument(const char* filename) {
  #ifdef _DEBUG1
  cout << "MLB::SessionManager::ImportDocument" << endl;
  #endif

  typedef std::vector<Magick::Image> ImageList;

  ImageList   imageList;
  MLB::Page*  page;
  char
    *zw,
    *imageName;
  int
    x, y,
    i;

  cursor->Set(GUI::Cursor::WATCH);
  progress->ShowProgress("Load Image",0);

  progress->IncrementProgress();

  try {
    Magick::readImages(&imageList, filename);
  } catch (Magick::Exception& exc)  {
    GUI::MessageWindow::Error("SessionManager::ImportDocument - Cannot read file: %q", exc.what() );
    progress->HideProgress();
    cursor->Reset();
    return false;
  }

  progress->IncrementProgress();

  if( 1 <= imageList.size() ) {
    if( 1 == imageList.size() ) {
      // only one image
      imageName = (char*)filename;
      if (!GetActiveSession() || !GetActivePage()) {
	progress->HideProgress();
	cursor->Reset();
	return false;
      }
      GetActivePage()->InsertImage(*imageList.begin(), imageName, 0, 0, true);
    } else {
      // more than one image
      // insert in new private document !
      Session* session = new Session(CommunicationManager::cm->GetObjectID(), Session::PRIVATE, Session::INACTIVE);
      AddSession(session);
      SetActiveSession(session->GetID());
      GUI::Attributes::ResetAction();

      if (GetActivePage()) GetActivePage()->DeActivate();

      i = 0;
      for (ImageList::iterator it = imageList.begin(); it !=imageList.end(); ++it) {
	progress->IncrementProgress();
	page = new Page(CommunicationManager::cm->GetObjectID(), Session::PRIVATE);
        if (!GetActivePage()) SetActivePage(page);
	session->InsertContainerCurrent(page);
	zw =  GetPraefix((char*)filename);
	imageName = new char[strlen(zw)+10];
	sprintf(imageName,"%s%d.png",zw,i++);
	delete[] zw;
	page->InsertImage( *it, imageName, 0, 0, false );
	delete[] imageName;
	//page->Activate(true);
	DrawPageTree();
      }
      GetActivePage()->Activate(true);

      GUI::Toplevel::SetMenuState("sessionedit normal");
    }
  }

  progress->HideProgress();
  cursor->Reset();

  return true;
}


void MLB::SessionManager::AddSession(MLB::Session* session) {
  if (GetSession(session->GetID())) return; // session already exists
  sessions.push_back(session);
}


MLB::Session* MLB::SessionManager::GetSession(uint32 id) {
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i)
    if ((*i)->GetID() == id)
      return *i;

  return NULL; // not found
}


MLB::Session* MLB::SessionManager::RemoveSession(uint32 id) {
  if (sessions.empty()) return NULL;

  Session* toDel = NULL;

  if (GetActiveFolder() && (GetParentSession(GetActiveFolder())->GetID() == id)
      && GetActivePage()  && (GetParentSession(GetActivePage())->GetID() != id)) {
    SetActiveFolder((Folder*)GetParentContainer(GetActivePage()->GetID()));
  }

  if (GetActivePage() && (GetParentSession(GetActivePage())->GetID() == id)) {
    GetActivePage()->DeActivate();
    SetActivePage(NULL);
    SetActiveFolder(NULL);
  }

  ++activeSession;

  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    if ((*i)->GetID() == id) {
      toDel = *i;

      if (toDel->GetSessionType() == Session::SHARED) {
	globalPage = 0;
	globalFolder = 0;
      }

      sessions.erase(i);

      if (sessions.size() == 0) {
	GUI::PageView::ActivateNewPageButton(false);
	activeSession = sessions.end();
      }

      return toDel;
    }
  }

  return NULL; // not found
}


void MLB::SessionManager::DeleteAllSessions() {
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i)
    delete *i;

  sessions.clear();
  SetActivePage(NULL);
  SetActiveFolder(NULL);
  globalPage = 0;
  globalFolder = 0;
  activeSession = sessions.end();

  ClearPageTree();
  DrawPageTree();
}


void MLB::SessionManager::SetActiveSession(uint32 id) {
  if (activeSession != sessions.end())
    (*activeSession)->SetSessionStatus(Session::INACTIVE); // deactivate

  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    if ((*i)->GetID() == id) {
      activeSession = i;
      (*i)->SetSessionStatus(Session::ACTIVE);
      break;
    }
  }

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    GUI::PageView::ActivateAnnotationButton(true);
  } else { // PRIVATE
    GUI::PageView::ActivateAnnotationButton(false);
    //    if ((*activeSession)->GetCurrentPage()) (*activeSession)->GetCurrentPage()->Activate();
  }
}


MLB::Session* MLB::SessionManager::GetActiveSession() {
  if (activeSession == sessions.end()) return NULL; // no sessions yet

  return *activeSession;
}


MLB::Session* MLB::SessionManager::GetSharedSession() {
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i)
    if ((*i)->GetSessionType() == Session::SHARED)
      return *i;

  return NULL; // not found
}


void MLB::SessionManager::CloseSharedSession() {
  Session* shared = GetSharedSession();

  uint32 newID = MLB::CommunicationManager::cm->GetObjectID();

  shared->SetSessionType(Session::PRIVATE);
  shared->SetName("Private");
  shared->SetID(newID);
  shared->ReplaceIDs();

  globalPage = 0;
  globalFolder = 0;

  SetActiveSession(newID);

  ClearPageTree();
  ClearNodeNames();
  DrawPageTree();
}


void MLB::SessionManager::DeleteActiveSession() {
  if (activeSession == sessions.end()) return; // no sessions yet

  Session* toDel = *activeSession;

  if (toDel->GetType() == Session::SHARED)
    MLB::CommunicationManager::cm->CloseRTPI();

  // deactivate active page if included in the session to close
  if (GetActivePage() && GetParentSession(GetActivePage()) == toDel) {
    GetActivePage()->DeActivate();
    SetActivePage(NULL);
  }
  if (GetActiveFolder() && GetParentSession(GetActiveFolder()) == toDel) {
    SetActiveFolder(NULL);
  }

  toDel->DeleteAll(NULL);
  RemoveSession(toDel->GetID());

  delete toDel;

  if (!sessions.empty()) {
    SetActiveSession((*sessions.begin())->GetID());
  } else {
    activeSession = sessions.end();
  }

  ClearPageTree();
  DrawPageTree();

  if (sessions.empty()) {
    GUI::Toplevel::SetMenuState("sessionedit disabled");
  }
}


void MLB::SessionManager::InsertNewFolder(int argc, const char* argv[]) {
  if (!GetActiveSession()) {
    GUI::MessageWindow::Error("Open Session first\n");
    GUI::Toplevel::SetStatus("Open Document first");
    return;
  }

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && !SynchMode())
    return;

  if (!GetActiveFolder())
    SetActiveFolder(*activeSession);

  if (GetActiveSession()->thumbnailFolder)
    if (IsParentOf(GetActiveSession()->thumbnailFolder->GetNodePath(), GetActiveFolder()->GetNodePath()))
      return; // not possible

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    uint32 newID = CommunicationManager::cm->GetObjectID();

    RefCount<Buffer> buf = new Buffer(Folder::GetStateSize("Chapter"));
    Folder::FlushState(buf->getCont(), "Chapter", GetActiveFolder()->GetInsertObjectOrder());
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(newID, GetActiveFolder()->GetID(), CommunicationManager::FOLDER, 3, buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

    //buf   = new Buffer(Folder::GetChangeParentObjectSize());
    //if (GetActivePage()) Folder::FlushChangeParentObject(buf->getCont(), 1 + GetActiveFolder()->GetObjectOrder(GetActivePage()->GetID()));
    //else Folder::FlushChangeParentObject(buf->getCont(), 0);
    ////if (GetActiveFolder()->Empty()) Folder::FlushChangeParentObject(buf, 0);
    ////else Folder::FlushChangeParentObject(buf, 1 + GetActiveFolder()->GetObjectOrder(GetActiveFolder()->GetCurrentContainer()->GetID()));
    //adu = CommunicationManager::cm->CreateEvent(newID, GetActiveFolder()->GetID(), buf);
    //AduManager::am->InsertAdu(adu);
    //CommunicationManager::cm->SendAduReliable(adu);

    buf   = new Buffer(Folder::GetSetActiveSize());
    Folder::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(newID, buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

  } else { // PRIVATE Session

    Folder *folder = new Folder(CommunicationManager::cm->GetObjectID());
    GetActiveFolder()->InsertContainerCurrent(folder);
    SetActiveFolder(folder);

    DrawPageTree();
  }
}

void MLB::SessionManager::InsertNewPage(int argc, const char* argv[]) {
  if (!GetActiveSession()) return;

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && !SynchMode())
    return;

  if (activeSession == sessions.end()) {
    GUI::MessageWindow::Error("SessionManager::InsertNewPage - Open Session first\n");
    GUI::Toplevel::SetStatus("Open Document first");
    return;
  }

  if (!GetActiveFolder())
    SetActiveFolder(*activeSession);

  if (GetActiveSession()->thumbnailFolder)
    if (IsParentOf(GetActiveSession()->thumbnailFolder->GetNodePath(), GetActiveFolder()->GetNodePath()))
      return; // not possible

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    uint32 newID = CommunicationManager::cm->GetObjectID();

    RefCount<Buffer> buf = new Buffer(Page::GetStateSize("Page"));
    Page::FlushState(buf->getCont(), atoi(argv[1]), atoi(argv[2]), argv[3], "Page", GetActiveFolder()->GetInsertObjectOrder(), true);
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(newID, GetActiveFolder()->GetID(), CommunicationManager::PAGE, 3, buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

    //buf   = new Buffer(Page::GetChangeParentObjectSize());
    //if (GetActivePage()) Page::FlushChangeParentObject(buf->getCont(), 1 + GetActiveFolder()->GetObjectOrder(GetActivePage()->GetID()));
    //else Page::FlushChangeParentObject(buf->getCont(), 0);
    ////else if (GetActiveFolder()->Empty()) Page::FlushChangeParentObject(buf, 0);
    ////else Page::FlushChangeParentObject(buf, 1 + GetActiveFolder()->GetObjectOrder(GetActiveFolder()->GetCurrentContainer()->GetID()));
    //adu = CommunicationManager::cm->CreateEvent(newID, GetActiveFolder()->GetID(), buf);
    //AduManager::am->InsertAdu(adu);
    //CommunicationManager::cm->SendAduReliable(adu);

    if (GetActivePage()) {
      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont(), false);
      adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }

    buf   = new Buffer(Page::GetSetActiveSize());
    Page::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(newID, buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);
  }
  else {// PRIVATE session
    Page* page = new Page(CommunicationManager::cm->GetObjectID(), (*activeSession)->GetSessionType(), "Page", atoi(argv[1]), atoi(argv[2]), argv[3]);

    GetActiveFolder()->InsertContainerCurrent(page);
    GetActiveFolder()->SetCurrentPage(page);
    page->SetBackgroundColor(argv[3]);
    page->Activate();

    DrawPageTree();
  }
}


void MLB::SessionManager::InsertNewPage(MLB::Page *_page) {
  if (!GetActiveSession()) return;

  if (activeSession == sessions.end()) {
    GUI::MessageWindow::Error("SessionManager::InsertNewPage - Open Session first\n");
    GUI::Toplevel::SetStatus("Open Document first");
    return;
  }

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && !SynchMode())
    return;

  if (!GetActiveFolder())
    SetActiveFolder(*activeSession);

  if (GetActiveSession()->thumbnailFolder)
    if (IsParentOf(GetActiveSession()->thumbnailFolder->GetNodePath(), GetActiveFolder()->GetNodePath()))
      return; // not possible

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    RefCount<Buffer> buf = new Buffer(Page::GetStateSize(_page->GetName()));
    Page::FlushState(buf->getCont(), _page->GetWidth(), _page->GetHeight(), _page->GetColor(),
		     _page->GetName(), GetActiveFolder()->GetInsertObjectOrder() , true);
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(_page->GetID(), GetActiveFolder()->GetID(), CommunicationManager::PAGE, 3, buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

    //buf   = new Buffer(Page::GetChangeParentObjectSize());
    //if (GetActivePage()) Page::FlushChangeParentObject(buf->getCont(), 1 + GetActiveFolder()->GetObjectOrder(GetActivePage()->GetID()));
    //else Page::FlushChangeParentObject(buf->getCont(), 0);
    ////if (GetActiveFolder()->Empty()) Page::FlushChangeParentObject(buf, 0);
    ////else Page::FlushChangeParentObject(buf, 1 + GetActiveFolder()->GetObjectOrder(GetActiveFolder()->GetCurrentContainer()->GetID()));
    //adu = CommunicationManager::cm->CreateEvent(_page->GetID(), GetActiveFolder()->GetID(), buf);
    //AduManager::am->InsertAdu(adu);
    //CommunicationManager::cm->SendAduReliable(adu);

    if (GetActivePage()) {
      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont(), false);
      adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }

    buf   = new Buffer(Page::GetSetActiveSize());
    Page::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(_page->GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

    delete _page;
  }
  else {// PRIVATE session
    GetActiveFolder()->InsertContainerCurrent(_page);
    GetActiveFolder()->SetCurrentPage(_page);
    _page->Activate();

    DrawPageTree();
  }
}


void MLB::SessionManager::DuplicatePage() {
  if (!GetActiveSession() || !GetActiveFolder() || !GetActivePage()) return;

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && !SynchMode())
    return;

  if (GetActiveSession()->thumbnailFolder)
    if (IsParentOf(GetActiveSession()->thumbnailFolder->GetNodePath(), GetActivePage()->GetNodePath()))
      return; // not possible

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAduVector> adus = new RTPI::RtpiAduVector;

    uint32 parentID = CommunicationManager::cm->GetObjectID();
    buf = new Buffer(GetActivePage()->GetStateSize());
    GetActivePage()->FlushState(buf->getCont());
    adus->push_back(CommunicationManager::cm->CreateState(parentID, GetActiveFolder()->GetID(), CommunicationManager::PAGE, 3, buf));

    for (std::list<DrawObject*>::iterator i = GetActivePage()->objects.begin(); i != GetActivePage()->objects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;
      if ((*i)->Annotation()) continue;

      if ((*i)->GetType() == GenericObject::IMAGE) {
	buf  = new Buffer(((PhotoImage*)*i)->GetStateSize());
	((PhotoImage*)*i)->FlushState(buf->getCont());
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::IMAGE, 3, buf));
	continue;
      }

      buf = new Buffer((*i)->GetStateSize());
      (*i)->FlushState(buf->getCont());

      switch ((*i)->GetType()) {
      case GenericObject::RECTANGLE: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::RECTANGLE, 3, buf));
	break;
      }
      case GenericObject::OVAL: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::OVAL, 3, buf));
	break;
      }
      case GenericObject::LINE: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::LINE, 3, buf));
	break;
      }
      case GenericObject::IMAGE: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::IMAGE, 3, buf));
	break;
      }
      case GenericObject::FREEHAND:
      case GenericObject::POLYLINE: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::POLYLINE, 3, buf));
	break;
      }
      case GenericObject::POLYGON: {
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::POLYGON, 3, buf));
	break;
      }
//        case GenericObject::GROUP: {
//  	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::GROUP, 3, buf));
//  	break;
//        }
      }
    }

    for (std::list<Group*>::iterator j = GetActivePage()->groups.begin(); j != GetActivePage()->groups.end(); ++j) {
      if ((*j)->IsTextGroup()) {
	buf   = new Buffer((*j)->GetStateSizeTG());
	(*j)->FlushStateTG(buf->getCont());
	adus->push_back(CommunicationManager::cm->CreateState(CommunicationManager::cm->GetObjectID(), parentID, CommunicationManager::TEXT, 3, buf));
      }
    }

    AduManager::am->InsertAdus(adus);
    CommunicationManager::cm->SendAdusReliable(adus);

  } else {// PRIVATE session
    Page* page = new Page(CommunicationManager::cm->GetObjectID(), (*activeSession)->GetSessionType(), "Page",
			  GetActivePage()->GetWidth(), GetActivePage()->GetHeight(), GetActivePage()->GetColor());

    for (std::list<DrawObject*>::iterator i = GetActivePage()->objects.begin(); i != GetActivePage()->objects.end(); ++i) {
      if ((*i)->GetType() == GenericObject::ASCIITEXT) continue;

      page->AppendObject((*i)->Copy());
    }

    Group* group;
    for (std::list<Group*>::iterator j = GetActivePage()->groups.begin(); j != GetActivePage()->groups.end(); ++j) {
      group = (Group*)(*j)->Copy();
      page->AppendObject(group);
      for (std::list<DrawObject*>::iterator g = group->groupObjects.begin(); g != group->groupObjects.end(); ++g) {
	page->AppendObject(*g);
      }
    }

    GetActiveFolder()->InsertContainerCurrent(page);
    GetActiveFolder()->SetCurrentPage(page);
    page->Activate();

    DrawPageTree();
  }
}


void MLB::SessionManager::DeletePages(int from, int nbr) {
  if (!GetActiveSession() || (GetActiveSession()->GetSessionType() == Session::SHARED)) return;

  cursor->Set(GUI::Cursor::WATCH);
  
  if (GetActivePage()) GetActivePage()->DeActivate();

  // find first page to delete
  GenericContainer* page = GetActiveSession()->GetFirstPage();
  for (int j = 1; j < from; ++j) {
    if (!page) break;
    do {
      page = GetActiveSession()->GetNext(page->GetID());
    } while (page && (page->GetType() != GenericObject::PAGE));
  }

  GenericContainer* tmp;
  for (int i = 1; i <= nbr; ++i) {
    if (!page) break;
    
    // 1. remember page to delete
    tmp = page;

    // 2. find next page
    do {
      page = GetActiveSession()->GetNext(page->GetID());
    } while (page && (page->GetType() != GenericObject::PAGE));

    // 3. delete page
    ((GenericContainer*)GetActiveSession())->DeleteObject(tmp->GetID());
    delete tmp;
  }

  ClearPageTree();
  DrawPageTree();

  if (GetActiveSession()->GetFirstPage()) 
    SelectContainer(GetActiveSession()->GetFirstPage()->GetNodePath());

  cursor->Reset();
}


std::vector<uint32>* MLB::SessionManager::CreateIdVector(std::string nodePath) {
  if (nodePath.length() == 0) return NULL;

  std::vector<uint32>* idVector = new std::vector<uint32>;
  std::string::size_type posBegin;
  std::string::size_type posEnd;
  std::string::size_type n;
  uint32 id;
  char zw[20];

  posBegin = 1;
  while (posBegin < nodePath.size()) {
    posEnd = nodePath.find('.',posBegin);
    if (posEnd == std::string::npos)
      posEnd = nodePath.size();
    n = posEnd-posBegin;
    nodePath.copy(zw,n,posBegin);
    zw[n] = '\0';
    id = StringToUint32(zw);
    idVector->push_back(id);
    posBegin = posEnd+1;
  }

  if (idVector->size() == 0) {
    delete idVector;
    return NULL;
  }

  return idVector;
}

 
void MLB::SessionManager::DeleteContainer(std::string nodePath) {
  if (!GetActiveSession()) return;
  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && !SynchMode()) return;

  std::vector<uint32>* idVector = CreateIdVector(nodePath);
  if (!idVector) return;

  if ((*activeSession)->GetSessionType() == Session::SHARED) {
    GenericContainer* obj = (GenericContainer*)((*activeSession)->GetObject(idVector->back()));
    if (obj) {
      if (obj->GetType() == GenericObject::SESSION) return;  // not possible
      RefCount<Buffer> buf = new Buffer(obj->GetDeleteSize());
      obj->FlushDelete(buf->getCont());
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(obj->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);

      GenericContainer* next = ((Folder*)GetParentContainer(obj->GetID()))->GetNextContainer(obj);
      Page* newActive = NULL;
      if (next) {
	if (next->GetType() == GenericObject::PAGE) {
	  newActive = (Page*)next;
	} else {
	  if (((Folder*)next)->GetCurrentPage())
	    newActive = ((Folder*)next)->GetCurrentPage();
	}
      }
      if (newActive) {
	if (GetActivePage()) {
	  buf   = new Buffer(Page::GetSetActiveSize());
	  Page::FlushSetActive(buf->getCont(), false);
	  adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
	  AduManager::am->InsertAdu(adu);
	  CommunicationManager::cm->SendAduReliable(adu);
	}

	buf   = new Buffer(Page::GetSetActiveSize());
	Page::FlushSetActive(buf->getCont());
	adu = CommunicationManager::cm->CreateEvent(newActive->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
      }
    }
  }
  else { //PRIVATE session
    GenericContainer* obj = (GenericContainer*)GetObject(idVector->back());
    if (obj) {
      if (obj->GetType() == GenericObject::SESSION) {
	SetActiveSession(obj->GetID());
	DeleteActiveSession();
      } else {

	if (GetActivePage()) {
	  GetActivePage()->DeActivate();
	  SetActivePage(NULL);
	}

	SetActiveFolder(NULL);

//  	if (obj->GetType() == GenericObject::PAGE) {
//  	  if (GetActivePage() == obj) {
//  	    ((Page*)obj)->DeActivate();
//  	    SetActivePage(NULL);
//  	  }
//  	}
	
	obj->DeleteAll(NULL);

	Folder* parent = (Folder*)GetParentContainer(obj->GetID());
	GenericContainer* next = parent->GetNextContainer(obj);

  if (parent->GetCurrentPage() == obj) parent->SetCurrentPage(NULL);
  
  if (GetParentSession(obj)->thumbnailFolder == obj) {
    GetParentSession(obj)->thumbnailFolder = NULL;
  }

  GetParentContainer(obj->GetID())->DeleteObject(obj->GetID());

	if (!parent->Empty() && next)
	  SelectContainer(next->GetNodePath());

	delete obj;
      }
    }

    ClearPageTree();
    DrawPageTree();
  }

  delete idVector;
}


void MLB::SessionManager::RemoteDeleteContainer(MLB::GenericContainer* obj) {
  obj->DeleteObjectsFromRTPI();

//    if (obj->GetType() == GenericObject::PAGE) {
//      if (GetActivePage() == obj) {
//        ((Page*)obj)->DeActivate();
//        SetActivePage(NULL);
//      }
//    }

  if (GetActivePage()) {
    GetActivePage()->DeActivate();
    SetActivePage(NULL);
  }
  SetActiveFolder(NULL);
  
  obj->DeleteAll(NULL);

  Folder* parent = (Folder*)GetParentContainer(obj->GetID());
  if (!parent) return;
  GenericContainer* next = parent->GetNextContainer(obj);

  ((GenericContainer*)GetSharedSession())->DeleteObject(obj->GetID());

  if (!parent->Empty() && next)
    RemoteSelectContainer(next->GetNodePath());

  delete obj;

  ClearPageTree();
  DrawPageTree();
}


void MLB::SessionManager::SaveContainer(std::string nodePath) {
  std::vector<uint32>* idVector;
  GenericContainer* obj;
  FILE* datei;
  std::string filename;
  char* zw;

  if (!GetActiveSession()) return;

  idVector = CreateIdVector(nodePath);
  if (!idVector) return;

  obj = (GenericContainer *)GetObject(idVector->back());
  if (obj) {
    zw = GetFilename(GetActiveSession()->GetDocumentName());
    filename = GetPraefix(zw);
    delete[] zw;
    filename += nodePath ;
    filename += ".mlb";

    datei = fopen(filename.c_str(),"wb");
    if (datei) {
      fprintf(datei,"<MLB name=\"Session\">\n");
      obj->SaveContainer(datei,1);
      fprintf(datei,"</MLB>\n");
      fclose(datei);
    }
    else {
      GUI::MessageWindow::Error("Can't open %s\n",filename.c_str());
      filename.erase();
      return;
    }
    filename.erase();
    return;
  }

  delete idVector;
}


void MLB::SessionManager::SelectContainer(std::string nodePath) {
  std::vector<uint32>* idVector = CreateIdVector(nodePath);
  if (!idVector) return;

  if (!SynchMode()) {
    if (GetActivePage()) GetActivePage()->DeActivate();

    SetActiveSession((*idVector)[1]);
    SetActivePage(NULL);
    SetActiveFolder(NULL);

    (*activeSession)->SetCurrentContainer(idVector,2);
    DrawPageTree();
    delete idVector;
    return;
  }

  Session* targetSession = (Session*)GetObject((*idVector)[1]);

  if (targetSession->GetSessionType() == Session::SHARED) {
    GenericContainer* newActive = (GenericContainer*)GetObject(idVector->back());

    if ((newActive->GetType() == GenericObject::FOLDER) || (newActive->GetType() == GenericObject::SESSION)) {
      SetActiveSession((*idVector)[1]);
      (*activeSession)->SetCurrentContainer(idVector,2);
      DrawPageTree();
    } else if (newActive->GetType() == GenericObject::PAGE) {
      RefCount<Buffer> buf;
      RefCount<RTPI::RtpiAdu> adu;
      if (GetActivePage() && (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED)) {
	buf   = new Buffer(Page::GetSetActiveSize());
	Page::FlushSetActive(buf->getCont(), false);
	adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
	AduManager::am->InsertAdu(adu);
	CommunicationManager::cm->SendAduReliable(adu);
      }

      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont());
      adu = CommunicationManager::cm->CreateEvent(newActive->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }

  }
  else { // PRIVATE Session
    if (GetActivePage() && (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED)) {
      CommunicationManager::cm->DeactivateAllSubcomponents();
    }
    SetActiveSession((*idVector)[1]);
    (*activeSession)->SetCurrentContainer(idVector,2);
    DrawPageTree();
  }

  delete idVector;
}


void MLB::SessionManager::RemoteSelectContainer(std::string nodePath) {
  std::vector<uint32>* idVector = CreateIdVector(nodePath);
  if (!idVector) return;

  GenericObject* tmp = GetObject(idVector->back());
  if (tmp && (tmp->GetType() == GenericObject::PAGE) && (GetActivePage() == tmp)) return; // page already active

  CommunicationManager::cm->DeactivateAllSubcomponents();
  if (GetActivePage()) GetActivePage()->DeActivate();

  if (GetActiveFolder() && (GetActiveFolder()->GetType() == GenericObject::FOLDER)
      && (GetParentSession(GetActiveFolder())->GetSessionType() == Session::SHARED))
    CommunicationManager::cm->ChangeSubcomponentActiveness(GetActiveFolder()->GetID(), false);

  SetActiveSession((*idVector)[1]);
  SetActiveFolder(NULL);
  SetActivePage(NULL);

  (*activeSession)->SetCurrentContainer(idVector,2);

  if (GetActivePage()) {
    GetActivePage()->ChangeSubcomponentsActiveness(true);
    if (LogPageChanges()) WritePageLog();
  }

  if (GetActiveFolder() && (GetActiveFolder()->GetType() == GenericObject::FOLDER))
    CommunicationManager::cm->ChangeSubcomponentActiveness(GetActiveFolder()->GetID(), true);


  GUI::PageView::ActivateAnnotationButton(true);
}


void MLB::SessionManager::ActivateFirstPage() {
  if (!GetActivePage()) return;

  Session* parent = GetParentSession(GetActivePage());
  if (!parent) return;
  
  GenericContainer* page = parent->GetFirstPage();
  if (!page) return; // not found

  if ((parent->GetSessionType() == Session::SHARED) && SynchMode()) {
    RefCount<Buffer> buf; 
    RefCount<RTPI::RtpiAdu> adu;

    if (GetActivePage() && (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED)) {
      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont(), false);
      adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }
    
    buf   = new Buffer(Page::GetSetActiveSize());
    Page::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(page->GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

  } else {
    SelectContainer(page->GetNodePath());
    ClearPageTree();
    DrawPageTree();
  }
}


void MLB::SessionManager::ActivatePreviousPage() {
  if (!GetActiveSession() || !GetActivePage()) return;

  bool found = false;
  Page* prevPage = GetActiveSession()->GetPrevPage(activePage,&found);

  if (!prevPage) return; // no next page

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && SynchMode()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;

    if (GetActivePage() && (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED)) {
      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont(), false);
      adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }

    buf   = new Buffer(Page::GetSetActiveSize());
    Page::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(prevPage->GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

  } else { //  PRIVATE Session
    prevPage->Activate();
    DrawPageTree();
  }

  return;
}


void MLB::SessionManager::ActivateNextPage() {
  if (!GetActiveSession() || !GetActivePage()) return;

  bool found = false;
  Page* nextPage = GetActiveSession()->GetNextPage(activePage,&found);

  if (!nextPage) return; // no next page

  if ((GetActiveSession()->GetSessionType() == Session::SHARED) && SynchMode()) {
    RefCount<Buffer> buf;
    RefCount<RTPI::RtpiAdu> adu;

    if (GetActivePage() && (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED)) {
      buf   = new Buffer(Page::GetSetActiveSize());
      Page::FlushSetActive(buf->getCont(), false);
      adu = CommunicationManager::cm->CreateEvent(GetActivePage()->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);
    }

    buf   = new Buffer(Page::GetSetActiveSize());
    Page::FlushSetActive(buf->getCont());
    adu = CommunicationManager::cm->CreateEvent(nextPage->GetID(), buf);
    AduManager::am->InsertAdu(adu);
    CommunicationManager::cm->SendAduReliable(adu);

  } else { //  PRIVATE Session
    nextPage->Activate();
    DrawPageTree();
  }

  return;
}


void MLB::SessionManager::ImportPage() {
  GenericContainer* obj = NULL;

  Session* shared = GetSharedSession();
  if (!shared) return; // no shared session open

  // find first private session with a page
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    if (*i == shared) continue; // ignore active session
    obj = (*i)->GetFirstPage();
    if (obj) {
      Session* session = SessionManager::sm->GetParentSession(obj);
      bool thumbnailPage = session->thumbnailFolder && SessionManager::sm->IsParentOf(session->thumbnailFolder->GetNodePath(), ((Page*)obj)->GetNodePath());
      if (thumbnailPage) continue;
      ChangeHierachy(obj->GetID(), shared->GetID(), "end");
      return;
    }
  }
}


void MLB::SessionManager::ChangeHierachy(uint32 objID, uint32 targetID, const char* position) {
  GenericContainer* obj = (GenericContainer*)GetObject(objID);
  if (!obj) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find object with id %u\n", objID);
    return;
  }

  if (obj->GetType() == GenericObject::SESSION)
    return; // not possible

  Session* sourceSession = GetParentSession(obj);
  if (!sourceSession) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find parent session of %u\n", obj->GetID());
    return;
  }

  GenericContainer* source = sourceSession->GetContainer(obj->GetID());
  if (!source) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find parent container of %u\n", obj->GetID());
    return;
  }

  GenericContainer* target = (GenericContainer*)GetObject(targetID);
  if (!target) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find target container %u\n", targetID);
    return;
  }

  Session* targetSession = GetParentSession(target);
  if (!targetSession) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find target session of %u\n", targetID);
    return;
  }


  if (obj == target)
    return; // not possible

  if (IsParentOf(obj->GetNodePath(), target->GetNodePath()))
    return; // not possible

  if (sourceSession->thumbnailFolder)
    if (IsParentOf(sourceSession->thumbnailFolder->GetNodePath(), obj->GetNodePath()))
      return; // not possible

  if (targetSession->thumbnailFolder)
    if (IsParentOf(targetSession->thumbnailFolder->GetNodePath(), target->GetNodePath()))
      return; // not possible
  
  if (target->GetType() == GenericObject::PAGE)
    return; // not possible

  uint32 order;
  int pos;

  //
  // obj is moved in scope of the same session
  //
  if (targetSession == sourceSession) {

    if (targetSession->GetSessionType() == Session::SHARED) {
      if (strcmp(position, "end") == 0) {
	if (target->Empty()) order = MLB_OO_MIDDLE;
	else order = target->GetEndObjectOrder() + MLB_OO_OFFSET;
      } else {
	pos = atoi(position);
	if (pos >= target->GetSize()) pos = target->GetSize();
	order = target->GetNewObjectOrder(pos);
      }

      #ifdef DEBUG
      GUI::MessageWindow::Info("Session::ChangeHierachy - move obj %u from folder %u to folder %u position %u\n",
		     obj->GetID(), source->GetID(), target->GetID(), pos);
      #endif

      RefCount<Buffer> buf = new Buffer(Page::GetChangeParentObjectSize());
      if (obj->GetType() == GenericObject::PAGE) Page::FlushChangeParentObject(buf->getCont(), order);
      else Folder::FlushChangeParentObject(buf->getCont(), order);
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(obj->GetID(), target->GetID(), buf);
      AduManager::am->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);

    } else {
      source->DeleteObject(obj->GetID());
      ((Folder*)source)->SetCurrentPage(NULL);

      if (strcmp(position, "end") == 0) {
	target->InsertContainerEnd(obj);
	pos = target->GetSize() - 1;
      } else {
	pos = atoi(position);
	if (pos >= target->GetSize()) pos = target->GetSize();
	obj->SetOrder(target->GetNewObjectOrder(pos));
	target->InsertContainer(obj);
      }

      #ifdef DEBUG
      GUI::MessageWindow::Info("Session::ChangeHierachy - move obj %u from folder %u to folder %u position %u\n",
		     obj->GetID(), source->GetID(), target->GetID(), pos);
      #endif

      ClearPageTree();
      DrawPageTree();

      if (GetActivePage()) GetActivePage()->DeActivate();

      if (obj->GetType() == GenericObject::FOLDER) {
	Page* fpage = (Page*)(((Folder*)obj)->GetFirstPage());
	if (fpage) {
	  SelectContainer(fpage->GetNodePath());
	} else {
	  SelectContainer(obj->GetNodePath());
	}
      } else { // PAGE
	SelectContainer(obj->GetNodePath());
      }
    }
    return;
  }

  //
  // obj is moved between sessions
  //

  //
  // 1. both sessions are private
  //
  if ((sourceSession->GetSessionType() == Session::PRIVATE) && (targetSession->GetSessionType() == Session::PRIVATE)) {
    // move obj from source session to target session
    source->DeleteObject(obj->GetID());
    ((Folder*)source)->SetCurrentPage(NULL);

    if (strcmp(position, "end") == 0) {
      target->InsertContainerEnd(obj);
      pos = target->GetSize() - 1;
    } else {
      pos = atoi(position);
      if (pos >= target->GetSize()) pos = target->GetSize();
      obj->SetOrder(target->GetNewObjectOrder(pos));
      target->InsertContainer(obj);
    }

    #ifdef DEBUG
    GUI::MessageWindow::Info("Session::ChangeHierachy - PRIV -> PRIV: move obj %u from folder %u to folder %u position %u\n",
		   obj->GetID(), source->GetID(), target->GetID(), pos);
    #endif

    ClearPageTree();
    DrawPageTree();

    if (GetActivePage()) GetActivePage()->DeActivate();

    if (obj->GetType() == GenericObject::FOLDER) {
      Page* fpage = (Page*)(((Folder*)obj)->GetFirstPage());
      if (fpage) {
	SelectContainer(fpage->GetNodePath());
      } else {
	SelectContainer(obj->GetNodePath());
      }
    } else { // PAGE
      SelectContainer(obj->GetNodePath());
    }
    return;
  }


  //
  // 2. source session is shared and target session is private
  //
  if ((sourceSession->GetSessionType() == Session::SHARED) && (targetSession->GetSessionType() == Session::PRIVATE)) {
    // 1. move all objects from source (shared) session to target (private) session
    source->DeleteObject(obj->GetID());
    ((Folder*)source)->SetCurrentPage(NULL);

    if (strcmp(position, "end") == 0) {
      target->InsertContainerEnd(obj);
      pos = target->GetSize() - 1;
    } else {
      pos = atoi(position);
      if (pos >= target->GetSize()) pos = target->GetSize();
      obj->SetOrder(target->GetNewObjectOrder(pos));
      target->InsertContainer(obj);
    }

    #ifdef DEBUG
    GUI::MessageWindow::Info("Session::ChangeHierachy - SHAR -> PRIV:  move obj %u from folder %u to folder %u position %u\n",
		   obj->GetID(), source->GetID(), target->GetID(), pos);
    #endif

    ClearPageTree();
    DrawPageTree();

    if (GetActivePage()) GetActivePage()->DeActivate();

    if (obj->GetType() == GenericObject::FOLDER) {
      Page* fpage = (Page*)(((Folder*)obj)->GetFirstPage());
      if (fpage) {
	SelectContainer(fpage->GetNodePath());
      } else {
	SelectContainer(obj->GetNodePath());
      }
    } else { // PAGE
      SelectContainer(obj->GetNodePath());
    }

    // 2. send delete for obj (obj is folder or page)
    RefCount<Buffer> buf = new Buffer(obj->GetDeleteSize());
    obj->FlushDelete(buf->getCont());
    RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateEvent(obj->GetID(), buf);
    CommunicationManager::cm->SendAduReliable(adu);
    TimeWarp::tw->InsertAdu(adu);

    // 3. remove all objects from RTP/I namespace
    obj->DeleteObjectsFromRTPI();

    return;
  }


  //
  // 3. source session is private and target session is shared
  //
  if ((sourceSession->GetSessionType() == Session::PRIVATE) && (targetSession->GetSessionType() == Session::SHARED)) {
    // 1. move all objects from source (private) session to target (shared) session
    source->DeleteObject(obj->GetID());
    ((Folder*)source)->SetCurrentPage(NULL);

    if (strcmp(position, "end") == 0) {
      target->InsertContainerEnd(obj);
      pos = target->GetSize() - 1;
    } else {
      pos = atoi(position);
      if (pos >= target->GetSize()) pos = target->GetSize();
      obj->SetOrder(target->GetNewObjectOrder(pos));
      target->InsertContainer(obj);
    }

    #ifdef DEBUG
    GUI::MessageWindow::Info("Session::ChangeHierachy - PRIV -> SHAR:  move obj %u from folder %u to folder %u position %u\n",
		   obj->GetID(), source->GetID(), target->GetID(), pos);
    #endif

    ClearPageTree();
    DrawPageTree();

    if (GetActivePage()) {
      if (GetParentSession(GetActivePage())->GetSessionType() == Session::SHARED) {
	CommunicationManager::cm->DeactivateAllSubcomponents();
      }
      GetActivePage()->DeActivate();
    }

    // 2. send state for obj
    if (obj->GetType() == GenericObject::FOLDER) {
      // send state for folder
      RefCount<Buffer> buf = new Buffer(((Folder*)obj)->GetStateSize());
      ((Folder*)obj)->FlushState(buf->getCont());
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(obj->GetID(), target->GetID(), CommunicationManager::FOLDER, 3, buf);
      TimeWarp::tw->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);

      ((Folder*)obj)->InitTelepointers();

      // send state for all pages and objects
      ((Folder*)obj)->SendAllObjects(rtpitime());

      Page* fpage = (Page*)(((Folder*)obj)->GetFirstPage());
      if (fpage) {
	SelectContainer(fpage->GetNodePath());
      } else {
	SelectContainer(obj->GetNodePath());
      }
    } else { // PAGE
      ((Page*)obj)->InitTelepointer();

      // UCE connection
      if (UceManager::connected) {
	PhotoImage* image = ((Page*)obj)->GetFirstImage();
	if (!image) return;
	UCE::UCE_Interface::ucei->cmd_load(obj->GetID(), obj->GetName(), obj->GetOrder(), image->GetBlobLength(), image->GetBlobData());
      }

      // send state for page
      RefCount<Buffer> buf = new Buffer(((Page*)obj)->GetStateSize());
      ((Page*)obj)->FlushState(buf->getCont());
      RefCount<RTPI::RtpiAdu> adu = CommunicationManager::cm->CreateState(obj->GetID(), target->GetID(), CommunicationManager::PAGE, 3, buf);
      TimeWarp::tw->InsertAdu(adu);
      CommunicationManager::cm->SendAduReliable(adu);

      // send state for all page objects
      ((Page*)obj)->SendAllObjects(rtpitime());

      SelectContainer(obj->GetNodePath());
    }
    return;
  }
}


void MLB::SessionManager::RemoteChangeHierachy(uint32 objID, uint32 targetID) {
  Session* parentSession = GetSharedSession();

  GenericContainer* obj = (GenericContainer*)parentSession->GetObject(objID);
  if (!obj) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find object with id %u\n", objID);
    return;
  }

  GenericContainer* source = parentSession->GetContainer(obj->GetID());
  if (!source) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find parent container of %u\n", obj->GetID());
    return;
  }

  GenericContainer* target = (GenericContainer*)parentSession->GetObject(targetID);
  if (!target) {
    GUI::MessageWindow::Error("Session::ChangeHierachy - did not find target container %u\n", targetID);
    return;
  }

  if (target->GetType() == GenericObject::PAGE)
    return; // not possible

  source->DeleteObject(obj->GetID());
  ((Folder*)source)->SetCurrentPage(NULL);

  target->InsertContainer(obj);

  ClearPageTree();
  DrawPageTree();

  if (obj->GetType() == GenericObject::FOLDER) {
    Page* fpage = (Page*)(((Folder*)obj)->GetFirstPage());
    if (fpage) {
      RemoteSelectContainer(fpage->GetNodePath());
    } else {
      RemoteSelectContainer(obj->GetNodePath());
    }
  } else { // PAGE
    RemoteSelectContainer(obj->GetNodePath());
  }
}


bool MLB::SessionManager::IsParentOf(std::string nodePath1, std::string nodePath2) {
//  #ifdef WIN32
//    if (nodePath2.compare(0, nodePath1.length(), nodePath1) == 0)
//  #else
//    if (nodePath2.compare(nodePath1, 0, nodePath1.length()) == 0)
//  #endif
  if (nodePath2.compare(0, nodePath1.length(), nodePath1) == 0)
    return true; // nodePath1 is part of nodePath2 -> node1 is parent of node2
  return false;
}


void MLB::SessionManager::DrawPageTree() {
  char buf[200];
  int count = 0;
  bool open;

  sprintf(buf,".%u",0);
  nodePath = buf;

  if (nodeName.empty()) {
    sprintf(buf, "%u", 0);
    nodeName = std::string("n:") + buf;
    GUI::PageView::Init(nodeName.c_str());
  }

  if (!sessions.empty()) {
    for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
      open = false;
      (*i)->DrawPageTree("",nodePath,nodeName,count,&open);
      count++;
    }
  }
  
  GUI::PageView::Redraw();
}


void MLB::SessionManager::ClearPageTree() {
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i)
    (*i)->ClearPageTree();

  GUI::PageView::Delete(nodeName.c_str());

  nodeName = "";
}


void MLB::SessionManager::ClearNodeNames() {
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i)
    (*i)->ClearNodeNames();

  nodeName = "";
}


MLB::GenericObject* MLB::SessionManager::GetObject(uint32 id) {
  GenericObject* obj = NULL;

  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    if ((*i)->GetID() == id) return (*i);
    obj = (*i)->GetObject(id);
    if (obj) return obj;
  }

  return NULL;
}


MLB::GenericObject* MLB::SessionManager::GetGroup(uint32 id) {
   GenericObject* obj = NULL;
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    obj = (*i)->GetGroup(id);
    if (obj) return obj;
  }

  return NULL;
}


// Find direct parent container of Object obj
MLB::GenericContainer* MLB::SessionManager::GetParentContainer(uint32 id) {
  GenericContainer* container = NULL;
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    container = (*i)->GetContainer(id);
    if (container) return container;
  }

  return NULL;
}


// Find parent session of Object obj
MLB::Session* MLB::SessionManager::GetParentSession(MLB::GenericObject* obj) {
  if (obj->GetType() == GenericObject::SESSION)
    return (Session*)obj;  // obj is session itself

  GenericContainer* container = NULL;
  for (std::list<Session*>::iterator i = sessions.begin(); i != sessions.end(); ++i) {
    container = (*i)->GetContainer(obj->GetID());
    if (container) return *i;
  }

  return NULL;
}


void MLB::SessionManager::SetAnnotationMode(bool annotate) {
  annotationMode = annotate;
  if (activePage && GetParentSession(activePage)->GetSessionType() == Session::SHARED)
    activePage->Draw();
}


void MLB::SessionManager::SetSynchMode(bool synch) {
  synchMode = synch;

  // not synchronized
  if (!synchMode) {
    globalPage = (activePage) ? activePage->GetID() : 0; // remember active page
    return;
  }

  // snychronized
  if (globalPage != 0) {
    Page* gPage = (Page*)GetObject(globalPage);
    if (!gPage || (gPage == activePage)) return;
    activePage->DeActivate();
    RemoteSelectContainer(gPage->GetNodePath());
    ClearPageTree();
    DrawPageTree();
  }
}


void MLB::SessionManager::WritePageLog() {
  static char* logfile = NULL;

  if (logfile == NULL) {
    time_t t;
    time(&t);
    struct tm* today = localtime(&t);
    char time[11];
    strftime(time, sizeof(time), "%d.%m.%Y", today);
    logfile = new char[200];
    sprintf(logfile, "pages_%s_%s_%u.log", CommunicationManager::cm->GetSessionIP().c_str(), time, rtpitime());
  }

  FILE* log = fopen(logfile, "a");
  if (!log) {
    GUI::MessageWindow::Error("SessionManager::WritePageLog - Cannot open pages.log\n");
    return;
  }
  
  if (!GetActivePage()) return;

  fprintf(log, "%u %u\n", GetActivePage()->GetID(), rtpitime());
  fclose(log);
}


void MLB::SessionManager::HandleRecording(int argc, const char* argv[])
{
  int
    status;

  status = atoi(argv[1]);

  if (status == Record)
   StartRecording();
  else if (status == Pause)
   PauseRecording();
  else if (status == Stop)
    StopRecording();
}


int MLB::SessionManager::StartRecording()
{
  char
    *zw,
    *recordName;

  if (!GetActiveSession()) {
    GUI::Record::Stop();
    GUI::MessageWindow::Error("No Session\n");
    return 0;
  }

  if (!recording) {
    zw = GetPraefix(GetActiveSession()->GetDocumentName());
    recordName = GetFilename(zw);

//	useJedas=false;
//	useAppPlay=false;

    aofRecording = new Record::AOF(recordName,frameRate);

    delete[] zw;
    delete[] recordName;

    evqFp = aofRecording->GetEvqFp();
    objFp = aofRecording->GetObjFp();

    recCount = 0;
    recording = true;
    recordedPages.clear();
    aofRecording->Start();
  }
  else {
    if (paused) {
      aofRecording->Continue();
      paused = false;
    }
  }
  if (SessionManager::sm->GetActivePage()) {
    SessionManager::sm->GetActivePage()->SaveAsAOF();
  }

  return 1;
}

int MLB::SessionManager::PauseRecording()
{
  if (!GetActiveSession()) {
    GUI::Record::Stop();
    GUI::MessageWindow::Error("No Session\n");
    return 0;
  }

  paused = true;
  aofRecording->Pause();

  return 1;
}

char *MLB::SessionManager::GetDataPath()
{
  if (aofRecording)
    return aofRecording->GetDataPath();
  else
    return "./";
}

void MLB::SessionManager::PrintHtmlPage()
{
  std::list<GenericContainer*>::iterator
    temp;
  FILE
    *indexFp;
  std::string
    indexName;
  Page
    *page,
    *prev,
    *next,
    *first,
    *last;

  cursor->Set(GUI::Cursor::WATCH);
  progress->ShowProgress("Create Html Overview",recordedPages.size());

  page = activePage;

#ifdef WIN32
  if (_mkdir (aofRecording->GetMiscPath()) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Can't create directory %s\n",aofRecording->GetMiscPath());
      return;
    }
  }
  if (_mkdir (aofRecording->GetApPath()) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Can't create directory %s\n",aofRecording->GetApPath());
      return;
    }
  }
#else
  if (mkdir (aofRecording->GetMiscPath(),S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Can't create directory %s\n",aofRecording->GetMiscPath());
      return;
    }
  }
  if (mkdir (aofRecording->GetApPath(),S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH) == -1) {
    if (errno != EEXIST) {
      GUI::MessageWindow::Error("Can't create directory %s\n",aofRecording->GetApPath());
      return;
    }
  }
#endif

  indexName = aofRecording->GetMiscPath();
  indexName += "indexframe.tmpl";

  indexFp = fopen((char *)indexName.c_str(),"wb");
  if (!indexFp) {
    GUI::MessageWindow::Error ("Cant't open %s\n",indexName.c_str());
    return;
  }

  fprintf(indexFp,"<html>\n");
  fprintf(indexFp,"  <head>\n");
  fprintf(indexFp,"    <title>Vortrag: &Uuml;bersichtsseite</title>\n");
  fprintf(indexFp,"    <base target=\"viewframe\">\n");
  fprintf(indexFp,"  </head>\n");
  fprintf(indexFp,"\n");
  fprintf(indexFp,"  <body bgcolor=\"#c0c0c0\">\n");
  fprintf(indexFp,"    <table cellpadding=2 cellspacing=1 border=0>\n");
  fprintf(indexFp,"\n");
  fprintf(indexFp,"      <tr><td>\n");
  fprintf(indexFp,"\n");
  fprintf(indexFp,"        <table cellpadding=1 cellspacing=2 border=0>\n");
  fprintf(indexFp,"          <tr>\n");
  fprintf(indexFp,"            <td valign=top><font face=\"Helvetica,Arial\"><strong>Titel:</strong></td>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\">TITEL</td>\n");
  fprintf(indexFp,"          </tr>\n");
  fprintf(indexFp,"          <tr>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\"><strong>L&auml;nge:</strong></td>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\">%d:%2.2d</td>\n",0,0);

  fprintf(indexFp,"          </tr>\n");
  fprintf(indexFp,"          <tr>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\"><strong>Autor:</strong></td>\n");
  fprintf(indexFp,"            <td><font face=\"Helvetica,Arial\">%s</td>\n","Unknown");
  fprintf(indexFp,"          </tr>\n");
  fprintf(indexFp,"        </table>\n\n");
  fprintf(indexFp,"      </td></tr>\n\n");
  fprintf(indexFp,"\n");

  if (!recordedPages.empty()) {
    first = (Page *)(recordedPages.front());
    last  = (Page *)(recordedPages.back());

    prev = NULL;
    for (temp = recordedPages.begin(); temp != recordedPages.end(); ++temp) {
      progress->IncrementProgress();

      ((Page *)(*temp))->Activate();
      DrawPageTree();
      temp++;
      if (temp != recordedPages.end())
	next = (Page *)(*temp);
      else
	next = NULL;
      temp--;

      activePage->WriteHtml(aofRecording,first,prev,next,last);

      fprintf(indexFp,"\n");
      fprintf(indexFp,"      <tr><td><hr></td></tr>\n");
      fprintf(indexFp,"\n");
      fprintf(indexFp,"      <tr>\n");
      fprintf(indexFp,"        <td align=center><strong><font face=\"Helvetica,Arial\"><a name=\"%d\">%s</font></strong><p>\n",activePage->GetPageNumber(),activePage->GetName());
      fprintf(indexFp,"        <a href=\"slide_%d.htm\">\n",activePage->GetPageNumber());
      fprintf(indexFp,"        <img width=%d height=%d border=1 src=\"%s_%d.png\" alt=\"Display slide: %s_%d.png\"></a><p>\n",activePage->GetWidth()/4,activePage->GetHeight()/4,aofRecording->GetPrefix(),activePage->GetPageNumber(),aofRecording->GetPrefix(),activePage->GetPageNumber());
      fprintf(indexFp,"        <a href=\"../ap/%s_%d.AOFPH\"><img border=1 src=\"aof_s.gif\" alt=\"Start replay at/jump to slide transition\"></a>\n",aofRecording->GetPrefix(),activePage->GetPageNumber());
      fprintf(indexFp,"\n");
      fprintf(indexFp,"        </td>\n");
      fprintf(indexFp,"      </tr>\n");


      prev = (Page *)(*temp);
    }
  }


  fprintf(indexFp,"    </table>\n");
  fprintf(indexFp,"  </body>\n");
  fprintf(indexFp,"</html>\n");

  fclose (indexFp);
  page->Activate();
  DrawPageTree();

  progress->HideProgress();
  cursor->Reset();
}

int MLB::SessionManager::StopRecording()
{

  if (!GetActiveSession()) {
    return 0;
  }

  if (recording) {
    recording = false;
    aofRecording->Stop();
    PrintHtmlPage();
    ResetRecording();
    delete aofRecording;
    aofRecording = NULL;
    return 1;
  }

  return 0;
}

void MLB::SessionManager::ResetRecording()
{
  std::list<Session *>::iterator
    temp;

  if (!sessions.empty()) {
    temp = sessions.begin();
    while (temp!=sessions.end()) {
      (*temp)->ResetRecording();
      ++temp;
    }
  }
  recCount = 0;
  recordedPages.clear();
}

void MLB::SessionManager::AddPageAsRecorded(MLB::Page *page, int time)
{
  std::list<GenericContainer *>::iterator
    temp;

  temp = recordedPages.begin();
  while (temp != recordedPages.end()) {
    if ((*temp) == page) return;
    temp++;
  }

  recordedPages.push_back(page);
}


void MLB::SessionManager::ChangeFrameRate(int argc, const char* argv[])
{
  int
    rate;

  if (recording)
    GUI::MessageWindow::Error("changing has no effect to current recording\n");


  rate = atoi(argv[1]);
  if ((rate != 8000) && (rate != 11025) &&(rate != 16000) &&
      (rate != 22050))
    frameRate = 16000;
  else
    frameRate = rate;
}

void MLB::SessionManager::SetZoom(int _zoom) {
  zoom = _zoom/100.0;
  if (GetActivePage()) GetActivePage()->ReDraw();
}

//-------------------------------------------------------------------------//

/**   Changes animation edit/presentation mode. If needed shows or hides
  *   all the objects on all session's pages.
  *   @param mode   edit mode: true if edit, false if presentation
  **/
void MLB::SessionManager::SetAnimationEditMode( bool mode )
{
  // nothing to do
  if( animationEditMode==mode ) return;

  animationEditMode = mode;
  for( std::list<Session*>::iterator i=sessions.begin(); i!=sessions.end(); ++i )
    (*i)->SetAnimationEditMode( mode );

} // SetAnimationEditMode()

//=========================================================================//
