/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_session.h
 authors : Juergen Vogel, Gabriela Maass
 date    : 18.02.2000
 desc    :

***********************************************************************************************************/

#ifndef __MLB_SESSION__
#define __MLB_SESSION__

#ifdef WIN32
#  pragma warning(once: 4786)
#endif


#include "mlb_folder.h"
#include "mlb_page.h"
#include "mlb_font.h"
#include "mlb_gui.h"
#include "mlb_record.h"

namespace MLB {

  int CalculateColor(Color color);

  class Session : public Folder {
  protected:
    int sessionType;
    int sessionStatus;
    std::string documentName;
    std::string documentPath;
    //bool animationEditMode;
  public:
    Folder* thumbnailFolder;
    std::map<uint32,uint32> thumbnailRefs;
    std::map<uint32,uint32> thumbnailIDs;

  public:
    enum types {
      // type of session
      PRIVATE  = 0,
      SHARED   = 1,

      // status of session
      ACTIVE   = 1,
      INACTIVE = 2
    };


    Session(uint32 id, int _type, int _status);
    ~Session();

    void SetSessionType(int _type) {sessionType = _type;}
    int  GetSessionType() {return sessionType;}

    void SetSessionStatus(int _status) {sessionStatus = _status;}
    int  GetSessionStatus() {return sessionStatus;}

    /*** Animation functions ***/
    //inline bool GetAnimationEditMode() { return animationEditMode; }
    //void        SetAnimationEditMode( bool mode );

    virtual int  GetType() {return GenericObject::SESSION;}
    virtual void SetName(char* _name);
    char* GetDocumentName() {return (char*)documentName.c_str();}
    char* GetDocumentPath() {return (char*)documentPath.c_str();}

    virtual void DrawPageTree(std::string _nodeNumber, std::string _nodePath, std::string _parentNode, int _pos, bool *_open);
    virtual void SaveContainer(FILE *_fp, int _level=0);

    void LoadSession(char *_filename);
    void InitPs(FILE *_fp);
    void EndPs(FILE *_fp);
    void DrawPageTree();
    void SaveDocumentAsPS(const char* _psName);
    void SaveDocumentAsHtml(const char* name);
    void SaveDocumentAsImage(const char* name, const char* type);

    int Save(int argc, const char* argv[]);
    int Open(int argc, const char* argv[]);
    int Open(char *filename);
    void ChangeContainerName(string nodePath, char* name);

    void ChangeLineColor(int argc, const char* argv[]);
    void ChangeFillColor(int argc, const char* argv[]);
    void ChangeLineWidth(int argc, const char* argv[]);
    void ChangeLineStyle(int argc, const char* argv[]);
    void ChangeCapStyle(int argc, const char* argv[]);
    void ChangeJoinStyle(int argc, const char* argv[]);
    void ChangeArrowStyle(int argc, const char* argv[]);
    void ChangeArrowShape(int argc, const char* argv[]);
    void ChangeFont(int argc, const char* argv[]);
    void ChangeObject(int argc, const char* argv[], int _type);
    void NewTextLine(int argc, const char* argv[]);
    void EndText(int argc, const char* argv[]);
    bool LoadImage(int argc, const char* argv[]);
    void PlaceImage(int argc, const char* argv[]);

    void AppendNewObject(int argc, const char* argv[], int _type);
    void MoveObjects(int argc, const char* argv[]);
    void CopyObjects(int argc, const char* argv[]);
    void RaiseObjects();
    void LowerObjects();
    void DeleteObject(int argc, const char* argv[]);
    void SetPageDimension(int argc, const char* argv[]);
    void SetPageColor(int argc, const char* argv[]);
    void ActivateOnlinePointer(int argc, const char* argv[]);
    void ActivateOnlineBox(int argc, const char* argv[]);
    void SetWindowPointer(int argc, const char* argv[]);
    void UnSetWindowPointer(int argc, const char* argv[]);
    void SelectObjects(int argc, const char* argv[]);
    void UnSelectObjects(int argc, const char* argv[]);
    void Draw();
    void DrawSelected();
    void ShowTicks(int argc, const char* argv[]);

    void CutObjects();
    void CopyObjects();
    void PasteObjects();

    void Change(int argc, const char* argv[]);

    void CreateThumbnails(int argc, const char* argv[]);
    void SelectThumbnail(int argc, const char* argv[]);
  };


  class SessionManager {
  public:
    SessionManager(const char* path);
    ~SessionManager();

    void PrintActiveSubcomponents();

    void        SetAnimationEditMode( bool mode );
    inline bool GetAnimationEditMode() { return animationEditMode; }

    bool ImportDocument(const char* filename);

    Session* GetSharedSession(); // only one shared session!
    void CloseSharedSession();

    void     AddSession(Session* session);
    Session* GetSession(uint32 id);
    Session* RemoveSession(uint32 id);
    void     DeleteAllSessions();

    void     SetActiveSession(uint32 id);
    Session* GetActiveSession();
    void     DeleteActiveSession();

    void   SetActivePage(Page *_page) {activePage = _page; Et_GlobalEvalF( Et_Interp, "Animations::update" ); }
    Page   *GetActivePage() {return activePage;}

    void   SetActiveFolder(Folder *_folder) {activeFolder = _folder;}
    Folder *GetActiveFolder() {return activeFolder;}

    void   SetGlobalPage(uint32 id) {globalPage = id;}
    Page*  GetGlobalPage() {return (Page*)GetObject(globalPage);}
    void   SetGlobalFolder(uint32 id) {globalFolder = id;}

    GenericObject*    GetObject(uint32 id);
    GenericObject*    GetGroup(uint32 id);
    GenericContainer* GetParentContainer(uint32 id);
    Session*          GetParentSession(GenericObject* obj);

    void InsertNewFolder(int argc, const char* argv[]);
    void InsertNewPage(int argc, const char* argv[]);
    void InsertNewPage(Page* page);
    void DuplicatePage();

    void SaveContainer(string _nodePath);

    void SelectContainer(string _nodePath);
    void RemoteSelectContainer(string _nodePath);

    void ActivateFirstPage();
    void ActivatePreviousPage();
    void ActivateNextPage();
    void ImportPage();

    void DeletePages(int from, int nbr);

    void DeleteContainer(string _nodePath);
    void RemoteDeleteContainer(GenericContainer* obj);

    void ChangeHierachy(uint32 container, uint32 target, const char* position);
    void RemoteChangeHierachy(uint32 container, uint32 target);
    bool IsParentOf(std::string node1, std::string node2);

    void DrawPageTree();
    void ClearPageTree();
    void ClearNodeNames();

    void SetAnnotationMode(bool annotate);
    bool AnnotationMode() {return annotationMode;}

    void SetSynchMode(bool synch);
    bool SynchMode() {return synchMode;}

    void FreezeImages(bool freeze) {freezeImages = freeze;}
    bool FreezeImages() {return freezeImages;}

    void SetLogPageChanges(bool log) {logPageChanges = log;}
    bool LogPageChanges() {return logPageChanges;}
    void WritePageLog();

    void   SetZoom(int _zoom);
    double GetZoom() {return zoom;}

    FILE* GetObjFp() {return objFp;}
    FILE* GetEvqFp() {return evqFp;}

    void HandleRecording(int argc, const char* argv[]);
    int StartRecording();
    void PrintHtmlPage();
    int StopRecording();
    int PauseRecording();
    char *GetDataPath();
    virtual void ResetRecording();
    void AddPageAsRecorded(Page *page, int time);
    void ChangeFrameRate(int argc, const char* argv[]);
    void SetAudioMute (int i) {mute=i;}
    int GetAudioMute () {return mute;}

    void UseJedas() {useJedas=true;}
    void UseAppPlay() {useAppPlay=true;}
    
    bool GetUseJedas() {return useJedas;}
    bool GetUseAppPlay() {return useAppPlay;}
    
    static SessionManager* sm;
    int pageCount;

    std::list<DrawObject*> pasteBuffer;

  private:
    friend class Session;

    enum type {
      Record = 0,
      Pause  = 1,
      Stop   = 2
    };

    std::vector<uint32>* CreateIdVector(string nodePath);

    double zoom;
    bool freezeImages;
    bool animationEditMode;

    std::list<Session*>::iterator activeSession;
    std::list<Session*> sessions;
    std::string nodeName;
    std::string nodePath;
    Page* activePage;
    Folder* activeFolder;
    uint32 globalPage;
    uint32 globalFolder;
    bool annotationMode;
    bool  synchMode;

    bool logPageChanges;

    std::list<GenericContainer*> recordedPages;
    int frameRate;
    FILE* objFp;
    FILE* evqFp;

    bool useJedas;
    bool useAppPlay;
    int mute;

    std::string programPath;

  public:
    GUI::Progress* progress;
    GUI::Cursor* cursor;
    GUI::MessageWindow* message;
    int recCount;
    Record::AOF* aofRecording;
    bool recording;
    bool paused;
    FontList* fontList;
  };
}

#endif
