/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : mlb_telepointer.h
 authors : Juergen Vogel, Gabriela Maass
 date    : 18.02.2000
 desc    :

***********************************************************************************************************/

#ifndef __MLB_TELEPOINTER_H__
#define __MLB_TELEPOINTER_H__


#include <rtpi/rtpi.h>
#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpihandler.h>
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpiInterfaces.h>
#include <rtpi/rtpiTransport.h>
#include <rtpi/rtpiSeqNumber.h>
#include <util/types.h>
#include <util/Buffer.h>

#include <mlb_generic.h>
#include <mlb_objects.h>

namespace MLB {
  class  Telepointer: public ColoredObject {
  public:
    enum types {
      // display
      ACTIVE   = 0,
      INACTIVE = 1
    };

    Telepointer(uint32 id, uint32 participant);
    Telepointer(int _x, int _y, char *_color, int _status, uint32 _id, uint32 participant);
    ~Telepointer();

    void CreateImages();

    uint32 GetParticipantID() {return participant;}
    virtual int GetType() {return GenericObject::TELEPOINTER;}

    virtual DrawObject *Copy() {return NULL;}

    virtual void Draw(Page* page);
    virtual void ReDraw(Page* page);
    virtual void Hide(Page* page);
    void Remove(Page* page);

    virtual bool IsEditablePoint(int _x, int _y) {return false;}

    void Update(int _x, int _y);
    void Update(char *_color);
    void Set(const char *_name, int _xHot, int _yHot);
    void SetActive(Page* page);
    void SetInactive(Page* page);
    bool IsActive();

    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false) {}
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects() {return 0;}
    virtual void AdjustGui() {};

    // network interface
    static void FlushType(char* buf, std::string type);
    static std::string ParseType(char* buf);

    int FlushState(char* buf);
    int ParseState(char* buf);

    int FlushChangeVisibility(char* buf);
    int FlushChangeVisibility(char* buf, bool v);
    int FlushMove(char* buf);
    int FlushChangeType(char* buf);
    int FlushChangeLineColor(char* buf);

    int FlushDelete(char* buf);

  private:
    uint32 participant;  // owner of the telepointer
    Pixmap bitmap;
    char   *name;
    int    xHot, yHot;
    int    status;
  };


  class  Telebox: public ColoredObject {
  public:
    enum types {
      // display
      ACTIVE   = 0,
      INACTIVE = 1
    };

    Telebox(uint32 id, uint32 participant);
    Telebox(int _x, int _y, int _width, int _height, char *_color, int _lineWidth, char *_lineStyle, int _status, uint32 _id, uint32 participant);
    ~Telebox();

    uint32 GetParticipantID() {return participant;}
    virtual int GetType() {return GenericObject::TELEBOX;}

    virtual DrawObject *Copy() {return NULL;}

    virtual void Draw(Page* page);
    virtual void ReDraw(Page* page);
    virtual void Hide(Page* page);
    void Remove(Page* page);

    virtual bool IsEditablePoint(int _x, int _y) {return false;}

    void Update(int _x, int _y, int _width, int _height);
    void Update(char *_color, int _lineWidth, char *_lineStyle);
    void SetActive(Page* page);
    void SetInactive(Page* page);
    bool IsActive();

    virtual void SaveAsPs(FILE *fp, int pageWidth, int PageHeight, bool rotate=false) {}
    virtual void SaveAsAOF(FILE *fp);
    virtual int RecordableObjects() {return 0;}
    virtual void AdjustGui() {};

    // network interface
    int FlushState(char* buf);
    int ParseState(char* buf);
    int FlushChangeVisibility(char* buf);
    int FlushChangeVisibility(char* buf, bool v);
    int FlushChangeSize(char* buf);
    int FlushChangeLinewidth(char* buf);
    int FlushChangeLinestyle(char* buf);
    int FlushChangeLineColor(char* buf);
    int FlushDelete(char* buf);

  private:
    uint32 participant;  // owner of the telebox
    int    status;
  };

    
  class TelepointerManager : public RTPI::RtcpiReportInterface, public RTPI::RtpiFeedbackInterface {
  public:
    enum types {
      // general
      VERSION = 0,
      APPLICATION_TYPE = 4, // telepointer

      // operation Types
      ADU_STATE  = 0,
      ADU_CHANGE = 1,
      ADU_DELETE = 2,

      // subtypes
      CHG_VISIBILITY = 0,
      MOVE           = 1,
      CHG_TYPE       = 2,
      CHG_LINECOLOR  = 3,
      CHG_SIZE       = 4,
      CHG_LINEWIDTH  = 5,
      CHG_LINESTYLE  = 6
    };

    TelepointerManager();
    ~TelepointerManager();

    bool OpenConnection() {return openConnection;}

    bool OpenRTPI();
    bool CloseRTPI();

    // handle incoming data
    void HandleReceivedRtpi(SMP::Indication ind);
    void HandleReceivedRtcpi(SMP::Indication ind);

    // outgoing data
    void SendState(uint32 subID, uint32 parentID, int priority, char* payload, int size);
    void SendEvent(uint32 subID, char* payload, int size);
    void SendDelete(uint32 subID, char* payload, int size); // special event
    void SendCue(uint32 subID, char* payload, int size);
    void SendStateQuery(uint32 subID, int priority);

    // functions implemented due to  RtcpiReportInterface
    void    send_rtcpi(RefCount<Buffer>);
    u_int32 get_local_address(int i=0)   {return MLB::CommunicationManager::cm->get_local_address(i);}
    int     addr_is_local(u_int32 addr)  {return MLB::CommunicationManager::cm->addr_is_local(addr);}
    u_int32 get_session_bandwith()       {return MLB::CommunicationManager::cm->bandwidth/10;}
    u_int32 get_new_pid()                {return MLB::CommunicationManager::cm->pid;}
    u_int32 get_new_subid()              {return MLB::CommunicationManager::cm->get_new_subid();}
    string  get_session_name(string ="") {return MLB::CommunicationManager::cm->name;}
    string  get_session_email()          {return MLB::CommunicationManager::cm->email;}
    string  get_session_phone()          {return MLB::CommunicationManager::cm->phone;}
    string  get_session_loc()            {return MLB::CommunicationManager::cm->location;}
    string  get_session_note()           {return ""; }
    string  get_session_tool()           {return "telepointer"; }

    // functions implemented due to RtpiFeedbackInterface
    void subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason);
    void subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason);
    void subcomponent_removed(uint32 id);
    void participant_added(uint32 pid);
    void participant_removed(uint32 pid);
    void participant_changed(uint32 pid);

    // helper functions
    void RegisterSubcomponent(uint32 subID);
    void UnregisterSubcomponent(uint32 subID);
    void ChangeSubcomponentActiveness(uint32 subID, bool active);

    static TelepointerManager* tm;
    static char* sendBuffer;

  private:
    // UDP
    bool openConnection;

    // SMP
    int rtpiSocket;
    int rtcpiSocket;

    // RTP/I
    bool init; // while init, don't answer state queries
    RTPI::Rtpi* rtpi;

  };
}


#endif
