/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************/

#include "mlb_undomanager.h"
#include "mlb_session.h"
#include "mlb_communication.h"


MLB::UndoManager* MLB::UndoManager::um = NULL;


void MLB::UndoManager::ClearList(RefCount<UndoAduList> aduList) {
  for (UndoAduList::iterator i = aduList->begin(); i != aduList->end(); ++i) {
    (*i)->clear();
  }
  aduList->clear();
}


void MLB::UndoManager::SetUndoState(bool active) {
  if (Et_GlobalEvalF(Et_Interp, "ToolBar::SetUndoButtonState %d", (int)active) != TCL_OK) {
    GUI::MessageWindow::Error("TclError in UndoManager::SetUndoState: %q\n",Et_Interp->result);
  }
}


void MLB::UndoManager::SetRedoState(bool active) {
  if (Et_GlobalEvalF(Et_Interp, "ToolBar::SetRedoButtonState %d", (int)active) != TCL_OK) {
    GUI::MessageWindow::Error("TclError in UndoManager::SetRedoState: %q\n",Et_Interp->result);
  }
}


void MLB::UndoManager::InsertAdu(RefCount<RTPI::RtpiAdu> adu) { 
  RefCount<RTPI::RtpiAduList> aduList = new RTPI::RtpiAduList(); 
  aduList->push_back(adu);
  InsertAduList(aduList); 
}   


void MLB::UndoManager::InsertAduList(RefCount<RTPI::RtpiAduList> aduList) { 
  if (!TimerManager::tm->TimerRunning(this, TimerObject::UNDO, 0)) {
    SetUndoState(true);
  }
  SetRedoState(false);
  undoList->push_back(aduList); 
  ClearList(redoList);
}   


void MLB::UndoManager::DeleteAdu(uint32 id) { 
  UndoAduList::iterator i;
  RTPI::RtpiAduList::iterator j;

  i = undoList->begin(); 
  while (i != undoList->end()) {
    j = (*i)->begin(); 
    while (j != (*i)->end()) {
      if ((*j)->GetSubcomponentID() == id) j = (*i)->erase(j); else ++j;
    }
    if ((*i)->empty()) i = undoList->erase(i); else ++i;
  }

  i = redoList->begin(); 
  while (i != redoList->end()) {
    j = (*i)->begin(); 
    while (j != (*i)->end()) {
      if ((*j)->GetSubcomponentID() == id) j = (*i)->erase(j); else ++j;
    }
    if ((*i)->empty()) i = redoList->erase(i); else ++i;
  }
  
  annotationSet->erase(id);
  
  if (undoList->empty()) SetUndoState(false);
  if (redoList->empty()) SetRedoState(false);
}   


void MLB::UndoManager::DeleteAdus(uint32 parentID) {
  std::set<uint32> deleteID;
  uint32 pID;
  UndoAduList::iterator i;
  RTPI::RtpiAduList::iterator j;

  i = undoList->begin(); 
  while (i != undoList->end()) {
    j = (*i)->begin(); 
    while (j != (*i)->end()) {
      switch ((*j)->GetType()) {
        case RTPI::RtpiHeader::PKT_EVENT: pID = ((RTPI::RtpiEvent*)&(**j))->GetObjectID(); break;
        case RTPI::RtpiHeader::PKT_STATE: pID = ((RTPI::RtpiState*)&(**j))->GetObjectID(); break;
      }
      if (pID == parentID) { 
        deleteID.insert((*j)->GetSubcomponentID());
        j = (*i)->erase(j);
      } else ++j;
    }
    if ((*i)->empty()) i = undoList->erase(i); else ++i;
  }
  
  i = redoList->begin(); 
  while (i != redoList->end()) {
    j = (*i)->begin(); 
    while (j != (*i)->end()) {
      switch ((*j)->GetType()) {
        case RTPI::RtpiHeader::PKT_EVENT: pID = ((RTPI::RtpiEvent*)&(**j))->GetObjectID(); break;
        case RTPI::RtpiHeader::PKT_STATE: pID = ((RTPI::RtpiState*)&(**j))->GetObjectID(); break;
      }
      if (pID == parentID) { 
        deleteID.insert((*j)->GetSubcomponentID());
        j = (*i)->erase(j);
      } else ++j;
    }
    if ((*i)->empty()) i = redoList->erase(i); else ++i;
  }
  
  for (AnnotationSet::iterator si = deleteID.begin(); si != deleteID.end(); ++si) {
    uint32 id = *si;
    annotationSet->erase(id);
  }
  deleteID.clear();
  
  if (undoList->empty()) SetUndoState(false);
  if (redoList->empty()) SetRedoState(false);
}


void MLB::UndoManager::ChangeAduID(RefCount<RTPI::RtpiAdu> adu, uint32 id, uint32 newID) { 
  if (adu->GetPacketHeader()->GetSubcomponentID() == id) {
    adu->GetPacketHeader()->SetSubcomponentID(newID); 
  }     
}


void MLB::UndoManager::ChangeAduID(uint32 id, uint32 newID) { 
  for (UndoAduList::iterator i = undoList->begin(); i != undoList->end(); ++i) {
    for (RTPI::RtpiAduList::iterator j = (*i)->begin(); j != (*i)->end(); ++j) {
      if ((*j)->GetSubcomponentID() == id) {
        (*j)->GetPacketHeader()->SetSubcomponentID(newID); 
      }
    }     
  }
  for (UndoAduList::iterator k = redoList->begin(); k != redoList->end(); ++k) {
    for (RTPI::RtpiAduList::iterator l = (*k)->begin(); l != (*k)->end(); ++l) {
      if ((*l)->GetSubcomponentID() == id) {
        (*l)->GetPacketHeader()->SetSubcomponentID(newID); 
      }
    }     
  }
}


RefCount<RTPI::RtpiAdu> MLB::UndoManager::CreateDelete(RefCount<RTPI::RtpiAdu> adu) {
  if (adu->GetType() != RTPI::RtpiHeader::PKT_STATE) return NULL;
  
  RTPI::RtpiState* state = (RTPI::RtpiState*)&(*adu);
  uint8*  p8   = (uint8*)(state->GetPayload());
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];
  uint32 objID = state->GetSubcomponentID();
  uint32 pID   = state->GetObjectID();

  RefCount<Buffer> buf = new Buffer(DrawObject::GetDeleteSize());  

  switch (objType) {
  //
  // Folder
  //
  case CommunicationManager::FOLDER: {
    //Folder::FlushDeleteS(buf->getCont());
    //return CommunicationManager::cm->CreateEvent(objID, pID, buf);
    return NULL;
  }
  //
  // Page
  //
  case CommunicationManager::PAGE: {
    //Page::FlushDeleteS(buf->getCont());
    //return CommunicationManager::cm->CreateEvent(objID, pID, buf);
    return NULL;
  }
  //
  // Group
  //
//  case CommunicationManager::GROUP: {}
  //
  // Rectangle
  //
  case CommunicationManager::RECTANGLE: {
    Rectangle::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  //
  // Oval
  //
  case CommunicationManager::OVAL: {
    Oval::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  //
  // Line
  //
  case CommunicationManager::LINE: {
    Line::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  //
  // PolyLine
  //
  case CommunicationManager::POLYLINE: {
    PolyLine::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }  
  //
  // Polygon
  //
  case CommunicationManager::POLYGON: {
    Polygon::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  //
  // Text
  //
  case CommunicationManager::TEXT: {
    Group::FlushDeleteTGS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  //
  // PhotoImage
  //
  case CommunicationManager::IMAGE: {
    PhotoImage::FlushDeleteS(buf->getCont());
    return CommunicationManager::cm->CreateEvent(objID, pID, buf);
  }
  }
  
  // error
  return NULL;
}


/*
 * MLB::UndoManager::CreateStateInverse
 */
RefCount<RTPI::RtpiAdu> MLB::UndoManager::CreateStateInverse(RefCount<RTPI::RtpiAdu> adu) {
  if (adu->GetType() != RTPI::RtpiHeader::PKT_STATE) return NULL;

  RTPI::RtpiState* state = (RTPI::RtpiState*)&(*adu);
  uint8*  p8   = (uint8*)(state->GetPayload());
  int  version = (p8[0] & MLB_MASK_V) >> 6;
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];

  uint32 objID = state->GetSubcomponentID();
  GenericObject* obj = SessionManager::sm->GetObject(objID);
  
  if (!obj) return CreateDelete(adu);
 
  return NULL;
}


/*
 * MLB::UndoManager::CreateEventInverse
 */
RefCount<RTPI::RtpiAdu> MLB::UndoManager::CreateEventInverse(RefCount<RTPI::RtpiAdu> adu) {
  if (adu->GetType() != RTPI::RtpiHeader::PKT_EVENT) return NULL;

  RTPI::RtpiEvent* event = (RTPI::RtpiEvent*)&(*adu);
  char* buf    = event->GetPayload();
  uint8*  p8   = (uint8*)buf;
  uint16* p16  = (uint16*)buf;
  int  opType  = (p8[0] & MLB_MASK_OPT) >> 2;
  int  objType = p8[1];
  int  subtype = p8[2];
    
  uint32 objID = event->GetSubcomponentID();
  uint32 pID   = event->GetObjectID();

  GenericObject* obj = SessionManager::sm->GetObject(objID);
  
  if (!obj) return NULL;
  
  RefCount<RTPI::RtpiAdu> invAdu;
  RefCount<Buffer> buffer;

  uint32 id;

  switch (objType) {
    //
    // Folder
    //
  case CommunicationManager::FOLDER: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::SET_ACTIVE: { /* nothing */ return NULL; }
        case CommunicationManager::CHG_NAME: {
          buffer = new Buffer(((GenericContainer*)obj)->GetChangeNameSize(((GenericContainer*)obj)->GetName()));
          ((GenericContainer*)obj)->FlushChangeName(buffer->getCont(), ((GenericContainer*)obj)->GetName());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_PARENT: { /* nothing */ return NULL; }
        }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        return NULL;
      }
    }  
    //
    // Page
    //  
    case CommunicationManager::PAGE: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::SET_ACTIVE: { /* nothing */ return NULL; }
        case CommunicationManager::CHG_SIZE: {
          buffer = new Buffer(((Page*)obj)->GetChangeSizeSize());
          ((Page*)obj)->FlushChangeSize(buffer->getCont(), ((Page*)obj)->GetWidth(), ((Page*)obj)->GetHeight());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_BACKGRCOLOR: {
          buffer = new Buffer(((Page*)obj)->GetChangeColorSize());
          ((Page*)obj)->FlushChangeColor(buffer->getCont(), ((Page*)obj)->GetColor());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_NAME: {
          buffer = new Buffer(((GenericContainer*)obj)->GetChangeNameSize(((GenericContainer*)obj)->GetName()));
          ((GenericContainer*)obj)->FlushChangeName(buffer->getCont(), ((GenericContainer*)obj)->GetName());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_PARENT: 
        case CommunicationManager::CHG_ACTIVE: { /* nothing */ return NULL; }
        }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        return NULL;
      }
    }
    //
    // Group
    //  
//  case CommunicationManager::GROUP: {
//    if (opType == CommunicationManager::ADU_CHANGE) {
//      switch (subtype) {
//      case CommunicationManager::MOVE: {}
//      case CommunicationManager::CHG_ACTIVE: {}
//      case CommunicationManager::RAISE: {}
//      case CommunicationManager::LOWER: {}
//    }
//    if (opType == CommunicationManager::ADU_DELETE) {}
//  }
    //
    // Rectangle
    //
    case CommunicationManager::RECTANGLE: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());  
          ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_SIZE: {
          int x, y, width, height;
          buffer = new Buffer(((DrawObject*)obj)->GetChangeSizeSize());
	        ((DrawObject*)obj)->ComputeCoords(((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY(), x, y, width, height);
	        ((DrawObject*)obj)->FlushChangeSize(buffer->getCont(), x, y, width, height);
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINEWIDTH: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinewidthSize());
	        ((ColoredObject*)obj)->FlushChangeLinewidth(buffer->getCont(), ((ColoredObject*)obj)->GetLineWidth());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINESTYLE: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinestyleSize());
	        ((ColoredObject*)obj)->FlushChangeLinestyle(buffer->getCont(), ((ColoredObject*)obj)->GetLineStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINECOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinecolorSize());
	        ((ColoredObject*)obj)->FlushChangeLinecolor(buffer->getCont(), ((ColoredObject*)obj)->GetLineColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FILLCOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeFillcolorSize());
	        ((ColoredObject*)obj)->FlushChangeFillcolor(buffer->getCont(), ((ColoredObject*)obj)->GetFillColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
      }
    }
    //
    // Oval
    //
    case CommunicationManager::OVAL: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());  
          ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_SIZE: {
          int x, y, width, height;
          buffer = new Buffer(((DrawObject*)obj)->GetChangeSizeSize());
	        ((DrawObject*)obj)->ComputeCoords(((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY(), x, y, width, height);
	        ((DrawObject*)obj)->FlushChangeSize(buffer->getCont(), x, y, width, height);
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINEWIDTH: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinewidthSize());
	        ((ColoredObject*)obj)->FlushChangeLinewidth(buffer->getCont(), ((ColoredObject*)obj)->GetLineWidth());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINESTYLE: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinestyleSize());
	        ((ColoredObject*)obj)->FlushChangeLinestyle(buffer->getCont(), ((ColoredObject*)obj)->GetLineStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINECOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinecolorSize());
	        ((ColoredObject*)obj)->FlushChangeLinecolor(buffer->getCont(), ((ColoredObject*)obj)->GetLineColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FILLCOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeFillcolorSize());
	        ((ColoredObject*)obj)->FlushChangeFillcolor(buffer->getCont(), ((ColoredObject*)obj)->GetFillColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
	      }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
      }
    }
    //
    // Line
    //
    case CommunicationManager::LINE: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());  
          ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::MOVE_POINT_1: {
          buffer = new Buffer(((Line*)obj)->GetMovePointSize());
          ((Line*)obj)->FlushMovePoint(buffer->getCont(), 1, ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::MOVE_POINT_2: {
          buffer = new Buffer(((Line*)obj)->GetMovePointSize());
          ((Line*)obj)->FlushMovePoint(buffer->getCont(), 2, ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINEWIDTH: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinewidthSize());
	        ((ColoredObject*)obj)->FlushChangeLinewidth(buffer->getCont(), ((ColoredObject*)obj)->GetLineWidth());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINESTYLE: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinestyleSize());
	        ((ColoredObject*)obj)->FlushChangeLinestyle(buffer->getCont(), ((ColoredObject*)obj)->GetLineStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINECOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinecolorSize());
	        ((ColoredObject*)obj)->FlushChangeLinecolor(buffer->getCont(), ((ColoredObject*)obj)->GetLineColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_CAPSTYLE: {
          buffer = new Buffer(((Line*)obj)->GetChangeCapstyleSize());
	        ((Line*)obj)->FlushChangeCapstyle(buffer->getCont(), ((Line*)obj)->GetCapStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_JOINSTYLE: {
          buffer = new Buffer(((Line*)obj)->GetChangeJoinstyleSize());
	        ((Line*)obj)->FlushChangeJoinstyle(buffer->getCont(), ((Line*)obj)->GetJoinStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_ARROWSTYLE: {
          buffer = new Buffer(((Line*)obj)->GetChangeArrowstyleSize());
	        ((Line*)obj)->FlushChangeArrowstyle(buffer->getCont(), ((Line*)obj)->GetArrowStyle(), ((Line*)obj)->GetArrowShape());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
      }
    }
    //
    // Polyline
    //
    case CommunicationManager::POLYLINE: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());  
          ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::MOVE_POINT: {
          buffer = new Buffer(((PolyObject*)obj)->GetMovePointSize());
	        ((PolyObject*)obj)->FlushMovePoint(buffer->getCont(), ntohs(p16[2]), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINEWIDTH: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinewidthSize());
	        ((ColoredObject*)obj)->FlushChangeLinewidth(buffer->getCont(), ((ColoredObject*)obj)->GetLineWidth());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINESTYLE: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinestyleSize());
	        ((ColoredObject*)obj)->FlushChangeLinestyle(buffer->getCont(), ((ColoredObject*)obj)->GetLineStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINECOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinecolorSize());
	        ((ColoredObject*)obj)->FlushChangeLinecolor(buffer->getCont(), ((ColoredObject*)obj)->GetLineColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_CAPSTYLE: {
          buffer = new Buffer(((PolyLine*)obj)->GetChangeCapstyleSize());
	        ((PolyLine*)obj)->FlushChangeCapstyle(buffer->getCont(), ((PolyLine*)obj)->GetCapStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_JOINSTYLE: {
          buffer = new Buffer(((PolyLine*)obj)->GetChangeJoinstyleSize());
	        ((PolyLine*)obj)->FlushChangeJoinstyle(buffer->getCont(), ((PolyLine*)obj)->GetJoinStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_ARROWSTYLE: {
          buffer = new Buffer(((PolyLine*)obj)->GetChangeArrowstyleSize());
	        ((PolyLine*)obj)->FlushChangeArrowstyle(buffer->getCont(), ((PolyLine*)obj)->GetArrowStyle(), ((PolyLine*)obj)->GetArrowShape());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::ADD_POINT: { /* nothing */ return NULL; }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
	      } 
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
      }
    }
    //
    // Polygon
    //
    case CommunicationManager::POLYGON: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());  
          ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::MOVE_POINT: {
          buffer = new Buffer(((PolyObject*)obj)->GetMovePointSize());
	        ((PolyObject*)obj)->FlushMovePoint(buffer->getCont(), ntohs(p16[2]), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINEWIDTH: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinewidthSize());
	        ((ColoredObject*)obj)->FlushChangeLinewidth(buffer->getCont(), ((ColoredObject*)obj)->GetLineWidth());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINESTYLE: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinestyleSize());
	        ((ColoredObject*)obj)->FlushChangeLinestyle(buffer->getCont(), ((ColoredObject*)obj)->GetLineStyle());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_LINECOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeLinecolorSize());
	        ((ColoredObject*)obj)->FlushChangeLinecolor(buffer->getCont(), ((ColoredObject*)obj)->GetLineColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FILLCOLOR: {
          buffer = new Buffer(((ColoredObject*)obj)->GetChangeFillcolorSize());
	        ((ColoredObject*)obj)->FlushChangeFillcolor(buffer->getCont(), ((ColoredObject*)obj)->GetFillColor());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::ADD_POINT: { /* nothing */ return NULL; }
        case CommunicationManager::CLOSE_POLYGON: { /* nothing */ return NULL; }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
	      }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
        }
        return invAdu;
      }
    }
    //
    // Text
    //
    case CommunicationManager::TEXT: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((Group*)obj)->GetMoveSize());
	        ((Group*)obj)->FlushMoveTG(buffer->getCont(), ((Group*)obj)->GetX(), ((Group*)obj)->GetY());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_SIZE: { /* nothing */ return NULL; }
        case CommunicationManager::INSERT_FONT: {
          Text* txt = (Text*)((Group*)obj)->GetActiveItem();
          if (!txt) return NULL;
          buffer = new Buffer(((Group*)obj)->GetChangeFontSizeTG(txt->GetFontFamily()));
          ((Group*)obj)->FlushInsertNewFontTG(buffer->getCont(), txt->GetColor(), txt->GetFontFamily(), txt->GetFontSize(), txt->GetFontWeight(), txt->GetFontSlant(), txt->GetTextPosition());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FONT: {
          Text* txt = (Text*)((Group*)obj)->GetFirst();
          if (!txt) return NULL;
          buffer = new Buffer(((Group*)obj)->GetChangeFontSizeTG(txt->GetFontFamily()));
          ((Group*)obj)->FlushChangeFontTG(buffer->getCont(), txt->GetColor(), txt->GetFontFamily(), txt->GetFontSize(), txt->GetFontWeight(), txt->GetFontSlant());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FONTFAMILY: { /* nothing */ return NULL; }
        case CommunicationManager::CHG_FONTSIZE: { /* nothing */ return NULL; }
        case CommunicationManager::CHG_LINECOLOR: {
          Text* txt = (Text*)((Group*)obj)->GetFirst();
          if (!txt) return NULL;
          buffer = new Buffer(((Group*)obj)->GetChangeFontColorSizeTG());
          ((Group*)obj)->FlushChangeFontColorTG(buffer->getCont(), txt->GetColor());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::INSERT_FCOLOR: {
          Text* txt = (Text*)((Group*)obj)->GetActiveItem();
          if (!txt) return NULL;
          buffer = new Buffer(((Group*)obj)->GetChangeFontColorSizeTG());
          ((Group*)obj)->FlushInsertNewFontColorTG(buffer->getCont(), txt->GetColor());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_FONTWEIGHT: 
        case CommunicationManager::CHG_FONTSLANT: 
        case CommunicationManager::CHG_FONTPOS: 
        case CommunicationManager::INSERT_CHARS: 
        case CommunicationManager::DELETE_CHARS: { /* nothing */ return NULL; }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((Group*)obj)->GetRaiseSize());  
          ((Group*)obj)->FlushRaiseTG(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((Group*)obj)->GetLowerSize());  
          ((Group*)obj)->FlushLowerTG(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_ACTIVE: { /* nothing */ return NULL; }
  	    }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
			}
    }  
    //
    // Image
    //  
    case CommunicationManager::IMAGE: {
      if (opType == CommunicationManager::ADU_CHANGE) {
        switch (subtype) {
        case CommunicationManager::CHG_VISIBILITY: { /* nothing */ return NULL; }
        case CommunicationManager::MOVE: {
          buffer = new Buffer(((DrawObject*)obj)->GetMoveSize());
	        ((DrawObject*)obj)->FlushMove(buffer->getCont(), ((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetY());
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::CHG_SIZE: {
          int x, y, width, height;
          buffer = new Buffer(((DrawObject*)obj)->GetChangeSizeSize());
	        ((DrawObject*)obj)->ComputeCoords(((DrawObject*)obj)->GetX(), ((DrawObject*)obj)->GetX(), x, y, width, height);
	        ((DrawObject*)obj)->FlushChangeSize(buffer->getCont(), x, y, width, height);
	        return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::RAISE: {
          buffer = new Buffer(((DrawObject*)obj)->GetRaiseSize());  
          ((DrawObject*)obj)->FlushRaise(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
        case CommunicationManager::LOWER: {
          buffer = new Buffer(((DrawObject*)obj)->GetLowerSize());  
          ((DrawObject*)obj)->FlushLower(buffer->getCont(), obj->GetOrder());
          return CommunicationManager::cm->CreateEvent(obj->GetID(), pID, buffer);
        }
	      }
      }
      if (opType == CommunicationManager::ADU_DELETE) {
        invAdu = obj->GetStateAdu(0, 3); 
        if (invAdu != NULL) { 
          id = CommunicationManager::cm->GetObjectID();
          ChangeAduID(invAdu->GetPacketHeader()->GetSubcomponentID(), id);
          invAdu->GetPacketHeader()->SetSubcomponentID(id); 
          if (((DrawObject*)obj)->Annotation()) SetAnnotation(id);
        }
        return invAdu;
      }
    }
  }
  return NULL;
}


/* 
 * MLB::UndoManager::CreateInverse
 */
RefCount<RTPI::RtpiAdu> MLB::UndoManager::CreateInverse(RefCount<RTPI::RtpiAdu> adu) {
  switch (adu->GetType()) {
    case RTPI::RtpiHeader::PKT_EVENT: return CreateEventInverse(adu);
    case RTPI::RtpiHeader::PKT_STATE: return CreateStateInverse(adu);
  }
  return NULL;
}


/*
 * MLB::UndoManager::Undo()
 */
void MLB::UndoManager::Undo() { 
  if (undoList->empty()) return;
  
  RefCount<RTPI::RtpiAduList> redoAduList = new RTPI::RtpiAduList();
  GenericObject* obj;
  Session* session;
  Page* page;
  uint objID;
  bool annotation;
  int runningTime = 0;

  // pop aduList
  RefCount<RTPI::RtpiAduList> undoAduList = undoList->back();
  undoList->pop_back();
  
  while (!undoAduList->empty()) {
    RefCount<RTPI::RtpiAdu> undoAdu = undoAduList->back();
    undoAduList->pop_back();
  
    // generate redo-adu
    RefCount<RTPI::RtpiAdu> redoAdu = CreateInverse(undoAdu);
    if (redoAdu == NULL) /* error */ break;
  
    redoAduList->push_back(redoAdu);

    // Annotation ?
    objID = undoAdu->GetSubcomponentID();
    obj   = SessionManager::sm->GetObject(objID);
    if (obj && obj->GetType() != GenericObject::FOLDER && obj->GetType() != GenericObject::PAGE) {
      annotation = ((DrawObject*)obj)->Annotation();
      session    = SessionManager::sm->GetParentSession(obj);
    } else {
      annotation = false;
      session    = NULL;
    }

    // current page ?
    if (session && (SessionManager::sm->GetActivePage() != (page = (Page*)session->GetContainer(objID)))) {
      if (page) SessionManager::sm->SelectContainer(page->GetNodePath());
    }

    // send adu
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
      if (runningTime == 0) {
        runningTime = AduManager::am->GetLocalLag() + 50;
        SetUndoState(false);
      }

      undoAdu->GetAduHeader()->SetTimestamp(rtpitime() + AduManager::am->GetLocalLag());
      AduManager::am->InsertAdu(undoAdu);
      CommunicationManager::cm->SendAduReliable(undoAdu);

    } else {
      undoAdu->GetAduHeader()->SetTimestamp(rtpitime());

      switch (undoAdu->GetType()) {
        case RTPI::RtpiHeader::PKT_EVENT: AduManager::am->HandleEvent(undoAdu); break;
        case RTPI::RtpiHeader::PKT_STATE: AduManager::am->HandleState(undoAdu); break;
      }
    }
  }
  redoList->push_back(redoAduList);
  SetRedoState(true);

  if (runningTime == 0) {
    if (undoList->empty()) SetUndoState(false);
  } else {
    TimerManager::tm->SetTimer(this, new TimerObject(runningTime, TimerObject::UNDO, 0, 0));
  }
}

    
/*
 * MLB::UndoManager::Redo()
 */
void MLB::UndoManager::Redo() {
  if (redoList->empty()) return;
  
  RefCount<RTPI::RtpiAduList> undoAduList = new RTPI::RtpiAduList();
  GenericObject* obj;
  Session* session;
  Page* page;
  uint32 objID;
  bool annotation;
  int runningTime = 0;

  // pop aduList
  RefCount<RTPI::RtpiAduList> redoAduList = redoList->back();
  redoList->pop_back();
  
  while (!redoAduList->empty()) {
    RefCount<RTPI::RtpiAdu> redoAdu = redoAduList->back();
    redoAduList->pop_back();

    // generate undo-adu
    RefCount<RTPI::RtpiAdu> undoAdu = CreateInverse(redoAdu);
    if (undoAdu == NULL) /* error */ break;
    
    undoAduList->push_back(undoAdu);

    // Annotation ?
    objID = redoAdu->GetSubcomponentID();
    obj   = SessionManager::sm->GetObject(objID);
    if (obj && obj->GetType() != GenericObject::FOLDER && obj->GetType() != GenericObject::PAGE) {
      annotation = ((DrawObject*)obj)->Annotation();
      session    = SessionManager::sm->GetParentSession(obj);
    } else {      
      if (!obj) annotation = (annotationSet->find(objID) != annotationSet->end());
      session    = NULL;
    }

    // current page ?    
    if (session && (SessionManager::sm->GetActivePage() != (page = (Page*)session->GetContainer(objID)))) {
      if (page) SessionManager::sm->SelectContainer(page->GetNodePath());
    }

    // send adu
    if (SessionManager::sm->GetActiveSession()->GetSessionType() == Session::SHARED && !annotation) {
      if (runningTime == 0) {
        runningTime = AduManager::am->GetLocalLag() + 50;
        SetRedoState(false);
      }

      redoAdu->GetAduHeader()->SetTimestamp(rtpitime() + AduManager::am->GetLocalLag());
      AduManager::am->InsertAdu(redoAdu);
      CommunicationManager::cm->SendAduReliable(redoAdu);

    } else {
      redoAdu->GetAduHeader()->SetTimestamp(rtpitime());

      switch (redoAdu->GetType()) {
        case RTPI::RtpiHeader::PKT_EVENT: AduManager::am->HandleEvent(redoAdu); break;
        case RTPI::RtpiHeader::PKT_STATE: AduManager::am->HandleState(redoAdu); break;
      }
    }
  }
  undoList->push_back(undoAduList);
  SetUndoState(true);

  if (runningTime == 0) {
    if (redoList->empty()) SetRedoState(false);
  } else {
    TimerManager::tm->SetTimer(this, new TimerObject(runningTime, TimerObject::UNDO, 1, 0));
  }
}


/*
 * MLB::UndoManager::ExcecuteTimer()
 */
void MLB::UndoManager::ExecuteTimer(TimerObject* timer) {
  if (timer->id == 0) {
    if (!undoList->empty()) SetUndoState(true);
  } else {
    if (!redoList->empty()) SetRedoState(true);
  }
}
