/***********************************************************************************************************
 * mlb - multimedia lecture board 
 * Copyright 1999 - 2003
 * 
 * University of Mannheim, Germany
 * University of Freiburg, Germany
 *
 * Development:
 * Leader: Juergen Vogel
 * Team Mannheim: Juergen Vogel, Marcel Busse, Andrius Kurtinaitis, and Werner Geyer.
 * Team Freiburg: Gabriela Maass, Jochen Lienhard, and Tobias Lauer.
 *
 * Development was sponsored by DFN (ANETTE project).
 *
 * License Information:
 * The mlb uses the following libraries: Tcl/Tk, BWidgets, ImageMagick, FreeType, RTP/I and SMP
 * 
 * This program is free software; you can redistribute it and/or modify it under the terms of 
 * the GNU General Public License as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; 
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU General Public License for more details.
 * The GNU General Public License can be obtained under http://www.gnu.org/licenses/gpl.html
 *
 * The Universities of Mannheim and Freiburg will appreciate receiving any enhancements and 
 * modifications of this software made by others.
 *
 * Contact:
 * Juergen Vogel - vogel@informatik.uni-mannheim.de
 * www.informatik.uni-mannheim.de/informatik/pi4/projects/mlb
 *
 **********************************************************************************************************

 program : vt_manager.h 
 authors : Juergen Vogel
 date    : 25.08.2001
 desc    : 

************************************************************************************************************/


#ifndef __VT_MANAGER_H__
#define __VT_MANAGER_H__

#define VT_MASK_V    0xC0
#define VT_MASK_TYPE 0x3F

#include <rtpi/rtpiPacket.h>
#include <rtpi/rtcpiPacket.h>
#include <rtpi/rtpihandler.h>
#include <rtpi/rtpiparticipant.h>
#include <rtpi/rtpiInterfaces.h>
#include <rtpi/rtpiTransport.h>
#include <rtpi/rtpiSeqNumber.h>
#include <util/types.h>
#include <util/Buffer.h>

#include <mlb_communication.h>

#include <tcl.h>

#include <vector>
#include <list>
#include <string>

namespace VT {
  class Answer {
  public:
    Answer() : pid(0), createTime(0) {}
    Answer(uint32 _pid, uint32 _createTime) : pid(_pid), createTime(_createTime) {}
    virtual ~Answer() {}

    virtual int FlushCreate(char* buf) = 0;
    virtual int ParseCreate(char* buf) = 0;
    virtual int FlushDelete(char* buf) = 0;

    uint32 pid;        // originating participant
    uint32 createTime; // time the answer was created (in secs)
  };

  
  class TextAnswer : public Answer {
  public:
    TextAnswer() : Answer() {};
    TextAnswer(uint32 _pid, uint32 _createTime) : Answer(_pid, _createTime) {}
    TextAnswer(uint32 _pid, uint32 _createTime, std::string _text) : Answer(_pid, _createTime), text(_text) {}
    ~TextAnswer() {}

    int FlushCreate(char* buf);
    int ParseCreate(char* buf);
    int FlushDelete(char* buf);

    std::string text;  // given text answer
  };

  
  class MultipleChoiceAnswer : public Answer {
  public:
    MultipleChoiceAnswer() : Answer() {}
    MultipleChoiceAnswer(uint32 _pid, uint32 _createTime) : Answer(_pid, _createTime) {}
    MultipleChoiceAnswer(uint32 _pid, uint32 _createTime, int _value) : Answer(_pid, _createTime), value(_value) {}
    ~MultipleChoiceAnswer() {}

    int FlushCreate(char* buf);
    int ParseCreate(char* buf);
    int FlushDelete(char* buf);

    int    value;      // selected value
  };


  class Question {
  public:
    Question(uint32 _id) : id (_id) {}
    Question(uint32 _id, uint32 _ttl, std::string _text) : id(_id), ttl(_ttl), text(_text) {}
    virtual ~Question();

    void DeleteAllAnswers();

    virtual void DisplayAdd() = 0;
    virtual void DisplayUpdate() = 0;
    virtual void DisplayResult() = 0;

    virtual std::string CalculateResult() = 0;

    virtual int FlushState(char* buf) = 0;
    virtual int ParseState(char* buf, uint32 ts) = 0;

    virtual int FlushDelete(char* buf) = 0;

    uint32 id;                              // subid of question
    uint32 ttl;                             // time to live (received answers)
    std::string text;                       // question text
    std::list<Answer*> answers;             // received answers
  };


  class TextQuestion : public Question {
  public:
    TextQuestion(uint32 _id) : Question(_id) {}
    TextQuestion(uint32 _id, uint32 _ttl, std::string text) : Question(_id, _ttl, text) {}
    virtual ~TextQuestion() {}

    void DisplayAdd();
    void DisplayUpdate();
    void DisplayResult();

    std::string CalculateResult();

    int FlushState(char* buf);
    int ParseState(char* buf, uint32 ts);

    int FlushDelete(char* buf);
  };

  
  class MultipleChoiceQuestion : public Question {
  public:
    MultipleChoiceQuestion(uint32 _id) : Question(_id) {}
    MultipleChoiceQuestion(uint32 _id, uint32 _ttl, std::string text, int nbr, std::string _expl);
    ~MultipleChoiceQuestion() {}

    void DisplayAdd();
    void DisplayUpdate();
    void DisplayResult();

    void Update(std::string text, int nbr, std::string _expl);
    std::string CalculateResult();

    int FlushState(char* buf);
    int ParseState(char* buf, uint32 ts);

    int FlushDelete(char* buf);

    std::vector<std::string> explanations;  // meaning of values
  };
  

  class VtManager : public RTPI::RtcpiReportInterface, public RTPI::RtpiFeedbackInterface, public MLB::TimerFeedbackInterface {
  public:
    enum types {
      // general
      VERSION = 0,
      APPLICATION_TYPE = 5, // ft
      
      // question types
      MC_QUESTION = 0,
      TX_QUESTION = 1,

      // operation Types
      DELETE_MC_QUESTION = 0,
      DELETE_TX_QUESTION = 1,
      CREATE_MC_ANSWER   = 2,
      DELETE_MC_ANSWER   = 3,
      CREATE_TX_ANSWER   = 4,
      DELETE_TX_ANSWER   = 5
    };
    
    VtManager();
    ~VtManager();
    
    void Init();

    bool OpenConnection() {return openConnection;}

    bool OpenRTPI();
    bool CloseRTPI();

    // handle incoming data
    void HandleReceivedRtpi(SMP::Indication ind);
    void HandleReceivedRtcpi(SMP::Indication ind);

    // functions called by gui
    void CreateMCQuestion(int argc, const char** argv);
    void CreateTextQuestion(int argc, const char** argv);
    void EditMCQuestion(int argc, const char** argv);
    void EditTextQuestion(int argc, const char** argv);
    void DeleteQuestion(uint32 id);
    void ChangeMCAnswer(uint32 id, int value);
    void ChangeTextAnswer(uint32 id, std::string text);

    // outgoing data
    void SendState(uint32 id, int priority, char* payload, int size);
    void SendEvent(uint32 id, char* payload, int size);
    void SendStateQuery(uint32 id, int priority);

    // functions implemented due to  RtcpiReportInterface
    void    send_rtcpi(RefCount<Buffer>);
    u_int32 get_local_address(int i=0)   {return MLB::CommunicationManager::cm->get_local_address(i);}
    int     addr_is_local(u_int32 addr)  {return MLB::CommunicationManager::cm->addr_is_local(addr);}
    u_int32 get_session_bandwith()       {return MLB::CommunicationManager::cm->bandwidth/10;}
    u_int32 get_new_subid()              {return MLB::CommunicationManager::cm->get_new_subid();} 
    u_int32 get_new_pid()                {return MLB::CommunicationManager::cm->pid;}
    string  get_session_name(string ="") {return MLB::CommunicationManager::cm->name;}
    string  get_session_email()          {return MLB::CommunicationManager::cm->email;}
    string  get_session_phone()          {return MLB::CommunicationManager::cm->phone;}
    string  get_session_loc()            {return MLB::CommunicationManager::cm->location;}
    string  get_session_note()           {return ""; }
    string  get_session_tool()           {return "voting"; }

    // functions implemented due to RtpiFeedbackInterface
    void subcomponent_added(RTPI::SubComponentMapElement* subcomp, int reason);
    void subcomponent_changed(RTPI::SubComponentMapElement* subcomp, uint8 reason);
    void subcomponent_removed(uint32 id);
    void participant_added(uint32 pid);
    void participant_removed(uint32 pid);
    void participant_changed(uint32 pid);

    static VtManager* vtm;

    static const unsigned int T;
    static char* sendBuffer;

  private:
    // helper functions
    void RegisterSubcomponent(uint32 subID);
    void UnregisterSubcomponent(uint32 subID);
    void ChangeSubcomponentActiveness(uint32 subID, bool active);

    // list of questions
    Question* GetQuestion(uint32 id);
    std::list<Question*> questions;
    
    // UDP 
    bool openConnection;

    // SMP 
    int rtpiSocket;
    int rtcpiSocket;

    // RTP/I 
    bool init; // while init, don't answer state queries
    RTPI::Rtpi* rtpi;

    // timer management
    void ExecuteTimer(MLB::TimerObject* obj);
  };
}


#endif
